// AUTOGENERATED FILE - DO NOT MODIFY!
// This file was generated by Djinni from tts_backend.djinni

#include "java/jni/Unit.hpp"  // my header

namespace prism::jni {

Unit::Unit() = default;

Unit::~Unit() = default;

auto Unit::fromCpp(JNIEnv* jniEnv, const CppType& c) -> ::djinni::LocalRef<JniType> {
    (void)c; // Suppress warnings in release builds for empty records
    const auto& data = ::djinni::JniClass<Unit>::get();
    auto r = ::djinni::LocalRef<JniType>{jniEnv->NewObject(data.clazz.get(), data.jconstructor)};
    ::djinni::jniExceptionCheck(jniEnv);
    return r;
}

auto Unit::toCpp(JNIEnv* jniEnv, JniType j) -> CppType {
    ::djinni::JniLocalScope jscope(jniEnv, 1);
    assert(j != nullptr);
    (void)j; // Suppress warnings in release builds for empty records
    return {};
}

} // namespace prism::jni
