LIBRARY "PCTKUSR.dll"
EXPORTS
    PCTKStatus
    PCTKGetVersion
    PCTKPRead
    PCTKPReadEx
    PCTKPReadW
    PCTKPReadExW
    PCTKVReset
    PCTKGetVStatus
    PCTKVoiceGuide
    PCTKBeep
    PCTKCGuide
    PCTKCGuideEx
    PCTKCGuideW
    PCTKCGuideExW
    PCTKGetStatus
    PCTKSetStatus
    PCTKCommand
    PCTKLoadUserDict
    PCTKPinStatus
    PCTKPinFocus
    PCTKPinFocusW
    PCTKPinIsFocus
    PCTKPinWrite
    PCTKPinWriteW
    PCTKPinEDWrite
    PCTKPinEDWriteW
    PCTKPinStatusCell
    PCTKPinStatusCellW
    PCTKPinReset
    SoundMessage
    SoundStatus
    SoundModifyMode
    SoundPause
    AGSEvent
    GetUIActionMode
    IsImmInput
    PCTKEventHook
    PCTKGetVoiceLog
    SetKeyBreak
    EncodeFlags
    dic_regist
    dic_regist_detail
    dic_reg_from_file
    dic_text_out
    dic_reg_detail_from_file
    dic_detail_text_out
    PCTKSTATUS
    PCTKGETVERSION
    PCTKGETVSTATUS
    PCTKGETSTATUS
    PCTKSETSTATUS
    PCTKPREAD
    PCTKPREADEX
    PCTKPREADW
    PCTKPREADEXW
    PCTKCGUIDE
    PCTKCGUIDEEX
    PCTKCGUIDEW
    PCTKCGUIDEEXW
    PCTKCOMMAND
    PCTKVOICEGUIDE
    PCTKVRESET
    PCTKBEEP
    PCTKLOADUSERDICT
    PCTKEVENTHOOK
    PCTKGETVOICELOG
