LIBRARY "PCTKUSR.dll"
EXPORTS
    PCTKStatus@0
    PCTKGetVersion@0
    PCTKPRead@12
    PCTKPReadEx@16
    PCTKPReadW@12
    PCTKPReadExW@16
    PCTKVReset@0
    PCTKGetVStatus@0
    PCTKVoiceGuide@4
    PCTKBeep@8
    PCTKCGuide@8
    PCTKCGuideEx@12
    PCTKCGuideW@8
    PCTKCGuideExW@12
    PCTKGetStatus@12
    PCTKSetStatus@12
    PCTKCommand@12
    PCTKLoadUserDict@0
    PCTKPinStatus@0
    PCTKPinFocus@20
    PCTKPinFocusW@20
    PCTKPinIsFocus@4
    PCTKPinWrite@12
    PCTKPinWriteW@12
    PCTKPinEDWrite@28
    PCTKPinEDWriteW@28
    PCTKPinStatusCell@8
    PCTKPinStatusCellW@8
    PCTKPinReset@0
    SoundMessage@8
    SoundStatus@0
    SoundModifyMode@8
    SoundPause@4
    AGSEvent@12
    GetUIActionMode@0
    IsImmInput@4
    PCTKEventHook@0
    PCTKGetVoiceLog@0
    SetKeyBreak@0
    EncodeFlags@0
    dic_regist@0
    dic_regist_detail@0
    dic_reg_from_file@0
    dic_text_out@0
    dic_reg_detail_from_file@0
    dic_detail_text_out@0
    PCTKSTATUS@0
    PCTKGETVERSION@0
    PCTKGETVSTATUS@0
    PCTKGETSTATUS@12
    PCTKSETSTATUS@12
    PCTKPREAD@12
    PCTKPREADEX@16
    PCTKPREADW@12
    PCTKPREADEXW@16
    PCTKCGUIDE@8
    PCTKCGUIDEEX@12
    PCTKCGUIDEW@8
    PCTKCGUIDEEXW@12
    PCTKCOMMAND@12
    PCTKVOICEGUIDE@4
    PCTKVRESET@0
    PCTKBEEP@8
    PCTKLOADUSERDICT@0
    PCTKEVENTHOOK@0
    PCTKGETVOICELOG@0
