// SPDX-License-Identifier: MPL-2.0
package com.github.ethindp.prism;

public enum BackendError {
  OK,
  NOT_INITIALIZED,
  INVALID_PARAM,
  NOT_IMPLEMENTED,
  NO_VOICES,
  VOICE_NOT_FOUND,
  SPEAK_FAILURE,
  MEMORY_FAILURE,
  RANGE_OUT_OF_BOUNDS,
  INTERNAL_BACKEND_ERROR,
  NOT_SPEAKING,
  NOT_PAUSED,
  ALREADY_PAUSED,
  INVALID_UTF8,
  INVALID_OPERATION,
  ALREADY_INITIALIZED,
  BACKEND_NOT_AVAILABLE,
  UNKNOWN,
  ;
}
