// AUTOGENERATED FILE - DO NOT MODIFY!
// This file was generated by Djinni from tts_backend.djinni

#pragma once

#include <utility>

namespace prism::java {

struct Unit final {
};

} // namespace prism::java
