// AUTOGENERATED FILE - DO NOT MODIFY!
// This file was generated by Djinni from tts_backend.djinni

#pragma once

#include "java/Unit.hpp"
#include "java/support/jni/djinni_support.hpp"

namespace prism::jni {

class Unit final {
public:
    using CppType = ::prism::java::Unit;
    using JniType = jobject;

    using Boxed = Unit;

    ~Unit();

    static CppType toCpp(JNIEnv* jniEnv, JniType j);
    static ::djinni::LocalRef<JniType> fromCpp(JNIEnv* jniEnv, const CppType& c);

private:
    Unit();
    friend ::djinni::JniClass<Unit>;

    const ::djinni::GlobalRef<jclass> clazz { ::djinni::jniFindClass("com/github/ethindp/prism/Unit") };
    const jmethodID jconstructor { ::djinni::jniGetMethodID(clazz.get(), "<init>", "()V") };
};

} // namespace prism::jni
