// AUTOGENERATED FILE - DO NOT MODIFY!
// This file was generated by Djinni from tts_backend.djinni

#pragma once

#include "java/BackendError.hpp"
#include "java/support/jni/djinni_support.hpp"

namespace prism::jni {

class BackendError final : ::djinni::JniEnum {
public:
    using CppType = ::prism::java::BackendError;
    using JniType = jobject;

    using Boxed = BackendError;

    static CppType toCpp(JNIEnv* jniEnv, JniType j) { return static_cast<CppType>(::djinni::JniClass<BackendError>::get().ordinal(jniEnv, j)); }
    static ::djinni::LocalRef<JniType> fromCpp(JNIEnv* jniEnv, CppType c) { return ::djinni::JniClass<BackendError>::get().create(jniEnv, static_cast<jint>(c)); }

private:
    BackendError() : JniEnum("com/github/ethindp/prism/BackendError") {}
    friend ::djinni::JniClass<BackendError>;
};

} // namespace prism::jni
