LIBRARY "BoyCtrl-x64.dll"
EXPORTS
    BoyCtrlActivateYTApp
    BoyCtrlCancelTextToAudio
    BoyCtrlGetReaderState
    BoyCtrlInitialize
    BoyCtrlInitializeAnsi
    BoyCtrlInitializeU8
    BoyCtrlIsReaderRunning
    BoyCtrlPauseScreenReader
    BoyCtrlReportInfo
    BoyCtrlSetAnyKeyStopSpeaking
    BoyCtrlSpeak
    BoyCtrlSpeak2
    BoyCtrlSpeakAnsi
    BoyCtrlSpeakEx
    BoyCtrlSpeakU8
    BoyCtrlStartTextToAudio
    BoyCtrlStopSpeaking
    BoyCtrlStopSpeaking2
    BoyCtrlStopSpeakingEx
    BoyCtrlUninitialize
    BoyCtrlVerify
