

/* this ALWAYS GENERATED file contains the definitions for the interfaces */


 /* File created by MIDL compiler version 8.00.0595 */
/* at Fri Sep 06 15:13:18 2013
 */
/* Compiler settings for ..\..\wineyes.IDL:
    Oicf, W1, Zp8, env=Win32 (32b run), target_arch=X86 8.00.0595 
    protocol : dce , ms_ext, c_ext, robust
    error checks: allocation ref bounds_check enum stub_data 
    VC __declspec() decoration level: 
         __declspec(uuid()), __declspec(selectany), __declspec(novtable)
         DECLSPEC_UUID(), MIDL_INTERFACE()
*/
/* @@MIDL_FILE_HEADING(  ) */

#pragma warning( disable: 4049 )  /* more than 64k source lines */


/* verify that the <rpcndr.h> version is high enough to compile this file*/
#ifndef __REQUIRED_RPCNDR_H_VERSION__
#define __REQUIRED_RPCNDR_H_VERSION__ 475
#endif

#include "rpc.h"
#include "rpcndr.h"

#ifndef __RPCNDR_H_VERSION__
#error this stub requires an updated version of <rpcndr.h>
#endif // __RPCNDR_H_VERSION__


#ifndef __wineyes_h_h__
#define __wineyes_h_h__

#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

/* Forward Declarations */ 

#ifndef ___Application_FWD_DEFINED__
#define ___Application_FWD_DEFINED__
typedef interface _Application _Application;

#endif 	/* ___Application_FWD_DEFINED__ */


#ifndef ___Window_FWD_DEFINED__
#define ___Window_FWD_DEFINED__
typedef interface _Window _Window;

#endif 	/* ___Window_FWD_DEFINED__ */


#ifndef __WindowRectangle_FWD_DEFINED__
#define __WindowRectangle_FWD_DEFINED__
typedef interface WindowRectangle WindowRectangle;

#endif 	/* __WindowRectangle_FWD_DEFINED__ */


#ifndef __ScreenRectangle_FWD_DEFINED__
#define __ScreenRectangle_FWD_DEFINED__
typedef interface ScreenRectangle ScreenRectangle;

#endif 	/* __ScreenRectangle_FWD_DEFINED__ */


#ifndef __ClientRectangle_FWD_DEFINED__
#define __ClientRectangle_FWD_DEFINED__
typedef interface ClientRectangle ClientRectangle;

#endif 	/* __ClientRectangle_FWD_DEFINED__ */


#ifndef __ClientPoint_FWD_DEFINED__
#define __ClientPoint_FWD_DEFINED__
typedef interface ClientPoint ClientPoint;

#endif 	/* __ClientPoint_FWD_DEFINED__ */


#ifndef __ScreenPoint_FWD_DEFINED__
#define __ScreenPoint_FWD_DEFINED__
typedef interface ScreenPoint ScreenPoint;

#endif 	/* __ScreenPoint_FWD_DEFINED__ */


#ifndef __Accessible_FWD_DEFINED__
#define __Accessible_FWD_DEFINED__
typedef interface Accessible Accessible;

#endif 	/* __Accessible_FWD_DEFINED__ */


#ifndef __Accessibles_FWD_DEFINED__
#define __Accessibles_FWD_DEFINED__
typedef interface Accessibles Accessibles;

#endif 	/* __Accessibles_FWD_DEFINED__ */


#ifndef __AccessibleRole_FWD_DEFINED__
#define __AccessibleRole_FWD_DEFINED__
typedef interface AccessibleRole AccessibleRole;

#endif 	/* __AccessibleRole_FWD_DEFINED__ */


#ifndef __AccessibleState_FWD_DEFINED__
#define __AccessibleState_FWD_DEFINED__
typedef interface AccessibleState AccessibleState;

#endif 	/* __AccessibleState_FWD_DEFINED__ */


#ifndef __WindowPoint_FWD_DEFINED__
#define __WindowPoint_FWD_DEFINED__
typedef interface WindowPoint WindowPoint;

#endif 	/* __WindowPoint_FWD_DEFINED__ */


#ifndef __Windows_FWD_DEFINED__
#define __Windows_FWD_DEFINED__
typedef interface Windows Windows;

#endif 	/* __Windows_FWD_DEFINED__ */


#ifndef __WindowStyle_FWD_DEFINED__
#define __WindowStyle_FWD_DEFINED__
typedef interface WindowStyle WindowStyle;

#endif 	/* __WindowStyle_FWD_DEFINED__ */


#ifndef __Desktop_FWD_DEFINED__
#define __Desktop_FWD_DEFINED__
typedef interface Desktop Desktop;

#endif 	/* __Desktop_FWD_DEFINED__ */


#ifndef __Process_FWD_DEFINED__
#define __Process_FWD_DEFINED__
typedef interface Process Process;

#endif 	/* __Process_FWD_DEFINED__ */


#ifndef __Version_FWD_DEFINED__
#define __Version_FWD_DEFINED__
typedef interface Version Version;

#endif 	/* __Version_FWD_DEFINED__ */


#ifndef __Clips_FWD_DEFINED__
#define __Clips_FWD_DEFINED__
typedef interface Clips Clips;

#endif 	/* __Clips_FWD_DEFINED__ */


#ifndef __Clip_FWD_DEFINED__
#define __Clip_FWD_DEFINED__
typedef interface Clip Clip;

#endif 	/* __Clip_FWD_DEFINED__ */


#ifndef __Color_FWD_DEFINED__
#define __Color_FWD_DEFINED__
typedef interface Color Color;

#endif 	/* __Color_FWD_DEFINED__ */


#ifndef __CharacterOffsets_FWD_DEFINED__
#define __CharacterOffsets_FWD_DEFINED__
typedef interface CharacterOffsets CharacterOffsets;

#endif 	/* __CharacterOffsets_FWD_DEFINED__ */


#ifndef __SetFile_FWD_DEFINED__
#define __SetFile_FWD_DEFINED__
typedef interface SetFile SetFile;

#endif 	/* __SetFile_FWD_DEFINED__ */


#ifndef __ScreenSettings_FWD_DEFINED__
#define __ScreenSettings_FWD_DEFINED__
typedef interface ScreenSettings ScreenSettings;

#endif 	/* __ScreenSettings_FWD_DEFINED__ */


#ifndef __PunctuationSettings_FWD_DEFINED__
#define __PunctuationSettings_FWD_DEFINED__
typedef interface PunctuationSettings PunctuationSettings;

#endif 	/* __PunctuationSettings_FWD_DEFINED__ */


#ifndef __LineFilterSettings_FWD_DEFINED__
#define __LineFilterSettings_FWD_DEFINED__
typedef interface LineFilterSettings LineFilterSettings;

#endif 	/* __LineFilterSettings_FWD_DEFINED__ */


#ifndef __KeyboardSettings_FWD_DEFINED__
#define __KeyboardSettings_FWD_DEFINED__
typedef interface KeyboardSettings KeyboardSettings;

#endif 	/* __KeyboardSettings_FWD_DEFINED__ */


#ifndef __SpecialKeySettings_FWD_DEFINED__
#define __SpecialKeySettings_FWD_DEFINED__
typedef interface SpecialKeySettings SpecialKeySettings;

#endif 	/* __SpecialKeySettings_FWD_DEFINED__ */


#ifndef __MouseSettings_FWD_DEFINED__
#define __MouseSettings_FWD_DEFINED__
typedef interface MouseSettings MouseSettings;

#endif 	/* __MouseSettings_FWD_DEFINED__ */


#ifndef __MouseSearchAttributesSettings_FWD_DEFINED__
#define __MouseSearchAttributesSettings_FWD_DEFINED__
typedef interface MouseSearchAttributesSettings MouseSearchAttributesSettings;

#endif 	/* __MouseSearchAttributesSettings_FWD_DEFINED__ */


#ifndef __GeneralSettings_FWD_DEFINED__
#define __GeneralSettings_FWD_DEFINED__
typedef interface GeneralSettings GeneralSettings;

#endif 	/* __GeneralSettings_FWD_DEFINED__ */


#ifndef __Verbosity_FWD_DEFINED__
#define __Verbosity_FWD_DEFINED__
typedef interface Verbosity Verbosity;

#endif 	/* __Verbosity_FWD_DEFINED__ */


#ifndef __MenuVerbosity_FWD_DEFINED__
#define __MenuVerbosity_FWD_DEFINED__
typedef interface MenuVerbosity MenuVerbosity;

#endif 	/* __MenuVerbosity_FWD_DEFINED__ */


#ifndef __ActivatedVerbosity_FWD_DEFINED__
#define __ActivatedVerbosity_FWD_DEFINED__
typedef interface ActivatedVerbosity ActivatedVerbosity;

#endif 	/* __ActivatedVerbosity_FWD_DEFINED__ */


#ifndef __FocusedVerbosity_FWD_DEFINED__
#define __FocusedVerbosity_FWD_DEFINED__
typedef interface FocusedVerbosity FocusedVerbosity;

#endif 	/* __FocusedVerbosity_FWD_DEFINED__ */


#ifndef __AttributeChangeVerbosity_FWD_DEFINED__
#define __AttributeChangeVerbosity_FWD_DEFINED__
typedef interface AttributeChangeVerbosity AttributeChangeVerbosity;

#endif 	/* __AttributeChangeVerbosity_FWD_DEFINED__ */


#ifndef __MiscVerbosity_FWD_DEFINED__
#define __MiscVerbosity_FWD_DEFINED__
typedef interface MiscVerbosity MiscVerbosity;

#endif 	/* __MiscVerbosity_FWD_DEFINED__ */


#ifndef __BrowseModeVerbosity_FWD_DEFINED__
#define __BrowseModeVerbosity_FWD_DEFINED__
typedef interface BrowseModeVerbosity BrowseModeVerbosity;

#endif 	/* __BrowseModeVerbosity_FWD_DEFINED__ */


#ifndef __BrowseModeAccessKeyVerbosity_FWD_DEFINED__
#define __BrowseModeAccessKeyVerbosity_FWD_DEFINED__
typedef interface BrowseModeAccessKeyVerbosity BrowseModeAccessKeyVerbosity;

#endif 	/* __BrowseModeAccessKeyVerbosity_FWD_DEFINED__ */


#ifndef __BrowseModeAcronymVerbosity_FWD_DEFINED__
#define __BrowseModeAcronymVerbosity_FWD_DEFINED__
typedef interface BrowseModeAcronymVerbosity BrowseModeAcronymVerbosity;

#endif 	/* __BrowseModeAcronymVerbosity_FWD_DEFINED__ */


#ifndef __BrowseModeFlashVerbosity_FWD_DEFINED__
#define __BrowseModeFlashVerbosity_FWD_DEFINED__
typedef interface BrowseModeFlashVerbosity BrowseModeFlashVerbosity;

#endif 	/* __BrowseModeFlashVerbosity_FWD_DEFINED__ */


#ifndef __BrowseModeFormVerbosity_FWD_DEFINED__
#define __BrowseModeFormVerbosity_FWD_DEFINED__
typedef interface BrowseModeFormVerbosity BrowseModeFormVerbosity;

#endif 	/* __BrowseModeFormVerbosity_FWD_DEFINED__ */


#ifndef __BrowseModeHeadingVerbosity_FWD_DEFINED__
#define __BrowseModeHeadingVerbosity_FWD_DEFINED__
typedef interface BrowseModeHeadingVerbosity BrowseModeHeadingVerbosity;

#endif 	/* __BrowseModeHeadingVerbosity_FWD_DEFINED__ */


#ifndef __BrowseModeLanguageVerbosity_FWD_DEFINED__
#define __BrowseModeLanguageVerbosity_FWD_DEFINED__
typedef interface BrowseModeLanguageVerbosity BrowseModeLanguageVerbosity;

#endif 	/* __BrowseModeLanguageVerbosity_FWD_DEFINED__ */


#ifndef __BrowseModeLinkVerbosity_FWD_DEFINED__
#define __BrowseModeLinkVerbosity_FWD_DEFINED__
typedef interface BrowseModeLinkVerbosity BrowseModeLinkVerbosity;

#endif 	/* __BrowseModeLinkVerbosity_FWD_DEFINED__ */


#ifndef __BrowseModeListVerbosity_FWD_DEFINED__
#define __BrowseModeListVerbosity_FWD_DEFINED__
typedef interface BrowseModeListVerbosity BrowseModeListVerbosity;

#endif 	/* __BrowseModeListVerbosity_FWD_DEFINED__ */


#ifndef __BrowseModeLongDescVerbosity_FWD_DEFINED__
#define __BrowseModeLongDescVerbosity_FWD_DEFINED__
typedef interface BrowseModeLongDescVerbosity BrowseModeLongDescVerbosity;

#endif 	/* __BrowseModeLongDescVerbosity_FWD_DEFINED__ */


#ifndef __BrowseModeMiscVerbosity_FWD_DEFINED__
#define __BrowseModeMiscVerbosity_FWD_DEFINED__
typedef interface BrowseModeMiscVerbosity BrowseModeMiscVerbosity;

#endif 	/* __BrowseModeMiscVerbosity_FWD_DEFINED__ */


#ifndef __BrowseModeNewPageVerbosity_FWD_DEFINED__
#define __BrowseModeNewPageVerbosity_FWD_DEFINED__
typedef interface BrowseModeNewPageVerbosity BrowseModeNewPageVerbosity;

#endif 	/* __BrowseModeNewPageVerbosity_FWD_DEFINED__ */


#ifndef __BrowseModeObjectVerbosity_FWD_DEFINED__
#define __BrowseModeObjectVerbosity_FWD_DEFINED__
typedef interface BrowseModeObjectVerbosity BrowseModeObjectVerbosity;

#endif 	/* __BrowseModeObjectVerbosity_FWD_DEFINED__ */


#ifndef __BrowseModeParagraphVerbosity_FWD_DEFINED__
#define __BrowseModeParagraphVerbosity_FWD_DEFINED__
typedef interface BrowseModeParagraphVerbosity BrowseModeParagraphVerbosity;

#endif 	/* __BrowseModeParagraphVerbosity_FWD_DEFINED__ */


#ifndef __BrowseModeQuoteVerbosity_FWD_DEFINED__
#define __BrowseModeQuoteVerbosity_FWD_DEFINED__
typedef interface BrowseModeQuoteVerbosity BrowseModeQuoteVerbosity;

#endif 	/* __BrowseModeQuoteVerbosity_FWD_DEFINED__ */


#ifndef __BrowseModeTableVerbosity_FWD_DEFINED__
#define __BrowseModeTableVerbosity_FWD_DEFINED__
typedef interface BrowseModeTableVerbosity BrowseModeTableVerbosity;

#endif 	/* __BrowseModeTableVerbosity_FWD_DEFINED__ */


#ifndef __WordVerbosity_FWD_DEFINED__
#define __WordVerbosity_FWD_DEFINED__
typedef interface WordVerbosity WordVerbosity;

#endif 	/* __WordVerbosity_FWD_DEFINED__ */


#ifndef __WordColumnVerbosity_FWD_DEFINED__
#define __WordColumnVerbosity_FWD_DEFINED__
typedef interface WordColumnVerbosity WordColumnVerbosity;

#endif 	/* __WordColumnVerbosity_FWD_DEFINED__ */


#ifndef __WordFieldVerbosity_FWD_DEFINED__
#define __WordFieldVerbosity_FWD_DEFINED__
typedef interface WordFieldVerbosity WordFieldVerbosity;

#endif 	/* __WordFieldVerbosity_FWD_DEFINED__ */


#ifndef __WordHeaderFooterVerbosity_FWD_DEFINED__
#define __WordHeaderFooterVerbosity_FWD_DEFINED__
typedef interface WordHeaderFooterVerbosity WordHeaderFooterVerbosity;

#endif 	/* __WordHeaderFooterVerbosity_FWD_DEFINED__ */


#ifndef __WordMiscVerbosity_FWD_DEFINED__
#define __WordMiscVerbosity_FWD_DEFINED__
typedef interface WordMiscVerbosity WordMiscVerbosity;

#endif 	/* __WordMiscVerbosity_FWD_DEFINED__ */


#ifndef __WordPictureObjectVerbosity_FWD_DEFINED__
#define __WordPictureObjectVerbosity_FWD_DEFINED__
typedef interface WordPictureObjectVerbosity WordPictureObjectVerbosity;

#endif 	/* __WordPictureObjectVerbosity_FWD_DEFINED__ */


#ifndef __WordReferenceVerbosity_FWD_DEFINED__
#define __WordReferenceVerbosity_FWD_DEFINED__
typedef interface WordReferenceVerbosity WordReferenceVerbosity;

#endif 	/* __WordReferenceVerbosity_FWD_DEFINED__ */


#ifndef __WordRevisionVerbosity_FWD_DEFINED__
#define __WordRevisionVerbosity_FWD_DEFINED__
typedef interface WordRevisionVerbosity WordRevisionVerbosity;

#endif 	/* __WordRevisionVerbosity_FWD_DEFINED__ */


#ifndef __WordSectionVerbosity_FWD_DEFINED__
#define __WordSectionVerbosity_FWD_DEFINED__
typedef interface WordSectionVerbosity WordSectionVerbosity;

#endif 	/* __WordSectionVerbosity_FWD_DEFINED__ */


#ifndef __WordSpellingGrammarVerbosity_FWD_DEFINED__
#define __WordSpellingGrammarVerbosity_FWD_DEFINED__
typedef interface WordSpellingGrammarVerbosity WordSpellingGrammarVerbosity;

#endif 	/* __WordSpellingGrammarVerbosity_FWD_DEFINED__ */


#ifndef __WordTableVerbosity_FWD_DEFINED__
#define __WordTableVerbosity_FWD_DEFINED__
typedef interface WordTableVerbosity WordTableVerbosity;

#endif 	/* __WordTableVerbosity_FWD_DEFINED__ */


#ifndef __ExcelVerbosity_FWD_DEFINED__
#define __ExcelVerbosity_FWD_DEFINED__
typedef interface ExcelVerbosity ExcelVerbosity;

#endif 	/* __ExcelVerbosity_FWD_DEFINED__ */


#ifndef __ExcelCellVerbosity_FWD_DEFINED__
#define __ExcelCellVerbosity_FWD_DEFINED__
typedef interface ExcelCellVerbosity ExcelCellVerbosity;

#endif 	/* __ExcelCellVerbosity_FWD_DEFINED__ */


#ifndef __ExcelMiscVerbosity_FWD_DEFINED__
#define __ExcelMiscVerbosity_FWD_DEFINED__
typedef interface ExcelMiscVerbosity ExcelMiscVerbosity;

#endif 	/* __ExcelMiscVerbosity_FWD_DEFINED__ */


#ifndef __ExcelSelectionVerbosity_FWD_DEFINED__
#define __ExcelSelectionVerbosity_FWD_DEFINED__
typedef interface ExcelSelectionVerbosity ExcelSelectionVerbosity;

#endif 	/* __ExcelSelectionVerbosity_FWD_DEFINED__ */


#ifndef __ExcelSummaryVerbosity_FWD_DEFINED__
#define __ExcelSummaryVerbosity_FWD_DEFINED__
typedef interface ExcelSummaryVerbosity ExcelSummaryVerbosity;

#endif 	/* __ExcelSummaryVerbosity_FWD_DEFINED__ */


#ifndef __PowerPointVerbosity_FWD_DEFINED__
#define __PowerPointVerbosity_FWD_DEFINED__
typedef interface PowerPointVerbosity PowerPointVerbosity;

#endif 	/* __PowerPointVerbosity_FWD_DEFINED__ */


#ifndef __PowerPointSlideVerbosity_FWD_DEFINED__
#define __PowerPointSlideVerbosity_FWD_DEFINED__
typedef interface PowerPointSlideVerbosity PowerPointSlideVerbosity;

#endif 	/* __PowerPointSlideVerbosity_FWD_DEFINED__ */


#ifndef __PowerPointPlaybackVerbosity_FWD_DEFINED__
#define __PowerPointPlaybackVerbosity_FWD_DEFINED__
typedef interface PowerPointPlaybackVerbosity PowerPointPlaybackVerbosity;

#endif 	/* __PowerPointPlaybackVerbosity_FWD_DEFINED__ */


#ifndef __MouseVerbosity_FWD_DEFINED__
#define __MouseVerbosity_FWD_DEFINED__
typedef interface MouseVerbosity MouseVerbosity;

#endif 	/* __MouseVerbosity_FWD_DEFINED__ */


#ifndef __BrailleSettings_FWD_DEFINED__
#define __BrailleSettings_FWD_DEFINED__
typedef interface BrailleSettings BrailleSettings;

#endif 	/* __BrailleSettings_FWD_DEFINED__ */


#ifndef __BrailleScrollingOptions_FWD_DEFINED__
#define __BrailleScrollingOptions_FWD_DEFINED__
typedef interface BrailleScrollingOptions BrailleScrollingOptions;

#endif 	/* __BrailleScrollingOptions_FWD_DEFINED__ */


#ifndef __BrailleControlInformation_FWD_DEFINED__
#define __BrailleControlInformation_FWD_DEFINED__
typedef interface BrailleControlInformation BrailleControlInformation;

#endif 	/* __BrailleControlInformation_FWD_DEFINED__ */


#ifndef __BrailleControlTypes_FWD_DEFINED__
#define __BrailleControlTypes_FWD_DEFINED__
typedef interface BrailleControlTypes BrailleControlTypes;

#endif 	/* __BrailleControlTypes_FWD_DEFINED__ */


#ifndef __BrailleControlType_FWD_DEFINED__
#define __BrailleControlType_FWD_DEFINED__
typedef interface BrailleControlType BrailleControlType;

#endif 	/* __BrailleControlType_FWD_DEFINED__ */


#ifndef __BrailleControlStates_FWD_DEFINED__
#define __BrailleControlStates_FWD_DEFINED__
typedef interface BrailleControlStates BrailleControlStates;

#endif 	/* __BrailleControlStates_FWD_DEFINED__ */


#ifndef __BrailleControlState_FWD_DEFINED__
#define __BrailleControlState_FWD_DEFINED__
typedef interface BrailleControlState BrailleControlState;

#endif 	/* __BrailleControlState_FWD_DEFINED__ */


#ifndef __BrailleDotPatterns_FWD_DEFINED__
#define __BrailleDotPatterns_FWD_DEFINED__
typedef interface BrailleDotPatterns BrailleDotPatterns;

#endif 	/* __BrailleDotPatterns_FWD_DEFINED__ */


#ifndef __BrailleTables_FWD_DEFINED__
#define __BrailleTables_FWD_DEFINED__
typedef interface BrailleTables BrailleTables;

#endif 	/* __BrailleTables_FWD_DEFINED__ */


#ifndef __BrailleTable_FWD_DEFINED__
#define __BrailleTable_FWD_DEFINED__
typedef interface BrailleTable BrailleTable;

#endif 	/* __BrailleTable_FWD_DEFINED__ */


#ifndef __BrailleTableEntry_FWD_DEFINED__
#define __BrailleTableEntry_FWD_DEFINED__
typedef interface BrailleTableEntry BrailleTableEntry;

#endif 	/* __BrailleTableEntry_FWD_DEFINED__ */


#ifndef __BrailleOptions_FWD_DEFINED__
#define __BrailleOptions_FWD_DEFINED__
typedef interface BrailleOptions BrailleOptions;

#endif 	/* __BrailleOptions_FWD_DEFINED__ */


#ifndef __BrailleGraphics_FWD_DEFINED__
#define __BrailleGraphics_FWD_DEFINED__
typedef interface BrailleGraphics BrailleGraphics;

#endif 	/* __BrailleGraphics_FWD_DEFINED__ */


#ifndef __UserWindows_FWD_DEFINED__
#define __UserWindows_FWD_DEFINED__
typedef interface UserWindows UserWindows;

#endif 	/* __UserWindows_FWD_DEFINED__ */


#ifndef __UserWindow_FWD_DEFINED__
#define __UserWindow_FWD_DEFINED__
typedef interface UserWindow UserWindow;

#endif 	/* __UserWindow_FWD_DEFINED__ */


#ifndef __UserWindowCoordinates_FWD_DEFINED__
#define __UserWindowCoordinates_FWD_DEFINED__
typedef interface UserWindowCoordinates UserWindowCoordinates;

#endif 	/* __UserWindowCoordinates_FWD_DEFINED__ */


#ifndef __HyperactiveWindows_FWD_DEFINED__
#define __HyperactiveWindows_FWD_DEFINED__
typedef interface HyperactiveWindows HyperactiveWindows;

#endif 	/* __HyperactiveWindows_FWD_DEFINED__ */


#ifndef __HyperactiveWindow_FWD_DEFINED__
#define __HyperactiveWindow_FWD_DEFINED__
typedef interface HyperactiveWindow HyperactiveWindow;

#endif 	/* __HyperactiveWindow_FWD_DEFINED__ */


#ifndef __HyperactiveWindowCommands_FWD_DEFINED__
#define __HyperactiveWindowCommands_FWD_DEFINED__
typedef interface HyperactiveWindowCommands HyperactiveWindowCommands;

#endif 	/* __HyperactiveWindowCommands_FWD_DEFINED__ */


#ifndef __HyperactiveWindowCommand_FWD_DEFINED__
#define __HyperactiveWindowCommand_FWD_DEFINED__
typedef interface HyperactiveWindowCommand HyperactiveWindowCommand;

#endif 	/* __HyperactiveWindowCommand_FWD_DEFINED__ */


#ifndef __TriggerSearchAttribute_FWD_DEFINED__
#define __TriggerSearchAttribute_FWD_DEFINED__
typedef interface TriggerSearchAttribute TriggerSearchAttribute;

#endif 	/* __TriggerSearchAttribute_FWD_DEFINED__ */


#ifndef __HotKeys_FWD_DEFINED__
#define __HotKeys_FWD_DEFINED__
typedef interface HotKeys HotKeys;

#endif 	/* __HotKeys_FWD_DEFINED__ */


#ifndef __Hotkey_FWD_DEFINED__
#define __Hotkey_FWD_DEFINED__
typedef interface Hotkey Hotkey;

#endif 	/* __Hotkey_FWD_DEFINED__ */


#ifndef __CursorKeys_FWD_DEFINED__
#define __CursorKeys_FWD_DEFINED__
typedef interface CursorKeys CursorKeys;

#endif 	/* __CursorKeys_FWD_DEFINED__ */


#ifndef __CursorKey_FWD_DEFINED__
#define __CursorKey_FWD_DEFINED__
typedef interface CursorKey CursorKey;

#endif 	/* __CursorKey_FWD_DEFINED__ */


#ifndef __CursorKeyAction_FWD_DEFINED__
#define __CursorKeyAction_FWD_DEFINED__
typedef interface CursorKeyAction CursorKeyAction;

#endif 	/* __CursorKeyAction_FWD_DEFINED__ */


#ifndef __CharacterDictionary_FWD_DEFINED__
#define __CharacterDictionary_FWD_DEFINED__
typedef interface CharacterDictionary CharacterDictionary;

#endif 	/* __CharacterDictionary_FWD_DEFINED__ */


#ifndef __CharacterEntry_FWD_DEFINED__
#define __CharacterEntry_FWD_DEFINED__
typedef interface CharacterEntry CharacterEntry;

#endif 	/* __CharacterEntry_FWD_DEFINED__ */


#ifndef __ColorDictionary_FWD_DEFINED__
#define __ColorDictionary_FWD_DEFINED__
typedef interface ColorDictionary ColorDictionary;

#endif 	/* __ColorDictionary_FWD_DEFINED__ */


#ifndef __ColorEntry_FWD_DEFINED__
#define __ColorEntry_FWD_DEFINED__
typedef interface ColorEntry ColorEntry;

#endif 	/* __ColorEntry_FWD_DEFINED__ */


#ifndef __ExceptionDictionary_FWD_DEFINED__
#define __ExceptionDictionary_FWD_DEFINED__
typedef interface ExceptionDictionary ExceptionDictionary;

#endif 	/* __ExceptionDictionary_FWD_DEFINED__ */


#ifndef __ExceptionEntry_FWD_DEFINED__
#define __ExceptionEntry_FWD_DEFINED__
typedef interface ExceptionEntry ExceptionEntry;

#endif 	/* __ExceptionEntry_FWD_DEFINED__ */


#ifndef __GraphicDictionary_FWD_DEFINED__
#define __GraphicDictionary_FWD_DEFINED__
typedef interface GraphicDictionary GraphicDictionary;

#endif 	/* __GraphicDictionary_FWD_DEFINED__ */


#ifndef __GraphicEntry_FWD_DEFINED__
#define __GraphicEntry_FWD_DEFINED__
typedef interface GraphicEntry GraphicEntry;

#endif 	/* __GraphicEntry_FWD_DEFINED__ */


#ifndef __KeyLabelDictionary_FWD_DEFINED__
#define __KeyLabelDictionary_FWD_DEFINED__
typedef interface KeyLabelDictionary KeyLabelDictionary;

#endif 	/* __KeyLabelDictionary_FWD_DEFINED__ */


#ifndef __KeyLabelEntry_FWD_DEFINED__
#define __KeyLabelEntry_FWD_DEFINED__
typedef interface KeyLabelEntry KeyLabelEntry;

#endif 	/* __KeyLabelEntry_FWD_DEFINED__ */


#ifndef __GlobalSettings_FWD_DEFINED__
#define __GlobalSettings_FWD_DEFINED__
typedef interface GlobalSettings GlobalSettings;

#endif 	/* __GlobalSettings_FWD_DEFINED__ */


#ifndef __SetFileScope_FWD_DEFINED__
#define __SetFileScope_FWD_DEFINED__
typedef interface SetFileScope SetFileScope;

#endif 	/* __SetFileScope_FWD_DEFINED__ */


#ifndef __ScreenSettingsScope_FWD_DEFINED__
#define __ScreenSettingsScope_FWD_DEFINED__
typedef interface ScreenSettingsScope ScreenSettingsScope;

#endif 	/* __ScreenSettingsScope_FWD_DEFINED__ */


#ifndef __PunctuationSettingsScope_FWD_DEFINED__
#define __PunctuationSettingsScope_FWD_DEFINED__
typedef interface PunctuationSettingsScope PunctuationSettingsScope;

#endif 	/* __PunctuationSettingsScope_FWD_DEFINED__ */


#ifndef __KeyboardSettingsScope_FWD_DEFINED__
#define __KeyboardSettingsScope_FWD_DEFINED__
typedef interface KeyboardSettingsScope KeyboardSettingsScope;

#endif 	/* __KeyboardSettingsScope_FWD_DEFINED__ */


#ifndef __SpecialKeySettingsScope_FWD_DEFINED__
#define __SpecialKeySettingsScope_FWD_DEFINED__
typedef interface SpecialKeySettingsScope SpecialKeySettingsScope;

#endif 	/* __SpecialKeySettingsScope_FWD_DEFINED__ */


#ifndef __MouseSettingsScope_FWD_DEFINED__
#define __MouseSettingsScope_FWD_DEFINED__
typedef interface MouseSettingsScope MouseSettingsScope;

#endif 	/* __MouseSettingsScope_FWD_DEFINED__ */


#ifndef __GeneralSettingsScope_FWD_DEFINED__
#define __GeneralSettingsScope_FWD_DEFINED__
typedef interface GeneralSettingsScope GeneralSettingsScope;

#endif 	/* __GeneralSettingsScope_FWD_DEFINED__ */


#ifndef __BrailleSettingsScope_FWD_DEFINED__
#define __BrailleSettingsScope_FWD_DEFINED__
typedef interface BrailleSettingsScope BrailleSettingsScope;

#endif 	/* __BrailleSettingsScope_FWD_DEFINED__ */


#ifndef __BrailleScrollingOptionsScope_FWD_DEFINED__
#define __BrailleScrollingOptionsScope_FWD_DEFINED__
typedef interface BrailleScrollingOptionsScope BrailleScrollingOptionsScope;

#endif 	/* __BrailleScrollingOptionsScope_FWD_DEFINED__ */


#ifndef __BrailleControlInformationScope_FWD_DEFINED__
#define __BrailleControlInformationScope_FWD_DEFINED__
typedef interface BrailleControlInformationScope BrailleControlInformationScope;

#endif 	/* __BrailleControlInformationScope_FWD_DEFINED__ */


#ifndef __BrailleOptionsScope_FWD_DEFINED__
#define __BrailleOptionsScope_FWD_DEFINED__
typedef interface BrailleOptionsScope BrailleOptionsScope;

#endif 	/* __BrailleOptionsScope_FWD_DEFINED__ */


#ifndef __BrailleGraphicsScope_FWD_DEFINED__
#define __BrailleGraphicsScope_FWD_DEFINED__
typedef interface BrailleGraphicsScope BrailleGraphicsScope;

#endif 	/* __BrailleGraphicsScope_FWD_DEFINED__ */


#ifndef __HotkeysScope_FWD_DEFINED__
#define __HotkeysScope_FWD_DEFINED__
typedef interface HotkeysScope HotkeysScope;

#endif 	/* __HotkeysScope_FWD_DEFINED__ */


#ifndef __HotkeyScope_FWD_DEFINED__
#define __HotkeyScope_FWD_DEFINED__
typedef interface HotkeyScope HotkeyScope;

#endif 	/* __HotkeyScope_FWD_DEFINED__ */


#ifndef __VerbosityScope_FWD_DEFINED__
#define __VerbosityScope_FWD_DEFINED__
typedef interface VerbosityScope VerbosityScope;

#endif 	/* __VerbosityScope_FWD_DEFINED__ */


#ifndef __MenuVerbosityScope_FWD_DEFINED__
#define __MenuVerbosityScope_FWD_DEFINED__
typedef interface MenuVerbosityScope MenuVerbosityScope;

#endif 	/* __MenuVerbosityScope_FWD_DEFINED__ */


#ifndef __ActivatedVerbosityScope_FWD_DEFINED__
#define __ActivatedVerbosityScope_FWD_DEFINED__
typedef interface ActivatedVerbosityScope ActivatedVerbosityScope;

#endif 	/* __ActivatedVerbosityScope_FWD_DEFINED__ */


#ifndef __FocusedVerbosityScope_FWD_DEFINED__
#define __FocusedVerbosityScope_FWD_DEFINED__
typedef interface FocusedVerbosityScope FocusedVerbosityScope;

#endif 	/* __FocusedVerbosityScope_FWD_DEFINED__ */


#ifndef __AttributeChangeVerbosityScope_FWD_DEFINED__
#define __AttributeChangeVerbosityScope_FWD_DEFINED__
typedef interface AttributeChangeVerbosityScope AttributeChangeVerbosityScope;

#endif 	/* __AttributeChangeVerbosityScope_FWD_DEFINED__ */


#ifndef __MiscVerbosityScope_FWD_DEFINED__
#define __MiscVerbosityScope_FWD_DEFINED__
typedef interface MiscVerbosityScope MiscVerbosityScope;

#endif 	/* __MiscVerbosityScope_FWD_DEFINED__ */


#ifndef __BrowseModeVerbosityScope_FWD_DEFINED__
#define __BrowseModeVerbosityScope_FWD_DEFINED__
typedef interface BrowseModeVerbosityScope BrowseModeVerbosityScope;

#endif 	/* __BrowseModeVerbosityScope_FWD_DEFINED__ */


#ifndef __BrowseModeAccessKeyVerbosityScope_FWD_DEFINED__
#define __BrowseModeAccessKeyVerbosityScope_FWD_DEFINED__
typedef interface BrowseModeAccessKeyVerbosityScope BrowseModeAccessKeyVerbosityScope;

#endif 	/* __BrowseModeAccessKeyVerbosityScope_FWD_DEFINED__ */


#ifndef __BrowseModeAcronymVerbosityScope_FWD_DEFINED__
#define __BrowseModeAcronymVerbosityScope_FWD_DEFINED__
typedef interface BrowseModeAcronymVerbosityScope BrowseModeAcronymVerbosityScope;

#endif 	/* __BrowseModeAcronymVerbosityScope_FWD_DEFINED__ */


#ifndef __BrowseModeFlashVerbosityScope_FWD_DEFINED__
#define __BrowseModeFlashVerbosityScope_FWD_DEFINED__
typedef interface BrowseModeFlashVerbosityScope BrowseModeFlashVerbosityScope;

#endif 	/* __BrowseModeFlashVerbosityScope_FWD_DEFINED__ */


#ifndef __BrowseModeFormVerbosityScope_FWD_DEFINED__
#define __BrowseModeFormVerbosityScope_FWD_DEFINED__
typedef interface BrowseModeFormVerbosityScope BrowseModeFormVerbosityScope;

#endif 	/* __BrowseModeFormVerbosityScope_FWD_DEFINED__ */


#ifndef __BrowseModeHeadingVerbosityScope_FWD_DEFINED__
#define __BrowseModeHeadingVerbosityScope_FWD_DEFINED__
typedef interface BrowseModeHeadingVerbosityScope BrowseModeHeadingVerbosityScope;

#endif 	/* __BrowseModeHeadingVerbosityScope_FWD_DEFINED__ */


#ifndef __BrowseModeLanguageVerbosityScope_FWD_DEFINED__
#define __BrowseModeLanguageVerbosityScope_FWD_DEFINED__
typedef interface BrowseModeLanguageVerbosityScope BrowseModeLanguageVerbosityScope;

#endif 	/* __BrowseModeLanguageVerbosityScope_FWD_DEFINED__ */


#ifndef __BrowseModeLinkVerbosityScope_FWD_DEFINED__
#define __BrowseModeLinkVerbosityScope_FWD_DEFINED__
typedef interface BrowseModeLinkVerbosityScope BrowseModeLinkVerbosityScope;

#endif 	/* __BrowseModeLinkVerbosityScope_FWD_DEFINED__ */


#ifndef __BrowseModeListVerbosityScope_FWD_DEFINED__
#define __BrowseModeListVerbosityScope_FWD_DEFINED__
typedef interface BrowseModeListVerbosityScope BrowseModeListVerbosityScope;

#endif 	/* __BrowseModeListVerbosityScope_FWD_DEFINED__ */


#ifndef __BrowseModeLongDescVerbosityScope_FWD_DEFINED__
#define __BrowseModeLongDescVerbosityScope_FWD_DEFINED__
typedef interface BrowseModeLongDescVerbosityScope BrowseModeLongDescVerbosityScope;

#endif 	/* __BrowseModeLongDescVerbosityScope_FWD_DEFINED__ */


#ifndef __BrowseModeMiscVerbosityScope_FWD_DEFINED__
#define __BrowseModeMiscVerbosityScope_FWD_DEFINED__
typedef interface BrowseModeMiscVerbosityScope BrowseModeMiscVerbosityScope;

#endif 	/* __BrowseModeMiscVerbosityScope_FWD_DEFINED__ */


#ifndef __BrowseModeNewPageVerbosityScope_FWD_DEFINED__
#define __BrowseModeNewPageVerbosityScope_FWD_DEFINED__
typedef interface BrowseModeNewPageVerbosityScope BrowseModeNewPageVerbosityScope;

#endif 	/* __BrowseModeNewPageVerbosityScope_FWD_DEFINED__ */


#ifndef __BrowseModeObjectVerbosityScope_FWD_DEFINED__
#define __BrowseModeObjectVerbosityScope_FWD_DEFINED__
typedef interface BrowseModeObjectVerbosityScope BrowseModeObjectVerbosityScope;

#endif 	/* __BrowseModeObjectVerbosityScope_FWD_DEFINED__ */


#ifndef __BrowseModeParagraphVerbosityScope_FWD_DEFINED__
#define __BrowseModeParagraphVerbosityScope_FWD_DEFINED__
typedef interface BrowseModeParagraphVerbosityScope BrowseModeParagraphVerbosityScope;

#endif 	/* __BrowseModeParagraphVerbosityScope_FWD_DEFINED__ */


#ifndef __BrowseModeQuoteVerbosityScope_FWD_DEFINED__
#define __BrowseModeQuoteVerbosityScope_FWD_DEFINED__
typedef interface BrowseModeQuoteVerbosityScope BrowseModeQuoteVerbosityScope;

#endif 	/* __BrowseModeQuoteVerbosityScope_FWD_DEFINED__ */


#ifndef __BrowseModeTableVerbosityScope_FWD_DEFINED__
#define __BrowseModeTableVerbosityScope_FWD_DEFINED__
typedef interface BrowseModeTableVerbosityScope BrowseModeTableVerbosityScope;

#endif 	/* __BrowseModeTableVerbosityScope_FWD_DEFINED__ */


#ifndef __MouseVerbosityScope_FWD_DEFINED__
#define __MouseVerbosityScope_FWD_DEFINED__
typedef interface MouseVerbosityScope MouseVerbosityScope;

#endif 	/* __MouseVerbosityScope_FWD_DEFINED__ */


#ifndef __Settings2_FWD_DEFINED__
#define __Settings2_FWD_DEFINED__
typedef interface Settings2 Settings2;

#endif 	/* __Settings2_FWD_DEFINED__ */


#ifndef ___Mouse_FWD_DEFINED__
#define ___Mouse_FWD_DEFINED__
typedef interface _Mouse _Mouse;

#endif 	/* ___Mouse_FWD_DEFINED__ */


#ifndef ___MousePointer_FWD_DEFINED__
#define ___MousePointer_FWD_DEFINED__
typedef interface _MousePointer _MousePointer;

#endif 	/* ___MousePointer_FWD_DEFINED__ */


#ifndef __Cursor_FWD_DEFINED__
#define __Cursor_FWD_DEFINED__
typedef interface Cursor Cursor;

#endif 	/* __Cursor_FWD_DEFINED__ */


#ifndef ___Keyboard_FWD_DEFINED__
#define ___Keyboard_FWD_DEFINED__
typedef interface _Keyboard _Keyboard;

#endif 	/* ___Keyboard_FWD_DEFINED__ */


#ifndef __Keys_FWD_DEFINED__
#define __Keys_FWD_DEFINED__
typedef interface Keys Keys;

#endif 	/* __Keys_FWD_DEFINED__ */


#ifndef ___Key_FWD_DEFINED__
#define ___Key_FWD_DEFINED__
typedef interface _Key _Key;

#endif 	/* ___Key_FWD_DEFINED__ */


#ifndef __RegisteredKey_FWD_DEFINED__
#define __RegisteredKey_FWD_DEFINED__
typedef interface RegisteredKey RegisteredKey;

#endif 	/* __RegisteredKey_FWD_DEFINED__ */


#ifndef __RegisteredHotkeys_FWD_DEFINED__
#define __RegisteredHotkeys_FWD_DEFINED__
typedef interface RegisteredHotkeys RegisteredHotkeys;

#endif 	/* __RegisteredHotkeys_FWD_DEFINED__ */


#ifndef __RegisteredHotkey_FWD_DEFINED__
#define __RegisteredHotkey_FWD_DEFINED__
typedef interface RegisteredHotkey RegisteredHotkey;

#endif 	/* __RegisteredHotkey_FWD_DEFINED__ */


#ifndef __LoadedScript_FWD_DEFINED__
#define __LoadedScript_FWD_DEFINED__
typedef interface LoadedScript LoadedScript;

#endif 	/* __LoadedScript_FWD_DEFINED__ */


#ifndef __CharacterDictionaries_FWD_DEFINED__
#define __CharacterDictionaries_FWD_DEFINED__
typedef interface CharacterDictionaries CharacterDictionaries;

#endif 	/* __CharacterDictionaries_FWD_DEFINED__ */


#ifndef __ColorDictionaries_FWD_DEFINED__
#define __ColorDictionaries_FWD_DEFINED__
typedef interface ColorDictionaries ColorDictionaries;

#endif 	/* __ColorDictionaries_FWD_DEFINED__ */


#ifndef __ExceptionDictionaries_FWD_DEFINED__
#define __ExceptionDictionaries_FWD_DEFINED__
typedef interface ExceptionDictionaries ExceptionDictionaries;

#endif 	/* __ExceptionDictionaries_FWD_DEFINED__ */


#ifndef __GraphicDictionaries_FWD_DEFINED__
#define __GraphicDictionaries_FWD_DEFINED__
typedef interface GraphicDictionaries GraphicDictionaries;

#endif 	/* __GraphicDictionaries_FWD_DEFINED__ */


#ifndef __KeyLabelDictionaries_FWD_DEFINED__
#define __KeyLabelDictionaries_FWD_DEFINED__
typedef interface KeyLabelDictionaries KeyLabelDictionaries;

#endif 	/* __KeyLabelDictionaries_FWD_DEFINED__ */


#ifndef ___BrailleDisplays_FWD_DEFINED__
#define ___BrailleDisplays_FWD_DEFINED__
typedef interface _BrailleDisplays _BrailleDisplays;

#endif 	/* ___BrailleDisplays_FWD_DEFINED__ */


#ifndef ___BrailleDisplay_FWD_DEFINED__
#define ___BrailleDisplay_FWD_DEFINED__
typedef interface _BrailleDisplay _BrailleDisplay;

#endif 	/* ___BrailleDisplay_FWD_DEFINED__ */


#ifndef __BrailleHotKey_FWD_DEFINED__
#define __BrailleHotKey_FWD_DEFINED__
typedef interface BrailleHotKey BrailleHotKey;

#endif 	/* __BrailleHotKey_FWD_DEFINED__ */


#ifndef __BrailleActions_FWD_DEFINED__
#define __BrailleActions_FWD_DEFINED__
typedef interface BrailleActions BrailleActions;

#endif 	/* __BrailleActions_FWD_DEFINED__ */


#ifndef __BrailleAction_FWD_DEFINED__
#define __BrailleAction_FWD_DEFINED__
typedef interface BrailleAction BrailleAction;

#endif 	/* __BrailleAction_FWD_DEFINED__ */


#ifndef __BrailleHotKeys_FWD_DEFINED__
#define __BrailleHotKeys_FWD_DEFINED__
typedef interface BrailleHotKeys BrailleHotKeys;

#endif 	/* __BrailleHotKeys_FWD_DEFINED__ */


#ifndef __BrailleDisplaysEvents_FWD_DEFINED__
#define __BrailleDisplaysEvents_FWD_DEFINED__
typedef interface BrailleDisplaysEvents BrailleDisplaysEvents;

#endif 	/* __BrailleDisplaysEvents_FWD_DEFINED__ */


#ifndef ___BrailleDisplaysEvents_FWD_DEFINED__
#define ___BrailleDisplaysEvents_FWD_DEFINED__
typedef interface _BrailleDisplaysEvents _BrailleDisplaysEvents;

#endif 	/* ___BrailleDisplaysEvents_FWD_DEFINED__ */


#ifndef ___SharedObjects_FWD_DEFINED__
#define ___SharedObjects_FWD_DEFINED__
typedef interface _SharedObjects _SharedObjects;

#endif 	/* ___SharedObjects_FWD_DEFINED__ */


#ifndef ___ClientInformation_FWD_DEFINED__
#define ___ClientInformation_FWD_DEFINED__
typedef interface _ClientInformation _ClientInformation;

#endif 	/* ___ClientInformation_FWD_DEFINED__ */


#ifndef __Signature_FWD_DEFINED__
#define __Signature_FWD_DEFINED__
typedef interface Signature Signature;

#endif 	/* __Signature_FWD_DEFINED__ */


#ifndef __WECursor_FWD_DEFINED__
#define __WECursor_FWD_DEFINED__
typedef interface WECursor WECursor;

#endif 	/* __WECursor_FWD_DEFINED__ */


#ifndef __Text_FWD_DEFINED__
#define __Text_FWD_DEFINED__
typedef interface Text Text;

#endif 	/* __Text_FWD_DEFINED__ */


#ifndef __Screen_FWD_DEFINED__
#define __Screen_FWD_DEFINED__
typedef interface Screen Screen;

#endif 	/* __Screen_FWD_DEFINED__ */


#ifndef ___Synthesizers_FWD_DEFINED__
#define ___Synthesizers_FWD_DEFINED__
typedef interface _Synthesizers _Synthesizers;

#endif 	/* ___Synthesizers_FWD_DEFINED__ */


#ifndef ___Synthesizer_FWD_DEFINED__
#define ___Synthesizer_FWD_DEFINED__
typedef interface _Synthesizer _Synthesizer;

#endif 	/* ___Synthesizer_FWD_DEFINED__ */


#ifndef __Engine_FWD_DEFINED__
#define __Engine_FWD_DEFINED__
typedef interface Engine Engine;

#endif 	/* __Engine_FWD_DEFINED__ */


#ifndef __Engines_FWD_DEFINED__
#define __Engines_FWD_DEFINED__
typedef interface Engines Engines;

#endif 	/* __Engines_FWD_DEFINED__ */


#ifndef __IniFile_FWD_DEFINED__
#define __IniFile_FWD_DEFINED__
typedef interface IniFile IniFile;

#endif 	/* __IniFile_FWD_DEFINED__ */


#ifndef ___Speech_FWD_DEFINED__
#define ___Speech_FWD_DEFINED__
typedef interface _Speech _Speech;

#endif 	/* ___Speech_FWD_DEFINED__ */


#ifndef ___MSAAEvents_FWD_DEFINED__
#define ___MSAAEvents_FWD_DEFINED__
typedef interface _MSAAEvents _MSAAEvents;

#endif 	/* ___MSAAEvents_FWD_DEFINED__ */


#ifndef __MSAAEventBlock_FWD_DEFINED__
#define __MSAAEventBlock_FWD_DEFINED__
typedef interface MSAAEventBlock MSAAEventBlock;

#endif 	/* __MSAAEventBlock_FWD_DEFINED__ */


#ifndef __Clipboard_FWD_DEFINED__
#define __Clipboard_FWD_DEFINED__
typedef interface Clipboard Clipboard;

#endif 	/* __Clipboard_FWD_DEFINED__ */


#ifndef ___BrowseMode_FWD_DEFINED__
#define ___BrowseMode_FWD_DEFINED__
typedef interface _BrowseMode _BrowseMode;

#endif 	/* ___BrowseMode_FWD_DEFINED__ */


#ifndef __BrowseModeLines_FWD_DEFINED__
#define __BrowseModeLines_FWD_DEFINED__
typedef interface BrowseModeLines BrowseModeLines;

#endif 	/* __BrowseModeLines_FWD_DEFINED__ */


#ifndef __BrowseModeLine_FWD_DEFINED__
#define __BrowseModeLine_FWD_DEFINED__
typedef interface BrowseModeLine BrowseModeLine;

#endif 	/* __BrowseModeLine_FWD_DEFINED__ */


#ifndef __WEDialogs_FWD_DEFINED__
#define __WEDialogs_FWD_DEFINED__
typedef interface WEDialogs WEDialogs;

#endif 	/* __WEDialogs_FWD_DEFINED__ */


#ifndef ___Braille_FWD_DEFINED__
#define ___Braille_FWD_DEFINED__
typedef interface _Braille _Braille;

#endif 	/* ___Braille_FWD_DEFINED__ */


#ifndef __BrailleAttributes_FWD_DEFINED__
#define __BrailleAttributes_FWD_DEFINED__
typedef interface BrailleAttributes BrailleAttributes;

#endif 	/* __BrailleAttributes_FWD_DEFINED__ */


#ifndef __OSVersion_FWD_DEFINED__
#define __OSVersion_FWD_DEFINED__
typedef interface OSVersion OSVersion;

#endif 	/* __OSVersion_FWD_DEFINED__ */


#ifndef ___Dialog_FWD_DEFINED__
#define ___Dialog_FWD_DEFINED__
typedef interface _Dialog _Dialog;

#endif 	/* ___Dialog_FWD_DEFINED__ */


#ifndef __Menu_FWD_DEFINED__
#define __Menu_FWD_DEFINED__
typedef interface Menu Menu;

#endif 	/* __Menu_FWD_DEFINED__ */


#ifndef __StartupOptions_FWD_DEFINED__
#define __StartupOptions_FWD_DEFINED__
typedef interface StartupOptions StartupOptions;

#endif 	/* __StartupOptions_FWD_DEFINED__ */


#ifndef __WEPM_FWD_DEFINED__
#define __WEPM_FWD_DEFINED__
typedef interface WEPM WEPM;

#endif 	/* __WEPM_FWD_DEFINED__ */


#ifndef __Profiles_FWD_DEFINED__
#define __Profiles_FWD_DEFINED__
typedef interface Profiles Profiles;

#endif 	/* __Profiles_FWD_DEFINED__ */


#ifndef __Profile_FWD_DEFINED__
#define __Profile_FWD_DEFINED__
typedef interface Profile Profile;

#endif 	/* __Profile_FWD_DEFINED__ */


#ifndef __Utilities_FWD_DEFINED__
#define __Utilities_FWD_DEFINED__
typedef interface Utilities Utilities;

#endif 	/* __Utilities_FWD_DEFINED__ */


#ifndef __LocaleInformation_FWD_DEFINED__
#define __LocaleInformation_FWD_DEFINED__
typedef interface LocaleInformation LocaleInformation;

#endif 	/* __LocaleInformation_FWD_DEFINED__ */


#ifndef __LoadedScripts_FWD_DEFINED__
#define __LoadedScripts_FWD_DEFINED__
typedef interface LoadedScripts LoadedScripts;

#endif 	/* __LoadedScripts_FWD_DEFINED__ */


#ifndef __WindowMessage_FWD_DEFINED__
#define __WindowMessage_FWD_DEFINED__
typedef interface WindowMessage WindowMessage;

#endif 	/* __WindowMessage_FWD_DEFINED__ */


#ifndef ___WindowEvents_FWD_DEFINED__
#define ___WindowEvents_FWD_DEFINED__
typedef interface _WindowEvents _WindowEvents;

#endif 	/* ___WindowEvents_FWD_DEFINED__ */


#ifndef ___Control_FWD_DEFINED__
#define ___Control_FWD_DEFINED__
typedef interface _Control _Control;

#endif 	/* ___Control_FWD_DEFINED__ */


#ifndef __TreeView_FWD_DEFINED__
#define __TreeView_FWD_DEFINED__
typedef interface TreeView TreeView;

#endif 	/* __TreeView_FWD_DEFINED__ */


#ifndef __TreeViewItem_FWD_DEFINED__
#define __TreeViewItem_FWD_DEFINED__
typedef interface TreeViewItem TreeViewItem;

#endif 	/* __TreeViewItem_FWD_DEFINED__ */


#ifndef __TreeViewItems_FWD_DEFINED__
#define __TreeViewItems_FWD_DEFINED__
typedef interface TreeViewItems TreeViewItems;

#endif 	/* __TreeViewItems_FWD_DEFINED__ */


#ifndef ___ApplicationEvents_FWD_DEFINED__
#define ___ApplicationEvents_FWD_DEFINED__
typedef interface _ApplicationEvents _ApplicationEvents;

#endif 	/* ___ApplicationEvents_FWD_DEFINED__ */


#ifndef __ApplicationEvents_FWD_DEFINED__
#define __ApplicationEvents_FWD_DEFINED__
typedef interface ApplicationEvents ApplicationEvents;

#endif 	/* __ApplicationEvents_FWD_DEFINED__ */


#ifndef ___SharedObjectEvents_FWD_DEFINED__
#define ___SharedObjectEvents_FWD_DEFINED__
typedef interface _SharedObjectEvents _SharedObjectEvents;

#endif 	/* ___SharedObjectEvents_FWD_DEFINED__ */


#ifndef __SharedObjectEvents_FWD_DEFINED__
#define __SharedObjectEvents_FWD_DEFINED__
typedef interface SharedObjectEvents SharedObjectEvents;

#endif 	/* __SharedObjectEvents_FWD_DEFINED__ */


#ifndef ___ClientInformationEvents_FWD_DEFINED__
#define ___ClientInformationEvents_FWD_DEFINED__
typedef interface _ClientInformationEvents _ClientInformationEvents;

#endif 	/* ___ClientInformationEvents_FWD_DEFINED__ */


#ifndef __ClientInformationEvents_FWD_DEFINED__
#define __ClientInformationEvents_FWD_DEFINED__
typedef interface ClientInformationEvents ClientInformationEvents;

#endif 	/* __ClientInformationEvents_FWD_DEFINED__ */


#ifndef ___Script_FWD_DEFINED__
#define ___Script_FWD_DEFINED__
typedef interface _Script _Script;

#endif 	/* ___Script_FWD_DEFINED__ */


#ifndef ___ScriptEvents_FWD_DEFINED__
#define ___ScriptEvents_FWD_DEFINED__
typedef interface _ScriptEvents _ScriptEvents;

#endif 	/* ___ScriptEvents_FWD_DEFINED__ */


#ifndef __ScriptEvents_FWD_DEFINED__
#define __ScriptEvents_FWD_DEFINED__
typedef interface ScriptEvents ScriptEvents;

#endif 	/* __ScriptEvents_FWD_DEFINED__ */


#ifndef ___MouseEvents_FWD_DEFINED__
#define ___MouseEvents_FWD_DEFINED__
typedef interface _MouseEvents _MouseEvents;

#endif 	/* ___MouseEvents_FWD_DEFINED__ */


#ifndef __MouseEvents_FWD_DEFINED__
#define __MouseEvents_FWD_DEFINED__
typedef interface MouseEvents MouseEvents;

#endif 	/* __MouseEvents_FWD_DEFINED__ */


#ifndef __IWEConnectionPointSink_FWD_DEFINED__
#define __IWEConnectionPointSink_FWD_DEFINED__
typedef interface IWEConnectionPointSink IWEConnectionPointSink;

#endif 	/* __IWEConnectionPointSink_FWD_DEFINED__ */


#ifndef __WindowEvents_FWD_DEFINED__
#define __WindowEvents_FWD_DEFINED__
typedef interface WindowEvents WindowEvents;

#endif 	/* __WindowEvents_FWD_DEFINED__ */


#ifndef ___KeyEvents_FWD_DEFINED__
#define ___KeyEvents_FWD_DEFINED__
typedef interface _KeyEvents _KeyEvents;

#endif 	/* ___KeyEvents_FWD_DEFINED__ */


#ifndef __KeyEvents_FWD_DEFINED__
#define __KeyEvents_FWD_DEFINED__
typedef interface KeyEvents KeyEvents;

#endif 	/* __KeyEvents_FWD_DEFINED__ */


#ifndef ___BrailleEvents_FWD_DEFINED__
#define ___BrailleEvents_FWD_DEFINED__
typedef interface _BrailleEvents _BrailleEvents;

#endif 	/* ___BrailleEvents_FWD_DEFINED__ */


#ifndef __BrailleEvents_FWD_DEFINED__
#define __BrailleEvents_FWD_DEFINED__
typedef interface BrailleEvents BrailleEvents;

#endif 	/* __BrailleEvents_FWD_DEFINED__ */


#ifndef ___BrailleDisplayEvents_FWD_DEFINED__
#define ___BrailleDisplayEvents_FWD_DEFINED__
typedef interface _BrailleDisplayEvents _BrailleDisplayEvents;

#endif 	/* ___BrailleDisplayEvents_FWD_DEFINED__ */


#ifndef __BrailleDisplayEvents_FWD_DEFINED__
#define __BrailleDisplayEvents_FWD_DEFINED__
typedef interface BrailleDisplayEvents BrailleDisplayEvents;

#endif 	/* __BrailleDisplayEvents_FWD_DEFINED__ */


#ifndef ___SpeechEvents_FWD_DEFINED__
#define ___SpeechEvents_FWD_DEFINED__
typedef interface _SpeechEvents _SpeechEvents;

#endif 	/* ___SpeechEvents_FWD_DEFINED__ */


#ifndef __SpeechEvents_FWD_DEFINED__
#define __SpeechEvents_FWD_DEFINED__
typedef interface SpeechEvents SpeechEvents;

#endif 	/* __SpeechEvents_FWD_DEFINED__ */


#ifndef ___MouseCursorEvents_FWD_DEFINED__
#define ___MouseCursorEvents_FWD_DEFINED__
typedef interface _MouseCursorEvents _MouseCursorEvents;

#endif 	/* ___MouseCursorEvents_FWD_DEFINED__ */


#ifndef __MouseCursorEvents_FWD_DEFINED__
#define __MouseCursorEvents_FWD_DEFINED__
typedef interface MouseCursorEvents MouseCursorEvents;

#endif 	/* __MouseCursorEvents_FWD_DEFINED__ */


#ifndef ___EditBox_FWD_DEFINED__
#define ___EditBox_FWD_DEFINED__
typedef interface _EditBox _EditBox;

#endif 	/* ___EditBox_FWD_DEFINED__ */


#ifndef ___Pushbutton_FWD_DEFINED__
#define ___Pushbutton_FWD_DEFINED__
typedef interface _Pushbutton _Pushbutton;

#endif 	/* ___Pushbutton_FWD_DEFINED__ */


#ifndef ___CheckBox_FWD_DEFINED__
#define ___CheckBox_FWD_DEFINED__
typedef interface _CheckBox _CheckBox;

#endif 	/* ___CheckBox_FWD_DEFINED__ */


#ifndef ___RadioButton_FWD_DEFINED__
#define ___RadioButton_FWD_DEFINED__
typedef interface _RadioButton _RadioButton;

#endif 	/* ___RadioButton_FWD_DEFINED__ */


#ifndef __GroupBox_FWD_DEFINED__
#define __GroupBox_FWD_DEFINED__
typedef interface GroupBox GroupBox;

#endif 	/* __GroupBox_FWD_DEFINED__ */


#ifndef __Static_FWD_DEFINED__
#define __Static_FWD_DEFINED__
typedef interface Static Static;

#endif 	/* __Static_FWD_DEFINED__ */


#ifndef ___ComboBox_FWD_DEFINED__
#define ___ComboBox_FWD_DEFINED__
typedef interface _ComboBox _ComboBox;

#endif 	/* ___ComboBox_FWD_DEFINED__ */


#ifndef ___ListBox_FWD_DEFINED__
#define ___ListBox_FWD_DEFINED__
typedef interface _ListBox _ListBox;

#endif 	/* ___ListBox_FWD_DEFINED__ */


#ifndef ___UpDown_FWD_DEFINED__
#define ___UpDown_FWD_DEFINED__
typedef interface _UpDown _UpDown;

#endif 	/* ___UpDown_FWD_DEFINED__ */


#ifndef __ProgressBar_FWD_DEFINED__
#define __ProgressBar_FWD_DEFINED__
typedef interface ProgressBar ProgressBar;

#endif 	/* __ProgressBar_FWD_DEFINED__ */


#ifndef ___TrackBar_FWD_DEFINED__
#define ___TrackBar_FWD_DEFINED__
typedef interface _TrackBar _TrackBar;

#endif 	/* ___TrackBar_FWD_DEFINED__ */


#ifndef __TabControl_FWD_DEFINED__
#define __TabControl_FWD_DEFINED__
typedef interface TabControl TabControl;

#endif 	/* __TabControl_FWD_DEFINED__ */


#ifndef __TreeViewEventInfo_FWD_DEFINED__
#define __TreeViewEventInfo_FWD_DEFINED__
typedef interface TreeViewEventInfo TreeViewEventInfo;

#endif 	/* __TreeViewEventInfo_FWD_DEFINED__ */


#ifndef __ListViewItem_FWD_DEFINED__
#define __ListViewItem_FWD_DEFINED__
typedef interface ListViewItem ListViewItem;

#endif 	/* __ListViewItem_FWD_DEFINED__ */


#ifndef __ListViewItems_FWD_DEFINED__
#define __ListViewItems_FWD_DEFINED__
typedef interface ListViewItems ListViewItems;

#endif 	/* __ListViewItems_FWD_DEFINED__ */


#ifndef __ListViewColumn_FWD_DEFINED__
#define __ListViewColumn_FWD_DEFINED__
typedef interface ListViewColumn ListViewColumn;

#endif 	/* __ListViewColumn_FWD_DEFINED__ */


#ifndef __ListViewColumns_FWD_DEFINED__
#define __ListViewColumns_FWD_DEFINED__
typedef interface ListViewColumns ListViewColumns;

#endif 	/* __ListViewColumns_FWD_DEFINED__ */


#ifndef __ListView_FWD_DEFINED__
#define __ListView_FWD_DEFINED__
typedef interface ListView ListView;

#endif 	/* __ListView_FWD_DEFINED__ */


#ifndef __UpDownEventInfo_FWD_DEFINED__
#define __UpDownEventInfo_FWD_DEFINED__
typedef interface UpDownEventInfo UpDownEventInfo;

#endif 	/* __UpDownEventInfo_FWD_DEFINED__ */


#ifndef ___SynthesizersEvents_FWD_DEFINED__
#define ___SynthesizersEvents_FWD_DEFINED__
typedef interface _SynthesizersEvents _SynthesizersEvents;

#endif 	/* ___SynthesizersEvents_FWD_DEFINED__ */


#ifndef __SynthesizersEvents_FWD_DEFINED__
#define __SynthesizersEvents_FWD_DEFINED__
typedef interface SynthesizersEvents SynthesizersEvents;

#endif 	/* __SynthesizersEvents_FWD_DEFINED__ */


#ifndef ____MSAAEvents_FWD_DEFINED__
#define ____MSAAEvents_FWD_DEFINED__
typedef interface __MSAAEvents __MSAAEvents;

#endif 	/* ____MSAAEvents_FWD_DEFINED__ */


#ifndef __MSAAEvents_FWD_DEFINED__
#define __MSAAEvents_FWD_DEFINED__
typedef interface MSAAEvents MSAAEvents;

#endif 	/* __MSAAEvents_FWD_DEFINED__ */


#ifndef ___BrowseModeEvents_FWD_DEFINED__
#define ___BrowseModeEvents_FWD_DEFINED__
typedef interface _BrowseModeEvents _BrowseModeEvents;

#endif 	/* ___BrowseModeEvents_FWD_DEFINED__ */


#ifndef __BrowseModeEvents_FWD_DEFINED__
#define __BrowseModeEvents_FWD_DEFINED__
typedef interface BrowseModeEvents BrowseModeEvents;

#endif 	/* __BrowseModeEvents_FWD_DEFINED__ */


#ifndef __BrailleDisplays_FWD_DEFINED__
#define __BrailleDisplays_FWD_DEFINED__

#ifdef __cplusplus
typedef class BrailleDisplays BrailleDisplays;
#else
typedef struct BrailleDisplays BrailleDisplays;
#endif /* __cplusplus */

#endif 	/* __BrailleDisplays_FWD_DEFINED__ */


#ifndef __Application_FWD_DEFINED__
#define __Application_FWD_DEFINED__

#ifdef __cplusplus
typedef class Application Application;
#else
typedef struct Application Application;
#endif /* __cplusplus */

#endif 	/* __Application_FWD_DEFINED__ */


#ifndef __Keyboard_FWD_DEFINED__
#define __Keyboard_FWD_DEFINED__

#ifdef __cplusplus
typedef class Keyboard Keyboard;
#else
typedef struct Keyboard Keyboard;
#endif /* __cplusplus */

#endif 	/* __Keyboard_FWD_DEFINED__ */


#ifndef __Key_FWD_DEFINED__
#define __Key_FWD_DEFINED__

#ifdef __cplusplus
typedef class Key Key;
#else
typedef struct Key Key;
#endif /* __cplusplus */

#endif 	/* __Key_FWD_DEFINED__ */


#ifndef __Script_FWD_DEFINED__
#define __Script_FWD_DEFINED__

#ifdef __cplusplus
typedef class Script Script;
#else
typedef struct Script Script;
#endif /* __cplusplus */

#endif 	/* __Script_FWD_DEFINED__ */


#ifndef __Window_FWD_DEFINED__
#define __Window_FWD_DEFINED__

#ifdef __cplusplus
typedef class Window Window;
#else
typedef struct Window Window;
#endif /* __cplusplus */

#endif 	/* __Window_FWD_DEFINED__ */


#ifndef __ClientInformation_FWD_DEFINED__
#define __ClientInformation_FWD_DEFINED__

#ifdef __cplusplus
typedef class ClientInformation ClientInformation;
#else
typedef struct ClientInformation ClientInformation;
#endif /* __cplusplus */

#endif 	/* __ClientInformation_FWD_DEFINED__ */


#ifndef __Mouse_FWD_DEFINED__
#define __Mouse_FWD_DEFINED__

#ifdef __cplusplus
typedef class Mouse Mouse;
#else
typedef struct Mouse Mouse;
#endif /* __cplusplus */

#endif 	/* __Mouse_FWD_DEFINED__ */


#ifndef __Dialog_FWD_DEFINED__
#define __Dialog_FWD_DEFINED__

#ifdef __cplusplus
typedef class Dialog Dialog;
#else
typedef struct Dialog Dialog;
#endif /* __cplusplus */

#endif 	/* __Dialog_FWD_DEFINED__ */


#ifndef __MousePointer_FWD_DEFINED__
#define __MousePointer_FWD_DEFINED__

#ifdef __cplusplus
typedef class MousePointer MousePointer;
#else
typedef struct MousePointer MousePointer;
#endif /* __cplusplus */

#endif 	/* __MousePointer_FWD_DEFINED__ */


#ifndef __Control_FWD_DEFINED__
#define __Control_FWD_DEFINED__

#ifdef __cplusplus
typedef class Control Control;
#else
typedef struct Control Control;
#endif /* __cplusplus */

#endif 	/* __Control_FWD_DEFINED__ */


#ifndef __EditBox_FWD_DEFINED__
#define __EditBox_FWD_DEFINED__

#ifdef __cplusplus
typedef class EditBox EditBox;
#else
typedef struct EditBox EditBox;
#endif /* __cplusplus */

#endif 	/* __EditBox_FWD_DEFINED__ */


#ifndef __Pushbutton_FWD_DEFINED__
#define __Pushbutton_FWD_DEFINED__

#ifdef __cplusplus
typedef class Pushbutton Pushbutton;
#else
typedef struct Pushbutton Pushbutton;
#endif /* __cplusplus */

#endif 	/* __Pushbutton_FWD_DEFINED__ */


#ifndef __CheckBox_FWD_DEFINED__
#define __CheckBox_FWD_DEFINED__

#ifdef __cplusplus
typedef class CheckBox CheckBox;
#else
typedef struct CheckBox CheckBox;
#endif /* __cplusplus */

#endif 	/* __CheckBox_FWD_DEFINED__ */


#ifndef __RadioButton_FWD_DEFINED__
#define __RadioButton_FWD_DEFINED__

#ifdef __cplusplus
typedef class RadioButton RadioButton;
#else
typedef struct RadioButton RadioButton;
#endif /* __cplusplus */

#endif 	/* __RadioButton_FWD_DEFINED__ */


#ifndef __ComboBox_FWD_DEFINED__
#define __ComboBox_FWD_DEFINED__

#ifdef __cplusplus
typedef class ComboBox ComboBox;
#else
typedef struct ComboBox ComboBox;
#endif /* __cplusplus */

#endif 	/* __ComboBox_FWD_DEFINED__ */


#ifndef __ListBox_FWD_DEFINED__
#define __ListBox_FWD_DEFINED__

#ifdef __cplusplus
typedef class ListBox ListBox;
#else
typedef struct ListBox ListBox;
#endif /* __cplusplus */

#endif 	/* __ListBox_FWD_DEFINED__ */


#ifndef __UpDown_FWD_DEFINED__
#define __UpDown_FWD_DEFINED__

#ifdef __cplusplus
typedef class UpDown UpDown;
#else
typedef struct UpDown UpDown;
#endif /* __cplusplus */

#endif 	/* __UpDown_FWD_DEFINED__ */


#ifndef __TrackBar_FWD_DEFINED__
#define __TrackBar_FWD_DEFINED__

#ifdef __cplusplus
typedef class TrackBar TrackBar;
#else
typedef struct TrackBar TrackBar;
#endif /* __cplusplus */

#endif 	/* __TrackBar_FWD_DEFINED__ */


#ifndef __Synthesizer_FWD_DEFINED__
#define __Synthesizer_FWD_DEFINED__

#ifdef __cplusplus
typedef class Synthesizer Synthesizer;
#else
typedef struct Synthesizer Synthesizer;
#endif /* __cplusplus */

#endif 	/* __Synthesizer_FWD_DEFINED__ */


#ifndef __Speech_FWD_DEFINED__
#define __Speech_FWD_DEFINED__

#ifdef __cplusplus
typedef class Speech Speech;
#else
typedef struct Speech Speech;
#endif /* __cplusplus */

#endif 	/* __Speech_FWD_DEFINED__ */


#ifndef __MSAAEventSource_FWD_DEFINED__
#define __MSAAEventSource_FWD_DEFINED__

#ifdef __cplusplus
typedef class MSAAEventSource MSAAEventSource;
#else
typedef struct MSAAEventSource MSAAEventSource;
#endif /* __cplusplus */

#endif 	/* __MSAAEventSource_FWD_DEFINED__ */


#ifndef __BrowseMode_FWD_DEFINED__
#define __BrowseMode_FWD_DEFINED__

#ifdef __cplusplus
typedef class BrowseMode BrowseMode;
#else
typedef struct BrowseMode BrowseMode;
#endif /* __cplusplus */

#endif 	/* __BrowseMode_FWD_DEFINED__ */


#ifndef __SharedObjects_FWD_DEFINED__
#define __SharedObjects_FWD_DEFINED__

#ifdef __cplusplus
typedef class SharedObjects SharedObjects;
#else
typedef struct SharedObjects SharedObjects;
#endif /* __cplusplus */

#endif 	/* __SharedObjects_FWD_DEFINED__ */


#ifndef __brailleDisplay_FWD_DEFINED__
#define __brailleDisplay_FWD_DEFINED__

#ifdef __cplusplus
typedef class brailleDisplay brailleDisplay;
#else
typedef struct brailleDisplay brailleDisplay;
#endif /* __cplusplus */

#endif 	/* __brailleDisplay_FWD_DEFINED__ */


#ifndef __Braille_FWD_DEFINED__
#define __Braille_FWD_DEFINED__

#ifdef __cplusplus
typedef class Braille Braille;
#else
typedef struct Braille Braille;
#endif /* __cplusplus */

#endif 	/* __Braille_FWD_DEFINED__ */


#ifndef __Synthesizers_FWD_DEFINED__
#define __Synthesizers_FWD_DEFINED__

#ifdef __cplusplus
typedef class Synthesizers Synthesizers;
#else
typedef struct Synthesizers Synthesizers;
#endif /* __cplusplus */

#endif 	/* __Synthesizers_FWD_DEFINED__ */


#ifdef __cplusplus
extern "C"{
#endif 



#ifndef __WindowEyes_LIBRARY_DEFINED__
#define __WindowEyes_LIBRARY_DEFINED__

/* library WindowEyes */
/* [custom][custom][custom][helpcontext][helpfile][helpstring][version][uuid] */ 
































































































































































































































































typedef /* [helpcontext] */ 
enum WindowEyesScriptingErrorCodes
    {
        WE_E_BADCALLBACK	= 0x8004e001,
        WE_E_BADDICTIONARY	= 0x8004e002,
        WE_E_DIALOGXMLPARSE	= 0x8004e003,
        WE_E_CURSORKEYASSIGNACTION	= 0x8004d004,
        WE_E_DUPLICATESHAREDOBJECT	= 0x8004d005,
        WE_E_CANONLYAPPENDTEXTTOTEXT	= 0x8004d006,
        WE_E_CANONLYBECALLEDFROMANOVERLAP	= 0x8004d007,
        WE_E_PORTTOOLONG	= 0x8004d008,
        WE_E_FREEZETOOMUCH	= 0x8004d009,
        WE_E_NOTWITHNEWTVITEM	= 0x8004d00b,
        WE_E_DIALOGCIRCULARCLASSES	= 0x8004d00c,
        WE_E_INTERNAL_OVERLAP	= 0x8004d00d,
        WE_E_XLMPARSENOLANGUAGE	= 0x8004d00e,
        WE_E_BRAILLEOFF	= 0x8004d00f,
        WE_E_NOTCURRENTDISPLAY	= 0x8004d010,
        WE_E_BRAILLEOBJECTINVALID	= 0x8004d011,
        WE_E_NODICTIONARY	= 0x8004d012,
        WE_E_NOBRAILLETRANSLATIONTABLE	= 0x8004d013,
        WE_E_NOTIDENTIFIED	= 0x8004e004,
        WE_E_ALREADYIDENTIFIED	= 0x8004e005,
        WE_E_INVALIDDOTPATTERN	= 0x8004e006,
        WE_E_CONTAINSDUPLICATEDOTS	= 0x8004e007,
        WE_E_SPEECHBOXMODEEXCEEDSCHARACTERLIMIT	= 0x8004e008,
        WE_E_WECOMISMISSING	= 0x8004e009,
        WE_E_CANTSAVESET	= 0x8004e00a,
        WE_E_NOGLOBALPROPERTY	= 0x8004e00b
    } 	WindowEyesScriptingErrorCodes;

typedef /* [helpcontext] */ 
enum BlankLinesMode
    {
        blmOff	= 0,
        blmSayBlank	= 1,
        blmReadToEndOff	= 2
    } 	BlankLinesMode;

typedef /* [helpcontext] */ 
enum CapitalizationMode
    {
        cmOff	= 0,
        cmSayCaps	= 1,
        cmPitch	= 2
    } 	CapitalizationMode;

typedef /* [helpcontext] */ 
enum KeyboardInterruptabilityMode
    {
        kimAllKeys	= 0,
        kimCursorKeysOff	= 1,
        kimCursorKeysOnly	= 2,
        kimControlAlt	= 3,
        kimOff	= 4
    } 	KeyboardInterruptabilityMode;

typedef /* [helpcontext] */ 
enum KeyboardVoiceMode
    {
        kvmOff	= 0,
        kvmCharacters	= 1,
        kvmWords	= 2,
        kvmWordsNumbers	= 3,
        kvmCharactersWords	= 4,
        kvmCharactersWordsNumbers	= 5,
        kvmWordsInterrupt	= 6,
        kvmWordsNumbersInterrupt	= 7
    } 	KeyboardVoiceMode;

typedef /* [helpcontext] */ 
enum LineFilterMode
    {
        lfmOff	= 0,
        lfmOn	= 1,
        lfmReadToEnd	= 2
    } 	LineFilterMode;

typedef /* [helpcontext] */ 
enum MixedCaseMode
    {
        mcmOff	= 0,
        mcmOn	= 1,
        mcmOnWithAcronyms	= 2
    } 	MixedCaseMode;

typedef /* [helpcontext] */ 
enum NumberMode
    {
        nmOff	= 0,
        nmOn	= 1,
        nmSynthesizer	= 2
    } 	NumberMode;

typedef /* [helpcontext] */ 
enum AutoMode
    {
        amOff	= 0,
        amAuto	= 1,
        amOn	= 2
    } 	AutoMode;

typedef /* [helpcontext] */ 
enum PunctuationMode
    {
        pmOff	= 0,
        pmDictionaryOnly	= 1,
        pmOn	= 2
    } 	PunctuationMode;

typedef /* [helpcontext] */ 
enum RepeatMode
    {
        rmOff	= 0,
        rmSayRepeats	= 1,
        rmIgnore	= 2
    } 	RepeatMode;

typedef /* [helpcontext] */ 
enum MouseVoiceMode
    {
        mvmOff	= 0,
        mvmOnPointerIdOff	= 1,
        mvmOnPointerIdDelayed	= 2,
        mvmOnPointerIdDelayedDictionaryOnly	= 3,
        mvmOnPointerIdImmediate	= 4,
        mvmOnPointerIdImmediateDictionaryOnly	= 5
    } 	MouseVoiceMode;

typedef /* [helpcontext] */ 
enum MouseInterruptabilityMode
    {
        mimMove	= 0,
        mimClick	= 1,
        mimMoveClick	= 2,
        mimOff	= 3
    } 	MouseInterruptabilityMode;

typedef /* [helpcontext] */ 
enum MouseBoundaryMode
    {
        mbmFullScreen	= 0,
        mbmActiveWindow	= 1,
        mbmFocusedWindow	= 2,
        mbmUserWindow	= 3
    } 	MouseBoundaryMode;

typedef /* [helpcontext] */ 
enum MouseUpDownMode
    {
        mudmCharacter	= 0,
        mudmWord	= 1,
        mudmLine	= 2,
        mudBeginningOfLineToMouse	= 3,
        mudFromMouseToEndOfLine	= 4
    } 	MouseUpDownMode;

typedef /* [helpcontext] */ 
enum HighlightTextMode
    {
        htmAuto	= 0,
        htmHighlight	= 1
    } 	HighlightTextMode;

typedef /* [helpcontext] */ 
enum GraphicDictionaryMode
    {
        gdmOff	= 0,
        gdmDictionaryOnly	= 1,
        gdmOn	= 2,
        gdmGraphicOnly	= 3
    } 	GraphicDictionaryMode;

typedef /* [helpcontext] */ 
enum ColorDictionaryMode
    {
        cdmOff	= 0,
        cdmExact	= 1,
        cmdExactOrClosest	= 2
    } 	ColorDictionaryMode;

typedef /* [helpcontext] */ 
enum MenuMode
    {
        mmBeginner	= 0,
        mmIntermediate	= 1,
        mmAdvanced	= 2
    } 	MenuMode;

typedef /* [helpcontext] */ 
enum GlobalMode
    {
        gmVoiceVerbosity	= 0,
        gmOff	= 1,
        gmVoice	= 2,
        gmVerbosity	= 3
    } 	GlobalMode;

typedef /* [helpcontext] */ 
enum HyperactiveMode
    {
        hmOff	= 0,
        hmMessagesOff	= 1,
        hmMessagesOn	= 2
    } 	HyperactiveMode;

typedef /* [helpcontext] */ 
enum AutoLoadMode
    {
        almOff	= 0,
        almMessagesOff	= 1,
        almMessagesOn	= 2
    } 	AutoLoadMode;

typedef /* [helpcontext] */ 
enum BrailleAttributeMarking
    {
        bamHighlight	= 0,
        bamBold	= 1,
        bamUnderline	= 2,
        bamItalic	= 3,
        bamStrikeout	= 4,
        bamAll	= 5,
        bamNone	= 6
    } 	BrailleAttributeMarking;

typedef /* [helpcontext] */ 
enum BrailleQuickMessageMode
    {
        bqmDisabled	= 0,
        bqmOnWithNoTimeout	= 1,
        bqmOnWithTimeout	= 2
    } 	BrailleQuickMessageMode;

typedef /* [helpcontext] */ 
enum BrailleGraphicsVerbosity
    {
        bgvNone	= 0,
        bgvBraille	= 1,
        bgvBrailleSpeech	= 2,
        bgvGraphic	= 3,
        bgvBrailleGraphic	= 4,
        bgvBrailleSpeechGraphic	= 5
    } 	BrailleGraphicsVerbosity;

typedef /* [helpcontext] */ 
enum BrailleInformationOrder
    {
        bioTypeNameData	= 0,
        bioTypeDataName	= 1,
        bioNameDataType	= 2,
        bioNameTypeData	= 3,
        bioDataTypeName	= 4,
        bioDataNameType	= 5
    } 	BrailleInformationOrder;

typedef /* [helpcontext] */ 
enum Language
    {
        langEnglish	= 0,
        langPolish	= 1,
        langGerman	= 2,
        langCroatian	= 3,
        langFrench	= 4,
        langItalian	= 5,
        langCzech	= 6,
        langNorwegian	= 7,
        langPortuguese	= 8,
        langDutch	= 9,
        langTurkish	= 10,
        langKorean	= 11,
        langSwedish	= 12,
        langSpanish	= 13,
        langRussian	= 14,
        langArabic	= 15,
        langHungarian	= 16,
        langHebrew	= 17
    } 	Language;

typedef /* [helpcontext] */ 
enum UserWindowStatus
    {
        uwsNeutral	= 0,
        uwsSilent	= 1,
        uwsSpeak	= 2,
        uwsFloat	= 3
    } 	UserWindowStatus;

typedef /* [helpcontext] */ 
enum HyperWindowCommand
    {
        hwcUndefined	= 0,
        hwcSpeakWindow	= 1,
        hwcSpeakString	= 2,
        hwcSpeakHighlight	= 3,
        hwcActivateWindow	= 4,
        hwcLoadSet	= 5,
        hwcExecuteHotKey	= 6,
        hwcBeep	= 7,
        hwcRouteMouseToWindow	= 8
    } 	HyperWindowCommand;

typedef /* [helpcontext] */ 
enum HyperWindowTrigger
    {
        hwtAnyChange	= 0,
        hwtContainsAttribute	= 1,
        hwtDoesNotContainAttribute	= 2,
        hwtContainsString	= 3,
        hwtDoesNotContainString	= 4,
        hwtContainsCursor	= 5,
        hwtDoesNotContainCursor	= 6,
        hwtContainsFocus	= 7,
        hwtDoesNotContainFocus	= 8
    } 	HyperWindowTrigger;

typedef /* [helpcontext] */ 
enum FloatWindowPosition
    {
        fwpConstant	= 0,
        fwpBeginningOfHighlight	= 1,
        fwpEndOfHighlight	= 2,
        fwpCursorPlusConstant	= 3,
        fwpCursorMinusConstant	= 4,
        fwpCursorPlusAttribute	= 5,
        fwpCursorMinusAttribute	= 6,
        fwpConstantPlusAttribute	= 7,
        fwpConstantMinusAttribute	= 8,
        fwpBeginningOfFocusWindow	= 9,
        fwpEndOfFocusWindow	= 10,
        fwpBeginningOfFocusWindowPlusConstant	= 11,
        fwpBeginningOfFocusWindowMinusConstant	= 12,
        fwpEndOfFocusWindowPlusConstant	= 13,
        fwpEndOfFocusWindowMinusConstant	= 14,
        fwpBeginningOfStringPlusConstant	= 15,
        fwpBeginningOfStringMinusConstant	= 16,
        fwpEndOfStringPlusConstant	= 17,
        fwpEndOfStringMinusConstant	= 18,
        fwpBeginningOfHighlightPlusConstant	= 19,
        fwpBeginningOfHighlightMinusConstant	= 20,
        fwpEndOfHighlightPlusConstant	= 21,
        fwpEndOfHighlightMinusConstant	= 22
    } 	FloatWindowPosition;

typedef /* [helpcontext] */ 
enum KeyFilterState
    {
        kfsEither	= 0,
        kfsDown	= 1,
        kfsUp	= 2
    } 	KeyFilterState;

typedef /* [helpcontext] */ 
enum KeyModifiers
    {
        kmNone	= 0,
        kmShift	= 1,
        kmControl	= 2,
        kmAlt	= 4,
        kmWindows	= 8,
        kmApplication	= 16,
        kmInsert	= 32,
        kmNumlock	= 64,
        kmAltGR	= 128,
        kmNumpad	= 256
    } 	KeyModifiers;

typedef /* [helpcontext] */ 
enum VerbosityShortcut
    {
        scNone	= 0,
        scBeforeOption	= 1,
        scAfterOption	= 2
    } 	VerbosityShortcut;

typedef /* [helpcontext] */ 
enum VerbosityListviewHeaders
    {
        lvhNone	= 0,
        lvhAllButFirstColumn	= 1,
        lvhAll	= 2
    } 	VerbosityListviewHeaders;

typedef /* [helpcontext] */ 
enum VerbosityMeasurementUnits
    {
        muInches	= 0,
        muCentimeters	= 1,
        muMillimeters	= 2,
        muPoints	= 3
    } 	VerbosityMeasurementUnits;

typedef /* [helpcontext] */ 
enum ClipType
    {
        ctUnknown	= 0,
        ctText	= 1,
        ctBitmap	= 2,
        ctFakeText	= 3
    } 	ClipType;

typedef /* [helpcontext] */ 
enum FontStyle
    {
        fsPlain	= 0,
        fsItalic	= 1,
        fsStruckout	= 2,
        fsUnderlined	= 4,
        fsBold	= 8,
        fsNofont	= 128
    } 	FontStyle;

typedef /* [helpcontext] */ 
enum WordPageVerbosity
    {
        wpgNone	= 0,
        wpgCurrent	= 1,
        wpgCurrentTotal	= 2
    } 	WordPageVerbosity;

typedef /* [helpcontext] */ 
enum PowerPointEffectsVerbosity
    {
        pefNone	= 0,
        pefSummary	= 1,
        pefSummaryDetails	= 2
    } 	PowerPointEffectsVerbosity;

typedef /* [helpcontext] */ 
enum ExcelPositionVerbosity
    {
        epNever	= 0,
        epAlways	= 1,
        epIfNoHeader	= 2
    } 	ExcelPositionVerbosity;

typedef /* [helpcontext] */ 
enum ExcelCountsVerbosity
    {
        ecWorksheet	= 0,
        ecWorkbook	= 1
    } 	ExcelCountsVerbosity;

typedef /* [helpcontext] */ 
enum BrowseModeAcronymIndication
    {
        baiNone	= 0,
        baiInclueType	= 1,
        baiExpand	= 2,
        baiReplace	= 3
    } 	BrowseModeAcronymIndication;

typedef /* [helpcontext] */ 
enum BrowseModeSuspendRefresh
    {
        bsrNone	= 0,
        bsrAll	= 1,
        bsrFlash	= 2
    } 	BrowseModeSuspendRefresh;

typedef /* [helpcontext] */ 
enum BrowseModeTableHeaders
    {
        bthOff	= 0,
        bthAttributeOnly	= 1,
        bthColumn	= 2,
        bthRow	= 3,
        bthColumnOrRow	= 4,
        bthColumnAndRow	= 5,
        bthRxCy	= 6
    } 	BrowseModeTableHeaders;

typedef /* [helpcontext] */ 
enum BrowseModeTableIndication
    {
        btiNone	= 0,
        btiAll	= 1,
        btiSmart	= 2
    } 	BrowseModeTableIndication;

typedef /* [helpcontext] */ 
enum BrowseModeLinkIndication
    {
        bliNone	= 0,
        bliBefore	= 1,
        bliAfter	= 2
    } 	BrowseModeLinkIndication;

typedef /* [helpcontext] */ 
enum BrowseModeNewPageRead
    {
        bnrFirst24	= 0,
        bnrEntire	= 1,
        bnrNone	= 2
    } 	BrowseModeNewPageRead;

typedef /* [helpcontext] */ 
enum TextAlignment
    {
        taLeft	= 0,
        taCenter	= 1,
        taRight	= 2
    } 	TextAlignment;

typedef /* [helpcontext] */ 
enum VerticalAlignment
    {
        vaTop	= 0,
        vaCenter	= 1,
        vaBottom	= 2
    } 	VerticalAlignment;

typedef /* [helpcontext] */ 
enum CheckState
    {
        csUnchecked	= 0,
        csChecked	= 1,
        csPartiallyChecked	= 2
    } 	CheckState;

typedef /* [helpcontext] */ 
enum ComboBoxType
    {
        cbtDropDown	= 0,
        cbtDropDownList	= 1,
        cbtSimple	= 2
    } 	ComboBoxType;

typedef /* [helpcontext] */ 
enum ListBoxType
    {
        lbtSingleSelect	= 0,
        lbtMultipleSelect	= 1,
        lbtExtendedSelect	= 2,
        lbtNoSelect	= 3
    } 	ListBoxType;

typedef /* [helpcontext] */ 
enum TickPlacement
    {
        tpNone	= 0,
        tpEnds	= 1,
        tpAuto	= 2
    } 	TickPlacement;

typedef /* [helpcontext] */ 
enum TickLocation
    {
        tlLeft	= 0,
        tlTop	= 1,
        tlRight	= 2,
        tlBottom	= 3,
        tlBoth	= 4
    } 	TickLocation;

typedef /* [helpcontext] */ 
enum VirtualKeyCode
    {
        vk_LBUTTON	= 1,
        vk_RBUTTON	= 2,
        vk_CANCEL	= 3,
        vk_MBUTTON	= 4,
        vk_XBUTTON1	= 5,
        vk_XBUTTON2	= 6,
        vk_BACK	= 8,
        vk_BACKSPACE	= 8,
        vk_TAB	= 9,
        vk_CLEAR	= 12,
        vk_RETURN	= 13,
        vk_ENTER	= 13,
        vk_SHIFT	= 16,
        vk_CONTROL	= 17,
        vk_MENU	= 18,
        vk_ALT	= 18,
        vk_PAUSE	= 19,
        vk_CAPITAL	= 20,
        vk_KANA	= 21,
        vk_HANGEUL	= 21,
        vk_HANGUL	= 21,
        vk_JUNJA	= 23,
        vk_FINAL	= 24,
        vk_HANJA	= 25,
        vk_KANJI	= 25,
        vk_ESCAPE	= 27,
        vk_CONVERT	= 28,
        vk_NONCONVERT	= 29,
        vk_ACCEPT	= 30,
        vk_MODECHANGE	= 31,
        vk_SPACE	= 32,
        vk_PRIOR	= 33,
        vk_PAGEUP	= 33,
        vk_NEXT	= 34,
        vk_PAGEDOWN	= 34,
        vk_END	= 35,
        vk_HOME	= 36,
        vk_LEFT	= 37,
        vk_UP	= 38,
        vk_RIGHT	= 39,
        vk_DOWN	= 40,
        vk_SELECT	= 41,
        vk_PRINT	= 42,
        vk_EXECUTE	= 43,
        vk_SNAPSHOT	= 44,
        vk_INSERT	= 45,
        vk_DELETE	= 46,
        vk_HELP	= 47,
        vk_0	= 48,
        vk_1	= 49,
        vk_2	= 50,
        vk_3	= 51,
        vk_4	= 52,
        vk_5	= 53,
        vk_6	= 54,
        vk_7	= 55,
        vk_8	= 56,
        vk_9	= 57,
        vk_A	= 65,
        vk_B	= 66,
        vk_C	= 67,
        vk_D	= 68,
        vk_E	= 69,
        vk_F	= 70,
        vk_G	= 71,
        vk_H	= 72,
        vk_I	= 73,
        vk_J	= 74,
        vk_K	= 75,
        vk_L	= 76,
        vk_M	= 77,
        vk_N	= 78,
        vk_O	= 79,
        vk_P	= 80,
        vk_Q	= 81,
        vk_R	= 82,
        vk_S	= 83,
        vk_T	= 84,
        vk_U	= 85,
        vk_V	= 86,
        vk_W	= 87,
        vk_X	= 88,
        vk_Y	= 89,
        vk_Z	= 90,
        vk_LWIN	= 91,
        vk_RWIN	= 92,
        vk_APPS	= 93,
        vk_SLEEP	= 95,
        vk_NUMPAD0	= 96,
        vk_NUMPAD1	= 97,
        vk_NUMPAD2	= 98,
        vk_NUMPAD3	= 99,
        vk_NUMPAD4	= 100,
        vk_NUMPAD5	= 101,
        vk_NUMPAD6	= 102,
        vk_NUMPAD7	= 103,
        vk_NUMPAD8	= 104,
        vk_NUMPAD9	= 105,
        vk_MULTIPLY	= 106,
        vk_ADD	= 107,
        vk_SEPARATOR	= 108,
        vk_SUBTRACT	= 109,
        vk_DECIMAL	= 110,
        vk_DIVIDE	= 111,
        vk_F1	= 112,
        vk_F2	= 113,
        vk_F3	= 114,
        vk_F4	= 115,
        vk_F5	= 116,
        vk_F6	= 117,
        vk_F7	= 118,
        vk_F8	= 119,
        vk_F9	= 120,
        vk_F10	= 121,
        vk_F11	= 122,
        vk_F12	= 123,
        vk_F13	= 124,
        vk_F14	= 125,
        vk_F15	= 126,
        vk_F16	= 127,
        vk_F17	= 128,
        vk_F18	= 129,
        vk_F19	= 130,
        vk_F20	= 131,
        vk_F21	= 132,
        vk_F22	= 133,
        vk_F23	= 134,
        vk_F24	= 135,
        vk_NUMLOCK	= 144,
        vk_SCROLL	= 145,
        vk_OEM_NEC_EQUAL	= 146,
        vk_OEM_FJ_JISHO	= 146,
        vk_OEM_FJ_MASSHOU	= 147,
        vk_OEM_FJ_TOUROKU	= 148,
        vk_OEM_FJ_LOYA	= 149,
        vk_OEM_FJ_ROYA	= 150,
        vk_LSHIFT	= 160,
        vk_RSHIFT	= 161,
        vk_LCONTROL	= 162,
        vk_RCONTROL	= 163,
        vk_LMENU	= 164,
        vk_RMENU	= 165,
        vk_LALT	= 164,
        vk_RALT	= 165,
        vk_BROWSER_BACK	= 166,
        vk_BROWSER_FORWARD	= 167,
        vk_BROWSER_REFRESH	= 168,
        vk_BROWSER_STOP	= 169,
        vk_BROWSER_SEARCH	= 170,
        vk_BROWSER_FAVORITES	= 171,
        vk_BROWSER_HOME	= 172,
        vk_VOLUME_MUTE	= 173,
        vk_VOLUME_DOWN	= 174,
        vk_VOLUME_UP	= 175,
        vk_MEDIA_NEXT_TRACK	= 176,
        vk_MEDIA_PREV_TRACK	= 177,
        vk_MEDIA_STOP	= 178,
        vk_MEDIA_PLAY_PAUSE	= 179,
        vk_LAUNCH_MAIL	= 180,
        vk_LAUNCH_MEDIA_SELECT	= 181,
        vk_LAUNCH_APP1	= 182,
        vk_LAUNCH_APP2	= 183,
        vk_OEM_1	= 186,
        vk_SEMICOLON	= 186,
        vk_OEM_PLUS	= 187,
        vk_PLUS	= 187,
        vk_OEM_COMMA	= 188,
        vk_COMMA	= 188,
        vk_OEM_MINUS	= 189,
        vk_MINUS	= 189,
        vk_OEM_PERIOD	= 190,
        vk_PERIOD	= 190,
        vk_OEM_2	= 191,
        vk_SLASH	= 191,
        vk_OEM_3	= 192,
        vk_BACKQUOTE	= 192,
        vk_GRAVEACCENT	= 192,
        vk_OEM_4	= 219,
        vk_LEFTBRACKET	= 219,
        vk_OEM_5	= 220,
        vk_BACKSLASH	= 220,
        vk_OEM_6	= 221,
        vk_RIGHTBRACKET	= 221,
        vk_OEM_7	= 222,
        vk_QUOTE	= 222,
        vk_OEM_8	= 223,
        vk_OEM_AX	= 225,
        vk_OEM_102	= 226,
        vk_ICO_HELP	= 227,
        vk_ICO_00	= 228,
        vk_PROCESSKEY	= 229,
        vk_ICO_CLEAR	= 230,
        vk_PACKET	= 231,
        vk_OEM_RESET	= 233,
        vk_OEM_JUMP	= 234,
        vk_OEM_PA1	= 235,
        vk_OEM_PA2	= 236,
        vk_OEM_PA3	= 237,
        vk_OEM_WSCTRL	= 238,
        vk_OEM_CUSEL	= 239,
        vk_OEM_ATTN	= 240,
        vk_OEM_FINISH	= 241,
        vk_OEM_COPY	= 242,
        vk_OEM_AUTO	= 243,
        vk_OEM_ENLW	= 244,
        vk_OEM_BACKTAB	= 245,
        vk_ATTN	= 246,
        vk_CRSEL	= 247,
        vk_EXSEL	= 248,
        vk_EREOF	= 249,
        vk_PLAY	= 250,
        vk_ZOOM	= 251,
        vk_NONAME	= 252,
        vk_PA1	= 253,
        vk_OEM_CLEAR	= 254
    } 	VirtualKeyCode;

typedef /* [helpcontext] */ 
enum WindowType
    {
        wtNotWindow	= 0,
        wtOriginalClass	= 0,
        wtOther	= 1,
        wtDialog	= 2,
        wtPropertySheet	= 3,
        wtMenu	= 4,
        wtMDIChild	= 5,
        wtMDIClient	= 6,
        wtMDIFrame	= 7,
        wtButton	= 8,
        wtCheckBox	= 9,
        wtComboBox	= 10,
        wtCustomControl	= 11,
        wtEdit	= 12,
        wtGroupBox	= 13,
        wtIconText	= 14,
        wtListBox	= 15,
        wtRadioButton	= 16,
        wtScrollBar	= 17,
        wtStatic	= 18,
        wtTab	= 19,
        wtTrackBar	= 20,
        wtListView	= 21,
        wtTreeView	= 22,
        wtHotkey	= 23,
        wtRichEdit	= 24,
        wtAnimate	= 25,
        wtProgress	= 26,
        wtStatus	= 27,
        wtUpDown	= 28,
        wtVoid	= 29,
        wtToolTip	= 31
    } 	WindowType;

typedef /* [helpcontext] */ 
enum WindowStatus
    {
        wsMinimized	= 0,
        wsMaximized	= 1,
        wsNormal	= 2,
        wsRestored	= 3
    } 	WindowStatus;

typedef /* [helpcontext] */ 
enum HotkeyID
    {
        hkNone	= 0,
        hkPriorCharacter	= 1,
        hkCharacter	= 2,
        hkNextCharacter	= 3,
        hkPriorWord	= 4,
        hkWord	= 5,
        hkNextWord	= 6,
        hkPriorLine	= 7,
        hkLine	= 8,
        hkNextLine	= 9,
        hkPriorSentence	= 10,
        hkSentence	= 11,
        hkNextSentence	= 12,
        hkPriorParagraph	= 13,
        hkParagraph	= 14,
        hkNextParagraph	= 15,
        hkTopToCursor	= 16,
        hkCursorToBottom	= 17,
        hkLeftToCursor	= 18,
        hkCursorToRight	= 19,
        hkUserWindow0	= 20,
        hkUserWindow1	= 21,
        hkUserWindow2	= 22,
        hkUserWindow3	= 23,
        hkUserWindow4	= 24,
        hkUserWindow5	= 25,
        hkUserWindow6	= 26,
        hkUserWindow7	= 27,
        hkUserWindow8	= 28,
        hkUserWindow9	= 29,
        hkUserWindow10	= 30,
        hkUserWindow11	= 31,
        hkUserWindow12	= 32,
        hkUserWindow13	= 33,
        hkUserWindow14	= 34,
        hkUserWindow15	= 35,
        hkUserWindow16	= 36,
        hkUserWindow17	= 37,
        hkUserWindow18	= 38,
        hkUserWindow19	= 39,
        hkAnyWindow	= 40,
        hkActiveUserWindow	= 41,
        hkReadToEnd	= 42,
        hkUnicodeAndAttributes	= 43,
        hkFieldData	= 44,
        hkFieldName	= 45,
        hkLabelFieldName	= 46,
        hkSpeakSummary	= 47,
        hkDefaultButton	= 48,
        hkAnyLine	= 49,
        hkHighlight	= 50,
        hkSetHighlightColor	= 51,
        hkHighlightRotor	= 52,
        hkTimeAndDate	= 53,
        hkPosition	= 54,
        hkSilence	= 55,
        hkVerbosity	= 56,
        hkMenu	= 57,
        hkBypass	= 58,
        hkAttributeChangeToggle	= 59,
        hkMouseBoundaryRotor	= 60,
        hkSingleClickLeftButton	= 61,
        hkSingleClickMiddleButton	= 62,
        hkSingleClickRightButton	= 63,
        hkDoubleClickLeftButton	= 64,
        hkDoubleClickMiddleButton	= 65,
        hkDoubleClickRightButton	= 66,
        hkToggleLeftButton	= 67,
        hkToggleMiddleButton	= 68,
        hkToggleRightButton	= 69,
        hkDragAndDrop	= 70,
        hkMouseHorizontalLock	= 71,
        hkMouseVerticalLock	= 72,
        hkWEMouseToggle	= 73,
        hkMouseUp	= 74,
        hkMouseDown	= 75,
        hkMouseLeft	= 76,
        hkMouseRight	= 77,
        hkMouseTopLeft	= 78,
        hkMouseTopRight	= 79,
        hkMouseBottomLeft	= 80,
        hkMouseBottomRight	= 81,
        hkMousePriorCharacter	= 82,
        hkMouseCharacter	= 83,
        hkMouseNextCharacter	= 84,
        hkMousePriorWord	= 85,
        hkMouseWord	= 86,
        hkMouseNextWord	= 87,
        hkMousePriorLine	= 88,
        hkMouseLine	= 89,
        hkMouseNextLine	= 90,
        hkMousePriorSentence	= 91,
        hkMouseSentence	= 92,
        hkMouseNextSentence	= 93,
        hkMousePriorParagraph	= 94,
        hkMouseParagraph	= 95,
        hkMouseNextParagraph	= 96,
        hkMousePriorIcon	= 97,
        hkMouseIcon	= 98,
        hkMouseNextIcon	= 99,
        hkMousePriorClip	= 100,
        hkMouseClip	= 101,
        hkMouseNextClip	= 102,
        hkTopToMouse	= 103,
        hkMouseToBottom	= 104,
        hkLeftToMouse	= 105,
        hkMouseToRight	= 106,
        hkMouseBeginningOfLine	= 107,
        hkMouseEndOfLine	= 108,
        hkMouseUpDownSpeak	= 109,
        hkMouseToFocus	= 110,
        hkCursorToMouse	= 111,
        hkSyncWeAndMouse	= 112,
        hkRouteToSpecifiedHighlight	= 113,
        hkRouteMouseToWindow	= 114,
        hkMousePosition	= 115,
        hkMouseUnicodeAndAttributes	= 116,
        hkMouseSearch	= 117,
        hkMouseContinueSearch	= 118,
        hkMousePriorAttribute	= 119,
        hkMouseNextAttribute	= 120,
        hkMousePriorControl	= 121,
        hkMouseNextControl	= 122,
        hkDescribePointer	= 123,
        hkCapturePointer	= 124,
        hkEditDictionaries	= 125,
        hkSpeakAllToggle	= 126,
        hkToggleAllVoicesApp	= 127,
        hkToggleAllVoices	= 128,
        hkEntireScreen	= 129,
        hkReadActiveWindow	= 130,
        hkTitleApp	= 131,
        hkFocusTitle	= 132,
        hkStatusLine	= 133,
        hkLastFlashAndTooltip	= 134,
        hkProgressOrScrollbar	= 135,
        hkReadMenuBar	= 136,
        hkHighlightedBlock	= 137,
        hkKeyDescriber	= 138,
        hkContextSensitiveHelp	= 139,
        hkVoiceDown	= 140,
        hkVoiceUp	= 141,
        hkVoiceSelectionRotor	= 142,
        hkVoiceParameterRotor	= 143,
        hkNumbersRotor	= 144,
        hkSignalCursorPositionToggle	= 145,
        hkWordMode	= 146,
        hkLineFilterStatusRotor	= 147,
        hkCapitalizationAlertRotor	= 148,
        hkFormatAlertToggle	= 149,
        hkSetFormatAlertOffset	= 150,
        hkSelectUserWindow	= 151,
        hkUserWindowCoordinates	= 152,
        hkUserWindowTopLeft	= 153,
        hkUserWindowBottomRight	= 154,
        hkUserWindowDefinition	= 155,
        hkUserWindowStatus	= 156,
        hkReadUserWindowCoordinates	= 157,
        hkSetUserWindowColor	= 158,
        hkSelectHyperactiveWindow	= 159,
        hkHyperactiveWindowCoordinates	= 160,
        hkHyperactiveWindowTopLeft	= 161,
        hkHyperactiveWindowBottomRight	= 162,
        hkHyperactiveWindowDefinition	= 163,
        hkReadHyperactiveWindowCoordinates	= 164,
        hkHyperactiveWindowsRotor	= 165,
        hkAutodetectCursor	= 166,
        hkAutoLabelGraphics	= 167,
        hkBrowseMode	= 168,
        hkSwitchLanguage	= 169,
        hkSystemTray	= 170,
        hkReclass	= 171,
        hkRedraw	= 172,
        hkLoadSet	= 173,
        hkSaveSet	= 174,
        hkReloadScript	= 175,
        hkDumpOSM	= 176,
        hkLinkNext	= 177,
        hkLinkPrior	= 178,
        hkControlNext	= 179,
        hkControlPrior	= 180,
        hkSpecifiedControlNext	= 181,
        hkSpecifiedControlPrior	= 182,
        hkTextNext	= 183,
        hkTextPrior	= 184,
        hkGraphicNext	= 185,
        hkGraphicPrior	= 186,
        hkFrameNext	= 187,
        hkFramePrior	= 188,
        hkDifferentLineNext	= 189,
        hkDifferentLinePrior	= 190,
        hkVisitedLinkNext	= 191,
        hkVisitedLinkPrior	= 192,
        hkAnchorNext	= 193,
        hkAnchorPrior	= 194,
        hkFieldsetNext	= 195,
        hkFieldSetPrior	= 196,
        hkFormNext	= 197,
        hkFormPrior	= 198,
        hkHeadingNext	= 199,
        hkHeadingPrior	= 200,
        hkParagraphNext	= 201,
        hkParagraphPrior	= 202,
        hkBlockQuoteNext	= 203,
        hkBlockQuotePrior	= 204,
        hkListNext	= 205,
        hkListPrior	= 206,
        hkListItemNext	= 207,
        hkListItemPrior	= 208,
        hkTableNext	= 209,
        hkTablePrior	= 210,
        hkEnterTableMode	= 211,
        hkExitTableMode	= 212,
        hkExitAllTableModes	= 213,
        hkBrowseModeCellRight	= 214,
        hkBrowseModeCellLeft	= 215,
        hkBrowseModeCellUp	= 216,
        hkBrowseModeCellDown	= 217,
        hkEndOfRow	= 218,
        hkBeginningOfRow	= 219,
        hkTopOfColumn	= 220,
        hkBottomOfColumn	= 221,
        hkHeaderRotor	= 222,
        hkBrowseModeElementProperties	= 223,
        hkPlacemarkerDialog	= 224,
        hkPlacemarkerNext	= 225,
        hkPlacemarkerPrior	= 226,
        hkPlacemarkerHotkey	= 227,
        hkBrowseModePageNavigation	= 228,
        hkPreviousPosition	= 229,
        hkAutoLoad	= 230,
        hkRefreshRotor	= 231,
        hkLineNext	= 232,
        hkLinePrior	= 233,
        hkSpellingOrGrammarNext	= 234,
        hkSpellingOrGrammarPrior	= 235,
        hkRevisionNext	= 236,
        hkRevisionPrior	= 237,
        hkTableCellRight	= 238,
        hkTableCellLeft	= 239,
        hkTableCellUp	= 240,
        hkTableCellDown	= 241,
        hkToFirstCellOfColumn	= 242,
        hkToLastCellOfColumn	= 243,
        hkToFirstCellOfRow	= 244,
        hkToLastCellOfRow	= 245,
        hkToTopLeftCell	= 246,
        hkToBottomRightCell	= 247,
        hkRow	= 248,
        hkFromRowStart	= 249,
        hkRowToEnd	= 250,
        hkFirstRowCell	= 251,
        hkColumn	= 252,
        hkFromColumnStart	= 253,
        hkColumnToEnd	= 254,
        hkFirstColumnCell	= 255,
        hkCurrentCell	= 256,
        hkElementProperties	= 257,
        hkPageNavigation	= 258,
        hkHeaderAndTotals	= 259,
        hkMonitorCells	= 260,
        hkMonitorCell1	= 261,
        hkMonitorCell2	= 262,
        hkMonitorCell3	= 263,
        hkMonitorCell4	= 264,
        hkMonitorCell5	= 265,
        hkMonitorCell6	= 266,
        hkMonitorCell7	= 267,
        hkMonitorCell8	= 268,
        hkMonitorCell9	= 269,
        hkMonitorCell10	= 270,
        hkAnyMonitorCell	= 271,
        hkReadTotalCell	= 272,
        hkReadComment	= 273,
        hkReadFormula	= 274,
        hkPreviousCell	= 275,
        hkOutlookCalendar	= 276,
        hkShutdownWindowEyes	= 277,
        hkBatteryLevel	= 278,
        hkWordEnhanced	= 0x80000001,
        hkOfficeAttributeStatus	= 0x80000002,
        hkOfficeSmartTab	= 0x80000003,
        hkScriptStatusRotor	= 279,
        hkLandmarkNext	= 280,
        hkLandmarkPrevious	= 281,
        hkButtonNext	= 282,
        hkButtonPrevious	= 283,
        hkComboboxNext	= 284,
        hkComboboxPrevious	= 285,
        hkEditboxNext	= 286,
        hkEditboxPrevious	= 287,
        hkRadioButtonNext	= 288,
        hkRadioButtonPrevious	= 289,
        hkCheckboxNext	= 290,
        hkCheckboxPrevious	= 291,
        hkFormInteraction	= 292
    } 	HotkeyID;

typedef /* [helpcontext] */ 
enum BrailleHotkeyID
    {
        bhkNone	= 0xffffffff,
        bhkPriorBrailleLine	= 1001,
        bhkNextBrailleLine	= 1002,
        bhkScrollBrailleLeft	= 1003,
        bhkScrollBrailleRight	= 1004,
        bhkBrailleTrackingToggle	= 1005,
        bhkBrailleGraphicMode	= 1006,
        bhk6Or8DotToggle	= 1008,
        bhkScrollModeToggle	= 1009,
        bhkAttributesToShowRotor	= 1010,
        bhkShowAttributesToggle	= 1011,
        bhkTopLine	= 1012,
        bhkBottomLine	= 1013,
        bhkToFocus	= 1014,
        bhkToMenuBar	= 1015,
        bhkToTitleBar	= 1016,
        bhkToStatusLine	= 1017,
        bhkQuickMessageRotor	= 1018,
        bhkDisplayIndividualAttributesRotor	= 1019,
        bhkBrailleWindowToggle	= 1020,
        bhkIndentToggle	= 1021,
        bhkSpacingToggle	= 1022,
        bhkDisplaySpecifiedCellAttributes	= 1023,
        bhkGrade2BrailleToggle	= 1024,
        bhkUntranslateAtCursor	= 1025,
        bhkUntranslateAtMouse	= 1026,
        bhkAutoUntranslateAtCursor	= 1027,
        bhkExpandSymbolsToggle	= 1028,
        bhkStickyShift	= 1029,
        bhkStickyControl	= 1030,
        bhkStickyAlt	= 1031,
        bhkStickyWindows	= 1032,
        bhkStickyMenu	= 1033,
        bhkStickyInsert	= 1034,
        bhkClearStickyKeys	= 1035,
        bhkNameToggle	= 1036,
        bhkTypeToggle	= 1037,
        bhkStickyAltGr	= 1038,
        bhkAutoRouteCursorToggle	= 1039,
        bhkRotorDisplayMode	= 1040,
        bhkLeftOneCell	= 1041,
        bhkRightOneCell	= 1042
    } 	BrailleHotkeyID;

typedef /* [helpcontext] */ 
enum MouseButton
    {
        mbNone	= 0xffffffff,
        mbLeft	= 0,
        mbMiddle	= 1,
        mbRight	= 2
    } 	MouseButton;

typedef /* [helpcontext] */ 
enum Voice
    {
        voiceKeyboard	= 0,
        voiceScreen	= 1,
        voiceMouse	= 2
    } 	Voice;

typedef /* [helpcontext] */ 
enum MouseBoundary
    {
        mbFullScreen	= 0,
        mbActiveWindow	= 1,
        mbFocusWindow	= 2,
        mbUserWindow	= 3
    } 	MouseBoundary;

typedef /* [helpcontext] */ 
enum KeyDisposition
    {
        kdDiscard	= 0,
        kdPass	= 1,
        kdProcess	= 2,
        kdSilence	= 0x40000000
    } 	KeyDisposition;

typedef /* [helpcontext] */ 
enum BaudRate
    {
        br300	= 0,
        br1200	= 1,
        br2400	= 2,
        br4800	= 3,
        br9600	= 4,
        br38400	= 5
    } 	BaudRate;

typedef /* [helpcontext] */ 
enum DataBits
    {
        db7	= 0,
        db8	= 1
    } 	DataBits;

typedef /* [helpcontext] */ 
enum StopBits
    {
        sb1	= 0,
        sb2	= 1
    } 	StopBits;

typedef /* [helpcontext] */ 
enum Parity
    {
        parityNone	= 0,
        parityOdd	= 1,
        parityEven	= 2,
        parityMark	= 3
    } 	Parity;

typedef /* [helpcontext] */ 
enum FlowControl
    {
        fcXonXoff	= 0,
        fcRtsCts	= 1,
        fcDsrDtr	= 2
    } 	FlowControl;

typedef /* [helpcontext] */ 
enum PortType
    {
        ptCom	= 0,
        ptLpt	= 1
    } 	PortType;

typedef /* [helpcontext] */ 
enum WindowTypeStatus
    {
        wtsOff	= 0,
        wtsBefore	= 1,
        wtsAfter	= 2
    } 	WindowTypeStatus;

typedef /* [helpcontext] */ 
enum DialogEvent
    {
        dialogCreated	= 272,
        dialogClosing	= 16,
        buttonClicked	= 0,
        buttonDisable	= 4,
        buttonDoubleClicked	= 5,
        buttonFocus	= 6,
        buttonBlur	= 7,
        checkboxClicked	= 0,
        checkboxDisable	= 4,
        checkboxDoubleClicked	= 5,
        checkboxFocus	= 6,
        checkboxBlur	= 7,
        radiobuttonClicked	= 0,
        radiobuttonDisable	= 4,
        radiobuttonDoubleClicked	= 5,
        radiobuttonFocus	= 6,
        radiobuttonBlur	= 7,
        staticClicked	= 0,
        staticDoubleClicked	= 1,
        staticEnabled	= 2,
        staticDisabled	= 3,
        editboxFocus	= 256,
        editboxBlur	= 512,
        editboxChange	= 768,
        editboxUpdate	= 1024,
        editboxMemoryError	= 1280,
        editboxTooMuchText	= 1281,
        editboxHorizontalScroll	= 1537,
        editboxVerticalScroll	= 1538,
        editboxAlignmentLTR	= 1792,
        editboxAlignmentRTL	= 1793,
        listboxMemoryError	= 0xfffffffe,
        listboxSelectionChange	= 1,
        listboxDoubleClicked	= 2,
        listboxSelectionCancel	= 3,
        listboxFocus	= 4,
        listboxBlur	= 5,
        comboboxMemoryError	= 0xffffffff,
        comboboxSelectionChange	= 1,
        comboboxDoubleClicked	= 2,
        comboboxFocus	= 3,
        comboboxBlur	= 4,
        comboboxEditChange	= 5,
        comboboxEditUpdate	= 6,
        comboboxDropdown	= 7,
        comboboxCloseup	= 8,
        comboboxSelectionEndOk	= 9,
        comboboxSelectionEndCancel	= 10,
        trackbarLineUp	= 0,
        trackbarLineDown	= 1,
        trackbarPageUp	= 2,
        trackbarPageDown	= 3,
        trackbarTop	= 6,
        trackbarBottom	= 7,
        trackbarEndTrack	= 8,
        updownUp	= 0,
        updownLeft	= 0,
        updownDown	= 1,
        updownRight	= 1,
        updownChange	= 0xfffffd2e,
        listviewFocus	= 0xfffffff9,
        listviewBlur	= 0xfffffff8,
        listviewItemChanged	= 0xffffff9b,
        listviewKeydown	= 0xffffff65,
        listviewClicked	= 0xfffffffe,
        treeviewFocus	= 0xfffffff9,
        treeviewBlur	= 0xfffffff8,
        treeviewSelectionChange	= 0xfffffe3d,
        treeviewItemExpanded	= 0xfffffe39,
        treeviewSelectionChanging	= 0xfffffe3e,
        treeviewItemExpanding	= 0xfffffe3a,
        treeviewKeydown	= 0xfffffe64,
        treeviewClicked	= 0xfffffffe,
        menuSelection	= 294
    } 	DialogEvent;

typedef /* [helpcontext] */ 
enum listviewView
    {
        lvvIcon	= 0,
        lvvList	= 1,
        lvvReport	= 2,
        lvvSmallIcon	= 3,
        lvvTile	= 4
    } 	listviewView;

typedef /* [helpcontext] */ 
enum listviewSort
    {
        lvsNone	= 0,
        lvsAscending	= 1,
        lvsDescending	= 2
    } 	listviewSort;

typedef /* [helpcontext] */ 
enum listviewCheck
    {
        lvcUnchecked	= 0,
        lvcChecked	= 1,
        lvcNocheck	= 2
    } 	listviewCheck;

typedef /* [helpcontext] */ 
enum CommonFileDialogFlags
    {
        cfdfAllowMultiSelect	= 512,
        cfdfCreatePrompt	= 8192,
        cfdfDoNotAddToRecent	= 0x2000000,
        cfdfEnableDialogResizing	= 0x800000,
        cfdfUseExplorerStyleInterface	= 0x80000,
        cfdfFileMustExist	= 4096,
        cfdfForceShowHidden	= 0x10000000,
        cfdfHideReadOnlyCheck	= 4,
        cfdfUseLongFileNames	= 0x200000,
        cfdfNoChangeDirectory	= 8,
        cfdfDoNotDereferenceLinks	= 0x100000,
        cfdfNoLongFileNames	= 0x40000,
        cfdfNoNetworkButton	= 0x20000,
        cfdfDoNotAllowReadOnlyFiles	= 32768,
        cfdfWarnIfFileExists	= 2,
        cfdfPathMustExist	= 2048,
        cfdfNoPlacesBar	= 0x20000000
    } 	CommonFileDialogFlags;

typedef /* [helpcontext] */ 
enum CursorKeyActions
    {
        ckaUndefined	= 0,
        ckaPriorCharacterBefore	= 1,
        ckaPriorCharacterAfter	= 2,
        ckaCharacterBefore	= 3,
        ckaCharacterAfter	= 4,
        ckaNextCharacter	= 5,
        ckaWordPrior	= 6,
        ckaWord	= 7,
        ckaWordNext	= 8,
        ckaLinePrior	= 9,
        ckaLine	= 10,
        ckaLineNext	= 11,
        ckaSentence	= 12,
        ckaParagraph	= 13,
        ckaToEndOfLine	= 14,
        ckaHighlight	= 15,
        ckaHighlightedText	= 16,
        ckaCursorPosition	= 17,
        ckaFieldData	= 18,
        ckaFieldName	= 19,
        ckaFieldNameAndData	= 20,
        ckaTitleStatus	= 21,
        ckaSpeakWindow	= 22,
        ckaActiveWindow	= 23,
        ckaMouseToWindow	= 24,
        ckaMouseTopLeft	= 25,
        ckaLoadSet	= 26,
        ckaIgnoreSpeakWindows	= 27,
        ckaExecuteHotkey	= 28,
        ckaWordEnhanced	= 29,
        ckaOfficeAttributeStatus	= 30,
        ckaOfficeSmartTab	= 31
    } 	CursorKeyActions;

typedef /* [helpcontext] */ 
enum OfficeAttributeStatus
    {
        oasBold	= 0,
        oasItalic	= 1,
        oasUnderline	= 2,
        oasShowAll	= 3,
        oasFontSize	= 4
    } 	OfficeAttributeStatus;

typedef /* [helpcontext] */ 
enum AccessibleSelection
    {
        accTakeFocus	= 1,
        accTakeSelection	= 2,
        accExtendSelection	= 4,
        accAddSelection	= 8,
        accRemoveSelection	= 16
    } 	AccessibleSelection;

typedef /* [helpcontext] */ 
enum AccessibleRoleEnum
    {
        role_SYSTEM_TITLEBAR	= 1,
        role_SYSTEM_MENUBAR	= 2,
        role_SYSTEM_SCROLLBAR	= 3,
        role_SYSTEM_GRIP	= 4,
        role_SYSTEM_SOUND	= 5,
        role_SYSTEM_CURSOR	= 6,
        role_SYSTEM_CARET	= 7,
        role_SYSTEM_ALERT	= 8,
        role_SYSTEM_WINDOW	= 9,
        role_SYSTEM_CLIENT	= 10,
        role_SYSTEM_MENUPOPUP	= 11,
        role_SYSTEM_MENUITEM	= 12,
        role_SYSTEM_TOOLTIP	= 13,
        role_SYSTEM_APPLICATION	= 14,
        role_SYSTEM_DOCUMENT	= 15,
        role_SYSTEM_PANE	= 16,
        role_SYSTEM_CHART	= 17,
        role_SYSTEM_DIALOG	= 18,
        role_SYSTEM_BORDER	= 19,
        role_SYSTEM_GROUPING	= 20,
        role_SYSTEM_SEPARATOR	= 21,
        role_SYSTEM_TOOLBAR	= 22,
        role_SYSTEM_STATUSBAR	= 23,
        role_SYSTEM_TABLE	= 24,
        role_SYSTEM_COLUMNHEADER	= 25,
        role_SYSTEM_ROWHEADER	= 26,
        role_SYSTEM_COLUMN	= 27,
        role_SYSTEM_ROW	= 28,
        role_SYSTEM_CELL	= 29,
        role_SYSTEM_LINK	= 30,
        role_SYSTEM_HELPBALLOON	= 31,
        role_SYSTEM_CHARACTER	= 32,
        role_SYSTEM_LIST	= 33,
        role_SYSTEM_LISTITEM	= 34,
        role_SYSTEM_OUTLINE	= 35,
        role_SYSTEM_OUTLINEITEM	= 36,
        role_SYSTEM_PAGETAB	= 37,
        role_SYSTEM_PROPERTYPAGE	= 38,
        role_SYSTEM_INDICATOR	= 39,
        role_SYSTEM_GRAPHIC	= 40,
        role_SYSTEM_STATICTEXT	= 41,
        role_SYSTEM_TEXT	= 42,
        role_SYSTEM_PUSHBUTTON	= 43,
        role_SYSTEM_CHECKBUTTON	= 44,
        role_SYSTEM_RADIOBUTTON	= 45,
        role_SYSTEM_COMBOBOX	= 46,
        role_SYSTEM_DROPLIST	= 47,
        role_SYSTEM_PROGRESSBAR	= 48,
        role_SYSTEM_DIAL	= 49,
        role_SYSTEM_HOTKEYFIELD	= 50,
        role_SYSTEM_SLIDER	= 51,
        role_SYSTEM_SPINBUTTON	= 52,
        role_SYSTEM_DIAGRAM	= 53,
        role_SYSTEM_ANIMATION	= 54,
        role_SYSTEM_EQUATION	= 55,
        role_SYSTEM_BUTTONDROPDOWN	= 56,
        role_SYSTEM_BUTTONMENU	= 57,
        role_SYSTEM_BUTTONDROPDOWNGRID	= 58,
        role_SYSTEM_WHITESPACE	= 59,
        role_SYSTEM_PAGETABLIST	= 60,
        role_SYSTEM_CLOCK	= 61,
        role_SYSTEM_SPLITBUTTON	= 62,
        role_SYSTEM_IPADDRESS	= 63,
        role_SYSTEM_OUTLINEBUTTON	= 64
    } 	AccessibleRoleEnum;

typedef /* [helpcontext] */ 
enum AccessibleStateEnum
    {
        state_SYSTEM_NORMAL	= 0,
        state_SYSTEM_UNAVAILABLE	= 1,
        state_SYSTEM_SELECTED	= 2,
        state_SYSTEM_FOCUSED	= 4,
        state_SYSTEM_PRESSED	= 8,
        state_SYSTEM_CHECKED	= 16,
        state_SYSTEM_MIXED	= 32,
        state_SYSTEM_INDETERMINATE	= 32,
        state_SYSTEM_READONLY	= 64,
        state_SYSTEM_HOTTRACKED	= 128,
        state_SYSTEM_DEFAULT	= 256,
        state_SYSTEM_EXPANDED	= 512,
        state_SYSTEM_COLLAPSED	= 1024,
        state_SYSTEM_BUSY	= 2048,
        state_SYSTEM_FLOATING	= 4096,
        state_SYSTEM_MARQUEED	= 8192,
        state_SYSTEM_ANIMATED	= 16384,
        state_SYSTEM_INVISIBLE	= 32768,
        state_SYSTEM_OFFSCREEN	= 0x10000,
        state_SYSTEM_SIZEABLE	= 0x20000,
        state_SYSTEM_MOVEABLE	= 0x40000,
        state_SYSTEM_SELFVOICING	= 0x80000,
        state_SYSTEM_FOCUSABLE	= 0x100000,
        state_SYSTEM_SELECTABLE	= 0x200000,
        state_SYSTEM_LINKED	= 0x400000,
        state_SYSTEM_TRAVERSED	= 0x800000,
        state_SYSTEM_MULTISELECTABLE	= 0x1000000,
        state_SYSTEM_EXTSELECTABLE	= 0x2000000,
        state_SYSTEM_ALERT_LOW	= 0x4000000,
        state_SYSTEM_ALERT_MEDIUM	= 0x8000000,
        state_SYSTEM_ALERT_HIGH	= 0x10000000,
        state_SYSTEM_PROTECTED	= 0x20000000,
        state_SYSTEM_HASPOPUP	= 0x40000000
    } 	AccessibleStateEnum;

typedef /* [helpcontext] */ 
enum AccessibleObjectID
    {
        objid_WINDOW	= 0,
        objid_SYSMENU	= 0xffffffff,
        objid_TITLEBAR	= 0xfffffffe,
        objid_MENU	= 0xfffffffd,
        objid_CLIENT	= 0xfffffffc,
        objid_VSCROLL	= 0xfffffffb,
        objid_HSCROLL	= 0xfffffffa,
        objid_SIZEGRIP	= 0xfffffff9
    } 	AccessibleObjectID;

typedef /* [helpcontext] */ 
enum MSAAEventID
    {
        event_SYSTEM_SOUND	= 1,
        event_SYSTEM_ALERT	= 2,
        event_SYSTEM_FOREGROUND	= 3,
        event_SYSTEM_MENUSTART	= 4,
        event_SYSTEM_MENUEND	= 5,
        event_SYSTEM_MENUPOPUPSTART	= 6,
        event_SYSTEM_MENUPOPUPEND	= 7,
        event_SYSTEM_CAPTURESTART	= 8,
        event_SYSTEM_CAPTUREEND	= 9,
        event_SYSTEM_MOVESIZESTART	= 10,
        event_SYSTEM_MOVESIZEEND	= 11,
        event_SYSTEM_CONTEXTHELPSTART	= 12,
        event_SYSTEM_CONTEXTHELPEND	= 13,
        event_SYSTEM_DRAGDROPSTART	= 14,
        event_SYSTEM_DRAGDROPEND	= 15,
        event_SYSTEM_DIALOGSTART	= 16,
        event_SYSTEM_DIALOGEND	= 17,
        event_SYSTEM_SCROLLINGSTART	= 18,
        event_SYSTEM_SCROLLINGEND	= 19,
        event_SYSTEM_SWITCHSTART	= 20,
        event_SYSTEM_SWITCHEND	= 21,
        event_SYSTEM_MINIMIZESTART	= 22,
        event_SYSTEM_MINIMIZEEND	= 23,
        event_CONSOLE_CARET	= 16385,
        event_CONSOLE_UPDATE_REGION	= 16386,
        event_CONSOLE_UPDATE_SIMPLE	= 16387,
        event_CONSOLE_UPDATE_SCROLL	= 16388,
        event_CONSOLE_LAYOUT	= 16389,
        event_CONSOLE_START_APPLICATION	= 16390,
        event_CONSOLE_END_APPLICATION	= 16391,
        event_OBJECT_SHOW	= 32770,
        event_OBJECT_HIDE	= 32771,
        event_OBJECT_REORDER	= 32772,
        event_OBJECT_FOCUS	= 32773,
        event_OBJECT_SELECTION	= 32774,
        event_OBJECT_SELECTIONADD	= 32775,
        event_OBJECT_SELECTIONREMOVE	= 32776,
        event_OBJECT_SELECTIONWITHIN	= 32777,
        event_OBJECT_STATECHANGE	= 32778,
        event_OBJECT_LOCATIONCHANGE	= 32779,
        event_OBJECT_NAMECHANGE	= 32780,
        event_OBJECT_DESCRIPTIONCHANGE	= 32781,
        event_OBJECT_VALUECHANGE	= 32782,
        event_OBJECT_PARENTCHANGE	= 32783,
        event_OBJECT_HELPCHANGE	= 32784,
        event_OBJECT_DEFACTIONCHANGE	= 32785,
        event_OBJECT_ACCELERATORCHANGE	= 32786,
        event_OBJECT_INVOKED	= 32787,
        event_OBJECT_TEXTSELECTIONCHANGED	= 32788,
        event_OBJECT_CONTENTSCROLLED	= 32789
    } 	MSAAEventID;

typedef /* [helpcontext] */ 
enum MessageBeeps
    {
        mbAsterisk	= 0,
        mbDefaultBeep	= 1,
        mbExclamation	= 2,
        mbSystemExit	= 3,
        mbCriticalStop	= 4,
        mbQuestion	= 5,
        mbSystemStart	= 6,
        mbSystemWelcome	= 7
    } 	MessageBeeps;

typedef /* [helpcontext] */ 
enum FilterMatch
    {
        fmExact	= 0,
        fmStartsWith	= 1,
        fmContains	= 2,
        fmEndsWith	= 3
    } 	FilterMatch;

typedef /* [helpcontext] */ 
enum CursorType
    {
        ctMousePointer	= 0,
        ctWECursor	= 1
    } 	CursorType;

typedef /* [helpcontext] */ 
enum OnScreenVisibility
    {
        osvNone	= 0,
        osvComplete	= 1,
        osvPartial	= 2
    } 	OnScreenVisibility;

typedef /* [helpcontext] */ 
enum SystemMetric
    {
        sm_CXSCREEN	= 0,
        sm_CYSCREEN	= 1,
        sm_CXVSCROLL	= 2,
        sm_CYHSCROLL	= 3,
        sm_CYCAPTION	= 4,
        sm_CXBORDER	= 5,
        sm_CYBORDER	= 6,
        sm_CXDLGFRAME	= 7,
        sm_CYDLGFRAME	= 8,
        sm_CYVTHUMB	= 9,
        sm_CXHTHUMB	= 10,
        sm_CXICON	= 11,
        sm_CYICON	= 12,
        sm_CXCURSOR	= 13,
        sm_CYCURSOR	= 14,
        sm_CYMENU	= 15,
        sm_CXFULLSCREEN	= 16,
        sm_CYFULLSCREEN	= 17,
        sm_CYKANJIWINDOW	= 18,
        sm_MOUSEPRESENT	= 19,
        sm_CYVSCROLL	= 20,
        sm_CXHSCROLL	= 21,
        sm_DEBUG	= 22,
        sm_SWAPBUTTON	= 23,
        sm_RESERVED1	= 24,
        sm_RESERVED2	= 25,
        sm_RESERVED3	= 26,
        sm_RESERVED4	= 27,
        sm_CXMIN	= 28,
        sm_CYMIN	= 29,
        sm_CXSIZE	= 30,
        sm_CYSIZE	= 31,
        sm_CXFRAME	= 32,
        sm_CYFRAME	= 33,
        sm_CXMINTRACK	= 34,
        sm_CYMINTRACK	= 35,
        sm_CXDOUBLECLK	= 36,
        sm_CYDOUBLECLK	= 37,
        sm_CXICONSPACING	= 38,
        sm_CYICONSPACING	= 39,
        sm_MENUDROPALIGNMENT	= 40,
        sm_PENWINDOWS	= 41,
        sm_DBCSENABLED	= 42,
        sm_CMOUSEBUTTONS	= 43,
        sm_CXFIXEDFRAME	= 7,
        sm_CYFIXEDFRAME	= 8,
        sm_CXSIZEFRAME	= 32,
        sm_CYSIZEFRAME	= 33,
        sm_SECURE	= 44,
        sm_CXEDGE	= 45,
        sm_CYEDGE	= 46,
        sm_CXMINSPACING	= 47,
        sm_CYMINSPACING	= 48,
        sm_CXSMICON	= 49,
        sm_CYSMICON	= 50,
        sm_CYSMCAPTION	= 51,
        sm_CXSMSIZE	= 52,
        sm_CYSMSIZE	= 53,
        sm_CXMENUSIZE	= 54,
        sm_CYMENUSIZE	= 55,
        sm_ARRANGE	= 56,
        sm_CXMINIMIZED	= 57,
        sm_CYMINIMIZED	= 58,
        sm_CXMAXTRACK	= 59,
        sm_CYMAXTRACK	= 60,
        sm_CXMAXIMIZED	= 61,
        sm_CYMAXIMIZED	= 62,
        sm_NETWORK	= 63,
        sm_CLEANBOOT	= 67,
        sm_CXDRAG	= 68,
        sm_CYDRAG	= 69,
        sm_SHOWSOUNDS	= 70,
        sm_CXMENUCHECK	= 71,
        sm_CYMENUCHECK	= 72,
        sm_SLOWMACHINE	= 73,
        sm_MIDEASTENABLED	= 74,
        sm_MOUSEWHEELPRESENT	= 75,
        sm_XVIRTUALSCREEN	= 76,
        sm_YVIRTUALSCREEN	= 77,
        sm_CXVIRTUALSCREEN	= 78,
        sm_CYVIRTUALSCREEN	= 79,
        sm_CMONITORS	= 80,
        sm_SAMEDISPLAYFORMAT	= 81,
        sm_IMMENABLED	= 82,
        sm_CXFOCUSBORDER	= 83,
        sm_CYFOCUSBORDER	= 84,
        sm_TABLETPC	= 86,
        sm_MEDIACENTER	= 87,
        sm_STARTER	= 88,
        sm_SERVERR2	= 89,
        sm_MOUSEHORIZONTALWHEELPRESENT	= 91,
        sm_CXPADDEDBORDER	= 92,
        sm_REMOTESESSION	= 4096,
        sm_SHUTTINGDOWN	= 8192,
        sm_REMOTECONTROL	= 8193,
        sm_CARETBLINKINGENABLED	= 8194
    } 	SystemMetric;

typedef /* [helpcontext] */ 
enum BrowseModeState
    {
        bmsDisabled	= 0,
        bmsEnabled	= 1,
        bmsSuspended	= 2,
        bmsLoading	= 3,
        bmsActive	= 4
    } 	BrowseModeState;

typedef /* [helpcontext] */ 
enum OSProductSuiteMask
    {
        psmBackOffice	= 4,
        psmBlade	= 1024,
        psmCluster	= 16384,
        psmDataCenter	= 128,
        psmEnterprise	= 2,
        psmEmbedded	= 64,
        psmPersonal	= 512,
        psmSingleUserTS	= 256,
        psmSmallBusiness	= 1,
        psmSmallBusinessRestricted	= 32,
        psmStorage	= 8192,
        psmTerminal	= 16
    } 	OSProductSuiteMask;

typedef /* [helpcontext] */ 
enum OSProductType
    {
        ptDomainController	= 2,
        ptServer	= 3,
        ptWorkstation	= 1
    } 	OSProductType;

typedef /* [helpcontext] */ 
enum MenuFlags
    {
        mfDisabled	= 1,
        mfPulldown	= 2,
        mfChecked	= 4,
        mfDialog	= 8,
        mfSystem	= 16,
        mfSystemChild	= 32
    } 	MenuFlags;

typedef /* [helpcontext] */ 
enum BrailleDisplayType
    {
        bdtNONE	= 0,
        bdtPOWERBRAILLE	= 1,
        bdtALVA	= 2,
        bdtBRAILLEWINDOW	= 3,
        bdtBRAILLELITE40	= 4,
        bdtBRAILLELITE18	= 5,
        bdtVARIO	= 6,
        bdtALVA280	= 7,
        bdtALVA240	= 8,
        bdtHT	= 9,
        bdtBRAILLEX	= 10,
        bdtBRAILLEXEL	= 11,
        bdtBRAILLENOTE	= 12,
        bdtVOYAGER	= 13,
        bdtDM80	= 14,
        bdtTVB2040	= 15,
        bdtTVBMINI	= 16,
        bdtME20	= 17,
        bdtME40	= 18,
        bdtZEPHYR	= 19,
        bdtMETEC	= 21,
        bdtEASY	= 22,
        bdtINKA	= 23,
        bdtBRAILLEX2	= 24,
        bdtFREEDOM	= 25,
        bdtTHDZ	= 26,
        bdtHANSONE	= 27,
        bdtEUROBRL	= 28,
        bdtSUPERVARIO	= 29,
        bdtPRONTO	= 30,
        bdtHEDO	= 31,
        bdtHT_COM	= 32,
        bdtTHDZ_PB	= 33,
        bdtVARIOCONNECT	= 34,
        bdtTRIO	= 35,
        bdtSYNCBRAILLE	= 36,
        bdtALVANEW	= 37,
        bdtBRAILLEMEMO	= 38,
        bdtLILLI	= 39,
        bdtOEM	= 100
    } 	BrailleDisplayType;

typedef /* [helpcontext] */ 
enum BrailleFreezeLocation
    {
        bflLeft	= 0,
        bflStart	= 0,
        bflRight	= 1,
        bflEnd	= 1
    } 	BrailleFreezeLocation;

typedef /* [helpcontext] */ 
enum BrailleControlTypeEnum
    {
        bctNone	= 0,
        bctControlID	= 1,
        bctFieldName	= 2,
        bctFieldData	= 3,
        bctSeparator	= 4,
        bctQuickMessageMode	= 5
    } 	BrailleControlTypeEnum;

typedef /* [helpcontext] */ 
enum BrailleActionType
    {
        batNothing	= 0,
        batHotkey	= 1,
        batBrailleHotkey	= 2,
        batKeypress	= 3,
        batCharacter	= 4
    } 	BrailleActionType;

typedef /* [helpcontext] */ 
enum BraillePatternBlinkStyle
    {
        bbsAlwaysUp	= 0,
        bbsBlink	= 1
    } 	BraillePatternBlinkStyle;

typedef /* [helpcontext] */ 
enum BrailleDisplayMode
    {
        bdmStructured	= 0,
        bdmLine	= 1,
        bdmSpeechBox	= 2
    } 	BrailleDisplayMode;

typedef /* [helpcontext] */ 
enum StartupOptionsCurrentUser
    {
        scuNever	= 0,
        scuAlways	= 1,
        scuUseCurrent	= 2
    } 	StartupOptionsCurrentUser;

typedef /* [helpcontext] */ 
enum AccessibleProperty
    {
        apName	= 1,
        apValue	= 2,
        apRole	= 4,
        apState	= 8,
        apRectangle	= 16,
        apShortcutKey	= 32,
        apDefaultAction	= 1024,
        apDescription	= 2048,
        apWindow	= 16384,
        apAll	= 19519
    } 	AccessibleProperty;

typedef /* [helpcontext] */ 
enum ProfileType
    {
        ptProfile	= 0,
        ptAllUsers	= 1,
        ptOSUser	= 2
    } 	ProfileType;

typedef /* [helpcontext] */ 
enum TreeViewInsert
    {
        tviRoot	= 0xffff0000,
        tviFirst	= 0xffff0001,
        tviLast	= 0xffff0002,
        tviSort	= 0xffff0003
    } 	TreeViewInsert;

typedef /* [helpcontext] */ 
enum treeviewCheck
    {
        tvcUnchecked	= 0,
        tvcChecked	= 1,
        tvcNocheck	= 2
    } 	treeviewCheck;

typedef /* [helpcontext] */ 
enum treeviewExpandAction
    {
        tveCollapse	= 1,
        tveExpand	= 2,
        tveToggle	= 3
    } 	treeviewExpandAction;

typedef /* [helpcontext] */ 
enum WindowFilterScope
    {
        wfsWindowOnly	= 1,
        wfsChildrenOnly	= 2,
        wfsWindowAndChildren	= 3
    } 	WindowFilterScope;

typedef /* [helpcontext] */ 
enum ListViewColumnWidth
    {
        lvcwAutoSize	= 0xffffffff,
        lvcwAutoSizeUseHeader	= 0xfffffffe
    } 	ListViewColumnWidth;

typedef /* [helpcontext] */ 
enum LoadedScriptStatus
    {
        ssRunning	= 1,
        ssRunningWithErrors	= 2,
        ssStarting	= 3,
        ssStopping	= 0x10000000,
        ssStopped	= 0x80000000,
        ssFileNotFound	= 0x80000001,
        ssDecryptionError1	= 0x80000002,
        ssDecryptionError2	= 0x80000003,
        ssAccessDenied	= 0x80000004,
        ssUntrusted	= 0x80000005
    } 	LoadedScriptStatus;

typedef /* [helpcontext] */ 
enum MenuItemFlags
    {
        mifChecked	= 8,
        mifUnchecked	= 0,
        mifDisabled	= 2,
        mifEnabled	= 0,
        mifGrayed	= 1,
        mifSeparator	= 2048
    } 	MenuItemFlags;

typedef /* [helpcontext] */ 
enum WindowEyesCopyProtection
    {
        cpNone	= 0,
        cpEverlock	= 1,
        cpDongle	= 2,
        cpUnknown	= 0xffffffff
    } 	WindowEyesCopyProtection;

typedef /* [helpcontext] */ 
enum MenuActiveStatus
    {
        maInactive	= 0,
        maContext	= 1,
        maNonContext	= 2
    } 	MenuActiveStatus;

typedef /* [helpcontext] */ 
enum WindowEyesLicenseType
    {
        ltNormal	= 0,
        ltDemo	= 1,
        ltEvaluation	= 2,
        ltLeaseToOwn	= 3
    } 	WindowEyesLicenseType;

typedef /* [helpcontext] */ 
enum SystemButtonId
    {
        idOK	= 1,
        idCANCEL	= 2,
        idABORT	= 3,
        idRETRY	= 4,
        idIGNORE	= 5,
        idYES	= 6,
        idNO	= 7,
        idCLOSE	= 8,
        idHELP	= 9,
        idTRYAGAIN	= 10,
        idCONTINUE	= 11
    } 	SystemButtonId;

typedef /* [helpcontext] */ 
enum MouseWheelDirection
    {
        wdForward	= 0,
        wdBackward	= 1
    } 	MouseWheelDirection;

typedef /* [helpcontext] */ 
enum SetFileScopes
    {
        scLocal	= 0,
        scGlobal	= 1
    } 	SetFileScopes;

typedef /* [helpcontext] */ 
enum BrailleTracking
    {
        btFocus	= 0,
        btMouse	= 1
    } 	BrailleTracking;

typedef 
enum HotkeyPriority
    {
        hpLow	= 0,
        hpMedium	= 1,
        hpHigh	= 2
    } 	HotkeyPriority;

typedef 
enum BrowseModeToggleInidicators
    {
        bmtiOff	= 0,
        bmtiOnWithSpeech	= 1,
        bmtiOnWithSound	= 2,
        bmtiOnWithSpeechAndSound	= 3
    } 	BrowseModeToggleInidicators;

typedef 
enum StartupMessages
    {
        smNone	= 0,
        smWindowEyes	= 1,
        smWindowEyesWithVersion	= 2
    } 	StartupMessages;

typedef Clip TextClip;


EXTERN_C const IID LIBID_WindowEyes;

#ifndef ___Application_INTERFACE_DEFINED__
#define ___Application_INTERFACE_DEFINED__

/* interface _Application */
/* [object][oleautomation][dual][helpcontext][helpstring][version][uuid] */ 


EXTERN_C const IID IID__Application;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("5E437833-ECC1-46EE-AAD6-0DE913811D3E")
    _Application : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ActiveWindow( 
            /* [retval][out] */ _Window **ActiveWindow) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ActiveWindow( 
            /* [in] */ _Window *ActiveWindow) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FocusedWindow( 
            /* [retval][out] */ _Window **FocusedWindow) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FocusedWindow( 
            /* [in] */ _Window *FocusedWindow) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Windows( 
            /* [retval][out] */ Windows **Windows) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Mouse( 
            /* [retval][out] */ _Mouse **Mouse) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Cursor( 
            /* [retval][out] */ Cursor **Cursor) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Keyboard( 
            /* [retval][out] */ _Keyboard **Keyboard) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ActiveSettings( 
            /* [retval][out] */ SetFile **settings) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ActiveCharacterDictionaries( 
            /* [retval][out] */ CharacterDictionaries **dicts) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ActiveColorDictionaries( 
            /* [retval][out] */ ColorDictionaries **dicts) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ActiveExceptionDictionaries( 
            /* [retval][out] */ ExceptionDictionaries **dicts) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ActiveGraphicDictionaries( 
            /* [retval][out] */ GraphicDictionaries **dicts) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ActiveKeyLabelDictionaries( 
            /* [retval][out] */ KeyLabelDictionaries **dicts) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FullName( 
            /* [retval][out] */ BSTR *FullName) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Name( 
            /* [retval][out] */ BSTR *Name) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Visible( 
            /* [retval][out] */ VARIANT_BOOL *Visible) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BrailleDisplays( 
            /* [retval][out] */ _BrailleDisplays **display) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CurrentLanguage( 
            /* [retval][out] */ Language *lang) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Clips( 
            /* [optional][in] */ VARIANT RealTime,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SerialNumber( 
            /* [retval][out] */ BSTR *serial) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UserName( 
            /* [retval][out] */ BSTR *UserName) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CompanyName( 
            /* [retval][out] */ BSTR *CompanyName) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Version( 
            /* [retval][out] */ BSTR *weversion) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReleaseDate( 
            /* [retval][out] */ DATE *date) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DesktopWindow( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SharedObjects( 
            /* [retval][out] */ _SharedObjects **SharedObjects) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ClientInformation( 
            /* [retval][out] */ _ClientInformation **ClientInformation) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WECursor( 
            /* [retval][out] */ WECursor **WECursor) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Text( 
            /* [retval][out] */ Text **Text) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ActiveMDIChild( 
            /* [retval][out] */ _Window **child) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Screen( 
            /* [retval][out] */ Screen **Screen) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Synthesizers( 
            /* [retval][out] */ _Synthesizers **Synthesizers) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Highlight( 
            /* [optional][in] */ VARIANT RealTime,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IniFile( 
            /* [in] */ BSTR file,
            /* [retval][out] */ IniFile **IniFile) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Speech( 
            /* [retval][out] */ _Speech **Speech) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MSAAEventSource( 
            /* [retval][out] */ _MSAAEvents **msaa) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TickCount( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Clipboard( 
            /* [retval][out] */ Clipboard **Clipboard) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ActiveCursorType( 
            /* [retval][out] */ CursorType *type) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ActiveCursorType( 
            /* [in] */ CursorType type) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BrowseMode( 
            /* [retval][out] */ _BrowseMode **BrowseMode) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WEDialogs( 
            /* [retval][out] */ WEDialogs **WEDialogs) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Braille( 
            /* [retval][out] */ _Braille **Braille) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_OSVersion( 
            /* [retval][out] */ OSVersion **obj) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Quit( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE PixelColor( 
            /* [in] */ ScreenPoint *point,
            /* [retval][out] */ Color **Color) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Color( 
            /* [in] */ int Red,
            /* [in] */ int Green,
            /* [in] */ int Blue,
            /* [retval][out] */ Color **Color) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ScreenPoint( 
            /* [in] */ int X,
            /* [in] */ int Y,
            /* [retval][out] */ ScreenPoint **sPoint) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ScreenRectangle( 
            /* [in] */ int Left,
            /* [in] */ int Top,
            /* [in] */ int Right,
            /* [in] */ int Bottom,
            /* [retval][out] */ ScreenRectangle **sRect) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ExecuteHotkey( 
            /* [in] */ HotkeyID Hotkey,
            /* [optional][in] */ VARIANT KeyAction) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ClientIdentify( 
            /* [in] */ VARIANT Script) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ExecuteBrailleHotkey( 
            /* [in] */ BrailleHotkeyID Hotkey) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Dialog( 
            /* [in] */ BSTR xmlName,
            /* [in] */ BSTR dialogName,
            /* [in] */ VARIANT eventHandler,
            /* [optional][in] */ VARIANT Parameter,
            /* [optional][in] */ VARIANT ParentWindow,
            /* [retval][out] */ _Dialog **Dialog) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Strings( 
            /* [in] */ BSTR xmlName,
            /* [retval][out] */ IDispatch **dictionary) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE PlaySound( 
            /* [in] */ VARIANT sound) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE GetSystemMetric( 
            /* [in] */ SystemMetric Metric,
            /* [retval][out] */ int *result) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE StartScript( 
            /* [in] */ BSTR ScriptName) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_RunInSystemTray( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_StartupOptions( 
            /* [retval][out] */ StartupOptions **StartupOptions) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WEPM( 
            /* [in] */ BSTR WEPMFile,
            /* [retval][out] */ WEPM **WEPM) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Profiles( 
            /* [retval][out] */ Profiles **Profiles) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Utilities( 
            /* [retval][out] */ Utilities **Utilities) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEndActive( 
            /* [retval][out] */ VARIANT_BOOL *ReadToEndActive) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE RefreshSets( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Menu( 
            /* [in] */ BSTR xmlName,
            /* [in] */ BSTR MenuName,
            /* [in] */ VARIANT eventHandler,
            /* [optional][in] */ VARIANT Parameter,
            /* [retval][out] */ Menu **Menu) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CurrentLocaleID( 
            /* [retval][out] */ long *LocaleID) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CopyProtection( 
            /* [retval][out] */ WindowEyesCopyProtection *protection) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MenuActive( 
            /* [retval][out] */ MenuActiveStatus *MenuActiveStatus) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LicenseType( 
            /* [retval][out] */ WindowEyesLicenseType *LicenseType) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE AttachWindowMessage( 
            /* [in] */ VARIANT function,
            /* [in] */ long message,
            /* [optional][in] */ VARIANT filterBy,
            /* [optional][in] */ VARIANT IncludeChildren,
            /* [optional][in] */ VARIANT functionParameter,
            /* [retval][out] */ WindowMessage **WindowMessage) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_RunInSystemTray( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Process( 
            /* [retval][out] */ Process **Process) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UseAlternateResource( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_UseAlternateResource( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _ApplicationVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _Application * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _Application * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _Application * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _Application * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _Application * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _Application * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _Application * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _Application * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _Application * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ActiveWindow )( 
            _Application * This,
            /* [retval][out] */ _Window **ActiveWindow);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ActiveWindow )( 
            _Application * This,
            /* [in] */ _Window *ActiveWindow);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FocusedWindow )( 
            _Application * This,
            /* [retval][out] */ _Window **FocusedWindow);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FocusedWindow )( 
            _Application * This,
            /* [in] */ _Window *FocusedWindow);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Windows )( 
            _Application * This,
            /* [retval][out] */ Windows **Windows);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Mouse )( 
            _Application * This,
            /* [retval][out] */ _Mouse **Mouse);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Cursor )( 
            _Application * This,
            /* [retval][out] */ Cursor **Cursor);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Keyboard )( 
            _Application * This,
            /* [retval][out] */ _Keyboard **Keyboard);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ActiveSettings )( 
            _Application * This,
            /* [retval][out] */ SetFile **settings);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ActiveCharacterDictionaries )( 
            _Application * This,
            /* [retval][out] */ CharacterDictionaries **dicts);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ActiveColorDictionaries )( 
            _Application * This,
            /* [retval][out] */ ColorDictionaries **dicts);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ActiveExceptionDictionaries )( 
            _Application * This,
            /* [retval][out] */ ExceptionDictionaries **dicts);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ActiveGraphicDictionaries )( 
            _Application * This,
            /* [retval][out] */ GraphicDictionaries **dicts);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ActiveKeyLabelDictionaries )( 
            _Application * This,
            /* [retval][out] */ KeyLabelDictionaries **dicts);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FullName )( 
            _Application * This,
            /* [retval][out] */ BSTR *FullName);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Name )( 
            _Application * This,
            /* [retval][out] */ BSTR *Name);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            _Application * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BrailleDisplays )( 
            _Application * This,
            /* [retval][out] */ _BrailleDisplays **display);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CurrentLanguage )( 
            _Application * This,
            /* [retval][out] */ Language *lang);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Clips )( 
            _Application * This,
            /* [optional][in] */ VARIANT RealTime,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SerialNumber )( 
            _Application * This,
            /* [retval][out] */ BSTR *serial);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UserName )( 
            _Application * This,
            /* [retval][out] */ BSTR *UserName);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CompanyName )( 
            _Application * This,
            /* [retval][out] */ BSTR *CompanyName);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Version )( 
            _Application * This,
            /* [retval][out] */ BSTR *weversion);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReleaseDate )( 
            _Application * This,
            /* [retval][out] */ DATE *date);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DesktopWindow )( 
            _Application * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SharedObjects )( 
            _Application * This,
            /* [retval][out] */ _SharedObjects **SharedObjects);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ClientInformation )( 
            _Application * This,
            /* [retval][out] */ _ClientInformation **ClientInformation);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WECursor )( 
            _Application * This,
            /* [retval][out] */ WECursor **WECursor);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            _Application * This,
            /* [retval][out] */ Text **Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ActiveMDIChild )( 
            _Application * This,
            /* [retval][out] */ _Window **child);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Screen )( 
            _Application * This,
            /* [retval][out] */ Screen **Screen);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Synthesizers )( 
            _Application * This,
            /* [retval][out] */ _Synthesizers **Synthesizers);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Highlight )( 
            _Application * This,
            /* [optional][in] */ VARIANT RealTime,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IniFile )( 
            _Application * This,
            /* [in] */ BSTR file,
            /* [retval][out] */ IniFile **IniFile);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Speech )( 
            _Application * This,
            /* [retval][out] */ _Speech **Speech);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MSAAEventSource )( 
            _Application * This,
            /* [retval][out] */ _MSAAEvents **msaa);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TickCount )( 
            _Application * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Clipboard )( 
            _Application * This,
            /* [retval][out] */ Clipboard **Clipboard);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ActiveCursorType )( 
            _Application * This,
            /* [retval][out] */ CursorType *type);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ActiveCursorType )( 
            _Application * This,
            /* [in] */ CursorType type);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BrowseMode )( 
            _Application * This,
            /* [retval][out] */ _BrowseMode **BrowseMode);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WEDialogs )( 
            _Application * This,
            /* [retval][out] */ WEDialogs **WEDialogs);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Braille )( 
            _Application * This,
            /* [retval][out] */ _Braille **Braille);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_OSVersion )( 
            _Application * This,
            /* [retval][out] */ OSVersion **obj);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Quit )( 
            _Application * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *PixelColor )( 
            _Application * This,
            /* [in] */ ScreenPoint *point,
            /* [retval][out] */ Color **Color);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Color )( 
            _Application * This,
            /* [in] */ int Red,
            /* [in] */ int Green,
            /* [in] */ int Blue,
            /* [retval][out] */ Color **Color);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ScreenPoint )( 
            _Application * This,
            /* [in] */ int X,
            /* [in] */ int Y,
            /* [retval][out] */ ScreenPoint **sPoint);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ScreenRectangle )( 
            _Application * This,
            /* [in] */ int Left,
            /* [in] */ int Top,
            /* [in] */ int Right,
            /* [in] */ int Bottom,
            /* [retval][out] */ ScreenRectangle **sRect);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ExecuteHotkey )( 
            _Application * This,
            /* [in] */ HotkeyID Hotkey,
            /* [optional][in] */ VARIANT KeyAction);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ClientIdentify )( 
            _Application * This,
            /* [in] */ VARIANT Script);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ExecuteBrailleHotkey )( 
            _Application * This,
            /* [in] */ BrailleHotkeyID Hotkey);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Dialog )( 
            _Application * This,
            /* [in] */ BSTR xmlName,
            /* [in] */ BSTR dialogName,
            /* [in] */ VARIANT eventHandler,
            /* [optional][in] */ VARIANT Parameter,
            /* [optional][in] */ VARIANT ParentWindow,
            /* [retval][out] */ _Dialog **Dialog);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Strings )( 
            _Application * This,
            /* [in] */ BSTR xmlName,
            /* [retval][out] */ IDispatch **dictionary);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *PlaySound )( 
            _Application * This,
            /* [in] */ VARIANT sound);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *GetSystemMetric )( 
            _Application * This,
            /* [in] */ SystemMetric Metric,
            /* [retval][out] */ int *result);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *StartScript )( 
            _Application * This,
            /* [in] */ BSTR ScriptName);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_RunInSystemTray )( 
            _Application * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_StartupOptions )( 
            _Application * This,
            /* [retval][out] */ StartupOptions **StartupOptions);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WEPM )( 
            _Application * This,
            /* [in] */ BSTR WEPMFile,
            /* [retval][out] */ WEPM **WEPM);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Profiles )( 
            _Application * This,
            /* [retval][out] */ Profiles **Profiles);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Utilities )( 
            _Application * This,
            /* [retval][out] */ Utilities **Utilities);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEndActive )( 
            _Application * This,
            /* [retval][out] */ VARIANT_BOOL *ReadToEndActive);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *RefreshSets )( 
            _Application * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Menu )( 
            _Application * This,
            /* [in] */ BSTR xmlName,
            /* [in] */ BSTR MenuName,
            /* [in] */ VARIANT eventHandler,
            /* [optional][in] */ VARIANT Parameter,
            /* [retval][out] */ Menu **Menu);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CurrentLocaleID )( 
            _Application * This,
            /* [retval][out] */ long *LocaleID);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CopyProtection )( 
            _Application * This,
            /* [retval][out] */ WindowEyesCopyProtection *protection);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MenuActive )( 
            _Application * This,
            /* [retval][out] */ MenuActiveStatus *MenuActiveStatus);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LicenseType )( 
            _Application * This,
            /* [retval][out] */ WindowEyesLicenseType *LicenseType);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *AttachWindowMessage )( 
            _Application * This,
            /* [in] */ VARIANT function,
            /* [in] */ long message,
            /* [optional][in] */ VARIANT filterBy,
            /* [optional][in] */ VARIANT IncludeChildren,
            /* [optional][in] */ VARIANT functionParameter,
            /* [retval][out] */ WindowMessage **WindowMessage);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_RunInSystemTray )( 
            _Application * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Process )( 
            _Application * This,
            /* [retval][out] */ Process **Process);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UseAlternateResource )( 
            _Application * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_UseAlternateResource )( 
            _Application * This,
            /* [in] */ VARIANT_BOOL prop);
        
        END_INTERFACE
    } _ApplicationVtbl;

    interface _Application
    {
        CONST_VTBL struct _ApplicationVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _Application_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _Application_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _Application_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _Application_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _Application_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _Application_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _Application_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _Application_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _Application_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _Application_get_ActiveWindow(This,ActiveWindow)	\
    ( (This)->lpVtbl -> get_ActiveWindow(This,ActiveWindow) ) 

#define _Application_put_ActiveWindow(This,ActiveWindow)	\
    ( (This)->lpVtbl -> put_ActiveWindow(This,ActiveWindow) ) 

#define _Application_get_FocusedWindow(This,FocusedWindow)	\
    ( (This)->lpVtbl -> get_FocusedWindow(This,FocusedWindow) ) 

#define _Application_put_FocusedWindow(This,FocusedWindow)	\
    ( (This)->lpVtbl -> put_FocusedWindow(This,FocusedWindow) ) 

#define _Application_get_Windows(This,Windows)	\
    ( (This)->lpVtbl -> get_Windows(This,Windows) ) 

#define _Application_get_Mouse(This,Mouse)	\
    ( (This)->lpVtbl -> get_Mouse(This,Mouse) ) 

#define _Application_get_Cursor(This,Cursor)	\
    ( (This)->lpVtbl -> get_Cursor(This,Cursor) ) 

#define _Application_get_Keyboard(This,Keyboard)	\
    ( (This)->lpVtbl -> get_Keyboard(This,Keyboard) ) 

#define _Application_get_ActiveSettings(This,settings)	\
    ( (This)->lpVtbl -> get_ActiveSettings(This,settings) ) 

#define _Application_get_ActiveCharacterDictionaries(This,dicts)	\
    ( (This)->lpVtbl -> get_ActiveCharacterDictionaries(This,dicts) ) 

#define _Application_get_ActiveColorDictionaries(This,dicts)	\
    ( (This)->lpVtbl -> get_ActiveColorDictionaries(This,dicts) ) 

#define _Application_get_ActiveExceptionDictionaries(This,dicts)	\
    ( (This)->lpVtbl -> get_ActiveExceptionDictionaries(This,dicts) ) 

#define _Application_get_ActiveGraphicDictionaries(This,dicts)	\
    ( (This)->lpVtbl -> get_ActiveGraphicDictionaries(This,dicts) ) 

#define _Application_get_ActiveKeyLabelDictionaries(This,dicts)	\
    ( (This)->lpVtbl -> get_ActiveKeyLabelDictionaries(This,dicts) ) 

#define _Application_get_FullName(This,FullName)	\
    ( (This)->lpVtbl -> get_FullName(This,FullName) ) 

#define _Application_get_Name(This,Name)	\
    ( (This)->lpVtbl -> get_Name(This,Name) ) 

#define _Application_get_Visible(This,Visible)	\
    ( (This)->lpVtbl -> get_Visible(This,Visible) ) 

#define _Application_get_BrailleDisplays(This,display)	\
    ( (This)->lpVtbl -> get_BrailleDisplays(This,display) ) 

#define _Application_get_CurrentLanguage(This,lang)	\
    ( (This)->lpVtbl -> get_CurrentLanguage(This,lang) ) 

#define _Application_get_Clips(This,RealTime,Clips)	\
    ( (This)->lpVtbl -> get_Clips(This,RealTime,Clips) ) 

#define _Application_get_SerialNumber(This,serial)	\
    ( (This)->lpVtbl -> get_SerialNumber(This,serial) ) 

#define _Application_get_UserName(This,UserName)	\
    ( (This)->lpVtbl -> get_UserName(This,UserName) ) 

#define _Application_get_CompanyName(This,CompanyName)	\
    ( (This)->lpVtbl -> get_CompanyName(This,CompanyName) ) 

#define _Application_get_Version(This,weversion)	\
    ( (This)->lpVtbl -> get_Version(This,weversion) ) 

#define _Application_get_ReleaseDate(This,date)	\
    ( (This)->lpVtbl -> get_ReleaseDate(This,date) ) 

#define _Application_get_DesktopWindow(This,Window)	\
    ( (This)->lpVtbl -> get_DesktopWindow(This,Window) ) 

#define _Application_get_SharedObjects(This,SharedObjects)	\
    ( (This)->lpVtbl -> get_SharedObjects(This,SharedObjects) ) 

#define _Application_get_ClientInformation(This,ClientInformation)	\
    ( (This)->lpVtbl -> get_ClientInformation(This,ClientInformation) ) 

#define _Application_get_WECursor(This,WECursor)	\
    ( (This)->lpVtbl -> get_WECursor(This,WECursor) ) 

#define _Application_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define _Application_get_ActiveMDIChild(This,child)	\
    ( (This)->lpVtbl -> get_ActiveMDIChild(This,child) ) 

#define _Application_get_Screen(This,Screen)	\
    ( (This)->lpVtbl -> get_Screen(This,Screen) ) 

#define _Application_get_Synthesizers(This,Synthesizers)	\
    ( (This)->lpVtbl -> get_Synthesizers(This,Synthesizers) ) 

#define _Application_get_Highlight(This,RealTime,Clips)	\
    ( (This)->lpVtbl -> get_Highlight(This,RealTime,Clips) ) 

#define _Application_get_IniFile(This,file,IniFile)	\
    ( (This)->lpVtbl -> get_IniFile(This,file,IniFile) ) 

#define _Application_get_Speech(This,Speech)	\
    ( (This)->lpVtbl -> get_Speech(This,Speech) ) 

#define _Application_get_MSAAEventSource(This,msaa)	\
    ( (This)->lpVtbl -> get_MSAAEventSource(This,msaa) ) 

#define _Application_get_TickCount(This,prop)	\
    ( (This)->lpVtbl -> get_TickCount(This,prop) ) 

#define _Application_get_Clipboard(This,Clipboard)	\
    ( (This)->lpVtbl -> get_Clipboard(This,Clipboard) ) 

#define _Application_get_ActiveCursorType(This,type)	\
    ( (This)->lpVtbl -> get_ActiveCursorType(This,type) ) 

#define _Application_put_ActiveCursorType(This,type)	\
    ( (This)->lpVtbl -> put_ActiveCursorType(This,type) ) 

#define _Application_get_BrowseMode(This,BrowseMode)	\
    ( (This)->lpVtbl -> get_BrowseMode(This,BrowseMode) ) 

#define _Application_get_WEDialogs(This,WEDialogs)	\
    ( (This)->lpVtbl -> get_WEDialogs(This,WEDialogs) ) 

#define _Application_get_Braille(This,Braille)	\
    ( (This)->lpVtbl -> get_Braille(This,Braille) ) 

#define _Application_get_OSVersion(This,obj)	\
    ( (This)->lpVtbl -> get_OSVersion(This,obj) ) 

#define _Application_Quit(This)	\
    ( (This)->lpVtbl -> Quit(This) ) 

#define _Application_PixelColor(This,point,Color)	\
    ( (This)->lpVtbl -> PixelColor(This,point,Color) ) 

#define _Application_Color(This,Red,Green,Blue,Color)	\
    ( (This)->lpVtbl -> Color(This,Red,Green,Blue,Color) ) 

#define _Application_ScreenPoint(This,X,Y,sPoint)	\
    ( (This)->lpVtbl -> ScreenPoint(This,X,Y,sPoint) ) 

#define _Application_ScreenRectangle(This,Left,Top,Right,Bottom,sRect)	\
    ( (This)->lpVtbl -> ScreenRectangle(This,Left,Top,Right,Bottom,sRect) ) 

#define _Application_ExecuteHotkey(This,Hotkey,KeyAction)	\
    ( (This)->lpVtbl -> ExecuteHotkey(This,Hotkey,KeyAction) ) 

#define _Application_ClientIdentify(This,Script)	\
    ( (This)->lpVtbl -> ClientIdentify(This,Script) ) 

#define _Application_ExecuteBrailleHotkey(This,Hotkey)	\
    ( (This)->lpVtbl -> ExecuteBrailleHotkey(This,Hotkey) ) 

#define _Application_Dialog(This,xmlName,dialogName,eventHandler,Parameter,ParentWindow,Dialog)	\
    ( (This)->lpVtbl -> Dialog(This,xmlName,dialogName,eventHandler,Parameter,ParentWindow,Dialog) ) 

#define _Application_Strings(This,xmlName,dictionary)	\
    ( (This)->lpVtbl -> Strings(This,xmlName,dictionary) ) 

#define _Application_PlaySound(This,sound)	\
    ( (This)->lpVtbl -> PlaySound(This,sound) ) 

#define _Application_GetSystemMetric(This,Metric,result)	\
    ( (This)->lpVtbl -> GetSystemMetric(This,Metric,result) ) 

#define _Application_StartScript(This,ScriptName)	\
    ( (This)->lpVtbl -> StartScript(This,ScriptName) ) 

#define _Application_get_RunInSystemTray(This,prop)	\
    ( (This)->lpVtbl -> get_RunInSystemTray(This,prop) ) 

#define _Application_get_StartupOptions(This,StartupOptions)	\
    ( (This)->lpVtbl -> get_StartupOptions(This,StartupOptions) ) 

#define _Application_get_WEPM(This,WEPMFile,WEPM)	\
    ( (This)->lpVtbl -> get_WEPM(This,WEPMFile,WEPM) ) 

#define _Application_get_Profiles(This,Profiles)	\
    ( (This)->lpVtbl -> get_Profiles(This,Profiles) ) 

#define _Application_get_Utilities(This,Utilities)	\
    ( (This)->lpVtbl -> get_Utilities(This,Utilities) ) 

#define _Application_get_ReadToEndActive(This,ReadToEndActive)	\
    ( (This)->lpVtbl -> get_ReadToEndActive(This,ReadToEndActive) ) 

#define _Application_RefreshSets(This)	\
    ( (This)->lpVtbl -> RefreshSets(This) ) 

#define _Application_Menu(This,xmlName,MenuName,eventHandler,Parameter,Menu)	\
    ( (This)->lpVtbl -> Menu(This,xmlName,MenuName,eventHandler,Parameter,Menu) ) 

#define _Application_get_CurrentLocaleID(This,LocaleID)	\
    ( (This)->lpVtbl -> get_CurrentLocaleID(This,LocaleID) ) 

#define _Application_get_CopyProtection(This,protection)	\
    ( (This)->lpVtbl -> get_CopyProtection(This,protection) ) 

#define _Application_get_MenuActive(This,MenuActiveStatus)	\
    ( (This)->lpVtbl -> get_MenuActive(This,MenuActiveStatus) ) 

#define _Application_get_LicenseType(This,LicenseType)	\
    ( (This)->lpVtbl -> get_LicenseType(This,LicenseType) ) 

#define _Application_AttachWindowMessage(This,function,message,filterBy,IncludeChildren,functionParameter,WindowMessage)	\
    ( (This)->lpVtbl -> AttachWindowMessage(This,function,message,filterBy,IncludeChildren,functionParameter,WindowMessage) ) 

#define _Application_put_RunInSystemTray(This,prop)	\
    ( (This)->lpVtbl -> put_RunInSystemTray(This,prop) ) 

#define _Application_get_Process(This,Process)	\
    ( (This)->lpVtbl -> get_Process(This,Process) ) 

#define _Application_get_UseAlternateResource(This,prop)	\
    ( (This)->lpVtbl -> get_UseAlternateResource(This,prop) ) 

#define _Application_put_UseAlternateResource(This,prop)	\
    ( (This)->lpVtbl -> put_UseAlternateResource(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___Application_INTERFACE_DEFINED__ */


#ifndef ___Window_INTERFACE_DEFINED__
#define ___Window_INTERFACE_DEFINED__

/* interface _Window */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__Window;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("138E1F63-840E-4E85-898E-14E19842C754")
    _Window : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Visible( 
            /* [retval][out] */ VARIANT_BOOL *Visible) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Enabled( 
            /* [retval][out] */ VARIANT_BOOL *Enabled) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Enabled( 
            /* [in] */ VARIANT_BOOL Enabled) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Handle( 
            /* [retval][out] */ long *Handle) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_rectangle( 
            /* [retval][out] */ WindowRectangle **rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_rectangle( 
            /* [in] */ VARIANT rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_type( 
            /* [retval][out] */ WindowType *type) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_type( 
            /* [in] */ WindowType type) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_OriginalType( 
            /* [retval][out] */ WindowType *OriginalType) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DirectChildren( 
            /* [retval][out] */ Windows **DirectChildren) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Children( 
            /* [retval][out] */ Windows **Children) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Overlap( 
            /* [retval][out] */ _Window **Overlap) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Style( 
            /* [retval][out] */ WindowStyle **Style) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CaretWidth( 
            /* [retval][out] */ long *Width) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Name( 
            /* [retval][out] */ BSTR *Name) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ClassName( 
            /* [retval][out] */ BSTR *ClassName) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_OriginalClassName( 
            /* [retval][out] */ BSTR *ClassName) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ModuleName( 
            /* [retval][out] */ BSTR *ModuleName) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Desktop( 
            /* [retval][out] */ Desktop **Desktop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Process( 
            /* [retval][out] */ Process **Process) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ClientArea( 
            /* [retval][out] */ ClientRectangle **rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Title( 
            /* [retval][out] */ BSTR *Title) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Status( 
            /* [retval][out] */ WindowStatus *Status) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Status( 
            /* [in] */ WindowStatus Status) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SetFileName( 
            /* [retval][out] */ BSTR *Name) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Selection( 
            /* [optional][in] */ VARIANT RealTime,
            /* [retval][out] */ Clips **Selection) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FieldData( 
            /* [retval][out] */ BSTR *Data) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FieldName( 
            /* [retval][out] */ BSTR *Name) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Accessible( 
            /* [optional][in] */ VARIANT objectID,
            /* [retval][out] */ Accessible **Accessible) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CurrentUserWindowID( 
            /* [retval][out] */ long *uwin) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CurrentUserWindowID( 
            /* [in] */ long uwin) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CurrentHyperactiveWindowID( 
            /* [retval][out] */ BSTR *hwin) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CurrentHyperactiveWindowID( 
            /* [in] */ BSTR hwin) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NativeObjectModel( 
            /* [retval][out] */ IDispatch **objectModel) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AlwaysOnTop( 
            /* [retval][out] */ VARIANT_BOOL *TopMost) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AlwaysOnTop( 
            /* [in] */ VARIANT_BOOL TopMost) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_settings( 
            /* [retval][out] */ SetFile **settings) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsValid( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_OnScreenVisibility( 
            /* [retval][out] */ OnScreenVisibility *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FormatAlertOffset( 
            /* [retval][out] */ WindowPoint **offset) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FormatAlertOffset( 
            /* [in] */ WindowPoint *offset) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Clips( 
            /* [optional][in] */ VARIANT IncludeChildren,
            /* [optional][in] */ VARIANT RealTime,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE WindowPoint( 
            /* [in] */ long X,
            /* [in] */ long Y,
            /* [retval][out] */ WindowPoint **point) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE WindowRectangle( 
            /* [in] */ long Left,
            /* [in] */ long Top,
            /* [in] */ long Right,
            /* [in] */ long Bottom,
            /* [retval][out] */ WindowRectangle **rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ClientPoint( 
            /* [in] */ long X,
            /* [in] */ long Y,
            /* [retval][out] */ ClientPoint **point) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ClientRectangle( 
            /* [in] */ long Left,
            /* [in] */ long Top,
            /* [in] */ long Right,
            /* [in] */ long Bottom,
            /* [retval][out] */ ClientRectangle **rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Control( 
            /* [optional][in] */ VARIANT id,
            /* [retval][out] */ IDispatch **Control) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Close( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Focus( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Activate( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SendMessage( 
            /* [in] */ long message,
            /* [in] */ VARIANT wParam,
            /* [in] */ VARIANT lParam,
            /* [retval][out] */ long *result) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE PostMessage( 
            /* [in] */ long message,
            /* [in] */ long wParam,
            /* [in] */ long lParam) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Redraw( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE LoadSet( 
            /* [in] */ BSTR SetFileName,
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Settings2( 
            /* [retval][out] */ Settings2 **settings) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _WindowVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _Window * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _Window * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _Window * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _Window * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _Window * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _Window * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _Window * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _Window * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _Window * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            _Window * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Enabled )( 
            _Window * This,
            /* [retval][out] */ VARIANT_BOOL *Enabled);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Enabled )( 
            _Window * This,
            /* [in] */ VARIANT_BOOL Enabled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Handle )( 
            _Window * This,
            /* [retval][out] */ long *Handle);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_rectangle )( 
            _Window * This,
            /* [retval][out] */ WindowRectangle **rectangle);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_rectangle )( 
            _Window * This,
            /* [in] */ VARIANT rectangle);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_type )( 
            _Window * This,
            /* [retval][out] */ WindowType *type);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_type )( 
            _Window * This,
            /* [in] */ WindowType type);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_OriginalType )( 
            _Window * This,
            /* [retval][out] */ WindowType *OriginalType);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DirectChildren )( 
            _Window * This,
            /* [retval][out] */ Windows **DirectChildren);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Children )( 
            _Window * This,
            /* [retval][out] */ Windows **Children);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Overlap )( 
            _Window * This,
            /* [retval][out] */ _Window **Overlap);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Style )( 
            _Window * This,
            /* [retval][out] */ WindowStyle **Style);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CaretWidth )( 
            _Window * This,
            /* [retval][out] */ long *Width);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Name )( 
            _Window * This,
            /* [retval][out] */ BSTR *Name);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ClassName )( 
            _Window * This,
            /* [retval][out] */ BSTR *ClassName);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_OriginalClassName )( 
            _Window * This,
            /* [retval][out] */ BSTR *ClassName);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ModuleName )( 
            _Window * This,
            /* [retval][out] */ BSTR *ModuleName);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Desktop )( 
            _Window * This,
            /* [retval][out] */ Desktop **Desktop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Process )( 
            _Window * This,
            /* [retval][out] */ Process **Process);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ClientArea )( 
            _Window * This,
            /* [retval][out] */ ClientRectangle **rectangle);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Title )( 
            _Window * This,
            /* [retval][out] */ BSTR *Title);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Status )( 
            _Window * This,
            /* [retval][out] */ WindowStatus *Status);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Status )( 
            _Window * This,
            /* [in] */ WindowStatus Status);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SetFileName )( 
            _Window * This,
            /* [retval][out] */ BSTR *Name);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Selection )( 
            _Window * This,
            /* [optional][in] */ VARIANT RealTime,
            /* [retval][out] */ Clips **Selection);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FieldData )( 
            _Window * This,
            /* [retval][out] */ BSTR *Data);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FieldName )( 
            _Window * This,
            /* [retval][out] */ BSTR *Name);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Accessible )( 
            _Window * This,
            /* [optional][in] */ VARIANT objectID,
            /* [retval][out] */ Accessible **Accessible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CurrentUserWindowID )( 
            _Window * This,
            /* [retval][out] */ long *uwin);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CurrentUserWindowID )( 
            _Window * This,
            /* [in] */ long uwin);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CurrentHyperactiveWindowID )( 
            _Window * This,
            /* [retval][out] */ BSTR *hwin);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CurrentHyperactiveWindowID )( 
            _Window * This,
            /* [in] */ BSTR hwin);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NativeObjectModel )( 
            _Window * This,
            /* [retval][out] */ IDispatch **objectModel);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AlwaysOnTop )( 
            _Window * This,
            /* [retval][out] */ VARIANT_BOOL *TopMost);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AlwaysOnTop )( 
            _Window * This,
            /* [in] */ VARIANT_BOOL TopMost);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_settings )( 
            _Window * This,
            /* [retval][out] */ SetFile **settings);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsValid )( 
            _Window * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_OnScreenVisibility )( 
            _Window * This,
            /* [retval][out] */ OnScreenVisibility *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FormatAlertOffset )( 
            _Window * This,
            /* [retval][out] */ WindowPoint **offset);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FormatAlertOffset )( 
            _Window * This,
            /* [in] */ WindowPoint *offset);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Clips )( 
            _Window * This,
            /* [optional][in] */ VARIANT IncludeChildren,
            /* [optional][in] */ VARIANT RealTime,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *WindowPoint )( 
            _Window * This,
            /* [in] */ long X,
            /* [in] */ long Y,
            /* [retval][out] */ WindowPoint **point);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *WindowRectangle )( 
            _Window * This,
            /* [in] */ long Left,
            /* [in] */ long Top,
            /* [in] */ long Right,
            /* [in] */ long Bottom,
            /* [retval][out] */ WindowRectangle **rectangle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ClientPoint )( 
            _Window * This,
            /* [in] */ long X,
            /* [in] */ long Y,
            /* [retval][out] */ ClientPoint **point);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ClientRectangle )( 
            _Window * This,
            /* [in] */ long Left,
            /* [in] */ long Top,
            /* [in] */ long Right,
            /* [in] */ long Bottom,
            /* [retval][out] */ ClientRectangle **rectangle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Control )( 
            _Window * This,
            /* [optional][in] */ VARIANT id,
            /* [retval][out] */ IDispatch **Control);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Close )( 
            _Window * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Focus )( 
            _Window * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Activate )( 
            _Window * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SendMessage )( 
            _Window * This,
            /* [in] */ long message,
            /* [in] */ VARIANT wParam,
            /* [in] */ VARIANT lParam,
            /* [retval][out] */ long *result);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *PostMessage )( 
            _Window * This,
            /* [in] */ long message,
            /* [in] */ long wParam,
            /* [in] */ long lParam);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Redraw )( 
            _Window * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *LoadSet )( 
            _Window * This,
            /* [in] */ BSTR SetFileName,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Settings2 )( 
            _Window * This,
            /* [retval][out] */ Settings2 **settings);
        
        END_INTERFACE
    } _WindowVtbl;

    interface _Window
    {
        CONST_VTBL struct _WindowVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _Window_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _Window_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _Window_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _Window_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _Window_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _Window_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _Window_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _Window_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _Window_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _Window_get_Visible(This,Visible)	\
    ( (This)->lpVtbl -> get_Visible(This,Visible) ) 

#define _Window_get_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> get_Enabled(This,Enabled) ) 

#define _Window_put_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> put_Enabled(This,Enabled) ) 

#define _Window_get_Handle(This,Handle)	\
    ( (This)->lpVtbl -> get_Handle(This,Handle) ) 

#define _Window_get_rectangle(This,rectangle)	\
    ( (This)->lpVtbl -> get_rectangle(This,rectangle) ) 

#define _Window_put_rectangle(This,rectangle)	\
    ( (This)->lpVtbl -> put_rectangle(This,rectangle) ) 

#define _Window_get_type(This,type)	\
    ( (This)->lpVtbl -> get_type(This,type) ) 

#define _Window_put_type(This,type)	\
    ( (This)->lpVtbl -> put_type(This,type) ) 

#define _Window_get_OriginalType(This,OriginalType)	\
    ( (This)->lpVtbl -> get_OriginalType(This,OriginalType) ) 

#define _Window_get_DirectChildren(This,DirectChildren)	\
    ( (This)->lpVtbl -> get_DirectChildren(This,DirectChildren) ) 

#define _Window_get_Children(This,Children)	\
    ( (This)->lpVtbl -> get_Children(This,Children) ) 

#define _Window_get_Overlap(This,Overlap)	\
    ( (This)->lpVtbl -> get_Overlap(This,Overlap) ) 

#define _Window_get_Style(This,Style)	\
    ( (This)->lpVtbl -> get_Style(This,Style) ) 

#define _Window_get_CaretWidth(This,Width)	\
    ( (This)->lpVtbl -> get_CaretWidth(This,Width) ) 

#define _Window_get_Name(This,Name)	\
    ( (This)->lpVtbl -> get_Name(This,Name) ) 

#define _Window_get_ClassName(This,ClassName)	\
    ( (This)->lpVtbl -> get_ClassName(This,ClassName) ) 

#define _Window_get_OriginalClassName(This,ClassName)	\
    ( (This)->lpVtbl -> get_OriginalClassName(This,ClassName) ) 

#define _Window_get_ModuleName(This,ModuleName)	\
    ( (This)->lpVtbl -> get_ModuleName(This,ModuleName) ) 

#define _Window_get_Desktop(This,Desktop)	\
    ( (This)->lpVtbl -> get_Desktop(This,Desktop) ) 

#define _Window_get_Process(This,Process)	\
    ( (This)->lpVtbl -> get_Process(This,Process) ) 

#define _Window_get_ClientArea(This,rectangle)	\
    ( (This)->lpVtbl -> get_ClientArea(This,rectangle) ) 

#define _Window_get_Title(This,Title)	\
    ( (This)->lpVtbl -> get_Title(This,Title) ) 

#define _Window_get_Status(This,Status)	\
    ( (This)->lpVtbl -> get_Status(This,Status) ) 

#define _Window_put_Status(This,Status)	\
    ( (This)->lpVtbl -> put_Status(This,Status) ) 

#define _Window_get_SetFileName(This,Name)	\
    ( (This)->lpVtbl -> get_SetFileName(This,Name) ) 

#define _Window_get_Selection(This,RealTime,Selection)	\
    ( (This)->lpVtbl -> get_Selection(This,RealTime,Selection) ) 

#define _Window_get_FieldData(This,Data)	\
    ( (This)->lpVtbl -> get_FieldData(This,Data) ) 

#define _Window_get_FieldName(This,Name)	\
    ( (This)->lpVtbl -> get_FieldName(This,Name) ) 

#define _Window_get_Accessible(This,objectID,Accessible)	\
    ( (This)->lpVtbl -> get_Accessible(This,objectID,Accessible) ) 

#define _Window_get_CurrentUserWindowID(This,uwin)	\
    ( (This)->lpVtbl -> get_CurrentUserWindowID(This,uwin) ) 

#define _Window_put_CurrentUserWindowID(This,uwin)	\
    ( (This)->lpVtbl -> put_CurrentUserWindowID(This,uwin) ) 

#define _Window_get_CurrentHyperactiveWindowID(This,hwin)	\
    ( (This)->lpVtbl -> get_CurrentHyperactiveWindowID(This,hwin) ) 

#define _Window_put_CurrentHyperactiveWindowID(This,hwin)	\
    ( (This)->lpVtbl -> put_CurrentHyperactiveWindowID(This,hwin) ) 

#define _Window_get_NativeObjectModel(This,objectModel)	\
    ( (This)->lpVtbl -> get_NativeObjectModel(This,objectModel) ) 

#define _Window_get_AlwaysOnTop(This,TopMost)	\
    ( (This)->lpVtbl -> get_AlwaysOnTop(This,TopMost) ) 

#define _Window_put_AlwaysOnTop(This,TopMost)	\
    ( (This)->lpVtbl -> put_AlwaysOnTop(This,TopMost) ) 

#define _Window_get_settings(This,settings)	\
    ( (This)->lpVtbl -> get_settings(This,settings) ) 

#define _Window_get_IsValid(This,prop)	\
    ( (This)->lpVtbl -> get_IsValid(This,prop) ) 

#define _Window_get_OnScreenVisibility(This,prop)	\
    ( (This)->lpVtbl -> get_OnScreenVisibility(This,prop) ) 

#define _Window_get_FormatAlertOffset(This,offset)	\
    ( (This)->lpVtbl -> get_FormatAlertOffset(This,offset) ) 

#define _Window_put_FormatAlertOffset(This,offset)	\
    ( (This)->lpVtbl -> put_FormatAlertOffset(This,offset) ) 

#define _Window_Clips(This,IncludeChildren,RealTime,Clips)	\
    ( (This)->lpVtbl -> Clips(This,IncludeChildren,RealTime,Clips) ) 

#define _Window_WindowPoint(This,X,Y,point)	\
    ( (This)->lpVtbl -> WindowPoint(This,X,Y,point) ) 

#define _Window_WindowRectangle(This,Left,Top,Right,Bottom,rectangle)	\
    ( (This)->lpVtbl -> WindowRectangle(This,Left,Top,Right,Bottom,rectangle) ) 

#define _Window_ClientPoint(This,X,Y,point)	\
    ( (This)->lpVtbl -> ClientPoint(This,X,Y,point) ) 

#define _Window_ClientRectangle(This,Left,Top,Right,Bottom,rectangle)	\
    ( (This)->lpVtbl -> ClientRectangle(This,Left,Top,Right,Bottom,rectangle) ) 

#define _Window_Control(This,id,Control)	\
    ( (This)->lpVtbl -> Control(This,id,Control) ) 

#define _Window_Close(This)	\
    ( (This)->lpVtbl -> Close(This) ) 

#define _Window_Focus(This)	\
    ( (This)->lpVtbl -> Focus(This) ) 

#define _Window_Activate(This)	\
    ( (This)->lpVtbl -> Activate(This) ) 

#define _Window_SendMessage(This,message,wParam,lParam,result)	\
    ( (This)->lpVtbl -> SendMessage(This,message,wParam,lParam,result) ) 

#define _Window_PostMessage(This,message,wParam,lParam)	\
    ( (This)->lpVtbl -> PostMessage(This,message,wParam,lParam) ) 

#define _Window_Redraw(This)	\
    ( (This)->lpVtbl -> Redraw(This) ) 

#define _Window_LoadSet(This,SetFileName,success)	\
    ( (This)->lpVtbl -> LoadSet(This,SetFileName,success) ) 

#define _Window_get_Settings2(This,settings)	\
    ( (This)->lpVtbl -> get_Settings2(This,settings) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___Window_INTERFACE_DEFINED__ */


#ifndef __WindowRectangle_INTERFACE_DEFINED__
#define __WindowRectangle_INTERFACE_DEFINED__

/* interface WindowRectangle */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WindowRectangle;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("F0C26627-1E1E-4C1B-8086-A26BFF1290F6")
    WindowRectangle : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Left( 
            /* [retval][out] */ long *Left) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Left( 
            /* [in] */ long Left) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Top( 
            /* [retval][out] */ long *Top) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Top( 
            /* [in] */ long Top) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Right( 
            /* [retval][out] */ long *Right) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Right( 
            /* [in] */ long Right) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Bottom( 
            /* [retval][out] */ long *Bottom) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Bottom( 
            /* [in] */ long Bottom) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Window( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ScreenRectangle( 
            /* [retval][out] */ ScreenRectangle **rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ScreenRectangle( 
            /* [in] */ ScreenRectangle *rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ClientRectangle( 
            /* [retval][out] */ ClientRectangle **rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ClientRectangle( 
            /* [in] */ ClientRectangle *rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE IsPointInRectangle( 
            /* [in] */ WindowPoint *point,
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WindowRectangleVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WindowRectangle * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WindowRectangle * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WindowRectangle * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WindowRectangle * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WindowRectangle * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WindowRectangle * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WindowRectangle * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WindowRectangle * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WindowRectangle * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Left )( 
            WindowRectangle * This,
            /* [retval][out] */ long *Left);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Left )( 
            WindowRectangle * This,
            /* [in] */ long Left);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Top )( 
            WindowRectangle * This,
            /* [retval][out] */ long *Top);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Top )( 
            WindowRectangle * This,
            /* [in] */ long Top);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Right )( 
            WindowRectangle * This,
            /* [retval][out] */ long *Right);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Right )( 
            WindowRectangle * This,
            /* [in] */ long Right);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Bottom )( 
            WindowRectangle * This,
            /* [retval][out] */ long *Bottom);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Bottom )( 
            WindowRectangle * This,
            /* [in] */ long Bottom);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            WindowRectangle * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ScreenRectangle )( 
            WindowRectangle * This,
            /* [retval][out] */ ScreenRectangle **rectangle);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ScreenRectangle )( 
            WindowRectangle * This,
            /* [in] */ ScreenRectangle *rectangle);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ClientRectangle )( 
            WindowRectangle * This,
            /* [retval][out] */ ClientRectangle **rectangle);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ClientRectangle )( 
            WindowRectangle * This,
            /* [in] */ ClientRectangle *rectangle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *IsPointInRectangle )( 
            WindowRectangle * This,
            /* [in] */ WindowPoint *point,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } WindowRectangleVtbl;

    interface WindowRectangle
    {
        CONST_VTBL struct WindowRectangleVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WindowRectangle_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WindowRectangle_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WindowRectangle_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WindowRectangle_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WindowRectangle_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WindowRectangle_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WindowRectangle_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WindowRectangle_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WindowRectangle_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WindowRectangle_get_Left(This,Left)	\
    ( (This)->lpVtbl -> get_Left(This,Left) ) 

#define WindowRectangle_put_Left(This,Left)	\
    ( (This)->lpVtbl -> put_Left(This,Left) ) 

#define WindowRectangle_get_Top(This,Top)	\
    ( (This)->lpVtbl -> get_Top(This,Top) ) 

#define WindowRectangle_put_Top(This,Top)	\
    ( (This)->lpVtbl -> put_Top(This,Top) ) 

#define WindowRectangle_get_Right(This,Right)	\
    ( (This)->lpVtbl -> get_Right(This,Right) ) 

#define WindowRectangle_put_Right(This,Right)	\
    ( (This)->lpVtbl -> put_Right(This,Right) ) 

#define WindowRectangle_get_Bottom(This,Bottom)	\
    ( (This)->lpVtbl -> get_Bottom(This,Bottom) ) 

#define WindowRectangle_put_Bottom(This,Bottom)	\
    ( (This)->lpVtbl -> put_Bottom(This,Bottom) ) 

#define WindowRectangle_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define WindowRectangle_get_ScreenRectangle(This,rectangle)	\
    ( (This)->lpVtbl -> get_ScreenRectangle(This,rectangle) ) 

#define WindowRectangle_put_ScreenRectangle(This,rectangle)	\
    ( (This)->lpVtbl -> put_ScreenRectangle(This,rectangle) ) 

#define WindowRectangle_get_ClientRectangle(This,rectangle)	\
    ( (This)->lpVtbl -> get_ClientRectangle(This,rectangle) ) 

#define WindowRectangle_put_ClientRectangle(This,rectangle)	\
    ( (This)->lpVtbl -> put_ClientRectangle(This,rectangle) ) 

#define WindowRectangle_IsPointInRectangle(This,point,prop)	\
    ( (This)->lpVtbl -> IsPointInRectangle(This,point,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WindowRectangle_INTERFACE_DEFINED__ */


#ifndef __ScreenRectangle_INTERFACE_DEFINED__
#define __ScreenRectangle_INTERFACE_DEFINED__

/* interface ScreenRectangle */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ScreenRectangle;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("D75594B7-6DF2-4FB6-A001-28E93EFF520C")
    ScreenRectangle : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Left( 
            /* [retval][out] */ long *Left) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Left( 
            /* [in] */ long Left) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Top( 
            /* [retval][out] */ long *Top) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Top( 
            /* [in] */ long Top) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Right( 
            /* [retval][out] */ long *Right) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Right( 
            /* [in] */ long Right) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Bottom( 
            /* [retval][out] */ long *Bottom) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Bottom( 
            /* [in] */ long Bottom) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsEmpty( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE WindowRectangle( 
            /* [in] */ _Window *Window,
            /* [retval][out] */ WindowRectangle **rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ClientRectangle( 
            /* [in] */ _Window *Window,
            /* [retval][out] */ ClientRectangle **rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE IsPointInRectangle( 
            /* [in] */ ScreenPoint *point,
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Intersects( 
            /* [in] */ ScreenRectangle *rect,
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Intersection( 
            /* [in] */ ScreenRectangle *rectangle,
            /* [retval][out] */ ScreenRectangle **Intersection) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Equals( 
            /* [in] */ ScreenRectangle *TargetRectangle,
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Contains( 
            /* [in] */ ScreenRectangle *TargetRectangle,
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ScreenRectangleVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ScreenRectangle * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ScreenRectangle * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ScreenRectangle * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ScreenRectangle * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ScreenRectangle * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ScreenRectangle * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ScreenRectangle * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ScreenRectangle * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ScreenRectangle * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Left )( 
            ScreenRectangle * This,
            /* [retval][out] */ long *Left);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Left )( 
            ScreenRectangle * This,
            /* [in] */ long Left);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Top )( 
            ScreenRectangle * This,
            /* [retval][out] */ long *Top);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Top )( 
            ScreenRectangle * This,
            /* [in] */ long Top);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Right )( 
            ScreenRectangle * This,
            /* [retval][out] */ long *Right);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Right )( 
            ScreenRectangle * This,
            /* [in] */ long Right);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Bottom )( 
            ScreenRectangle * This,
            /* [retval][out] */ long *Bottom);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Bottom )( 
            ScreenRectangle * This,
            /* [in] */ long Bottom);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsEmpty )( 
            ScreenRectangle * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *WindowRectangle )( 
            ScreenRectangle * This,
            /* [in] */ _Window *Window,
            /* [retval][out] */ WindowRectangle **rectangle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ClientRectangle )( 
            ScreenRectangle * This,
            /* [in] */ _Window *Window,
            /* [retval][out] */ ClientRectangle **rectangle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *IsPointInRectangle )( 
            ScreenRectangle * This,
            /* [in] */ ScreenPoint *point,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Intersects )( 
            ScreenRectangle * This,
            /* [in] */ ScreenRectangle *rect,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Intersection )( 
            ScreenRectangle * This,
            /* [in] */ ScreenRectangle *rectangle,
            /* [retval][out] */ ScreenRectangle **Intersection);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Equals )( 
            ScreenRectangle * This,
            /* [in] */ ScreenRectangle *TargetRectangle,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Contains )( 
            ScreenRectangle * This,
            /* [in] */ ScreenRectangle *TargetRectangle,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } ScreenRectangleVtbl;

    interface ScreenRectangle
    {
        CONST_VTBL struct ScreenRectangleVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ScreenRectangle_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ScreenRectangle_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ScreenRectangle_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ScreenRectangle_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ScreenRectangle_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ScreenRectangle_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ScreenRectangle_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ScreenRectangle_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ScreenRectangle_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ScreenRectangle_get_Left(This,Left)	\
    ( (This)->lpVtbl -> get_Left(This,Left) ) 

#define ScreenRectangle_put_Left(This,Left)	\
    ( (This)->lpVtbl -> put_Left(This,Left) ) 

#define ScreenRectangle_get_Top(This,Top)	\
    ( (This)->lpVtbl -> get_Top(This,Top) ) 

#define ScreenRectangle_put_Top(This,Top)	\
    ( (This)->lpVtbl -> put_Top(This,Top) ) 

#define ScreenRectangle_get_Right(This,Right)	\
    ( (This)->lpVtbl -> get_Right(This,Right) ) 

#define ScreenRectangle_put_Right(This,Right)	\
    ( (This)->lpVtbl -> put_Right(This,Right) ) 

#define ScreenRectangle_get_Bottom(This,Bottom)	\
    ( (This)->lpVtbl -> get_Bottom(This,Bottom) ) 

#define ScreenRectangle_put_Bottom(This,Bottom)	\
    ( (This)->lpVtbl -> put_Bottom(This,Bottom) ) 

#define ScreenRectangle_get_IsEmpty(This,prop)	\
    ( (This)->lpVtbl -> get_IsEmpty(This,prop) ) 

#define ScreenRectangle_WindowRectangle(This,Window,rectangle)	\
    ( (This)->lpVtbl -> WindowRectangle(This,Window,rectangle) ) 

#define ScreenRectangle_ClientRectangle(This,Window,rectangle)	\
    ( (This)->lpVtbl -> ClientRectangle(This,Window,rectangle) ) 

#define ScreenRectangle_IsPointInRectangle(This,point,prop)	\
    ( (This)->lpVtbl -> IsPointInRectangle(This,point,prop) ) 

#define ScreenRectangle_Intersects(This,rect,prop)	\
    ( (This)->lpVtbl -> Intersects(This,rect,prop) ) 

#define ScreenRectangle_Intersection(This,rectangle,Intersection)	\
    ( (This)->lpVtbl -> Intersection(This,rectangle,Intersection) ) 

#define ScreenRectangle_Equals(This,TargetRectangle,prop)	\
    ( (This)->lpVtbl -> Equals(This,TargetRectangle,prop) ) 

#define ScreenRectangle_Contains(This,TargetRectangle,prop)	\
    ( (This)->lpVtbl -> Contains(This,TargetRectangle,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ScreenRectangle_INTERFACE_DEFINED__ */


#ifndef __ClientRectangle_INTERFACE_DEFINED__
#define __ClientRectangle_INTERFACE_DEFINED__

/* interface ClientRectangle */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ClientRectangle;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("BC5BEE46-5487-4DDE-ADFD-BE2725307C0B")
    ClientRectangle : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Left( 
            /* [retval][out] */ long *Left) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Left( 
            /* [in] */ long Left) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Top( 
            /* [retval][out] */ long *Top) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Top( 
            /* [in] */ long Top) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Right( 
            /* [retval][out] */ long *Right) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Right( 
            /* [in] */ long Right) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Bottom( 
            /* [retval][out] */ long *Bottom) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Bottom( 
            /* [in] */ long Bottom) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Window( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ScreenRectangle( 
            /* [retval][out] */ ScreenRectangle **rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ScreenRectangle( 
            /* [in] */ ScreenRectangle *rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WindowRectangle( 
            /* [retval][out] */ WindowRectangle **rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_WindowRectangle( 
            /* [in] */ WindowRectangle *rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE IsPointInRectangle( 
            /* [in] */ ClientPoint *point,
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ClientRectangleVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ClientRectangle * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ClientRectangle * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ClientRectangle * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ClientRectangle * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ClientRectangle * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ClientRectangle * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ClientRectangle * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ClientRectangle * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ClientRectangle * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Left )( 
            ClientRectangle * This,
            /* [retval][out] */ long *Left);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Left )( 
            ClientRectangle * This,
            /* [in] */ long Left);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Top )( 
            ClientRectangle * This,
            /* [retval][out] */ long *Top);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Top )( 
            ClientRectangle * This,
            /* [in] */ long Top);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Right )( 
            ClientRectangle * This,
            /* [retval][out] */ long *Right);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Right )( 
            ClientRectangle * This,
            /* [in] */ long Right);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Bottom )( 
            ClientRectangle * This,
            /* [retval][out] */ long *Bottom);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Bottom )( 
            ClientRectangle * This,
            /* [in] */ long Bottom);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            ClientRectangle * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ScreenRectangle )( 
            ClientRectangle * This,
            /* [retval][out] */ ScreenRectangle **rectangle);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ScreenRectangle )( 
            ClientRectangle * This,
            /* [in] */ ScreenRectangle *rectangle);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WindowRectangle )( 
            ClientRectangle * This,
            /* [retval][out] */ WindowRectangle **rectangle);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_WindowRectangle )( 
            ClientRectangle * This,
            /* [in] */ WindowRectangle *rectangle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *IsPointInRectangle )( 
            ClientRectangle * This,
            /* [in] */ ClientPoint *point,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } ClientRectangleVtbl;

    interface ClientRectangle
    {
        CONST_VTBL struct ClientRectangleVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ClientRectangle_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ClientRectangle_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ClientRectangle_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ClientRectangle_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ClientRectangle_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ClientRectangle_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ClientRectangle_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ClientRectangle_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ClientRectangle_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ClientRectangle_get_Left(This,Left)	\
    ( (This)->lpVtbl -> get_Left(This,Left) ) 

#define ClientRectangle_put_Left(This,Left)	\
    ( (This)->lpVtbl -> put_Left(This,Left) ) 

#define ClientRectangle_get_Top(This,Top)	\
    ( (This)->lpVtbl -> get_Top(This,Top) ) 

#define ClientRectangle_put_Top(This,Top)	\
    ( (This)->lpVtbl -> put_Top(This,Top) ) 

#define ClientRectangle_get_Right(This,Right)	\
    ( (This)->lpVtbl -> get_Right(This,Right) ) 

#define ClientRectangle_put_Right(This,Right)	\
    ( (This)->lpVtbl -> put_Right(This,Right) ) 

#define ClientRectangle_get_Bottom(This,Bottom)	\
    ( (This)->lpVtbl -> get_Bottom(This,Bottom) ) 

#define ClientRectangle_put_Bottom(This,Bottom)	\
    ( (This)->lpVtbl -> put_Bottom(This,Bottom) ) 

#define ClientRectangle_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define ClientRectangle_get_ScreenRectangle(This,rectangle)	\
    ( (This)->lpVtbl -> get_ScreenRectangle(This,rectangle) ) 

#define ClientRectangle_put_ScreenRectangle(This,rectangle)	\
    ( (This)->lpVtbl -> put_ScreenRectangle(This,rectangle) ) 

#define ClientRectangle_get_WindowRectangle(This,rectangle)	\
    ( (This)->lpVtbl -> get_WindowRectangle(This,rectangle) ) 

#define ClientRectangle_put_WindowRectangle(This,rectangle)	\
    ( (This)->lpVtbl -> put_WindowRectangle(This,rectangle) ) 

#define ClientRectangle_IsPointInRectangle(This,point,prop)	\
    ( (This)->lpVtbl -> IsPointInRectangle(This,point,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ClientRectangle_INTERFACE_DEFINED__ */


#ifndef __ClientPoint_INTERFACE_DEFINED__
#define __ClientPoint_INTERFACE_DEFINED__

/* interface ClientPoint */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ClientPoint;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("38373772-1B30-4C81-90BE-12AC169ECC79")
    ClientPoint : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_X( 
            /* [retval][out] */ long *X) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_X( 
            /* [in] */ long X) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Y( 
            /* [retval][out] */ long *Y) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Y( 
            /* [in] */ long Y) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Window( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ScreenPoint( 
            /* [retval][out] */ ScreenPoint **point) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ScreenPoint( 
            /* [in] */ ScreenPoint *point) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WindowPoint( 
            /* [retval][out] */ WindowPoint **point) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_WindowPoint( 
            /* [in] */ WindowPoint *point) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ClientPointVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ClientPoint * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ClientPoint * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ClientPoint * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ClientPoint * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ClientPoint * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ClientPoint * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ClientPoint * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ClientPoint * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ClientPoint * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_X )( 
            ClientPoint * This,
            /* [retval][out] */ long *X);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_X )( 
            ClientPoint * This,
            /* [in] */ long X);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Y )( 
            ClientPoint * This,
            /* [retval][out] */ long *Y);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Y )( 
            ClientPoint * This,
            /* [in] */ long Y);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            ClientPoint * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ScreenPoint )( 
            ClientPoint * This,
            /* [retval][out] */ ScreenPoint **point);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ScreenPoint )( 
            ClientPoint * This,
            /* [in] */ ScreenPoint *point);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WindowPoint )( 
            ClientPoint * This,
            /* [retval][out] */ WindowPoint **point);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_WindowPoint )( 
            ClientPoint * This,
            /* [in] */ WindowPoint *point);
        
        END_INTERFACE
    } ClientPointVtbl;

    interface ClientPoint
    {
        CONST_VTBL struct ClientPointVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ClientPoint_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ClientPoint_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ClientPoint_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ClientPoint_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ClientPoint_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ClientPoint_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ClientPoint_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ClientPoint_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ClientPoint_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ClientPoint_get_X(This,X)	\
    ( (This)->lpVtbl -> get_X(This,X) ) 

#define ClientPoint_put_X(This,X)	\
    ( (This)->lpVtbl -> put_X(This,X) ) 

#define ClientPoint_get_Y(This,Y)	\
    ( (This)->lpVtbl -> get_Y(This,Y) ) 

#define ClientPoint_put_Y(This,Y)	\
    ( (This)->lpVtbl -> put_Y(This,Y) ) 

#define ClientPoint_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define ClientPoint_get_ScreenPoint(This,point)	\
    ( (This)->lpVtbl -> get_ScreenPoint(This,point) ) 

#define ClientPoint_put_ScreenPoint(This,point)	\
    ( (This)->lpVtbl -> put_ScreenPoint(This,point) ) 

#define ClientPoint_get_WindowPoint(This,point)	\
    ( (This)->lpVtbl -> get_WindowPoint(This,point) ) 

#define ClientPoint_put_WindowPoint(This,point)	\
    ( (This)->lpVtbl -> put_WindowPoint(This,point) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ClientPoint_INTERFACE_DEFINED__ */


#ifndef __ScreenPoint_INTERFACE_DEFINED__
#define __ScreenPoint_INTERFACE_DEFINED__

/* interface ScreenPoint */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ScreenPoint;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("EC281A02-CAA5-4BBF-A7D3-1EBFFCE77444")
    ScreenPoint : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_X( 
            /* [retval][out] */ long *X) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_X( 
            /* [in] */ long X) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Y( 
            /* [retval][out] */ long *Y) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Y( 
            /* [in] */ long Y) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Accessible( 
            /* [retval][out] */ Accessible **Accessible) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Window( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE WindowPoint( 
            /* [in] */ _Window *Window,
            /* [retval][out] */ WindowPoint **point) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ClientPoint( 
            /* [in] */ _Window *Window,
            /* [retval][out] */ ClientPoint **point) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ScreenPointVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ScreenPoint * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ScreenPoint * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ScreenPoint * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ScreenPoint * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ScreenPoint * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ScreenPoint * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ScreenPoint * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ScreenPoint * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ScreenPoint * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_X )( 
            ScreenPoint * This,
            /* [retval][out] */ long *X);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_X )( 
            ScreenPoint * This,
            /* [in] */ long X);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Y )( 
            ScreenPoint * This,
            /* [retval][out] */ long *Y);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Y )( 
            ScreenPoint * This,
            /* [in] */ long Y);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Accessible )( 
            ScreenPoint * This,
            /* [retval][out] */ Accessible **Accessible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            ScreenPoint * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *WindowPoint )( 
            ScreenPoint * This,
            /* [in] */ _Window *Window,
            /* [retval][out] */ WindowPoint **point);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ClientPoint )( 
            ScreenPoint * This,
            /* [in] */ _Window *Window,
            /* [retval][out] */ ClientPoint **point);
        
        END_INTERFACE
    } ScreenPointVtbl;

    interface ScreenPoint
    {
        CONST_VTBL struct ScreenPointVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ScreenPoint_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ScreenPoint_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ScreenPoint_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ScreenPoint_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ScreenPoint_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ScreenPoint_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ScreenPoint_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ScreenPoint_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ScreenPoint_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ScreenPoint_get_X(This,X)	\
    ( (This)->lpVtbl -> get_X(This,X) ) 

#define ScreenPoint_put_X(This,X)	\
    ( (This)->lpVtbl -> put_X(This,X) ) 

#define ScreenPoint_get_Y(This,Y)	\
    ( (This)->lpVtbl -> get_Y(This,Y) ) 

#define ScreenPoint_put_Y(This,Y)	\
    ( (This)->lpVtbl -> put_Y(This,Y) ) 

#define ScreenPoint_get_Accessible(This,Accessible)	\
    ( (This)->lpVtbl -> get_Accessible(This,Accessible) ) 

#define ScreenPoint_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define ScreenPoint_WindowPoint(This,Window,point)	\
    ( (This)->lpVtbl -> WindowPoint(This,Window,point) ) 

#define ScreenPoint_ClientPoint(This,Window,point)	\
    ( (This)->lpVtbl -> ClientPoint(This,Window,point) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ScreenPoint_INTERFACE_DEFINED__ */


#ifndef __Accessible_INTERFACE_DEFINED__
#define __Accessible_INTERFACE_DEFINED__

/* interface Accessible */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Accessible;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("1291AFC2-AD5A-4CC2-B763-A440EC9D8B18")
    Accessible : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Value( 
            /* [retval][out] */ BSTR *Value) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Value( 
            /* [in] */ BSTR Value) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Down( 
            /* [retval][out] */ Accessible **Down) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Up( 
            /* [retval][out] */ Accessible **Up) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Left( 
            /* [retval][out] */ Accessible **Left) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Right( 
            /* [retval][out] */ Accessible **Right) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Next( 
            /* [retval][out] */ Accessible **Next) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Previous( 
            /* [retval][out] */ Accessible **Previous) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FirstChild( 
            /* [retval][out] */ Accessible **FirstChild) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LastChild( 
            /* [retval][out] */ Accessible **LastChild) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DefaultAction( 
            /* [retval][out] */ BSTR *DefaultAction) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Description( 
            /* [retval][out] */ BSTR *Description) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Help( 
            /* [retval][out] */ BSTR *Help) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Name( 
            /* [retval][out] */ BSTR *Name) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_KeyboardShortcut( 
            /* [retval][out] */ BSTR *KeyboardShortcut) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Focus( 
            /* [retval][out] */ Accessible **Focus) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Selection( 
            /* [retval][out] */ Accessibles **Selection) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_role( 
            /* [retval][out] */ AccessibleRole **role) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_state( 
            /* [retval][out] */ AccessibleState **state) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Location( 
            /* [retval][out] */ ScreenRectangle **Location) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Children( 
            /* [retval][out] */ Accessibles **Children) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Window( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IAccessible( 
            /* [retval][out] */ IDispatch **iacc) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ChildID( 
            /* [retval][out] */ VARIANT *ChildID) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EventWindow( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE DoDefaultAction( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE HitTest( 
            /* [in] */ ScreenPoint *point,
            /* [retval][out] */ Accessible **child) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Select( 
            /* [in] */ AccessibleSelection type) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SimulateEvent( 
            /* [in] */ MSAAEventID id,
            /* [optional][in] */ VARIANT values) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Prefetch( 
            /* [in] */ AccessibleProperty values) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct AccessibleVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Accessible * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Accessible * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Accessible * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Accessible * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Accessible * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Accessible * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Accessible * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Value )( 
            Accessible * This,
            /* [retval][out] */ BSTR *Value);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Value )( 
            Accessible * This,
            /* [in] */ BSTR Value);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Accessible * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Accessible * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Down )( 
            Accessible * This,
            /* [retval][out] */ Accessible **Down);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Up )( 
            Accessible * This,
            /* [retval][out] */ Accessible **Up);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Left )( 
            Accessible * This,
            /* [retval][out] */ Accessible **Left);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Right )( 
            Accessible * This,
            /* [retval][out] */ Accessible **Right);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Next )( 
            Accessible * This,
            /* [retval][out] */ Accessible **Next);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Previous )( 
            Accessible * This,
            /* [retval][out] */ Accessible **Previous);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FirstChild )( 
            Accessible * This,
            /* [retval][out] */ Accessible **FirstChild);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LastChild )( 
            Accessible * This,
            /* [retval][out] */ Accessible **LastChild);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DefaultAction )( 
            Accessible * This,
            /* [retval][out] */ BSTR *DefaultAction);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Description )( 
            Accessible * This,
            /* [retval][out] */ BSTR *Description);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Help )( 
            Accessible * This,
            /* [retval][out] */ BSTR *Help);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Name )( 
            Accessible * This,
            /* [retval][out] */ BSTR *Name);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_KeyboardShortcut )( 
            Accessible * This,
            /* [retval][out] */ BSTR *KeyboardShortcut);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focus )( 
            Accessible * This,
            /* [retval][out] */ Accessible **Focus);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Selection )( 
            Accessible * This,
            /* [retval][out] */ Accessibles **Selection);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_role )( 
            Accessible * This,
            /* [retval][out] */ AccessibleRole **role);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_state )( 
            Accessible * This,
            /* [retval][out] */ AccessibleState **state);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Location )( 
            Accessible * This,
            /* [retval][out] */ ScreenRectangle **Location);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Children )( 
            Accessible * This,
            /* [retval][out] */ Accessibles **Children);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            Accessible * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IAccessible )( 
            Accessible * This,
            /* [retval][out] */ IDispatch **iacc);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ChildID )( 
            Accessible * This,
            /* [retval][out] */ VARIANT *ChildID);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EventWindow )( 
            Accessible * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *DoDefaultAction )( 
            Accessible * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *HitTest )( 
            Accessible * This,
            /* [in] */ ScreenPoint *point,
            /* [retval][out] */ Accessible **child);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Select )( 
            Accessible * This,
            /* [in] */ AccessibleSelection type);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SimulateEvent )( 
            Accessible * This,
            /* [in] */ MSAAEventID id,
            /* [optional][in] */ VARIANT values);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Prefetch )( 
            Accessible * This,
            /* [in] */ AccessibleProperty values);
        
        END_INTERFACE
    } AccessibleVtbl;

    interface Accessible
    {
        CONST_VTBL struct AccessibleVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Accessible_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Accessible_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Accessible_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Accessible_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Accessible_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Accessible_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Accessible_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Accessible_get_Value(This,Value)	\
    ( (This)->lpVtbl -> get_Value(This,Value) ) 

#define Accessible_put_Value(This,Value)	\
    ( (This)->lpVtbl -> put_Value(This,Value) ) 

#define Accessible_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Accessible_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Accessible_get_Down(This,Down)	\
    ( (This)->lpVtbl -> get_Down(This,Down) ) 

#define Accessible_get_Up(This,Up)	\
    ( (This)->lpVtbl -> get_Up(This,Up) ) 

#define Accessible_get_Left(This,Left)	\
    ( (This)->lpVtbl -> get_Left(This,Left) ) 

#define Accessible_get_Right(This,Right)	\
    ( (This)->lpVtbl -> get_Right(This,Right) ) 

#define Accessible_get_Next(This,Next)	\
    ( (This)->lpVtbl -> get_Next(This,Next) ) 

#define Accessible_get_Previous(This,Previous)	\
    ( (This)->lpVtbl -> get_Previous(This,Previous) ) 

#define Accessible_get_FirstChild(This,FirstChild)	\
    ( (This)->lpVtbl -> get_FirstChild(This,FirstChild) ) 

#define Accessible_get_LastChild(This,LastChild)	\
    ( (This)->lpVtbl -> get_LastChild(This,LastChild) ) 

#define Accessible_get_DefaultAction(This,DefaultAction)	\
    ( (This)->lpVtbl -> get_DefaultAction(This,DefaultAction) ) 

#define Accessible_get_Description(This,Description)	\
    ( (This)->lpVtbl -> get_Description(This,Description) ) 

#define Accessible_get_Help(This,Help)	\
    ( (This)->lpVtbl -> get_Help(This,Help) ) 

#define Accessible_get_Name(This,Name)	\
    ( (This)->lpVtbl -> get_Name(This,Name) ) 

#define Accessible_get_KeyboardShortcut(This,KeyboardShortcut)	\
    ( (This)->lpVtbl -> get_KeyboardShortcut(This,KeyboardShortcut) ) 

#define Accessible_get_Focus(This,Focus)	\
    ( (This)->lpVtbl -> get_Focus(This,Focus) ) 

#define Accessible_get_Selection(This,Selection)	\
    ( (This)->lpVtbl -> get_Selection(This,Selection) ) 

#define Accessible_get_role(This,role)	\
    ( (This)->lpVtbl -> get_role(This,role) ) 

#define Accessible_get_state(This,state)	\
    ( (This)->lpVtbl -> get_state(This,state) ) 

#define Accessible_get_Location(This,Location)	\
    ( (This)->lpVtbl -> get_Location(This,Location) ) 

#define Accessible_get_Children(This,Children)	\
    ( (This)->lpVtbl -> get_Children(This,Children) ) 

#define Accessible_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define Accessible_get_IAccessible(This,iacc)	\
    ( (This)->lpVtbl -> get_IAccessible(This,iacc) ) 

#define Accessible_get_ChildID(This,ChildID)	\
    ( (This)->lpVtbl -> get_ChildID(This,ChildID) ) 

#define Accessible_get_EventWindow(This,Window)	\
    ( (This)->lpVtbl -> get_EventWindow(This,Window) ) 

#define Accessible_DoDefaultAction(This)	\
    ( (This)->lpVtbl -> DoDefaultAction(This) ) 

#define Accessible_HitTest(This,point,child)	\
    ( (This)->lpVtbl -> HitTest(This,point,child) ) 

#define Accessible_Select(This,type)	\
    ( (This)->lpVtbl -> Select(This,type) ) 

#define Accessible_SimulateEvent(This,id,values)	\
    ( (This)->lpVtbl -> SimulateEvent(This,id,values) ) 

#define Accessible_Prefetch(This,values)	\
    ( (This)->lpVtbl -> Prefetch(This,values) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Accessible_INTERFACE_DEFINED__ */


#ifndef __Accessibles_INTERFACE_DEFINED__
#define __Accessibles_INTERFACE_DEFINED__

/* interface Accessibles */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Accessibles;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("50660384-FD5C-4D0E-9AAF-720CD7B0E278")
    Accessibles : public IDispatch
    {
    public:
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ Accessible **Accessible) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE FilterByRoleAndState( 
            /* [optional][in] */ VARIANT FilterRole,
            /* [optional][in] */ VARIANT FilterState,
            /* [optional][in] */ VARIANT MatchStateExactly,
            /* [retval][out] */ Accessibles **Accessibles) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct AccessiblesVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Accessibles * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Accessibles * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Accessibles * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Accessibles * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Accessibles * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Accessibles * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Accessibles * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            Accessibles * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Accessibles * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Accessibles * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            Accessibles * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            Accessibles * This,
            /* [in] */ long index,
            /* [retval][out] */ Accessible **Accessible);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *FilterByRoleAndState )( 
            Accessibles * This,
            /* [optional][in] */ VARIANT FilterRole,
            /* [optional][in] */ VARIANT FilterState,
            /* [optional][in] */ VARIANT MatchStateExactly,
            /* [retval][out] */ Accessibles **Accessibles);
        
        END_INTERFACE
    } AccessiblesVtbl;

    interface Accessibles
    {
        CONST_VTBL struct AccessiblesVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Accessibles_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Accessibles_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Accessibles_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Accessibles_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Accessibles_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Accessibles_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Accessibles_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Accessibles_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define Accessibles_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Accessibles_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Accessibles_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define Accessibles_Item(This,index,Accessible)	\
    ( (This)->lpVtbl -> Item(This,index,Accessible) ) 

#define Accessibles_FilterByRoleAndState(This,FilterRole,FilterState,MatchStateExactly,Accessibles)	\
    ( (This)->lpVtbl -> FilterByRoleAndState(This,FilterRole,FilterState,MatchStateExactly,Accessibles) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Accessibles_INTERFACE_DEFINED__ */


#ifndef __AccessibleRole_INTERFACE_DEFINED__
#define __AccessibleRole_INTERFACE_DEFINED__

/* interface AccessibleRole */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_AccessibleRole;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("67C68BDD-7CA8-4877-87F1-D137A367EB4B")
    AccessibleRole : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Text( 
            /* [retval][out] */ BSTR *Text) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Value( 
            /* [retval][out] */ AccessibleRoleEnum *role) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct AccessibleRoleVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            AccessibleRole * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            AccessibleRole * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            AccessibleRole * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            AccessibleRole * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            AccessibleRole * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            AccessibleRole * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            AccessibleRole * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            AccessibleRole * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            AccessibleRole * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            AccessibleRole * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Value )( 
            AccessibleRole * This,
            /* [retval][out] */ AccessibleRoleEnum *role);
        
        END_INTERFACE
    } AccessibleRoleVtbl;

    interface AccessibleRole
    {
        CONST_VTBL struct AccessibleRoleVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define AccessibleRole_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define AccessibleRole_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define AccessibleRole_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define AccessibleRole_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define AccessibleRole_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define AccessibleRole_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define AccessibleRole_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define AccessibleRole_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define AccessibleRole_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define AccessibleRole_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define AccessibleRole_get_Value(This,role)	\
    ( (This)->lpVtbl -> get_Value(This,role) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __AccessibleRole_INTERFACE_DEFINED__ */


#ifndef __AccessibleState_INTERFACE_DEFINED__
#define __AccessibleState_INTERFACE_DEFINED__

/* interface AccessibleState */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_AccessibleState;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("0CF3CFCE-234A-4AD7-A012-BC1EFADBAAE0")
    AccessibleState : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Text( 
            /* [retval][out] */ BSTR *Text) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Value( 
            /* [retval][out] */ AccessibleStateEnum *state) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Animated( 
            /* [retval][out] */ VARIANT_BOOL *Animated) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Busy( 
            /* [retval][out] */ VARIANT_BOOL *Animated) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Checked( 
            /* [retval][out] */ VARIANT_BOOL *Checked) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Collapsed( 
            /* [retval][out] */ VARIANT_BOOL *Collapsed) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Default( 
            /* [retval][out] */ VARIANT_BOOL *defaultButton) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Expanded( 
            /* [retval][out] */ VARIANT_BOOL *Expanded) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ExtendedSelectable( 
            /* [retval][out] */ VARIANT_BOOL *ExtendedSelectable) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Focusable( 
            /* [retval][out] */ VARIANT_BOOL *Focusable) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Focused( 
            /* [retval][out] */ VARIANT_BOOL *Focused) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HasPopup( 
            /* [retval][out] */ VARIANT_BOOL *HasPopup) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HotTracked( 
            /* [retval][out] */ VARIANT_BOOL *HotTracked) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Invisible( 
            /* [retval][out] */ VARIANT_BOOL *Invisible) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Linked( 
            /* [retval][out] */ VARIANT_BOOL *Linked) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Marqueed( 
            /* [retval][out] */ VARIANT_BOOL *Marqueed) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Mixed( 
            /* [retval][out] */ VARIANT_BOOL *Mixed) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Moveable( 
            /* [retval][out] */ VARIANT_BOOL *Moveable) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MultiSelectable( 
            /* [retval][out] */ VARIANT_BOOL *MultiSelectable) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Offscreen( 
            /* [retval][out] */ VARIANT_BOOL *Offscreen) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Pressed( 
            /* [retval][out] */ VARIANT_BOOL *Pressed) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Protected( 
            /* [retval][out] */ VARIANT_BOOL *protectedEdit) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadOnly( 
            /* [retval][out] */ VARIANT_BOOL *ReadOnly) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Selectable( 
            /* [retval][out] */ VARIANT_BOOL *Selectable) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Selected( 
            /* [retval][out] */ VARIANT_BOOL *Selected) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SelfVoicing( 
            /* [retval][out] */ VARIANT_BOOL *SelfVoicing) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Sizeable( 
            /* [retval][out] */ VARIANT_BOOL *Sizeable) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Traversed( 
            /* [retval][out] */ VARIANT_BOOL *Traversed) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Unavailable( 
            /* [retval][out] */ VARIANT_BOOL *Unavailable) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct AccessibleStateVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            AccessibleState * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            AccessibleState * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            AccessibleState * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            AccessibleState * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            AccessibleState * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            AccessibleState * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            AccessibleState * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            AccessibleState * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            AccessibleState * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            AccessibleState * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Value )( 
            AccessibleState * This,
            /* [retval][out] */ AccessibleStateEnum *state);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Animated )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Animated);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Busy )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Animated);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Checked )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Checked);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Collapsed )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Collapsed);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Default )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *defaultButton);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Expanded )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Expanded);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ExtendedSelectable )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *ExtendedSelectable);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focusable )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Focusable);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Focused);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HasPopup )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *HasPopup);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HotTracked )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *HotTracked);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Invisible )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Invisible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Linked )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Linked);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Marqueed )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Marqueed);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Mixed )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Mixed);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Moveable )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Moveable);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MultiSelectable )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *MultiSelectable);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Offscreen )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Offscreen);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Pressed )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Pressed);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Protected )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *protectedEdit);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadOnly )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *ReadOnly);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Selectable )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Selectable);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Selected )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Selected);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SelfVoicing )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *SelfVoicing);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Sizeable )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Sizeable);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Traversed )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Traversed);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Unavailable )( 
            AccessibleState * This,
            /* [retval][out] */ VARIANT_BOOL *Unavailable);
        
        END_INTERFACE
    } AccessibleStateVtbl;

    interface AccessibleState
    {
        CONST_VTBL struct AccessibleStateVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define AccessibleState_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define AccessibleState_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define AccessibleState_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define AccessibleState_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define AccessibleState_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define AccessibleState_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define AccessibleState_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define AccessibleState_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define AccessibleState_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define AccessibleState_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define AccessibleState_get_Value(This,state)	\
    ( (This)->lpVtbl -> get_Value(This,state) ) 

#define AccessibleState_get_Animated(This,Animated)	\
    ( (This)->lpVtbl -> get_Animated(This,Animated) ) 

#define AccessibleState_get_Busy(This,Animated)	\
    ( (This)->lpVtbl -> get_Busy(This,Animated) ) 

#define AccessibleState_get_Checked(This,Checked)	\
    ( (This)->lpVtbl -> get_Checked(This,Checked) ) 

#define AccessibleState_get_Collapsed(This,Collapsed)	\
    ( (This)->lpVtbl -> get_Collapsed(This,Collapsed) ) 

#define AccessibleState_get_Default(This,defaultButton)	\
    ( (This)->lpVtbl -> get_Default(This,defaultButton) ) 

#define AccessibleState_get_Expanded(This,Expanded)	\
    ( (This)->lpVtbl -> get_Expanded(This,Expanded) ) 

#define AccessibleState_get_ExtendedSelectable(This,ExtendedSelectable)	\
    ( (This)->lpVtbl -> get_ExtendedSelectable(This,ExtendedSelectable) ) 

#define AccessibleState_get_Focusable(This,Focusable)	\
    ( (This)->lpVtbl -> get_Focusable(This,Focusable) ) 

#define AccessibleState_get_Focused(This,Focused)	\
    ( (This)->lpVtbl -> get_Focused(This,Focused) ) 

#define AccessibleState_get_HasPopup(This,HasPopup)	\
    ( (This)->lpVtbl -> get_HasPopup(This,HasPopup) ) 

#define AccessibleState_get_HotTracked(This,HotTracked)	\
    ( (This)->lpVtbl -> get_HotTracked(This,HotTracked) ) 

#define AccessibleState_get_Invisible(This,Invisible)	\
    ( (This)->lpVtbl -> get_Invisible(This,Invisible) ) 

#define AccessibleState_get_Linked(This,Linked)	\
    ( (This)->lpVtbl -> get_Linked(This,Linked) ) 

#define AccessibleState_get_Marqueed(This,Marqueed)	\
    ( (This)->lpVtbl -> get_Marqueed(This,Marqueed) ) 

#define AccessibleState_get_Mixed(This,Mixed)	\
    ( (This)->lpVtbl -> get_Mixed(This,Mixed) ) 

#define AccessibleState_get_Moveable(This,Moveable)	\
    ( (This)->lpVtbl -> get_Moveable(This,Moveable) ) 

#define AccessibleState_get_MultiSelectable(This,MultiSelectable)	\
    ( (This)->lpVtbl -> get_MultiSelectable(This,MultiSelectable) ) 

#define AccessibleState_get_Offscreen(This,Offscreen)	\
    ( (This)->lpVtbl -> get_Offscreen(This,Offscreen) ) 

#define AccessibleState_get_Pressed(This,Pressed)	\
    ( (This)->lpVtbl -> get_Pressed(This,Pressed) ) 

#define AccessibleState_get_Protected(This,protectedEdit)	\
    ( (This)->lpVtbl -> get_Protected(This,protectedEdit) ) 

#define AccessibleState_get_ReadOnly(This,ReadOnly)	\
    ( (This)->lpVtbl -> get_ReadOnly(This,ReadOnly) ) 

#define AccessibleState_get_Selectable(This,Selectable)	\
    ( (This)->lpVtbl -> get_Selectable(This,Selectable) ) 

#define AccessibleState_get_Selected(This,Selected)	\
    ( (This)->lpVtbl -> get_Selected(This,Selected) ) 

#define AccessibleState_get_SelfVoicing(This,SelfVoicing)	\
    ( (This)->lpVtbl -> get_SelfVoicing(This,SelfVoicing) ) 

#define AccessibleState_get_Sizeable(This,Sizeable)	\
    ( (This)->lpVtbl -> get_Sizeable(This,Sizeable) ) 

#define AccessibleState_get_Traversed(This,Traversed)	\
    ( (This)->lpVtbl -> get_Traversed(This,Traversed) ) 

#define AccessibleState_get_Unavailable(This,Unavailable)	\
    ( (This)->lpVtbl -> get_Unavailable(This,Unavailable) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __AccessibleState_INTERFACE_DEFINED__ */


#ifndef __WindowPoint_INTERFACE_DEFINED__
#define __WindowPoint_INTERFACE_DEFINED__

/* interface WindowPoint */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WindowPoint;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("71A1DFD7-CD71-49F6-9E46-62BAB10DD8B8")
    WindowPoint : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_X( 
            /* [retval][out] */ long *X) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_X( 
            /* [in] */ long X) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Y( 
            /* [retval][out] */ long *Y) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Y( 
            /* [in] */ long Y) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Window( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ScreenPoint( 
            /* [retval][out] */ ScreenPoint **point) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ScreenPoint( 
            /* [in] */ ScreenPoint *point) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ClientPoint( 
            /* [retval][out] */ ClientPoint **point) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ClientPoint( 
            /* [in] */ ClientPoint *point) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WindowPointVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WindowPoint * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WindowPoint * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WindowPoint * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WindowPoint * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WindowPoint * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WindowPoint * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WindowPoint * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WindowPoint * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WindowPoint * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_X )( 
            WindowPoint * This,
            /* [retval][out] */ long *X);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_X )( 
            WindowPoint * This,
            /* [in] */ long X);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Y )( 
            WindowPoint * This,
            /* [retval][out] */ long *Y);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Y )( 
            WindowPoint * This,
            /* [in] */ long Y);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            WindowPoint * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ScreenPoint )( 
            WindowPoint * This,
            /* [retval][out] */ ScreenPoint **point);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ScreenPoint )( 
            WindowPoint * This,
            /* [in] */ ScreenPoint *point);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ClientPoint )( 
            WindowPoint * This,
            /* [retval][out] */ ClientPoint **point);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ClientPoint )( 
            WindowPoint * This,
            /* [in] */ ClientPoint *point);
        
        END_INTERFACE
    } WindowPointVtbl;

    interface WindowPoint
    {
        CONST_VTBL struct WindowPointVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WindowPoint_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WindowPoint_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WindowPoint_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WindowPoint_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WindowPoint_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WindowPoint_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WindowPoint_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WindowPoint_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WindowPoint_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WindowPoint_get_X(This,X)	\
    ( (This)->lpVtbl -> get_X(This,X) ) 

#define WindowPoint_put_X(This,X)	\
    ( (This)->lpVtbl -> put_X(This,X) ) 

#define WindowPoint_get_Y(This,Y)	\
    ( (This)->lpVtbl -> get_Y(This,Y) ) 

#define WindowPoint_put_Y(This,Y)	\
    ( (This)->lpVtbl -> put_Y(This,Y) ) 

#define WindowPoint_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define WindowPoint_get_ScreenPoint(This,point)	\
    ( (This)->lpVtbl -> get_ScreenPoint(This,point) ) 

#define WindowPoint_put_ScreenPoint(This,point)	\
    ( (This)->lpVtbl -> put_ScreenPoint(This,point) ) 

#define WindowPoint_get_ClientPoint(This,point)	\
    ( (This)->lpVtbl -> get_ClientPoint(This,point) ) 

#define WindowPoint_put_ClientPoint(This,point)	\
    ( (This)->lpVtbl -> put_ClientPoint(This,point) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WindowPoint_INTERFACE_DEFINED__ */


#ifndef __Windows_INTERFACE_DEFINED__
#define __Windows_INTERFACE_DEFINED__

/* interface Windows */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Windows;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("ABF05245-9E29-49FF-B6EB-6E8578099578")
    Windows : public IDispatch
    {
    public:
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Find( 
            /* [in] */ long Handle,
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE FilterByClassAndModule( 
            /* [in] */ BSTR ClassName,
            /* [in] */ BSTR ModuleName,
            /* [optional][in] */ VARIANT FilterMatch,
            /* [retval][out] */ Windows **result) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE FilterByName( 
            /* [in] */ BSTR Name,
            /* [optional][in] */ VARIANT FilterMatch,
            /* [optional][in] */ VARIANT MatchCase,
            /* [retval][out] */ Windows **result) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE FilterByType( 
            /* [in] */ WindowType type,
            /* [retval][out] */ Windows **result) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE FilterByLocation( 
            /* [in] */ VARIANT Location,
            /* [retval][out] */ Windows **result) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE FilterByTitle( 
            /* [in] */ BSTR Title,
            /* [optional][in] */ VARIANT FilterMatch,
            /* [optional][in] */ VARIANT MatchCase,
            /* [retval][out] */ Windows **result) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WindowsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Windows * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Windows * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Windows * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Windows * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Windows * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Windows * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Windows * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            Windows * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Windows * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Windows * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            Windows * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            Windows * This,
            /* [in] */ long index,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Find )( 
            Windows * This,
            /* [in] */ long Handle,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *FilterByClassAndModule )( 
            Windows * This,
            /* [in] */ BSTR ClassName,
            /* [in] */ BSTR ModuleName,
            /* [optional][in] */ VARIANT FilterMatch,
            /* [retval][out] */ Windows **result);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *FilterByName )( 
            Windows * This,
            /* [in] */ BSTR Name,
            /* [optional][in] */ VARIANT FilterMatch,
            /* [optional][in] */ VARIANT MatchCase,
            /* [retval][out] */ Windows **result);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *FilterByType )( 
            Windows * This,
            /* [in] */ WindowType type,
            /* [retval][out] */ Windows **result);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *FilterByLocation )( 
            Windows * This,
            /* [in] */ VARIANT Location,
            /* [retval][out] */ Windows **result);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *FilterByTitle )( 
            Windows * This,
            /* [in] */ BSTR Title,
            /* [optional][in] */ VARIANT FilterMatch,
            /* [optional][in] */ VARIANT MatchCase,
            /* [retval][out] */ Windows **result);
        
        END_INTERFACE
    } WindowsVtbl;

    interface Windows
    {
        CONST_VTBL struct WindowsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Windows_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Windows_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Windows_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Windows_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Windows_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Windows_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Windows_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Windows_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define Windows_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Windows_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Windows_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define Windows_Item(This,index,Window)	\
    ( (This)->lpVtbl -> Item(This,index,Window) ) 

#define Windows_Find(This,Handle,Window)	\
    ( (This)->lpVtbl -> Find(This,Handle,Window) ) 

#define Windows_FilterByClassAndModule(This,ClassName,ModuleName,FilterMatch,result)	\
    ( (This)->lpVtbl -> FilterByClassAndModule(This,ClassName,ModuleName,FilterMatch,result) ) 

#define Windows_FilterByName(This,Name,FilterMatch,MatchCase,result)	\
    ( (This)->lpVtbl -> FilterByName(This,Name,FilterMatch,MatchCase,result) ) 

#define Windows_FilterByType(This,type,result)	\
    ( (This)->lpVtbl -> FilterByType(This,type,result) ) 

#define Windows_FilterByLocation(This,Location,result)	\
    ( (This)->lpVtbl -> FilterByLocation(This,Location,result) ) 

#define Windows_FilterByTitle(This,Title,FilterMatch,MatchCase,result)	\
    ( (This)->lpVtbl -> FilterByTitle(This,Title,FilterMatch,MatchCase,result) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Windows_INTERFACE_DEFINED__ */


#ifndef __WindowStyle_INTERFACE_DEFINED__
#define __WindowStyle_INTERFACE_DEFINED__

/* interface WindowStyle */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WindowStyle;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("8887EE67-8612-4174-82ED-48B65388B6CE")
    WindowStyle : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Style( 
            /* [retval][out] */ long *Style) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Style( 
            /* [in] */ long Style) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Overlapped( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Popup( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_child( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Minimize( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Visible( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Disabled( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ClipSiblings( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ClipChildren( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Maximize( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Caption( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Border( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DialogFrame( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_VerticalScroll( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HorizontalScroll( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SystemMenu( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ThickFrame( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Group( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TabStop( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MinimizeBox( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MaximizeBox( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DialogModalFrame( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NoParentNotify( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TopMost( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AcceptFiles( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Transparent( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MDIChild( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ToolWindow( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WindowEdge( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ClientEdge( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ContextHelp( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Right( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Left( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_RTLReading( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LTRReading( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LeftScrollbar( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_RightScrollbar( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ControlParent( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_StaticEdge( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ApplicationWindow( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Layered( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NoInheritLayout( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LayoutRTL( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Composited( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NoActivate( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
        virtual /* [helpcontext][propget][id] */ HRESULT STDMETHODCALLTYPE get_NoPaddedBorder( 
            /* [retval][out] */ VARIANT_BOOL *set) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WindowStyleVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WindowStyle * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WindowStyle * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WindowStyle * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WindowStyle * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WindowStyle * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WindowStyle * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WindowStyle * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Style )( 
            WindowStyle * This,
            /* [retval][out] */ long *Style);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Style )( 
            WindowStyle * This,
            /* [in] */ long Style);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WindowStyle * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WindowStyle * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Overlapped )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Popup )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_child )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Minimize )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Disabled )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ClipSiblings )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ClipChildren )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Maximize )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Caption )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Border )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DialogFrame )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_VerticalScroll )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HorizontalScroll )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SystemMenu )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ThickFrame )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Group )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabStop )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MinimizeBox )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MaximizeBox )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DialogModalFrame )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NoParentNotify )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TopMost )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AcceptFiles )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Transparent )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MDIChild )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ToolWindow )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WindowEdge )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ClientEdge )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ContextHelp )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Right )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Left )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_RTLReading )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LTRReading )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LeftScrollbar )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_RightScrollbar )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ControlParent )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_StaticEdge )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ApplicationWindow )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Layered )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NoInheritLayout )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LayoutRTL )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Composited )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NoActivate )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        /* [helpcontext][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NoPaddedBorder )( 
            WindowStyle * This,
            /* [retval][out] */ VARIANT_BOOL *set);
        
        END_INTERFACE
    } WindowStyleVtbl;

    interface WindowStyle
    {
        CONST_VTBL struct WindowStyleVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WindowStyle_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WindowStyle_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WindowStyle_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WindowStyle_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WindowStyle_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WindowStyle_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WindowStyle_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WindowStyle_get_Style(This,Style)	\
    ( (This)->lpVtbl -> get_Style(This,Style) ) 

#define WindowStyle_put_Style(This,Style)	\
    ( (This)->lpVtbl -> put_Style(This,Style) ) 

#define WindowStyle_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WindowStyle_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WindowStyle_get_Overlapped(This,set)	\
    ( (This)->lpVtbl -> get_Overlapped(This,set) ) 

#define WindowStyle_get_Popup(This,set)	\
    ( (This)->lpVtbl -> get_Popup(This,set) ) 

#define WindowStyle_get_child(This,set)	\
    ( (This)->lpVtbl -> get_child(This,set) ) 

#define WindowStyle_get_Minimize(This,set)	\
    ( (This)->lpVtbl -> get_Minimize(This,set) ) 

#define WindowStyle_get_Visible(This,set)	\
    ( (This)->lpVtbl -> get_Visible(This,set) ) 

#define WindowStyle_get_Disabled(This,set)	\
    ( (This)->lpVtbl -> get_Disabled(This,set) ) 

#define WindowStyle_get_ClipSiblings(This,set)	\
    ( (This)->lpVtbl -> get_ClipSiblings(This,set) ) 

#define WindowStyle_get_ClipChildren(This,set)	\
    ( (This)->lpVtbl -> get_ClipChildren(This,set) ) 

#define WindowStyle_get_Maximize(This,set)	\
    ( (This)->lpVtbl -> get_Maximize(This,set) ) 

#define WindowStyle_get_Caption(This,set)	\
    ( (This)->lpVtbl -> get_Caption(This,set) ) 

#define WindowStyle_get_Border(This,set)	\
    ( (This)->lpVtbl -> get_Border(This,set) ) 

#define WindowStyle_get_DialogFrame(This,set)	\
    ( (This)->lpVtbl -> get_DialogFrame(This,set) ) 

#define WindowStyle_get_VerticalScroll(This,set)	\
    ( (This)->lpVtbl -> get_VerticalScroll(This,set) ) 

#define WindowStyle_get_HorizontalScroll(This,set)	\
    ( (This)->lpVtbl -> get_HorizontalScroll(This,set) ) 

#define WindowStyle_get_SystemMenu(This,set)	\
    ( (This)->lpVtbl -> get_SystemMenu(This,set) ) 

#define WindowStyle_get_ThickFrame(This,set)	\
    ( (This)->lpVtbl -> get_ThickFrame(This,set) ) 

#define WindowStyle_get_Group(This,set)	\
    ( (This)->lpVtbl -> get_Group(This,set) ) 

#define WindowStyle_get_TabStop(This,set)	\
    ( (This)->lpVtbl -> get_TabStop(This,set) ) 

#define WindowStyle_get_MinimizeBox(This,set)	\
    ( (This)->lpVtbl -> get_MinimizeBox(This,set) ) 

#define WindowStyle_get_MaximizeBox(This,set)	\
    ( (This)->lpVtbl -> get_MaximizeBox(This,set) ) 

#define WindowStyle_get_DialogModalFrame(This,set)	\
    ( (This)->lpVtbl -> get_DialogModalFrame(This,set) ) 

#define WindowStyle_get_NoParentNotify(This,set)	\
    ( (This)->lpVtbl -> get_NoParentNotify(This,set) ) 

#define WindowStyle_get_TopMost(This,set)	\
    ( (This)->lpVtbl -> get_TopMost(This,set) ) 

#define WindowStyle_get_AcceptFiles(This,set)	\
    ( (This)->lpVtbl -> get_AcceptFiles(This,set) ) 

#define WindowStyle_get_Transparent(This,set)	\
    ( (This)->lpVtbl -> get_Transparent(This,set) ) 

#define WindowStyle_get_MDIChild(This,set)	\
    ( (This)->lpVtbl -> get_MDIChild(This,set) ) 

#define WindowStyle_get_ToolWindow(This,set)	\
    ( (This)->lpVtbl -> get_ToolWindow(This,set) ) 

#define WindowStyle_get_WindowEdge(This,set)	\
    ( (This)->lpVtbl -> get_WindowEdge(This,set) ) 

#define WindowStyle_get_ClientEdge(This,set)	\
    ( (This)->lpVtbl -> get_ClientEdge(This,set) ) 

#define WindowStyle_get_ContextHelp(This,set)	\
    ( (This)->lpVtbl -> get_ContextHelp(This,set) ) 

#define WindowStyle_get_Right(This,set)	\
    ( (This)->lpVtbl -> get_Right(This,set) ) 

#define WindowStyle_get_Left(This,set)	\
    ( (This)->lpVtbl -> get_Left(This,set) ) 

#define WindowStyle_get_RTLReading(This,set)	\
    ( (This)->lpVtbl -> get_RTLReading(This,set) ) 

#define WindowStyle_get_LTRReading(This,set)	\
    ( (This)->lpVtbl -> get_LTRReading(This,set) ) 

#define WindowStyle_get_LeftScrollbar(This,set)	\
    ( (This)->lpVtbl -> get_LeftScrollbar(This,set) ) 

#define WindowStyle_get_RightScrollbar(This,set)	\
    ( (This)->lpVtbl -> get_RightScrollbar(This,set) ) 

#define WindowStyle_get_ControlParent(This,set)	\
    ( (This)->lpVtbl -> get_ControlParent(This,set) ) 

#define WindowStyle_get_StaticEdge(This,set)	\
    ( (This)->lpVtbl -> get_StaticEdge(This,set) ) 

#define WindowStyle_get_ApplicationWindow(This,set)	\
    ( (This)->lpVtbl -> get_ApplicationWindow(This,set) ) 

#define WindowStyle_get_Layered(This,set)	\
    ( (This)->lpVtbl -> get_Layered(This,set) ) 

#define WindowStyle_get_NoInheritLayout(This,set)	\
    ( (This)->lpVtbl -> get_NoInheritLayout(This,set) ) 

#define WindowStyle_get_LayoutRTL(This,set)	\
    ( (This)->lpVtbl -> get_LayoutRTL(This,set) ) 

#define WindowStyle_get_Composited(This,set)	\
    ( (This)->lpVtbl -> get_Composited(This,set) ) 

#define WindowStyle_get_NoActivate(This,set)	\
    ( (This)->lpVtbl -> get_NoActivate(This,set) ) 

#define WindowStyle_get_NoPaddedBorder(This,set)	\
    ( (This)->lpVtbl -> get_NoPaddedBorder(This,set) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WindowStyle_INTERFACE_DEFINED__ */


#ifndef __Desktop_INTERFACE_DEFINED__
#define __Desktop_INTERFACE_DEFINED__

/* interface Desktop */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Desktop;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("991A7510-8885-4A88-B086-D71834E1A032")
    Desktop : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Name( 
            /* [retval][out] */ BSTR *desktopName) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct DesktopVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Desktop * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Desktop * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Desktop * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Desktop * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Desktop * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Desktop * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Desktop * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Desktop * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Desktop * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Name )( 
            Desktop * This,
            /* [retval][out] */ BSTR *desktopName);
        
        END_INTERFACE
    } DesktopVtbl;

    interface Desktop
    {
        CONST_VTBL struct DesktopVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Desktop_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Desktop_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Desktop_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Desktop_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Desktop_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Desktop_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Desktop_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Desktop_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Desktop_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Desktop_get_Name(This,desktopName)	\
    ( (This)->lpVtbl -> get_Name(This,desktopName) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Desktop_INTERFACE_DEFINED__ */


#ifndef __Process_INTERFACE_DEFINED__
#define __Process_INTERFACE_DEFINED__

/* interface Process */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Process;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("713468F4-0BB7-4DE2-B7AE-373C6AE2949F")
    Process : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ProcessID( 
            /* [retval][out] */ long *id) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ModuleName( 
            /* [retval][out] */ BSTR *ModuleName) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ExecutablePath( 
            /* [retval][out] */ BSTR *path) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Comments( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CompanyName( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FileDescription( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FileVersion( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_InternalName( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LegalCopyright( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LegalTrademarks( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_OriginalFilename( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ProductName( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ProductVersion( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_PrivateBuild( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpecialBuild( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Is64Bit( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Version( 
            /* [retval][out] */ Version **Version) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ProcessVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Process * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Process * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Process * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Process * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Process * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Process * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Process * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Process * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Process * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ProcessID )( 
            Process * This,
            /* [retval][out] */ long *id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ModuleName )( 
            Process * This,
            /* [retval][out] */ BSTR *ModuleName);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ExecutablePath )( 
            Process * This,
            /* [retval][out] */ BSTR *path);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Comments )( 
            Process * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CompanyName )( 
            Process * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FileDescription )( 
            Process * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FileVersion )( 
            Process * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_InternalName )( 
            Process * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LegalCopyright )( 
            Process * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LegalTrademarks )( 
            Process * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_OriginalFilename )( 
            Process * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ProductName )( 
            Process * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ProductVersion )( 
            Process * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_PrivateBuild )( 
            Process * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpecialBuild )( 
            Process * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Is64Bit )( 
            Process * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Version )( 
            Process * This,
            /* [retval][out] */ Version **Version);
        
        END_INTERFACE
    } ProcessVtbl;

    interface Process
    {
        CONST_VTBL struct ProcessVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Process_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Process_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Process_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Process_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Process_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Process_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Process_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Process_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Process_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Process_get_ProcessID(This,id)	\
    ( (This)->lpVtbl -> get_ProcessID(This,id) ) 

#define Process_get_ModuleName(This,ModuleName)	\
    ( (This)->lpVtbl -> get_ModuleName(This,ModuleName) ) 

#define Process_get_ExecutablePath(This,path)	\
    ( (This)->lpVtbl -> get_ExecutablePath(This,path) ) 

#define Process_get_Comments(This,prop)	\
    ( (This)->lpVtbl -> get_Comments(This,prop) ) 

#define Process_get_CompanyName(This,prop)	\
    ( (This)->lpVtbl -> get_CompanyName(This,prop) ) 

#define Process_get_FileDescription(This,prop)	\
    ( (This)->lpVtbl -> get_FileDescription(This,prop) ) 

#define Process_get_FileVersion(This,prop)	\
    ( (This)->lpVtbl -> get_FileVersion(This,prop) ) 

#define Process_get_InternalName(This,prop)	\
    ( (This)->lpVtbl -> get_InternalName(This,prop) ) 

#define Process_get_LegalCopyright(This,prop)	\
    ( (This)->lpVtbl -> get_LegalCopyright(This,prop) ) 

#define Process_get_LegalTrademarks(This,prop)	\
    ( (This)->lpVtbl -> get_LegalTrademarks(This,prop) ) 

#define Process_get_OriginalFilename(This,prop)	\
    ( (This)->lpVtbl -> get_OriginalFilename(This,prop) ) 

#define Process_get_ProductName(This,prop)	\
    ( (This)->lpVtbl -> get_ProductName(This,prop) ) 

#define Process_get_ProductVersion(This,prop)	\
    ( (This)->lpVtbl -> get_ProductVersion(This,prop) ) 

#define Process_get_PrivateBuild(This,prop)	\
    ( (This)->lpVtbl -> get_PrivateBuild(This,prop) ) 

#define Process_get_SpecialBuild(This,prop)	\
    ( (This)->lpVtbl -> get_SpecialBuild(This,prop) ) 

#define Process_get_Is64Bit(This,prop)	\
    ( (This)->lpVtbl -> get_Is64Bit(This,prop) ) 

#define Process_get_Version(This,Version)	\
    ( (This)->lpVtbl -> get_Version(This,Version) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Process_INTERFACE_DEFINED__ */


#ifndef __Version_INTERFACE_DEFINED__
#define __Version_INTERFACE_DEFINED__

/* interface Version */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Version;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("B457E747-8F0A-44D6-B812-48A3E7DA4F81")
    Version : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Major( 
            /* [retval][out] */ long *Major) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Minor( 
            /* [retval][out] */ long *Minor) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Build( 
            /* [retval][out] */ long *Build) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Revision( 
            /* [retval][out] */ long *Revision) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_String( 
            /* [retval][out] */ BSTR *ProductVersion) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct VersionVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Version * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Version * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Version * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Version * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Version * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Version * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Version * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Version * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Version * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Major )( 
            Version * This,
            /* [retval][out] */ long *Major);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Minor )( 
            Version * This,
            /* [retval][out] */ long *Minor);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Build )( 
            Version * This,
            /* [retval][out] */ long *Build);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Revision )( 
            Version * This,
            /* [retval][out] */ long *Revision);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_String )( 
            Version * This,
            /* [retval][out] */ BSTR *ProductVersion);
        
        END_INTERFACE
    } VersionVtbl;

    interface Version
    {
        CONST_VTBL struct VersionVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Version_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Version_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Version_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Version_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Version_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Version_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Version_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Version_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Version_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Version_get_Major(This,Major)	\
    ( (This)->lpVtbl -> get_Major(This,Major) ) 

#define Version_get_Minor(This,Minor)	\
    ( (This)->lpVtbl -> get_Minor(This,Minor) ) 

#define Version_get_Build(This,Build)	\
    ( (This)->lpVtbl -> get_Build(This,Build) ) 

#define Version_get_Revision(This,Revision)	\
    ( (This)->lpVtbl -> get_Revision(This,Revision) ) 

#define Version_get_String(This,ProductVersion)	\
    ( (This)->lpVtbl -> get_String(This,ProductVersion) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Version_INTERFACE_DEFINED__ */


#ifndef __Clips_INTERFACE_DEFINED__
#define __Clips_INTERFACE_DEFINED__

/* interface Clips */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Clips;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("539C9C74-8FF7-4718-AC9D-8D3F1B8CAF66")
    Clips : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ClipsText( 
            /* [retval][out] */ BSTR *Text) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ Clip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ClipAtPoint( 
            /* [in] */ WindowPoint *point,
            /* [retval][out] */ Clip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE FilterByType( 
            /* [in] */ ClipType ClipType,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE FilterByRectangle( 
            /* [in] */ WindowRectangle *rect,
            /* [optional][in] */ VARIANT clipClips,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SearchText( 
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT MatchCase,
            /* [optional][in] */ VARIANT Highlighted,
            /* [optional][in] */ VARIANT FontStyle,
            /* [optional][in] */ VARIANT FontName,
            /* [optional][in] */ VARIANT FontSize,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Find( 
            /* [in] */ _Window *Window,
            /* [in] */ long clipid,
            /* [retval][out] */ Clip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE FilterByColor( 
            /* [in] */ Color *ForegroundColor,
            /* [in] */ Color *BackgroundColor,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE CharacterIndexAtPoint( 
            /* [in] */ ScreenPoint *point,
            /* [retval][out] */ long *index) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ClipsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Clips * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Clips * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Clips * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Clips * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Clips * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Clips * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Clips * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Clips * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Clips * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            Clips * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ClipsText )( 
            Clips * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            Clips * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            Clips * This,
            /* [in] */ long index,
            /* [retval][out] */ Clip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ClipAtPoint )( 
            Clips * This,
            /* [in] */ WindowPoint *point,
            /* [retval][out] */ Clip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *FilterByType )( 
            Clips * This,
            /* [in] */ ClipType ClipType,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *FilterByRectangle )( 
            Clips * This,
            /* [in] */ WindowRectangle *rect,
            /* [optional][in] */ VARIANT clipClips,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SearchText )( 
            Clips * This,
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT MatchCase,
            /* [optional][in] */ VARIANT Highlighted,
            /* [optional][in] */ VARIANT FontStyle,
            /* [optional][in] */ VARIANT FontName,
            /* [optional][in] */ VARIANT FontSize,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Find )( 
            Clips * This,
            /* [in] */ _Window *Window,
            /* [in] */ long clipid,
            /* [retval][out] */ Clip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *FilterByColor )( 
            Clips * This,
            /* [in] */ Color *ForegroundColor,
            /* [in] */ Color *BackgroundColor,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *CharacterIndexAtPoint )( 
            Clips * This,
            /* [in] */ ScreenPoint *point,
            /* [retval][out] */ long *index);
        
        END_INTERFACE
    } ClipsVtbl;

    interface Clips
    {
        CONST_VTBL struct ClipsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Clips_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Clips_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Clips_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Clips_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Clips_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Clips_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Clips_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Clips_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Clips_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Clips_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define Clips_get_ClipsText(This,Text)	\
    ( (This)->lpVtbl -> get_ClipsText(This,Text) ) 

#define Clips_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define Clips_Item(This,index,Clip)	\
    ( (This)->lpVtbl -> Item(This,index,Clip) ) 

#define Clips_ClipAtPoint(This,point,Clip)	\
    ( (This)->lpVtbl -> ClipAtPoint(This,point,Clip) ) 

#define Clips_FilterByType(This,ClipType,Clips)	\
    ( (This)->lpVtbl -> FilterByType(This,ClipType,Clips) ) 

#define Clips_FilterByRectangle(This,rect,clipClips,Clips)	\
    ( (This)->lpVtbl -> FilterByRectangle(This,rect,clipClips,Clips) ) 

#define Clips_SearchText(This,Text,MatchCase,Highlighted,FontStyle,FontName,FontSize,Clips)	\
    ( (This)->lpVtbl -> SearchText(This,Text,MatchCase,Highlighted,FontStyle,FontName,FontSize,Clips) ) 

#define Clips_Find(This,Window,clipid,Clip)	\
    ( (This)->lpVtbl -> Find(This,Window,clipid,Clip) ) 

#define Clips_FilterByColor(This,ForegroundColor,BackgroundColor,Clips)	\
    ( (This)->lpVtbl -> FilterByColor(This,ForegroundColor,BackgroundColor,Clips) ) 

#define Clips_CharacterIndexAtPoint(This,point,index)	\
    ( (This)->lpVtbl -> CharacterIndexAtPoint(This,point,index) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Clips_INTERFACE_DEFINED__ */


#ifndef __Clip_INTERFACE_DEFINED__
#define __Clip_INTERFACE_DEFINED__

/* interface Clip */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Clip;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("E39DC37D-7C07-4859-94FD-A661B8C1F2AC")
    Clip : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Window( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_id( 
            /* [retval][out] */ long *clipid) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_type( 
            /* [retval][out] */ ClipType *ClipType) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Position( 
            /* [retval][out] */ WindowPoint **Position) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_rectangle( 
            /* [retval][out] */ WindowRectangle **rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Text( 
            /* [retval][out] */ BSTR *Text) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Length( 
            /* [retval][out] */ long *Length) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ForegroundColor( 
            /* [retval][out] */ Color **Color) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ForegroundColor( 
            /* [in] */ Color *Color) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BackgroundColor( 
            /* [retval][out] */ Color **Color) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_BackgroundColor( 
            /* [in] */ Color *Color) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FontName( 
            /* [retval][out] */ BSTR *Name) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FontName( 
            /* [in] */ BSTR Name) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FontSize( 
            /* [retval][out] */ long *size) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FontSize( 
            /* [in] */ long size) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FontStyle( 
            /* [retval][out] */ FontStyle *Style) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FontStyle( 
            /* [in] */ FontStyle Style) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CharacterOffsets( 
            /* [retval][out] */ CharacterOffsets **offsets) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CRC( 
            /* [retval][out] */ long *CRC) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Delete( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Line( 
            /* [optional][in] */ VARIANT RealTime,
            /* [retval][out] */ Clips **Line) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE CharacterIndexAtPoint( 
            /* [in] */ ScreenPoint *point,
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsStatic( 
            /* [retval][out] */ VARIANT_BOOL *IsStatic) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsValid( 
            /* [retval][out] */ VARIANT_BOOL *IsValid) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ClipVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Clip * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Clip * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Clip * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Clip * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Clip * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Clip * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Clip * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Clip * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Clip * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            Clip * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            Clip * This,
            /* [retval][out] */ long *clipid);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_type )( 
            Clip * This,
            /* [retval][out] */ ClipType *ClipType);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            Clip * This,
            /* [retval][out] */ WindowPoint **Position);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_rectangle )( 
            Clip * This,
            /* [retval][out] */ WindowRectangle **rectangle);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            Clip * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Length )( 
            Clip * This,
            /* [retval][out] */ long *Length);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ForegroundColor )( 
            Clip * This,
            /* [retval][out] */ Color **Color);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ForegroundColor )( 
            Clip * This,
            /* [in] */ Color *Color);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BackgroundColor )( 
            Clip * This,
            /* [retval][out] */ Color **Color);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_BackgroundColor )( 
            Clip * This,
            /* [in] */ Color *Color);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FontName )( 
            Clip * This,
            /* [retval][out] */ BSTR *Name);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FontName )( 
            Clip * This,
            /* [in] */ BSTR Name);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FontSize )( 
            Clip * This,
            /* [retval][out] */ long *size);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FontSize )( 
            Clip * This,
            /* [in] */ long size);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FontStyle )( 
            Clip * This,
            /* [retval][out] */ FontStyle *Style);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FontStyle )( 
            Clip * This,
            /* [in] */ FontStyle Style);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CharacterOffsets )( 
            Clip * This,
            /* [retval][out] */ CharacterOffsets **offsets);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CRC )( 
            Clip * This,
            /* [retval][out] */ long *CRC);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Delete )( 
            Clip * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Line )( 
            Clip * This,
            /* [optional][in] */ VARIANT RealTime,
            /* [retval][out] */ Clips **Line);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *CharacterIndexAtPoint )( 
            Clip * This,
            /* [in] */ ScreenPoint *point,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsStatic )( 
            Clip * This,
            /* [retval][out] */ VARIANT_BOOL *IsStatic);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsValid )( 
            Clip * This,
            /* [retval][out] */ VARIANT_BOOL *IsValid);
        
        END_INTERFACE
    } ClipVtbl;

    interface Clip
    {
        CONST_VTBL struct ClipVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Clip_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Clip_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Clip_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Clip_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Clip_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Clip_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Clip_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Clip_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Clip_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Clip_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define Clip_get_id(This,clipid)	\
    ( (This)->lpVtbl -> get_id(This,clipid) ) 

#define Clip_get_type(This,ClipType)	\
    ( (This)->lpVtbl -> get_type(This,ClipType) ) 

#define Clip_get_Position(This,Position)	\
    ( (This)->lpVtbl -> get_Position(This,Position) ) 

#define Clip_get_rectangle(This,rectangle)	\
    ( (This)->lpVtbl -> get_rectangle(This,rectangle) ) 

#define Clip_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define Clip_get_Length(This,Length)	\
    ( (This)->lpVtbl -> get_Length(This,Length) ) 

#define Clip_get_ForegroundColor(This,Color)	\
    ( (This)->lpVtbl -> get_ForegroundColor(This,Color) ) 

#define Clip_put_ForegroundColor(This,Color)	\
    ( (This)->lpVtbl -> put_ForegroundColor(This,Color) ) 

#define Clip_get_BackgroundColor(This,Color)	\
    ( (This)->lpVtbl -> get_BackgroundColor(This,Color) ) 

#define Clip_put_BackgroundColor(This,Color)	\
    ( (This)->lpVtbl -> put_BackgroundColor(This,Color) ) 

#define Clip_get_FontName(This,Name)	\
    ( (This)->lpVtbl -> get_FontName(This,Name) ) 

#define Clip_put_FontName(This,Name)	\
    ( (This)->lpVtbl -> put_FontName(This,Name) ) 

#define Clip_get_FontSize(This,size)	\
    ( (This)->lpVtbl -> get_FontSize(This,size) ) 

#define Clip_put_FontSize(This,size)	\
    ( (This)->lpVtbl -> put_FontSize(This,size) ) 

#define Clip_get_FontStyle(This,Style)	\
    ( (This)->lpVtbl -> get_FontStyle(This,Style) ) 

#define Clip_put_FontStyle(This,Style)	\
    ( (This)->lpVtbl -> put_FontStyle(This,Style) ) 

#define Clip_get_CharacterOffsets(This,offsets)	\
    ( (This)->lpVtbl -> get_CharacterOffsets(This,offsets) ) 

#define Clip_get_CRC(This,CRC)	\
    ( (This)->lpVtbl -> get_CRC(This,CRC) ) 

#define Clip_Delete(This)	\
    ( (This)->lpVtbl -> Delete(This) ) 

#define Clip_Line(This,RealTime,Line)	\
    ( (This)->lpVtbl -> Line(This,RealTime,Line) ) 

#define Clip_CharacterIndexAtPoint(This,point,index)	\
    ( (This)->lpVtbl -> CharacterIndexAtPoint(This,point,index) ) 

#define Clip_get_IsStatic(This,IsStatic)	\
    ( (This)->lpVtbl -> get_IsStatic(This,IsStatic) ) 

#define Clip_get_IsValid(This,IsValid)	\
    ( (This)->lpVtbl -> get_IsValid(This,IsValid) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Clip_INTERFACE_DEFINED__ */


#ifndef __Color_INTERFACE_DEFINED__
#define __Color_INTERFACE_DEFINED__

/* interface Color */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Color;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("E8B9CFBA-3E39-4EA7-BF3E-87F52BD8F2A3")
    Color : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Red( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Red( 
            /* [in] */ long prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Green( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Green( 
            /* [in] */ long prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Blue( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Blue( 
            /* [in] */ long prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ColorVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Color * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Color * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Color * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Color * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Color * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Color * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Color * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Color * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Color * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Red )( 
            Color * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Red )( 
            Color * This,
            /* [in] */ long prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Green )( 
            Color * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Green )( 
            Color * This,
            /* [in] */ long prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Blue )( 
            Color * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Blue )( 
            Color * This,
            /* [in] */ long prop);
        
        END_INTERFACE
    } ColorVtbl;

    interface Color
    {
        CONST_VTBL struct ColorVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Color_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Color_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Color_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Color_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Color_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Color_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Color_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Color_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Color_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Color_get_Red(This,prop)	\
    ( (This)->lpVtbl -> get_Red(This,prop) ) 

#define Color_put_Red(This,prop)	\
    ( (This)->lpVtbl -> put_Red(This,prop) ) 

#define Color_get_Green(This,prop)	\
    ( (This)->lpVtbl -> get_Green(This,prop) ) 

#define Color_put_Green(This,prop)	\
    ( (This)->lpVtbl -> put_Green(This,prop) ) 

#define Color_get_Blue(This,prop)	\
    ( (This)->lpVtbl -> get_Blue(This,prop) ) 

#define Color_put_Blue(This,prop)	\
    ( (This)->lpVtbl -> put_Blue(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Color_INTERFACE_DEFINED__ */


#ifndef __CharacterOffsets_INTERFACE_DEFINED__
#define __CharacterOffsets_INTERFACE_DEFINED__

/* interface CharacterOffsets */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_CharacterOffsets;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("3AD09936-1812-4F02-A430-A257E9A68BFC")
    CharacterOffsets : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ long *offset) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct CharacterOffsetsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            CharacterOffsets * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            CharacterOffsets * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            CharacterOffsets * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            CharacterOffsets * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            CharacterOffsets * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            CharacterOffsets * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            CharacterOffsets * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            CharacterOffsets * This,
            /* [in] */ long index,
            /* [retval][out] */ long *offset);
        
        END_INTERFACE
    } CharacterOffsetsVtbl;

    interface CharacterOffsets
    {
        CONST_VTBL struct CharacterOffsetsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define CharacterOffsets_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define CharacterOffsets_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define CharacterOffsets_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define CharacterOffsets_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define CharacterOffsets_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define CharacterOffsets_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define CharacterOffsets_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define CharacterOffsets_Item(This,index,offset)	\
    ( (This)->lpVtbl -> Item(This,index,offset) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __CharacterOffsets_INTERFACE_DEFINED__ */


#ifndef __SetFile_INTERFACE_DEFINED__
#define __SetFile_INTERFACE_DEFINED__

/* interface SetFile */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_SetFile;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("66BE9EB6-315D-4B0E-81A7-EC3B04CC0A16")
    SetFile : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Screen( 
            /* [retval][out] */ ScreenSettings **Screen) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Keyboard( 
            /* [retval][out] */ KeyboardSettings **Keyboard) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Mouse( 
            /* [retval][out] */ MouseSettings **Mouse) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_General( 
            /* [retval][out] */ GeneralSettings **General) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Verbosity( 
            /* [retval][out] */ Verbosity **Verbosity) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Braille( 
            /* [retval][out] */ BrailleSettings **Braille) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UserWindows( 
            /* [retval][out] */ UserWindows **UserWindows) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HyperactiveWindows( 
            /* [retval][out] */ HyperactiveWindows **HyperactiveWindows) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HotKeys( 
            /* [retval][out] */ HotKeys **HotKeys) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CursorKeys( 
            /* [retval][out] */ CursorKeys **CursorKeys) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CharacterDictionary( 
            /* [retval][out] */ CharacterDictionary **dict) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ColorDictionary( 
            /* [retval][out] */ ColorDictionary **dict) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ExceptionDictionary( 
            /* [retval][out] */ ExceptionDictionary **dict) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_GraphicDictionary( 
            /* [retval][out] */ GraphicDictionary **dict) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_KeyLabelDictionary( 
            /* [retval][out] */ KeyLabelDictionary **dict) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Global( 
            /* [retval][out] */ GlobalSettings **settings) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Filename( 
            /* [retval][out] */ BSTR *Name) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Save( void) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Scope( 
            /* [retval][out] */ SetFileScope **SetFileScope) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct SetFileVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            SetFile * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            SetFile * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            SetFile * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            SetFile * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            SetFile * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            SetFile * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            SetFile * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            SetFile * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            SetFile * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Screen )( 
            SetFile * This,
            /* [retval][out] */ ScreenSettings **Screen);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Keyboard )( 
            SetFile * This,
            /* [retval][out] */ KeyboardSettings **Keyboard);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Mouse )( 
            SetFile * This,
            /* [retval][out] */ MouseSettings **Mouse);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_General )( 
            SetFile * This,
            /* [retval][out] */ GeneralSettings **General);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Verbosity )( 
            SetFile * This,
            /* [retval][out] */ Verbosity **Verbosity);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Braille )( 
            SetFile * This,
            /* [retval][out] */ BrailleSettings **Braille);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UserWindows )( 
            SetFile * This,
            /* [retval][out] */ UserWindows **UserWindows);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HyperactiveWindows )( 
            SetFile * This,
            /* [retval][out] */ HyperactiveWindows **HyperactiveWindows);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HotKeys )( 
            SetFile * This,
            /* [retval][out] */ HotKeys **HotKeys);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CursorKeys )( 
            SetFile * This,
            /* [retval][out] */ CursorKeys **CursorKeys);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CharacterDictionary )( 
            SetFile * This,
            /* [retval][out] */ CharacterDictionary **dict);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ColorDictionary )( 
            SetFile * This,
            /* [retval][out] */ ColorDictionary **dict);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ExceptionDictionary )( 
            SetFile * This,
            /* [retval][out] */ ExceptionDictionary **dict);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_GraphicDictionary )( 
            SetFile * This,
            /* [retval][out] */ GraphicDictionary **dict);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_KeyLabelDictionary )( 
            SetFile * This,
            /* [retval][out] */ KeyLabelDictionary **dict);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Global )( 
            SetFile * This,
            /* [retval][out] */ GlobalSettings **settings);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Filename )( 
            SetFile * This,
            /* [retval][out] */ BSTR *Name);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Save )( 
            SetFile * This);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Scope )( 
            SetFile * This,
            /* [retval][out] */ SetFileScope **SetFileScope);
        
        END_INTERFACE
    } SetFileVtbl;

    interface SetFile
    {
        CONST_VTBL struct SetFileVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define SetFile_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define SetFile_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define SetFile_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define SetFile_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define SetFile_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define SetFile_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define SetFile_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define SetFile_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define SetFile_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define SetFile_get_Screen(This,Screen)	\
    ( (This)->lpVtbl -> get_Screen(This,Screen) ) 

#define SetFile_get_Keyboard(This,Keyboard)	\
    ( (This)->lpVtbl -> get_Keyboard(This,Keyboard) ) 

#define SetFile_get_Mouse(This,Mouse)	\
    ( (This)->lpVtbl -> get_Mouse(This,Mouse) ) 

#define SetFile_get_General(This,General)	\
    ( (This)->lpVtbl -> get_General(This,General) ) 

#define SetFile_get_Verbosity(This,Verbosity)	\
    ( (This)->lpVtbl -> get_Verbosity(This,Verbosity) ) 

#define SetFile_get_Braille(This,Braille)	\
    ( (This)->lpVtbl -> get_Braille(This,Braille) ) 

#define SetFile_get_UserWindows(This,UserWindows)	\
    ( (This)->lpVtbl -> get_UserWindows(This,UserWindows) ) 

#define SetFile_get_HyperactiveWindows(This,HyperactiveWindows)	\
    ( (This)->lpVtbl -> get_HyperactiveWindows(This,HyperactiveWindows) ) 

#define SetFile_get_HotKeys(This,HotKeys)	\
    ( (This)->lpVtbl -> get_HotKeys(This,HotKeys) ) 

#define SetFile_get_CursorKeys(This,CursorKeys)	\
    ( (This)->lpVtbl -> get_CursorKeys(This,CursorKeys) ) 

#define SetFile_get_CharacterDictionary(This,dict)	\
    ( (This)->lpVtbl -> get_CharacterDictionary(This,dict) ) 

#define SetFile_get_ColorDictionary(This,dict)	\
    ( (This)->lpVtbl -> get_ColorDictionary(This,dict) ) 

#define SetFile_get_ExceptionDictionary(This,dict)	\
    ( (This)->lpVtbl -> get_ExceptionDictionary(This,dict) ) 

#define SetFile_get_GraphicDictionary(This,dict)	\
    ( (This)->lpVtbl -> get_GraphicDictionary(This,dict) ) 

#define SetFile_get_KeyLabelDictionary(This,dict)	\
    ( (This)->lpVtbl -> get_KeyLabelDictionary(This,dict) ) 

#define SetFile_get_Global(This,settings)	\
    ( (This)->lpVtbl -> get_Global(This,settings) ) 

#define SetFile_get_Filename(This,Name)	\
    ( (This)->lpVtbl -> get_Filename(This,Name) ) 

#define SetFile_Save(This)	\
    ( (This)->lpVtbl -> Save(This) ) 

#define SetFile_get_Scope(This,SetFileScope)	\
    ( (This)->lpVtbl -> get_Scope(This,SetFileScope) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __SetFile_INTERFACE_DEFINED__ */


#ifndef __ScreenSettings_INTERFACE_DEFINED__
#define __ScreenSettings_INTERFACE_DEFINED__

/* interface ScreenSettings */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ScreenSettings;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("E52BE567-CCE5-4C69-AB71-2CAA455459A2")
    ScreenSettings : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Voice( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Voice( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Rate( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Rate( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Pitch( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Pitch( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Tone( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Tone( 
            /* [in] */ VARIANT prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Volume( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Volume( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Punctuation( 
            /* [retval][out] */ PunctuationSettings **punc) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Numbers( 
            /* [retval][out] */ NumberMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Numbers( 
            /* [in] */ NumberMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CapitalizationAlert( 
            /* [retval][out] */ CapitalizationMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CapitalizationAlert( 
            /* [in] */ CapitalizationMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FormatAlert( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FormatAlert( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpellAlphaNumeric( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpellAlphaNumeric( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SplitMixedCaseWords( 
            /* [retval][out] */ MixedCaseMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SplitMixedCaseWords( 
            /* [in] */ MixedCaseMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_RepeatFilter( 
            /* [retval][out] */ RepeatMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_RepeatFilter( 
            /* [in] */ RepeatMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_RepeatValue( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_RepeatValue( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BlankLines( 
            /* [retval][out] */ BlankLinesMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_BlankLines( 
            /* [in] */ BlankLinesMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SignalCursorPosition( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SignalCursorPosition( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AttributeChanges( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AttributeChanges( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadParagraphStyle( 
            /* [retval][out] */ AutoMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadParagraphStyle( 
            /* [in] */ AutoMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LineFilter( 
            /* [retval][out] */ LineFilterSettings **LineFilter) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ScreenSettingsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ScreenSettings * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ScreenSettings * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ScreenSettings * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ScreenSettings * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ScreenSettings * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ScreenSettings * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ScreenSettings * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ScreenSettings * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ScreenSettings * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Voice )( 
            ScreenSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Voice )( 
            ScreenSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Rate )( 
            ScreenSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Rate )( 
            ScreenSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Pitch )( 
            ScreenSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Pitch )( 
            ScreenSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Tone )( 
            ScreenSettings * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Tone )( 
            ScreenSettings * This,
            /* [in] */ VARIANT prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Volume )( 
            ScreenSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Volume )( 
            ScreenSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Punctuation )( 
            ScreenSettings * This,
            /* [retval][out] */ PunctuationSettings **punc);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Numbers )( 
            ScreenSettings * This,
            /* [retval][out] */ NumberMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Numbers )( 
            ScreenSettings * This,
            /* [in] */ NumberMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CapitalizationAlert )( 
            ScreenSettings * This,
            /* [retval][out] */ CapitalizationMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CapitalizationAlert )( 
            ScreenSettings * This,
            /* [in] */ CapitalizationMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FormatAlert )( 
            ScreenSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FormatAlert )( 
            ScreenSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpellAlphaNumeric )( 
            ScreenSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpellAlphaNumeric )( 
            ScreenSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SplitMixedCaseWords )( 
            ScreenSettings * This,
            /* [retval][out] */ MixedCaseMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SplitMixedCaseWords )( 
            ScreenSettings * This,
            /* [in] */ MixedCaseMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_RepeatFilter )( 
            ScreenSettings * This,
            /* [retval][out] */ RepeatMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_RepeatFilter )( 
            ScreenSettings * This,
            /* [in] */ RepeatMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_RepeatValue )( 
            ScreenSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_RepeatValue )( 
            ScreenSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BlankLines )( 
            ScreenSettings * This,
            /* [retval][out] */ BlankLinesMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_BlankLines )( 
            ScreenSettings * This,
            /* [in] */ BlankLinesMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SignalCursorPosition )( 
            ScreenSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SignalCursorPosition )( 
            ScreenSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AttributeChanges )( 
            ScreenSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AttributeChanges )( 
            ScreenSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadParagraphStyle )( 
            ScreenSettings * This,
            /* [retval][out] */ AutoMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadParagraphStyle )( 
            ScreenSettings * This,
            /* [in] */ AutoMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LineFilter )( 
            ScreenSettings * This,
            /* [retval][out] */ LineFilterSettings **LineFilter);
        
        END_INTERFACE
    } ScreenSettingsVtbl;

    interface ScreenSettings
    {
        CONST_VTBL struct ScreenSettingsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ScreenSettings_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ScreenSettings_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ScreenSettings_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ScreenSettings_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ScreenSettings_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ScreenSettings_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ScreenSettings_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ScreenSettings_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ScreenSettings_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ScreenSettings_get_Voice(This,prop)	\
    ( (This)->lpVtbl -> get_Voice(This,prop) ) 

#define ScreenSettings_put_Voice(This,prop)	\
    ( (This)->lpVtbl -> put_Voice(This,prop) ) 

#define ScreenSettings_get_Rate(This,prop)	\
    ( (This)->lpVtbl -> get_Rate(This,prop) ) 

#define ScreenSettings_put_Rate(This,prop)	\
    ( (This)->lpVtbl -> put_Rate(This,prop) ) 

#define ScreenSettings_get_Pitch(This,prop)	\
    ( (This)->lpVtbl -> get_Pitch(This,prop) ) 

#define ScreenSettings_put_Pitch(This,prop)	\
    ( (This)->lpVtbl -> put_Pitch(This,prop) ) 

#define ScreenSettings_get_Tone(This,prop)	\
    ( (This)->lpVtbl -> get_Tone(This,prop) ) 

#define ScreenSettings_put_Tone(This,prop)	\
    ( (This)->lpVtbl -> put_Tone(This,prop) ) 

#define ScreenSettings_get_Volume(This,prop)	\
    ( (This)->lpVtbl -> get_Volume(This,prop) ) 

#define ScreenSettings_put_Volume(This,prop)	\
    ( (This)->lpVtbl -> put_Volume(This,prop) ) 

#define ScreenSettings_get_Punctuation(This,punc)	\
    ( (This)->lpVtbl -> get_Punctuation(This,punc) ) 

#define ScreenSettings_get_Numbers(This,prop)	\
    ( (This)->lpVtbl -> get_Numbers(This,prop) ) 

#define ScreenSettings_put_Numbers(This,prop)	\
    ( (This)->lpVtbl -> put_Numbers(This,prop) ) 

#define ScreenSettings_get_CapitalizationAlert(This,prop)	\
    ( (This)->lpVtbl -> get_CapitalizationAlert(This,prop) ) 

#define ScreenSettings_put_CapitalizationAlert(This,prop)	\
    ( (This)->lpVtbl -> put_CapitalizationAlert(This,prop) ) 

#define ScreenSettings_get_FormatAlert(This,prop)	\
    ( (This)->lpVtbl -> get_FormatAlert(This,prop) ) 

#define ScreenSettings_put_FormatAlert(This,prop)	\
    ( (This)->lpVtbl -> put_FormatAlert(This,prop) ) 

#define ScreenSettings_get_SpellAlphaNumeric(This,prop)	\
    ( (This)->lpVtbl -> get_SpellAlphaNumeric(This,prop) ) 

#define ScreenSettings_put_SpellAlphaNumeric(This,prop)	\
    ( (This)->lpVtbl -> put_SpellAlphaNumeric(This,prop) ) 

#define ScreenSettings_get_SplitMixedCaseWords(This,prop)	\
    ( (This)->lpVtbl -> get_SplitMixedCaseWords(This,prop) ) 

#define ScreenSettings_put_SplitMixedCaseWords(This,prop)	\
    ( (This)->lpVtbl -> put_SplitMixedCaseWords(This,prop) ) 

#define ScreenSettings_get_RepeatFilter(This,prop)	\
    ( (This)->lpVtbl -> get_RepeatFilter(This,prop) ) 

#define ScreenSettings_put_RepeatFilter(This,prop)	\
    ( (This)->lpVtbl -> put_RepeatFilter(This,prop) ) 

#define ScreenSettings_get_RepeatValue(This,prop)	\
    ( (This)->lpVtbl -> get_RepeatValue(This,prop) ) 

#define ScreenSettings_put_RepeatValue(This,prop)	\
    ( (This)->lpVtbl -> put_RepeatValue(This,prop) ) 

#define ScreenSettings_get_BlankLines(This,prop)	\
    ( (This)->lpVtbl -> get_BlankLines(This,prop) ) 

#define ScreenSettings_put_BlankLines(This,prop)	\
    ( (This)->lpVtbl -> put_BlankLines(This,prop) ) 

#define ScreenSettings_get_SignalCursorPosition(This,prop)	\
    ( (This)->lpVtbl -> get_SignalCursorPosition(This,prop) ) 

#define ScreenSettings_put_SignalCursorPosition(This,prop)	\
    ( (This)->lpVtbl -> put_SignalCursorPosition(This,prop) ) 

#define ScreenSettings_get_AttributeChanges(This,prop)	\
    ( (This)->lpVtbl -> get_AttributeChanges(This,prop) ) 

#define ScreenSettings_put_AttributeChanges(This,prop)	\
    ( (This)->lpVtbl -> put_AttributeChanges(This,prop) ) 

#define ScreenSettings_get_ReadParagraphStyle(This,prop)	\
    ( (This)->lpVtbl -> get_ReadParagraphStyle(This,prop) ) 

#define ScreenSettings_put_ReadParagraphStyle(This,prop)	\
    ( (This)->lpVtbl -> put_ReadParagraphStyle(This,prop) ) 

#define ScreenSettings_get_LineFilter(This,LineFilter)	\
    ( (This)->lpVtbl -> get_LineFilter(This,LineFilter) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ScreenSettings_INTERFACE_DEFINED__ */


#ifndef __PunctuationSettings_INTERFACE_DEFINED__
#define __PunctuationSettings_INTERFACE_DEFINED__

/* interface PunctuationSettings */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_PunctuationSettings;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("BE2E7D51-873F-4F95-94E9-532566223B83")
    PunctuationSettings : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Textual( 
            /* [retval][out] */ PunctuationMode *mode) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Textual( 
            /* [in] */ PunctuationMode mode) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Math( 
            /* [retval][out] */ PunctuationMode *mode) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Math( 
            /* [in] */ PunctuationMode mode) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Miscellaneous( 
            /* [retval][out] */ PunctuationMode *mode) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Miscellaneous( 
            /* [in] */ PunctuationMode mode) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Space( 
            /* [retval][out] */ PunctuationMode *mode) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Space( 
            /* [in] */ PunctuationMode mode) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct PunctuationSettingsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            PunctuationSettings * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            PunctuationSettings * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            PunctuationSettings * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            PunctuationSettings * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            PunctuationSettings * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            PunctuationSettings * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            PunctuationSettings * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            PunctuationSettings * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            PunctuationSettings * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Textual )( 
            PunctuationSettings * This,
            /* [retval][out] */ PunctuationMode *mode);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Textual )( 
            PunctuationSettings * This,
            /* [in] */ PunctuationMode mode);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Math )( 
            PunctuationSettings * This,
            /* [retval][out] */ PunctuationMode *mode);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Math )( 
            PunctuationSettings * This,
            /* [in] */ PunctuationMode mode);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Miscellaneous )( 
            PunctuationSettings * This,
            /* [retval][out] */ PunctuationMode *mode);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Miscellaneous )( 
            PunctuationSettings * This,
            /* [in] */ PunctuationMode mode);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Space )( 
            PunctuationSettings * This,
            /* [retval][out] */ PunctuationMode *mode);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Space )( 
            PunctuationSettings * This,
            /* [in] */ PunctuationMode mode);
        
        END_INTERFACE
    } PunctuationSettingsVtbl;

    interface PunctuationSettings
    {
        CONST_VTBL struct PunctuationSettingsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define PunctuationSettings_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define PunctuationSettings_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define PunctuationSettings_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define PunctuationSettings_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define PunctuationSettings_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define PunctuationSettings_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define PunctuationSettings_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define PunctuationSettings_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define PunctuationSettings_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define PunctuationSettings_get_Textual(This,mode)	\
    ( (This)->lpVtbl -> get_Textual(This,mode) ) 

#define PunctuationSettings_put_Textual(This,mode)	\
    ( (This)->lpVtbl -> put_Textual(This,mode) ) 

#define PunctuationSettings_get_Math(This,mode)	\
    ( (This)->lpVtbl -> get_Math(This,mode) ) 

#define PunctuationSettings_put_Math(This,mode)	\
    ( (This)->lpVtbl -> put_Math(This,mode) ) 

#define PunctuationSettings_get_Miscellaneous(This,mode)	\
    ( (This)->lpVtbl -> get_Miscellaneous(This,mode) ) 

#define PunctuationSettings_put_Miscellaneous(This,mode)	\
    ( (This)->lpVtbl -> put_Miscellaneous(This,mode) ) 

#define PunctuationSettings_get_Space(This,mode)	\
    ( (This)->lpVtbl -> get_Space(This,mode) ) 

#define PunctuationSettings_put_Space(This,mode)	\
    ( (This)->lpVtbl -> put_Space(This,mode) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __PunctuationSettings_INTERFACE_DEFINED__ */


#ifndef __LineFilterSettings_INTERFACE_DEFINED__
#define __LineFilterSettings_INTERFACE_DEFINED__

/* interface LineFilterSettings */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_LineFilterSettings;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("AE828844-7DD7-4FA3-9B9A-80D18E241449")
    LineFilterSettings : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Status( 
            /* [retval][out] */ LineFilterMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Status( 
            /* [in] */ LineFilterMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Position( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Position( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Characters( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Characters( 
            /* [in] */ BSTR prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct LineFilterSettingsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            LineFilterSettings * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            LineFilterSettings * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            LineFilterSettings * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            LineFilterSettings * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            LineFilterSettings * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            LineFilterSettings * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            LineFilterSettings * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            LineFilterSettings * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            LineFilterSettings * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Status )( 
            LineFilterSettings * This,
            /* [retval][out] */ LineFilterMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Status )( 
            LineFilterSettings * This,
            /* [in] */ LineFilterMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            LineFilterSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            LineFilterSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Characters )( 
            LineFilterSettings * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Characters )( 
            LineFilterSettings * This,
            /* [in] */ BSTR prop);
        
        END_INTERFACE
    } LineFilterSettingsVtbl;

    interface LineFilterSettings
    {
        CONST_VTBL struct LineFilterSettingsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define LineFilterSettings_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define LineFilterSettings_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define LineFilterSettings_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define LineFilterSettings_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define LineFilterSettings_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define LineFilterSettings_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define LineFilterSettings_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define LineFilterSettings_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define LineFilterSettings_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define LineFilterSettings_get_Status(This,prop)	\
    ( (This)->lpVtbl -> get_Status(This,prop) ) 

#define LineFilterSettings_put_Status(This,prop)	\
    ( (This)->lpVtbl -> put_Status(This,prop) ) 

#define LineFilterSettings_get_Position(This,prop)	\
    ( (This)->lpVtbl -> get_Position(This,prop) ) 

#define LineFilterSettings_put_Position(This,prop)	\
    ( (This)->lpVtbl -> put_Position(This,prop) ) 

#define LineFilterSettings_get_Characters(This,prop)	\
    ( (This)->lpVtbl -> get_Characters(This,prop) ) 

#define LineFilterSettings_put_Characters(This,prop)	\
    ( (This)->lpVtbl -> put_Characters(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __LineFilterSettings_INTERFACE_DEFINED__ */


#ifndef __KeyboardSettings_INTERFACE_DEFINED__
#define __KeyboardSettings_INTERFACE_DEFINED__

/* interface KeyboardSettings */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_KeyboardSettings;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("16EC8F0F-849A-4B7F-AFF9-B30FA22F8F54")
    KeyboardSettings : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Voice( 
            /* [retval][out] */ KeyboardVoiceMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Voice( 
            /* [in] */ KeyboardVoiceMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Rate( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Rate( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Pitch( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Pitch( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Tone( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Tone( 
            /* [in] */ VARIANT prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Volume( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Volume( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Punctuation( 
            /* [retval][out] */ PunctuationSettings **punc) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpecialKeys( 
            /* [retval][out] */ SpecialKeySettings **SpecialKeys) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CapitalMode( 
            /* [retval][out] */ CapitalizationMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CapitalMode( 
            /* [in] */ CapitalizationMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Interruptability( 
            /* [retval][out] */ KeyboardInterruptabilityMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Interruptability( 
            /* [in] */ KeyboardInterruptabilityMode prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct KeyboardSettingsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            KeyboardSettings * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            KeyboardSettings * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            KeyboardSettings * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            KeyboardSettings * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            KeyboardSettings * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            KeyboardSettings * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            KeyboardSettings * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            KeyboardSettings * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            KeyboardSettings * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Voice )( 
            KeyboardSettings * This,
            /* [retval][out] */ KeyboardVoiceMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Voice )( 
            KeyboardSettings * This,
            /* [in] */ KeyboardVoiceMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Rate )( 
            KeyboardSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Rate )( 
            KeyboardSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Pitch )( 
            KeyboardSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Pitch )( 
            KeyboardSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Tone )( 
            KeyboardSettings * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Tone )( 
            KeyboardSettings * This,
            /* [in] */ VARIANT prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Volume )( 
            KeyboardSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Volume )( 
            KeyboardSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Punctuation )( 
            KeyboardSettings * This,
            /* [retval][out] */ PunctuationSettings **punc);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpecialKeys )( 
            KeyboardSettings * This,
            /* [retval][out] */ SpecialKeySettings **SpecialKeys);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CapitalMode )( 
            KeyboardSettings * This,
            /* [retval][out] */ CapitalizationMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CapitalMode )( 
            KeyboardSettings * This,
            /* [in] */ CapitalizationMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Interruptability )( 
            KeyboardSettings * This,
            /* [retval][out] */ KeyboardInterruptabilityMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Interruptability )( 
            KeyboardSettings * This,
            /* [in] */ KeyboardInterruptabilityMode prop);
        
        END_INTERFACE
    } KeyboardSettingsVtbl;

    interface KeyboardSettings
    {
        CONST_VTBL struct KeyboardSettingsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define KeyboardSettings_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define KeyboardSettings_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define KeyboardSettings_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define KeyboardSettings_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define KeyboardSettings_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define KeyboardSettings_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define KeyboardSettings_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define KeyboardSettings_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define KeyboardSettings_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define KeyboardSettings_get_Voice(This,prop)	\
    ( (This)->lpVtbl -> get_Voice(This,prop) ) 

#define KeyboardSettings_put_Voice(This,prop)	\
    ( (This)->lpVtbl -> put_Voice(This,prop) ) 

#define KeyboardSettings_get_Rate(This,prop)	\
    ( (This)->lpVtbl -> get_Rate(This,prop) ) 

#define KeyboardSettings_put_Rate(This,prop)	\
    ( (This)->lpVtbl -> put_Rate(This,prop) ) 

#define KeyboardSettings_get_Pitch(This,prop)	\
    ( (This)->lpVtbl -> get_Pitch(This,prop) ) 

#define KeyboardSettings_put_Pitch(This,prop)	\
    ( (This)->lpVtbl -> put_Pitch(This,prop) ) 

#define KeyboardSettings_get_Tone(This,prop)	\
    ( (This)->lpVtbl -> get_Tone(This,prop) ) 

#define KeyboardSettings_put_Tone(This,prop)	\
    ( (This)->lpVtbl -> put_Tone(This,prop) ) 

#define KeyboardSettings_get_Volume(This,prop)	\
    ( (This)->lpVtbl -> get_Volume(This,prop) ) 

#define KeyboardSettings_put_Volume(This,prop)	\
    ( (This)->lpVtbl -> put_Volume(This,prop) ) 

#define KeyboardSettings_get_Punctuation(This,punc)	\
    ( (This)->lpVtbl -> get_Punctuation(This,punc) ) 

#define KeyboardSettings_get_SpecialKeys(This,SpecialKeys)	\
    ( (This)->lpVtbl -> get_SpecialKeys(This,SpecialKeys) ) 

#define KeyboardSettings_get_CapitalMode(This,prop)	\
    ( (This)->lpVtbl -> get_CapitalMode(This,prop) ) 

#define KeyboardSettings_put_CapitalMode(This,prop)	\
    ( (This)->lpVtbl -> put_CapitalMode(This,prop) ) 

#define KeyboardSettings_get_Interruptability(This,prop)	\
    ( (This)->lpVtbl -> get_Interruptability(This,prop) ) 

#define KeyboardSettings_put_Interruptability(This,prop)	\
    ( (This)->lpVtbl -> put_Interruptability(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __KeyboardSettings_INTERFACE_DEFINED__ */


#ifndef __SpecialKeySettings_INTERFACE_DEFINED__
#define __SpecialKeySettings_INTERFACE_DEFINED__

/* interface SpecialKeySettings */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_SpecialKeySettings;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("D1870316-B206-4081-B219-525B338F5B74")
    SpecialKeySettings : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FunctionKeys( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FunctionKeys( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ToggleKeys( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ToggleKeys( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_KeyEnhancements( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_KeyEnhancements( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EscapeKey( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EscapeKey( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EnterKey( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EnterKey( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EnableBothInsertsAsModifier( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EnableBothInsertsAsModifier( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CursorMovement( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CursorMovement( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EnableCapsLockAsInsertKey( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EnableCapsLockAsInsertKey( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct SpecialKeySettingsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            SpecialKeySettings * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            SpecialKeySettings * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            SpecialKeySettings * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            SpecialKeySettings * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            SpecialKeySettings * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            SpecialKeySettings * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            SpecialKeySettings * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            SpecialKeySettings * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            SpecialKeySettings * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FunctionKeys )( 
            SpecialKeySettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FunctionKeys )( 
            SpecialKeySettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ToggleKeys )( 
            SpecialKeySettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ToggleKeys )( 
            SpecialKeySettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_KeyEnhancements )( 
            SpecialKeySettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_KeyEnhancements )( 
            SpecialKeySettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EscapeKey )( 
            SpecialKeySettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EscapeKey )( 
            SpecialKeySettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EnterKey )( 
            SpecialKeySettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EnterKey )( 
            SpecialKeySettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EnableBothInsertsAsModifier )( 
            SpecialKeySettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EnableBothInsertsAsModifier )( 
            SpecialKeySettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CursorMovement )( 
            SpecialKeySettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CursorMovement )( 
            SpecialKeySettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EnableCapsLockAsInsertKey )( 
            SpecialKeySettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EnableCapsLockAsInsertKey )( 
            SpecialKeySettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        END_INTERFACE
    } SpecialKeySettingsVtbl;

    interface SpecialKeySettings
    {
        CONST_VTBL struct SpecialKeySettingsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define SpecialKeySettings_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define SpecialKeySettings_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define SpecialKeySettings_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define SpecialKeySettings_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define SpecialKeySettings_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define SpecialKeySettings_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define SpecialKeySettings_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define SpecialKeySettings_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define SpecialKeySettings_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define SpecialKeySettings_get_FunctionKeys(This,prop)	\
    ( (This)->lpVtbl -> get_FunctionKeys(This,prop) ) 

#define SpecialKeySettings_put_FunctionKeys(This,prop)	\
    ( (This)->lpVtbl -> put_FunctionKeys(This,prop) ) 

#define SpecialKeySettings_get_ToggleKeys(This,prop)	\
    ( (This)->lpVtbl -> get_ToggleKeys(This,prop) ) 

#define SpecialKeySettings_put_ToggleKeys(This,prop)	\
    ( (This)->lpVtbl -> put_ToggleKeys(This,prop) ) 

#define SpecialKeySettings_get_KeyEnhancements(This,prop)	\
    ( (This)->lpVtbl -> get_KeyEnhancements(This,prop) ) 

#define SpecialKeySettings_put_KeyEnhancements(This,prop)	\
    ( (This)->lpVtbl -> put_KeyEnhancements(This,prop) ) 

#define SpecialKeySettings_get_EscapeKey(This,prop)	\
    ( (This)->lpVtbl -> get_EscapeKey(This,prop) ) 

#define SpecialKeySettings_put_EscapeKey(This,prop)	\
    ( (This)->lpVtbl -> put_EscapeKey(This,prop) ) 

#define SpecialKeySettings_get_EnterKey(This,prop)	\
    ( (This)->lpVtbl -> get_EnterKey(This,prop) ) 

#define SpecialKeySettings_put_EnterKey(This,prop)	\
    ( (This)->lpVtbl -> put_EnterKey(This,prop) ) 

#define SpecialKeySettings_get_EnableBothInsertsAsModifier(This,prop)	\
    ( (This)->lpVtbl -> get_EnableBothInsertsAsModifier(This,prop) ) 

#define SpecialKeySettings_put_EnableBothInsertsAsModifier(This,prop)	\
    ( (This)->lpVtbl -> put_EnableBothInsertsAsModifier(This,prop) ) 

#define SpecialKeySettings_get_CursorMovement(This,prop)	\
    ( (This)->lpVtbl -> get_CursorMovement(This,prop) ) 

#define SpecialKeySettings_put_CursorMovement(This,prop)	\
    ( (This)->lpVtbl -> put_CursorMovement(This,prop) ) 

#define SpecialKeySettings_get_EnableCapsLockAsInsertKey(This,prop)	\
    ( (This)->lpVtbl -> get_EnableCapsLockAsInsertKey(This,prop) ) 

#define SpecialKeySettings_put_EnableCapsLockAsInsertKey(This,prop)	\
    ( (This)->lpVtbl -> put_EnableCapsLockAsInsertKey(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __SpecialKeySettings_INTERFACE_DEFINED__ */


#ifndef __MouseSettings_INTERFACE_DEFINED__
#define __MouseSettings_INTERFACE_DEFINED__

/* interface MouseSettings */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_MouseSettings;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("6B979D0D-51A4-4495-89E9-8A221B471E10")
    MouseSettings : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Voice( 
            /* [retval][out] */ MouseVoiceMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Voice( 
            /* [in] */ MouseVoiceMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Rate( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Rate( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Pitch( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Pitch( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Tone( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Tone( 
            /* [in] */ VARIANT prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Volume( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Volume( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Boundary( 
            /* [retval][out] */ MouseBoundaryMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Boundary( 
            /* [in] */ MouseBoundaryMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Buttons( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Buttons( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_PointerUpDelta( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_PointerUpDelta( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_PointerDownDelta( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_PointerDownDelta( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_PointerLeftDelta( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_PointerLeftDelta( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_PointerRightDelta( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_PointerRightDelta( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Interruptability( 
            /* [retval][out] */ MouseInterruptabilityMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Interruptability( 
            /* [in] */ MouseInterruptabilityMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TrackFocus( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TrackFocus( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UpDownSpeak( 
            /* [retval][out] */ MouseUpDownMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_UpDownSpeak( 
            /* [in] */ MouseUpDownMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SearchAttributes( 
            /* [retval][out] */ MouseSearchAttributesSettings **searchAttribs) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct MouseSettingsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            MouseSettings * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            MouseSettings * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            MouseSettings * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            MouseSettings * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            MouseSettings * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            MouseSettings * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            MouseSettings * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            MouseSettings * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            MouseSettings * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Voice )( 
            MouseSettings * This,
            /* [retval][out] */ MouseVoiceMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Voice )( 
            MouseSettings * This,
            /* [in] */ MouseVoiceMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Rate )( 
            MouseSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Rate )( 
            MouseSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Pitch )( 
            MouseSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Pitch )( 
            MouseSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Tone )( 
            MouseSettings * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Tone )( 
            MouseSettings * This,
            /* [in] */ VARIANT prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Volume )( 
            MouseSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Volume )( 
            MouseSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Boundary )( 
            MouseSettings * This,
            /* [retval][out] */ MouseBoundaryMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Boundary )( 
            MouseSettings * This,
            /* [in] */ MouseBoundaryMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Buttons )( 
            MouseSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Buttons )( 
            MouseSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_PointerUpDelta )( 
            MouseSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_PointerUpDelta )( 
            MouseSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_PointerDownDelta )( 
            MouseSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_PointerDownDelta )( 
            MouseSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_PointerLeftDelta )( 
            MouseSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_PointerLeftDelta )( 
            MouseSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_PointerRightDelta )( 
            MouseSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_PointerRightDelta )( 
            MouseSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Interruptability )( 
            MouseSettings * This,
            /* [retval][out] */ MouseInterruptabilityMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Interruptability )( 
            MouseSettings * This,
            /* [in] */ MouseInterruptabilityMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TrackFocus )( 
            MouseSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TrackFocus )( 
            MouseSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UpDownSpeak )( 
            MouseSettings * This,
            /* [retval][out] */ MouseUpDownMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_UpDownSpeak )( 
            MouseSettings * This,
            /* [in] */ MouseUpDownMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SearchAttributes )( 
            MouseSettings * This,
            /* [retval][out] */ MouseSearchAttributesSettings **searchAttribs);
        
        END_INTERFACE
    } MouseSettingsVtbl;

    interface MouseSettings
    {
        CONST_VTBL struct MouseSettingsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define MouseSettings_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define MouseSettings_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define MouseSettings_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define MouseSettings_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define MouseSettings_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define MouseSettings_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define MouseSettings_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define MouseSettings_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define MouseSettings_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define MouseSettings_get_Voice(This,prop)	\
    ( (This)->lpVtbl -> get_Voice(This,prop) ) 

#define MouseSettings_put_Voice(This,prop)	\
    ( (This)->lpVtbl -> put_Voice(This,prop) ) 

#define MouseSettings_get_Rate(This,prop)	\
    ( (This)->lpVtbl -> get_Rate(This,prop) ) 

#define MouseSettings_put_Rate(This,prop)	\
    ( (This)->lpVtbl -> put_Rate(This,prop) ) 

#define MouseSettings_get_Pitch(This,prop)	\
    ( (This)->lpVtbl -> get_Pitch(This,prop) ) 

#define MouseSettings_put_Pitch(This,prop)	\
    ( (This)->lpVtbl -> put_Pitch(This,prop) ) 

#define MouseSettings_get_Tone(This,prop)	\
    ( (This)->lpVtbl -> get_Tone(This,prop) ) 

#define MouseSettings_put_Tone(This,prop)	\
    ( (This)->lpVtbl -> put_Tone(This,prop) ) 

#define MouseSettings_get_Volume(This,prop)	\
    ( (This)->lpVtbl -> get_Volume(This,prop) ) 

#define MouseSettings_put_Volume(This,prop)	\
    ( (This)->lpVtbl -> put_Volume(This,prop) ) 

#define MouseSettings_get_Boundary(This,prop)	\
    ( (This)->lpVtbl -> get_Boundary(This,prop) ) 

#define MouseSettings_put_Boundary(This,prop)	\
    ( (This)->lpVtbl -> put_Boundary(This,prop) ) 

#define MouseSettings_get_Buttons(This,prop)	\
    ( (This)->lpVtbl -> get_Buttons(This,prop) ) 

#define MouseSettings_put_Buttons(This,prop)	\
    ( (This)->lpVtbl -> put_Buttons(This,prop) ) 

#define MouseSettings_get_PointerUpDelta(This,prop)	\
    ( (This)->lpVtbl -> get_PointerUpDelta(This,prop) ) 

#define MouseSettings_put_PointerUpDelta(This,prop)	\
    ( (This)->lpVtbl -> put_PointerUpDelta(This,prop) ) 

#define MouseSettings_get_PointerDownDelta(This,prop)	\
    ( (This)->lpVtbl -> get_PointerDownDelta(This,prop) ) 

#define MouseSettings_put_PointerDownDelta(This,prop)	\
    ( (This)->lpVtbl -> put_PointerDownDelta(This,prop) ) 

#define MouseSettings_get_PointerLeftDelta(This,prop)	\
    ( (This)->lpVtbl -> get_PointerLeftDelta(This,prop) ) 

#define MouseSettings_put_PointerLeftDelta(This,prop)	\
    ( (This)->lpVtbl -> put_PointerLeftDelta(This,prop) ) 

#define MouseSettings_get_PointerRightDelta(This,prop)	\
    ( (This)->lpVtbl -> get_PointerRightDelta(This,prop) ) 

#define MouseSettings_put_PointerRightDelta(This,prop)	\
    ( (This)->lpVtbl -> put_PointerRightDelta(This,prop) ) 

#define MouseSettings_get_Interruptability(This,prop)	\
    ( (This)->lpVtbl -> get_Interruptability(This,prop) ) 

#define MouseSettings_put_Interruptability(This,prop)	\
    ( (This)->lpVtbl -> put_Interruptability(This,prop) ) 

#define MouseSettings_get_TrackFocus(This,prop)	\
    ( (This)->lpVtbl -> get_TrackFocus(This,prop) ) 

#define MouseSettings_put_TrackFocus(This,prop)	\
    ( (This)->lpVtbl -> put_TrackFocus(This,prop) ) 

#define MouseSettings_get_UpDownSpeak(This,prop)	\
    ( (This)->lpVtbl -> get_UpDownSpeak(This,prop) ) 

#define MouseSettings_put_UpDownSpeak(This,prop)	\
    ( (This)->lpVtbl -> put_UpDownSpeak(This,prop) ) 

#define MouseSettings_get_SearchAttributes(This,searchAttribs)	\
    ( (This)->lpVtbl -> get_SearchAttributes(This,searchAttribs) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __MouseSettings_INTERFACE_DEFINED__ */


#ifndef __MouseSearchAttributesSettings_INTERFACE_DEFINED__
#define __MouseSearchAttributesSettings_INTERFACE_DEFINED__

/* interface MouseSearchAttributesSettings */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_MouseSearchAttributesSettings;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("6766D2E3-4358-4F48-A9AA-766094DAC00D")
    MouseSearchAttributesSettings : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Underlined( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Underlined( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Bold( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Bold( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Highlighted( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Highlighted( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Italic( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Italic( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Strikeout( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Strikeout( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct MouseSearchAttributesSettingsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            MouseSearchAttributesSettings * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            MouseSearchAttributesSettings * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            MouseSearchAttributesSettings * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            MouseSearchAttributesSettings * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            MouseSearchAttributesSettings * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            MouseSearchAttributesSettings * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            MouseSearchAttributesSettings * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            MouseSearchAttributesSettings * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            MouseSearchAttributesSettings * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Underlined )( 
            MouseSearchAttributesSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Underlined )( 
            MouseSearchAttributesSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Bold )( 
            MouseSearchAttributesSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Bold )( 
            MouseSearchAttributesSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Highlighted )( 
            MouseSearchAttributesSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Highlighted )( 
            MouseSearchAttributesSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Italic )( 
            MouseSearchAttributesSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Italic )( 
            MouseSearchAttributesSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Strikeout )( 
            MouseSearchAttributesSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Strikeout )( 
            MouseSearchAttributesSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        END_INTERFACE
    } MouseSearchAttributesSettingsVtbl;

    interface MouseSearchAttributesSettings
    {
        CONST_VTBL struct MouseSearchAttributesSettingsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define MouseSearchAttributesSettings_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define MouseSearchAttributesSettings_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define MouseSearchAttributesSettings_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define MouseSearchAttributesSettings_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define MouseSearchAttributesSettings_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define MouseSearchAttributesSettings_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define MouseSearchAttributesSettings_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define MouseSearchAttributesSettings_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define MouseSearchAttributesSettings_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define MouseSearchAttributesSettings_get_Underlined(This,prop)	\
    ( (This)->lpVtbl -> get_Underlined(This,prop) ) 

#define MouseSearchAttributesSettings_put_Underlined(This,prop)	\
    ( (This)->lpVtbl -> put_Underlined(This,prop) ) 

#define MouseSearchAttributesSettings_get_Bold(This,prop)	\
    ( (This)->lpVtbl -> get_Bold(This,prop) ) 

#define MouseSearchAttributesSettings_put_Bold(This,prop)	\
    ( (This)->lpVtbl -> put_Bold(This,prop) ) 

#define MouseSearchAttributesSettings_get_Highlighted(This,prop)	\
    ( (This)->lpVtbl -> get_Highlighted(This,prop) ) 

#define MouseSearchAttributesSettings_put_Highlighted(This,prop)	\
    ( (This)->lpVtbl -> put_Highlighted(This,prop) ) 

#define MouseSearchAttributesSettings_get_Italic(This,prop)	\
    ( (This)->lpVtbl -> get_Italic(This,prop) ) 

#define MouseSearchAttributesSettings_put_Italic(This,prop)	\
    ( (This)->lpVtbl -> put_Italic(This,prop) ) 

#define MouseSearchAttributesSettings_get_Strikeout(This,prop)	\
    ( (This)->lpVtbl -> get_Strikeout(This,prop) ) 

#define MouseSearchAttributesSettings_put_Strikeout(This,prop)	\
    ( (This)->lpVtbl -> put_Strikeout(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __MouseSearchAttributesSettings_INTERFACE_DEFINED__ */


#ifndef __GeneralSettings_INTERFACE_DEFINED__
#define __GeneralSettings_INTERFACE_DEFINED__

/* interface GeneralSettings */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_GeneralSettings;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("560A43E2-CFBC-44EE-82A1-1149EBB9569F")
    GeneralSettings : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Voice( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Voice( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Braille( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Braille( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HotKeys( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_HotKeys( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CursorKeys( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CursorKeys( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HighlightTrack( 
            /* [retval][out] */ AutoMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_HighlightTrack( 
            /* [in] */ AutoMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CharacterDictionary( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CharacterDictionary( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_KeyLabelDictionary( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_KeyLabelDictionary( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ExceptionDictionary( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ExceptionDictionary( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_GraphicDictionary( 
            /* [retval][out] */ GraphicDictionaryMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_GraphicDictionary( 
            /* [in] */ GraphicDictionaryMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ColorDictionary( 
            /* [retval][out] */ ColorDictionaryMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ColorDictionary( 
            /* [in] */ ColorDictionaryMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AllowSpeakWindowsInEditBoxes( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AllowSpeakWindowsInEditBoxes( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CursorDelay( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CursorDelay( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TriggerDelay( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TriggerDelay( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpaceThreshold( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpaceThreshold( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AllowAutoRedraw( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AllowAutoRedraw( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpeakTooltips( 
            /* [retval][out] */ AutoMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpeakTooltips( 
            /* [in] */ AutoMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BrowseMode( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_BrowseMode( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeBoxInHighlight( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeBoxInHighlight( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HighlightedText( 
            /* [retval][out] */ HighlightTextMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_HighlightedText( 
            /* [in] */ HighlightTextMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeUserWindowWithBox( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeUserWindowWithBox( 
            /* [in] */ short prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct GeneralSettingsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            GeneralSettings * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            GeneralSettings * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            GeneralSettings * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            GeneralSettings * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            GeneralSettings * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            GeneralSettings * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            GeneralSettings * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            GeneralSettings * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            GeneralSettings * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Voice )( 
            GeneralSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Voice )( 
            GeneralSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Braille )( 
            GeneralSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Braille )( 
            GeneralSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HotKeys )( 
            GeneralSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_HotKeys )( 
            GeneralSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CursorKeys )( 
            GeneralSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CursorKeys )( 
            GeneralSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HighlightTrack )( 
            GeneralSettings * This,
            /* [retval][out] */ AutoMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_HighlightTrack )( 
            GeneralSettings * This,
            /* [in] */ AutoMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CharacterDictionary )( 
            GeneralSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CharacterDictionary )( 
            GeneralSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_KeyLabelDictionary )( 
            GeneralSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_KeyLabelDictionary )( 
            GeneralSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ExceptionDictionary )( 
            GeneralSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ExceptionDictionary )( 
            GeneralSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_GraphicDictionary )( 
            GeneralSettings * This,
            /* [retval][out] */ GraphicDictionaryMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_GraphicDictionary )( 
            GeneralSettings * This,
            /* [in] */ GraphicDictionaryMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ColorDictionary )( 
            GeneralSettings * This,
            /* [retval][out] */ ColorDictionaryMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ColorDictionary )( 
            GeneralSettings * This,
            /* [in] */ ColorDictionaryMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AllowSpeakWindowsInEditBoxes )( 
            GeneralSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AllowSpeakWindowsInEditBoxes )( 
            GeneralSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CursorDelay )( 
            GeneralSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CursorDelay )( 
            GeneralSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TriggerDelay )( 
            GeneralSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TriggerDelay )( 
            GeneralSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpaceThreshold )( 
            GeneralSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpaceThreshold )( 
            GeneralSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AllowAutoRedraw )( 
            GeneralSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AllowAutoRedraw )( 
            GeneralSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpeakTooltips )( 
            GeneralSettings * This,
            /* [retval][out] */ AutoMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpeakTooltips )( 
            GeneralSettings * This,
            /* [in] */ AutoMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BrowseMode )( 
            GeneralSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_BrowseMode )( 
            GeneralSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeBoxInHighlight )( 
            GeneralSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeBoxInHighlight )( 
            GeneralSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HighlightedText )( 
            GeneralSettings * This,
            /* [retval][out] */ HighlightTextMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_HighlightedText )( 
            GeneralSettings * This,
            /* [in] */ HighlightTextMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeUserWindowWithBox )( 
            GeneralSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeUserWindowWithBox )( 
            GeneralSettings * This,
            /* [in] */ short prop);
        
        END_INTERFACE
    } GeneralSettingsVtbl;

    interface GeneralSettings
    {
        CONST_VTBL struct GeneralSettingsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define GeneralSettings_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define GeneralSettings_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define GeneralSettings_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define GeneralSettings_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define GeneralSettings_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define GeneralSettings_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define GeneralSettings_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define GeneralSettings_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define GeneralSettings_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define GeneralSettings_get_Voice(This,prop)	\
    ( (This)->lpVtbl -> get_Voice(This,prop) ) 

#define GeneralSettings_put_Voice(This,prop)	\
    ( (This)->lpVtbl -> put_Voice(This,prop) ) 

#define GeneralSettings_get_Braille(This,prop)	\
    ( (This)->lpVtbl -> get_Braille(This,prop) ) 

#define GeneralSettings_put_Braille(This,prop)	\
    ( (This)->lpVtbl -> put_Braille(This,prop) ) 

#define GeneralSettings_get_HotKeys(This,prop)	\
    ( (This)->lpVtbl -> get_HotKeys(This,prop) ) 

#define GeneralSettings_put_HotKeys(This,prop)	\
    ( (This)->lpVtbl -> put_HotKeys(This,prop) ) 

#define GeneralSettings_get_CursorKeys(This,prop)	\
    ( (This)->lpVtbl -> get_CursorKeys(This,prop) ) 

#define GeneralSettings_put_CursorKeys(This,prop)	\
    ( (This)->lpVtbl -> put_CursorKeys(This,prop) ) 

#define GeneralSettings_get_HighlightTrack(This,prop)	\
    ( (This)->lpVtbl -> get_HighlightTrack(This,prop) ) 

#define GeneralSettings_put_HighlightTrack(This,prop)	\
    ( (This)->lpVtbl -> put_HighlightTrack(This,prop) ) 

#define GeneralSettings_get_CharacterDictionary(This,prop)	\
    ( (This)->lpVtbl -> get_CharacterDictionary(This,prop) ) 

#define GeneralSettings_put_CharacterDictionary(This,prop)	\
    ( (This)->lpVtbl -> put_CharacterDictionary(This,prop) ) 

#define GeneralSettings_get_KeyLabelDictionary(This,prop)	\
    ( (This)->lpVtbl -> get_KeyLabelDictionary(This,prop) ) 

#define GeneralSettings_put_KeyLabelDictionary(This,prop)	\
    ( (This)->lpVtbl -> put_KeyLabelDictionary(This,prop) ) 

#define GeneralSettings_get_ExceptionDictionary(This,prop)	\
    ( (This)->lpVtbl -> get_ExceptionDictionary(This,prop) ) 

#define GeneralSettings_put_ExceptionDictionary(This,prop)	\
    ( (This)->lpVtbl -> put_ExceptionDictionary(This,prop) ) 

#define GeneralSettings_get_GraphicDictionary(This,prop)	\
    ( (This)->lpVtbl -> get_GraphicDictionary(This,prop) ) 

#define GeneralSettings_put_GraphicDictionary(This,prop)	\
    ( (This)->lpVtbl -> put_GraphicDictionary(This,prop) ) 

#define GeneralSettings_get_ColorDictionary(This,prop)	\
    ( (This)->lpVtbl -> get_ColorDictionary(This,prop) ) 

#define GeneralSettings_put_ColorDictionary(This,prop)	\
    ( (This)->lpVtbl -> put_ColorDictionary(This,prop) ) 

#define GeneralSettings_get_AllowSpeakWindowsInEditBoxes(This,prop)	\
    ( (This)->lpVtbl -> get_AllowSpeakWindowsInEditBoxes(This,prop) ) 

#define GeneralSettings_put_AllowSpeakWindowsInEditBoxes(This,prop)	\
    ( (This)->lpVtbl -> put_AllowSpeakWindowsInEditBoxes(This,prop) ) 

#define GeneralSettings_get_CursorDelay(This,prop)	\
    ( (This)->lpVtbl -> get_CursorDelay(This,prop) ) 

#define GeneralSettings_put_CursorDelay(This,prop)	\
    ( (This)->lpVtbl -> put_CursorDelay(This,prop) ) 

#define GeneralSettings_get_TriggerDelay(This,prop)	\
    ( (This)->lpVtbl -> get_TriggerDelay(This,prop) ) 

#define GeneralSettings_put_TriggerDelay(This,prop)	\
    ( (This)->lpVtbl -> put_TriggerDelay(This,prop) ) 

#define GeneralSettings_get_SpaceThreshold(This,prop)	\
    ( (This)->lpVtbl -> get_SpaceThreshold(This,prop) ) 

#define GeneralSettings_put_SpaceThreshold(This,prop)	\
    ( (This)->lpVtbl -> put_SpaceThreshold(This,prop) ) 

#define GeneralSettings_get_AllowAutoRedraw(This,prop)	\
    ( (This)->lpVtbl -> get_AllowAutoRedraw(This,prop) ) 

#define GeneralSettings_put_AllowAutoRedraw(This,prop)	\
    ( (This)->lpVtbl -> put_AllowAutoRedraw(This,prop) ) 

#define GeneralSettings_get_SpeakTooltips(This,prop)	\
    ( (This)->lpVtbl -> get_SpeakTooltips(This,prop) ) 

#define GeneralSettings_put_SpeakTooltips(This,prop)	\
    ( (This)->lpVtbl -> put_SpeakTooltips(This,prop) ) 

#define GeneralSettings_get_BrowseMode(This,prop)	\
    ( (This)->lpVtbl -> get_BrowseMode(This,prop) ) 

#define GeneralSettings_put_BrowseMode(This,prop)	\
    ( (This)->lpVtbl -> put_BrowseMode(This,prop) ) 

#define GeneralSettings_get_IncludeBoxInHighlight(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeBoxInHighlight(This,prop) ) 

#define GeneralSettings_put_IncludeBoxInHighlight(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeBoxInHighlight(This,prop) ) 

#define GeneralSettings_get_HighlightedText(This,prop)	\
    ( (This)->lpVtbl -> get_HighlightedText(This,prop) ) 

#define GeneralSettings_put_HighlightedText(This,prop)	\
    ( (This)->lpVtbl -> put_HighlightedText(This,prop) ) 

#define GeneralSettings_get_IncludeUserWindowWithBox(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeUserWindowWithBox(This,prop) ) 

#define GeneralSettings_put_IncludeUserWindowWithBox(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeUserWindowWithBox(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __GeneralSettings_INTERFACE_DEFINED__ */


#ifndef __Verbosity_INTERFACE_DEFINED__
#define __Verbosity_INTERFACE_DEFINED__

/* interface Verbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Verbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("9534C04C-344A-4E89-B5DD-00F6B34C7E8C")
    Verbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Menu( 
            /* [retval][out] */ MenuVerbosity **Menu) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Activated( 
            /* [retval][out] */ ActivatedVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Focused( 
            /* [retval][out] */ FocusedVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Attribute( 
            /* [retval][out] */ AttributeChangeVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Miscellaneous( 
            /* [retval][out] */ MiscVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BrowseMode( 
            /* [retval][out] */ BrowseModeVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Word( 
            /* [retval][out] */ WordVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Excel( 
            /* [retval][out] */ ExcelVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_PowerPoint( 
            /* [retval][out] */ PowerPointVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Mouse( 
            /* [retval][out] */ MouseVerbosity **Mouse) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct VerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Verbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Verbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Verbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Verbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Verbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Verbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Verbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Verbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Verbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Menu )( 
            Verbosity * This,
            /* [retval][out] */ MenuVerbosity **Menu);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Activated )( 
            Verbosity * This,
            /* [retval][out] */ ActivatedVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            Verbosity * This,
            /* [retval][out] */ FocusedVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Attribute )( 
            Verbosity * This,
            /* [retval][out] */ AttributeChangeVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Miscellaneous )( 
            Verbosity * This,
            /* [retval][out] */ MiscVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BrowseMode )( 
            Verbosity * This,
            /* [retval][out] */ BrowseModeVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Word )( 
            Verbosity * This,
            /* [retval][out] */ WordVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Excel )( 
            Verbosity * This,
            /* [retval][out] */ ExcelVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_PowerPoint )( 
            Verbosity * This,
            /* [retval][out] */ PowerPointVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Mouse )( 
            Verbosity * This,
            /* [retval][out] */ MouseVerbosity **Mouse);
        
        END_INTERFACE
    } VerbosityVtbl;

    interface Verbosity
    {
        CONST_VTBL struct VerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Verbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Verbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Verbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Verbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Verbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Verbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Verbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Verbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Verbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Verbosity_get_Menu(This,Menu)	\
    ( (This)->lpVtbl -> get_Menu(This,Menu) ) 

#define Verbosity_get_Activated(This,obj)	\
    ( (This)->lpVtbl -> get_Activated(This,obj) ) 

#define Verbosity_get_Focused(This,obj)	\
    ( (This)->lpVtbl -> get_Focused(This,obj) ) 

#define Verbosity_get_Attribute(This,obj)	\
    ( (This)->lpVtbl -> get_Attribute(This,obj) ) 

#define Verbosity_get_Miscellaneous(This,obj)	\
    ( (This)->lpVtbl -> get_Miscellaneous(This,obj) ) 

#define Verbosity_get_BrowseMode(This,obj)	\
    ( (This)->lpVtbl -> get_BrowseMode(This,obj) ) 

#define Verbosity_get_Word(This,obj)	\
    ( (This)->lpVtbl -> get_Word(This,obj) ) 

#define Verbosity_get_Excel(This,obj)	\
    ( (This)->lpVtbl -> get_Excel(This,obj) ) 

#define Verbosity_get_PowerPoint(This,obj)	\
    ( (This)->lpVtbl -> get_PowerPoint(This,obj) ) 

#define Verbosity_get_Mouse(This,Mouse)	\
    ( (This)->lpVtbl -> get_Mouse(This,Mouse) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Verbosity_INTERFACE_DEFINED__ */


#ifndef __MenuVerbosity_INTERFACE_DEFINED__
#define __MenuVerbosity_INTERFACE_DEFINED__

/* interface MenuVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_MenuVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("CDF3524C-57EF-440A-B55E-9F5379B233E9")
    MenuVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Opened( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Opened( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Closed( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Closed( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Option( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Option( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Accelerators( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Accelerators( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Descriptions( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Descriptions( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Pulldown( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Pulldown( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Shortcut( 
            /* [in] */ VerbosityShortcut prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Shortcut( 
            /* [retval][out] */ VerbosityShortcut *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct MenuVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            MenuVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            MenuVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            MenuVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            MenuVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            MenuVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            MenuVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            MenuVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            MenuVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            MenuVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Opened )( 
            MenuVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Opened )( 
            MenuVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Closed )( 
            MenuVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Closed )( 
            MenuVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Option )( 
            MenuVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Option )( 
            MenuVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Accelerators )( 
            MenuVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Accelerators )( 
            MenuVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Descriptions )( 
            MenuVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Descriptions )( 
            MenuVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Pulldown )( 
            MenuVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Pulldown )( 
            MenuVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Shortcut )( 
            MenuVerbosity * This,
            /* [in] */ VerbosityShortcut prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Shortcut )( 
            MenuVerbosity * This,
            /* [retval][out] */ VerbosityShortcut *prop);
        
        END_INTERFACE
    } MenuVerbosityVtbl;

    interface MenuVerbosity
    {
        CONST_VTBL struct MenuVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define MenuVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define MenuVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define MenuVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define MenuVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define MenuVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define MenuVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define MenuVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define MenuVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define MenuVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define MenuVerbosity_put_Opened(This,prop)	\
    ( (This)->lpVtbl -> put_Opened(This,prop) ) 

#define MenuVerbosity_get_Opened(This,prop)	\
    ( (This)->lpVtbl -> get_Opened(This,prop) ) 

#define MenuVerbosity_put_Closed(This,prop)	\
    ( (This)->lpVtbl -> put_Closed(This,prop) ) 

#define MenuVerbosity_get_Closed(This,prop)	\
    ( (This)->lpVtbl -> get_Closed(This,prop) ) 

#define MenuVerbosity_put_Option(This,prop)	\
    ( (This)->lpVtbl -> put_Option(This,prop) ) 

#define MenuVerbosity_get_Option(This,prop)	\
    ( (This)->lpVtbl -> get_Option(This,prop) ) 

#define MenuVerbosity_put_Accelerators(This,prop)	\
    ( (This)->lpVtbl -> put_Accelerators(This,prop) ) 

#define MenuVerbosity_get_Accelerators(This,prop)	\
    ( (This)->lpVtbl -> get_Accelerators(This,prop) ) 

#define MenuVerbosity_put_Descriptions(This,prop)	\
    ( (This)->lpVtbl -> put_Descriptions(This,prop) ) 

#define MenuVerbosity_get_Descriptions(This,prop)	\
    ( (This)->lpVtbl -> get_Descriptions(This,prop) ) 

#define MenuVerbosity_put_Pulldown(This,prop)	\
    ( (This)->lpVtbl -> put_Pulldown(This,prop) ) 

#define MenuVerbosity_get_Pulldown(This,prop)	\
    ( (This)->lpVtbl -> get_Pulldown(This,prop) ) 

#define MenuVerbosity_put_Shortcut(This,prop)	\
    ( (This)->lpVtbl -> put_Shortcut(This,prop) ) 

#define MenuVerbosity_get_Shortcut(This,prop)	\
    ( (This)->lpVtbl -> get_Shortcut(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __MenuVerbosity_INTERFACE_DEFINED__ */


#ifndef __ActivatedVerbosity_INTERFACE_DEFINED__
#define __ActivatedVerbosity_INTERFACE_DEFINED__

/* interface ActivatedVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ActivatedVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("163FA9DA-59C4-471C-9E93-907D4489090B")
    ActivatedVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Activated( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Activated( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_WindowType( 
            /* [in] */ WindowTypeStatus prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WindowType( 
            /* [retval][out] */ WindowTypeStatus *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_WindowTitle( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WindowTitle( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EntireMessageBox( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EntireMessageBox( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EntireDialog( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EntireDialog( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ActivatedVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ActivatedVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ActivatedVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ActivatedVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ActivatedVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ActivatedVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ActivatedVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ActivatedVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ActivatedVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ActivatedVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Activated )( 
            ActivatedVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Activated )( 
            ActivatedVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_WindowType )( 
            ActivatedVerbosity * This,
            /* [in] */ WindowTypeStatus prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WindowType )( 
            ActivatedVerbosity * This,
            /* [retval][out] */ WindowTypeStatus *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_WindowTitle )( 
            ActivatedVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WindowTitle )( 
            ActivatedVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EntireMessageBox )( 
            ActivatedVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EntireMessageBox )( 
            ActivatedVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EntireDialog )( 
            ActivatedVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EntireDialog )( 
            ActivatedVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } ActivatedVerbosityVtbl;

    interface ActivatedVerbosity
    {
        CONST_VTBL struct ActivatedVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ActivatedVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ActivatedVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ActivatedVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ActivatedVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ActivatedVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ActivatedVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ActivatedVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ActivatedVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ActivatedVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ActivatedVerbosity_put_Activated(This,prop)	\
    ( (This)->lpVtbl -> put_Activated(This,prop) ) 

#define ActivatedVerbosity_get_Activated(This,prop)	\
    ( (This)->lpVtbl -> get_Activated(This,prop) ) 

#define ActivatedVerbosity_put_WindowType(This,prop)	\
    ( (This)->lpVtbl -> put_WindowType(This,prop) ) 

#define ActivatedVerbosity_get_WindowType(This,prop)	\
    ( (This)->lpVtbl -> get_WindowType(This,prop) ) 

#define ActivatedVerbosity_put_WindowTitle(This,prop)	\
    ( (This)->lpVtbl -> put_WindowTitle(This,prop) ) 

#define ActivatedVerbosity_get_WindowTitle(This,prop)	\
    ( (This)->lpVtbl -> get_WindowTitle(This,prop) ) 

#define ActivatedVerbosity_put_EntireMessageBox(This,prop)	\
    ( (This)->lpVtbl -> put_EntireMessageBox(This,prop) ) 

#define ActivatedVerbosity_get_EntireMessageBox(This,prop)	\
    ( (This)->lpVtbl -> get_EntireMessageBox(This,prop) ) 

#define ActivatedVerbosity_put_EntireDialog(This,prop)	\
    ( (This)->lpVtbl -> put_EntireDialog(This,prop) ) 

#define ActivatedVerbosity_get_EntireDialog(This,prop)	\
    ( (This)->lpVtbl -> get_EntireDialog(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ActivatedVerbosity_INTERFACE_DEFINED__ */


#ifndef __FocusedVerbosity_INTERFACE_DEFINED__
#define __FocusedVerbosity_INTERFACE_DEFINED__

/* interface FocusedVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_FocusedVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("FD82F541-E59D-4092-8A52-50E8289CCF9B")
    FocusedVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Focused( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Focused( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_WindowType( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WindowType( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_WindowTypeFirst( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WindowTypeFirst( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FieldName( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FieldName( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FieldData( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FieldData( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CustomControl( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CustomControl( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IdentifyMSAAControls( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IdentifyMSAAControls( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateLink( 
            /* [in] */ BrowseModeLinkIndication prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateLink( 
            /* [retval][out] */ BrowseModeLinkIndication *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct FocusedVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            FocusedVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            FocusedVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            FocusedVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            FocusedVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            FocusedVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            FocusedVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            FocusedVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            FocusedVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            FocusedVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Focused )( 
            FocusedVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            FocusedVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_WindowType )( 
            FocusedVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WindowType )( 
            FocusedVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_WindowTypeFirst )( 
            FocusedVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WindowTypeFirst )( 
            FocusedVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FieldName )( 
            FocusedVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FieldName )( 
            FocusedVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FieldData )( 
            FocusedVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FieldData )( 
            FocusedVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CustomControl )( 
            FocusedVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CustomControl )( 
            FocusedVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IdentifyMSAAControls )( 
            FocusedVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IdentifyMSAAControls )( 
            FocusedVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateLink )( 
            FocusedVerbosity * This,
            /* [in] */ BrowseModeLinkIndication prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateLink )( 
            FocusedVerbosity * This,
            /* [retval][out] */ BrowseModeLinkIndication *prop);
        
        END_INTERFACE
    } FocusedVerbosityVtbl;

    interface FocusedVerbosity
    {
        CONST_VTBL struct FocusedVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define FocusedVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define FocusedVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define FocusedVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define FocusedVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define FocusedVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define FocusedVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define FocusedVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define FocusedVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define FocusedVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define FocusedVerbosity_put_Focused(This,prop)	\
    ( (This)->lpVtbl -> put_Focused(This,prop) ) 

#define FocusedVerbosity_get_Focused(This,prop)	\
    ( (This)->lpVtbl -> get_Focused(This,prop) ) 

#define FocusedVerbosity_put_WindowType(This,prop)	\
    ( (This)->lpVtbl -> put_WindowType(This,prop) ) 

#define FocusedVerbosity_get_WindowType(This,prop)	\
    ( (This)->lpVtbl -> get_WindowType(This,prop) ) 

#define FocusedVerbosity_put_WindowTypeFirst(This,prop)	\
    ( (This)->lpVtbl -> put_WindowTypeFirst(This,prop) ) 

#define FocusedVerbosity_get_WindowTypeFirst(This,prop)	\
    ( (This)->lpVtbl -> get_WindowTypeFirst(This,prop) ) 

#define FocusedVerbosity_put_FieldName(This,prop)	\
    ( (This)->lpVtbl -> put_FieldName(This,prop) ) 

#define FocusedVerbosity_get_FieldName(This,prop)	\
    ( (This)->lpVtbl -> get_FieldName(This,prop) ) 

#define FocusedVerbosity_put_FieldData(This,prop)	\
    ( (This)->lpVtbl -> put_FieldData(This,prop) ) 

#define FocusedVerbosity_get_FieldData(This,prop)	\
    ( (This)->lpVtbl -> get_FieldData(This,prop) ) 

#define FocusedVerbosity_put_CustomControl(This,prop)	\
    ( (This)->lpVtbl -> put_CustomControl(This,prop) ) 

#define FocusedVerbosity_get_CustomControl(This,prop)	\
    ( (This)->lpVtbl -> get_CustomControl(This,prop) ) 

#define FocusedVerbosity_put_IdentifyMSAAControls(This,prop)	\
    ( (This)->lpVtbl -> put_IdentifyMSAAControls(This,prop) ) 

#define FocusedVerbosity_get_IdentifyMSAAControls(This,prop)	\
    ( (This)->lpVtbl -> get_IdentifyMSAAControls(This,prop) ) 

#define FocusedVerbosity_put_IndicateLink(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateLink(This,prop) ) 

#define FocusedVerbosity_get_IndicateLink(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateLink(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __FocusedVerbosity_INTERFACE_DEFINED__ */


#ifndef __AttributeChangeVerbosity_INTERFACE_DEFINED__
#define __AttributeChangeVerbosity_INTERFACE_DEFINED__

/* interface AttributeChangeVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_AttributeChangeVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("B55EFF9A-8946-4ED2-8C60-3EC69A854117")
    AttributeChangeVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Color( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Color( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Font( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Font( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_size( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_size( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Style( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Style( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct AttributeChangeVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            AttributeChangeVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            AttributeChangeVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            AttributeChangeVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            AttributeChangeVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            AttributeChangeVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            AttributeChangeVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            AttributeChangeVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            AttributeChangeVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            AttributeChangeVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Color )( 
            AttributeChangeVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Color )( 
            AttributeChangeVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Font )( 
            AttributeChangeVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Font )( 
            AttributeChangeVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_size )( 
            AttributeChangeVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_size )( 
            AttributeChangeVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Style )( 
            AttributeChangeVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Style )( 
            AttributeChangeVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } AttributeChangeVerbosityVtbl;

    interface AttributeChangeVerbosity
    {
        CONST_VTBL struct AttributeChangeVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define AttributeChangeVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define AttributeChangeVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define AttributeChangeVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define AttributeChangeVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define AttributeChangeVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define AttributeChangeVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define AttributeChangeVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define AttributeChangeVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define AttributeChangeVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define AttributeChangeVerbosity_put_Color(This,prop)	\
    ( (This)->lpVtbl -> put_Color(This,prop) ) 

#define AttributeChangeVerbosity_get_Color(This,prop)	\
    ( (This)->lpVtbl -> get_Color(This,prop) ) 

#define AttributeChangeVerbosity_put_Font(This,prop)	\
    ( (This)->lpVtbl -> put_Font(This,prop) ) 

#define AttributeChangeVerbosity_get_Font(This,prop)	\
    ( (This)->lpVtbl -> get_Font(This,prop) ) 

#define AttributeChangeVerbosity_put_size(This,prop)	\
    ( (This)->lpVtbl -> put_size(This,prop) ) 

#define AttributeChangeVerbosity_get_size(This,prop)	\
    ( (This)->lpVtbl -> get_size(This,prop) ) 

#define AttributeChangeVerbosity_put_Style(This,prop)	\
    ( (This)->lpVtbl -> put_Style(This,prop) ) 

#define AttributeChangeVerbosity_get_Style(This,prop)	\
    ( (This)->lpVtbl -> get_Style(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __AttributeChangeVerbosity_INTERFACE_DEFINED__ */


#ifndef __MiscVerbosity_INTERFACE_DEFINED__
#define __MiscVerbosity_INTERFACE_DEFINED__

/* interface MiscVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_MiscVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("4201F55C-A4F1-40E7-878D-9E7B97BF8EBF")
    MiscVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Shortcut( 
            /* [in] */ VerbosityShortcut prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Shortcut( 
            /* [retval][out] */ VerbosityShortcut *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TopBottomIndex( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TopBottomIndex( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CurrentIndex( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CurrentIndex( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SilencePasswordEdits( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SilencePasswordEdits( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ListviewHeaders( 
            /* [in] */ VerbosityListviewHeaders prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ListviewHeaders( 
            /* [retval][out] */ VerbosityListviewHeaders *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct MiscVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            MiscVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            MiscVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            MiscVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            MiscVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            MiscVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            MiscVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            MiscVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            MiscVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            MiscVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Shortcut )( 
            MiscVerbosity * This,
            /* [in] */ VerbosityShortcut prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Shortcut )( 
            MiscVerbosity * This,
            /* [retval][out] */ VerbosityShortcut *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TopBottomIndex )( 
            MiscVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TopBottomIndex )( 
            MiscVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CurrentIndex )( 
            MiscVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CurrentIndex )( 
            MiscVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SilencePasswordEdits )( 
            MiscVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SilencePasswordEdits )( 
            MiscVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ListviewHeaders )( 
            MiscVerbosity * This,
            /* [in] */ VerbosityListviewHeaders prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ListviewHeaders )( 
            MiscVerbosity * This,
            /* [retval][out] */ VerbosityListviewHeaders *prop);
        
        END_INTERFACE
    } MiscVerbosityVtbl;

    interface MiscVerbosity
    {
        CONST_VTBL struct MiscVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define MiscVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define MiscVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define MiscVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define MiscVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define MiscVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define MiscVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define MiscVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define MiscVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define MiscVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define MiscVerbosity_put_Shortcut(This,prop)	\
    ( (This)->lpVtbl -> put_Shortcut(This,prop) ) 

#define MiscVerbosity_get_Shortcut(This,prop)	\
    ( (This)->lpVtbl -> get_Shortcut(This,prop) ) 

#define MiscVerbosity_put_TopBottomIndex(This,prop)	\
    ( (This)->lpVtbl -> put_TopBottomIndex(This,prop) ) 

#define MiscVerbosity_get_TopBottomIndex(This,prop)	\
    ( (This)->lpVtbl -> get_TopBottomIndex(This,prop) ) 

#define MiscVerbosity_put_CurrentIndex(This,prop)	\
    ( (This)->lpVtbl -> put_CurrentIndex(This,prop) ) 

#define MiscVerbosity_get_CurrentIndex(This,prop)	\
    ( (This)->lpVtbl -> get_CurrentIndex(This,prop) ) 

#define MiscVerbosity_put_SilencePasswordEdits(This,prop)	\
    ( (This)->lpVtbl -> put_SilencePasswordEdits(This,prop) ) 

#define MiscVerbosity_get_SilencePasswordEdits(This,prop)	\
    ( (This)->lpVtbl -> get_SilencePasswordEdits(This,prop) ) 

#define MiscVerbosity_put_ListviewHeaders(This,prop)	\
    ( (This)->lpVtbl -> put_ListviewHeaders(This,prop) ) 

#define MiscVerbosity_get_ListviewHeaders(This,prop)	\
    ( (This)->lpVtbl -> get_ListviewHeaders(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __MiscVerbosity_INTERFACE_DEFINED__ */


#ifndef __BrowseModeVerbosity_INTERFACE_DEFINED__
#define __BrowseModeVerbosity_INTERFACE_DEFINED__

/* interface BrowseModeVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("6061A38D-DD88-484B-9D72-372AD1BE73A8")
    BrowseModeVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AccessKeys( 
            /* [retval][out] */ BrowseModeAccessKeyVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AcronymsAbbreviations( 
            /* [retval][out] */ BrowseModeAcronymVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Flash( 
            /* [retval][out] */ BrowseModeFlashVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Forms( 
            /* [retval][out] */ BrowseModeFormVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Headings( 
            /* [retval][out] */ BrowseModeHeadingVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Languages( 
            /* [retval][out] */ BrowseModeLanguageVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Links( 
            /* [retval][out] */ BrowseModeLinkVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Lists( 
            /* [retval][out] */ BrowseModeListVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LongDesc( 
            /* [retval][out] */ BrowseModeLongDescVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Miscellaneous( 
            /* [retval][out] */ BrowseModeMiscVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NewPageInformation( 
            /* [retval][out] */ BrowseModeNewPageVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Objects( 
            /* [retval][out] */ BrowseModeObjectVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Paragraphs( 
            /* [retval][out] */ BrowseModeParagraphVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_QuotesBlockQuotes( 
            /* [retval][out] */ BrowseModeQuoteVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Tables( 
            /* [retval][out] */ BrowseModeTableVerbosity **obj) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AccessKeys )( 
            BrowseModeVerbosity * This,
            /* [retval][out] */ BrowseModeAccessKeyVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AcronymsAbbreviations )( 
            BrowseModeVerbosity * This,
            /* [retval][out] */ BrowseModeAcronymVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Flash )( 
            BrowseModeVerbosity * This,
            /* [retval][out] */ BrowseModeFlashVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Forms )( 
            BrowseModeVerbosity * This,
            /* [retval][out] */ BrowseModeFormVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Headings )( 
            BrowseModeVerbosity * This,
            /* [retval][out] */ BrowseModeHeadingVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Languages )( 
            BrowseModeVerbosity * This,
            /* [retval][out] */ BrowseModeLanguageVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Links )( 
            BrowseModeVerbosity * This,
            /* [retval][out] */ BrowseModeLinkVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Lists )( 
            BrowseModeVerbosity * This,
            /* [retval][out] */ BrowseModeListVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LongDesc )( 
            BrowseModeVerbosity * This,
            /* [retval][out] */ BrowseModeLongDescVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Miscellaneous )( 
            BrowseModeVerbosity * This,
            /* [retval][out] */ BrowseModeMiscVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NewPageInformation )( 
            BrowseModeVerbosity * This,
            /* [retval][out] */ BrowseModeNewPageVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Objects )( 
            BrowseModeVerbosity * This,
            /* [retval][out] */ BrowseModeObjectVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Paragraphs )( 
            BrowseModeVerbosity * This,
            /* [retval][out] */ BrowseModeParagraphVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_QuotesBlockQuotes )( 
            BrowseModeVerbosity * This,
            /* [retval][out] */ BrowseModeQuoteVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Tables )( 
            BrowseModeVerbosity * This,
            /* [retval][out] */ BrowseModeTableVerbosity **obj);
        
        END_INTERFACE
    } BrowseModeVerbosityVtbl;

    interface BrowseModeVerbosity
    {
        CONST_VTBL struct BrowseModeVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeVerbosity_get_AccessKeys(This,obj)	\
    ( (This)->lpVtbl -> get_AccessKeys(This,obj) ) 

#define BrowseModeVerbosity_get_AcronymsAbbreviations(This,obj)	\
    ( (This)->lpVtbl -> get_AcronymsAbbreviations(This,obj) ) 

#define BrowseModeVerbosity_get_Flash(This,obj)	\
    ( (This)->lpVtbl -> get_Flash(This,obj) ) 

#define BrowseModeVerbosity_get_Forms(This,obj)	\
    ( (This)->lpVtbl -> get_Forms(This,obj) ) 

#define BrowseModeVerbosity_get_Headings(This,obj)	\
    ( (This)->lpVtbl -> get_Headings(This,obj) ) 

#define BrowseModeVerbosity_get_Languages(This,obj)	\
    ( (This)->lpVtbl -> get_Languages(This,obj) ) 

#define BrowseModeVerbosity_get_Links(This,obj)	\
    ( (This)->lpVtbl -> get_Links(This,obj) ) 

#define BrowseModeVerbosity_get_Lists(This,obj)	\
    ( (This)->lpVtbl -> get_Lists(This,obj) ) 

#define BrowseModeVerbosity_get_LongDesc(This,obj)	\
    ( (This)->lpVtbl -> get_LongDesc(This,obj) ) 

#define BrowseModeVerbosity_get_Miscellaneous(This,obj)	\
    ( (This)->lpVtbl -> get_Miscellaneous(This,obj) ) 

#define BrowseModeVerbosity_get_NewPageInformation(This,obj)	\
    ( (This)->lpVtbl -> get_NewPageInformation(This,obj) ) 

#define BrowseModeVerbosity_get_Objects(This,obj)	\
    ( (This)->lpVtbl -> get_Objects(This,obj) ) 

#define BrowseModeVerbosity_get_Paragraphs(This,obj)	\
    ( (This)->lpVtbl -> get_Paragraphs(This,obj) ) 

#define BrowseModeVerbosity_get_QuotesBlockQuotes(This,obj)	\
    ( (This)->lpVtbl -> get_QuotesBlockQuotes(This,obj) ) 

#define BrowseModeVerbosity_get_Tables(This,obj)	\
    ( (This)->lpVtbl -> get_Tables(This,obj) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeVerbosity_INTERFACE_DEFINED__ */


#ifndef __BrowseModeAccessKeyVerbosity_INTERFACE_DEFINED__
#define __BrowseModeAccessKeyVerbosity_INTERFACE_DEFINED__

/* interface BrowseModeAccessKeyVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeAccessKeyVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("70AFC45C-6452-46A2-A921-5634B2610AC4")
    BrowseModeAccessKeyVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indicate( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indicate( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeAccessKeyVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeAccessKeyVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeAccessKeyVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeAccessKeyVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeAccessKeyVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeAccessKeyVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeAccessKeyVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeAccessKeyVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeAccessKeyVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeAccessKeyVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indicate )( 
            BrowseModeAccessKeyVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indicate )( 
            BrowseModeAccessKeyVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeAccessKeyVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeAccessKeyVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } BrowseModeAccessKeyVerbosityVtbl;

    interface BrowseModeAccessKeyVerbosity
    {
        CONST_VTBL struct BrowseModeAccessKeyVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeAccessKeyVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeAccessKeyVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeAccessKeyVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeAccessKeyVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeAccessKeyVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeAccessKeyVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeAccessKeyVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeAccessKeyVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeAccessKeyVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeAccessKeyVerbosity_put_Indicate(This,prop)	\
    ( (This)->lpVtbl -> put_Indicate(This,prop) ) 

#define BrowseModeAccessKeyVerbosity_get_Indicate(This,prop)	\
    ( (This)->lpVtbl -> get_Indicate(This,prop) ) 

#define BrowseModeAccessKeyVerbosity_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeAccessKeyVerbosity_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeAccessKeyVerbosity_INTERFACE_DEFINED__ */


#ifndef __BrowseModeAcronymVerbosity_INTERFACE_DEFINED__
#define __BrowseModeAcronymVerbosity_INTERFACE_DEFINED__

/* interface BrowseModeAcronymVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeAcronymVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("7200CD64-3E31-4758-8F1E-69AC82EFADCC")
    BrowseModeAcronymVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indication( 
            /* [in] */ BrowseModeAcronymIndication prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indication( 
            /* [retval][out] */ BrowseModeAcronymIndication *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeAcronymVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeAcronymVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeAcronymVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeAcronymVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeAcronymVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeAcronymVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeAcronymVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeAcronymVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeAcronymVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeAcronymVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeAcronymVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeAcronymVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indication )( 
            BrowseModeAcronymVerbosity * This,
            /* [in] */ BrowseModeAcronymIndication prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indication )( 
            BrowseModeAcronymVerbosity * This,
            /* [retval][out] */ BrowseModeAcronymIndication *prop);
        
        END_INTERFACE
    } BrowseModeAcronymVerbosityVtbl;

    interface BrowseModeAcronymVerbosity
    {
        CONST_VTBL struct BrowseModeAcronymVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeAcronymVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeAcronymVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeAcronymVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeAcronymVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeAcronymVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeAcronymVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeAcronymVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeAcronymVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeAcronymVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeAcronymVerbosity_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeAcronymVerbosity_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#define BrowseModeAcronymVerbosity_put_Indication(This,prop)	\
    ( (This)->lpVtbl -> put_Indication(This,prop) ) 

#define BrowseModeAcronymVerbosity_get_Indication(This,prop)	\
    ( (This)->lpVtbl -> get_Indication(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeAcronymVerbosity_INTERFACE_DEFINED__ */


#ifndef __BrowseModeFlashVerbosity_INTERFACE_DEFINED__
#define __BrowseModeFlashVerbosity_INTERFACE_DEFINED__

/* interface BrowseModeFlashVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeFlashVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("A387C9D1-D42E-4976-ACA0-D1720C3D3A79")
    BrowseModeFlashVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indicate( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indicate( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeFlashVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeFlashVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeFlashVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeFlashVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeFlashVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeFlashVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeFlashVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeFlashVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeFlashVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeFlashVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indicate )( 
            BrowseModeFlashVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indicate )( 
            BrowseModeFlashVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeFlashVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeFlashVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } BrowseModeFlashVerbosityVtbl;

    interface BrowseModeFlashVerbosity
    {
        CONST_VTBL struct BrowseModeFlashVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeFlashVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeFlashVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeFlashVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeFlashVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeFlashVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeFlashVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeFlashVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeFlashVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeFlashVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeFlashVerbosity_put_Indicate(This,prop)	\
    ( (This)->lpVtbl -> put_Indicate(This,prop) ) 

#define BrowseModeFlashVerbosity_get_Indicate(This,prop)	\
    ( (This)->lpVtbl -> get_Indicate(This,prop) ) 

#define BrowseModeFlashVerbosity_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeFlashVerbosity_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeFlashVerbosity_INTERFACE_DEFINED__ */


#ifndef __BrowseModeFormVerbosity_INTERFACE_DEFINED__
#define __BrowseModeFormVerbosity_INTERFACE_DEFINED__

/* interface BrowseModeFormVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeFormVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("9420D69A-3034-4284-9BAA-96B90E75B405")
    BrowseModeFormVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateForms( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateForms( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEndForms( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEndForms( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateFieldsets( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateFieldsets( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEndFieldsets( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEndFieldsets( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeFormVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeFormVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeFormVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeFormVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeFormVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeFormVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeFormVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeFormVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeFormVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeFormVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateForms )( 
            BrowseModeFormVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateForms )( 
            BrowseModeFormVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEndForms )( 
            BrowseModeFormVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEndForms )( 
            BrowseModeFormVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateFieldsets )( 
            BrowseModeFormVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateFieldsets )( 
            BrowseModeFormVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEndFieldsets )( 
            BrowseModeFormVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEndFieldsets )( 
            BrowseModeFormVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } BrowseModeFormVerbosityVtbl;

    interface BrowseModeFormVerbosity
    {
        CONST_VTBL struct BrowseModeFormVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeFormVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeFormVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeFormVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeFormVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeFormVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeFormVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeFormVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeFormVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeFormVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeFormVerbosity_put_IndicateForms(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateForms(This,prop) ) 

#define BrowseModeFormVerbosity_get_IndicateForms(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateForms(This,prop) ) 

#define BrowseModeFormVerbosity_put_ReadToEndForms(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEndForms(This,prop) ) 

#define BrowseModeFormVerbosity_get_ReadToEndForms(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEndForms(This,prop) ) 

#define BrowseModeFormVerbosity_put_IndicateFieldsets(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateFieldsets(This,prop) ) 

#define BrowseModeFormVerbosity_get_IndicateFieldsets(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateFieldsets(This,prop) ) 

#define BrowseModeFormVerbosity_put_ReadToEndFieldsets(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEndFieldsets(This,prop) ) 

#define BrowseModeFormVerbosity_get_ReadToEndFieldsets(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEndFieldsets(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeFormVerbosity_INTERFACE_DEFINED__ */


#ifndef __BrowseModeHeadingVerbosity_INTERFACE_DEFINED__
#define __BrowseModeHeadingVerbosity_INTERFACE_DEFINED__

/* interface BrowseModeHeadingVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeHeadingVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("ED926AC6-D32D-44EC-8E2C-9D9DC33584CB")
    BrowseModeHeadingVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indicate( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indicate( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeHeadingVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeHeadingVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeHeadingVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeHeadingVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeHeadingVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeHeadingVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeHeadingVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeHeadingVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeHeadingVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeHeadingVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indicate )( 
            BrowseModeHeadingVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indicate )( 
            BrowseModeHeadingVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeHeadingVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeHeadingVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } BrowseModeHeadingVerbosityVtbl;

    interface BrowseModeHeadingVerbosity
    {
        CONST_VTBL struct BrowseModeHeadingVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeHeadingVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeHeadingVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeHeadingVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeHeadingVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeHeadingVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeHeadingVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeHeadingVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeHeadingVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeHeadingVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeHeadingVerbosity_put_Indicate(This,prop)	\
    ( (This)->lpVtbl -> put_Indicate(This,prop) ) 

#define BrowseModeHeadingVerbosity_get_Indicate(This,prop)	\
    ( (This)->lpVtbl -> get_Indicate(This,prop) ) 

#define BrowseModeHeadingVerbosity_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeHeadingVerbosity_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeHeadingVerbosity_INTERFACE_DEFINED__ */


#ifndef __BrowseModeLanguageVerbosity_INTERFACE_DEFINED__
#define __BrowseModeLanguageVerbosity_INTERFACE_DEFINED__

/* interface BrowseModeLanguageVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeLanguageVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("BBC0FAF7-050E-44A3-A1D1-1F107F857E1D")
    BrowseModeLanguageVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indicate( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indicate( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AutoSwitchSynthesizer( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AutoSwitchSynthesizer( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeLanguageVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeLanguageVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeLanguageVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeLanguageVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeLanguageVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeLanguageVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeLanguageVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeLanguageVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeLanguageVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeLanguageVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indicate )( 
            BrowseModeLanguageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indicate )( 
            BrowseModeLanguageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeLanguageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeLanguageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AutoSwitchSynthesizer )( 
            BrowseModeLanguageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AutoSwitchSynthesizer )( 
            BrowseModeLanguageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } BrowseModeLanguageVerbosityVtbl;

    interface BrowseModeLanguageVerbosity
    {
        CONST_VTBL struct BrowseModeLanguageVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeLanguageVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeLanguageVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeLanguageVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeLanguageVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeLanguageVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeLanguageVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeLanguageVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeLanguageVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeLanguageVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeLanguageVerbosity_put_Indicate(This,prop)	\
    ( (This)->lpVtbl -> put_Indicate(This,prop) ) 

#define BrowseModeLanguageVerbosity_get_Indicate(This,prop)	\
    ( (This)->lpVtbl -> get_Indicate(This,prop) ) 

#define BrowseModeLanguageVerbosity_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeLanguageVerbosity_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#define BrowseModeLanguageVerbosity_put_AutoSwitchSynthesizer(This,prop)	\
    ( (This)->lpVtbl -> put_AutoSwitchSynthesizer(This,prop) ) 

#define BrowseModeLanguageVerbosity_get_AutoSwitchSynthesizer(This,prop)	\
    ( (This)->lpVtbl -> get_AutoSwitchSynthesizer(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeLanguageVerbosity_INTERFACE_DEFINED__ */


#ifndef __BrowseModeLinkVerbosity_INTERFACE_DEFINED__
#define __BrowseModeLinkVerbosity_INTERFACE_DEFINED__

/* interface BrowseModeLinkVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeLinkVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("46B44ECA-2291-48DC-B79E-676CB12B5456")
    BrowseModeLinkVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IdentifyVisitedLinks( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IdentifyVisitedLinks( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IdentifySamePageLinks( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IdentifySamePageLinks( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indication( 
            /* [in] */ BrowseModeLinkIndication prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indication( 
            /* [retval][out] */ BrowseModeLinkIndication *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeLinkVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeLinkVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeLinkVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeLinkVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeLinkVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeLinkVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeLinkVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeLinkVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeLinkVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeLinkVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeLinkVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeLinkVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IdentifyVisitedLinks )( 
            BrowseModeLinkVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IdentifyVisitedLinks )( 
            BrowseModeLinkVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IdentifySamePageLinks )( 
            BrowseModeLinkVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IdentifySamePageLinks )( 
            BrowseModeLinkVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indication )( 
            BrowseModeLinkVerbosity * This,
            /* [in] */ BrowseModeLinkIndication prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indication )( 
            BrowseModeLinkVerbosity * This,
            /* [retval][out] */ BrowseModeLinkIndication *prop);
        
        END_INTERFACE
    } BrowseModeLinkVerbosityVtbl;

    interface BrowseModeLinkVerbosity
    {
        CONST_VTBL struct BrowseModeLinkVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeLinkVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeLinkVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeLinkVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeLinkVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeLinkVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeLinkVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeLinkVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeLinkVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeLinkVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeLinkVerbosity_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeLinkVerbosity_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#define BrowseModeLinkVerbosity_put_IdentifyVisitedLinks(This,prop)	\
    ( (This)->lpVtbl -> put_IdentifyVisitedLinks(This,prop) ) 

#define BrowseModeLinkVerbosity_get_IdentifyVisitedLinks(This,prop)	\
    ( (This)->lpVtbl -> get_IdentifyVisitedLinks(This,prop) ) 

#define BrowseModeLinkVerbosity_put_IdentifySamePageLinks(This,prop)	\
    ( (This)->lpVtbl -> put_IdentifySamePageLinks(This,prop) ) 

#define BrowseModeLinkVerbosity_get_IdentifySamePageLinks(This,prop)	\
    ( (This)->lpVtbl -> get_IdentifySamePageLinks(This,prop) ) 

#define BrowseModeLinkVerbosity_put_Indication(This,prop)	\
    ( (This)->lpVtbl -> put_Indication(This,prop) ) 

#define BrowseModeLinkVerbosity_get_Indication(This,prop)	\
    ( (This)->lpVtbl -> get_Indication(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeLinkVerbosity_INTERFACE_DEFINED__ */


#ifndef __BrowseModeListVerbosity_INTERFACE_DEFINED__
#define __BrowseModeListVerbosity_INTERFACE_DEFINED__

/* interface BrowseModeListVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeListVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("A9D9C347-ED99-4E1D-8DD6-77F86FCA48A3")
    BrowseModeListVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indicate( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indicate( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateBullets( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateBullets( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEndBullets( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEndBullets( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeListVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeListVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeListVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeListVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeListVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeListVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeListVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeListVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeListVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeListVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indicate )( 
            BrowseModeListVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indicate )( 
            BrowseModeListVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeListVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeListVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateBullets )( 
            BrowseModeListVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateBullets )( 
            BrowseModeListVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEndBullets )( 
            BrowseModeListVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEndBullets )( 
            BrowseModeListVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } BrowseModeListVerbosityVtbl;

    interface BrowseModeListVerbosity
    {
        CONST_VTBL struct BrowseModeListVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeListVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeListVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeListVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeListVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeListVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeListVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeListVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeListVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeListVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeListVerbosity_put_Indicate(This,prop)	\
    ( (This)->lpVtbl -> put_Indicate(This,prop) ) 

#define BrowseModeListVerbosity_get_Indicate(This,prop)	\
    ( (This)->lpVtbl -> get_Indicate(This,prop) ) 

#define BrowseModeListVerbosity_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeListVerbosity_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#define BrowseModeListVerbosity_put_IndicateBullets(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateBullets(This,prop) ) 

#define BrowseModeListVerbosity_get_IndicateBullets(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateBullets(This,prop) ) 

#define BrowseModeListVerbosity_put_ReadToEndBullets(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEndBullets(This,prop) ) 

#define BrowseModeListVerbosity_get_ReadToEndBullets(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEndBullets(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeListVerbosity_INTERFACE_DEFINED__ */


#ifndef __BrowseModeLongDescVerbosity_INTERFACE_DEFINED__
#define __BrowseModeLongDescVerbosity_INTERFACE_DEFINED__

/* interface BrowseModeLongDescVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeLongDescVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("B53306E3-72BB-48C3-87F2-DB1FB5EF1DD8")
    BrowseModeLongDescVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indicate( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indicate( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeLongDescVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeLongDescVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeLongDescVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeLongDescVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeLongDescVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeLongDescVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeLongDescVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeLongDescVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeLongDescVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeLongDescVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indicate )( 
            BrowseModeLongDescVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indicate )( 
            BrowseModeLongDescVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeLongDescVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeLongDescVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } BrowseModeLongDescVerbosityVtbl;

    interface BrowseModeLongDescVerbosity
    {
        CONST_VTBL struct BrowseModeLongDescVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeLongDescVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeLongDescVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeLongDescVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeLongDescVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeLongDescVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeLongDescVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeLongDescVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeLongDescVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeLongDescVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeLongDescVerbosity_put_Indicate(This,prop)	\
    ( (This)->lpVtbl -> put_Indicate(This,prop) ) 

#define BrowseModeLongDescVerbosity_get_Indicate(This,prop)	\
    ( (This)->lpVtbl -> get_Indicate(This,prop) ) 

#define BrowseModeLongDescVerbosity_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeLongDescVerbosity_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeLongDescVerbosity_INTERFACE_DEFINED__ */


#ifndef __BrowseModeMiscVerbosity_INTERFACE_DEFINED__
#define __BrowseModeMiscVerbosity_INTERFACE_DEFINED__

/* interface BrowseModeMiscVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeMiscVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("46D8E757-B12B-4EC4-8F5C-0BB48A70807E")
    BrowseModeMiscVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateNewPopups( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateNewPopups( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeGraphicWithNoDescription( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeGraphicWithNoDescription( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_VerbosityHotKeyToSpecificElement( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_VerbosityHotKeyToSpecificElement( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AutoLoad( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AutoLoad( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NextTextMinimumLineLength( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NextTextMinimumLineLength( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NextTextConsecutiveLines( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NextTextConsecutiveLines( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SuspendRefresh( 
            /* [in] */ BrowseModeSuspendRefresh prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SuspendRefresh( 
            /* [retval][out] */ BrowseModeSuspendRefresh *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AutoToggle( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AutoToggle( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ModeIndicator( 
            /* [in] */ BrowseModeToggleInidicators prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ModeIndicator( 
            /* [retval][out] */ BrowseModeToggleInidicators *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_QuickReview( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_QuickReview( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeMiscVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeMiscVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeMiscVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeMiscVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeMiscVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeMiscVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeMiscVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeMiscVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeMiscVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeMiscVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateNewPopups )( 
            BrowseModeMiscVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateNewPopups )( 
            BrowseModeMiscVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeGraphicWithNoDescription )( 
            BrowseModeMiscVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeGraphicWithNoDescription )( 
            BrowseModeMiscVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_VerbosityHotKeyToSpecificElement )( 
            BrowseModeMiscVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_VerbosityHotKeyToSpecificElement )( 
            BrowseModeMiscVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AutoLoad )( 
            BrowseModeMiscVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AutoLoad )( 
            BrowseModeMiscVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NextTextMinimumLineLength )( 
            BrowseModeMiscVerbosity * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NextTextMinimumLineLength )( 
            BrowseModeMiscVerbosity * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NextTextConsecutiveLines )( 
            BrowseModeMiscVerbosity * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NextTextConsecutiveLines )( 
            BrowseModeMiscVerbosity * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SuspendRefresh )( 
            BrowseModeMiscVerbosity * This,
            /* [in] */ BrowseModeSuspendRefresh prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SuspendRefresh )( 
            BrowseModeMiscVerbosity * This,
            /* [retval][out] */ BrowseModeSuspendRefresh *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AutoToggle )( 
            BrowseModeMiscVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AutoToggle )( 
            BrowseModeMiscVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ModeIndicator )( 
            BrowseModeMiscVerbosity * This,
            /* [in] */ BrowseModeToggleInidicators prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ModeIndicator )( 
            BrowseModeMiscVerbosity * This,
            /* [retval][out] */ BrowseModeToggleInidicators *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_QuickReview )( 
            BrowseModeMiscVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_QuickReview )( 
            BrowseModeMiscVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } BrowseModeMiscVerbosityVtbl;

    interface BrowseModeMiscVerbosity
    {
        CONST_VTBL struct BrowseModeMiscVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeMiscVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeMiscVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeMiscVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeMiscVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeMiscVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeMiscVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeMiscVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeMiscVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeMiscVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeMiscVerbosity_put_IndicateNewPopups(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateNewPopups(This,prop) ) 

#define BrowseModeMiscVerbosity_get_IndicateNewPopups(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateNewPopups(This,prop) ) 

#define BrowseModeMiscVerbosity_put_IncludeGraphicWithNoDescription(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeGraphicWithNoDescription(This,prop) ) 

#define BrowseModeMiscVerbosity_get_IncludeGraphicWithNoDescription(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeGraphicWithNoDescription(This,prop) ) 

#define BrowseModeMiscVerbosity_put_VerbosityHotKeyToSpecificElement(This,prop)	\
    ( (This)->lpVtbl -> put_VerbosityHotKeyToSpecificElement(This,prop) ) 

#define BrowseModeMiscVerbosity_get_VerbosityHotKeyToSpecificElement(This,prop)	\
    ( (This)->lpVtbl -> get_VerbosityHotKeyToSpecificElement(This,prop) ) 

#define BrowseModeMiscVerbosity_put_AutoLoad(This,prop)	\
    ( (This)->lpVtbl -> put_AutoLoad(This,prop) ) 

#define BrowseModeMiscVerbosity_get_AutoLoad(This,prop)	\
    ( (This)->lpVtbl -> get_AutoLoad(This,prop) ) 

#define BrowseModeMiscVerbosity_put_NextTextMinimumLineLength(This,prop)	\
    ( (This)->lpVtbl -> put_NextTextMinimumLineLength(This,prop) ) 

#define BrowseModeMiscVerbosity_get_NextTextMinimumLineLength(This,prop)	\
    ( (This)->lpVtbl -> get_NextTextMinimumLineLength(This,prop) ) 

#define BrowseModeMiscVerbosity_put_NextTextConsecutiveLines(This,prop)	\
    ( (This)->lpVtbl -> put_NextTextConsecutiveLines(This,prop) ) 

#define BrowseModeMiscVerbosity_get_NextTextConsecutiveLines(This,prop)	\
    ( (This)->lpVtbl -> get_NextTextConsecutiveLines(This,prop) ) 

#define BrowseModeMiscVerbosity_put_SuspendRefresh(This,prop)	\
    ( (This)->lpVtbl -> put_SuspendRefresh(This,prop) ) 

#define BrowseModeMiscVerbosity_get_SuspendRefresh(This,prop)	\
    ( (This)->lpVtbl -> get_SuspendRefresh(This,prop) ) 

#define BrowseModeMiscVerbosity_put_AutoToggle(This,prop)	\
    ( (This)->lpVtbl -> put_AutoToggle(This,prop) ) 

#define BrowseModeMiscVerbosity_get_AutoToggle(This,prop)	\
    ( (This)->lpVtbl -> get_AutoToggle(This,prop) ) 

#define BrowseModeMiscVerbosity_put_ModeIndicator(This,prop)	\
    ( (This)->lpVtbl -> put_ModeIndicator(This,prop) ) 

#define BrowseModeMiscVerbosity_get_ModeIndicator(This,prop)	\
    ( (This)->lpVtbl -> get_ModeIndicator(This,prop) ) 

#define BrowseModeMiscVerbosity_put_QuickReview(This,prop)	\
    ( (This)->lpVtbl -> put_QuickReview(This,prop) ) 

#define BrowseModeMiscVerbosity_get_QuickReview(This,prop)	\
    ( (This)->lpVtbl -> get_QuickReview(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeMiscVerbosity_INTERFACE_DEFINED__ */


#ifndef __BrowseModeNewPageVerbosity_INTERFACE_DEFINED__
#define __BrowseModeNewPageVerbosity_INTERFACE_DEFINED__

/* interface BrowseModeNewPageVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeNewPageVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("0AD0EC69-44F3-4DF9-9956-7383F62101F8")
    BrowseModeNewPageVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_StatusMessage( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_StatusMessage( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfLines( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfLines( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfLinks( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfLinks( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfFrames( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfFrames( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfTables( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfTables( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfLists( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfLists( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfParagraphs( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfParagraphs( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfHeadings( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfHeadings( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfForms( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfForms( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfAcronymsAbbreviations( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfAcronymsAbbreviations( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfBlockQuotes( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfBlockQuotes( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfLanguages( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfLanguages( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfLongDesc( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfLongDesc( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfAnchors( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfAnchors( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfAccessKeys( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfAccessKeys( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfFlashObjects( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfFlashObjects( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfObjects( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfObjects( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_RefreshRate( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_RefreshRate( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadEntirePage( 
            /* [in] */ BrowseModeNewPageRead prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadEntirePage( 
            /* [retval][out] */ BrowseModeNewPageRead *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfLandmarks( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfLandmarks( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeNewPageVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeNewPageVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeNewPageVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeNewPageVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeNewPageVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_StatusMessage )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_StatusMessage )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfLines )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfLines )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfLinks )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfLinks )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfFrames )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfFrames )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfTables )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfTables )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfLists )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfLists )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfParagraphs )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfParagraphs )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfHeadings )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfHeadings )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfForms )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfForms )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfAcronymsAbbreviations )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfAcronymsAbbreviations )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfBlockQuotes )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfBlockQuotes )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfLanguages )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfLanguages )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfLongDesc )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfLongDesc )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfAnchors )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfAnchors )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfAccessKeys )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfAccessKeys )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfFlashObjects )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfFlashObjects )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfObjects )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfObjects )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_RefreshRate )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_RefreshRate )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadEntirePage )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ BrowseModeNewPageRead prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadEntirePage )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ BrowseModeNewPageRead *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfLandmarks )( 
            BrowseModeNewPageVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfLandmarks )( 
            BrowseModeNewPageVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } BrowseModeNewPageVerbosityVtbl;

    interface BrowseModeNewPageVerbosity
    {
        CONST_VTBL struct BrowseModeNewPageVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeNewPageVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeNewPageVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeNewPageVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeNewPageVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeNewPageVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeNewPageVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeNewPageVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeNewPageVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeNewPageVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeNewPageVerbosity_put_StatusMessage(This,prop)	\
    ( (This)->lpVtbl -> put_StatusMessage(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_StatusMessage(This,prop)	\
    ( (This)->lpVtbl -> get_StatusMessage(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_NumberOfLines(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfLines(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_NumberOfLines(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfLines(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_NumberOfLinks(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfLinks(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_NumberOfLinks(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfLinks(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_NumberOfFrames(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfFrames(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_NumberOfFrames(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfFrames(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_NumberOfTables(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfTables(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_NumberOfTables(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfTables(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_NumberOfLists(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfLists(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_NumberOfLists(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfLists(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_NumberOfParagraphs(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfParagraphs(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_NumberOfParagraphs(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfParagraphs(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_NumberOfHeadings(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfHeadings(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_NumberOfHeadings(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfHeadings(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_NumberOfForms(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfForms(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_NumberOfForms(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfForms(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_NumberOfAcronymsAbbreviations(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfAcronymsAbbreviations(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_NumberOfAcronymsAbbreviations(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfAcronymsAbbreviations(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_NumberOfBlockQuotes(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfBlockQuotes(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_NumberOfBlockQuotes(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfBlockQuotes(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_NumberOfLanguages(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfLanguages(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_NumberOfLanguages(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfLanguages(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_NumberOfLongDesc(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfLongDesc(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_NumberOfLongDesc(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfLongDesc(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_NumberOfAnchors(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfAnchors(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_NumberOfAnchors(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfAnchors(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_NumberOfAccessKeys(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfAccessKeys(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_NumberOfAccessKeys(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfAccessKeys(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_NumberOfFlashObjects(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfFlashObjects(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_NumberOfFlashObjects(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfFlashObjects(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_NumberOfObjects(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfObjects(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_NumberOfObjects(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfObjects(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_RefreshRate(This,prop)	\
    ( (This)->lpVtbl -> put_RefreshRate(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_RefreshRate(This,prop)	\
    ( (This)->lpVtbl -> get_RefreshRate(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_ReadEntirePage(This,prop)	\
    ( (This)->lpVtbl -> put_ReadEntirePage(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_ReadEntirePage(This,prop)	\
    ( (This)->lpVtbl -> get_ReadEntirePage(This,prop) ) 

#define BrowseModeNewPageVerbosity_put_NumberOfLandmarks(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfLandmarks(This,prop) ) 

#define BrowseModeNewPageVerbosity_get_NumberOfLandmarks(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfLandmarks(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeNewPageVerbosity_INTERFACE_DEFINED__ */


#ifndef __BrowseModeObjectVerbosity_INTERFACE_DEFINED__
#define __BrowseModeObjectVerbosity_INTERFACE_DEFINED__

/* interface BrowseModeObjectVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeObjectVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("057FEC88-70FE-448A-B3C0-9B3FFCCB3CB6")
    BrowseModeObjectVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeType( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeType( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ExposeAlternateText( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ExposeAlternateText( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeObjectVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeObjectVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeObjectVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeObjectVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeObjectVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeObjectVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeObjectVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeObjectVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeObjectVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeObjectVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeType )( 
            BrowseModeObjectVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeType )( 
            BrowseModeObjectVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ExposeAlternateText )( 
            BrowseModeObjectVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ExposeAlternateText )( 
            BrowseModeObjectVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeObjectVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeObjectVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } BrowseModeObjectVerbosityVtbl;

    interface BrowseModeObjectVerbosity
    {
        CONST_VTBL struct BrowseModeObjectVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeObjectVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeObjectVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeObjectVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeObjectVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeObjectVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeObjectVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeObjectVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeObjectVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeObjectVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeObjectVerbosity_put_IncludeType(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeType(This,prop) ) 

#define BrowseModeObjectVerbosity_get_IncludeType(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeType(This,prop) ) 

#define BrowseModeObjectVerbosity_put_ExposeAlternateText(This,prop)	\
    ( (This)->lpVtbl -> put_ExposeAlternateText(This,prop) ) 

#define BrowseModeObjectVerbosity_get_ExposeAlternateText(This,prop)	\
    ( (This)->lpVtbl -> get_ExposeAlternateText(This,prop) ) 

#define BrowseModeObjectVerbosity_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeObjectVerbosity_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeObjectVerbosity_INTERFACE_DEFINED__ */


#ifndef __BrowseModeParagraphVerbosity_INTERFACE_DEFINED__
#define __BrowseModeParagraphVerbosity_INTERFACE_DEFINED__

/* interface BrowseModeParagraphVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeParagraphVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("83D4752A-5745-4AA9-85F7-D7074DCCDABD")
    BrowseModeParagraphVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indicate( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indicate( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SmartParagraphMode( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SmartParagraphMode( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ShowBlankLines( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ShowBlankLines( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeParagraphVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeParagraphVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeParagraphVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeParagraphVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeParagraphVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeParagraphVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeParagraphVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeParagraphVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeParagraphVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeParagraphVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indicate )( 
            BrowseModeParagraphVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indicate )( 
            BrowseModeParagraphVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeParagraphVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeParagraphVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SmartParagraphMode )( 
            BrowseModeParagraphVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SmartParagraphMode )( 
            BrowseModeParagraphVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ShowBlankLines )( 
            BrowseModeParagraphVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ShowBlankLines )( 
            BrowseModeParagraphVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } BrowseModeParagraphVerbosityVtbl;

    interface BrowseModeParagraphVerbosity
    {
        CONST_VTBL struct BrowseModeParagraphVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeParagraphVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeParagraphVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeParagraphVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeParagraphVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeParagraphVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeParagraphVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeParagraphVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeParagraphVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeParagraphVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeParagraphVerbosity_put_Indicate(This,prop)	\
    ( (This)->lpVtbl -> put_Indicate(This,prop) ) 

#define BrowseModeParagraphVerbosity_get_Indicate(This,prop)	\
    ( (This)->lpVtbl -> get_Indicate(This,prop) ) 

#define BrowseModeParagraphVerbosity_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeParagraphVerbosity_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#define BrowseModeParagraphVerbosity_put_SmartParagraphMode(This,prop)	\
    ( (This)->lpVtbl -> put_SmartParagraphMode(This,prop) ) 

#define BrowseModeParagraphVerbosity_get_SmartParagraphMode(This,prop)	\
    ( (This)->lpVtbl -> get_SmartParagraphMode(This,prop) ) 

#define BrowseModeParagraphVerbosity_put_ShowBlankLines(This,prop)	\
    ( (This)->lpVtbl -> put_ShowBlankLines(This,prop) ) 

#define BrowseModeParagraphVerbosity_get_ShowBlankLines(This,prop)	\
    ( (This)->lpVtbl -> get_ShowBlankLines(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeParagraphVerbosity_INTERFACE_DEFINED__ */


#ifndef __BrowseModeQuoteVerbosity_INTERFACE_DEFINED__
#define __BrowseModeQuoteVerbosity_INTERFACE_DEFINED__

/* interface BrowseModeQuoteVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeQuoteVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("5353CF70-8FFF-4DB8-BF73-5AE56610A183")
    BrowseModeQuoteVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indicate( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indicate( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeQuoteVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeQuoteVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeQuoteVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeQuoteVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeQuoteVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeQuoteVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeQuoteVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeQuoteVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeQuoteVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeQuoteVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indicate )( 
            BrowseModeQuoteVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indicate )( 
            BrowseModeQuoteVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeQuoteVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeQuoteVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } BrowseModeQuoteVerbosityVtbl;

    interface BrowseModeQuoteVerbosity
    {
        CONST_VTBL struct BrowseModeQuoteVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeQuoteVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeQuoteVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeQuoteVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeQuoteVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeQuoteVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeQuoteVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeQuoteVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeQuoteVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeQuoteVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeQuoteVerbosity_put_Indicate(This,prop)	\
    ( (This)->lpVtbl -> put_Indicate(This,prop) ) 

#define BrowseModeQuoteVerbosity_get_Indicate(This,prop)	\
    ( (This)->lpVtbl -> get_Indicate(This,prop) ) 

#define BrowseModeQuoteVerbosity_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeQuoteVerbosity_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeQuoteVerbosity_INTERFACE_DEFINED__ */


#ifndef __BrowseModeTableVerbosity_INTERFACE_DEFINED__
#define __BrowseModeTableVerbosity_INTERFACE_DEFINED__

/* interface BrowseModeTableVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeTableVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("37931477-81C0-473C-B5C6-D1A0BF565F1A")
    BrowseModeTableVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeSummary( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeSummary( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeSize( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeSize( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeRowGrouping( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeRowGrouping( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeMergedCells( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeMergedCells( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_HeaderAttributeOverride( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HeaderAttributeOverride( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indication( 
            /* [in] */ BrowseModeTableIndication prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indication( 
            /* [retval][out] */ BrowseModeTableIndication *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_DefaultHeaderTo( 
            /* [in] */ BrowseModeTableHeaders prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DefaultHeaderTo( 
            /* [retval][out] */ BrowseModeTableHeaders *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeTableVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeTableVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeTableVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeTableVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeTableVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeTableVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeTableVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeTableVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeTableVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeTableVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeTableVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeTableVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeSummary )( 
            BrowseModeTableVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeSummary )( 
            BrowseModeTableVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeSize )( 
            BrowseModeTableVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeSize )( 
            BrowseModeTableVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeRowGrouping )( 
            BrowseModeTableVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeRowGrouping )( 
            BrowseModeTableVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeMergedCells )( 
            BrowseModeTableVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeMergedCells )( 
            BrowseModeTableVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_HeaderAttributeOverride )( 
            BrowseModeTableVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HeaderAttributeOverride )( 
            BrowseModeTableVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indication )( 
            BrowseModeTableVerbosity * This,
            /* [in] */ BrowseModeTableIndication prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indication )( 
            BrowseModeTableVerbosity * This,
            /* [retval][out] */ BrowseModeTableIndication *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_DefaultHeaderTo )( 
            BrowseModeTableVerbosity * This,
            /* [in] */ BrowseModeTableHeaders prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DefaultHeaderTo )( 
            BrowseModeTableVerbosity * This,
            /* [retval][out] */ BrowseModeTableHeaders *prop);
        
        END_INTERFACE
    } BrowseModeTableVerbosityVtbl;

    interface BrowseModeTableVerbosity
    {
        CONST_VTBL struct BrowseModeTableVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeTableVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeTableVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeTableVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeTableVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeTableVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeTableVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeTableVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeTableVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeTableVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeTableVerbosity_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeTableVerbosity_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#define BrowseModeTableVerbosity_put_IncludeSummary(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeSummary(This,prop) ) 

#define BrowseModeTableVerbosity_get_IncludeSummary(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeSummary(This,prop) ) 

#define BrowseModeTableVerbosity_put_IncludeSize(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeSize(This,prop) ) 

#define BrowseModeTableVerbosity_get_IncludeSize(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeSize(This,prop) ) 

#define BrowseModeTableVerbosity_put_IncludeRowGrouping(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeRowGrouping(This,prop) ) 

#define BrowseModeTableVerbosity_get_IncludeRowGrouping(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeRowGrouping(This,prop) ) 

#define BrowseModeTableVerbosity_put_IncludeMergedCells(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeMergedCells(This,prop) ) 

#define BrowseModeTableVerbosity_get_IncludeMergedCells(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeMergedCells(This,prop) ) 

#define BrowseModeTableVerbosity_put_HeaderAttributeOverride(This,prop)	\
    ( (This)->lpVtbl -> put_HeaderAttributeOverride(This,prop) ) 

#define BrowseModeTableVerbosity_get_HeaderAttributeOverride(This,prop)	\
    ( (This)->lpVtbl -> get_HeaderAttributeOverride(This,prop) ) 

#define BrowseModeTableVerbosity_put_Indication(This,prop)	\
    ( (This)->lpVtbl -> put_Indication(This,prop) ) 

#define BrowseModeTableVerbosity_get_Indication(This,prop)	\
    ( (This)->lpVtbl -> get_Indication(This,prop) ) 

#define BrowseModeTableVerbosity_put_DefaultHeaderTo(This,prop)	\
    ( (This)->lpVtbl -> put_DefaultHeaderTo(This,prop) ) 

#define BrowseModeTableVerbosity_get_DefaultHeaderTo(This,prop)	\
    ( (This)->lpVtbl -> get_DefaultHeaderTo(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeTableVerbosity_INTERFACE_DEFINED__ */


#ifndef __WordVerbosity_INTERFACE_DEFINED__
#define __WordVerbosity_INTERFACE_DEFINED__

/* interface WordVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WordVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("45F09BBF-5811-49BE-BF38-F97ADBAA7D9E")
    WordVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Columns( 
            /* [retval][out] */ WordColumnVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Fields( 
            /* [retval][out] */ WordFieldVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HeadersFooters( 
            /* [retval][out] */ WordHeaderFooterVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Miscellaneous( 
            /* [retval][out] */ WordMiscVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_PicturesObjects( 
            /* [retval][out] */ WordPictureObjectVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_References( 
            /* [retval][out] */ WordReferenceVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Revisions( 
            /* [retval][out] */ WordRevisionVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Sections( 
            /* [retval][out] */ WordSectionVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpellingGrammar( 
            /* [retval][out] */ WordSpellingGrammarVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Tables( 
            /* [retval][out] */ WordTableVerbosity **obj) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WordVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WordVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WordVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WordVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WordVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WordVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WordVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WordVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WordVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WordVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Columns )( 
            WordVerbosity * This,
            /* [retval][out] */ WordColumnVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Fields )( 
            WordVerbosity * This,
            /* [retval][out] */ WordFieldVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HeadersFooters )( 
            WordVerbosity * This,
            /* [retval][out] */ WordHeaderFooterVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Miscellaneous )( 
            WordVerbosity * This,
            /* [retval][out] */ WordMiscVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_PicturesObjects )( 
            WordVerbosity * This,
            /* [retval][out] */ WordPictureObjectVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_References )( 
            WordVerbosity * This,
            /* [retval][out] */ WordReferenceVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Revisions )( 
            WordVerbosity * This,
            /* [retval][out] */ WordRevisionVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Sections )( 
            WordVerbosity * This,
            /* [retval][out] */ WordSectionVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpellingGrammar )( 
            WordVerbosity * This,
            /* [retval][out] */ WordSpellingGrammarVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Tables )( 
            WordVerbosity * This,
            /* [retval][out] */ WordTableVerbosity **obj);
        
        END_INTERFACE
    } WordVerbosityVtbl;

    interface WordVerbosity
    {
        CONST_VTBL struct WordVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WordVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WordVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WordVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WordVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WordVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WordVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WordVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WordVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WordVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WordVerbosity_get_Columns(This,obj)	\
    ( (This)->lpVtbl -> get_Columns(This,obj) ) 

#define WordVerbosity_get_Fields(This,obj)	\
    ( (This)->lpVtbl -> get_Fields(This,obj) ) 

#define WordVerbosity_get_HeadersFooters(This,obj)	\
    ( (This)->lpVtbl -> get_HeadersFooters(This,obj) ) 

#define WordVerbosity_get_Miscellaneous(This,obj)	\
    ( (This)->lpVtbl -> get_Miscellaneous(This,obj) ) 

#define WordVerbosity_get_PicturesObjects(This,obj)	\
    ( (This)->lpVtbl -> get_PicturesObjects(This,obj) ) 

#define WordVerbosity_get_References(This,obj)	\
    ( (This)->lpVtbl -> get_References(This,obj) ) 

#define WordVerbosity_get_Revisions(This,obj)	\
    ( (This)->lpVtbl -> get_Revisions(This,obj) ) 

#define WordVerbosity_get_Sections(This,obj)	\
    ( (This)->lpVtbl -> get_Sections(This,obj) ) 

#define WordVerbosity_get_SpellingGrammar(This,obj)	\
    ( (This)->lpVtbl -> get_SpellingGrammar(This,obj) ) 

#define WordVerbosity_get_Tables(This,obj)	\
    ( (This)->lpVtbl -> get_Tables(This,obj) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WordVerbosity_INTERFACE_DEFINED__ */


#ifndef __WordColumnVerbosity_INTERFACE_DEFINED__
#define __WordColumnVerbosity_INTERFACE_DEFINED__

/* interface WordColumnVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WordColumnVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("E7F432C2-6A0D-43BC-B106-E5414C23FF24")
    WordColumnVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Identify( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Identify( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IdentifyReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IdentifyReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WordColumnVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WordColumnVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WordColumnVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WordColumnVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WordColumnVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WordColumnVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WordColumnVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WordColumnVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WordColumnVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WordColumnVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Identify )( 
            WordColumnVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Identify )( 
            WordColumnVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IdentifyReadToEnd )( 
            WordColumnVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IdentifyReadToEnd )( 
            WordColumnVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } WordColumnVerbosityVtbl;

    interface WordColumnVerbosity
    {
        CONST_VTBL struct WordColumnVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WordColumnVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WordColumnVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WordColumnVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WordColumnVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WordColumnVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WordColumnVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WordColumnVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WordColumnVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WordColumnVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WordColumnVerbosity_put_Identify(This,prop)	\
    ( (This)->lpVtbl -> put_Identify(This,prop) ) 

#define WordColumnVerbosity_get_Identify(This,prop)	\
    ( (This)->lpVtbl -> get_Identify(This,prop) ) 

#define WordColumnVerbosity_put_IdentifyReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_IdentifyReadToEnd(This,prop) ) 

#define WordColumnVerbosity_get_IdentifyReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_IdentifyReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WordColumnVerbosity_INTERFACE_DEFINED__ */


#ifndef __WordFieldVerbosity_INTERFACE_DEFINED__
#define __WordFieldVerbosity_INTERFACE_DEFINED__

/* interface WordFieldVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WordFieldVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("2113D7CE-5269-4CBC-9583-C1844036E1B6")
    WordFieldVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Identify( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Identify( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Count( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CountReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CountReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WordFieldVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WordFieldVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WordFieldVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WordFieldVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WordFieldVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WordFieldVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WordFieldVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WordFieldVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WordFieldVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WordFieldVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Identify )( 
            WordFieldVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Identify )( 
            WordFieldVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Count )( 
            WordFieldVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            WordFieldVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CountReadToEnd )( 
            WordFieldVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CountReadToEnd )( 
            WordFieldVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } WordFieldVerbosityVtbl;

    interface WordFieldVerbosity
    {
        CONST_VTBL struct WordFieldVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WordFieldVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WordFieldVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WordFieldVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WordFieldVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WordFieldVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WordFieldVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WordFieldVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WordFieldVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WordFieldVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WordFieldVerbosity_put_Identify(This,prop)	\
    ( (This)->lpVtbl -> put_Identify(This,prop) ) 

#define WordFieldVerbosity_get_Identify(This,prop)	\
    ( (This)->lpVtbl -> get_Identify(This,prop) ) 

#define WordFieldVerbosity_put_Count(This,prop)	\
    ( (This)->lpVtbl -> put_Count(This,prop) ) 

#define WordFieldVerbosity_get_Count(This,prop)	\
    ( (This)->lpVtbl -> get_Count(This,prop) ) 

#define WordFieldVerbosity_put_CountReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_CountReadToEnd(This,prop) ) 

#define WordFieldVerbosity_get_CountReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_CountReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WordFieldVerbosity_INTERFACE_DEFINED__ */


#ifndef __WordHeaderFooterVerbosity_INTERFACE_DEFINED__
#define __WordHeaderFooterVerbosity_INTERFACE_DEFINED__

/* interface WordHeaderFooterVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WordHeaderFooterVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("AF1222C1-C085-4E87-B836-4959FBEB7228")
    WordHeaderFooterVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_HeaderIdentify( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HeaderIdentify( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FooterIdentify( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FooterIdentify( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_HeaderIdentifyReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HeaderIdentifyReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FooterIdentifyReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FooterIdentifyReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_PageIdentify( 
            /* [in] */ WordPageVerbosity prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_PageIdentify( 
            /* [retval][out] */ WordPageVerbosity *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_PageIdentifyReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_PageIdentifyReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WordHeaderFooterVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WordHeaderFooterVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WordHeaderFooterVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WordHeaderFooterVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WordHeaderFooterVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WordHeaderFooterVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WordHeaderFooterVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WordHeaderFooterVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WordHeaderFooterVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WordHeaderFooterVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_HeaderIdentify )( 
            WordHeaderFooterVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HeaderIdentify )( 
            WordHeaderFooterVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FooterIdentify )( 
            WordHeaderFooterVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FooterIdentify )( 
            WordHeaderFooterVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_HeaderIdentifyReadToEnd )( 
            WordHeaderFooterVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HeaderIdentifyReadToEnd )( 
            WordHeaderFooterVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FooterIdentifyReadToEnd )( 
            WordHeaderFooterVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FooterIdentifyReadToEnd )( 
            WordHeaderFooterVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_PageIdentify )( 
            WordHeaderFooterVerbosity * This,
            /* [in] */ WordPageVerbosity prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_PageIdentify )( 
            WordHeaderFooterVerbosity * This,
            /* [retval][out] */ WordPageVerbosity *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_PageIdentifyReadToEnd )( 
            WordHeaderFooterVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_PageIdentifyReadToEnd )( 
            WordHeaderFooterVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } WordHeaderFooterVerbosityVtbl;

    interface WordHeaderFooterVerbosity
    {
        CONST_VTBL struct WordHeaderFooterVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WordHeaderFooterVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WordHeaderFooterVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WordHeaderFooterVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WordHeaderFooterVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WordHeaderFooterVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WordHeaderFooterVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WordHeaderFooterVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WordHeaderFooterVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WordHeaderFooterVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WordHeaderFooterVerbosity_put_HeaderIdentify(This,prop)	\
    ( (This)->lpVtbl -> put_HeaderIdentify(This,prop) ) 

#define WordHeaderFooterVerbosity_get_HeaderIdentify(This,prop)	\
    ( (This)->lpVtbl -> get_HeaderIdentify(This,prop) ) 

#define WordHeaderFooterVerbosity_put_FooterIdentify(This,prop)	\
    ( (This)->lpVtbl -> put_FooterIdentify(This,prop) ) 

#define WordHeaderFooterVerbosity_get_FooterIdentify(This,prop)	\
    ( (This)->lpVtbl -> get_FooterIdentify(This,prop) ) 

#define WordHeaderFooterVerbosity_put_HeaderIdentifyReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_HeaderIdentifyReadToEnd(This,prop) ) 

#define WordHeaderFooterVerbosity_get_HeaderIdentifyReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_HeaderIdentifyReadToEnd(This,prop) ) 

#define WordHeaderFooterVerbosity_put_FooterIdentifyReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_FooterIdentifyReadToEnd(This,prop) ) 

#define WordHeaderFooterVerbosity_get_FooterIdentifyReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_FooterIdentifyReadToEnd(This,prop) ) 

#define WordHeaderFooterVerbosity_put_PageIdentify(This,prop)	\
    ( (This)->lpVtbl -> put_PageIdentify(This,prop) ) 

#define WordHeaderFooterVerbosity_get_PageIdentify(This,prop)	\
    ( (This)->lpVtbl -> get_PageIdentify(This,prop) ) 

#define WordHeaderFooterVerbosity_put_PageIdentifyReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_PageIdentifyReadToEnd(This,prop) ) 

#define WordHeaderFooterVerbosity_get_PageIdentifyReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_PageIdentifyReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WordHeaderFooterVerbosity_INTERFACE_DEFINED__ */


#ifndef __WordMiscVerbosity_INTERFACE_DEFINED__
#define __WordMiscVerbosity_INTERFACE_DEFINED__

/* interface WordMiscVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WordMiscVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("85E8A2FB-7255-46FD-84EF-5CAFBF459AEF")
    WordMiscVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AutoCorrectIdentify( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AutoCorrectIdentify( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_BorderIdentify( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BorderIdentify( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SmartTagIdentify( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SmartTagIdentify( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_StyleIdentify( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_StyleIdentify( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AutoDockToolbars( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AutoDockToolbars( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WordMiscVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WordMiscVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WordMiscVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WordMiscVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WordMiscVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WordMiscVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WordMiscVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WordMiscVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WordMiscVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WordMiscVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AutoCorrectIdentify )( 
            WordMiscVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AutoCorrectIdentify )( 
            WordMiscVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_BorderIdentify )( 
            WordMiscVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BorderIdentify )( 
            WordMiscVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SmartTagIdentify )( 
            WordMiscVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SmartTagIdentify )( 
            WordMiscVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_StyleIdentify )( 
            WordMiscVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_StyleIdentify )( 
            WordMiscVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AutoDockToolbars )( 
            WordMiscVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AutoDockToolbars )( 
            WordMiscVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } WordMiscVerbosityVtbl;

    interface WordMiscVerbosity
    {
        CONST_VTBL struct WordMiscVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WordMiscVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WordMiscVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WordMiscVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WordMiscVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WordMiscVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WordMiscVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WordMiscVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WordMiscVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WordMiscVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WordMiscVerbosity_put_AutoCorrectIdentify(This,prop)	\
    ( (This)->lpVtbl -> put_AutoCorrectIdentify(This,prop) ) 

#define WordMiscVerbosity_get_AutoCorrectIdentify(This,prop)	\
    ( (This)->lpVtbl -> get_AutoCorrectIdentify(This,prop) ) 

#define WordMiscVerbosity_put_BorderIdentify(This,prop)	\
    ( (This)->lpVtbl -> put_BorderIdentify(This,prop) ) 

#define WordMiscVerbosity_get_BorderIdentify(This,prop)	\
    ( (This)->lpVtbl -> get_BorderIdentify(This,prop) ) 

#define WordMiscVerbosity_put_SmartTagIdentify(This,prop)	\
    ( (This)->lpVtbl -> put_SmartTagIdentify(This,prop) ) 

#define WordMiscVerbosity_get_SmartTagIdentify(This,prop)	\
    ( (This)->lpVtbl -> get_SmartTagIdentify(This,prop) ) 

#define WordMiscVerbosity_put_StyleIdentify(This,prop)	\
    ( (This)->lpVtbl -> put_StyleIdentify(This,prop) ) 

#define WordMiscVerbosity_get_StyleIdentify(This,prop)	\
    ( (This)->lpVtbl -> get_StyleIdentify(This,prop) ) 

#define WordMiscVerbosity_put_AutoDockToolbars(This,prop)	\
    ( (This)->lpVtbl -> put_AutoDockToolbars(This,prop) ) 

#define WordMiscVerbosity_get_AutoDockToolbars(This,prop)	\
    ( (This)->lpVtbl -> get_AutoDockToolbars(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WordMiscVerbosity_INTERFACE_DEFINED__ */


#ifndef __WordPictureObjectVerbosity_INTERFACE_DEFINED__
#define __WordPictureObjectVerbosity_INTERFACE_DEFINED__

/* interface WordPictureObjectVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WordPictureObjectVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("C1DA88D1-913A-44EB-9BF8-19057CC47D12")
    WordPictureObjectVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Identify( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Identify( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IdentifyReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IdentifyReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WordPictureObjectVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WordPictureObjectVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WordPictureObjectVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WordPictureObjectVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WordPictureObjectVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WordPictureObjectVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WordPictureObjectVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WordPictureObjectVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WordPictureObjectVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WordPictureObjectVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Identify )( 
            WordPictureObjectVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Identify )( 
            WordPictureObjectVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IdentifyReadToEnd )( 
            WordPictureObjectVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IdentifyReadToEnd )( 
            WordPictureObjectVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } WordPictureObjectVerbosityVtbl;

    interface WordPictureObjectVerbosity
    {
        CONST_VTBL struct WordPictureObjectVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WordPictureObjectVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WordPictureObjectVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WordPictureObjectVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WordPictureObjectVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WordPictureObjectVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WordPictureObjectVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WordPictureObjectVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WordPictureObjectVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WordPictureObjectVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WordPictureObjectVerbosity_put_Identify(This,prop)	\
    ( (This)->lpVtbl -> put_Identify(This,prop) ) 

#define WordPictureObjectVerbosity_get_Identify(This,prop)	\
    ( (This)->lpVtbl -> get_Identify(This,prop) ) 

#define WordPictureObjectVerbosity_put_IdentifyReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_IdentifyReadToEnd(This,prop) ) 

#define WordPictureObjectVerbosity_get_IdentifyReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_IdentifyReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WordPictureObjectVerbosity_INTERFACE_DEFINED__ */


#ifndef __WordReferenceVerbosity_INTERFACE_DEFINED__
#define __WordReferenceVerbosity_INTERFACE_DEFINED__

/* interface WordReferenceVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WordReferenceVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("037DFDED-C37E-44A0-AAFE-C4A1844654E6")
    WordReferenceVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Identify( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Identify( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Count( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CountReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CountReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WordReferenceVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WordReferenceVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WordReferenceVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WordReferenceVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WordReferenceVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WordReferenceVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WordReferenceVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WordReferenceVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WordReferenceVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WordReferenceVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Identify )( 
            WordReferenceVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Identify )( 
            WordReferenceVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Count )( 
            WordReferenceVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            WordReferenceVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CountReadToEnd )( 
            WordReferenceVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CountReadToEnd )( 
            WordReferenceVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } WordReferenceVerbosityVtbl;

    interface WordReferenceVerbosity
    {
        CONST_VTBL struct WordReferenceVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WordReferenceVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WordReferenceVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WordReferenceVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WordReferenceVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WordReferenceVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WordReferenceVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WordReferenceVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WordReferenceVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WordReferenceVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WordReferenceVerbosity_put_Identify(This,prop)	\
    ( (This)->lpVtbl -> put_Identify(This,prop) ) 

#define WordReferenceVerbosity_get_Identify(This,prop)	\
    ( (This)->lpVtbl -> get_Identify(This,prop) ) 

#define WordReferenceVerbosity_put_Count(This,prop)	\
    ( (This)->lpVtbl -> put_Count(This,prop) ) 

#define WordReferenceVerbosity_get_Count(This,prop)	\
    ( (This)->lpVtbl -> get_Count(This,prop) ) 

#define WordReferenceVerbosity_put_CountReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_CountReadToEnd(This,prop) ) 

#define WordReferenceVerbosity_get_CountReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_CountReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WordReferenceVerbosity_INTERFACE_DEFINED__ */


#ifndef __WordRevisionVerbosity_INTERFACE_DEFINED__
#define __WordRevisionVerbosity_INTERFACE_DEFINED__

/* interface WordRevisionVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WordRevisionVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("CEA173DD-411A-4789-AB4B-938F9D924B7B")
    WordRevisionVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Identify( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Identify( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IdentifyAuthor( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IdentifyAuthor( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IdentifyDate( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IdentifyDate( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Count( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CountReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CountReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WordRevisionVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WordRevisionVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WordRevisionVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WordRevisionVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WordRevisionVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WordRevisionVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WordRevisionVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WordRevisionVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WordRevisionVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WordRevisionVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Identify )( 
            WordRevisionVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Identify )( 
            WordRevisionVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IdentifyAuthor )( 
            WordRevisionVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IdentifyAuthor )( 
            WordRevisionVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IdentifyDate )( 
            WordRevisionVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IdentifyDate )( 
            WordRevisionVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Count )( 
            WordRevisionVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            WordRevisionVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CountReadToEnd )( 
            WordRevisionVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CountReadToEnd )( 
            WordRevisionVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } WordRevisionVerbosityVtbl;

    interface WordRevisionVerbosity
    {
        CONST_VTBL struct WordRevisionVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WordRevisionVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WordRevisionVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WordRevisionVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WordRevisionVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WordRevisionVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WordRevisionVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WordRevisionVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WordRevisionVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WordRevisionVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WordRevisionVerbosity_put_Identify(This,prop)	\
    ( (This)->lpVtbl -> put_Identify(This,prop) ) 

#define WordRevisionVerbosity_get_Identify(This,prop)	\
    ( (This)->lpVtbl -> get_Identify(This,prop) ) 

#define WordRevisionVerbosity_put_IdentifyAuthor(This,prop)	\
    ( (This)->lpVtbl -> put_IdentifyAuthor(This,prop) ) 

#define WordRevisionVerbosity_get_IdentifyAuthor(This,prop)	\
    ( (This)->lpVtbl -> get_IdentifyAuthor(This,prop) ) 

#define WordRevisionVerbosity_put_IdentifyDate(This,prop)	\
    ( (This)->lpVtbl -> put_IdentifyDate(This,prop) ) 

#define WordRevisionVerbosity_get_IdentifyDate(This,prop)	\
    ( (This)->lpVtbl -> get_IdentifyDate(This,prop) ) 

#define WordRevisionVerbosity_put_Count(This,prop)	\
    ( (This)->lpVtbl -> put_Count(This,prop) ) 

#define WordRevisionVerbosity_get_Count(This,prop)	\
    ( (This)->lpVtbl -> get_Count(This,prop) ) 

#define WordRevisionVerbosity_put_CountReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_CountReadToEnd(This,prop) ) 

#define WordRevisionVerbosity_get_CountReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_CountReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WordRevisionVerbosity_INTERFACE_DEFINED__ */


#ifndef __WordSectionVerbosity_INTERFACE_DEFINED__
#define __WordSectionVerbosity_INTERFACE_DEFINED__

/* interface WordSectionVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WordSectionVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("766C76FB-6AA8-4C0A-B9B1-2F946B4CBC40")
    WordSectionVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Identify( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Identify( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IdentifyReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IdentifyReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WordSectionVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WordSectionVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WordSectionVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WordSectionVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WordSectionVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WordSectionVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WordSectionVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WordSectionVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WordSectionVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WordSectionVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Identify )( 
            WordSectionVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Identify )( 
            WordSectionVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IdentifyReadToEnd )( 
            WordSectionVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IdentifyReadToEnd )( 
            WordSectionVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } WordSectionVerbosityVtbl;

    interface WordSectionVerbosity
    {
        CONST_VTBL struct WordSectionVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WordSectionVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WordSectionVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WordSectionVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WordSectionVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WordSectionVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WordSectionVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WordSectionVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WordSectionVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WordSectionVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WordSectionVerbosity_put_Identify(This,prop)	\
    ( (This)->lpVtbl -> put_Identify(This,prop) ) 

#define WordSectionVerbosity_get_Identify(This,prop)	\
    ( (This)->lpVtbl -> get_Identify(This,prop) ) 

#define WordSectionVerbosity_put_IdentifyReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_IdentifyReadToEnd(This,prop) ) 

#define WordSectionVerbosity_get_IdentifyReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_IdentifyReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WordSectionVerbosity_INTERFACE_DEFINED__ */


#ifndef __WordSpellingGrammarVerbosity_INTERFACE_DEFINED__
#define __WordSpellingGrammarVerbosity_INTERFACE_DEFINED__

/* interface WordSpellingGrammarVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WordSpellingGrammarVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("61682996-ACDE-4715-ABB5-F8DBFD629B40")
    WordSpellingGrammarVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpellingIdentify( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpellingIdentify( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_GrammarIdentify( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_GrammarIdentify( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_GrammarCount( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_GrammarCount( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_GrammarCountReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_GrammarCountReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpellingCount( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpellingCount( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpellingCountReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpellingCountReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WordSpellingGrammarVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WordSpellingGrammarVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WordSpellingGrammarVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WordSpellingGrammarVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WordSpellingGrammarVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WordSpellingGrammarVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WordSpellingGrammarVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WordSpellingGrammarVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WordSpellingGrammarVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WordSpellingGrammarVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpellingIdentify )( 
            WordSpellingGrammarVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpellingIdentify )( 
            WordSpellingGrammarVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_GrammarIdentify )( 
            WordSpellingGrammarVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_GrammarIdentify )( 
            WordSpellingGrammarVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_GrammarCount )( 
            WordSpellingGrammarVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_GrammarCount )( 
            WordSpellingGrammarVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_GrammarCountReadToEnd )( 
            WordSpellingGrammarVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_GrammarCountReadToEnd )( 
            WordSpellingGrammarVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpellingCount )( 
            WordSpellingGrammarVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpellingCount )( 
            WordSpellingGrammarVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpellingCountReadToEnd )( 
            WordSpellingGrammarVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpellingCountReadToEnd )( 
            WordSpellingGrammarVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } WordSpellingGrammarVerbosityVtbl;

    interface WordSpellingGrammarVerbosity
    {
        CONST_VTBL struct WordSpellingGrammarVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WordSpellingGrammarVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WordSpellingGrammarVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WordSpellingGrammarVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WordSpellingGrammarVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WordSpellingGrammarVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WordSpellingGrammarVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WordSpellingGrammarVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WordSpellingGrammarVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WordSpellingGrammarVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WordSpellingGrammarVerbosity_put_SpellingIdentify(This,prop)	\
    ( (This)->lpVtbl -> put_SpellingIdentify(This,prop) ) 

#define WordSpellingGrammarVerbosity_get_SpellingIdentify(This,prop)	\
    ( (This)->lpVtbl -> get_SpellingIdentify(This,prop) ) 

#define WordSpellingGrammarVerbosity_put_GrammarIdentify(This,prop)	\
    ( (This)->lpVtbl -> put_GrammarIdentify(This,prop) ) 

#define WordSpellingGrammarVerbosity_get_GrammarIdentify(This,prop)	\
    ( (This)->lpVtbl -> get_GrammarIdentify(This,prop) ) 

#define WordSpellingGrammarVerbosity_put_GrammarCount(This,prop)	\
    ( (This)->lpVtbl -> put_GrammarCount(This,prop) ) 

#define WordSpellingGrammarVerbosity_get_GrammarCount(This,prop)	\
    ( (This)->lpVtbl -> get_GrammarCount(This,prop) ) 

#define WordSpellingGrammarVerbosity_put_GrammarCountReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_GrammarCountReadToEnd(This,prop) ) 

#define WordSpellingGrammarVerbosity_get_GrammarCountReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_GrammarCountReadToEnd(This,prop) ) 

#define WordSpellingGrammarVerbosity_put_SpellingCount(This,prop)	\
    ( (This)->lpVtbl -> put_SpellingCount(This,prop) ) 

#define WordSpellingGrammarVerbosity_get_SpellingCount(This,prop)	\
    ( (This)->lpVtbl -> get_SpellingCount(This,prop) ) 

#define WordSpellingGrammarVerbosity_put_SpellingCountReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_SpellingCountReadToEnd(This,prop) ) 

#define WordSpellingGrammarVerbosity_get_SpellingCountReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_SpellingCountReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WordSpellingGrammarVerbosity_INTERFACE_DEFINED__ */


#ifndef __WordTableVerbosity_INTERFACE_DEFINED__
#define __WordTableVerbosity_INTERFACE_DEFINED__

/* interface WordTableVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WordTableVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("918C0986-A398-4679-8360-768C68E47733")
    WordTableVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IdentifyTable( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IdentifyTable( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IdentifyTableReadToEnd( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IdentifyTableReadToEnd( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IdentifyCell( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IdentifyCell( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WordTableVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WordTableVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WordTableVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WordTableVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WordTableVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WordTableVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WordTableVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WordTableVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WordTableVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WordTableVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IdentifyTable )( 
            WordTableVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IdentifyTable )( 
            WordTableVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IdentifyTableReadToEnd )( 
            WordTableVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IdentifyTableReadToEnd )( 
            WordTableVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IdentifyCell )( 
            WordTableVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IdentifyCell )( 
            WordTableVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } WordTableVerbosityVtbl;

    interface WordTableVerbosity
    {
        CONST_VTBL struct WordTableVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WordTableVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WordTableVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WordTableVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WordTableVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WordTableVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WordTableVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WordTableVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WordTableVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WordTableVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WordTableVerbosity_put_IdentifyTable(This,prop)	\
    ( (This)->lpVtbl -> put_IdentifyTable(This,prop) ) 

#define WordTableVerbosity_get_IdentifyTable(This,prop)	\
    ( (This)->lpVtbl -> get_IdentifyTable(This,prop) ) 

#define WordTableVerbosity_put_IdentifyTableReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_IdentifyTableReadToEnd(This,prop) ) 

#define WordTableVerbosity_get_IdentifyTableReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_IdentifyTableReadToEnd(This,prop) ) 

#define WordTableVerbosity_put_IdentifyCell(This,prop)	\
    ( (This)->lpVtbl -> put_IdentifyCell(This,prop) ) 

#define WordTableVerbosity_get_IdentifyCell(This,prop)	\
    ( (This)->lpVtbl -> get_IdentifyCell(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WordTableVerbosity_INTERFACE_DEFINED__ */


#ifndef __ExcelVerbosity_INTERFACE_DEFINED__
#define __ExcelVerbosity_INTERFACE_DEFINED__

/* interface ExcelVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ExcelVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("45968376-3971-459D-AD10-27149C4C3B7D")
    ExcelVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CellInformation( 
            /* [retval][out] */ ExcelCellVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Miscellaneous( 
            /* [retval][out] */ ExcelMiscVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Selection( 
            /* [retval][out] */ ExcelSelectionVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Summary( 
            /* [retval][out] */ ExcelSummaryVerbosity **obj) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ExcelVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ExcelVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ExcelVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ExcelVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ExcelVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ExcelVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ExcelVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ExcelVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ExcelVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ExcelVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CellInformation )( 
            ExcelVerbosity * This,
            /* [retval][out] */ ExcelCellVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Miscellaneous )( 
            ExcelVerbosity * This,
            /* [retval][out] */ ExcelMiscVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Selection )( 
            ExcelVerbosity * This,
            /* [retval][out] */ ExcelSelectionVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Summary )( 
            ExcelVerbosity * This,
            /* [retval][out] */ ExcelSummaryVerbosity **obj);
        
        END_INTERFACE
    } ExcelVerbosityVtbl;

    interface ExcelVerbosity
    {
        CONST_VTBL struct ExcelVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ExcelVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ExcelVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ExcelVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ExcelVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ExcelVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ExcelVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ExcelVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ExcelVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ExcelVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ExcelVerbosity_get_CellInformation(This,obj)	\
    ( (This)->lpVtbl -> get_CellInformation(This,obj) ) 

#define ExcelVerbosity_get_Miscellaneous(This,obj)	\
    ( (This)->lpVtbl -> get_Miscellaneous(This,obj) ) 

#define ExcelVerbosity_get_Selection(This,obj)	\
    ( (This)->lpVtbl -> get_Selection(This,obj) ) 

#define ExcelVerbosity_get_Summary(This,obj)	\
    ( (This)->lpVtbl -> get_Summary(This,obj) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ExcelVerbosity_INTERFACE_DEFINED__ */


#ifndef __ExcelCellVerbosity_INTERFACE_DEFINED__
#define __ExcelCellVerbosity_INTERFACE_DEFINED__

/* interface ExcelCellVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ExcelCellVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("7FE654A2-B097-410E-9CAD-4ADFAB97D6B0")
    ExcelCellVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Contents( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Contents( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Border( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Border( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpeakPositionBeforeContents( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpeakPositionBeforeContents( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Formula( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Formula( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Comment( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Comment( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Hyperlink( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Hyperlink( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Clipped( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Clipped( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Wrapped( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Wrapped( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Overlaps( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Overlaps( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Merged( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Merged( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Shrunk( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Shrunk( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AffectsCells( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AffectsCells( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AutoFilter( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AutoFilter( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Orientation( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Orientation( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_InputMessage( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_InputMessage( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ConditionalFormatting( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ConditionalFormatting( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberFormat( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberFormat( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpeakPosition( 
            /* [in] */ ExcelPositionVerbosity prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpeakPosition( 
            /* [retval][out] */ ExcelPositionVerbosity *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ExcelCellVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ExcelCellVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ExcelCellVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ExcelCellVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ExcelCellVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ExcelCellVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ExcelCellVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ExcelCellVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Contents )( 
            ExcelCellVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Contents )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Border )( 
            ExcelCellVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Border )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpeakPositionBeforeContents )( 
            ExcelCellVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpeakPositionBeforeContents )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Formula )( 
            ExcelCellVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Formula )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Comment )( 
            ExcelCellVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Comment )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Hyperlink )( 
            ExcelCellVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Hyperlink )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Clipped )( 
            ExcelCellVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Clipped )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Wrapped )( 
            ExcelCellVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Wrapped )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Overlaps )( 
            ExcelCellVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Overlaps )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Merged )( 
            ExcelCellVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Merged )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Shrunk )( 
            ExcelCellVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Shrunk )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AffectsCells )( 
            ExcelCellVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AffectsCells )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AutoFilter )( 
            ExcelCellVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AutoFilter )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Orientation )( 
            ExcelCellVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Orientation )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_InputMessage )( 
            ExcelCellVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_InputMessage )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ConditionalFormatting )( 
            ExcelCellVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ConditionalFormatting )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberFormat )( 
            ExcelCellVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberFormat )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpeakPosition )( 
            ExcelCellVerbosity * This,
            /* [in] */ ExcelPositionVerbosity prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpeakPosition )( 
            ExcelCellVerbosity * This,
            /* [retval][out] */ ExcelPositionVerbosity *prop);
        
        END_INTERFACE
    } ExcelCellVerbosityVtbl;

    interface ExcelCellVerbosity
    {
        CONST_VTBL struct ExcelCellVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ExcelCellVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ExcelCellVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ExcelCellVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ExcelCellVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ExcelCellVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ExcelCellVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ExcelCellVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ExcelCellVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ExcelCellVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ExcelCellVerbosity_put_Contents(This,prop)	\
    ( (This)->lpVtbl -> put_Contents(This,prop) ) 

#define ExcelCellVerbosity_get_Contents(This,prop)	\
    ( (This)->lpVtbl -> get_Contents(This,prop) ) 

#define ExcelCellVerbosity_put_Border(This,prop)	\
    ( (This)->lpVtbl -> put_Border(This,prop) ) 

#define ExcelCellVerbosity_get_Border(This,prop)	\
    ( (This)->lpVtbl -> get_Border(This,prop) ) 

#define ExcelCellVerbosity_put_SpeakPositionBeforeContents(This,prop)	\
    ( (This)->lpVtbl -> put_SpeakPositionBeforeContents(This,prop) ) 

#define ExcelCellVerbosity_get_SpeakPositionBeforeContents(This,prop)	\
    ( (This)->lpVtbl -> get_SpeakPositionBeforeContents(This,prop) ) 

#define ExcelCellVerbosity_put_Formula(This,prop)	\
    ( (This)->lpVtbl -> put_Formula(This,prop) ) 

#define ExcelCellVerbosity_get_Formula(This,prop)	\
    ( (This)->lpVtbl -> get_Formula(This,prop) ) 

#define ExcelCellVerbosity_put_Comment(This,prop)	\
    ( (This)->lpVtbl -> put_Comment(This,prop) ) 

#define ExcelCellVerbosity_get_Comment(This,prop)	\
    ( (This)->lpVtbl -> get_Comment(This,prop) ) 

#define ExcelCellVerbosity_put_Hyperlink(This,prop)	\
    ( (This)->lpVtbl -> put_Hyperlink(This,prop) ) 

#define ExcelCellVerbosity_get_Hyperlink(This,prop)	\
    ( (This)->lpVtbl -> get_Hyperlink(This,prop) ) 

#define ExcelCellVerbosity_put_Clipped(This,prop)	\
    ( (This)->lpVtbl -> put_Clipped(This,prop) ) 

#define ExcelCellVerbosity_get_Clipped(This,prop)	\
    ( (This)->lpVtbl -> get_Clipped(This,prop) ) 

#define ExcelCellVerbosity_put_Wrapped(This,prop)	\
    ( (This)->lpVtbl -> put_Wrapped(This,prop) ) 

#define ExcelCellVerbosity_get_Wrapped(This,prop)	\
    ( (This)->lpVtbl -> get_Wrapped(This,prop) ) 

#define ExcelCellVerbosity_put_Overlaps(This,prop)	\
    ( (This)->lpVtbl -> put_Overlaps(This,prop) ) 

#define ExcelCellVerbosity_get_Overlaps(This,prop)	\
    ( (This)->lpVtbl -> get_Overlaps(This,prop) ) 

#define ExcelCellVerbosity_put_Merged(This,prop)	\
    ( (This)->lpVtbl -> put_Merged(This,prop) ) 

#define ExcelCellVerbosity_get_Merged(This,prop)	\
    ( (This)->lpVtbl -> get_Merged(This,prop) ) 

#define ExcelCellVerbosity_put_Shrunk(This,prop)	\
    ( (This)->lpVtbl -> put_Shrunk(This,prop) ) 

#define ExcelCellVerbosity_get_Shrunk(This,prop)	\
    ( (This)->lpVtbl -> get_Shrunk(This,prop) ) 

#define ExcelCellVerbosity_put_AffectsCells(This,prop)	\
    ( (This)->lpVtbl -> put_AffectsCells(This,prop) ) 

#define ExcelCellVerbosity_get_AffectsCells(This,prop)	\
    ( (This)->lpVtbl -> get_AffectsCells(This,prop) ) 

#define ExcelCellVerbosity_put_AutoFilter(This,prop)	\
    ( (This)->lpVtbl -> put_AutoFilter(This,prop) ) 

#define ExcelCellVerbosity_get_AutoFilter(This,prop)	\
    ( (This)->lpVtbl -> get_AutoFilter(This,prop) ) 

#define ExcelCellVerbosity_put_Orientation(This,prop)	\
    ( (This)->lpVtbl -> put_Orientation(This,prop) ) 

#define ExcelCellVerbosity_get_Orientation(This,prop)	\
    ( (This)->lpVtbl -> get_Orientation(This,prop) ) 

#define ExcelCellVerbosity_put_InputMessage(This,prop)	\
    ( (This)->lpVtbl -> put_InputMessage(This,prop) ) 

#define ExcelCellVerbosity_get_InputMessage(This,prop)	\
    ( (This)->lpVtbl -> get_InputMessage(This,prop) ) 

#define ExcelCellVerbosity_put_ConditionalFormatting(This,prop)	\
    ( (This)->lpVtbl -> put_ConditionalFormatting(This,prop) ) 

#define ExcelCellVerbosity_get_ConditionalFormatting(This,prop)	\
    ( (This)->lpVtbl -> get_ConditionalFormatting(This,prop) ) 

#define ExcelCellVerbosity_put_NumberFormat(This,prop)	\
    ( (This)->lpVtbl -> put_NumberFormat(This,prop) ) 

#define ExcelCellVerbosity_get_NumberFormat(This,prop)	\
    ( (This)->lpVtbl -> get_NumberFormat(This,prop) ) 

#define ExcelCellVerbosity_put_SpeakPosition(This,prop)	\
    ( (This)->lpVtbl -> put_SpeakPosition(This,prop) ) 

#define ExcelCellVerbosity_get_SpeakPosition(This,prop)	\
    ( (This)->lpVtbl -> get_SpeakPosition(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ExcelCellVerbosity_INTERFACE_DEFINED__ */


#ifndef __ExcelMiscVerbosity_INTERFACE_DEFINED__
#define __ExcelMiscVerbosity_INTERFACE_DEFINED__

/* interface ExcelMiscVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ExcelMiscVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("C6F3B503-D467-4B69-BF02-D2ED035F9350")
    ExcelMiscVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpeakWorksheetNameOnFocus( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpeakWorksheetNameOnFocus( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpeakWorkbookNameOnFocus( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpeakWorkbookNameOnFocus( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AutoDockToolbars( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AutoDockToolbars( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_MeasurementUnits( 
            /* [in] */ VerbosityMeasurementUnits prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MeasurementUnits( 
            /* [retval][out] */ VerbosityMeasurementUnits *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ExcelMiscVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ExcelMiscVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ExcelMiscVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ExcelMiscVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ExcelMiscVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ExcelMiscVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ExcelMiscVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ExcelMiscVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ExcelMiscVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ExcelMiscVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpeakWorksheetNameOnFocus )( 
            ExcelMiscVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpeakWorksheetNameOnFocus )( 
            ExcelMiscVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpeakWorkbookNameOnFocus )( 
            ExcelMiscVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpeakWorkbookNameOnFocus )( 
            ExcelMiscVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AutoDockToolbars )( 
            ExcelMiscVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AutoDockToolbars )( 
            ExcelMiscVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_MeasurementUnits )( 
            ExcelMiscVerbosity * This,
            /* [in] */ VerbosityMeasurementUnits prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MeasurementUnits )( 
            ExcelMiscVerbosity * This,
            /* [retval][out] */ VerbosityMeasurementUnits *prop);
        
        END_INTERFACE
    } ExcelMiscVerbosityVtbl;

    interface ExcelMiscVerbosity
    {
        CONST_VTBL struct ExcelMiscVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ExcelMiscVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ExcelMiscVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ExcelMiscVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ExcelMiscVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ExcelMiscVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ExcelMiscVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ExcelMiscVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ExcelMiscVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ExcelMiscVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ExcelMiscVerbosity_put_SpeakWorksheetNameOnFocus(This,prop)	\
    ( (This)->lpVtbl -> put_SpeakWorksheetNameOnFocus(This,prop) ) 

#define ExcelMiscVerbosity_get_SpeakWorksheetNameOnFocus(This,prop)	\
    ( (This)->lpVtbl -> get_SpeakWorksheetNameOnFocus(This,prop) ) 

#define ExcelMiscVerbosity_put_SpeakWorkbookNameOnFocus(This,prop)	\
    ( (This)->lpVtbl -> put_SpeakWorkbookNameOnFocus(This,prop) ) 

#define ExcelMiscVerbosity_get_SpeakWorkbookNameOnFocus(This,prop)	\
    ( (This)->lpVtbl -> get_SpeakWorkbookNameOnFocus(This,prop) ) 

#define ExcelMiscVerbosity_put_AutoDockToolbars(This,prop)	\
    ( (This)->lpVtbl -> put_AutoDockToolbars(This,prop) ) 

#define ExcelMiscVerbosity_get_AutoDockToolbars(This,prop)	\
    ( (This)->lpVtbl -> get_AutoDockToolbars(This,prop) ) 

#define ExcelMiscVerbosity_put_MeasurementUnits(This,prop)	\
    ( (This)->lpVtbl -> put_MeasurementUnits(This,prop) ) 

#define ExcelMiscVerbosity_get_MeasurementUnits(This,prop)	\
    ( (This)->lpVtbl -> get_MeasurementUnits(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ExcelMiscVerbosity_INTERFACE_DEFINED__ */


#ifndef __ExcelSelectionVerbosity_INTERFACE_DEFINED__
#define __ExcelSelectionVerbosity_INTERFACE_DEFINED__

/* interface ExcelSelectionVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ExcelSelectionVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("D30EB596-6B56-496D-B0F0-DFE91C7517CF")
    ExcelSelectionVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpeakBothNewCorners( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpeakBothNewCorners( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpeakNumberOfRowsAndColumns( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpeakNumberOfRowsAndColumns( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ExcelSelectionVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ExcelSelectionVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ExcelSelectionVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ExcelSelectionVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ExcelSelectionVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ExcelSelectionVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ExcelSelectionVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ExcelSelectionVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ExcelSelectionVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ExcelSelectionVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpeakBothNewCorners )( 
            ExcelSelectionVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpeakBothNewCorners )( 
            ExcelSelectionVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpeakNumberOfRowsAndColumns )( 
            ExcelSelectionVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpeakNumberOfRowsAndColumns )( 
            ExcelSelectionVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } ExcelSelectionVerbosityVtbl;

    interface ExcelSelectionVerbosity
    {
        CONST_VTBL struct ExcelSelectionVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ExcelSelectionVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ExcelSelectionVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ExcelSelectionVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ExcelSelectionVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ExcelSelectionVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ExcelSelectionVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ExcelSelectionVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ExcelSelectionVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ExcelSelectionVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ExcelSelectionVerbosity_put_SpeakBothNewCorners(This,prop)	\
    ( (This)->lpVtbl -> put_SpeakBothNewCorners(This,prop) ) 

#define ExcelSelectionVerbosity_get_SpeakBothNewCorners(This,prop)	\
    ( (This)->lpVtbl -> get_SpeakBothNewCorners(This,prop) ) 

#define ExcelSelectionVerbosity_put_SpeakNumberOfRowsAndColumns(This,prop)	\
    ( (This)->lpVtbl -> put_SpeakNumberOfRowsAndColumns(This,prop) ) 

#define ExcelSelectionVerbosity_get_SpeakNumberOfRowsAndColumns(This,prop)	\
    ( (This)->lpVtbl -> get_SpeakNumberOfRowsAndColumns(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ExcelSelectionVerbosity_INTERFACE_DEFINED__ */


#ifndef __ExcelSummaryVerbosity_INTERFACE_DEFINED__
#define __ExcelSummaryVerbosity_INTERFACE_DEFINED__

/* interface ExcelSummaryVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ExcelSummaryVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("384A3D6D-5B6B-4229-8736-97890FD3CA15")
    ExcelSummaryVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_WorkbookName( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WorkbookName( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_WorksheetName( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WorksheetName( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_WorksheetCount( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WorksheetCount( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CommentCount( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CommentCount( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_HyperlinkCount( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HyperlinkCount( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ObjectCount( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ObjectCount( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ChartCount( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ChartCount( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Position( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Position( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CountsRelativeTo( 
            /* [in] */ ExcelCountsVerbosity prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CountsRelativeTo( 
            /* [retval][out] */ ExcelCountsVerbosity *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ExcelSummaryVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ExcelSummaryVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ExcelSummaryVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ExcelSummaryVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ExcelSummaryVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ExcelSummaryVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ExcelSummaryVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ExcelSummaryVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ExcelSummaryVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ExcelSummaryVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_WorkbookName )( 
            ExcelSummaryVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WorkbookName )( 
            ExcelSummaryVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_WorksheetName )( 
            ExcelSummaryVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WorksheetName )( 
            ExcelSummaryVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_WorksheetCount )( 
            ExcelSummaryVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WorksheetCount )( 
            ExcelSummaryVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CommentCount )( 
            ExcelSummaryVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CommentCount )( 
            ExcelSummaryVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_HyperlinkCount )( 
            ExcelSummaryVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HyperlinkCount )( 
            ExcelSummaryVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ObjectCount )( 
            ExcelSummaryVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ObjectCount )( 
            ExcelSummaryVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ChartCount )( 
            ExcelSummaryVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ChartCount )( 
            ExcelSummaryVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            ExcelSummaryVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            ExcelSummaryVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CountsRelativeTo )( 
            ExcelSummaryVerbosity * This,
            /* [in] */ ExcelCountsVerbosity prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CountsRelativeTo )( 
            ExcelSummaryVerbosity * This,
            /* [retval][out] */ ExcelCountsVerbosity *prop);
        
        END_INTERFACE
    } ExcelSummaryVerbosityVtbl;

    interface ExcelSummaryVerbosity
    {
        CONST_VTBL struct ExcelSummaryVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ExcelSummaryVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ExcelSummaryVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ExcelSummaryVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ExcelSummaryVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ExcelSummaryVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ExcelSummaryVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ExcelSummaryVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ExcelSummaryVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ExcelSummaryVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ExcelSummaryVerbosity_put_WorkbookName(This,prop)	\
    ( (This)->lpVtbl -> put_WorkbookName(This,prop) ) 

#define ExcelSummaryVerbosity_get_WorkbookName(This,prop)	\
    ( (This)->lpVtbl -> get_WorkbookName(This,prop) ) 

#define ExcelSummaryVerbosity_put_WorksheetName(This,prop)	\
    ( (This)->lpVtbl -> put_WorksheetName(This,prop) ) 

#define ExcelSummaryVerbosity_get_WorksheetName(This,prop)	\
    ( (This)->lpVtbl -> get_WorksheetName(This,prop) ) 

#define ExcelSummaryVerbosity_put_WorksheetCount(This,prop)	\
    ( (This)->lpVtbl -> put_WorksheetCount(This,prop) ) 

#define ExcelSummaryVerbosity_get_WorksheetCount(This,prop)	\
    ( (This)->lpVtbl -> get_WorksheetCount(This,prop) ) 

#define ExcelSummaryVerbosity_put_CommentCount(This,prop)	\
    ( (This)->lpVtbl -> put_CommentCount(This,prop) ) 

#define ExcelSummaryVerbosity_get_CommentCount(This,prop)	\
    ( (This)->lpVtbl -> get_CommentCount(This,prop) ) 

#define ExcelSummaryVerbosity_put_HyperlinkCount(This,prop)	\
    ( (This)->lpVtbl -> put_HyperlinkCount(This,prop) ) 

#define ExcelSummaryVerbosity_get_HyperlinkCount(This,prop)	\
    ( (This)->lpVtbl -> get_HyperlinkCount(This,prop) ) 

#define ExcelSummaryVerbosity_put_ObjectCount(This,prop)	\
    ( (This)->lpVtbl -> put_ObjectCount(This,prop) ) 

#define ExcelSummaryVerbosity_get_ObjectCount(This,prop)	\
    ( (This)->lpVtbl -> get_ObjectCount(This,prop) ) 

#define ExcelSummaryVerbosity_put_ChartCount(This,prop)	\
    ( (This)->lpVtbl -> put_ChartCount(This,prop) ) 

#define ExcelSummaryVerbosity_get_ChartCount(This,prop)	\
    ( (This)->lpVtbl -> get_ChartCount(This,prop) ) 

#define ExcelSummaryVerbosity_put_Position(This,prop)	\
    ( (This)->lpVtbl -> put_Position(This,prop) ) 

#define ExcelSummaryVerbosity_get_Position(This,prop)	\
    ( (This)->lpVtbl -> get_Position(This,prop) ) 

#define ExcelSummaryVerbosity_put_CountsRelativeTo(This,prop)	\
    ( (This)->lpVtbl -> put_CountsRelativeTo(This,prop) ) 

#define ExcelSummaryVerbosity_get_CountsRelativeTo(This,prop)	\
    ( (This)->lpVtbl -> get_CountsRelativeTo(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ExcelSummaryVerbosity_INTERFACE_DEFINED__ */


#ifndef __PowerPointVerbosity_INTERFACE_DEFINED__
#define __PowerPointVerbosity_INTERFACE_DEFINED__

/* interface PowerPointVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_PowerPointVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("41648B9D-A0EA-47C7-B684-F2B83D8B0B70")
    PowerPointVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SlidePane( 
            /* [retval][out] */ PowerPointSlideVerbosity **obj) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Playback( 
            /* [retval][out] */ PowerPointPlaybackVerbosity **obj) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct PowerPointVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            PowerPointVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            PowerPointVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            PowerPointVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            PowerPointVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            PowerPointVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            PowerPointVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            PowerPointVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            PowerPointVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            PowerPointVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SlidePane )( 
            PowerPointVerbosity * This,
            /* [retval][out] */ PowerPointSlideVerbosity **obj);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Playback )( 
            PowerPointVerbosity * This,
            /* [retval][out] */ PowerPointPlaybackVerbosity **obj);
        
        END_INTERFACE
    } PowerPointVerbosityVtbl;

    interface PowerPointVerbosity
    {
        CONST_VTBL struct PowerPointVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define PowerPointVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define PowerPointVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define PowerPointVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define PowerPointVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define PowerPointVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define PowerPointVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define PowerPointVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define PowerPointVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define PowerPointVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define PowerPointVerbosity_get_SlidePane(This,obj)	\
    ( (This)->lpVtbl -> get_SlidePane(This,obj) ) 

#define PowerPointVerbosity_get_Playback(This,obj)	\
    ( (This)->lpVtbl -> get_Playback(This,obj) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __PowerPointVerbosity_INTERFACE_DEFINED__ */


#ifndef __PowerPointSlideVerbosity_INTERFACE_DEFINED__
#define __PowerPointSlideVerbosity_INTERFACE_DEFINED__

/* interface PowerPointSlideVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_PowerPointSlideVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("11A3C83E-2A8E-43B1-A68A-32A6ACF79A9F")
    PowerPointSlideVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateSize( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateSize( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateOverlap( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateOverlap( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicatePlacement( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicatePlacement( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateTextOverflow( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateTextOverflow( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateObjectOverflow( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateObjectOverflow( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_MeasurementUnits( 
            /* [in] */ VerbosityMeasurementUnits prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MeasurementUnits( 
            /* [retval][out] */ VerbosityMeasurementUnits *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct PowerPointSlideVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            PowerPointSlideVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            PowerPointSlideVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            PowerPointSlideVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            PowerPointSlideVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            PowerPointSlideVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            PowerPointSlideVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            PowerPointSlideVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            PowerPointSlideVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            PowerPointSlideVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateSize )( 
            PowerPointSlideVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateSize )( 
            PowerPointSlideVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateOverlap )( 
            PowerPointSlideVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateOverlap )( 
            PowerPointSlideVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicatePlacement )( 
            PowerPointSlideVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicatePlacement )( 
            PowerPointSlideVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateTextOverflow )( 
            PowerPointSlideVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateTextOverflow )( 
            PowerPointSlideVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateObjectOverflow )( 
            PowerPointSlideVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateObjectOverflow )( 
            PowerPointSlideVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_MeasurementUnits )( 
            PowerPointSlideVerbosity * This,
            /* [in] */ VerbosityMeasurementUnits prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MeasurementUnits )( 
            PowerPointSlideVerbosity * This,
            /* [retval][out] */ VerbosityMeasurementUnits *prop);
        
        END_INTERFACE
    } PowerPointSlideVerbosityVtbl;

    interface PowerPointSlideVerbosity
    {
        CONST_VTBL struct PowerPointSlideVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define PowerPointSlideVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define PowerPointSlideVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define PowerPointSlideVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define PowerPointSlideVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define PowerPointSlideVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define PowerPointSlideVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define PowerPointSlideVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define PowerPointSlideVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define PowerPointSlideVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define PowerPointSlideVerbosity_put_IndicateSize(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateSize(This,prop) ) 

#define PowerPointSlideVerbosity_get_IndicateSize(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateSize(This,prop) ) 

#define PowerPointSlideVerbosity_put_IndicateOverlap(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateOverlap(This,prop) ) 

#define PowerPointSlideVerbosity_get_IndicateOverlap(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateOverlap(This,prop) ) 

#define PowerPointSlideVerbosity_put_IndicatePlacement(This,prop)	\
    ( (This)->lpVtbl -> put_IndicatePlacement(This,prop) ) 

#define PowerPointSlideVerbosity_get_IndicatePlacement(This,prop)	\
    ( (This)->lpVtbl -> get_IndicatePlacement(This,prop) ) 

#define PowerPointSlideVerbosity_put_IndicateTextOverflow(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateTextOverflow(This,prop) ) 

#define PowerPointSlideVerbosity_get_IndicateTextOverflow(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateTextOverflow(This,prop) ) 

#define PowerPointSlideVerbosity_put_IndicateObjectOverflow(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateObjectOverflow(This,prop) ) 

#define PowerPointSlideVerbosity_get_IndicateObjectOverflow(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateObjectOverflow(This,prop) ) 

#define PowerPointSlideVerbosity_put_MeasurementUnits(This,prop)	\
    ( (This)->lpVtbl -> put_MeasurementUnits(This,prop) ) 

#define PowerPointSlideVerbosity_get_MeasurementUnits(This,prop)	\
    ( (This)->lpVtbl -> get_MeasurementUnits(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __PowerPointSlideVerbosity_INTERFACE_DEFINED__ */


#ifndef __PowerPointPlaybackVerbosity_INTERFACE_DEFINED__
#define __PowerPointPlaybackVerbosity_INTERFACE_DEFINED__

/* interface PowerPointPlaybackVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_PowerPointPlaybackVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("60E556DC-75DF-4E41-BAD5-825C9EA1B6EA")
    PowerPointPlaybackVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ResetPosition( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ResetPosition( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeNotes( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeNotes( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateEffects( 
            /* [in] */ PowerPointEffectsVerbosity prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateEffects( 
            /* [retval][out] */ PowerPointEffectsVerbosity *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct PowerPointPlaybackVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            PowerPointPlaybackVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            PowerPointPlaybackVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            PowerPointPlaybackVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            PowerPointPlaybackVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            PowerPointPlaybackVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            PowerPointPlaybackVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            PowerPointPlaybackVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            PowerPointPlaybackVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            PowerPointPlaybackVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ResetPosition )( 
            PowerPointPlaybackVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ResetPosition )( 
            PowerPointPlaybackVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeNotes )( 
            PowerPointPlaybackVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeNotes )( 
            PowerPointPlaybackVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateEffects )( 
            PowerPointPlaybackVerbosity * This,
            /* [in] */ PowerPointEffectsVerbosity prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateEffects )( 
            PowerPointPlaybackVerbosity * This,
            /* [retval][out] */ PowerPointEffectsVerbosity *prop);
        
        END_INTERFACE
    } PowerPointPlaybackVerbosityVtbl;

    interface PowerPointPlaybackVerbosity
    {
        CONST_VTBL struct PowerPointPlaybackVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define PowerPointPlaybackVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define PowerPointPlaybackVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define PowerPointPlaybackVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define PowerPointPlaybackVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define PowerPointPlaybackVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define PowerPointPlaybackVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define PowerPointPlaybackVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define PowerPointPlaybackVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define PowerPointPlaybackVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define PowerPointPlaybackVerbosity_put_ResetPosition(This,prop)	\
    ( (This)->lpVtbl -> put_ResetPosition(This,prop) ) 

#define PowerPointPlaybackVerbosity_get_ResetPosition(This,prop)	\
    ( (This)->lpVtbl -> get_ResetPosition(This,prop) ) 

#define PowerPointPlaybackVerbosity_put_IncludeNotes(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeNotes(This,prop) ) 

#define PowerPointPlaybackVerbosity_get_IncludeNotes(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeNotes(This,prop) ) 

#define PowerPointPlaybackVerbosity_put_IndicateEffects(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateEffects(This,prop) ) 

#define PowerPointPlaybackVerbosity_get_IndicateEffects(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateEffects(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __PowerPointPlaybackVerbosity_INTERFACE_DEFINED__ */


#ifndef __MouseVerbosity_INTERFACE_DEFINED__
#define __MouseVerbosity_INTERFACE_DEFINED__

/* interface MouseVerbosity */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_MouseVerbosity;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("827DE804-EF0C-40C1-938C-85DE3FE974F4")
    MouseVerbosity : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_In( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_In( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Out( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Out( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ClipChange( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ClipChange( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_LineChange( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LineChange( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_WindowChange( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WindowChange( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct MouseVerbosityVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            MouseVerbosity * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            MouseVerbosity * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            MouseVerbosity * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            MouseVerbosity * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            MouseVerbosity * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            MouseVerbosity * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            MouseVerbosity * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            MouseVerbosity * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            MouseVerbosity * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_In )( 
            MouseVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_In )( 
            MouseVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Out )( 
            MouseVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Out )( 
            MouseVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ClipChange )( 
            MouseVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ClipChange )( 
            MouseVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_LineChange )( 
            MouseVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LineChange )( 
            MouseVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_WindowChange )( 
            MouseVerbosity * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WindowChange )( 
            MouseVerbosity * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } MouseVerbosityVtbl;

    interface MouseVerbosity
    {
        CONST_VTBL struct MouseVerbosityVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define MouseVerbosity_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define MouseVerbosity_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define MouseVerbosity_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define MouseVerbosity_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define MouseVerbosity_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define MouseVerbosity_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define MouseVerbosity_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define MouseVerbosity_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define MouseVerbosity_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define MouseVerbosity_put_In(This,prop)	\
    ( (This)->lpVtbl -> put_In(This,prop) ) 

#define MouseVerbosity_get_In(This,prop)	\
    ( (This)->lpVtbl -> get_In(This,prop) ) 

#define MouseVerbosity_put_Out(This,prop)	\
    ( (This)->lpVtbl -> put_Out(This,prop) ) 

#define MouseVerbosity_get_Out(This,prop)	\
    ( (This)->lpVtbl -> get_Out(This,prop) ) 

#define MouseVerbosity_put_ClipChange(This,prop)	\
    ( (This)->lpVtbl -> put_ClipChange(This,prop) ) 

#define MouseVerbosity_get_ClipChange(This,prop)	\
    ( (This)->lpVtbl -> get_ClipChange(This,prop) ) 

#define MouseVerbosity_put_LineChange(This,prop)	\
    ( (This)->lpVtbl -> put_LineChange(This,prop) ) 

#define MouseVerbosity_get_LineChange(This,prop)	\
    ( (This)->lpVtbl -> get_LineChange(This,prop) ) 

#define MouseVerbosity_put_WindowChange(This,prop)	\
    ( (This)->lpVtbl -> put_WindowChange(This,prop) ) 

#define MouseVerbosity_get_WindowChange(This,prop)	\
    ( (This)->lpVtbl -> get_WindowChange(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __MouseVerbosity_INTERFACE_DEFINED__ */


#ifndef __BrailleSettings_INTERFACE_DEFINED__
#define __BrailleSettings_INTERFACE_DEFINED__

/* interface BrailleSettings */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleSettings;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("E2C6826E-B450-4CF9-A217-55B671499E7F")
    BrailleSettings : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Scrolling( 
            /* [retval][out] */ BrailleScrollingOptions **Scrolling) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ControlInformation( 
            /* [retval][out] */ BrailleControlInformation **controlInfo) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DotPatterns( 
            /* [retval][out] */ BrailleDotPatterns **patterns) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TranslationTables( 
            /* [retval][out] */ BrailleTables **Table) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Options( 
            /* [retval][out] */ BrailleOptions **Options) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Graphics( 
            /* [retval][out] */ BrailleGraphics **Graphics) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BrailleWindow( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_BrailleWindow( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleSettingsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleSettings * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleSettings * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleSettings * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleSettings * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleSettings * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleSettings * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleSettings * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleSettings * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleSettings * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Scrolling )( 
            BrailleSettings * This,
            /* [retval][out] */ BrailleScrollingOptions **Scrolling);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ControlInformation )( 
            BrailleSettings * This,
            /* [retval][out] */ BrailleControlInformation **controlInfo);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DotPatterns )( 
            BrailleSettings * This,
            /* [retval][out] */ BrailleDotPatterns **patterns);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TranslationTables )( 
            BrailleSettings * This,
            /* [retval][out] */ BrailleTables **Table);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Options )( 
            BrailleSettings * This,
            /* [retval][out] */ BrailleOptions **Options);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Graphics )( 
            BrailleSettings * This,
            /* [retval][out] */ BrailleGraphics **Graphics);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BrailleWindow )( 
            BrailleSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_BrailleWindow )( 
            BrailleSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        END_INTERFACE
    } BrailleSettingsVtbl;

    interface BrailleSettings
    {
        CONST_VTBL struct BrailleSettingsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleSettings_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleSettings_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleSettings_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleSettings_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleSettings_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleSettings_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleSettings_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleSettings_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleSettings_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleSettings_get_Scrolling(This,Scrolling)	\
    ( (This)->lpVtbl -> get_Scrolling(This,Scrolling) ) 

#define BrailleSettings_get_ControlInformation(This,controlInfo)	\
    ( (This)->lpVtbl -> get_ControlInformation(This,controlInfo) ) 

#define BrailleSettings_get_DotPatterns(This,patterns)	\
    ( (This)->lpVtbl -> get_DotPatterns(This,patterns) ) 

#define BrailleSettings_get_TranslationTables(This,Table)	\
    ( (This)->lpVtbl -> get_TranslationTables(This,Table) ) 

#define BrailleSettings_get_Options(This,Options)	\
    ( (This)->lpVtbl -> get_Options(This,Options) ) 

#define BrailleSettings_get_Graphics(This,Graphics)	\
    ( (This)->lpVtbl -> get_Graphics(This,Graphics) ) 

#define BrailleSettings_get_BrailleWindow(This,prop)	\
    ( (This)->lpVtbl -> get_BrailleWindow(This,prop) ) 

#define BrailleSettings_put_BrailleWindow(This,prop)	\
    ( (This)->lpVtbl -> put_BrailleWindow(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleSettings_INTERFACE_DEFINED__ */


#ifndef __BrailleScrollingOptions_INTERFACE_DEFINED__
#define __BrailleScrollingOptions_INTERFACE_DEFINED__

/* interface BrailleScrollingOptions */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleScrollingOptions;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("9131D53A-78D6-45EA-831E-B9D2B910DB5C")
    BrailleScrollingOptions : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AutoRouteCursor( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AutoRouteCursor( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WholeWordMode( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_WholeWordMode( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ScrollDisplayWidth( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ScrollDisplayWidth( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpecificWidth( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpecificWidth( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateLineChange( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateLineChange( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleScrollingOptionsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleScrollingOptions * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleScrollingOptions * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleScrollingOptions * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleScrollingOptions * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleScrollingOptions * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleScrollingOptions * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleScrollingOptions * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleScrollingOptions * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleScrollingOptions * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AutoRouteCursor )( 
            BrailleScrollingOptions * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AutoRouteCursor )( 
            BrailleScrollingOptions * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WholeWordMode )( 
            BrailleScrollingOptions * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_WholeWordMode )( 
            BrailleScrollingOptions * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ScrollDisplayWidth )( 
            BrailleScrollingOptions * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ScrollDisplayWidth )( 
            BrailleScrollingOptions * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpecificWidth )( 
            BrailleScrollingOptions * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpecificWidth )( 
            BrailleScrollingOptions * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateLineChange )( 
            BrailleScrollingOptions * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateLineChange )( 
            BrailleScrollingOptions * This,
            /* [in] */ VARIANT_BOOL prop);
        
        END_INTERFACE
    } BrailleScrollingOptionsVtbl;

    interface BrailleScrollingOptions
    {
        CONST_VTBL struct BrailleScrollingOptionsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleScrollingOptions_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleScrollingOptions_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleScrollingOptions_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleScrollingOptions_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleScrollingOptions_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleScrollingOptions_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleScrollingOptions_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleScrollingOptions_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleScrollingOptions_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleScrollingOptions_get_AutoRouteCursor(This,prop)	\
    ( (This)->lpVtbl -> get_AutoRouteCursor(This,prop) ) 

#define BrailleScrollingOptions_put_AutoRouteCursor(This,prop)	\
    ( (This)->lpVtbl -> put_AutoRouteCursor(This,prop) ) 

#define BrailleScrollingOptions_get_WholeWordMode(This,prop)	\
    ( (This)->lpVtbl -> get_WholeWordMode(This,prop) ) 

#define BrailleScrollingOptions_put_WholeWordMode(This,prop)	\
    ( (This)->lpVtbl -> put_WholeWordMode(This,prop) ) 

#define BrailleScrollingOptions_get_ScrollDisplayWidth(This,prop)	\
    ( (This)->lpVtbl -> get_ScrollDisplayWidth(This,prop) ) 

#define BrailleScrollingOptions_put_ScrollDisplayWidth(This,prop)	\
    ( (This)->lpVtbl -> put_ScrollDisplayWidth(This,prop) ) 

#define BrailleScrollingOptions_get_SpecificWidth(This,prop)	\
    ( (This)->lpVtbl -> get_SpecificWidth(This,prop) ) 

#define BrailleScrollingOptions_put_SpecificWidth(This,prop)	\
    ( (This)->lpVtbl -> put_SpecificWidth(This,prop) ) 

#define BrailleScrollingOptions_get_IndicateLineChange(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateLineChange(This,prop) ) 

#define BrailleScrollingOptions_put_IndicateLineChange(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateLineChange(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleScrollingOptions_INTERFACE_DEFINED__ */


#ifndef __BrailleControlInformation_INTERFACE_DEFINED__
#define __BrailleControlInformation_INTERFACE_DEFINED__

/* interface BrailleControlInformation */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleControlInformation;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("D36E82D4-E54F-44FA-BBC5-E807B4283724")
    BrailleControlInformation : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_InformationOrder( 
            /* [retval][out] */ BrailleInformationOrder *order) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_InformationOrder( 
            /* [in] */ BrailleInformationOrder order) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeName( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeName( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeType( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeType( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeIndex( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeIndex( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SeparatorDotPattern( 
            /* [retval][out] */ BSTR *pattern) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SeparatorDotPattern( 
            /* [in] */ BSTR pattern) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Types( 
            /* [retval][out] */ BrailleControlTypes **Types) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpeechBoxModeCharacterLimit( 
            /* [retval][out] */ long *limit) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpeechBoxModeCharacterLimit( 
            /* [in] */ long limit) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_LineModeIncludeDataIndicators( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LineModeIncludeDataIndicators( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_LineModeLeftIndicatorDotPattern( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LineModeLeftIndicatorDotPattern( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_LineModeRightIndicatorDotPattern( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LineModeRightIndicatorDotPattern( 
            /* [retval][out] */ BSTR *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleControlInformationVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleControlInformation * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleControlInformation * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleControlInformation * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleControlInformation * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleControlInformation * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleControlInformation * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleControlInformation * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleControlInformation * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleControlInformation * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_InformationOrder )( 
            BrailleControlInformation * This,
            /* [retval][out] */ BrailleInformationOrder *order);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_InformationOrder )( 
            BrailleControlInformation * This,
            /* [in] */ BrailleInformationOrder order);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeName )( 
            BrailleControlInformation * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeName )( 
            BrailleControlInformation * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeType )( 
            BrailleControlInformation * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeType )( 
            BrailleControlInformation * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeIndex )( 
            BrailleControlInformation * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeIndex )( 
            BrailleControlInformation * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SeparatorDotPattern )( 
            BrailleControlInformation * This,
            /* [retval][out] */ BSTR *pattern);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SeparatorDotPattern )( 
            BrailleControlInformation * This,
            /* [in] */ BSTR pattern);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Types )( 
            BrailleControlInformation * This,
            /* [retval][out] */ BrailleControlTypes **Types);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpeechBoxModeCharacterLimit )( 
            BrailleControlInformation * This,
            /* [retval][out] */ long *limit);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpeechBoxModeCharacterLimit )( 
            BrailleControlInformation * This,
            /* [in] */ long limit);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_LineModeIncludeDataIndicators )( 
            BrailleControlInformation * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LineModeIncludeDataIndicators )( 
            BrailleControlInformation * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_LineModeLeftIndicatorDotPattern )( 
            BrailleControlInformation * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LineModeLeftIndicatorDotPattern )( 
            BrailleControlInformation * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_LineModeRightIndicatorDotPattern )( 
            BrailleControlInformation * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LineModeRightIndicatorDotPattern )( 
            BrailleControlInformation * This,
            /* [retval][out] */ BSTR *prop);
        
        END_INTERFACE
    } BrailleControlInformationVtbl;

    interface BrailleControlInformation
    {
        CONST_VTBL struct BrailleControlInformationVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleControlInformation_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleControlInformation_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleControlInformation_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleControlInformation_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleControlInformation_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleControlInformation_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleControlInformation_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleControlInformation_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleControlInformation_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleControlInformation_get_InformationOrder(This,order)	\
    ( (This)->lpVtbl -> get_InformationOrder(This,order) ) 

#define BrailleControlInformation_put_InformationOrder(This,order)	\
    ( (This)->lpVtbl -> put_InformationOrder(This,order) ) 

#define BrailleControlInformation_get_IncludeName(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeName(This,prop) ) 

#define BrailleControlInformation_put_IncludeName(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeName(This,prop) ) 

#define BrailleControlInformation_get_IncludeType(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeType(This,prop) ) 

#define BrailleControlInformation_put_IncludeType(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeType(This,prop) ) 

#define BrailleControlInformation_get_IncludeIndex(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeIndex(This,prop) ) 

#define BrailleControlInformation_put_IncludeIndex(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeIndex(This,prop) ) 

#define BrailleControlInformation_get_SeparatorDotPattern(This,pattern)	\
    ( (This)->lpVtbl -> get_SeparatorDotPattern(This,pattern) ) 

#define BrailleControlInformation_put_SeparatorDotPattern(This,pattern)	\
    ( (This)->lpVtbl -> put_SeparatorDotPattern(This,pattern) ) 

#define BrailleControlInformation_get_Types(This,Types)	\
    ( (This)->lpVtbl -> get_Types(This,Types) ) 

#define BrailleControlInformation_get_SpeechBoxModeCharacterLimit(This,limit)	\
    ( (This)->lpVtbl -> get_SpeechBoxModeCharacterLimit(This,limit) ) 

#define BrailleControlInformation_put_SpeechBoxModeCharacterLimit(This,limit)	\
    ( (This)->lpVtbl -> put_SpeechBoxModeCharacterLimit(This,limit) ) 

#define BrailleControlInformation_put_LineModeIncludeDataIndicators(This,prop)	\
    ( (This)->lpVtbl -> put_LineModeIncludeDataIndicators(This,prop) ) 

#define BrailleControlInformation_get_LineModeIncludeDataIndicators(This,prop)	\
    ( (This)->lpVtbl -> get_LineModeIncludeDataIndicators(This,prop) ) 

#define BrailleControlInformation_put_LineModeLeftIndicatorDotPattern(This,prop)	\
    ( (This)->lpVtbl -> put_LineModeLeftIndicatorDotPattern(This,prop) ) 

#define BrailleControlInformation_get_LineModeLeftIndicatorDotPattern(This,prop)	\
    ( (This)->lpVtbl -> get_LineModeLeftIndicatorDotPattern(This,prop) ) 

#define BrailleControlInformation_put_LineModeRightIndicatorDotPattern(This,prop)	\
    ( (This)->lpVtbl -> put_LineModeRightIndicatorDotPattern(This,prop) ) 

#define BrailleControlInformation_get_LineModeRightIndicatorDotPattern(This,prop)	\
    ( (This)->lpVtbl -> get_LineModeRightIndicatorDotPattern(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleControlInformation_INTERFACE_DEFINED__ */


#ifndef __BrailleControlTypes_INTERFACE_DEFINED__
#define __BrailleControlTypes_INTERFACE_DEFINED__

/* interface BrailleControlTypes */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleControlTypes;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("9FE95569-2C11-4E4E-935F-AAB4C342D3F5")
    BrailleControlTypes : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ BrailleControlTypeEnum index,
            /* [retval][out] */ BrailleControlType **type) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleControlTypesVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleControlTypes * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleControlTypes * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleControlTypes * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleControlTypes * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleControlTypes * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleControlTypes * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleControlTypes * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleControlTypes * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleControlTypes * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            BrailleControlTypes * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            BrailleControlTypes * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            BrailleControlTypes * This,
            /* [in] */ BrailleControlTypeEnum index,
            /* [retval][out] */ BrailleControlType **type);
        
        END_INTERFACE
    } BrailleControlTypesVtbl;

    interface BrailleControlTypes
    {
        CONST_VTBL struct BrailleControlTypesVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleControlTypes_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleControlTypes_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleControlTypes_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleControlTypes_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleControlTypes_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleControlTypes_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleControlTypes_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleControlTypes_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleControlTypes_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleControlTypes_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define BrailleControlTypes_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define BrailleControlTypes_Item(This,index,type)	\
    ( (This)->lpVtbl -> Item(This,index,type) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleControlTypes_INTERFACE_DEFINED__ */


#ifndef __BrailleControlType_INTERFACE_DEFINED__
#define __BrailleControlType_INTERFACE_DEFINED__

/* interface BrailleControlType */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleControlType;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("8F452D32-49C5-427F-9668-2D77B9A259B2")
    BrailleControlType : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Name( 
            /* [retval][out] */ BSTR *typeName) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DisplayTypeIfCursorDetected( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_DisplayTypeIfCursorDetected( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indication( 
            /* [retval][out] */ BSTR *typeIndication) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indication( 
            /* [in] */ BSTR typeIndication) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_States( 
            /* [retval][out] */ BrailleControlStates **States) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleControlTypeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleControlType * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleControlType * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleControlType * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleControlType * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleControlType * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleControlType * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleControlType * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleControlType * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleControlType * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Name )( 
            BrailleControlType * This,
            /* [retval][out] */ BSTR *typeName);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DisplayTypeIfCursorDetected )( 
            BrailleControlType * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_DisplayTypeIfCursorDetected )( 
            BrailleControlType * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indication )( 
            BrailleControlType * This,
            /* [retval][out] */ BSTR *typeIndication);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indication )( 
            BrailleControlType * This,
            /* [in] */ BSTR typeIndication);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_States )( 
            BrailleControlType * This,
            /* [retval][out] */ BrailleControlStates **States);
        
        END_INTERFACE
    } BrailleControlTypeVtbl;

    interface BrailleControlType
    {
        CONST_VTBL struct BrailleControlTypeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleControlType_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleControlType_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleControlType_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleControlType_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleControlType_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleControlType_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleControlType_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleControlType_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleControlType_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleControlType_get_Name(This,typeName)	\
    ( (This)->lpVtbl -> get_Name(This,typeName) ) 

#define BrailleControlType_get_DisplayTypeIfCursorDetected(This,prop)	\
    ( (This)->lpVtbl -> get_DisplayTypeIfCursorDetected(This,prop) ) 

#define BrailleControlType_put_DisplayTypeIfCursorDetected(This,prop)	\
    ( (This)->lpVtbl -> put_DisplayTypeIfCursorDetected(This,prop) ) 

#define BrailleControlType_get_Indication(This,typeIndication)	\
    ( (This)->lpVtbl -> get_Indication(This,typeIndication) ) 

#define BrailleControlType_put_Indication(This,typeIndication)	\
    ( (This)->lpVtbl -> put_Indication(This,typeIndication) ) 

#define BrailleControlType_get_States(This,States)	\
    ( (This)->lpVtbl -> get_States(This,States) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleControlType_INTERFACE_DEFINED__ */


#ifndef __BrailleControlStates_INTERFACE_DEFINED__
#define __BrailleControlStates_INTERFACE_DEFINED__

/* interface BrailleControlStates */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleControlStates;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("79936E4D-6993-420F-B6AB-88EC509B406F")
    BrailleControlStates : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ BrailleControlState **state) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleControlStatesVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleControlStates * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleControlStates * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleControlStates * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleControlStates * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleControlStates * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleControlStates * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleControlStates * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleControlStates * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleControlStates * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            BrailleControlStates * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            BrailleControlStates * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            BrailleControlStates * This,
            /* [in] */ long index,
            /* [retval][out] */ BrailleControlState **state);
        
        END_INTERFACE
    } BrailleControlStatesVtbl;

    interface BrailleControlStates
    {
        CONST_VTBL struct BrailleControlStatesVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleControlStates_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleControlStates_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleControlStates_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleControlStates_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleControlStates_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleControlStates_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleControlStates_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleControlStates_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleControlStates_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleControlStates_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define BrailleControlStates_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define BrailleControlStates_Item(This,index,state)	\
    ( (This)->lpVtbl -> Item(This,index,state) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleControlStates_INTERFACE_DEFINED__ */


#ifndef __BrailleControlState_INTERFACE_DEFINED__
#define __BrailleControlState_INTERFACE_DEFINED__

/* interface BrailleControlState */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleControlState;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("D0F53943-43F0-4979-A390-9CCB7A24DAA1")
    BrailleControlState : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Name( 
            /* [retval][out] */ BSTR *Name) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indication( 
            /* [retval][out] */ BSTR *Indication) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indication( 
            /* [in] */ BSTR Indication) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleControlStateVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleControlState * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleControlState * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleControlState * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleControlState * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleControlState * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleControlState * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleControlState * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleControlState * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleControlState * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Name )( 
            BrailleControlState * This,
            /* [retval][out] */ BSTR *Name);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indication )( 
            BrailleControlState * This,
            /* [retval][out] */ BSTR *Indication);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indication )( 
            BrailleControlState * This,
            /* [in] */ BSTR Indication);
        
        END_INTERFACE
    } BrailleControlStateVtbl;

    interface BrailleControlState
    {
        CONST_VTBL struct BrailleControlStateVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleControlState_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleControlState_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleControlState_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleControlState_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleControlState_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleControlState_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleControlState_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleControlState_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleControlState_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleControlState_get_Name(This,Name)	\
    ( (This)->lpVtbl -> get_Name(This,Name) ) 

#define BrailleControlState_get_Indication(This,Indication)	\
    ( (This)->lpVtbl -> get_Indication(This,Indication) ) 

#define BrailleControlState_put_Indication(This,Indication)	\
    ( (This)->lpVtbl -> put_Indication(This,Indication) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleControlState_INTERFACE_DEFINED__ */


#ifndef __BrailleDotPatterns_INTERFACE_DEFINED__
#define __BrailleDotPatterns_INTERFACE_DEFINED__

/* interface BrailleDotPatterns */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleDotPatterns;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("951F1617-0998-4D72-BD0D-976B65997322")
    BrailleDotPatterns : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Cursor( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Cursor( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CursorBlinkRate( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CursorBlinkRate( 
            /* [in] */ long prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Mouse( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Mouse( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MouseBlinkRate( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_MouseBlinkRate( 
            /* [in] */ long prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Bold( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Bold( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Highlight( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Highlight( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Italic( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Italic( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Strikeout( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Strikeout( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Underline( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Underline( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MouseBlinkStyle( 
            /* [retval][out] */ BraillePatternBlinkStyle *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_MouseBlinkStyle( 
            /* [in] */ BraillePatternBlinkStyle prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CursorBlinkStyle( 
            /* [retval][out] */ BraillePatternBlinkStyle *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CursorBlinkStyle( 
            /* [in] */ BraillePatternBlinkStyle prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleDotPatternsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleDotPatterns * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleDotPatterns * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleDotPatterns * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleDotPatterns * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleDotPatterns * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleDotPatterns * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleDotPatterns * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleDotPatterns * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleDotPatterns * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Cursor )( 
            BrailleDotPatterns * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Cursor )( 
            BrailleDotPatterns * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CursorBlinkRate )( 
            BrailleDotPatterns * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CursorBlinkRate )( 
            BrailleDotPatterns * This,
            /* [in] */ long prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Mouse )( 
            BrailleDotPatterns * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Mouse )( 
            BrailleDotPatterns * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MouseBlinkRate )( 
            BrailleDotPatterns * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_MouseBlinkRate )( 
            BrailleDotPatterns * This,
            /* [in] */ long prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Bold )( 
            BrailleDotPatterns * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Bold )( 
            BrailleDotPatterns * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Highlight )( 
            BrailleDotPatterns * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Highlight )( 
            BrailleDotPatterns * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Italic )( 
            BrailleDotPatterns * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Italic )( 
            BrailleDotPatterns * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Strikeout )( 
            BrailleDotPatterns * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Strikeout )( 
            BrailleDotPatterns * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Underline )( 
            BrailleDotPatterns * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Underline )( 
            BrailleDotPatterns * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MouseBlinkStyle )( 
            BrailleDotPatterns * This,
            /* [retval][out] */ BraillePatternBlinkStyle *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_MouseBlinkStyle )( 
            BrailleDotPatterns * This,
            /* [in] */ BraillePatternBlinkStyle prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CursorBlinkStyle )( 
            BrailleDotPatterns * This,
            /* [retval][out] */ BraillePatternBlinkStyle *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CursorBlinkStyle )( 
            BrailleDotPatterns * This,
            /* [in] */ BraillePatternBlinkStyle prop);
        
        END_INTERFACE
    } BrailleDotPatternsVtbl;

    interface BrailleDotPatterns
    {
        CONST_VTBL struct BrailleDotPatternsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleDotPatterns_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleDotPatterns_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleDotPatterns_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleDotPatterns_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleDotPatterns_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleDotPatterns_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleDotPatterns_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleDotPatterns_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleDotPatterns_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleDotPatterns_get_Cursor(This,prop)	\
    ( (This)->lpVtbl -> get_Cursor(This,prop) ) 

#define BrailleDotPatterns_put_Cursor(This,prop)	\
    ( (This)->lpVtbl -> put_Cursor(This,prop) ) 

#define BrailleDotPatterns_get_CursorBlinkRate(This,prop)	\
    ( (This)->lpVtbl -> get_CursorBlinkRate(This,prop) ) 

#define BrailleDotPatterns_put_CursorBlinkRate(This,prop)	\
    ( (This)->lpVtbl -> put_CursorBlinkRate(This,prop) ) 

#define BrailleDotPatterns_get_Mouse(This,prop)	\
    ( (This)->lpVtbl -> get_Mouse(This,prop) ) 

#define BrailleDotPatterns_put_Mouse(This,prop)	\
    ( (This)->lpVtbl -> put_Mouse(This,prop) ) 

#define BrailleDotPatterns_get_MouseBlinkRate(This,prop)	\
    ( (This)->lpVtbl -> get_MouseBlinkRate(This,prop) ) 

#define BrailleDotPatterns_put_MouseBlinkRate(This,prop)	\
    ( (This)->lpVtbl -> put_MouseBlinkRate(This,prop) ) 

#define BrailleDotPatterns_get_Bold(This,prop)	\
    ( (This)->lpVtbl -> get_Bold(This,prop) ) 

#define BrailleDotPatterns_put_Bold(This,prop)	\
    ( (This)->lpVtbl -> put_Bold(This,prop) ) 

#define BrailleDotPatterns_get_Highlight(This,prop)	\
    ( (This)->lpVtbl -> get_Highlight(This,prop) ) 

#define BrailleDotPatterns_put_Highlight(This,prop)	\
    ( (This)->lpVtbl -> put_Highlight(This,prop) ) 

#define BrailleDotPatterns_get_Italic(This,prop)	\
    ( (This)->lpVtbl -> get_Italic(This,prop) ) 

#define BrailleDotPatterns_put_Italic(This,prop)	\
    ( (This)->lpVtbl -> put_Italic(This,prop) ) 

#define BrailleDotPatterns_get_Strikeout(This,prop)	\
    ( (This)->lpVtbl -> get_Strikeout(This,prop) ) 

#define BrailleDotPatterns_put_Strikeout(This,prop)	\
    ( (This)->lpVtbl -> put_Strikeout(This,prop) ) 

#define BrailleDotPatterns_get_Underline(This,prop)	\
    ( (This)->lpVtbl -> get_Underline(This,prop) ) 

#define BrailleDotPatterns_put_Underline(This,prop)	\
    ( (This)->lpVtbl -> put_Underline(This,prop) ) 

#define BrailleDotPatterns_get_MouseBlinkStyle(This,prop)	\
    ( (This)->lpVtbl -> get_MouseBlinkStyle(This,prop) ) 

#define BrailleDotPatterns_put_MouseBlinkStyle(This,prop)	\
    ( (This)->lpVtbl -> put_MouseBlinkStyle(This,prop) ) 

#define BrailleDotPatterns_get_CursorBlinkStyle(This,prop)	\
    ( (This)->lpVtbl -> get_CursorBlinkStyle(This,prop) ) 

#define BrailleDotPatterns_put_CursorBlinkStyle(This,prop)	\
    ( (This)->lpVtbl -> put_CursorBlinkStyle(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleDotPatterns_INTERFACE_DEFINED__ */


#ifndef __BrailleTables_INTERFACE_DEFINED__
#define __BrailleTables_INTERFACE_DEFINED__

/* interface BrailleTables */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleTables;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("87657888-511D-4F43-BAC7-B13CD51131EC")
    BrailleTables : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Active( 
            /* [retval][out] */ BrailleTable **Table) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long tableID,
            /* [retval][out] */ BrailleTable **Table) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleTablesVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleTables * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleTables * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleTables * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleTables * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleTables * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleTables * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleTables * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleTables * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleTables * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            BrailleTables * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            BrailleTables * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Active )( 
            BrailleTables * This,
            /* [retval][out] */ BrailleTable **Table);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            BrailleTables * This,
            /* [in] */ long tableID,
            /* [retval][out] */ BrailleTable **Table);
        
        END_INTERFACE
    } BrailleTablesVtbl;

    interface BrailleTables
    {
        CONST_VTBL struct BrailleTablesVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleTables_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleTables_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleTables_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleTables_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleTables_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleTables_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleTables_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleTables_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleTables_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleTables_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define BrailleTables_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define BrailleTables_get_Active(This,Table)	\
    ( (This)->lpVtbl -> get_Active(This,Table) ) 

#define BrailleTables_Item(This,tableID,Table)	\
    ( (This)->lpVtbl -> Item(This,tableID,Table) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleTables_INTERFACE_DEFINED__ */


#ifndef __BrailleTable_INTERFACE_DEFINED__
#define __BrailleTable_INTERFACE_DEFINED__

/* interface BrailleTable */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleTable;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("0A5C485D-6719-4769-9D50-8A1E6EB1AD3E")
    BrailleTable : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadOnly( 
            /* [retval][out] */ VARIANT_BOOL *ReadOnly) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Name( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UnifiedEnglishBraille( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_UnifiedEnglishBraille( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ BrailleTableEntry **tableEntry) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Activate( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Add( 
            /* [in] */ long index,
            /* [in] */ BSTR DotPattern,
            /* [retval][out] */ BrailleTableEntry **tableEntry) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Delete( 
            /* [in] */ long index) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Filename( 
            /* [retval][out] */ BSTR *Filename) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleTableVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleTable * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleTable * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleTable * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleTable * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleTable * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleTable * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleTable * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleTable * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleTable * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadOnly )( 
            BrailleTable * This,
            /* [retval][out] */ VARIANT_BOOL *ReadOnly);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Name )( 
            BrailleTable * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            BrailleTable * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            BrailleTable * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UnifiedEnglishBraille )( 
            BrailleTable * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_UnifiedEnglishBraille )( 
            BrailleTable * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            BrailleTable * This,
            /* [in] */ long index,
            /* [retval][out] */ BrailleTableEntry **tableEntry);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Activate )( 
            BrailleTable * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Add )( 
            BrailleTable * This,
            /* [in] */ long index,
            /* [in] */ BSTR DotPattern,
            /* [retval][out] */ BrailleTableEntry **tableEntry);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Delete )( 
            BrailleTable * This,
            /* [in] */ long index);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Filename )( 
            BrailleTable * This,
            /* [retval][out] */ BSTR *Filename);
        
        END_INTERFACE
    } BrailleTableVtbl;

    interface BrailleTable
    {
        CONST_VTBL struct BrailleTableVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleTable_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleTable_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleTable_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleTable_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleTable_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleTable_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleTable_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleTable_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleTable_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleTable_get_ReadOnly(This,ReadOnly)	\
    ( (This)->lpVtbl -> get_ReadOnly(This,ReadOnly) ) 

#define BrailleTable_get_Name(This,prop)	\
    ( (This)->lpVtbl -> get_Name(This,prop) ) 

#define BrailleTable_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define BrailleTable_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define BrailleTable_get_UnifiedEnglishBraille(This,prop)	\
    ( (This)->lpVtbl -> get_UnifiedEnglishBraille(This,prop) ) 

#define BrailleTable_put_UnifiedEnglishBraille(This,prop)	\
    ( (This)->lpVtbl -> put_UnifiedEnglishBraille(This,prop) ) 

#define BrailleTable_Item(This,index,tableEntry)	\
    ( (This)->lpVtbl -> Item(This,index,tableEntry) ) 

#define BrailleTable_Activate(This)	\
    ( (This)->lpVtbl -> Activate(This) ) 

#define BrailleTable_Add(This,index,DotPattern,tableEntry)	\
    ( (This)->lpVtbl -> Add(This,index,DotPattern,tableEntry) ) 

#define BrailleTable_Delete(This,index)	\
    ( (This)->lpVtbl -> Delete(This,index) ) 

#define BrailleTable_get_Filename(This,Filename)	\
    ( (This)->lpVtbl -> get_Filename(This,Filename) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleTable_INTERFACE_DEFINED__ */


#ifndef __BrailleTableEntry_INTERFACE_DEFINED__
#define __BrailleTableEntry_INTERFACE_DEFINED__

/* interface BrailleTableEntry */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleTableEntry;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("5692B95A-BF10-451B-8FA8-1E49BFE47FE3")
    BrailleTableEntry : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_index( 
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DotPattern( 
            /* [retval][out] */ BSTR *pattern) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_DotPattern( 
            /* [in] */ BSTR pattern) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UnicodeBraille( 
            /* [retval][out] */ BSTR *pattern) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_UnicodeBraille( 
            /* [in] */ BSTR pattern) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleTableEntryVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleTableEntry * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleTableEntry * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleTableEntry * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleTableEntry * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleTableEntry * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleTableEntry * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleTableEntry * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleTableEntry * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleTableEntry * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_index )( 
            BrailleTableEntry * This,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DotPattern )( 
            BrailleTableEntry * This,
            /* [retval][out] */ BSTR *pattern);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_DotPattern )( 
            BrailleTableEntry * This,
            /* [in] */ BSTR pattern);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UnicodeBraille )( 
            BrailleTableEntry * This,
            /* [retval][out] */ BSTR *pattern);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_UnicodeBraille )( 
            BrailleTableEntry * This,
            /* [in] */ BSTR pattern);
        
        END_INTERFACE
    } BrailleTableEntryVtbl;

    interface BrailleTableEntry
    {
        CONST_VTBL struct BrailleTableEntryVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleTableEntry_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleTableEntry_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleTableEntry_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleTableEntry_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleTableEntry_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleTableEntry_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleTableEntry_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleTableEntry_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleTableEntry_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleTableEntry_get_index(This,index)	\
    ( (This)->lpVtbl -> get_index(This,index) ) 

#define BrailleTableEntry_get_DotPattern(This,pattern)	\
    ( (This)->lpVtbl -> get_DotPattern(This,pattern) ) 

#define BrailleTableEntry_put_DotPattern(This,pattern)	\
    ( (This)->lpVtbl -> put_DotPattern(This,pattern) ) 

#define BrailleTableEntry_get_UnicodeBraille(This,pattern)	\
    ( (This)->lpVtbl -> get_UnicodeBraille(This,pattern) ) 

#define BrailleTableEntry_put_UnicodeBraille(This,pattern)	\
    ( (This)->lpVtbl -> put_UnicodeBraille(This,pattern) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleTableEntry_INTERFACE_DEFINED__ */


#ifndef __BrailleOptions_INTERFACE_DEFINED__
#define __BrailleOptions_INTERFACE_DEFINED__

/* interface BrailleOptions */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleOptions;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("28E578E8-9600-4DAF-9BA9-6A241981957A")
    BrailleOptions : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Grade2Enabled( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Grade2Enabled( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Grade2AutoUntranslate( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Grade2AutoUntranslate( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Grade2TranslateSymbols( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Grade2TranslateSymbols( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EightDot( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EightDot( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AutoUpdateTimeout( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AutoUpdateTimeout( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AttributeMarking( 
            /* [retval][out] */ BrailleAttributeMarking *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AttributeMarking( 
            /* [in] */ BrailleAttributeMarking prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_QuickMessageMode( 
            /* [retval][out] */ BrailleQuickMessageMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_QuickMessageMode( 
            /* [in] */ BrailleQuickMessageMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_QuickMessageModeTimeout( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_QuickMessageModeTimeout( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ShowIndents( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ShowIndents( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ExtraSpacingPixels( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ExtraSpacingPixels( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EnableAttributeMarking( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EnableAttributeMarking( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ShowExtraSpacing( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ShowExtraSpacing( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateCapitalization( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateCapitalization( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleOptionsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleOptions * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleOptions * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleOptions * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleOptions * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleOptions * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleOptions * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleOptions * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleOptions * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleOptions * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Grade2Enabled )( 
            BrailleOptions * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Grade2Enabled )( 
            BrailleOptions * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Grade2AutoUntranslate )( 
            BrailleOptions * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Grade2AutoUntranslate )( 
            BrailleOptions * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Grade2TranslateSymbols )( 
            BrailleOptions * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Grade2TranslateSymbols )( 
            BrailleOptions * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EightDot )( 
            BrailleOptions * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EightDot )( 
            BrailleOptions * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AutoUpdateTimeout )( 
            BrailleOptions * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AutoUpdateTimeout )( 
            BrailleOptions * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AttributeMarking )( 
            BrailleOptions * This,
            /* [retval][out] */ BrailleAttributeMarking *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AttributeMarking )( 
            BrailleOptions * This,
            /* [in] */ BrailleAttributeMarking prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_QuickMessageMode )( 
            BrailleOptions * This,
            /* [retval][out] */ BrailleQuickMessageMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_QuickMessageMode )( 
            BrailleOptions * This,
            /* [in] */ BrailleQuickMessageMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_QuickMessageModeTimeout )( 
            BrailleOptions * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_QuickMessageModeTimeout )( 
            BrailleOptions * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ShowIndents )( 
            BrailleOptions * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ShowIndents )( 
            BrailleOptions * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ExtraSpacingPixels )( 
            BrailleOptions * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ExtraSpacingPixels )( 
            BrailleOptions * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EnableAttributeMarking )( 
            BrailleOptions * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EnableAttributeMarking )( 
            BrailleOptions * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ShowExtraSpacing )( 
            BrailleOptions * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ShowExtraSpacing )( 
            BrailleOptions * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateCapitalization )( 
            BrailleOptions * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateCapitalization )( 
            BrailleOptions * This,
            /* [in] */ VARIANT_BOOL prop);
        
        END_INTERFACE
    } BrailleOptionsVtbl;

    interface BrailleOptions
    {
        CONST_VTBL struct BrailleOptionsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleOptions_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleOptions_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleOptions_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleOptions_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleOptions_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleOptions_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleOptions_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleOptions_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleOptions_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleOptions_get_Grade2Enabled(This,prop)	\
    ( (This)->lpVtbl -> get_Grade2Enabled(This,prop) ) 

#define BrailleOptions_put_Grade2Enabled(This,prop)	\
    ( (This)->lpVtbl -> put_Grade2Enabled(This,prop) ) 

#define BrailleOptions_get_Grade2AutoUntranslate(This,prop)	\
    ( (This)->lpVtbl -> get_Grade2AutoUntranslate(This,prop) ) 

#define BrailleOptions_put_Grade2AutoUntranslate(This,prop)	\
    ( (This)->lpVtbl -> put_Grade2AutoUntranslate(This,prop) ) 

#define BrailleOptions_get_Grade2TranslateSymbols(This,prop)	\
    ( (This)->lpVtbl -> get_Grade2TranslateSymbols(This,prop) ) 

#define BrailleOptions_put_Grade2TranslateSymbols(This,prop)	\
    ( (This)->lpVtbl -> put_Grade2TranslateSymbols(This,prop) ) 

#define BrailleOptions_get_EightDot(This,prop)	\
    ( (This)->lpVtbl -> get_EightDot(This,prop) ) 

#define BrailleOptions_put_EightDot(This,prop)	\
    ( (This)->lpVtbl -> put_EightDot(This,prop) ) 

#define BrailleOptions_get_AutoUpdateTimeout(This,prop)	\
    ( (This)->lpVtbl -> get_AutoUpdateTimeout(This,prop) ) 

#define BrailleOptions_put_AutoUpdateTimeout(This,prop)	\
    ( (This)->lpVtbl -> put_AutoUpdateTimeout(This,prop) ) 

#define BrailleOptions_get_AttributeMarking(This,prop)	\
    ( (This)->lpVtbl -> get_AttributeMarking(This,prop) ) 

#define BrailleOptions_put_AttributeMarking(This,prop)	\
    ( (This)->lpVtbl -> put_AttributeMarking(This,prop) ) 

#define BrailleOptions_get_QuickMessageMode(This,prop)	\
    ( (This)->lpVtbl -> get_QuickMessageMode(This,prop) ) 

#define BrailleOptions_put_QuickMessageMode(This,prop)	\
    ( (This)->lpVtbl -> put_QuickMessageMode(This,prop) ) 

#define BrailleOptions_get_QuickMessageModeTimeout(This,prop)	\
    ( (This)->lpVtbl -> get_QuickMessageModeTimeout(This,prop) ) 

#define BrailleOptions_put_QuickMessageModeTimeout(This,prop)	\
    ( (This)->lpVtbl -> put_QuickMessageModeTimeout(This,prop) ) 

#define BrailleOptions_get_ShowIndents(This,prop)	\
    ( (This)->lpVtbl -> get_ShowIndents(This,prop) ) 

#define BrailleOptions_put_ShowIndents(This,prop)	\
    ( (This)->lpVtbl -> put_ShowIndents(This,prop) ) 

#define BrailleOptions_get_ExtraSpacingPixels(This,prop)	\
    ( (This)->lpVtbl -> get_ExtraSpacingPixels(This,prop) ) 

#define BrailleOptions_put_ExtraSpacingPixels(This,prop)	\
    ( (This)->lpVtbl -> put_ExtraSpacingPixels(This,prop) ) 

#define BrailleOptions_get_EnableAttributeMarking(This,prop)	\
    ( (This)->lpVtbl -> get_EnableAttributeMarking(This,prop) ) 

#define BrailleOptions_put_EnableAttributeMarking(This,prop)	\
    ( (This)->lpVtbl -> put_EnableAttributeMarking(This,prop) ) 

#define BrailleOptions_get_ShowExtraSpacing(This,prop)	\
    ( (This)->lpVtbl -> get_ShowExtraSpacing(This,prop) ) 

#define BrailleOptions_put_ShowExtraSpacing(This,prop)	\
    ( (This)->lpVtbl -> put_ShowExtraSpacing(This,prop) ) 

#define BrailleOptions_get_IndicateCapitalization(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateCapitalization(This,prop) ) 

#define BrailleOptions_put_IndicateCapitalization(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateCapitalization(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleOptions_INTERFACE_DEFINED__ */


#ifndef __BrailleGraphics_INTERFACE_DEFINED__
#define __BrailleGraphics_INTERFACE_DEFINED__

/* interface BrailleGraphics */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleGraphics;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("D88BA939-CD75-46D8-8114-8E65DAB44357")
    BrailleGraphics : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Verbosity( 
            /* [retval][out] */ BrailleGraphicsVerbosity *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Verbosity( 
            /* [in] */ BrailleGraphicsVerbosity prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Symbol( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Symbol( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UseSeparator( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_UseSeparator( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleGraphicsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleGraphics * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleGraphics * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleGraphics * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleGraphics * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleGraphics * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleGraphics * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleGraphics * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleGraphics * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleGraphics * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Verbosity )( 
            BrailleGraphics * This,
            /* [retval][out] */ BrailleGraphicsVerbosity *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Verbosity )( 
            BrailleGraphics * This,
            /* [in] */ BrailleGraphicsVerbosity prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Symbol )( 
            BrailleGraphics * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Symbol )( 
            BrailleGraphics * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UseSeparator )( 
            BrailleGraphics * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_UseSeparator )( 
            BrailleGraphics * This,
            /* [in] */ VARIANT_BOOL prop);
        
        END_INTERFACE
    } BrailleGraphicsVtbl;

    interface BrailleGraphics
    {
        CONST_VTBL struct BrailleGraphicsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleGraphics_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleGraphics_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleGraphics_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleGraphics_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleGraphics_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleGraphics_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleGraphics_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleGraphics_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleGraphics_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleGraphics_get_Verbosity(This,prop)	\
    ( (This)->lpVtbl -> get_Verbosity(This,prop) ) 

#define BrailleGraphics_put_Verbosity(This,prop)	\
    ( (This)->lpVtbl -> put_Verbosity(This,prop) ) 

#define BrailleGraphics_get_Symbol(This,prop)	\
    ( (This)->lpVtbl -> get_Symbol(This,prop) ) 

#define BrailleGraphics_put_Symbol(This,prop)	\
    ( (This)->lpVtbl -> put_Symbol(This,prop) ) 

#define BrailleGraphics_get_UseSeparator(This,prop)	\
    ( (This)->lpVtbl -> get_UseSeparator(This,prop) ) 

#define BrailleGraphics_put_UseSeparator(This,prop)	\
    ( (This)->lpVtbl -> put_UseSeparator(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleGraphics_INTERFACE_DEFINED__ */


#ifndef __UserWindows_INTERFACE_DEFINED__
#define __UserWindows_INTERFACE_DEFINED__

/* interface UserWindows */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_UserWindows;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("B1AB92CC-3C71-4A77-8FCF-12B3A71C5B09")
    UserWindows : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long userWinID,
            /* [retval][out] */ UserWindow **UserWindow) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct UserWindowsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            UserWindows * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            UserWindows * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            UserWindows * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            UserWindows * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            UserWindows * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            UserWindows * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            UserWindows * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            UserWindows * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            UserWindows * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            UserWindows * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            UserWindows * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            UserWindows * This,
            /* [in] */ long userWinID,
            /* [retval][out] */ UserWindow **UserWindow);
        
        END_INTERFACE
    } UserWindowsVtbl;

    interface UserWindows
    {
        CONST_VTBL struct UserWindowsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define UserWindows_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define UserWindows_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define UserWindows_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define UserWindows_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define UserWindows_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define UserWindows_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define UserWindows_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define UserWindows_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define UserWindows_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define UserWindows_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define UserWindows_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define UserWindows_Item(This,userWinID,UserWindow)	\
    ( (This)->lpVtbl -> Item(This,userWinID,UserWindow) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __UserWindows_INTERFACE_DEFINED__ */


#ifndef __UserWindow_INTERFACE_DEFINED__
#define __UserWindow_INTERFACE_DEFINED__

/* interface UserWindow */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_UserWindow;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("5CEF409D-BA9F-4B8A-BEFB-7C7A27997550")
    UserWindow : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Coordinates( 
            /* [retval][out] */ UserWindowCoordinates **coords) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Status( 
            /* [retval][out] */ UserWindowStatus *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Status( 
            /* [in] */ UserWindowStatus prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Comment( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Comment( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_MatchAnyForegroundColor( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MatchAnyForegroundColor( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ForegroundColor( 
            /* [retval][out] */ Color **Color) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ForegroundColor( 
            /* [in] */ Color *Color) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_MatchAnyBackgroundColor( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MatchAnyBackgroundColor( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BackgroundColor( 
            /* [retval][out] */ Color **Color) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_BackgroundColor( 
            /* [in] */ Color *Color) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_WindowToChainRead( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WindowToChainRead( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpeakAndSpell( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpeakAndSpell( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HighlightForegroundColor( 
            /* [retval][out] */ Color **Color) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_HighlightForegroundColor( 
            /* [in] */ Color *Color) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HighlightBackgroundColor( 
            /* [retval][out] */ Color **Color) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_HighlightBackgroundColor( 
            /* [in] */ Color *Color) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AutoDetermineHighlightColor( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AutoDetermineHighlightColor( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HighlightStatus( 
            /* [retval][out] */ AutoMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_HighlightStatus( 
            /* [in] */ AutoMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_rectangle( 
            /* [retval][out] */ WindowRectangle **WindowRectangle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ConfigurationDialog( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Speak( void) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct UserWindowVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            UserWindow * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            UserWindow * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            UserWindow * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            UserWindow * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            UserWindow * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            UserWindow * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            UserWindow * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            UserWindow * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            UserWindow * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Coordinates )( 
            UserWindow * This,
            /* [retval][out] */ UserWindowCoordinates **coords);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Status )( 
            UserWindow * This,
            /* [retval][out] */ UserWindowStatus *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Status )( 
            UserWindow * This,
            /* [in] */ UserWindowStatus prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Comment )( 
            UserWindow * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Comment )( 
            UserWindow * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_MatchAnyForegroundColor )( 
            UserWindow * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MatchAnyForegroundColor )( 
            UserWindow * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ForegroundColor )( 
            UserWindow * This,
            /* [retval][out] */ Color **Color);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ForegroundColor )( 
            UserWindow * This,
            /* [in] */ Color *Color);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_MatchAnyBackgroundColor )( 
            UserWindow * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MatchAnyBackgroundColor )( 
            UserWindow * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BackgroundColor )( 
            UserWindow * This,
            /* [retval][out] */ Color **Color);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_BackgroundColor )( 
            UserWindow * This,
            /* [in] */ Color *Color);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_WindowToChainRead )( 
            UserWindow * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WindowToChainRead )( 
            UserWindow * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpeakAndSpell )( 
            UserWindow * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpeakAndSpell )( 
            UserWindow * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HighlightForegroundColor )( 
            UserWindow * This,
            /* [retval][out] */ Color **Color);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_HighlightForegroundColor )( 
            UserWindow * This,
            /* [in] */ Color *Color);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HighlightBackgroundColor )( 
            UserWindow * This,
            /* [retval][out] */ Color **Color);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_HighlightBackgroundColor )( 
            UserWindow * This,
            /* [in] */ Color *Color);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AutoDetermineHighlightColor )( 
            UserWindow * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AutoDetermineHighlightColor )( 
            UserWindow * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HighlightStatus )( 
            UserWindow * This,
            /* [retval][out] */ AutoMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_HighlightStatus )( 
            UserWindow * This,
            /* [in] */ AutoMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_rectangle )( 
            UserWindow * This,
            /* [retval][out] */ WindowRectangle **WindowRectangle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ConfigurationDialog )( 
            UserWindow * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Speak )( 
            UserWindow * This);
        
        END_INTERFACE
    } UserWindowVtbl;

    interface UserWindow
    {
        CONST_VTBL struct UserWindowVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define UserWindow_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define UserWindow_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define UserWindow_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define UserWindow_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define UserWindow_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define UserWindow_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define UserWindow_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define UserWindow_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define UserWindow_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define UserWindow_get_Coordinates(This,coords)	\
    ( (This)->lpVtbl -> get_Coordinates(This,coords) ) 

#define UserWindow_get_Status(This,prop)	\
    ( (This)->lpVtbl -> get_Status(This,prop) ) 

#define UserWindow_put_Status(This,prop)	\
    ( (This)->lpVtbl -> put_Status(This,prop) ) 

#define UserWindow_get_Comment(This,prop)	\
    ( (This)->lpVtbl -> get_Comment(This,prop) ) 

#define UserWindow_put_Comment(This,prop)	\
    ( (This)->lpVtbl -> put_Comment(This,prop) ) 

#define UserWindow_put_MatchAnyForegroundColor(This,prop)	\
    ( (This)->lpVtbl -> put_MatchAnyForegroundColor(This,prop) ) 

#define UserWindow_get_MatchAnyForegroundColor(This,prop)	\
    ( (This)->lpVtbl -> get_MatchAnyForegroundColor(This,prop) ) 

#define UserWindow_get_ForegroundColor(This,Color)	\
    ( (This)->lpVtbl -> get_ForegroundColor(This,Color) ) 

#define UserWindow_put_ForegroundColor(This,Color)	\
    ( (This)->lpVtbl -> put_ForegroundColor(This,Color) ) 

#define UserWindow_put_MatchAnyBackgroundColor(This,prop)	\
    ( (This)->lpVtbl -> put_MatchAnyBackgroundColor(This,prop) ) 

#define UserWindow_get_MatchAnyBackgroundColor(This,prop)	\
    ( (This)->lpVtbl -> get_MatchAnyBackgroundColor(This,prop) ) 

#define UserWindow_get_BackgroundColor(This,Color)	\
    ( (This)->lpVtbl -> get_BackgroundColor(This,Color) ) 

#define UserWindow_put_BackgroundColor(This,Color)	\
    ( (This)->lpVtbl -> put_BackgroundColor(This,Color) ) 

#define UserWindow_put_WindowToChainRead(This,prop)	\
    ( (This)->lpVtbl -> put_WindowToChainRead(This,prop) ) 

#define UserWindow_get_WindowToChainRead(This,prop)	\
    ( (This)->lpVtbl -> get_WindowToChainRead(This,prop) ) 

#define UserWindow_put_SpeakAndSpell(This,prop)	\
    ( (This)->lpVtbl -> put_SpeakAndSpell(This,prop) ) 

#define UserWindow_get_SpeakAndSpell(This,prop)	\
    ( (This)->lpVtbl -> get_SpeakAndSpell(This,prop) ) 

#define UserWindow_get_HighlightForegroundColor(This,Color)	\
    ( (This)->lpVtbl -> get_HighlightForegroundColor(This,Color) ) 

#define UserWindow_put_HighlightForegroundColor(This,Color)	\
    ( (This)->lpVtbl -> put_HighlightForegroundColor(This,Color) ) 

#define UserWindow_get_HighlightBackgroundColor(This,Color)	\
    ( (This)->lpVtbl -> get_HighlightBackgroundColor(This,Color) ) 

#define UserWindow_put_HighlightBackgroundColor(This,Color)	\
    ( (This)->lpVtbl -> put_HighlightBackgroundColor(This,Color) ) 

#define UserWindow_put_AutoDetermineHighlightColor(This,prop)	\
    ( (This)->lpVtbl -> put_AutoDetermineHighlightColor(This,prop) ) 

#define UserWindow_get_AutoDetermineHighlightColor(This,prop)	\
    ( (This)->lpVtbl -> get_AutoDetermineHighlightColor(This,prop) ) 

#define UserWindow_get_HighlightStatus(This,prop)	\
    ( (This)->lpVtbl -> get_HighlightStatus(This,prop) ) 

#define UserWindow_put_HighlightStatus(This,prop)	\
    ( (This)->lpVtbl -> put_HighlightStatus(This,prop) ) 

#define UserWindow_get_rectangle(This,WindowRectangle)	\
    ( (This)->lpVtbl -> get_rectangle(This,WindowRectangle) ) 

#define UserWindow_ConfigurationDialog(This)	\
    ( (This)->lpVtbl -> ConfigurationDialog(This) ) 

#define UserWindow_Speak(This)	\
    ( (This)->lpVtbl -> Speak(This) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __UserWindow_INTERFACE_DEFINED__ */


#ifndef __UserWindowCoordinates_INTERFACE_DEFINED__
#define __UserWindowCoordinates_INTERFACE_DEFINED__

/* interface UserWindowCoordinates */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_UserWindowCoordinates;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("2A8A2434-956D-41A0-891C-71B3E4FE6EF2")
    UserWindowCoordinates : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Left( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Left( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Right( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Right( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Top( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Top( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Bottom( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Bottom( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_OffsetFromLeft( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_OffsetFromLeft( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_OffsetFromRight( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_OffsetFromRight( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_OffsetFromTop( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_OffsetFromTop( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_OffsetFromBottom( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_OffsetFromBottom( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LeftFloatPosition( 
            /* [retval][out] */ FloatWindowPosition *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_LeftFloatPosition( 
            /* [in] */ FloatWindowPosition prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_RightFloatPosition( 
            /* [retval][out] */ FloatWindowPosition *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_RightFloatPosition( 
            /* [in] */ FloatWindowPosition prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TopFloatPosition( 
            /* [retval][out] */ FloatWindowPosition *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TopFloatPosition( 
            /* [in] */ FloatWindowPosition prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BottomFloatPosition( 
            /* [retval][out] */ FloatWindowPosition *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_BottomFloatPosition( 
            /* [in] */ FloatWindowPosition prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LeftString( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_LeftString( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_RightString( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_RightString( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TopString( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TopString( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BottomString( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_BottomString( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ConfigurationDialog( void) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct UserWindowCoordinatesVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            UserWindowCoordinates * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            UserWindowCoordinates * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            UserWindowCoordinates * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            UserWindowCoordinates * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            UserWindowCoordinates * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            UserWindowCoordinates * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            UserWindowCoordinates * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Left )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Left )( 
            UserWindowCoordinates * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Right )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Right )( 
            UserWindowCoordinates * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Top )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Top )( 
            UserWindowCoordinates * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Bottom )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Bottom )( 
            UserWindowCoordinates * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_OffsetFromLeft )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_OffsetFromLeft )( 
            UserWindowCoordinates * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_OffsetFromRight )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_OffsetFromRight )( 
            UserWindowCoordinates * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_OffsetFromTop )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_OffsetFromTop )( 
            UserWindowCoordinates * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_OffsetFromBottom )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_OffsetFromBottom )( 
            UserWindowCoordinates * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LeftFloatPosition )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ FloatWindowPosition *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_LeftFloatPosition )( 
            UserWindowCoordinates * This,
            /* [in] */ FloatWindowPosition prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_RightFloatPosition )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ FloatWindowPosition *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_RightFloatPosition )( 
            UserWindowCoordinates * This,
            /* [in] */ FloatWindowPosition prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TopFloatPosition )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ FloatWindowPosition *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TopFloatPosition )( 
            UserWindowCoordinates * This,
            /* [in] */ FloatWindowPosition prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BottomFloatPosition )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ FloatWindowPosition *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_BottomFloatPosition )( 
            UserWindowCoordinates * This,
            /* [in] */ FloatWindowPosition prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LeftString )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_LeftString )( 
            UserWindowCoordinates * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_RightString )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_RightString )( 
            UserWindowCoordinates * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TopString )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TopString )( 
            UserWindowCoordinates * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BottomString )( 
            UserWindowCoordinates * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_BottomString )( 
            UserWindowCoordinates * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ConfigurationDialog )( 
            UserWindowCoordinates * This);
        
        END_INTERFACE
    } UserWindowCoordinatesVtbl;

    interface UserWindowCoordinates
    {
        CONST_VTBL struct UserWindowCoordinatesVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define UserWindowCoordinates_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define UserWindowCoordinates_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define UserWindowCoordinates_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define UserWindowCoordinates_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define UserWindowCoordinates_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define UserWindowCoordinates_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define UserWindowCoordinates_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define UserWindowCoordinates_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define UserWindowCoordinates_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define UserWindowCoordinates_get_Left(This,prop)	\
    ( (This)->lpVtbl -> get_Left(This,prop) ) 

#define UserWindowCoordinates_put_Left(This,prop)	\
    ( (This)->lpVtbl -> put_Left(This,prop) ) 

#define UserWindowCoordinates_get_Right(This,prop)	\
    ( (This)->lpVtbl -> get_Right(This,prop) ) 

#define UserWindowCoordinates_put_Right(This,prop)	\
    ( (This)->lpVtbl -> put_Right(This,prop) ) 

#define UserWindowCoordinates_get_Top(This,prop)	\
    ( (This)->lpVtbl -> get_Top(This,prop) ) 

#define UserWindowCoordinates_put_Top(This,prop)	\
    ( (This)->lpVtbl -> put_Top(This,prop) ) 

#define UserWindowCoordinates_get_Bottom(This,prop)	\
    ( (This)->lpVtbl -> get_Bottom(This,prop) ) 

#define UserWindowCoordinates_put_Bottom(This,prop)	\
    ( (This)->lpVtbl -> put_Bottom(This,prop) ) 

#define UserWindowCoordinates_get_OffsetFromLeft(This,prop)	\
    ( (This)->lpVtbl -> get_OffsetFromLeft(This,prop) ) 

#define UserWindowCoordinates_put_OffsetFromLeft(This,prop)	\
    ( (This)->lpVtbl -> put_OffsetFromLeft(This,prop) ) 

#define UserWindowCoordinates_get_OffsetFromRight(This,prop)	\
    ( (This)->lpVtbl -> get_OffsetFromRight(This,prop) ) 

#define UserWindowCoordinates_put_OffsetFromRight(This,prop)	\
    ( (This)->lpVtbl -> put_OffsetFromRight(This,prop) ) 

#define UserWindowCoordinates_get_OffsetFromTop(This,prop)	\
    ( (This)->lpVtbl -> get_OffsetFromTop(This,prop) ) 

#define UserWindowCoordinates_put_OffsetFromTop(This,prop)	\
    ( (This)->lpVtbl -> put_OffsetFromTop(This,prop) ) 

#define UserWindowCoordinates_get_OffsetFromBottom(This,prop)	\
    ( (This)->lpVtbl -> get_OffsetFromBottom(This,prop) ) 

#define UserWindowCoordinates_put_OffsetFromBottom(This,prop)	\
    ( (This)->lpVtbl -> put_OffsetFromBottom(This,prop) ) 

#define UserWindowCoordinates_get_LeftFloatPosition(This,prop)	\
    ( (This)->lpVtbl -> get_LeftFloatPosition(This,prop) ) 

#define UserWindowCoordinates_put_LeftFloatPosition(This,prop)	\
    ( (This)->lpVtbl -> put_LeftFloatPosition(This,prop) ) 

#define UserWindowCoordinates_get_RightFloatPosition(This,prop)	\
    ( (This)->lpVtbl -> get_RightFloatPosition(This,prop) ) 

#define UserWindowCoordinates_put_RightFloatPosition(This,prop)	\
    ( (This)->lpVtbl -> put_RightFloatPosition(This,prop) ) 

#define UserWindowCoordinates_get_TopFloatPosition(This,prop)	\
    ( (This)->lpVtbl -> get_TopFloatPosition(This,prop) ) 

#define UserWindowCoordinates_put_TopFloatPosition(This,prop)	\
    ( (This)->lpVtbl -> put_TopFloatPosition(This,prop) ) 

#define UserWindowCoordinates_get_BottomFloatPosition(This,prop)	\
    ( (This)->lpVtbl -> get_BottomFloatPosition(This,prop) ) 

#define UserWindowCoordinates_put_BottomFloatPosition(This,prop)	\
    ( (This)->lpVtbl -> put_BottomFloatPosition(This,prop) ) 

#define UserWindowCoordinates_get_LeftString(This,prop)	\
    ( (This)->lpVtbl -> get_LeftString(This,prop) ) 

#define UserWindowCoordinates_put_LeftString(This,prop)	\
    ( (This)->lpVtbl -> put_LeftString(This,prop) ) 

#define UserWindowCoordinates_get_RightString(This,prop)	\
    ( (This)->lpVtbl -> get_RightString(This,prop) ) 

#define UserWindowCoordinates_put_RightString(This,prop)	\
    ( (This)->lpVtbl -> put_RightString(This,prop) ) 

#define UserWindowCoordinates_get_TopString(This,prop)	\
    ( (This)->lpVtbl -> get_TopString(This,prop) ) 

#define UserWindowCoordinates_put_TopString(This,prop)	\
    ( (This)->lpVtbl -> put_TopString(This,prop) ) 

#define UserWindowCoordinates_get_BottomString(This,prop)	\
    ( (This)->lpVtbl -> get_BottomString(This,prop) ) 

#define UserWindowCoordinates_put_BottomString(This,prop)	\
    ( (This)->lpVtbl -> put_BottomString(This,prop) ) 

#define UserWindowCoordinates_ConfigurationDialog(This)	\
    ( (This)->lpVtbl -> ConfigurationDialog(This) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __UserWindowCoordinates_INTERFACE_DEFINED__ */


#ifndef __HyperactiveWindows_INTERFACE_DEFINED__
#define __HyperactiveWindows_INTERFACE_DEFINED__

/* interface HyperactiveWindows */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_HyperactiveWindows;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("742B480F-5A4E-4E2D-B544-D21FBEDE9A1C")
    HyperactiveWindows : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ BSTR hyperWinID,
            /* [retval][out] */ HyperactiveWindow **hyperWindow) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct HyperactiveWindowsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            HyperactiveWindows * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            HyperactiveWindows * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            HyperactiveWindows * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            HyperactiveWindows * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            HyperactiveWindows * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            HyperactiveWindows * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            HyperactiveWindows * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            HyperactiveWindows * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            HyperactiveWindows * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            HyperactiveWindows * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            HyperactiveWindows * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            HyperactiveWindows * This,
            /* [in] */ BSTR hyperWinID,
            /* [retval][out] */ HyperactiveWindow **hyperWindow);
        
        END_INTERFACE
    } HyperactiveWindowsVtbl;

    interface HyperactiveWindows
    {
        CONST_VTBL struct HyperactiveWindowsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define HyperactiveWindows_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define HyperactiveWindows_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define HyperactiveWindows_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define HyperactiveWindows_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define HyperactiveWindows_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define HyperactiveWindows_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define HyperactiveWindows_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define HyperactiveWindows_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define HyperactiveWindows_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define HyperactiveWindows_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define HyperactiveWindows_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define HyperactiveWindows_Item(This,hyperWinID,hyperWindow)	\
    ( (This)->lpVtbl -> Item(This,hyperWinID,hyperWindow) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __HyperactiveWindows_INTERFACE_DEFINED__ */


#ifndef __HyperactiveWindow_INTERFACE_DEFINED__
#define __HyperactiveWindow_INTERFACE_DEFINED__

/* interface HyperactiveWindow */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_HyperactiveWindow;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("86B2E2DF-4987-4E18-8791-E638BFB774AE")
    HyperactiveWindow : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Coordinates( 
            /* [retval][out] */ UserWindowCoordinates **coords) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Status( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Status( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Comment( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Comment( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Interruptability( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Interruptability( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Trigger( 
            /* [retval][out] */ HyperWindowTrigger *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Trigger( 
            /* [in] */ HyperWindowTrigger prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Commands( 
            /* [retval][out] */ HyperactiveWindowCommands **prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TriggerInformation( 
            /* [retval][out] */ VARIANT *commandInfo) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TriggerInformation( 
            /* [in] */ VARIANT commandInfo) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SearchAttribute( 
            /* [retval][out] */ TriggerSearchAttribute **prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_rectangle( 
            /* [retval][out] */ WindowRectangle **WindowRectangle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ConfigurationDialog( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Speak( void) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct HyperactiveWindowVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            HyperactiveWindow * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            HyperactiveWindow * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            HyperactiveWindow * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            HyperactiveWindow * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            HyperactiveWindow * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            HyperactiveWindow * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            HyperactiveWindow * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            HyperactiveWindow * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            HyperactiveWindow * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Coordinates )( 
            HyperactiveWindow * This,
            /* [retval][out] */ UserWindowCoordinates **coords);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Status )( 
            HyperactiveWindow * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Status )( 
            HyperactiveWindow * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Comment )( 
            HyperactiveWindow * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Comment )( 
            HyperactiveWindow * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Interruptability )( 
            HyperactiveWindow * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Interruptability )( 
            HyperactiveWindow * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Trigger )( 
            HyperactiveWindow * This,
            /* [retval][out] */ HyperWindowTrigger *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Trigger )( 
            HyperactiveWindow * This,
            /* [in] */ HyperWindowTrigger prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Commands )( 
            HyperactiveWindow * This,
            /* [retval][out] */ HyperactiveWindowCommands **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TriggerInformation )( 
            HyperactiveWindow * This,
            /* [retval][out] */ VARIANT *commandInfo);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TriggerInformation )( 
            HyperactiveWindow * This,
            /* [in] */ VARIANT commandInfo);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SearchAttribute )( 
            HyperactiveWindow * This,
            /* [retval][out] */ TriggerSearchAttribute **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_rectangle )( 
            HyperactiveWindow * This,
            /* [retval][out] */ WindowRectangle **WindowRectangle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ConfigurationDialog )( 
            HyperactiveWindow * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Speak )( 
            HyperactiveWindow * This);
        
        END_INTERFACE
    } HyperactiveWindowVtbl;

    interface HyperactiveWindow
    {
        CONST_VTBL struct HyperactiveWindowVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define HyperactiveWindow_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define HyperactiveWindow_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define HyperactiveWindow_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define HyperactiveWindow_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define HyperactiveWindow_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define HyperactiveWindow_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define HyperactiveWindow_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define HyperactiveWindow_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define HyperactiveWindow_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define HyperactiveWindow_get_Coordinates(This,coords)	\
    ( (This)->lpVtbl -> get_Coordinates(This,coords) ) 

#define HyperactiveWindow_get_Status(This,prop)	\
    ( (This)->lpVtbl -> get_Status(This,prop) ) 

#define HyperactiveWindow_put_Status(This,prop)	\
    ( (This)->lpVtbl -> put_Status(This,prop) ) 

#define HyperactiveWindow_get_Comment(This,prop)	\
    ( (This)->lpVtbl -> get_Comment(This,prop) ) 

#define HyperactiveWindow_put_Comment(This,prop)	\
    ( (This)->lpVtbl -> put_Comment(This,prop) ) 

#define HyperactiveWindow_get_Interruptability(This,prop)	\
    ( (This)->lpVtbl -> get_Interruptability(This,prop) ) 

#define HyperactiveWindow_put_Interruptability(This,prop)	\
    ( (This)->lpVtbl -> put_Interruptability(This,prop) ) 

#define HyperactiveWindow_get_Trigger(This,prop)	\
    ( (This)->lpVtbl -> get_Trigger(This,prop) ) 

#define HyperactiveWindow_put_Trigger(This,prop)	\
    ( (This)->lpVtbl -> put_Trigger(This,prop) ) 

#define HyperactiveWindow_get_Commands(This,prop)	\
    ( (This)->lpVtbl -> get_Commands(This,prop) ) 

#define HyperactiveWindow_get_TriggerInformation(This,commandInfo)	\
    ( (This)->lpVtbl -> get_TriggerInformation(This,commandInfo) ) 

#define HyperactiveWindow_put_TriggerInformation(This,commandInfo)	\
    ( (This)->lpVtbl -> put_TriggerInformation(This,commandInfo) ) 

#define HyperactiveWindow_get_SearchAttribute(This,prop)	\
    ( (This)->lpVtbl -> get_SearchAttribute(This,prop) ) 

#define HyperactiveWindow_get_rectangle(This,WindowRectangle)	\
    ( (This)->lpVtbl -> get_rectangle(This,WindowRectangle) ) 

#define HyperactiveWindow_ConfigurationDialog(This)	\
    ( (This)->lpVtbl -> ConfigurationDialog(This) ) 

#define HyperactiveWindow_Speak(This)	\
    ( (This)->lpVtbl -> Speak(This) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __HyperactiveWindow_INTERFACE_DEFINED__ */


#ifndef __HyperactiveWindowCommands_INTERFACE_DEFINED__
#define __HyperactiveWindowCommands_INTERFACE_DEFINED__

/* interface HyperactiveWindowCommands */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_HyperactiveWindowCommands;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("7EF5AC30-72B1-46BB-B16E-B728E83E7C04")
    HyperactiveWindowCommands : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ HyperactiveWindowCommand **hyperCommand) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct HyperactiveWindowCommandsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            HyperactiveWindowCommands * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            HyperactiveWindowCommands * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            HyperactiveWindowCommands * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            HyperactiveWindowCommands * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            HyperactiveWindowCommands * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            HyperactiveWindowCommands * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            HyperactiveWindowCommands * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            HyperactiveWindowCommands * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            HyperactiveWindowCommands * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            HyperactiveWindowCommands * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            HyperactiveWindowCommands * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            HyperactiveWindowCommands * This,
            /* [in] */ long index,
            /* [retval][out] */ HyperactiveWindowCommand **hyperCommand);
        
        END_INTERFACE
    } HyperactiveWindowCommandsVtbl;

    interface HyperactiveWindowCommands
    {
        CONST_VTBL struct HyperactiveWindowCommandsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define HyperactiveWindowCommands_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define HyperactiveWindowCommands_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define HyperactiveWindowCommands_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define HyperactiveWindowCommands_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define HyperactiveWindowCommands_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define HyperactiveWindowCommands_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define HyperactiveWindowCommands_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define HyperactiveWindowCommands_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define HyperactiveWindowCommands_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define HyperactiveWindowCommands_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define HyperactiveWindowCommands_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define HyperactiveWindowCommands_Item(This,index,hyperCommand)	\
    ( (This)->lpVtbl -> Item(This,index,hyperCommand) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __HyperactiveWindowCommands_INTERFACE_DEFINED__ */


#ifndef __HyperactiveWindowCommand_INTERFACE_DEFINED__
#define __HyperactiveWindowCommand_INTERFACE_DEFINED__

/* interface HyperactiveWindowCommand */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_HyperactiveWindowCommand;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("9BD2BD33-DF7E-44AC-8CF8-BBAE0BD47179")
    HyperactiveWindowCommand : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Command( 
            /* [retval][out] */ HyperWindowCommand *Command) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Command( 
            /* [in] */ HyperWindowCommand Command) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CommandInformation( 
            /* [retval][out] */ VARIANT *commandInfo) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CommandInformation( 
            /* [in] */ VARIANT commandInfo) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct HyperactiveWindowCommandVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            HyperactiveWindowCommand * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            HyperactiveWindowCommand * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            HyperactiveWindowCommand * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            HyperactiveWindowCommand * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            HyperactiveWindowCommand * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            HyperactiveWindowCommand * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            HyperactiveWindowCommand * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            HyperactiveWindowCommand * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            HyperactiveWindowCommand * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Command )( 
            HyperactiveWindowCommand * This,
            /* [retval][out] */ HyperWindowCommand *Command);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Command )( 
            HyperactiveWindowCommand * This,
            /* [in] */ HyperWindowCommand Command);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CommandInformation )( 
            HyperactiveWindowCommand * This,
            /* [retval][out] */ VARIANT *commandInfo);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CommandInformation )( 
            HyperactiveWindowCommand * This,
            /* [in] */ VARIANT commandInfo);
        
        END_INTERFACE
    } HyperactiveWindowCommandVtbl;

    interface HyperactiveWindowCommand
    {
        CONST_VTBL struct HyperactiveWindowCommandVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define HyperactiveWindowCommand_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define HyperactiveWindowCommand_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define HyperactiveWindowCommand_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define HyperactiveWindowCommand_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define HyperactiveWindowCommand_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define HyperactiveWindowCommand_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define HyperactiveWindowCommand_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define HyperactiveWindowCommand_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define HyperactiveWindowCommand_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define HyperactiveWindowCommand_get_Command(This,Command)	\
    ( (This)->lpVtbl -> get_Command(This,Command) ) 

#define HyperactiveWindowCommand_put_Command(This,Command)	\
    ( (This)->lpVtbl -> put_Command(This,Command) ) 

#define HyperactiveWindowCommand_get_CommandInformation(This,commandInfo)	\
    ( (This)->lpVtbl -> get_CommandInformation(This,commandInfo) ) 

#define HyperactiveWindowCommand_put_CommandInformation(This,commandInfo)	\
    ( (This)->lpVtbl -> put_CommandInformation(This,commandInfo) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __HyperactiveWindowCommand_INTERFACE_DEFINED__ */


#ifndef __TriggerSearchAttribute_INTERFACE_DEFINED__
#define __TriggerSearchAttribute_INTERFACE_DEFINED__

/* interface TriggerSearchAttribute */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_TriggerSearchAttribute;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("250971B8-CB84-46F3-8D24-641768707DE6")
    TriggerSearchAttribute : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ForegroundColor( 
            /* [retval][out] */ Color **Color) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ForegroundColor( 
            /* [in] */ Color *Color) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BackgroundColor( 
            /* [retval][out] */ Color **Color) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_BackgroundColor( 
            /* [in] */ Color *Color) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct TriggerSearchAttributeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            TriggerSearchAttribute * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            TriggerSearchAttribute * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            TriggerSearchAttribute * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            TriggerSearchAttribute * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            TriggerSearchAttribute * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            TriggerSearchAttribute * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            TriggerSearchAttribute * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            TriggerSearchAttribute * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            TriggerSearchAttribute * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ForegroundColor )( 
            TriggerSearchAttribute * This,
            /* [retval][out] */ Color **Color);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ForegroundColor )( 
            TriggerSearchAttribute * This,
            /* [in] */ Color *Color);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BackgroundColor )( 
            TriggerSearchAttribute * This,
            /* [retval][out] */ Color **Color);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_BackgroundColor )( 
            TriggerSearchAttribute * This,
            /* [in] */ Color *Color);
        
        END_INTERFACE
    } TriggerSearchAttributeVtbl;

    interface TriggerSearchAttribute
    {
        CONST_VTBL struct TriggerSearchAttributeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define TriggerSearchAttribute_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define TriggerSearchAttribute_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define TriggerSearchAttribute_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define TriggerSearchAttribute_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define TriggerSearchAttribute_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define TriggerSearchAttribute_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define TriggerSearchAttribute_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define TriggerSearchAttribute_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define TriggerSearchAttribute_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define TriggerSearchAttribute_get_ForegroundColor(This,Color)	\
    ( (This)->lpVtbl -> get_ForegroundColor(This,Color) ) 

#define TriggerSearchAttribute_put_ForegroundColor(This,Color)	\
    ( (This)->lpVtbl -> put_ForegroundColor(This,Color) ) 

#define TriggerSearchAttribute_get_BackgroundColor(This,Color)	\
    ( (This)->lpVtbl -> get_BackgroundColor(This,Color) ) 

#define TriggerSearchAttribute_put_BackgroundColor(This,Color)	\
    ( (This)->lpVtbl -> put_BackgroundColor(This,Color) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __TriggerSearchAttribute_INTERFACE_DEFINED__ */


#ifndef __HotKeys_INTERFACE_DEFINED__
#define __HotKeys_INTERFACE_DEFINED__

/* interface HotKeys */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_HotKeys;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("C8CA7C7F-E51D-43CD-8340-FD040E8DAEC6")
    HotKeys : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ Hotkey **Hotkey) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE GetDuplicates( 
            /* [in] */ VARIANT Key,
            /* [retval][out] */ HotKeys **HotKeys) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct HotKeysVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            HotKeys * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            HotKeys * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            HotKeys * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            HotKeys * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            HotKeys * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            HotKeys * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            HotKeys * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            HotKeys * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            HotKeys * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            HotKeys * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            HotKeys * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            HotKeys * This,
            /* [in] */ long index,
            /* [retval][out] */ Hotkey **Hotkey);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *GetDuplicates )( 
            HotKeys * This,
            /* [in] */ VARIANT Key,
            /* [retval][out] */ HotKeys **HotKeys);
        
        END_INTERFACE
    } HotKeysVtbl;

    interface HotKeys
    {
        CONST_VTBL struct HotKeysVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define HotKeys_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define HotKeys_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define HotKeys_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define HotKeys_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define HotKeys_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define HotKeys_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define HotKeys_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define HotKeys_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define HotKeys_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define HotKeys_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define HotKeys_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define HotKeys_Item(This,index,Hotkey)	\
    ( (This)->lpVtbl -> Item(This,index,Hotkey) ) 

#define HotKeys_GetDuplicates(This,Key,HotKeys)	\
    ( (This)->lpVtbl -> GetDuplicates(This,Key,HotKeys) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __HotKeys_INTERFACE_DEFINED__ */


#ifndef __Hotkey_INTERFACE_DEFINED__
#define __Hotkey_INTERFACE_DEFINED__

/* interface Hotkey */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Hotkey;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("72A98026-B0FC-41BA-8835-FF3F24A07987")
    Hotkey : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_id( 
            /* [retval][out] */ HotkeyID *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Key( 
            /* [retval][out] */ VARIANT *Key) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Key( 
            /* [in] */ VARIANT Key) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Description( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Execute( 
            /* [optional][in] */ VARIANT KeyAction) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Scope( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Scope( 
            /* [in] */ SetFileScopes prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct HotkeyVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Hotkey * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Hotkey * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Hotkey * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Hotkey * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Hotkey * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Hotkey * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Hotkey * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Hotkey * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Hotkey * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            Hotkey * This,
            /* [retval][out] */ HotkeyID *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Key )( 
            Hotkey * This,
            /* [retval][out] */ VARIANT *Key);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Key )( 
            Hotkey * This,
            /* [in] */ VARIANT Key);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Description )( 
            Hotkey * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Execute )( 
            Hotkey * This,
            /* [optional][in] */ VARIANT KeyAction);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Scope )( 
            Hotkey * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Scope )( 
            Hotkey * This,
            /* [in] */ SetFileScopes prop);
        
        END_INTERFACE
    } HotkeyVtbl;

    interface Hotkey
    {
        CONST_VTBL struct HotkeyVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Hotkey_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Hotkey_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Hotkey_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Hotkey_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Hotkey_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Hotkey_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Hotkey_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Hotkey_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Hotkey_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Hotkey_get_id(This,prop)	\
    ( (This)->lpVtbl -> get_id(This,prop) ) 

#define Hotkey_get_Key(This,Key)	\
    ( (This)->lpVtbl -> get_Key(This,Key) ) 

#define Hotkey_put_Key(This,Key)	\
    ( (This)->lpVtbl -> put_Key(This,Key) ) 

#define Hotkey_get_Description(This,prop)	\
    ( (This)->lpVtbl -> get_Description(This,prop) ) 

#define Hotkey_Execute(This,KeyAction)	\
    ( (This)->lpVtbl -> Execute(This,KeyAction) ) 

#define Hotkey_get_Scope(This,prop)	\
    ( (This)->lpVtbl -> get_Scope(This,prop) ) 

#define Hotkey_put_Scope(This,prop)	\
    ( (This)->lpVtbl -> put_Scope(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Hotkey_INTERFACE_DEFINED__ */


#ifndef __CursorKeys_INTERFACE_DEFINED__
#define __CursorKeys_INTERFACE_DEFINED__

/* interface CursorKeys */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_CursorKeys;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("193004DD-1572-4966-8A3F-6710C54BE4E3")
    CursorKeys : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ CursorKey **CursorKey) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE GetDuplicates( 
            /* [in] */ VARIANT Key,
            /* [retval][out] */ CursorKeys **CursorKeys) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeGlobalKeys( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeGlobalKeys( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct CursorKeysVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            CursorKeys * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            CursorKeys * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            CursorKeys * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            CursorKeys * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            CursorKeys * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            CursorKeys * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            CursorKeys * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            CursorKeys * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            CursorKeys * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            CursorKeys * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            CursorKeys * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            CursorKeys * This,
            /* [in] */ long index,
            /* [retval][out] */ CursorKey **CursorKey);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *GetDuplicates )( 
            CursorKeys * This,
            /* [in] */ VARIANT Key,
            /* [retval][out] */ CursorKeys **CursorKeys);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeGlobalKeys )( 
            CursorKeys * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeGlobalKeys )( 
            CursorKeys * This,
            /* [in] */ VARIANT_BOOL prop);
        
        END_INTERFACE
    } CursorKeysVtbl;

    interface CursorKeys
    {
        CONST_VTBL struct CursorKeysVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define CursorKeys_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define CursorKeys_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define CursorKeys_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define CursorKeys_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define CursorKeys_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define CursorKeys_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define CursorKeys_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define CursorKeys_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define CursorKeys_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define CursorKeys_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define CursorKeys_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define CursorKeys_Item(This,index,CursorKey)	\
    ( (This)->lpVtbl -> Item(This,index,CursorKey) ) 

#define CursorKeys_GetDuplicates(This,Key,CursorKeys)	\
    ( (This)->lpVtbl -> GetDuplicates(This,Key,CursorKeys) ) 

#define CursorKeys_get_IncludeGlobalKeys(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeGlobalKeys(This,prop) ) 

#define CursorKeys_put_IncludeGlobalKeys(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeGlobalKeys(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __CursorKeys_INTERFACE_DEFINED__ */


#ifndef __CursorKey_INTERFACE_DEFINED__
#define __CursorKey_INTERFACE_DEFINED__

/* interface CursorKey */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_CursorKey;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("275E3F64-D44A-4643-97D7-5C741320CAD3")
    CursorKey : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Key( 
            /* [retval][out] */ VARIANT *Key) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Key( 
            /* [in] */ VARIANT Key) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FirstAction( 
            /* [retval][out] */ CursorKeyAction **Action) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SecondAction( 
            /* [retval][out] */ CursorKeyAction **Action) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DelayFactor( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_DelayFactor( 
            /* [in] */ long prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Scope( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Scope( 
            /* [in] */ SetFileScopes prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct CursorKeyVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            CursorKey * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            CursorKey * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            CursorKey * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            CursorKey * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            CursorKey * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            CursorKey * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            CursorKey * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            CursorKey * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            CursorKey * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Key )( 
            CursorKey * This,
            /* [retval][out] */ VARIANT *Key);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Key )( 
            CursorKey * This,
            /* [in] */ VARIANT Key);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FirstAction )( 
            CursorKey * This,
            /* [retval][out] */ CursorKeyAction **Action);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SecondAction )( 
            CursorKey * This,
            /* [retval][out] */ CursorKeyAction **Action);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DelayFactor )( 
            CursorKey * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_DelayFactor )( 
            CursorKey * This,
            /* [in] */ long prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Scope )( 
            CursorKey * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Scope )( 
            CursorKey * This,
            /* [in] */ SetFileScopes prop);
        
        END_INTERFACE
    } CursorKeyVtbl;

    interface CursorKey
    {
        CONST_VTBL struct CursorKeyVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define CursorKey_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define CursorKey_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define CursorKey_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define CursorKey_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define CursorKey_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define CursorKey_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define CursorKey_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define CursorKey_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define CursorKey_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define CursorKey_get_Key(This,Key)	\
    ( (This)->lpVtbl -> get_Key(This,Key) ) 

#define CursorKey_put_Key(This,Key)	\
    ( (This)->lpVtbl -> put_Key(This,Key) ) 

#define CursorKey_get_FirstAction(This,Action)	\
    ( (This)->lpVtbl -> get_FirstAction(This,Action) ) 

#define CursorKey_get_SecondAction(This,Action)	\
    ( (This)->lpVtbl -> get_SecondAction(This,Action) ) 

#define CursorKey_get_DelayFactor(This,prop)	\
    ( (This)->lpVtbl -> get_DelayFactor(This,prop) ) 

#define CursorKey_put_DelayFactor(This,prop)	\
    ( (This)->lpVtbl -> put_DelayFactor(This,prop) ) 

#define CursorKey_get_Scope(This,prop)	\
    ( (This)->lpVtbl -> get_Scope(This,prop) ) 

#define CursorKey_put_Scope(This,prop)	\
    ( (This)->lpVtbl -> put_Scope(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __CursorKey_INTERFACE_DEFINED__ */


#ifndef __CursorKeyAction_INTERFACE_DEFINED__
#define __CursorKeyAction_INTERFACE_DEFINED__

/* interface CursorKeyAction */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_CursorKeyAction;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("E362FB3E-5315-4ECC-9B81-FC8281F64786")
    CursorKeyAction : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Description( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_id( 
            /* [retval][out] */ CursorKeyActions *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_id( 
            /* [in] */ CursorKeyActions prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parameter( 
            /* [retval][out] */ VARIANT *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Parameter( 
            /* [in] */ VARIANT prop) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Execute( void) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct CursorKeyActionVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            CursorKeyAction * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            CursorKeyAction * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            CursorKeyAction * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            CursorKeyAction * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            CursorKeyAction * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            CursorKeyAction * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            CursorKeyAction * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            CursorKeyAction * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            CursorKeyAction * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Description )( 
            CursorKeyAction * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            CursorKeyAction * This,
            /* [retval][out] */ CursorKeyActions *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_id )( 
            CursorKeyAction * This,
            /* [in] */ CursorKeyActions prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parameter )( 
            CursorKeyAction * This,
            /* [retval][out] */ VARIANT *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Parameter )( 
            CursorKeyAction * This,
            /* [in] */ VARIANT prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Execute )( 
            CursorKeyAction * This);
        
        END_INTERFACE
    } CursorKeyActionVtbl;

    interface CursorKeyAction
    {
        CONST_VTBL struct CursorKeyActionVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define CursorKeyAction_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define CursorKeyAction_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define CursorKeyAction_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define CursorKeyAction_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define CursorKeyAction_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define CursorKeyAction_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define CursorKeyAction_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define CursorKeyAction_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define CursorKeyAction_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define CursorKeyAction_get_Description(This,prop)	\
    ( (This)->lpVtbl -> get_Description(This,prop) ) 

#define CursorKeyAction_get_id(This,prop)	\
    ( (This)->lpVtbl -> get_id(This,prop) ) 

#define CursorKeyAction_put_id(This,prop)	\
    ( (This)->lpVtbl -> put_id(This,prop) ) 

#define CursorKeyAction_get_Parameter(This,prop)	\
    ( (This)->lpVtbl -> get_Parameter(This,prop) ) 

#define CursorKeyAction_put_Parameter(This,prop)	\
    ( (This)->lpVtbl -> put_Parameter(This,prop) ) 

#define CursorKeyAction_Execute(This)	\
    ( (This)->lpVtbl -> Execute(This) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __CursorKeyAction_INTERFACE_DEFINED__ */


#ifndef __CharacterDictionary_INTERFACE_DEFINED__
#define __CharacterDictionary_INTERFACE_DEFINED__

/* interface CharacterDictionary */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_CharacterDictionary;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("2E3A39F9-65DA-4B7F-BD59-9AC8490F17F2")
    CharacterDictionary : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Filename( 
            /* [retval][out] */ BSTR *Name) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Overlap( 
            /* [retval][out] */ _Window **Overlap) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsValid( 
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ VARIANT Character,
            /* [optional][in] */ VARIANT FontName,
            /* [retval][out] */ CharacterEntry **CharacterEntry) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Add( 
            /* [in] */ VARIANT Character,
            /* [in] */ BSTR Translation,
            /* [optional][in] */ VARIANT FontName,
            /* [retval][out] */ VARIANT_BOOL *succes) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Delete( 
            /* [in] */ VARIANT Character,
            /* [optional][in] */ VARIANT FontName,
            /* [retval][out] */ VARIANT_BOOL *succes) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Save( 
            /* [retval][out] */ VARIANT_BOOL *succes) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SaveAs( 
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *succes) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Close( 
            /* [retval][out] */ VARIANT_BOOL *succes) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Open( 
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *succes) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct CharacterDictionaryVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            CharacterDictionary * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            CharacterDictionary * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            CharacterDictionary * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            CharacterDictionary * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            CharacterDictionary * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            CharacterDictionary * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            CharacterDictionary * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            CharacterDictionary * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            CharacterDictionary * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            CharacterDictionary * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Filename )( 
            CharacterDictionary * This,
            /* [retval][out] */ BSTR *Name);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Overlap )( 
            CharacterDictionary * This,
            /* [retval][out] */ _Window **Overlap);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsValid )( 
            CharacterDictionary * This,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            CharacterDictionary * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            CharacterDictionary * This,
            /* [in] */ VARIANT Character,
            /* [optional][in] */ VARIANT FontName,
            /* [retval][out] */ CharacterEntry **CharacterEntry);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Add )( 
            CharacterDictionary * This,
            /* [in] */ VARIANT Character,
            /* [in] */ BSTR Translation,
            /* [optional][in] */ VARIANT FontName,
            /* [retval][out] */ VARIANT_BOOL *succes);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Delete )( 
            CharacterDictionary * This,
            /* [in] */ VARIANT Character,
            /* [optional][in] */ VARIANT FontName,
            /* [retval][out] */ VARIANT_BOOL *succes);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Save )( 
            CharacterDictionary * This,
            /* [retval][out] */ VARIANT_BOOL *succes);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SaveAs )( 
            CharacterDictionary * This,
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *succes);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Close )( 
            CharacterDictionary * This,
            /* [retval][out] */ VARIANT_BOOL *succes);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Open )( 
            CharacterDictionary * This,
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *succes);
        
        END_INTERFACE
    } CharacterDictionaryVtbl;

    interface CharacterDictionary
    {
        CONST_VTBL struct CharacterDictionaryVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define CharacterDictionary_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define CharacterDictionary_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define CharacterDictionary_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define CharacterDictionary_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define CharacterDictionary_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define CharacterDictionary_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define CharacterDictionary_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define CharacterDictionary_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define CharacterDictionary_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define CharacterDictionary_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define CharacterDictionary_get_Filename(This,Name)	\
    ( (This)->lpVtbl -> get_Filename(This,Name) ) 

#define CharacterDictionary_get_Overlap(This,Overlap)	\
    ( (This)->lpVtbl -> get_Overlap(This,Overlap) ) 

#define CharacterDictionary_get_IsValid(This,success)	\
    ( (This)->lpVtbl -> get_IsValid(This,success) ) 

#define CharacterDictionary_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define CharacterDictionary_Item(This,Character,FontName,CharacterEntry)	\
    ( (This)->lpVtbl -> Item(This,Character,FontName,CharacterEntry) ) 

#define CharacterDictionary_Add(This,Character,Translation,FontName,succes)	\
    ( (This)->lpVtbl -> Add(This,Character,Translation,FontName,succes) ) 

#define CharacterDictionary_Delete(This,Character,FontName,succes)	\
    ( (This)->lpVtbl -> Delete(This,Character,FontName,succes) ) 

#define CharacterDictionary_Save(This,succes)	\
    ( (This)->lpVtbl -> Save(This,succes) ) 

#define CharacterDictionary_SaveAs(This,Name,succes)	\
    ( (This)->lpVtbl -> SaveAs(This,Name,succes) ) 

#define CharacterDictionary_Close(This,succes)	\
    ( (This)->lpVtbl -> Close(This,succes) ) 

#define CharacterDictionary_Open(This,Name,succes)	\
    ( (This)->lpVtbl -> Open(This,Name,succes) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __CharacterDictionary_INTERFACE_DEFINED__ */


#ifndef __CharacterEntry_INTERFACE_DEFINED__
#define __CharacterEntry_INTERFACE_DEFINED__

/* interface CharacterEntry */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_CharacterEntry;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("9C333E1B-3BDA-41DF-AAB3-56F901F26D5E")
    CharacterEntry : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Character( 
            /* [in] */ VARIANT prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Character( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FontName( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FontName( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Translation( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Translation( 
            /* [retval][out] */ BSTR *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct CharacterEntryVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            CharacterEntry * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            CharacterEntry * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            CharacterEntry * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            CharacterEntry * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            CharacterEntry * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            CharacterEntry * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            CharacterEntry * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            CharacterEntry * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            CharacterEntry * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Character )( 
            CharacterEntry * This,
            /* [in] */ VARIANT prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Character )( 
            CharacterEntry * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FontName )( 
            CharacterEntry * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FontName )( 
            CharacterEntry * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Translation )( 
            CharacterEntry * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Translation )( 
            CharacterEntry * This,
            /* [retval][out] */ BSTR *prop);
        
        END_INTERFACE
    } CharacterEntryVtbl;

    interface CharacterEntry
    {
        CONST_VTBL struct CharacterEntryVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define CharacterEntry_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define CharacterEntry_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define CharacterEntry_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define CharacterEntry_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define CharacterEntry_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define CharacterEntry_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define CharacterEntry_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define CharacterEntry_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define CharacterEntry_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define CharacterEntry_put_Character(This,prop)	\
    ( (This)->lpVtbl -> put_Character(This,prop) ) 

#define CharacterEntry_get_Character(This,prop)	\
    ( (This)->lpVtbl -> get_Character(This,prop) ) 

#define CharacterEntry_put_FontName(This,prop)	\
    ( (This)->lpVtbl -> put_FontName(This,prop) ) 

#define CharacterEntry_get_FontName(This,prop)	\
    ( (This)->lpVtbl -> get_FontName(This,prop) ) 

#define CharacterEntry_put_Translation(This,prop)	\
    ( (This)->lpVtbl -> put_Translation(This,prop) ) 

#define CharacterEntry_get_Translation(This,prop)	\
    ( (This)->lpVtbl -> get_Translation(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __CharacterEntry_INTERFACE_DEFINED__ */


#ifndef __ColorDictionary_INTERFACE_DEFINED__
#define __ColorDictionary_INTERFACE_DEFINED__

/* interface ColorDictionary */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ColorDictionary;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("FADF1AF7-3AC3-4E41-8F34-B0528CED5D0A")
    ColorDictionary : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Filename( 
            /* [retval][out] */ BSTR *Name) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Overlap( 
            /* [retval][out] */ _Window **Overlap) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsValid( 
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ Color *Color,
            /* [optional][in] */ VARIANT ClosestMatch,
            /* [retval][out] */ ColorEntry **ColorEntry) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Add( 
            /* [in] */ Color *Color,
            /* [in] */ BSTR Translation,
            /* [retval][out] */ VARIANT_BOOL *succes) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Delete( 
            /* [in] */ Color *Color,
            /* [retval][out] */ VARIANT_BOOL *succes) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Save( 
            /* [retval][out] */ VARIANT_BOOL *succes) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SaveAs( 
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *succes) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Close( 
            /* [retval][out] */ VARIANT_BOOL *succes) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Open( 
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *succes) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ColorDictionaryVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ColorDictionary * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ColorDictionary * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ColorDictionary * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ColorDictionary * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ColorDictionary * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ColorDictionary * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ColorDictionary * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ColorDictionary * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ColorDictionary * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            ColorDictionary * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Filename )( 
            ColorDictionary * This,
            /* [retval][out] */ BSTR *Name);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Overlap )( 
            ColorDictionary * This,
            /* [retval][out] */ _Window **Overlap);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsValid )( 
            ColorDictionary * This,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            ColorDictionary * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            ColorDictionary * This,
            /* [in] */ Color *Color,
            /* [optional][in] */ VARIANT ClosestMatch,
            /* [retval][out] */ ColorEntry **ColorEntry);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Add )( 
            ColorDictionary * This,
            /* [in] */ Color *Color,
            /* [in] */ BSTR Translation,
            /* [retval][out] */ VARIANT_BOOL *succes);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Delete )( 
            ColorDictionary * This,
            /* [in] */ Color *Color,
            /* [retval][out] */ VARIANT_BOOL *succes);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Save )( 
            ColorDictionary * This,
            /* [retval][out] */ VARIANT_BOOL *succes);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SaveAs )( 
            ColorDictionary * This,
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *succes);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Close )( 
            ColorDictionary * This,
            /* [retval][out] */ VARIANT_BOOL *succes);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Open )( 
            ColorDictionary * This,
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *succes);
        
        END_INTERFACE
    } ColorDictionaryVtbl;

    interface ColorDictionary
    {
        CONST_VTBL struct ColorDictionaryVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ColorDictionary_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ColorDictionary_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ColorDictionary_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ColorDictionary_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ColorDictionary_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ColorDictionary_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ColorDictionary_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ColorDictionary_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ColorDictionary_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ColorDictionary_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define ColorDictionary_get_Filename(This,Name)	\
    ( (This)->lpVtbl -> get_Filename(This,Name) ) 

#define ColorDictionary_get_Overlap(This,Overlap)	\
    ( (This)->lpVtbl -> get_Overlap(This,Overlap) ) 

#define ColorDictionary_get_IsValid(This,success)	\
    ( (This)->lpVtbl -> get_IsValid(This,success) ) 

#define ColorDictionary_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define ColorDictionary_Item(This,Color,ClosestMatch,ColorEntry)	\
    ( (This)->lpVtbl -> Item(This,Color,ClosestMatch,ColorEntry) ) 

#define ColorDictionary_Add(This,Color,Translation,succes)	\
    ( (This)->lpVtbl -> Add(This,Color,Translation,succes) ) 

#define ColorDictionary_Delete(This,Color,succes)	\
    ( (This)->lpVtbl -> Delete(This,Color,succes) ) 

#define ColorDictionary_Save(This,succes)	\
    ( (This)->lpVtbl -> Save(This,succes) ) 

#define ColorDictionary_SaveAs(This,Name,succes)	\
    ( (This)->lpVtbl -> SaveAs(This,Name,succes) ) 

#define ColorDictionary_Close(This,succes)	\
    ( (This)->lpVtbl -> Close(This,succes) ) 

#define ColorDictionary_Open(This,Name,succes)	\
    ( (This)->lpVtbl -> Open(This,Name,succes) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ColorDictionary_INTERFACE_DEFINED__ */


#ifndef __ColorEntry_INTERFACE_DEFINED__
#define __ColorEntry_INTERFACE_DEFINED__

/* interface ColorEntry */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ColorEntry;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("2853890E-9D0C-4368-8E79-F9D7E88F8604")
    ColorEntry : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Value( 
            /* [in] */ Color *Color) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Value( 
            /* [retval][out] */ Color **Color) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Translation( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Translation( 
            /* [retval][out] */ BSTR *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ColorEntryVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ColorEntry * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ColorEntry * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ColorEntry * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ColorEntry * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ColorEntry * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ColorEntry * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ColorEntry * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ColorEntry * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ColorEntry * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Value )( 
            ColorEntry * This,
            /* [in] */ Color *Color);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Value )( 
            ColorEntry * This,
            /* [retval][out] */ Color **Color);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Translation )( 
            ColorEntry * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Translation )( 
            ColorEntry * This,
            /* [retval][out] */ BSTR *prop);
        
        END_INTERFACE
    } ColorEntryVtbl;

    interface ColorEntry
    {
        CONST_VTBL struct ColorEntryVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ColorEntry_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ColorEntry_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ColorEntry_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ColorEntry_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ColorEntry_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ColorEntry_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ColorEntry_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ColorEntry_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ColorEntry_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ColorEntry_put_Value(This,Color)	\
    ( (This)->lpVtbl -> put_Value(This,Color) ) 

#define ColorEntry_get_Value(This,Color)	\
    ( (This)->lpVtbl -> get_Value(This,Color) ) 

#define ColorEntry_put_Translation(This,prop)	\
    ( (This)->lpVtbl -> put_Translation(This,prop) ) 

#define ColorEntry_get_Translation(This,prop)	\
    ( (This)->lpVtbl -> get_Translation(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ColorEntry_INTERFACE_DEFINED__ */


#ifndef __ExceptionDictionary_INTERFACE_DEFINED__
#define __ExceptionDictionary_INTERFACE_DEFINED__

/* interface ExceptionDictionary */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ExceptionDictionary;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("7569E232-5C97-4CE8-A432-4A549E77A175")
    ExceptionDictionary : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Filename( 
            /* [retval][out] */ BSTR *Name) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Overlap( 
            /* [retval][out] */ _Window **Overlap) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsValid( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ BSTR Word,
            /* [retval][out] */ ExceptionEntry **ExceptionEntry) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Add( 
            /* [in] */ BSTR Word,
            /* [in] */ BSTR Translation,
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Delete( 
            /* [in] */ BSTR Word,
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Save( 
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SaveAs( 
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Close( 
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Open( 
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ExceptionDictionaryVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ExceptionDictionary * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ExceptionDictionary * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ExceptionDictionary * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ExceptionDictionary * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ExceptionDictionary * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ExceptionDictionary * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ExceptionDictionary * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ExceptionDictionary * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ExceptionDictionary * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            ExceptionDictionary * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Filename )( 
            ExceptionDictionary * This,
            /* [retval][out] */ BSTR *Name);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Overlap )( 
            ExceptionDictionary * This,
            /* [retval][out] */ _Window **Overlap);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsValid )( 
            ExceptionDictionary * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            ExceptionDictionary * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            ExceptionDictionary * This,
            /* [in] */ BSTR Word,
            /* [retval][out] */ ExceptionEntry **ExceptionEntry);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Add )( 
            ExceptionDictionary * This,
            /* [in] */ BSTR Word,
            /* [in] */ BSTR Translation,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Delete )( 
            ExceptionDictionary * This,
            /* [in] */ BSTR Word,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Save )( 
            ExceptionDictionary * This,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SaveAs )( 
            ExceptionDictionary * This,
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Close )( 
            ExceptionDictionary * This,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Open )( 
            ExceptionDictionary * This,
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        END_INTERFACE
    } ExceptionDictionaryVtbl;

    interface ExceptionDictionary
    {
        CONST_VTBL struct ExceptionDictionaryVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ExceptionDictionary_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ExceptionDictionary_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ExceptionDictionary_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ExceptionDictionary_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ExceptionDictionary_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ExceptionDictionary_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ExceptionDictionary_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ExceptionDictionary_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ExceptionDictionary_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ExceptionDictionary_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define ExceptionDictionary_get_Filename(This,Name)	\
    ( (This)->lpVtbl -> get_Filename(This,Name) ) 

#define ExceptionDictionary_get_Overlap(This,Overlap)	\
    ( (This)->lpVtbl -> get_Overlap(This,Overlap) ) 

#define ExceptionDictionary_get_IsValid(This,prop)	\
    ( (This)->lpVtbl -> get_IsValid(This,prop) ) 

#define ExceptionDictionary_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define ExceptionDictionary_Item(This,Word,ExceptionEntry)	\
    ( (This)->lpVtbl -> Item(This,Word,ExceptionEntry) ) 

#define ExceptionDictionary_Add(This,Word,Translation,success)	\
    ( (This)->lpVtbl -> Add(This,Word,Translation,success) ) 

#define ExceptionDictionary_Delete(This,Word,success)	\
    ( (This)->lpVtbl -> Delete(This,Word,success) ) 

#define ExceptionDictionary_Save(This,success)	\
    ( (This)->lpVtbl -> Save(This,success) ) 

#define ExceptionDictionary_SaveAs(This,Name,success)	\
    ( (This)->lpVtbl -> SaveAs(This,Name,success) ) 

#define ExceptionDictionary_Close(This,success)	\
    ( (This)->lpVtbl -> Close(This,success) ) 

#define ExceptionDictionary_Open(This,Name,success)	\
    ( (This)->lpVtbl -> Open(This,Name,success) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ExceptionDictionary_INTERFACE_DEFINED__ */


#ifndef __ExceptionEntry_INTERFACE_DEFINED__
#define __ExceptionEntry_INTERFACE_DEFINED__

/* interface ExceptionEntry */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ExceptionEntry;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("DD771ED5-2FDA-4C0F-BF4E-D33A379D885B")
    ExceptionEntry : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Word( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Word( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Translation( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Translation( 
            /* [retval][out] */ BSTR *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ExceptionEntryVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ExceptionEntry * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ExceptionEntry * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ExceptionEntry * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ExceptionEntry * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ExceptionEntry * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ExceptionEntry * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ExceptionEntry * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ExceptionEntry * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ExceptionEntry * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Word )( 
            ExceptionEntry * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Word )( 
            ExceptionEntry * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Translation )( 
            ExceptionEntry * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Translation )( 
            ExceptionEntry * This,
            /* [retval][out] */ BSTR *prop);
        
        END_INTERFACE
    } ExceptionEntryVtbl;

    interface ExceptionEntry
    {
        CONST_VTBL struct ExceptionEntryVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ExceptionEntry_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ExceptionEntry_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ExceptionEntry_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ExceptionEntry_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ExceptionEntry_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ExceptionEntry_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ExceptionEntry_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ExceptionEntry_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ExceptionEntry_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ExceptionEntry_put_Word(This,prop)	\
    ( (This)->lpVtbl -> put_Word(This,prop) ) 

#define ExceptionEntry_get_Word(This,prop)	\
    ( (This)->lpVtbl -> get_Word(This,prop) ) 

#define ExceptionEntry_put_Translation(This,prop)	\
    ( (This)->lpVtbl -> put_Translation(This,prop) ) 

#define ExceptionEntry_get_Translation(This,prop)	\
    ( (This)->lpVtbl -> get_Translation(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ExceptionEntry_INTERFACE_DEFINED__ */


#ifndef __GraphicDictionary_INTERFACE_DEFINED__
#define __GraphicDictionary_INTERFACE_DEFINED__

/* interface GraphicDictionary */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_GraphicDictionary;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("61CF7E57-96B5-40C8-823A-A25CC29C7F93")
    GraphicDictionary : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Filename( 
            /* [retval][out] */ BSTR *Name) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Overlap( 
            /* [retval][out] */ _Window **Overlap) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsValid( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long CRC,
            /* [retval][out] */ GraphicEntry **GraphicEntry) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Add( 
            /* [in] */ unsigned long CRC,
            /* [in] */ BSTR SpeechLabel,
            /* [in] */ BSTR BrailleLabel,
            /* [in] */ VARIANT_BOOL SpeechSilence,
            /* [in] */ VARIANT_BOOL ExcludeFromLines,
            /* [in] */ VARIANT_BOOL BrailleSilence,
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Delete( 
            /* [in] */ long CRC,
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Save( 
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SaveAs( 
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Close( 
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Open( 
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct GraphicDictionaryVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            GraphicDictionary * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            GraphicDictionary * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            GraphicDictionary * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            GraphicDictionary * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            GraphicDictionary * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            GraphicDictionary * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            GraphicDictionary * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            GraphicDictionary * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            GraphicDictionary * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            GraphicDictionary * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Filename )( 
            GraphicDictionary * This,
            /* [retval][out] */ BSTR *Name);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Overlap )( 
            GraphicDictionary * This,
            /* [retval][out] */ _Window **Overlap);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsValid )( 
            GraphicDictionary * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            GraphicDictionary * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            GraphicDictionary * This,
            /* [in] */ long CRC,
            /* [retval][out] */ GraphicEntry **GraphicEntry);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Add )( 
            GraphicDictionary * This,
            /* [in] */ unsigned long CRC,
            /* [in] */ BSTR SpeechLabel,
            /* [in] */ BSTR BrailleLabel,
            /* [in] */ VARIANT_BOOL SpeechSilence,
            /* [in] */ VARIANT_BOOL ExcludeFromLines,
            /* [in] */ VARIANT_BOOL BrailleSilence,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Delete )( 
            GraphicDictionary * This,
            /* [in] */ long CRC,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Save )( 
            GraphicDictionary * This,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SaveAs )( 
            GraphicDictionary * This,
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Close )( 
            GraphicDictionary * This,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Open )( 
            GraphicDictionary * This,
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        END_INTERFACE
    } GraphicDictionaryVtbl;

    interface GraphicDictionary
    {
        CONST_VTBL struct GraphicDictionaryVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define GraphicDictionary_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define GraphicDictionary_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define GraphicDictionary_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define GraphicDictionary_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define GraphicDictionary_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define GraphicDictionary_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define GraphicDictionary_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define GraphicDictionary_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define GraphicDictionary_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define GraphicDictionary_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define GraphicDictionary_get_Filename(This,Name)	\
    ( (This)->lpVtbl -> get_Filename(This,Name) ) 

#define GraphicDictionary_get_Overlap(This,Overlap)	\
    ( (This)->lpVtbl -> get_Overlap(This,Overlap) ) 

#define GraphicDictionary_get_IsValid(This,prop)	\
    ( (This)->lpVtbl -> get_IsValid(This,prop) ) 

#define GraphicDictionary_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define GraphicDictionary_Item(This,CRC,GraphicEntry)	\
    ( (This)->lpVtbl -> Item(This,CRC,GraphicEntry) ) 

#define GraphicDictionary_Add(This,CRC,SpeechLabel,BrailleLabel,SpeechSilence,ExcludeFromLines,BrailleSilence,success)	\
    ( (This)->lpVtbl -> Add(This,CRC,SpeechLabel,BrailleLabel,SpeechSilence,ExcludeFromLines,BrailleSilence,success) ) 

#define GraphicDictionary_Delete(This,CRC,success)	\
    ( (This)->lpVtbl -> Delete(This,CRC,success) ) 

#define GraphicDictionary_Save(This,success)	\
    ( (This)->lpVtbl -> Save(This,success) ) 

#define GraphicDictionary_SaveAs(This,Name,success)	\
    ( (This)->lpVtbl -> SaveAs(This,Name,success) ) 

#define GraphicDictionary_Close(This,success)	\
    ( (This)->lpVtbl -> Close(This,success) ) 

#define GraphicDictionary_Open(This,Name,success)	\
    ( (This)->lpVtbl -> Open(This,Name,success) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __GraphicDictionary_INTERFACE_DEFINED__ */


#ifndef __GraphicEntry_INTERFACE_DEFINED__
#define __GraphicEntry_INTERFACE_DEFINED__

/* interface GraphicEntry */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_GraphicEntry;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("E60B1E71-548F-4B34-BD81-99003EA6922A")
    GraphicEntry : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CRC( 
            /* [in] */ long CRC) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CRC( 
            /* [retval][out] */ long *CRC) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpeechLabel( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpeechLabel( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_BrailleLabel( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BrailleLabel( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpeechSilence( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpeechSilence( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_BrailleSilence( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BrailleSilence( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ExcludeFromLines( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ExcludeFromLines( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct GraphicEntryVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            GraphicEntry * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            GraphicEntry * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            GraphicEntry * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            GraphicEntry * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            GraphicEntry * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            GraphicEntry * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            GraphicEntry * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            GraphicEntry * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            GraphicEntry * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CRC )( 
            GraphicEntry * This,
            /* [in] */ long CRC);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CRC )( 
            GraphicEntry * This,
            /* [retval][out] */ long *CRC);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpeechLabel )( 
            GraphicEntry * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpeechLabel )( 
            GraphicEntry * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_BrailleLabel )( 
            GraphicEntry * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BrailleLabel )( 
            GraphicEntry * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpeechSilence )( 
            GraphicEntry * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpeechSilence )( 
            GraphicEntry * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_BrailleSilence )( 
            GraphicEntry * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BrailleSilence )( 
            GraphicEntry * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ExcludeFromLines )( 
            GraphicEntry * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ExcludeFromLines )( 
            GraphicEntry * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } GraphicEntryVtbl;

    interface GraphicEntry
    {
        CONST_VTBL struct GraphicEntryVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define GraphicEntry_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define GraphicEntry_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define GraphicEntry_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define GraphicEntry_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define GraphicEntry_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define GraphicEntry_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define GraphicEntry_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define GraphicEntry_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define GraphicEntry_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define GraphicEntry_put_CRC(This,CRC)	\
    ( (This)->lpVtbl -> put_CRC(This,CRC) ) 

#define GraphicEntry_get_CRC(This,CRC)	\
    ( (This)->lpVtbl -> get_CRC(This,CRC) ) 

#define GraphicEntry_put_SpeechLabel(This,prop)	\
    ( (This)->lpVtbl -> put_SpeechLabel(This,prop) ) 

#define GraphicEntry_get_SpeechLabel(This,prop)	\
    ( (This)->lpVtbl -> get_SpeechLabel(This,prop) ) 

#define GraphicEntry_put_BrailleLabel(This,prop)	\
    ( (This)->lpVtbl -> put_BrailleLabel(This,prop) ) 

#define GraphicEntry_get_BrailleLabel(This,prop)	\
    ( (This)->lpVtbl -> get_BrailleLabel(This,prop) ) 

#define GraphicEntry_put_SpeechSilence(This,prop)	\
    ( (This)->lpVtbl -> put_SpeechSilence(This,prop) ) 

#define GraphicEntry_get_SpeechSilence(This,prop)	\
    ( (This)->lpVtbl -> get_SpeechSilence(This,prop) ) 

#define GraphicEntry_put_BrailleSilence(This,prop)	\
    ( (This)->lpVtbl -> put_BrailleSilence(This,prop) ) 

#define GraphicEntry_get_BrailleSilence(This,prop)	\
    ( (This)->lpVtbl -> get_BrailleSilence(This,prop) ) 

#define GraphicEntry_put_ExcludeFromLines(This,prop)	\
    ( (This)->lpVtbl -> put_ExcludeFromLines(This,prop) ) 

#define GraphicEntry_get_ExcludeFromLines(This,prop)	\
    ( (This)->lpVtbl -> get_ExcludeFromLines(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __GraphicEntry_INTERFACE_DEFINED__ */


#ifndef __KeyLabelDictionary_INTERFACE_DEFINED__
#define __KeyLabelDictionary_INTERFACE_DEFINED__

/* interface KeyLabelDictionary */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_KeyLabelDictionary;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("781865FC-9E31-46C9-8140-E4AF15A8170B")
    KeyLabelDictionary : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Filename( 
            /* [retval][out] */ BSTR *Name) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Overlap( 
            /* [retval][out] */ _Window **Overlap) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsValid( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ VirtualKeyCode Key,
            /* [in] */ KeyModifiers mods,
            /* [retval][out] */ KeyLabelEntry **KeyLabelEntry) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Add( 
            /* [in] */ VirtualKeyCode Key,
            /* [in] */ KeyModifiers mods,
            /* [in] */ BSTR Translation,
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Delete( 
            /* [in] */ VirtualKeyCode Key,
            /* [in] */ KeyModifiers mods,
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Save( 
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SaveAs( 
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Close( 
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Open( 
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct KeyLabelDictionaryVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            KeyLabelDictionary * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            KeyLabelDictionary * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            KeyLabelDictionary * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            KeyLabelDictionary * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            KeyLabelDictionary * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            KeyLabelDictionary * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            KeyLabelDictionary * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            KeyLabelDictionary * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            KeyLabelDictionary * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            KeyLabelDictionary * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Filename )( 
            KeyLabelDictionary * This,
            /* [retval][out] */ BSTR *Name);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Overlap )( 
            KeyLabelDictionary * This,
            /* [retval][out] */ _Window **Overlap);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsValid )( 
            KeyLabelDictionary * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            KeyLabelDictionary * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            KeyLabelDictionary * This,
            /* [in] */ VirtualKeyCode Key,
            /* [in] */ KeyModifiers mods,
            /* [retval][out] */ KeyLabelEntry **KeyLabelEntry);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Add )( 
            KeyLabelDictionary * This,
            /* [in] */ VirtualKeyCode Key,
            /* [in] */ KeyModifiers mods,
            /* [in] */ BSTR Translation,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Delete )( 
            KeyLabelDictionary * This,
            /* [in] */ VirtualKeyCode Key,
            /* [in] */ KeyModifiers mods,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Save )( 
            KeyLabelDictionary * This,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SaveAs )( 
            KeyLabelDictionary * This,
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Close )( 
            KeyLabelDictionary * This,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Open )( 
            KeyLabelDictionary * This,
            /* [in] */ BSTR Name,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        END_INTERFACE
    } KeyLabelDictionaryVtbl;

    interface KeyLabelDictionary
    {
        CONST_VTBL struct KeyLabelDictionaryVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define KeyLabelDictionary_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define KeyLabelDictionary_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define KeyLabelDictionary_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define KeyLabelDictionary_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define KeyLabelDictionary_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define KeyLabelDictionary_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define KeyLabelDictionary_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define KeyLabelDictionary_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define KeyLabelDictionary_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define KeyLabelDictionary_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define KeyLabelDictionary_get_Filename(This,Name)	\
    ( (This)->lpVtbl -> get_Filename(This,Name) ) 

#define KeyLabelDictionary_get_Overlap(This,Overlap)	\
    ( (This)->lpVtbl -> get_Overlap(This,Overlap) ) 

#define KeyLabelDictionary_get_IsValid(This,prop)	\
    ( (This)->lpVtbl -> get_IsValid(This,prop) ) 

#define KeyLabelDictionary_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define KeyLabelDictionary_Item(This,Key,mods,KeyLabelEntry)	\
    ( (This)->lpVtbl -> Item(This,Key,mods,KeyLabelEntry) ) 

#define KeyLabelDictionary_Add(This,Key,mods,Translation,success)	\
    ( (This)->lpVtbl -> Add(This,Key,mods,Translation,success) ) 

#define KeyLabelDictionary_Delete(This,Key,mods,success)	\
    ( (This)->lpVtbl -> Delete(This,Key,mods,success) ) 

#define KeyLabelDictionary_Save(This,success)	\
    ( (This)->lpVtbl -> Save(This,success) ) 

#define KeyLabelDictionary_SaveAs(This,Name,success)	\
    ( (This)->lpVtbl -> SaveAs(This,Name,success) ) 

#define KeyLabelDictionary_Close(This,success)	\
    ( (This)->lpVtbl -> Close(This,success) ) 

#define KeyLabelDictionary_Open(This,Name,success)	\
    ( (This)->lpVtbl -> Open(This,Name,success) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __KeyLabelDictionary_INTERFACE_DEFINED__ */


#ifndef __KeyLabelEntry_INTERFACE_DEFINED__
#define __KeyLabelEntry_INTERFACE_DEFINED__

/* interface KeyLabelEntry */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_KeyLabelEntry;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("8EE54CC0-2617-450E-BFF0-55356DA30A9C")
    KeyLabelEntry : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Key( 
            /* [in] */ VirtualKeyCode Key) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Key( 
            /* [retval][out] */ VirtualKeyCode *Key) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_KeyModifiers( 
            /* [in] */ KeyModifiers mods) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_KeyModifiers( 
            /* [retval][out] */ KeyModifiers *mods) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Translation( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Translation( 
            /* [retval][out] */ BSTR *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct KeyLabelEntryVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            KeyLabelEntry * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            KeyLabelEntry * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            KeyLabelEntry * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            KeyLabelEntry * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            KeyLabelEntry * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            KeyLabelEntry * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            KeyLabelEntry * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            KeyLabelEntry * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            KeyLabelEntry * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Key )( 
            KeyLabelEntry * This,
            /* [in] */ VirtualKeyCode Key);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Key )( 
            KeyLabelEntry * This,
            /* [retval][out] */ VirtualKeyCode *Key);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_KeyModifiers )( 
            KeyLabelEntry * This,
            /* [in] */ KeyModifiers mods);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_KeyModifiers )( 
            KeyLabelEntry * This,
            /* [retval][out] */ KeyModifiers *mods);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Translation )( 
            KeyLabelEntry * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Translation )( 
            KeyLabelEntry * This,
            /* [retval][out] */ BSTR *prop);
        
        END_INTERFACE
    } KeyLabelEntryVtbl;

    interface KeyLabelEntry
    {
        CONST_VTBL struct KeyLabelEntryVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define KeyLabelEntry_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define KeyLabelEntry_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define KeyLabelEntry_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define KeyLabelEntry_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define KeyLabelEntry_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define KeyLabelEntry_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define KeyLabelEntry_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define KeyLabelEntry_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define KeyLabelEntry_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define KeyLabelEntry_put_Key(This,Key)	\
    ( (This)->lpVtbl -> put_Key(This,Key) ) 

#define KeyLabelEntry_get_Key(This,Key)	\
    ( (This)->lpVtbl -> get_Key(This,Key) ) 

#define KeyLabelEntry_put_KeyModifiers(This,mods)	\
    ( (This)->lpVtbl -> put_KeyModifiers(This,mods) ) 

#define KeyLabelEntry_get_KeyModifiers(This,mods)	\
    ( (This)->lpVtbl -> get_KeyModifiers(This,mods) ) 

#define KeyLabelEntry_put_Translation(This,prop)	\
    ( (This)->lpVtbl -> put_Translation(This,prop) ) 

#define KeyLabelEntry_get_Translation(This,prop)	\
    ( (This)->lpVtbl -> get_Translation(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __KeyLabelEntry_INTERFACE_DEFINED__ */


#ifndef __GlobalSettings_INTERFACE_DEFINED__
#define __GlobalSettings_INTERFACE_DEFINED__

/* interface GlobalSettings */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_GlobalSettings;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("D5C04D9E-F114-43BD-89C8-5277182A8A56")
    GlobalSettings : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MenuLevel( 
            /* [retval][out] */ MenuMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_MenuLevel( 
            /* [in] */ MenuMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_GlobalSettings( 
            /* [retval][out] */ GlobalMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_GlobalSettings( 
            /* [in] */ GlobalMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HyperactiveStatus( 
            /* [retval][out] */ HyperactiveMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_HyperactiveStatus( 
            /* [in] */ HyperactiveMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AutoloadStatus( 
            /* [retval][out] */ AutoLoadMode *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AutoloadStatus( 
            /* [in] */ AutoLoadMode prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ShowUserWindowOutline( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ShowUserWindowOutline( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CursorBlinkRate( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CursorBlinkRate( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MinimumGraphicSize( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_MinimumGraphicSize( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MaximumGraphicSize( 
            /* [retval][out] */ short *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_MaximumGraphicSize( 
            /* [in] */ short prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UseClipBoundariesWithMouse( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_UseClipBoundariesWithMouse( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EnhancedSupportForCurrentWordDocument( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EnhancedSupportForCurrentWordDocument( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct GlobalSettingsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            GlobalSettings * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            GlobalSettings * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            GlobalSettings * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            GlobalSettings * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            GlobalSettings * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            GlobalSettings * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            GlobalSettings * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            GlobalSettings * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            GlobalSettings * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MenuLevel )( 
            GlobalSettings * This,
            /* [retval][out] */ MenuMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_MenuLevel )( 
            GlobalSettings * This,
            /* [in] */ MenuMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_GlobalSettings )( 
            GlobalSettings * This,
            /* [retval][out] */ GlobalMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_GlobalSettings )( 
            GlobalSettings * This,
            /* [in] */ GlobalMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HyperactiveStatus )( 
            GlobalSettings * This,
            /* [retval][out] */ HyperactiveMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_HyperactiveStatus )( 
            GlobalSettings * This,
            /* [in] */ HyperactiveMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AutoloadStatus )( 
            GlobalSettings * This,
            /* [retval][out] */ AutoLoadMode *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AutoloadStatus )( 
            GlobalSettings * This,
            /* [in] */ AutoLoadMode prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ShowUserWindowOutline )( 
            GlobalSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ShowUserWindowOutline )( 
            GlobalSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CursorBlinkRate )( 
            GlobalSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CursorBlinkRate )( 
            GlobalSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MinimumGraphicSize )( 
            GlobalSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_MinimumGraphicSize )( 
            GlobalSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MaximumGraphicSize )( 
            GlobalSettings * This,
            /* [retval][out] */ short *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_MaximumGraphicSize )( 
            GlobalSettings * This,
            /* [in] */ short prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UseClipBoundariesWithMouse )( 
            GlobalSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_UseClipBoundariesWithMouse )( 
            GlobalSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EnhancedSupportForCurrentWordDocument )( 
            GlobalSettings * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EnhancedSupportForCurrentWordDocument )( 
            GlobalSettings * This,
            /* [in] */ VARIANT_BOOL prop);
        
        END_INTERFACE
    } GlobalSettingsVtbl;

    interface GlobalSettings
    {
        CONST_VTBL struct GlobalSettingsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define GlobalSettings_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define GlobalSettings_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define GlobalSettings_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define GlobalSettings_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define GlobalSettings_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define GlobalSettings_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define GlobalSettings_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define GlobalSettings_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define GlobalSettings_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define GlobalSettings_get_MenuLevel(This,prop)	\
    ( (This)->lpVtbl -> get_MenuLevel(This,prop) ) 

#define GlobalSettings_put_MenuLevel(This,prop)	\
    ( (This)->lpVtbl -> put_MenuLevel(This,prop) ) 

#define GlobalSettings_get_GlobalSettings(This,prop)	\
    ( (This)->lpVtbl -> get_GlobalSettings(This,prop) ) 

#define GlobalSettings_put_GlobalSettings(This,prop)	\
    ( (This)->lpVtbl -> put_GlobalSettings(This,prop) ) 

#define GlobalSettings_get_HyperactiveStatus(This,prop)	\
    ( (This)->lpVtbl -> get_HyperactiveStatus(This,prop) ) 

#define GlobalSettings_put_HyperactiveStatus(This,prop)	\
    ( (This)->lpVtbl -> put_HyperactiveStatus(This,prop) ) 

#define GlobalSettings_get_AutoloadStatus(This,prop)	\
    ( (This)->lpVtbl -> get_AutoloadStatus(This,prop) ) 

#define GlobalSettings_put_AutoloadStatus(This,prop)	\
    ( (This)->lpVtbl -> put_AutoloadStatus(This,prop) ) 

#define GlobalSettings_get_ShowUserWindowOutline(This,prop)	\
    ( (This)->lpVtbl -> get_ShowUserWindowOutline(This,prop) ) 

#define GlobalSettings_put_ShowUserWindowOutline(This,prop)	\
    ( (This)->lpVtbl -> put_ShowUserWindowOutline(This,prop) ) 

#define GlobalSettings_get_CursorBlinkRate(This,prop)	\
    ( (This)->lpVtbl -> get_CursorBlinkRate(This,prop) ) 

#define GlobalSettings_put_CursorBlinkRate(This,prop)	\
    ( (This)->lpVtbl -> put_CursorBlinkRate(This,prop) ) 

#define GlobalSettings_get_MinimumGraphicSize(This,prop)	\
    ( (This)->lpVtbl -> get_MinimumGraphicSize(This,prop) ) 

#define GlobalSettings_put_MinimumGraphicSize(This,prop)	\
    ( (This)->lpVtbl -> put_MinimumGraphicSize(This,prop) ) 

#define GlobalSettings_get_MaximumGraphicSize(This,prop)	\
    ( (This)->lpVtbl -> get_MaximumGraphicSize(This,prop) ) 

#define GlobalSettings_put_MaximumGraphicSize(This,prop)	\
    ( (This)->lpVtbl -> put_MaximumGraphicSize(This,prop) ) 

#define GlobalSettings_get_UseClipBoundariesWithMouse(This,prop)	\
    ( (This)->lpVtbl -> get_UseClipBoundariesWithMouse(This,prop) ) 

#define GlobalSettings_put_UseClipBoundariesWithMouse(This,prop)	\
    ( (This)->lpVtbl -> put_UseClipBoundariesWithMouse(This,prop) ) 

#define GlobalSettings_get_EnhancedSupportForCurrentWordDocument(This,prop)	\
    ( (This)->lpVtbl -> get_EnhancedSupportForCurrentWordDocument(This,prop) ) 

#define GlobalSettings_put_EnhancedSupportForCurrentWordDocument(This,prop)	\
    ( (This)->lpVtbl -> put_EnhancedSupportForCurrentWordDocument(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __GlobalSettings_INTERFACE_DEFINED__ */


#ifndef __SetFileScope_INTERFACE_DEFINED__
#define __SetFileScope_INTERFACE_DEFINED__

/* interface SetFileScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_SetFileScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("201A7966-E51C-48FD-BE98-E6F92F1D1116")
    SetFileScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Screen( 
            /* [retval][out] */ ScreenSettingsScope **Screen) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Keyboard( 
            /* [retval][out] */ KeyboardSettingsScope **Keyboard) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Mouse( 
            /* [retval][out] */ MouseSettingsScope **Mouse) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_General( 
            /* [retval][out] */ GeneralSettingsScope **General) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Braille( 
            /* [retval][out] */ BrailleSettingsScope **Braille) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HotKeys( 
            /* [retval][out] */ HotkeysScope **HotKeys) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Verbosity( 
            /* [retval][out] */ VerbosityScope **Verbosity) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct SetFileScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            SetFileScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            SetFileScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            SetFileScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            SetFileScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            SetFileScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            SetFileScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            SetFileScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            SetFileScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            SetFileScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Screen )( 
            SetFileScope * This,
            /* [retval][out] */ ScreenSettingsScope **Screen);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Keyboard )( 
            SetFileScope * This,
            /* [retval][out] */ KeyboardSettingsScope **Keyboard);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Mouse )( 
            SetFileScope * This,
            /* [retval][out] */ MouseSettingsScope **Mouse);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_General )( 
            SetFileScope * This,
            /* [retval][out] */ GeneralSettingsScope **General);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Braille )( 
            SetFileScope * This,
            /* [retval][out] */ BrailleSettingsScope **Braille);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HotKeys )( 
            SetFileScope * This,
            /* [retval][out] */ HotkeysScope **HotKeys);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Verbosity )( 
            SetFileScope * This,
            /* [retval][out] */ VerbosityScope **Verbosity);
        
        END_INTERFACE
    } SetFileScopeVtbl;

    interface SetFileScope
    {
        CONST_VTBL struct SetFileScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define SetFileScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define SetFileScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define SetFileScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define SetFileScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define SetFileScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define SetFileScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define SetFileScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define SetFileScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define SetFileScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define SetFileScope_get_Screen(This,Screen)	\
    ( (This)->lpVtbl -> get_Screen(This,Screen) ) 

#define SetFileScope_get_Keyboard(This,Keyboard)	\
    ( (This)->lpVtbl -> get_Keyboard(This,Keyboard) ) 

#define SetFileScope_get_Mouse(This,Mouse)	\
    ( (This)->lpVtbl -> get_Mouse(This,Mouse) ) 

#define SetFileScope_get_General(This,General)	\
    ( (This)->lpVtbl -> get_General(This,General) ) 

#define SetFileScope_get_Braille(This,Braille)	\
    ( (This)->lpVtbl -> get_Braille(This,Braille) ) 

#define SetFileScope_get_HotKeys(This,HotKeys)	\
    ( (This)->lpVtbl -> get_HotKeys(This,HotKeys) ) 

#define SetFileScope_get_Verbosity(This,Verbosity)	\
    ( (This)->lpVtbl -> get_Verbosity(This,Verbosity) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __SetFileScope_INTERFACE_DEFINED__ */


#ifndef __ScreenSettingsScope_INTERFACE_DEFINED__
#define __ScreenSettingsScope_INTERFACE_DEFINED__

/* interface ScreenSettingsScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_ScreenSettingsScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("E5EF0F40-35D1-4581-B5B9-F178CA9ABF5D")
    ScreenSettingsScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Voice( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Voice( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Rate( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Rate( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Pitch( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Pitch( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Tone( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Tone( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Volume( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Volume( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Punctuation( 
            /* [retval][out] */ PunctuationSettingsScope **punc) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Numbers( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Numbers( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CapitalizationAlert( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CapitalizationAlert( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FormatAlert( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FormatAlert( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpellAlphaNumeric( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpellAlphaNumeric( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SplitMixedCaseWords( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SplitMixedCaseWords( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_RepeatFilter( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_RepeatFilter( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BlankLines( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_BlankLines( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SignalCursorPosition( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SignalCursorPosition( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AttributeChanges( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AttributeChanges( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadParagraphStyle( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadParagraphStyle( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LineFilter( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_LineFilter( 
            /* [in] */ SetFileScopes prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ScreenSettingsScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ScreenSettingsScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ScreenSettingsScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ScreenSettingsScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ScreenSettingsScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ScreenSettingsScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ScreenSettingsScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ScreenSettingsScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Voice )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Voice )( 
            ScreenSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Rate )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Rate )( 
            ScreenSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Pitch )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Pitch )( 
            ScreenSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Tone )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Tone )( 
            ScreenSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Volume )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Volume )( 
            ScreenSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Punctuation )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ PunctuationSettingsScope **punc);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Numbers )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Numbers )( 
            ScreenSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CapitalizationAlert )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CapitalizationAlert )( 
            ScreenSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FormatAlert )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FormatAlert )( 
            ScreenSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpellAlphaNumeric )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpellAlphaNumeric )( 
            ScreenSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SplitMixedCaseWords )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SplitMixedCaseWords )( 
            ScreenSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_RepeatFilter )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_RepeatFilter )( 
            ScreenSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BlankLines )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_BlankLines )( 
            ScreenSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SignalCursorPosition )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SignalCursorPosition )( 
            ScreenSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AttributeChanges )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AttributeChanges )( 
            ScreenSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadParagraphStyle )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadParagraphStyle )( 
            ScreenSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LineFilter )( 
            ScreenSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_LineFilter )( 
            ScreenSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        END_INTERFACE
    } ScreenSettingsScopeVtbl;

    interface ScreenSettingsScope
    {
        CONST_VTBL struct ScreenSettingsScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ScreenSettingsScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ScreenSettingsScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ScreenSettingsScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ScreenSettingsScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ScreenSettingsScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ScreenSettingsScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ScreenSettingsScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ScreenSettingsScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ScreenSettingsScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ScreenSettingsScope_get_Voice(This,prop)	\
    ( (This)->lpVtbl -> get_Voice(This,prop) ) 

#define ScreenSettingsScope_put_Voice(This,prop)	\
    ( (This)->lpVtbl -> put_Voice(This,prop) ) 

#define ScreenSettingsScope_get_Rate(This,prop)	\
    ( (This)->lpVtbl -> get_Rate(This,prop) ) 

#define ScreenSettingsScope_put_Rate(This,prop)	\
    ( (This)->lpVtbl -> put_Rate(This,prop) ) 

#define ScreenSettingsScope_get_Pitch(This,prop)	\
    ( (This)->lpVtbl -> get_Pitch(This,prop) ) 

#define ScreenSettingsScope_put_Pitch(This,prop)	\
    ( (This)->lpVtbl -> put_Pitch(This,prop) ) 

#define ScreenSettingsScope_get_Tone(This,prop)	\
    ( (This)->lpVtbl -> get_Tone(This,prop) ) 

#define ScreenSettingsScope_put_Tone(This,prop)	\
    ( (This)->lpVtbl -> put_Tone(This,prop) ) 

#define ScreenSettingsScope_get_Volume(This,prop)	\
    ( (This)->lpVtbl -> get_Volume(This,prop) ) 

#define ScreenSettingsScope_put_Volume(This,prop)	\
    ( (This)->lpVtbl -> put_Volume(This,prop) ) 

#define ScreenSettingsScope_get_Punctuation(This,punc)	\
    ( (This)->lpVtbl -> get_Punctuation(This,punc) ) 

#define ScreenSettingsScope_get_Numbers(This,prop)	\
    ( (This)->lpVtbl -> get_Numbers(This,prop) ) 

#define ScreenSettingsScope_put_Numbers(This,prop)	\
    ( (This)->lpVtbl -> put_Numbers(This,prop) ) 

#define ScreenSettingsScope_get_CapitalizationAlert(This,prop)	\
    ( (This)->lpVtbl -> get_CapitalizationAlert(This,prop) ) 

#define ScreenSettingsScope_put_CapitalizationAlert(This,prop)	\
    ( (This)->lpVtbl -> put_CapitalizationAlert(This,prop) ) 

#define ScreenSettingsScope_get_FormatAlert(This,prop)	\
    ( (This)->lpVtbl -> get_FormatAlert(This,prop) ) 

#define ScreenSettingsScope_put_FormatAlert(This,prop)	\
    ( (This)->lpVtbl -> put_FormatAlert(This,prop) ) 

#define ScreenSettingsScope_get_SpellAlphaNumeric(This,prop)	\
    ( (This)->lpVtbl -> get_SpellAlphaNumeric(This,prop) ) 

#define ScreenSettingsScope_put_SpellAlphaNumeric(This,prop)	\
    ( (This)->lpVtbl -> put_SpellAlphaNumeric(This,prop) ) 

#define ScreenSettingsScope_get_SplitMixedCaseWords(This,prop)	\
    ( (This)->lpVtbl -> get_SplitMixedCaseWords(This,prop) ) 

#define ScreenSettingsScope_put_SplitMixedCaseWords(This,prop)	\
    ( (This)->lpVtbl -> put_SplitMixedCaseWords(This,prop) ) 

#define ScreenSettingsScope_get_RepeatFilter(This,prop)	\
    ( (This)->lpVtbl -> get_RepeatFilter(This,prop) ) 

#define ScreenSettingsScope_put_RepeatFilter(This,prop)	\
    ( (This)->lpVtbl -> put_RepeatFilter(This,prop) ) 

#define ScreenSettingsScope_get_BlankLines(This,prop)	\
    ( (This)->lpVtbl -> get_BlankLines(This,prop) ) 

#define ScreenSettingsScope_put_BlankLines(This,prop)	\
    ( (This)->lpVtbl -> put_BlankLines(This,prop) ) 

#define ScreenSettingsScope_get_SignalCursorPosition(This,prop)	\
    ( (This)->lpVtbl -> get_SignalCursorPosition(This,prop) ) 

#define ScreenSettingsScope_put_SignalCursorPosition(This,prop)	\
    ( (This)->lpVtbl -> put_SignalCursorPosition(This,prop) ) 

#define ScreenSettingsScope_get_AttributeChanges(This,prop)	\
    ( (This)->lpVtbl -> get_AttributeChanges(This,prop) ) 

#define ScreenSettingsScope_put_AttributeChanges(This,prop)	\
    ( (This)->lpVtbl -> put_AttributeChanges(This,prop) ) 

#define ScreenSettingsScope_get_ReadParagraphStyle(This,prop)	\
    ( (This)->lpVtbl -> get_ReadParagraphStyle(This,prop) ) 

#define ScreenSettingsScope_put_ReadParagraphStyle(This,prop)	\
    ( (This)->lpVtbl -> put_ReadParagraphStyle(This,prop) ) 

#define ScreenSettingsScope_get_LineFilter(This,prop)	\
    ( (This)->lpVtbl -> get_LineFilter(This,prop) ) 

#define ScreenSettingsScope_put_LineFilter(This,prop)	\
    ( (This)->lpVtbl -> put_LineFilter(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ScreenSettingsScope_INTERFACE_DEFINED__ */


#ifndef __PunctuationSettingsScope_INTERFACE_DEFINED__
#define __PunctuationSettingsScope_INTERFACE_DEFINED__

/* interface PunctuationSettingsScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_PunctuationSettingsScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("9FBAB55B-0436-4157-B22F-EE7F2ED9B6CD")
    PunctuationSettingsScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Textual( 
            /* [retval][out] */ SetFileScopes *Scope) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Textual( 
            /* [in] */ SetFileScopes Scope) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Math( 
            /* [retval][out] */ SetFileScopes *Scope) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Math( 
            /* [in] */ SetFileScopes Scope) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Miscellaneous( 
            /* [retval][out] */ SetFileScopes *Scope) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Miscellaneous( 
            /* [in] */ SetFileScopes Scope) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Space( 
            /* [retval][out] */ SetFileScopes *Scope) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Space( 
            /* [in] */ SetFileScopes Scope) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct PunctuationSettingsScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            PunctuationSettingsScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            PunctuationSettingsScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            PunctuationSettingsScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            PunctuationSettingsScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            PunctuationSettingsScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            PunctuationSettingsScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            PunctuationSettingsScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            PunctuationSettingsScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            PunctuationSettingsScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Textual )( 
            PunctuationSettingsScope * This,
            /* [retval][out] */ SetFileScopes *Scope);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Textual )( 
            PunctuationSettingsScope * This,
            /* [in] */ SetFileScopes Scope);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Math )( 
            PunctuationSettingsScope * This,
            /* [retval][out] */ SetFileScopes *Scope);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Math )( 
            PunctuationSettingsScope * This,
            /* [in] */ SetFileScopes Scope);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Miscellaneous )( 
            PunctuationSettingsScope * This,
            /* [retval][out] */ SetFileScopes *Scope);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Miscellaneous )( 
            PunctuationSettingsScope * This,
            /* [in] */ SetFileScopes Scope);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Space )( 
            PunctuationSettingsScope * This,
            /* [retval][out] */ SetFileScopes *Scope);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Space )( 
            PunctuationSettingsScope * This,
            /* [in] */ SetFileScopes Scope);
        
        END_INTERFACE
    } PunctuationSettingsScopeVtbl;

    interface PunctuationSettingsScope
    {
        CONST_VTBL struct PunctuationSettingsScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define PunctuationSettingsScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define PunctuationSettingsScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define PunctuationSettingsScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define PunctuationSettingsScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define PunctuationSettingsScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define PunctuationSettingsScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define PunctuationSettingsScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define PunctuationSettingsScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define PunctuationSettingsScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define PunctuationSettingsScope_get_Textual(This,Scope)	\
    ( (This)->lpVtbl -> get_Textual(This,Scope) ) 

#define PunctuationSettingsScope_put_Textual(This,Scope)	\
    ( (This)->lpVtbl -> put_Textual(This,Scope) ) 

#define PunctuationSettingsScope_get_Math(This,Scope)	\
    ( (This)->lpVtbl -> get_Math(This,Scope) ) 

#define PunctuationSettingsScope_put_Math(This,Scope)	\
    ( (This)->lpVtbl -> put_Math(This,Scope) ) 

#define PunctuationSettingsScope_get_Miscellaneous(This,Scope)	\
    ( (This)->lpVtbl -> get_Miscellaneous(This,Scope) ) 

#define PunctuationSettingsScope_put_Miscellaneous(This,Scope)	\
    ( (This)->lpVtbl -> put_Miscellaneous(This,Scope) ) 

#define PunctuationSettingsScope_get_Space(This,Scope)	\
    ( (This)->lpVtbl -> get_Space(This,Scope) ) 

#define PunctuationSettingsScope_put_Space(This,Scope)	\
    ( (This)->lpVtbl -> put_Space(This,Scope) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __PunctuationSettingsScope_INTERFACE_DEFINED__ */


#ifndef __KeyboardSettingsScope_INTERFACE_DEFINED__
#define __KeyboardSettingsScope_INTERFACE_DEFINED__

/* interface KeyboardSettingsScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_KeyboardSettingsScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("B3509CBF-064F-4796-84D1-904D315ABAFB")
    KeyboardSettingsScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Voice( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Voice( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Rate( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Rate( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Pitch( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Pitch( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Tone( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Tone( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Volume( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Volume( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Punctuation( 
            /* [retval][out] */ PunctuationSettingsScope **punc) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpecialKeys( 
            /* [retval][out] */ SpecialKeySettingsScope **SpecialKeys) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CapitalMode( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CapitalMode( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Interruptability( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Interruptability( 
            /* [in] */ SetFileScopes prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct KeyboardSettingsScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            KeyboardSettingsScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            KeyboardSettingsScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            KeyboardSettingsScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            KeyboardSettingsScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            KeyboardSettingsScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            KeyboardSettingsScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            KeyboardSettingsScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            KeyboardSettingsScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            KeyboardSettingsScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Voice )( 
            KeyboardSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Voice )( 
            KeyboardSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Rate )( 
            KeyboardSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Rate )( 
            KeyboardSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Pitch )( 
            KeyboardSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Pitch )( 
            KeyboardSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Tone )( 
            KeyboardSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Tone )( 
            KeyboardSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Volume )( 
            KeyboardSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Volume )( 
            KeyboardSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Punctuation )( 
            KeyboardSettingsScope * This,
            /* [retval][out] */ PunctuationSettingsScope **punc);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpecialKeys )( 
            KeyboardSettingsScope * This,
            /* [retval][out] */ SpecialKeySettingsScope **SpecialKeys);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CapitalMode )( 
            KeyboardSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CapitalMode )( 
            KeyboardSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Interruptability )( 
            KeyboardSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Interruptability )( 
            KeyboardSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        END_INTERFACE
    } KeyboardSettingsScopeVtbl;

    interface KeyboardSettingsScope
    {
        CONST_VTBL struct KeyboardSettingsScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define KeyboardSettingsScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define KeyboardSettingsScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define KeyboardSettingsScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define KeyboardSettingsScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define KeyboardSettingsScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define KeyboardSettingsScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define KeyboardSettingsScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define KeyboardSettingsScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define KeyboardSettingsScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define KeyboardSettingsScope_get_Voice(This,prop)	\
    ( (This)->lpVtbl -> get_Voice(This,prop) ) 

#define KeyboardSettingsScope_put_Voice(This,prop)	\
    ( (This)->lpVtbl -> put_Voice(This,prop) ) 

#define KeyboardSettingsScope_get_Rate(This,prop)	\
    ( (This)->lpVtbl -> get_Rate(This,prop) ) 

#define KeyboardSettingsScope_put_Rate(This,prop)	\
    ( (This)->lpVtbl -> put_Rate(This,prop) ) 

#define KeyboardSettingsScope_get_Pitch(This,prop)	\
    ( (This)->lpVtbl -> get_Pitch(This,prop) ) 

#define KeyboardSettingsScope_put_Pitch(This,prop)	\
    ( (This)->lpVtbl -> put_Pitch(This,prop) ) 

#define KeyboardSettingsScope_get_Tone(This,prop)	\
    ( (This)->lpVtbl -> get_Tone(This,prop) ) 

#define KeyboardSettingsScope_put_Tone(This,prop)	\
    ( (This)->lpVtbl -> put_Tone(This,prop) ) 

#define KeyboardSettingsScope_get_Volume(This,prop)	\
    ( (This)->lpVtbl -> get_Volume(This,prop) ) 

#define KeyboardSettingsScope_put_Volume(This,prop)	\
    ( (This)->lpVtbl -> put_Volume(This,prop) ) 

#define KeyboardSettingsScope_get_Punctuation(This,punc)	\
    ( (This)->lpVtbl -> get_Punctuation(This,punc) ) 

#define KeyboardSettingsScope_get_SpecialKeys(This,SpecialKeys)	\
    ( (This)->lpVtbl -> get_SpecialKeys(This,SpecialKeys) ) 

#define KeyboardSettingsScope_get_CapitalMode(This,prop)	\
    ( (This)->lpVtbl -> get_CapitalMode(This,prop) ) 

#define KeyboardSettingsScope_put_CapitalMode(This,prop)	\
    ( (This)->lpVtbl -> put_CapitalMode(This,prop) ) 

#define KeyboardSettingsScope_get_Interruptability(This,prop)	\
    ( (This)->lpVtbl -> get_Interruptability(This,prop) ) 

#define KeyboardSettingsScope_put_Interruptability(This,prop)	\
    ( (This)->lpVtbl -> put_Interruptability(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __KeyboardSettingsScope_INTERFACE_DEFINED__ */


#ifndef __SpecialKeySettingsScope_INTERFACE_DEFINED__
#define __SpecialKeySettingsScope_INTERFACE_DEFINED__

/* interface SpecialKeySettingsScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_SpecialKeySettingsScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("3E143270-1E9D-4F20-86A5-C305C1FC37C0")
    SpecialKeySettingsScope : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FunctionKeys( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FunctionKeys( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ToggleKeys( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ToggleKeys( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_KeyEnhancements( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_KeyEnhancements( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EscapeKey( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EscapeKey( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EnterKey( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EnterKey( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EnableBothInsertsAsModifier( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EnableBothInsertsAsModifier( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CursorMovement( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CursorMovement( 
            /* [in] */ SetFileScopes prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct SpecialKeySettingsScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            SpecialKeySettingsScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            SpecialKeySettingsScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            SpecialKeySettingsScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            SpecialKeySettingsScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            SpecialKeySettingsScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            SpecialKeySettingsScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            SpecialKeySettingsScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            SpecialKeySettingsScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            SpecialKeySettingsScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FunctionKeys )( 
            SpecialKeySettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FunctionKeys )( 
            SpecialKeySettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ToggleKeys )( 
            SpecialKeySettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ToggleKeys )( 
            SpecialKeySettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_KeyEnhancements )( 
            SpecialKeySettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_KeyEnhancements )( 
            SpecialKeySettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EscapeKey )( 
            SpecialKeySettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EscapeKey )( 
            SpecialKeySettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EnterKey )( 
            SpecialKeySettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EnterKey )( 
            SpecialKeySettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EnableBothInsertsAsModifier )( 
            SpecialKeySettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EnableBothInsertsAsModifier )( 
            SpecialKeySettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CursorMovement )( 
            SpecialKeySettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CursorMovement )( 
            SpecialKeySettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        END_INTERFACE
    } SpecialKeySettingsScopeVtbl;

    interface SpecialKeySettingsScope
    {
        CONST_VTBL struct SpecialKeySettingsScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define SpecialKeySettingsScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define SpecialKeySettingsScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define SpecialKeySettingsScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define SpecialKeySettingsScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define SpecialKeySettingsScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define SpecialKeySettingsScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define SpecialKeySettingsScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define SpecialKeySettingsScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define SpecialKeySettingsScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define SpecialKeySettingsScope_get_FunctionKeys(This,prop)	\
    ( (This)->lpVtbl -> get_FunctionKeys(This,prop) ) 

#define SpecialKeySettingsScope_put_FunctionKeys(This,prop)	\
    ( (This)->lpVtbl -> put_FunctionKeys(This,prop) ) 

#define SpecialKeySettingsScope_get_ToggleKeys(This,prop)	\
    ( (This)->lpVtbl -> get_ToggleKeys(This,prop) ) 

#define SpecialKeySettingsScope_put_ToggleKeys(This,prop)	\
    ( (This)->lpVtbl -> put_ToggleKeys(This,prop) ) 

#define SpecialKeySettingsScope_get_KeyEnhancements(This,prop)	\
    ( (This)->lpVtbl -> get_KeyEnhancements(This,prop) ) 

#define SpecialKeySettingsScope_put_KeyEnhancements(This,prop)	\
    ( (This)->lpVtbl -> put_KeyEnhancements(This,prop) ) 

#define SpecialKeySettingsScope_get_EscapeKey(This,prop)	\
    ( (This)->lpVtbl -> get_EscapeKey(This,prop) ) 

#define SpecialKeySettingsScope_put_EscapeKey(This,prop)	\
    ( (This)->lpVtbl -> put_EscapeKey(This,prop) ) 

#define SpecialKeySettingsScope_get_EnterKey(This,prop)	\
    ( (This)->lpVtbl -> get_EnterKey(This,prop) ) 

#define SpecialKeySettingsScope_put_EnterKey(This,prop)	\
    ( (This)->lpVtbl -> put_EnterKey(This,prop) ) 

#define SpecialKeySettingsScope_get_EnableBothInsertsAsModifier(This,prop)	\
    ( (This)->lpVtbl -> get_EnableBothInsertsAsModifier(This,prop) ) 

#define SpecialKeySettingsScope_put_EnableBothInsertsAsModifier(This,prop)	\
    ( (This)->lpVtbl -> put_EnableBothInsertsAsModifier(This,prop) ) 

#define SpecialKeySettingsScope_get_CursorMovement(This,prop)	\
    ( (This)->lpVtbl -> get_CursorMovement(This,prop) ) 

#define SpecialKeySettingsScope_put_CursorMovement(This,prop)	\
    ( (This)->lpVtbl -> put_CursorMovement(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __SpecialKeySettingsScope_INTERFACE_DEFINED__ */


#ifndef __MouseSettingsScope_INTERFACE_DEFINED__
#define __MouseSettingsScope_INTERFACE_DEFINED__

/* interface MouseSettingsScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_MouseSettingsScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("340B121B-9D18-4149-9088-F125BE7BC19D")
    MouseSettingsScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Voice( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Voice( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Rate( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Rate( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Pitch( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Pitch( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Tone( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Tone( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Volume( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Volume( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Boundary( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Boundary( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Buttons( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Buttons( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Movement( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Movement( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Interruptability( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Interruptability( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TrackFocus( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TrackFocus( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UpDownSpeak( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_UpDownSpeak( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SearchAttributes( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SearchAttributes( 
            /* [in] */ SetFileScopes prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct MouseSettingsScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            MouseSettingsScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            MouseSettingsScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            MouseSettingsScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            MouseSettingsScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            MouseSettingsScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            MouseSettingsScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            MouseSettingsScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            MouseSettingsScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            MouseSettingsScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Voice )( 
            MouseSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Voice )( 
            MouseSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Rate )( 
            MouseSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Rate )( 
            MouseSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Pitch )( 
            MouseSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Pitch )( 
            MouseSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Tone )( 
            MouseSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Tone )( 
            MouseSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Volume )( 
            MouseSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Volume )( 
            MouseSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Boundary )( 
            MouseSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Boundary )( 
            MouseSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Buttons )( 
            MouseSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Buttons )( 
            MouseSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Movement )( 
            MouseSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Movement )( 
            MouseSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Interruptability )( 
            MouseSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Interruptability )( 
            MouseSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TrackFocus )( 
            MouseSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TrackFocus )( 
            MouseSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UpDownSpeak )( 
            MouseSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_UpDownSpeak )( 
            MouseSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SearchAttributes )( 
            MouseSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SearchAttributes )( 
            MouseSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        END_INTERFACE
    } MouseSettingsScopeVtbl;

    interface MouseSettingsScope
    {
        CONST_VTBL struct MouseSettingsScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define MouseSettingsScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define MouseSettingsScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define MouseSettingsScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define MouseSettingsScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define MouseSettingsScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define MouseSettingsScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define MouseSettingsScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define MouseSettingsScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define MouseSettingsScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define MouseSettingsScope_get_Voice(This,prop)	\
    ( (This)->lpVtbl -> get_Voice(This,prop) ) 

#define MouseSettingsScope_put_Voice(This,prop)	\
    ( (This)->lpVtbl -> put_Voice(This,prop) ) 

#define MouseSettingsScope_get_Rate(This,prop)	\
    ( (This)->lpVtbl -> get_Rate(This,prop) ) 

#define MouseSettingsScope_put_Rate(This,prop)	\
    ( (This)->lpVtbl -> put_Rate(This,prop) ) 

#define MouseSettingsScope_get_Pitch(This,prop)	\
    ( (This)->lpVtbl -> get_Pitch(This,prop) ) 

#define MouseSettingsScope_put_Pitch(This,prop)	\
    ( (This)->lpVtbl -> put_Pitch(This,prop) ) 

#define MouseSettingsScope_get_Tone(This,prop)	\
    ( (This)->lpVtbl -> get_Tone(This,prop) ) 

#define MouseSettingsScope_put_Tone(This,prop)	\
    ( (This)->lpVtbl -> put_Tone(This,prop) ) 

#define MouseSettingsScope_get_Volume(This,prop)	\
    ( (This)->lpVtbl -> get_Volume(This,prop) ) 

#define MouseSettingsScope_put_Volume(This,prop)	\
    ( (This)->lpVtbl -> put_Volume(This,prop) ) 

#define MouseSettingsScope_get_Boundary(This,prop)	\
    ( (This)->lpVtbl -> get_Boundary(This,prop) ) 

#define MouseSettingsScope_put_Boundary(This,prop)	\
    ( (This)->lpVtbl -> put_Boundary(This,prop) ) 

#define MouseSettingsScope_get_Buttons(This,prop)	\
    ( (This)->lpVtbl -> get_Buttons(This,prop) ) 

#define MouseSettingsScope_put_Buttons(This,prop)	\
    ( (This)->lpVtbl -> put_Buttons(This,prop) ) 

#define MouseSettingsScope_get_Movement(This,prop)	\
    ( (This)->lpVtbl -> get_Movement(This,prop) ) 

#define MouseSettingsScope_put_Movement(This,prop)	\
    ( (This)->lpVtbl -> put_Movement(This,prop) ) 

#define MouseSettingsScope_get_Interruptability(This,prop)	\
    ( (This)->lpVtbl -> get_Interruptability(This,prop) ) 

#define MouseSettingsScope_put_Interruptability(This,prop)	\
    ( (This)->lpVtbl -> put_Interruptability(This,prop) ) 

#define MouseSettingsScope_get_TrackFocus(This,prop)	\
    ( (This)->lpVtbl -> get_TrackFocus(This,prop) ) 

#define MouseSettingsScope_put_TrackFocus(This,prop)	\
    ( (This)->lpVtbl -> put_TrackFocus(This,prop) ) 

#define MouseSettingsScope_get_UpDownSpeak(This,prop)	\
    ( (This)->lpVtbl -> get_UpDownSpeak(This,prop) ) 

#define MouseSettingsScope_put_UpDownSpeak(This,prop)	\
    ( (This)->lpVtbl -> put_UpDownSpeak(This,prop) ) 

#define MouseSettingsScope_get_SearchAttributes(This,prop)	\
    ( (This)->lpVtbl -> get_SearchAttributes(This,prop) ) 

#define MouseSettingsScope_put_SearchAttributes(This,prop)	\
    ( (This)->lpVtbl -> put_SearchAttributes(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __MouseSettingsScope_INTERFACE_DEFINED__ */


#ifndef __GeneralSettingsScope_INTERFACE_DEFINED__
#define __GeneralSettingsScope_INTERFACE_DEFINED__

/* interface GeneralSettingsScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_GeneralSettingsScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("652A6567-793B-4502-AD03-6A989D6E7C45")
    GeneralSettingsScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Voice( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Braille( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HotKeys( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CursorKeys( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HighlightTrack( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CharacterDictionary( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CharacterDictionary( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_KeyLabelDictionary( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_KeyLabelDictionary( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ExceptionDictionary( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ExceptionDictionary( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_GraphicDictionary( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_GraphicDictionary( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ColorDictionary( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ColorDictionary( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AllowSpeakWindowsInEditBoxes( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AllowSpeakWindowsInEditBoxes( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CursorDelay( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CursorDelay( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TriggerDelay( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TriggerDelay( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpaceThreshold( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpaceThreshold( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AllowAutoRedraw( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpeakTooltips( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpeakTooltips( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BrowseMode( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_BrowseMode( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeBoxInHighlight( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeBoxInHighlight( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HighlightedText( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_HighlightedText( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeUserWindowWithBox( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeUserWindowWithBox( 
            /* [in] */ SetFileScopes prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct GeneralSettingsScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            GeneralSettingsScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            GeneralSettingsScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            GeneralSettingsScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            GeneralSettingsScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            GeneralSettingsScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            GeneralSettingsScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            GeneralSettingsScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Voice )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Braille )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HotKeys )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CursorKeys )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HighlightTrack )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CharacterDictionary )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CharacterDictionary )( 
            GeneralSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_KeyLabelDictionary )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_KeyLabelDictionary )( 
            GeneralSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ExceptionDictionary )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ExceptionDictionary )( 
            GeneralSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_GraphicDictionary )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_GraphicDictionary )( 
            GeneralSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ColorDictionary )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ColorDictionary )( 
            GeneralSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AllowSpeakWindowsInEditBoxes )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AllowSpeakWindowsInEditBoxes )( 
            GeneralSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CursorDelay )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CursorDelay )( 
            GeneralSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TriggerDelay )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TriggerDelay )( 
            GeneralSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpaceThreshold )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpaceThreshold )( 
            GeneralSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AllowAutoRedraw )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpeakTooltips )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpeakTooltips )( 
            GeneralSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BrowseMode )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_BrowseMode )( 
            GeneralSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeBoxInHighlight )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeBoxInHighlight )( 
            GeneralSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HighlightedText )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_HighlightedText )( 
            GeneralSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeUserWindowWithBox )( 
            GeneralSettingsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeUserWindowWithBox )( 
            GeneralSettingsScope * This,
            /* [in] */ SetFileScopes prop);
        
        END_INTERFACE
    } GeneralSettingsScopeVtbl;

    interface GeneralSettingsScope
    {
        CONST_VTBL struct GeneralSettingsScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define GeneralSettingsScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define GeneralSettingsScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define GeneralSettingsScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define GeneralSettingsScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define GeneralSettingsScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define GeneralSettingsScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define GeneralSettingsScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define GeneralSettingsScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define GeneralSettingsScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define GeneralSettingsScope_get_Voice(This,prop)	\
    ( (This)->lpVtbl -> get_Voice(This,prop) ) 

#define GeneralSettingsScope_get_Braille(This,prop)	\
    ( (This)->lpVtbl -> get_Braille(This,prop) ) 

#define GeneralSettingsScope_get_HotKeys(This,prop)	\
    ( (This)->lpVtbl -> get_HotKeys(This,prop) ) 

#define GeneralSettingsScope_get_CursorKeys(This,prop)	\
    ( (This)->lpVtbl -> get_CursorKeys(This,prop) ) 

#define GeneralSettingsScope_get_HighlightTrack(This,prop)	\
    ( (This)->lpVtbl -> get_HighlightTrack(This,prop) ) 

#define GeneralSettingsScope_get_CharacterDictionary(This,prop)	\
    ( (This)->lpVtbl -> get_CharacterDictionary(This,prop) ) 

#define GeneralSettingsScope_put_CharacterDictionary(This,prop)	\
    ( (This)->lpVtbl -> put_CharacterDictionary(This,prop) ) 

#define GeneralSettingsScope_get_KeyLabelDictionary(This,prop)	\
    ( (This)->lpVtbl -> get_KeyLabelDictionary(This,prop) ) 

#define GeneralSettingsScope_put_KeyLabelDictionary(This,prop)	\
    ( (This)->lpVtbl -> put_KeyLabelDictionary(This,prop) ) 

#define GeneralSettingsScope_get_ExceptionDictionary(This,prop)	\
    ( (This)->lpVtbl -> get_ExceptionDictionary(This,prop) ) 

#define GeneralSettingsScope_put_ExceptionDictionary(This,prop)	\
    ( (This)->lpVtbl -> put_ExceptionDictionary(This,prop) ) 

#define GeneralSettingsScope_get_GraphicDictionary(This,prop)	\
    ( (This)->lpVtbl -> get_GraphicDictionary(This,prop) ) 

#define GeneralSettingsScope_put_GraphicDictionary(This,prop)	\
    ( (This)->lpVtbl -> put_GraphicDictionary(This,prop) ) 

#define GeneralSettingsScope_get_ColorDictionary(This,prop)	\
    ( (This)->lpVtbl -> get_ColorDictionary(This,prop) ) 

#define GeneralSettingsScope_put_ColorDictionary(This,prop)	\
    ( (This)->lpVtbl -> put_ColorDictionary(This,prop) ) 

#define GeneralSettingsScope_get_AllowSpeakWindowsInEditBoxes(This,prop)	\
    ( (This)->lpVtbl -> get_AllowSpeakWindowsInEditBoxes(This,prop) ) 

#define GeneralSettingsScope_put_AllowSpeakWindowsInEditBoxes(This,prop)	\
    ( (This)->lpVtbl -> put_AllowSpeakWindowsInEditBoxes(This,prop) ) 

#define GeneralSettingsScope_get_CursorDelay(This,prop)	\
    ( (This)->lpVtbl -> get_CursorDelay(This,prop) ) 

#define GeneralSettingsScope_put_CursorDelay(This,prop)	\
    ( (This)->lpVtbl -> put_CursorDelay(This,prop) ) 

#define GeneralSettingsScope_get_TriggerDelay(This,prop)	\
    ( (This)->lpVtbl -> get_TriggerDelay(This,prop) ) 

#define GeneralSettingsScope_put_TriggerDelay(This,prop)	\
    ( (This)->lpVtbl -> put_TriggerDelay(This,prop) ) 

#define GeneralSettingsScope_get_SpaceThreshold(This,prop)	\
    ( (This)->lpVtbl -> get_SpaceThreshold(This,prop) ) 

#define GeneralSettingsScope_put_SpaceThreshold(This,prop)	\
    ( (This)->lpVtbl -> put_SpaceThreshold(This,prop) ) 

#define GeneralSettingsScope_get_AllowAutoRedraw(This,prop)	\
    ( (This)->lpVtbl -> get_AllowAutoRedraw(This,prop) ) 

#define GeneralSettingsScope_get_SpeakTooltips(This,prop)	\
    ( (This)->lpVtbl -> get_SpeakTooltips(This,prop) ) 

#define GeneralSettingsScope_put_SpeakTooltips(This,prop)	\
    ( (This)->lpVtbl -> put_SpeakTooltips(This,prop) ) 

#define GeneralSettingsScope_get_BrowseMode(This,prop)	\
    ( (This)->lpVtbl -> get_BrowseMode(This,prop) ) 

#define GeneralSettingsScope_put_BrowseMode(This,prop)	\
    ( (This)->lpVtbl -> put_BrowseMode(This,prop) ) 

#define GeneralSettingsScope_get_IncludeBoxInHighlight(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeBoxInHighlight(This,prop) ) 

#define GeneralSettingsScope_put_IncludeBoxInHighlight(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeBoxInHighlight(This,prop) ) 

#define GeneralSettingsScope_get_HighlightedText(This,prop)	\
    ( (This)->lpVtbl -> get_HighlightedText(This,prop) ) 

#define GeneralSettingsScope_put_HighlightedText(This,prop)	\
    ( (This)->lpVtbl -> put_HighlightedText(This,prop) ) 

#define GeneralSettingsScope_get_IncludeUserWindowWithBox(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeUserWindowWithBox(This,prop) ) 

#define GeneralSettingsScope_put_IncludeUserWindowWithBox(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeUserWindowWithBox(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __GeneralSettingsScope_INTERFACE_DEFINED__ */


#ifndef __BrailleSettingsScope_INTERFACE_DEFINED__
#define __BrailleSettingsScope_INTERFACE_DEFINED__

/* interface BrailleSettingsScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrailleSettingsScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("BC7FCC1D-28AA-437D-8B73-B896B6F8B74C")
    BrailleSettingsScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Scrolling( 
            /* [retval][out] */ BrailleScrollingOptionsScope **Scrolling) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ControlInformation( 
            /* [retval][out] */ BrailleControlInformationScope **controlInfo) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Options( 
            /* [retval][out] */ BrailleOptionsScope **Options) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Graphics( 
            /* [retval][out] */ BrailleGraphicsScope **Graphics) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleSettingsScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleSettingsScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleSettingsScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleSettingsScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleSettingsScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleSettingsScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleSettingsScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleSettingsScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleSettingsScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleSettingsScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Scrolling )( 
            BrailleSettingsScope * This,
            /* [retval][out] */ BrailleScrollingOptionsScope **Scrolling);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ControlInformation )( 
            BrailleSettingsScope * This,
            /* [retval][out] */ BrailleControlInformationScope **controlInfo);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Options )( 
            BrailleSettingsScope * This,
            /* [retval][out] */ BrailleOptionsScope **Options);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Graphics )( 
            BrailleSettingsScope * This,
            /* [retval][out] */ BrailleGraphicsScope **Graphics);
        
        END_INTERFACE
    } BrailleSettingsScopeVtbl;

    interface BrailleSettingsScope
    {
        CONST_VTBL struct BrailleSettingsScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleSettingsScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleSettingsScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleSettingsScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleSettingsScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleSettingsScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleSettingsScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleSettingsScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleSettingsScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleSettingsScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleSettingsScope_get_Scrolling(This,Scrolling)	\
    ( (This)->lpVtbl -> get_Scrolling(This,Scrolling) ) 

#define BrailleSettingsScope_get_ControlInformation(This,controlInfo)	\
    ( (This)->lpVtbl -> get_ControlInformation(This,controlInfo) ) 

#define BrailleSettingsScope_get_Options(This,Options)	\
    ( (This)->lpVtbl -> get_Options(This,Options) ) 

#define BrailleSettingsScope_get_Graphics(This,Graphics)	\
    ( (This)->lpVtbl -> get_Graphics(This,Graphics) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleSettingsScope_INTERFACE_DEFINED__ */


#ifndef __BrailleScrollingOptionsScope_INTERFACE_DEFINED__
#define __BrailleScrollingOptionsScope_INTERFACE_DEFINED__

/* interface BrailleScrollingOptionsScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrailleScrollingOptionsScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("DCC11043-4D97-4B4F-B51D-4C02736E60E8")
    BrailleScrollingOptionsScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AutoRouteCursor( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AutoRouteCursor( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WholeWordMode( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_WholeWordMode( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ScrollingWidth( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ScrollingWidth( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateLineChange( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateLineChange( 
            /* [in] */ SetFileScopes prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleScrollingOptionsScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleScrollingOptionsScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleScrollingOptionsScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleScrollingOptionsScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleScrollingOptionsScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleScrollingOptionsScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleScrollingOptionsScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleScrollingOptionsScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleScrollingOptionsScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleScrollingOptionsScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AutoRouteCursor )( 
            BrailleScrollingOptionsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AutoRouteCursor )( 
            BrailleScrollingOptionsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WholeWordMode )( 
            BrailleScrollingOptionsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_WholeWordMode )( 
            BrailleScrollingOptionsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ScrollingWidth )( 
            BrailleScrollingOptionsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ScrollingWidth )( 
            BrailleScrollingOptionsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateLineChange )( 
            BrailleScrollingOptionsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateLineChange )( 
            BrailleScrollingOptionsScope * This,
            /* [in] */ SetFileScopes prop);
        
        END_INTERFACE
    } BrailleScrollingOptionsScopeVtbl;

    interface BrailleScrollingOptionsScope
    {
        CONST_VTBL struct BrailleScrollingOptionsScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleScrollingOptionsScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleScrollingOptionsScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleScrollingOptionsScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleScrollingOptionsScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleScrollingOptionsScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleScrollingOptionsScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleScrollingOptionsScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleScrollingOptionsScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleScrollingOptionsScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleScrollingOptionsScope_get_AutoRouteCursor(This,prop)	\
    ( (This)->lpVtbl -> get_AutoRouteCursor(This,prop) ) 

#define BrailleScrollingOptionsScope_put_AutoRouteCursor(This,prop)	\
    ( (This)->lpVtbl -> put_AutoRouteCursor(This,prop) ) 

#define BrailleScrollingOptionsScope_get_WholeWordMode(This,prop)	\
    ( (This)->lpVtbl -> get_WholeWordMode(This,prop) ) 

#define BrailleScrollingOptionsScope_put_WholeWordMode(This,prop)	\
    ( (This)->lpVtbl -> put_WholeWordMode(This,prop) ) 

#define BrailleScrollingOptionsScope_get_ScrollingWidth(This,prop)	\
    ( (This)->lpVtbl -> get_ScrollingWidth(This,prop) ) 

#define BrailleScrollingOptionsScope_put_ScrollingWidth(This,prop)	\
    ( (This)->lpVtbl -> put_ScrollingWidth(This,prop) ) 

#define BrailleScrollingOptionsScope_get_IndicateLineChange(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateLineChange(This,prop) ) 

#define BrailleScrollingOptionsScope_put_IndicateLineChange(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateLineChange(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleScrollingOptionsScope_INTERFACE_DEFINED__ */


#ifndef __BrailleControlInformationScope_INTERFACE_DEFINED__
#define __BrailleControlInformationScope_INTERFACE_DEFINED__

/* interface BrailleControlInformationScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrailleControlInformationScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("DB1A0515-EF94-410F-8AA5-AEAAC18D9069")
    BrailleControlInformationScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_InformationOrder( 
            /* [retval][out] */ SetFileScopes *order) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_InformationOrder( 
            /* [in] */ SetFileScopes order) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeName( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeName( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeType( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeType( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeIndex( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeIndex( 
            /* [in] */ SetFileScopes prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleControlInformationScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleControlInformationScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleControlInformationScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleControlInformationScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleControlInformationScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleControlInformationScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleControlInformationScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleControlInformationScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleControlInformationScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleControlInformationScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_InformationOrder )( 
            BrailleControlInformationScope * This,
            /* [retval][out] */ SetFileScopes *order);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_InformationOrder )( 
            BrailleControlInformationScope * This,
            /* [in] */ SetFileScopes order);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeName )( 
            BrailleControlInformationScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeName )( 
            BrailleControlInformationScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeType )( 
            BrailleControlInformationScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeType )( 
            BrailleControlInformationScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeIndex )( 
            BrailleControlInformationScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeIndex )( 
            BrailleControlInformationScope * This,
            /* [in] */ SetFileScopes prop);
        
        END_INTERFACE
    } BrailleControlInformationScopeVtbl;

    interface BrailleControlInformationScope
    {
        CONST_VTBL struct BrailleControlInformationScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleControlInformationScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleControlInformationScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleControlInformationScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleControlInformationScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleControlInformationScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleControlInformationScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleControlInformationScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleControlInformationScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleControlInformationScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleControlInformationScope_get_InformationOrder(This,order)	\
    ( (This)->lpVtbl -> get_InformationOrder(This,order) ) 

#define BrailleControlInformationScope_put_InformationOrder(This,order)	\
    ( (This)->lpVtbl -> put_InformationOrder(This,order) ) 

#define BrailleControlInformationScope_get_IncludeName(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeName(This,prop) ) 

#define BrailleControlInformationScope_put_IncludeName(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeName(This,prop) ) 

#define BrailleControlInformationScope_get_IncludeType(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeType(This,prop) ) 

#define BrailleControlInformationScope_put_IncludeType(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeType(This,prop) ) 

#define BrailleControlInformationScope_get_IncludeIndex(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeIndex(This,prop) ) 

#define BrailleControlInformationScope_put_IncludeIndex(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeIndex(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleControlInformationScope_INTERFACE_DEFINED__ */


#ifndef __BrailleOptionsScope_INTERFACE_DEFINED__
#define __BrailleOptionsScope_INTERFACE_DEFINED__

/* interface BrailleOptionsScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrailleOptionsScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("AE632BB5-D26F-4DB3-ABCA-575A6B5F3C37")
    BrailleOptionsScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Grade2Enabled( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Grade2Enabled( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Grade2AutoUntranslate( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Grade2AutoUntranslate( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Grade2TranslateSymbols( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Grade2TranslateSymbols( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EightDot( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EightDot( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AutoUpdateTimeout( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AutoUpdateTimeout( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AttributeMarking( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AttributeMarking( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_QuickMessageMode( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_QuickMessageMode( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_QuickMessageModeTimeout( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_QuickMessageModeTimeout( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ShowIndents( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ShowIndents( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ExtraSpacing( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ExtraSpacing( 
            /* [in] */ SetFileScopes prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleOptionsScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleOptionsScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleOptionsScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleOptionsScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleOptionsScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleOptionsScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleOptionsScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleOptionsScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleOptionsScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleOptionsScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Grade2Enabled )( 
            BrailleOptionsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Grade2Enabled )( 
            BrailleOptionsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Grade2AutoUntranslate )( 
            BrailleOptionsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Grade2AutoUntranslate )( 
            BrailleOptionsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Grade2TranslateSymbols )( 
            BrailleOptionsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Grade2TranslateSymbols )( 
            BrailleOptionsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EightDot )( 
            BrailleOptionsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EightDot )( 
            BrailleOptionsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AutoUpdateTimeout )( 
            BrailleOptionsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AutoUpdateTimeout )( 
            BrailleOptionsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AttributeMarking )( 
            BrailleOptionsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AttributeMarking )( 
            BrailleOptionsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_QuickMessageMode )( 
            BrailleOptionsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_QuickMessageMode )( 
            BrailleOptionsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_QuickMessageModeTimeout )( 
            BrailleOptionsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_QuickMessageModeTimeout )( 
            BrailleOptionsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ShowIndents )( 
            BrailleOptionsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ShowIndents )( 
            BrailleOptionsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ExtraSpacing )( 
            BrailleOptionsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ExtraSpacing )( 
            BrailleOptionsScope * This,
            /* [in] */ SetFileScopes prop);
        
        END_INTERFACE
    } BrailleOptionsScopeVtbl;

    interface BrailleOptionsScope
    {
        CONST_VTBL struct BrailleOptionsScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleOptionsScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleOptionsScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleOptionsScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleOptionsScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleOptionsScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleOptionsScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleOptionsScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleOptionsScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleOptionsScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleOptionsScope_get_Grade2Enabled(This,prop)	\
    ( (This)->lpVtbl -> get_Grade2Enabled(This,prop) ) 

#define BrailleOptionsScope_put_Grade2Enabled(This,prop)	\
    ( (This)->lpVtbl -> put_Grade2Enabled(This,prop) ) 

#define BrailleOptionsScope_get_Grade2AutoUntranslate(This,prop)	\
    ( (This)->lpVtbl -> get_Grade2AutoUntranslate(This,prop) ) 

#define BrailleOptionsScope_put_Grade2AutoUntranslate(This,prop)	\
    ( (This)->lpVtbl -> put_Grade2AutoUntranslate(This,prop) ) 

#define BrailleOptionsScope_get_Grade2TranslateSymbols(This,prop)	\
    ( (This)->lpVtbl -> get_Grade2TranslateSymbols(This,prop) ) 

#define BrailleOptionsScope_put_Grade2TranslateSymbols(This,prop)	\
    ( (This)->lpVtbl -> put_Grade2TranslateSymbols(This,prop) ) 

#define BrailleOptionsScope_get_EightDot(This,prop)	\
    ( (This)->lpVtbl -> get_EightDot(This,prop) ) 

#define BrailleOptionsScope_put_EightDot(This,prop)	\
    ( (This)->lpVtbl -> put_EightDot(This,prop) ) 

#define BrailleOptionsScope_get_AutoUpdateTimeout(This,prop)	\
    ( (This)->lpVtbl -> get_AutoUpdateTimeout(This,prop) ) 

#define BrailleOptionsScope_put_AutoUpdateTimeout(This,prop)	\
    ( (This)->lpVtbl -> put_AutoUpdateTimeout(This,prop) ) 

#define BrailleOptionsScope_get_AttributeMarking(This,prop)	\
    ( (This)->lpVtbl -> get_AttributeMarking(This,prop) ) 

#define BrailleOptionsScope_put_AttributeMarking(This,prop)	\
    ( (This)->lpVtbl -> put_AttributeMarking(This,prop) ) 

#define BrailleOptionsScope_get_QuickMessageMode(This,prop)	\
    ( (This)->lpVtbl -> get_QuickMessageMode(This,prop) ) 

#define BrailleOptionsScope_put_QuickMessageMode(This,prop)	\
    ( (This)->lpVtbl -> put_QuickMessageMode(This,prop) ) 

#define BrailleOptionsScope_get_QuickMessageModeTimeout(This,prop)	\
    ( (This)->lpVtbl -> get_QuickMessageModeTimeout(This,prop) ) 

#define BrailleOptionsScope_put_QuickMessageModeTimeout(This,prop)	\
    ( (This)->lpVtbl -> put_QuickMessageModeTimeout(This,prop) ) 

#define BrailleOptionsScope_get_ShowIndents(This,prop)	\
    ( (This)->lpVtbl -> get_ShowIndents(This,prop) ) 

#define BrailleOptionsScope_put_ShowIndents(This,prop)	\
    ( (This)->lpVtbl -> put_ShowIndents(This,prop) ) 

#define BrailleOptionsScope_get_ExtraSpacing(This,prop)	\
    ( (This)->lpVtbl -> get_ExtraSpacing(This,prop) ) 

#define BrailleOptionsScope_put_ExtraSpacing(This,prop)	\
    ( (This)->lpVtbl -> put_ExtraSpacing(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleOptionsScope_INTERFACE_DEFINED__ */


#ifndef __BrailleGraphicsScope_INTERFACE_DEFINED__
#define __BrailleGraphicsScope_INTERFACE_DEFINED__

/* interface BrailleGraphicsScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrailleGraphicsScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("1A8CAEAD-0FA1-42BD-A1EF-21D188EAA7D0")
    BrailleGraphicsScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Verbosity( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Verbosity( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Symbol( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Symbol( 
            /* [in] */ SetFileScopes prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleGraphicsScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleGraphicsScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleGraphicsScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleGraphicsScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleGraphicsScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleGraphicsScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleGraphicsScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleGraphicsScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleGraphicsScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleGraphicsScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Verbosity )( 
            BrailleGraphicsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Verbosity )( 
            BrailleGraphicsScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Symbol )( 
            BrailleGraphicsScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Symbol )( 
            BrailleGraphicsScope * This,
            /* [in] */ SetFileScopes prop);
        
        END_INTERFACE
    } BrailleGraphicsScopeVtbl;

    interface BrailleGraphicsScope
    {
        CONST_VTBL struct BrailleGraphicsScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleGraphicsScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleGraphicsScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleGraphicsScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleGraphicsScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleGraphicsScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleGraphicsScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleGraphicsScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleGraphicsScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleGraphicsScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleGraphicsScope_get_Verbosity(This,prop)	\
    ( (This)->lpVtbl -> get_Verbosity(This,prop) ) 

#define BrailleGraphicsScope_put_Verbosity(This,prop)	\
    ( (This)->lpVtbl -> put_Verbosity(This,prop) ) 

#define BrailleGraphicsScope_get_Symbol(This,prop)	\
    ( (This)->lpVtbl -> get_Symbol(This,prop) ) 

#define BrailleGraphicsScope_put_Symbol(This,prop)	\
    ( (This)->lpVtbl -> put_Symbol(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleGraphicsScope_INTERFACE_DEFINED__ */


#ifndef __HotkeysScope_INTERFACE_DEFINED__
#define __HotkeysScope_INTERFACE_DEFINED__

/* interface HotkeysScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_HotkeysScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("944F4E76-6622-4CE3-A310-B98881B94C9F")
    HotkeysScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ HotkeyScope **Hotkey) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct HotkeysScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            HotkeysScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            HotkeysScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            HotkeysScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            HotkeysScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            HotkeysScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            HotkeysScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            HotkeysScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            HotkeysScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            HotkeysScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            HotkeysScope * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            HotkeysScope * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            HotkeysScope * This,
            /* [in] */ long index,
            /* [retval][out] */ HotkeyScope **Hotkey);
        
        END_INTERFACE
    } HotkeysScopeVtbl;

    interface HotkeysScope
    {
        CONST_VTBL struct HotkeysScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define HotkeysScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define HotkeysScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define HotkeysScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define HotkeysScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define HotkeysScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define HotkeysScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define HotkeysScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define HotkeysScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define HotkeysScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define HotkeysScope_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define HotkeysScope_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define HotkeysScope_Item(This,index,Hotkey)	\
    ( (This)->lpVtbl -> Item(This,index,Hotkey) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __HotkeysScope_INTERFACE_DEFINED__ */


#ifndef __HotkeyScope_INTERFACE_DEFINED__
#define __HotkeyScope_INTERFACE_DEFINED__

/* interface HotkeyScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_HotkeyScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("6B44A13F-C36F-4586-86DB-8DE768AC7125")
    HotkeyScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Scope( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Scope( 
            /* [in] */ SetFileScopes prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct HotkeyScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            HotkeyScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            HotkeyScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            HotkeyScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            HotkeyScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            HotkeyScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            HotkeyScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            HotkeyScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            HotkeyScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            HotkeyScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Scope )( 
            HotkeyScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Scope )( 
            HotkeyScope * This,
            /* [in] */ SetFileScopes prop);
        
        END_INTERFACE
    } HotkeyScopeVtbl;

    interface HotkeyScope
    {
        CONST_VTBL struct HotkeyScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define HotkeyScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define HotkeyScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define HotkeyScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define HotkeyScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define HotkeyScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define HotkeyScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define HotkeyScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define HotkeyScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define HotkeyScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define HotkeyScope_get_Scope(This,prop)	\
    ( (This)->lpVtbl -> get_Scope(This,prop) ) 

#define HotkeyScope_put_Scope(This,prop)	\
    ( (This)->lpVtbl -> put_Scope(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __HotkeyScope_INTERFACE_DEFINED__ */


#ifndef __VerbosityScope_INTERFACE_DEFINED__
#define __VerbosityScope_INTERFACE_DEFINED__

/* interface VerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_VerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("DB786D7D-E09A-4255-A202-DA801FA0F0CB")
    VerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Menu( 
            /* [retval][out] */ MenuVerbosityScope **Menu) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Activated( 
            /* [retval][out] */ ActivatedVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Focused( 
            /* [retval][out] */ FocusedVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Attribute( 
            /* [retval][out] */ AttributeChangeVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Miscellaneous( 
            /* [retval][out] */ MiscVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BrowseMode( 
            /* [retval][out] */ BrowseModeVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Mouse( 
            /* [retval][out] */ MouseVerbosityScope **Mouse) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct VerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            VerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            VerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            VerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            VerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            VerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            VerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            VerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            VerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            VerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Menu )( 
            VerbosityScope * This,
            /* [retval][out] */ MenuVerbosityScope **Menu);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Activated )( 
            VerbosityScope * This,
            /* [retval][out] */ ActivatedVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            VerbosityScope * This,
            /* [retval][out] */ FocusedVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Attribute )( 
            VerbosityScope * This,
            /* [retval][out] */ AttributeChangeVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Miscellaneous )( 
            VerbosityScope * This,
            /* [retval][out] */ MiscVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BrowseMode )( 
            VerbosityScope * This,
            /* [retval][out] */ BrowseModeVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Mouse )( 
            VerbosityScope * This,
            /* [retval][out] */ MouseVerbosityScope **Mouse);
        
        END_INTERFACE
    } VerbosityScopeVtbl;

    interface VerbosityScope
    {
        CONST_VTBL struct VerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define VerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define VerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define VerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define VerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define VerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define VerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define VerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define VerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define VerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define VerbosityScope_get_Menu(This,Menu)	\
    ( (This)->lpVtbl -> get_Menu(This,Menu) ) 

#define VerbosityScope_get_Activated(This,obj)	\
    ( (This)->lpVtbl -> get_Activated(This,obj) ) 

#define VerbosityScope_get_Focused(This,obj)	\
    ( (This)->lpVtbl -> get_Focused(This,obj) ) 

#define VerbosityScope_get_Attribute(This,obj)	\
    ( (This)->lpVtbl -> get_Attribute(This,obj) ) 

#define VerbosityScope_get_Miscellaneous(This,obj)	\
    ( (This)->lpVtbl -> get_Miscellaneous(This,obj) ) 

#define VerbosityScope_get_BrowseMode(This,obj)	\
    ( (This)->lpVtbl -> get_BrowseMode(This,obj) ) 

#define VerbosityScope_get_Mouse(This,Mouse)	\
    ( (This)->lpVtbl -> get_Mouse(This,Mouse) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __VerbosityScope_INTERFACE_DEFINED__ */


#ifndef __MenuVerbosityScope_INTERFACE_DEFINED__
#define __MenuVerbosityScope_INTERFACE_DEFINED__

/* interface MenuVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_MenuVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("7B87A15F-1B6A-4083-AEE8-16A07E2707D1")
    MenuVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Opened( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Opened( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Closed( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Closed( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Option( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Option( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Accelerators( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Accelerators( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Descriptions( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Descriptions( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Shortcut( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Shortcut( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct MenuVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            MenuVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            MenuVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            MenuVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            MenuVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            MenuVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            MenuVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            MenuVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            MenuVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            MenuVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Opened )( 
            MenuVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Opened )( 
            MenuVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Closed )( 
            MenuVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Closed )( 
            MenuVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Option )( 
            MenuVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Option )( 
            MenuVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Accelerators )( 
            MenuVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Accelerators )( 
            MenuVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Descriptions )( 
            MenuVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Descriptions )( 
            MenuVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Shortcut )( 
            MenuVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Shortcut )( 
            MenuVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } MenuVerbosityScopeVtbl;

    interface MenuVerbosityScope
    {
        CONST_VTBL struct MenuVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define MenuVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define MenuVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define MenuVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define MenuVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define MenuVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define MenuVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define MenuVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define MenuVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define MenuVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define MenuVerbosityScope_put_Opened(This,prop)	\
    ( (This)->lpVtbl -> put_Opened(This,prop) ) 

#define MenuVerbosityScope_get_Opened(This,prop)	\
    ( (This)->lpVtbl -> get_Opened(This,prop) ) 

#define MenuVerbosityScope_put_Closed(This,prop)	\
    ( (This)->lpVtbl -> put_Closed(This,prop) ) 

#define MenuVerbosityScope_get_Closed(This,prop)	\
    ( (This)->lpVtbl -> get_Closed(This,prop) ) 

#define MenuVerbosityScope_put_Option(This,prop)	\
    ( (This)->lpVtbl -> put_Option(This,prop) ) 

#define MenuVerbosityScope_get_Option(This,prop)	\
    ( (This)->lpVtbl -> get_Option(This,prop) ) 

#define MenuVerbosityScope_put_Accelerators(This,prop)	\
    ( (This)->lpVtbl -> put_Accelerators(This,prop) ) 

#define MenuVerbosityScope_get_Accelerators(This,prop)	\
    ( (This)->lpVtbl -> get_Accelerators(This,prop) ) 

#define MenuVerbosityScope_put_Descriptions(This,prop)	\
    ( (This)->lpVtbl -> put_Descriptions(This,prop) ) 

#define MenuVerbosityScope_get_Descriptions(This,prop)	\
    ( (This)->lpVtbl -> get_Descriptions(This,prop) ) 

#define MenuVerbosityScope_put_Shortcut(This,prop)	\
    ( (This)->lpVtbl -> put_Shortcut(This,prop) ) 

#define MenuVerbosityScope_get_Shortcut(This,prop)	\
    ( (This)->lpVtbl -> get_Shortcut(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __MenuVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __ActivatedVerbosityScope_INTERFACE_DEFINED__
#define __ActivatedVerbosityScope_INTERFACE_DEFINED__

/* interface ActivatedVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_ActivatedVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("DC114DA2-23B9-4F7E-80D4-EF20D59A5254")
    ActivatedVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Activated( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Activated( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_WindowType( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WindowType( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_WindowTitle( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WindowTitle( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EntireMessageBox( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EntireMessageBox( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EntireDialog( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EntireDialog( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ActivatedVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ActivatedVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ActivatedVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ActivatedVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ActivatedVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ActivatedVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ActivatedVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ActivatedVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ActivatedVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ActivatedVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Activated )( 
            ActivatedVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Activated )( 
            ActivatedVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_WindowType )( 
            ActivatedVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WindowType )( 
            ActivatedVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_WindowTitle )( 
            ActivatedVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WindowTitle )( 
            ActivatedVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EntireMessageBox )( 
            ActivatedVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EntireMessageBox )( 
            ActivatedVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EntireDialog )( 
            ActivatedVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EntireDialog )( 
            ActivatedVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } ActivatedVerbosityScopeVtbl;

    interface ActivatedVerbosityScope
    {
        CONST_VTBL struct ActivatedVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ActivatedVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ActivatedVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ActivatedVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ActivatedVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ActivatedVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ActivatedVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ActivatedVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ActivatedVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ActivatedVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ActivatedVerbosityScope_put_Activated(This,prop)	\
    ( (This)->lpVtbl -> put_Activated(This,prop) ) 

#define ActivatedVerbosityScope_get_Activated(This,prop)	\
    ( (This)->lpVtbl -> get_Activated(This,prop) ) 

#define ActivatedVerbosityScope_put_WindowType(This,prop)	\
    ( (This)->lpVtbl -> put_WindowType(This,prop) ) 

#define ActivatedVerbosityScope_get_WindowType(This,prop)	\
    ( (This)->lpVtbl -> get_WindowType(This,prop) ) 

#define ActivatedVerbosityScope_put_WindowTitle(This,prop)	\
    ( (This)->lpVtbl -> put_WindowTitle(This,prop) ) 

#define ActivatedVerbosityScope_get_WindowTitle(This,prop)	\
    ( (This)->lpVtbl -> get_WindowTitle(This,prop) ) 

#define ActivatedVerbosityScope_put_EntireMessageBox(This,prop)	\
    ( (This)->lpVtbl -> put_EntireMessageBox(This,prop) ) 

#define ActivatedVerbosityScope_get_EntireMessageBox(This,prop)	\
    ( (This)->lpVtbl -> get_EntireMessageBox(This,prop) ) 

#define ActivatedVerbosityScope_put_EntireDialog(This,prop)	\
    ( (This)->lpVtbl -> put_EntireDialog(This,prop) ) 

#define ActivatedVerbosityScope_get_EntireDialog(This,prop)	\
    ( (This)->lpVtbl -> get_EntireDialog(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ActivatedVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __FocusedVerbosityScope_INTERFACE_DEFINED__
#define __FocusedVerbosityScope_INTERFACE_DEFINED__

/* interface FocusedVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_FocusedVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("43B5F04F-0ED8-47A7-A3EA-CDBCA2954BB2")
    FocusedVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Focused( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Focused( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_WindowType( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WindowType( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FieldName( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FieldName( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FieldData( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FieldData( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CustomControl( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CustomControl( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IdentifyMSAAControls( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IdentifyMSAAControls( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateLink( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateLink( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct FocusedVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            FocusedVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            FocusedVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            FocusedVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            FocusedVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            FocusedVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            FocusedVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            FocusedVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            FocusedVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            FocusedVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Focused )( 
            FocusedVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            FocusedVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_WindowType )( 
            FocusedVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WindowType )( 
            FocusedVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FieldName )( 
            FocusedVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FieldName )( 
            FocusedVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FieldData )( 
            FocusedVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FieldData )( 
            FocusedVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CustomControl )( 
            FocusedVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CustomControl )( 
            FocusedVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IdentifyMSAAControls )( 
            FocusedVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IdentifyMSAAControls )( 
            FocusedVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateLink )( 
            FocusedVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateLink )( 
            FocusedVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } FocusedVerbosityScopeVtbl;

    interface FocusedVerbosityScope
    {
        CONST_VTBL struct FocusedVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define FocusedVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define FocusedVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define FocusedVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define FocusedVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define FocusedVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define FocusedVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define FocusedVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define FocusedVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define FocusedVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define FocusedVerbosityScope_put_Focused(This,prop)	\
    ( (This)->lpVtbl -> put_Focused(This,prop) ) 

#define FocusedVerbosityScope_get_Focused(This,prop)	\
    ( (This)->lpVtbl -> get_Focused(This,prop) ) 

#define FocusedVerbosityScope_put_WindowType(This,prop)	\
    ( (This)->lpVtbl -> put_WindowType(This,prop) ) 

#define FocusedVerbosityScope_get_WindowType(This,prop)	\
    ( (This)->lpVtbl -> get_WindowType(This,prop) ) 

#define FocusedVerbosityScope_put_FieldName(This,prop)	\
    ( (This)->lpVtbl -> put_FieldName(This,prop) ) 

#define FocusedVerbosityScope_get_FieldName(This,prop)	\
    ( (This)->lpVtbl -> get_FieldName(This,prop) ) 

#define FocusedVerbosityScope_put_FieldData(This,prop)	\
    ( (This)->lpVtbl -> put_FieldData(This,prop) ) 

#define FocusedVerbosityScope_get_FieldData(This,prop)	\
    ( (This)->lpVtbl -> get_FieldData(This,prop) ) 

#define FocusedVerbosityScope_put_CustomControl(This,prop)	\
    ( (This)->lpVtbl -> put_CustomControl(This,prop) ) 

#define FocusedVerbosityScope_get_CustomControl(This,prop)	\
    ( (This)->lpVtbl -> get_CustomControl(This,prop) ) 

#define FocusedVerbosityScope_put_IdentifyMSAAControls(This,prop)	\
    ( (This)->lpVtbl -> put_IdentifyMSAAControls(This,prop) ) 

#define FocusedVerbosityScope_get_IdentifyMSAAControls(This,prop)	\
    ( (This)->lpVtbl -> get_IdentifyMSAAControls(This,prop) ) 

#define FocusedVerbosityScope_put_IndicateLink(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateLink(This,prop) ) 

#define FocusedVerbosityScope_get_IndicateLink(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateLink(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __FocusedVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __AttributeChangeVerbosityScope_INTERFACE_DEFINED__
#define __AttributeChangeVerbosityScope_INTERFACE_DEFINED__

/* interface AttributeChangeVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_AttributeChangeVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("AB2625DB-32A2-4779-ACA6-FC9CAF5B02B1")
    AttributeChangeVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Color( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Color( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Font( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Font( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_size( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_size( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Style( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Style( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct AttributeChangeVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            AttributeChangeVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            AttributeChangeVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            AttributeChangeVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            AttributeChangeVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            AttributeChangeVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            AttributeChangeVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            AttributeChangeVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            AttributeChangeVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            AttributeChangeVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Color )( 
            AttributeChangeVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Color )( 
            AttributeChangeVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Font )( 
            AttributeChangeVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Font )( 
            AttributeChangeVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_size )( 
            AttributeChangeVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_size )( 
            AttributeChangeVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Style )( 
            AttributeChangeVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Style )( 
            AttributeChangeVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } AttributeChangeVerbosityScopeVtbl;

    interface AttributeChangeVerbosityScope
    {
        CONST_VTBL struct AttributeChangeVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define AttributeChangeVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define AttributeChangeVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define AttributeChangeVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define AttributeChangeVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define AttributeChangeVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define AttributeChangeVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define AttributeChangeVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define AttributeChangeVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define AttributeChangeVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define AttributeChangeVerbosityScope_put_Color(This,prop)	\
    ( (This)->lpVtbl -> put_Color(This,prop) ) 

#define AttributeChangeVerbosityScope_get_Color(This,prop)	\
    ( (This)->lpVtbl -> get_Color(This,prop) ) 

#define AttributeChangeVerbosityScope_put_Font(This,prop)	\
    ( (This)->lpVtbl -> put_Font(This,prop) ) 

#define AttributeChangeVerbosityScope_get_Font(This,prop)	\
    ( (This)->lpVtbl -> get_Font(This,prop) ) 

#define AttributeChangeVerbosityScope_put_size(This,prop)	\
    ( (This)->lpVtbl -> put_size(This,prop) ) 

#define AttributeChangeVerbosityScope_get_size(This,prop)	\
    ( (This)->lpVtbl -> get_size(This,prop) ) 

#define AttributeChangeVerbosityScope_put_Style(This,prop)	\
    ( (This)->lpVtbl -> put_Style(This,prop) ) 

#define AttributeChangeVerbosityScope_get_Style(This,prop)	\
    ( (This)->lpVtbl -> get_Style(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __AttributeChangeVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __MiscVerbosityScope_INTERFACE_DEFINED__
#define __MiscVerbosityScope_INTERFACE_DEFINED__

/* interface MiscVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_MiscVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("ED5F090D-9C93-4653-8AEB-A522E07BF7F5")
    MiscVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Shortcut( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Shortcut( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TopBottomIndex( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TopBottomIndex( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CurrentIndex( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CurrentIndex( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SilencePasswordEdits( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SilencePasswordEdits( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ListviewHeaders( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ListviewHeaders( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct MiscVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            MiscVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            MiscVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            MiscVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            MiscVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            MiscVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            MiscVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            MiscVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            MiscVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            MiscVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Shortcut )( 
            MiscVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Shortcut )( 
            MiscVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TopBottomIndex )( 
            MiscVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TopBottomIndex )( 
            MiscVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CurrentIndex )( 
            MiscVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CurrentIndex )( 
            MiscVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SilencePasswordEdits )( 
            MiscVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SilencePasswordEdits )( 
            MiscVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ListviewHeaders )( 
            MiscVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ListviewHeaders )( 
            MiscVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } MiscVerbosityScopeVtbl;

    interface MiscVerbosityScope
    {
        CONST_VTBL struct MiscVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define MiscVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define MiscVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define MiscVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define MiscVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define MiscVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define MiscVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define MiscVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define MiscVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define MiscVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define MiscVerbosityScope_put_Shortcut(This,prop)	\
    ( (This)->lpVtbl -> put_Shortcut(This,prop) ) 

#define MiscVerbosityScope_get_Shortcut(This,prop)	\
    ( (This)->lpVtbl -> get_Shortcut(This,prop) ) 

#define MiscVerbosityScope_put_TopBottomIndex(This,prop)	\
    ( (This)->lpVtbl -> put_TopBottomIndex(This,prop) ) 

#define MiscVerbosityScope_get_TopBottomIndex(This,prop)	\
    ( (This)->lpVtbl -> get_TopBottomIndex(This,prop) ) 

#define MiscVerbosityScope_put_CurrentIndex(This,prop)	\
    ( (This)->lpVtbl -> put_CurrentIndex(This,prop) ) 

#define MiscVerbosityScope_get_CurrentIndex(This,prop)	\
    ( (This)->lpVtbl -> get_CurrentIndex(This,prop) ) 

#define MiscVerbosityScope_put_SilencePasswordEdits(This,prop)	\
    ( (This)->lpVtbl -> put_SilencePasswordEdits(This,prop) ) 

#define MiscVerbosityScope_get_SilencePasswordEdits(This,prop)	\
    ( (This)->lpVtbl -> get_SilencePasswordEdits(This,prop) ) 

#define MiscVerbosityScope_put_ListviewHeaders(This,prop)	\
    ( (This)->lpVtbl -> put_ListviewHeaders(This,prop) ) 

#define MiscVerbosityScope_get_ListviewHeaders(This,prop)	\
    ( (This)->lpVtbl -> get_ListviewHeaders(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __MiscVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __BrowseModeVerbosityScope_INTERFACE_DEFINED__
#define __BrowseModeVerbosityScope_INTERFACE_DEFINED__

/* interface BrowseModeVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("AC4B0D0E-7984-4150-AB27-B7691133DFFA")
    BrowseModeVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AccessKeys( 
            /* [retval][out] */ BrowseModeAccessKeyVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AcronymsAbbreviations( 
            /* [retval][out] */ BrowseModeAcronymVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Flash( 
            /* [retval][out] */ BrowseModeFlashVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Forms( 
            /* [retval][out] */ BrowseModeFormVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Headings( 
            /* [retval][out] */ BrowseModeHeadingVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Languages( 
            /* [retval][out] */ BrowseModeLanguageVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Links( 
            /* [retval][out] */ BrowseModeLinkVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Lists( 
            /* [retval][out] */ BrowseModeListVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LongDesc( 
            /* [retval][out] */ BrowseModeLongDescVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Miscellaneous( 
            /* [retval][out] */ BrowseModeMiscVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NewPageInformation( 
            /* [retval][out] */ BrowseModeNewPageVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Objects( 
            /* [retval][out] */ BrowseModeObjectVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Paragraphs( 
            /* [retval][out] */ BrowseModeParagraphVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_QuotesBlockQuotes( 
            /* [retval][out] */ BrowseModeQuoteVerbosityScope **obj) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Tables( 
            /* [retval][out] */ BrowseModeTableVerbosityScope **obj) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AccessKeys )( 
            BrowseModeVerbosityScope * This,
            /* [retval][out] */ BrowseModeAccessKeyVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AcronymsAbbreviations )( 
            BrowseModeVerbosityScope * This,
            /* [retval][out] */ BrowseModeAcronymVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Flash )( 
            BrowseModeVerbosityScope * This,
            /* [retval][out] */ BrowseModeFlashVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Forms )( 
            BrowseModeVerbosityScope * This,
            /* [retval][out] */ BrowseModeFormVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Headings )( 
            BrowseModeVerbosityScope * This,
            /* [retval][out] */ BrowseModeHeadingVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Languages )( 
            BrowseModeVerbosityScope * This,
            /* [retval][out] */ BrowseModeLanguageVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Links )( 
            BrowseModeVerbosityScope * This,
            /* [retval][out] */ BrowseModeLinkVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Lists )( 
            BrowseModeVerbosityScope * This,
            /* [retval][out] */ BrowseModeListVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LongDesc )( 
            BrowseModeVerbosityScope * This,
            /* [retval][out] */ BrowseModeLongDescVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Miscellaneous )( 
            BrowseModeVerbosityScope * This,
            /* [retval][out] */ BrowseModeMiscVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NewPageInformation )( 
            BrowseModeVerbosityScope * This,
            /* [retval][out] */ BrowseModeNewPageVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Objects )( 
            BrowseModeVerbosityScope * This,
            /* [retval][out] */ BrowseModeObjectVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Paragraphs )( 
            BrowseModeVerbosityScope * This,
            /* [retval][out] */ BrowseModeParagraphVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_QuotesBlockQuotes )( 
            BrowseModeVerbosityScope * This,
            /* [retval][out] */ BrowseModeQuoteVerbosityScope **obj);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Tables )( 
            BrowseModeVerbosityScope * This,
            /* [retval][out] */ BrowseModeTableVerbosityScope **obj);
        
        END_INTERFACE
    } BrowseModeVerbosityScopeVtbl;

    interface BrowseModeVerbosityScope
    {
        CONST_VTBL struct BrowseModeVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeVerbosityScope_get_AccessKeys(This,obj)	\
    ( (This)->lpVtbl -> get_AccessKeys(This,obj) ) 

#define BrowseModeVerbosityScope_get_AcronymsAbbreviations(This,obj)	\
    ( (This)->lpVtbl -> get_AcronymsAbbreviations(This,obj) ) 

#define BrowseModeVerbosityScope_get_Flash(This,obj)	\
    ( (This)->lpVtbl -> get_Flash(This,obj) ) 

#define BrowseModeVerbosityScope_get_Forms(This,obj)	\
    ( (This)->lpVtbl -> get_Forms(This,obj) ) 

#define BrowseModeVerbosityScope_get_Headings(This,obj)	\
    ( (This)->lpVtbl -> get_Headings(This,obj) ) 

#define BrowseModeVerbosityScope_get_Languages(This,obj)	\
    ( (This)->lpVtbl -> get_Languages(This,obj) ) 

#define BrowseModeVerbosityScope_get_Links(This,obj)	\
    ( (This)->lpVtbl -> get_Links(This,obj) ) 

#define BrowseModeVerbosityScope_get_Lists(This,obj)	\
    ( (This)->lpVtbl -> get_Lists(This,obj) ) 

#define BrowseModeVerbosityScope_get_LongDesc(This,obj)	\
    ( (This)->lpVtbl -> get_LongDesc(This,obj) ) 

#define BrowseModeVerbosityScope_get_Miscellaneous(This,obj)	\
    ( (This)->lpVtbl -> get_Miscellaneous(This,obj) ) 

#define BrowseModeVerbosityScope_get_NewPageInformation(This,obj)	\
    ( (This)->lpVtbl -> get_NewPageInformation(This,obj) ) 

#define BrowseModeVerbosityScope_get_Objects(This,obj)	\
    ( (This)->lpVtbl -> get_Objects(This,obj) ) 

#define BrowseModeVerbosityScope_get_Paragraphs(This,obj)	\
    ( (This)->lpVtbl -> get_Paragraphs(This,obj) ) 

#define BrowseModeVerbosityScope_get_QuotesBlockQuotes(This,obj)	\
    ( (This)->lpVtbl -> get_QuotesBlockQuotes(This,obj) ) 

#define BrowseModeVerbosityScope_get_Tables(This,obj)	\
    ( (This)->lpVtbl -> get_Tables(This,obj) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __BrowseModeAccessKeyVerbosityScope_INTERFACE_DEFINED__
#define __BrowseModeAccessKeyVerbosityScope_INTERFACE_DEFINED__

/* interface BrowseModeAccessKeyVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeAccessKeyVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("B3B06C36-650E-4F33-A94C-807FB4CCB4F5")
    BrowseModeAccessKeyVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indicate( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indicate( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeAccessKeyVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeAccessKeyVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeAccessKeyVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeAccessKeyVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeAccessKeyVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeAccessKeyVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeAccessKeyVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeAccessKeyVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeAccessKeyVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeAccessKeyVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indicate )( 
            BrowseModeAccessKeyVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indicate )( 
            BrowseModeAccessKeyVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeAccessKeyVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeAccessKeyVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } BrowseModeAccessKeyVerbosityScopeVtbl;

    interface BrowseModeAccessKeyVerbosityScope
    {
        CONST_VTBL struct BrowseModeAccessKeyVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeAccessKeyVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeAccessKeyVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeAccessKeyVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeAccessKeyVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeAccessKeyVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeAccessKeyVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeAccessKeyVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeAccessKeyVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeAccessKeyVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeAccessKeyVerbosityScope_put_Indicate(This,prop)	\
    ( (This)->lpVtbl -> put_Indicate(This,prop) ) 

#define BrowseModeAccessKeyVerbosityScope_get_Indicate(This,prop)	\
    ( (This)->lpVtbl -> get_Indicate(This,prop) ) 

#define BrowseModeAccessKeyVerbosityScope_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeAccessKeyVerbosityScope_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeAccessKeyVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __BrowseModeAcronymVerbosityScope_INTERFACE_DEFINED__
#define __BrowseModeAcronymVerbosityScope_INTERFACE_DEFINED__

/* interface BrowseModeAcronymVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeAcronymVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("44E8C463-F909-44FE-B36C-D7CFCF36979C")
    BrowseModeAcronymVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indication( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indication( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeAcronymVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeAcronymVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeAcronymVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeAcronymVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeAcronymVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeAcronymVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeAcronymVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeAcronymVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeAcronymVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeAcronymVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeAcronymVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeAcronymVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indication )( 
            BrowseModeAcronymVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indication )( 
            BrowseModeAcronymVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } BrowseModeAcronymVerbosityScopeVtbl;

    interface BrowseModeAcronymVerbosityScope
    {
        CONST_VTBL struct BrowseModeAcronymVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeAcronymVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeAcronymVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeAcronymVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeAcronymVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeAcronymVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeAcronymVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeAcronymVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeAcronymVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeAcronymVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeAcronymVerbosityScope_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeAcronymVerbosityScope_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#define BrowseModeAcronymVerbosityScope_put_Indication(This,prop)	\
    ( (This)->lpVtbl -> put_Indication(This,prop) ) 

#define BrowseModeAcronymVerbosityScope_get_Indication(This,prop)	\
    ( (This)->lpVtbl -> get_Indication(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeAcronymVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __BrowseModeFlashVerbosityScope_INTERFACE_DEFINED__
#define __BrowseModeFlashVerbosityScope_INTERFACE_DEFINED__

/* interface BrowseModeFlashVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeFlashVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("B46E1DE9-2FED-4FFC-A71F-AA01E1F8C169")
    BrowseModeFlashVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indicate( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indicate( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeFlashVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeFlashVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeFlashVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeFlashVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeFlashVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeFlashVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeFlashVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeFlashVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeFlashVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeFlashVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indicate )( 
            BrowseModeFlashVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indicate )( 
            BrowseModeFlashVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeFlashVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeFlashVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } BrowseModeFlashVerbosityScopeVtbl;

    interface BrowseModeFlashVerbosityScope
    {
        CONST_VTBL struct BrowseModeFlashVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeFlashVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeFlashVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeFlashVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeFlashVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeFlashVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeFlashVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeFlashVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeFlashVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeFlashVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeFlashVerbosityScope_put_Indicate(This,prop)	\
    ( (This)->lpVtbl -> put_Indicate(This,prop) ) 

#define BrowseModeFlashVerbosityScope_get_Indicate(This,prop)	\
    ( (This)->lpVtbl -> get_Indicate(This,prop) ) 

#define BrowseModeFlashVerbosityScope_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeFlashVerbosityScope_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeFlashVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __BrowseModeFormVerbosityScope_INTERFACE_DEFINED__
#define __BrowseModeFormVerbosityScope_INTERFACE_DEFINED__

/* interface BrowseModeFormVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeFormVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("95E8595F-B6E7-468C-9EEE-EE44600EF3F0")
    BrowseModeFormVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateForms( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateForms( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEndForms( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEndForms( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateFieldsets( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateFieldsets( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEndFieldsets( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEndFieldsets( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeFormVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeFormVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeFormVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeFormVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeFormVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeFormVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeFormVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeFormVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeFormVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeFormVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateForms )( 
            BrowseModeFormVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateForms )( 
            BrowseModeFormVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEndForms )( 
            BrowseModeFormVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEndForms )( 
            BrowseModeFormVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateFieldsets )( 
            BrowseModeFormVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateFieldsets )( 
            BrowseModeFormVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEndFieldsets )( 
            BrowseModeFormVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEndFieldsets )( 
            BrowseModeFormVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } BrowseModeFormVerbosityScopeVtbl;

    interface BrowseModeFormVerbosityScope
    {
        CONST_VTBL struct BrowseModeFormVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeFormVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeFormVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeFormVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeFormVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeFormVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeFormVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeFormVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeFormVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeFormVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeFormVerbosityScope_put_IndicateForms(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateForms(This,prop) ) 

#define BrowseModeFormVerbosityScope_get_IndicateForms(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateForms(This,prop) ) 

#define BrowseModeFormVerbosityScope_put_ReadToEndForms(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEndForms(This,prop) ) 

#define BrowseModeFormVerbosityScope_get_ReadToEndForms(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEndForms(This,prop) ) 

#define BrowseModeFormVerbosityScope_put_IndicateFieldsets(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateFieldsets(This,prop) ) 

#define BrowseModeFormVerbosityScope_get_IndicateFieldsets(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateFieldsets(This,prop) ) 

#define BrowseModeFormVerbosityScope_put_ReadToEndFieldsets(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEndFieldsets(This,prop) ) 

#define BrowseModeFormVerbosityScope_get_ReadToEndFieldsets(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEndFieldsets(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeFormVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __BrowseModeHeadingVerbosityScope_INTERFACE_DEFINED__
#define __BrowseModeHeadingVerbosityScope_INTERFACE_DEFINED__

/* interface BrowseModeHeadingVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeHeadingVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("64A65943-A1CC-4096-A9D3-119ECC57C080")
    BrowseModeHeadingVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indicate( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indicate( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeHeadingVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeHeadingVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeHeadingVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeHeadingVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeHeadingVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeHeadingVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeHeadingVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeHeadingVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeHeadingVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeHeadingVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indicate )( 
            BrowseModeHeadingVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indicate )( 
            BrowseModeHeadingVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeHeadingVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeHeadingVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } BrowseModeHeadingVerbosityScopeVtbl;

    interface BrowseModeHeadingVerbosityScope
    {
        CONST_VTBL struct BrowseModeHeadingVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeHeadingVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeHeadingVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeHeadingVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeHeadingVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeHeadingVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeHeadingVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeHeadingVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeHeadingVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeHeadingVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeHeadingVerbosityScope_put_Indicate(This,prop)	\
    ( (This)->lpVtbl -> put_Indicate(This,prop) ) 

#define BrowseModeHeadingVerbosityScope_get_Indicate(This,prop)	\
    ( (This)->lpVtbl -> get_Indicate(This,prop) ) 

#define BrowseModeHeadingVerbosityScope_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeHeadingVerbosityScope_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeHeadingVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __BrowseModeLanguageVerbosityScope_INTERFACE_DEFINED__
#define __BrowseModeLanguageVerbosityScope_INTERFACE_DEFINED__

/* interface BrowseModeLanguageVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeLanguageVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("9617234B-C55A-40EA-B30C-58D9114810B7")
    BrowseModeLanguageVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indicate( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indicate( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AutoSwitchSynthesizer( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AutoSwitchSynthesizer( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeLanguageVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeLanguageVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeLanguageVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeLanguageVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeLanguageVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeLanguageVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeLanguageVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeLanguageVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeLanguageVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeLanguageVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indicate )( 
            BrowseModeLanguageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indicate )( 
            BrowseModeLanguageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeLanguageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeLanguageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AutoSwitchSynthesizer )( 
            BrowseModeLanguageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AutoSwitchSynthesizer )( 
            BrowseModeLanguageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } BrowseModeLanguageVerbosityScopeVtbl;

    interface BrowseModeLanguageVerbosityScope
    {
        CONST_VTBL struct BrowseModeLanguageVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeLanguageVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeLanguageVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeLanguageVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeLanguageVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeLanguageVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeLanguageVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeLanguageVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeLanguageVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeLanguageVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeLanguageVerbosityScope_put_Indicate(This,prop)	\
    ( (This)->lpVtbl -> put_Indicate(This,prop) ) 

#define BrowseModeLanguageVerbosityScope_get_Indicate(This,prop)	\
    ( (This)->lpVtbl -> get_Indicate(This,prop) ) 

#define BrowseModeLanguageVerbosityScope_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeLanguageVerbosityScope_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#define BrowseModeLanguageVerbosityScope_put_AutoSwitchSynthesizer(This,prop)	\
    ( (This)->lpVtbl -> put_AutoSwitchSynthesizer(This,prop) ) 

#define BrowseModeLanguageVerbosityScope_get_AutoSwitchSynthesizer(This,prop)	\
    ( (This)->lpVtbl -> get_AutoSwitchSynthesizer(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeLanguageVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __BrowseModeLinkVerbosityScope_INTERFACE_DEFINED__
#define __BrowseModeLinkVerbosityScope_INTERFACE_DEFINED__

/* interface BrowseModeLinkVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeLinkVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("0696AB43-3979-44B1-AA09-7379B4E9E5A4")
    BrowseModeLinkVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IdentifyVisitedLinks( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IdentifyVisitedLinks( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IdentifySamePageLinks( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IdentifySamePageLinks( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indication( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indication( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeLinkVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeLinkVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeLinkVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeLinkVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeLinkVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeLinkVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeLinkVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeLinkVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeLinkVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeLinkVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeLinkVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeLinkVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IdentifyVisitedLinks )( 
            BrowseModeLinkVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IdentifyVisitedLinks )( 
            BrowseModeLinkVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IdentifySamePageLinks )( 
            BrowseModeLinkVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IdentifySamePageLinks )( 
            BrowseModeLinkVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indication )( 
            BrowseModeLinkVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indication )( 
            BrowseModeLinkVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } BrowseModeLinkVerbosityScopeVtbl;

    interface BrowseModeLinkVerbosityScope
    {
        CONST_VTBL struct BrowseModeLinkVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeLinkVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeLinkVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeLinkVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeLinkVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeLinkVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeLinkVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeLinkVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeLinkVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeLinkVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeLinkVerbosityScope_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeLinkVerbosityScope_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#define BrowseModeLinkVerbosityScope_put_IdentifyVisitedLinks(This,prop)	\
    ( (This)->lpVtbl -> put_IdentifyVisitedLinks(This,prop) ) 

#define BrowseModeLinkVerbosityScope_get_IdentifyVisitedLinks(This,prop)	\
    ( (This)->lpVtbl -> get_IdentifyVisitedLinks(This,prop) ) 

#define BrowseModeLinkVerbosityScope_put_IdentifySamePageLinks(This,prop)	\
    ( (This)->lpVtbl -> put_IdentifySamePageLinks(This,prop) ) 

#define BrowseModeLinkVerbosityScope_get_IdentifySamePageLinks(This,prop)	\
    ( (This)->lpVtbl -> get_IdentifySamePageLinks(This,prop) ) 

#define BrowseModeLinkVerbosityScope_put_Indication(This,prop)	\
    ( (This)->lpVtbl -> put_Indication(This,prop) ) 

#define BrowseModeLinkVerbosityScope_get_Indication(This,prop)	\
    ( (This)->lpVtbl -> get_Indication(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeLinkVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __BrowseModeListVerbosityScope_INTERFACE_DEFINED__
#define __BrowseModeListVerbosityScope_INTERFACE_DEFINED__

/* interface BrowseModeListVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeListVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("DDF6295A-63CC-4A94-A8E3-1F29B771FD34")
    BrowseModeListVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indicate( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indicate( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateBullets( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateBullets( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEndBullets( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEndBullets( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeListVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeListVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeListVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeListVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeListVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeListVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeListVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeListVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeListVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeListVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indicate )( 
            BrowseModeListVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indicate )( 
            BrowseModeListVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeListVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeListVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateBullets )( 
            BrowseModeListVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateBullets )( 
            BrowseModeListVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEndBullets )( 
            BrowseModeListVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEndBullets )( 
            BrowseModeListVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } BrowseModeListVerbosityScopeVtbl;

    interface BrowseModeListVerbosityScope
    {
        CONST_VTBL struct BrowseModeListVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeListVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeListVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeListVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeListVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeListVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeListVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeListVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeListVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeListVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeListVerbosityScope_put_Indicate(This,prop)	\
    ( (This)->lpVtbl -> put_Indicate(This,prop) ) 

#define BrowseModeListVerbosityScope_get_Indicate(This,prop)	\
    ( (This)->lpVtbl -> get_Indicate(This,prop) ) 

#define BrowseModeListVerbosityScope_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeListVerbosityScope_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#define BrowseModeListVerbosityScope_put_IndicateBullets(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateBullets(This,prop) ) 

#define BrowseModeListVerbosityScope_get_IndicateBullets(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateBullets(This,prop) ) 

#define BrowseModeListVerbosityScope_put_ReadToEndBullets(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEndBullets(This,prop) ) 

#define BrowseModeListVerbosityScope_get_ReadToEndBullets(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEndBullets(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeListVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __BrowseModeLongDescVerbosityScope_INTERFACE_DEFINED__
#define __BrowseModeLongDescVerbosityScope_INTERFACE_DEFINED__

/* interface BrowseModeLongDescVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeLongDescVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("A29DCA3D-496B-40C9-A72C-7AACE04FC8A9")
    BrowseModeLongDescVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indicate( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indicate( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeLongDescVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeLongDescVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeLongDescVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeLongDescVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeLongDescVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeLongDescVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeLongDescVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeLongDescVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeLongDescVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeLongDescVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indicate )( 
            BrowseModeLongDescVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indicate )( 
            BrowseModeLongDescVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeLongDescVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeLongDescVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } BrowseModeLongDescVerbosityScopeVtbl;

    interface BrowseModeLongDescVerbosityScope
    {
        CONST_VTBL struct BrowseModeLongDescVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeLongDescVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeLongDescVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeLongDescVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeLongDescVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeLongDescVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeLongDescVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeLongDescVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeLongDescVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeLongDescVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeLongDescVerbosityScope_put_Indicate(This,prop)	\
    ( (This)->lpVtbl -> put_Indicate(This,prop) ) 

#define BrowseModeLongDescVerbosityScope_get_Indicate(This,prop)	\
    ( (This)->lpVtbl -> get_Indicate(This,prop) ) 

#define BrowseModeLongDescVerbosityScope_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeLongDescVerbosityScope_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeLongDescVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __BrowseModeMiscVerbosityScope_INTERFACE_DEFINED__
#define __BrowseModeMiscVerbosityScope_INTERFACE_DEFINED__

/* interface BrowseModeMiscVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeMiscVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("51E2B4A8-060A-4FA6-94C5-E2F7B47A3D21")
    BrowseModeMiscVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IndicateNewPopups( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IndicateNewPopups( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeGraphicWithNoDescription( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeGraphicWithNoDescription( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_VerbosityHotKeyToSpecificElement( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_VerbosityHotKeyToSpecificElement( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AutoLoad( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AutoLoad( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NextTextMinimumLineLength( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NextTextMinimumLineLength( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NextTextConsecutiveLines( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NextTextConsecutiveLines( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SuspendRefresh( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SuspendRefresh( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AutoToggle( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AutoToggle( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ModeIndicator( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ModeIndicator( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_QuickReview( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_QuickReview( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeMiscVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeMiscVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeMiscVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeMiscVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeMiscVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeMiscVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeMiscVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeMiscVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeMiscVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeMiscVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IndicateNewPopups )( 
            BrowseModeMiscVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IndicateNewPopups )( 
            BrowseModeMiscVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeGraphicWithNoDescription )( 
            BrowseModeMiscVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeGraphicWithNoDescription )( 
            BrowseModeMiscVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_VerbosityHotKeyToSpecificElement )( 
            BrowseModeMiscVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_VerbosityHotKeyToSpecificElement )( 
            BrowseModeMiscVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AutoLoad )( 
            BrowseModeMiscVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AutoLoad )( 
            BrowseModeMiscVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NextTextMinimumLineLength )( 
            BrowseModeMiscVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NextTextMinimumLineLength )( 
            BrowseModeMiscVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NextTextConsecutiveLines )( 
            BrowseModeMiscVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NextTextConsecutiveLines )( 
            BrowseModeMiscVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SuspendRefresh )( 
            BrowseModeMiscVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SuspendRefresh )( 
            BrowseModeMiscVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AutoToggle )( 
            BrowseModeMiscVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AutoToggle )( 
            BrowseModeMiscVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ModeIndicator )( 
            BrowseModeMiscVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ModeIndicator )( 
            BrowseModeMiscVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_QuickReview )( 
            BrowseModeMiscVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_QuickReview )( 
            BrowseModeMiscVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } BrowseModeMiscVerbosityScopeVtbl;

    interface BrowseModeMiscVerbosityScope
    {
        CONST_VTBL struct BrowseModeMiscVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeMiscVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeMiscVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeMiscVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeMiscVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeMiscVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeMiscVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeMiscVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeMiscVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeMiscVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeMiscVerbosityScope_put_IndicateNewPopups(This,prop)	\
    ( (This)->lpVtbl -> put_IndicateNewPopups(This,prop) ) 

#define BrowseModeMiscVerbosityScope_get_IndicateNewPopups(This,prop)	\
    ( (This)->lpVtbl -> get_IndicateNewPopups(This,prop) ) 

#define BrowseModeMiscVerbosityScope_put_IncludeGraphicWithNoDescription(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeGraphicWithNoDescription(This,prop) ) 

#define BrowseModeMiscVerbosityScope_get_IncludeGraphicWithNoDescription(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeGraphicWithNoDescription(This,prop) ) 

#define BrowseModeMiscVerbosityScope_put_VerbosityHotKeyToSpecificElement(This,prop)	\
    ( (This)->lpVtbl -> put_VerbosityHotKeyToSpecificElement(This,prop) ) 

#define BrowseModeMiscVerbosityScope_get_VerbosityHotKeyToSpecificElement(This,prop)	\
    ( (This)->lpVtbl -> get_VerbosityHotKeyToSpecificElement(This,prop) ) 

#define BrowseModeMiscVerbosityScope_put_AutoLoad(This,prop)	\
    ( (This)->lpVtbl -> put_AutoLoad(This,prop) ) 

#define BrowseModeMiscVerbosityScope_get_AutoLoad(This,prop)	\
    ( (This)->lpVtbl -> get_AutoLoad(This,prop) ) 

#define BrowseModeMiscVerbosityScope_put_NextTextMinimumLineLength(This,prop)	\
    ( (This)->lpVtbl -> put_NextTextMinimumLineLength(This,prop) ) 

#define BrowseModeMiscVerbosityScope_get_NextTextMinimumLineLength(This,prop)	\
    ( (This)->lpVtbl -> get_NextTextMinimumLineLength(This,prop) ) 

#define BrowseModeMiscVerbosityScope_put_NextTextConsecutiveLines(This,prop)	\
    ( (This)->lpVtbl -> put_NextTextConsecutiveLines(This,prop) ) 

#define BrowseModeMiscVerbosityScope_get_NextTextConsecutiveLines(This,prop)	\
    ( (This)->lpVtbl -> get_NextTextConsecutiveLines(This,prop) ) 

#define BrowseModeMiscVerbosityScope_put_SuspendRefresh(This,prop)	\
    ( (This)->lpVtbl -> put_SuspendRefresh(This,prop) ) 

#define BrowseModeMiscVerbosityScope_get_SuspendRefresh(This,prop)	\
    ( (This)->lpVtbl -> get_SuspendRefresh(This,prop) ) 

#define BrowseModeMiscVerbosityScope_put_AutoToggle(This,prop)	\
    ( (This)->lpVtbl -> put_AutoToggle(This,prop) ) 

#define BrowseModeMiscVerbosityScope_get_AutoToggle(This,prop)	\
    ( (This)->lpVtbl -> get_AutoToggle(This,prop) ) 

#define BrowseModeMiscVerbosityScope_put_ModeIndicator(This,prop)	\
    ( (This)->lpVtbl -> put_ModeIndicator(This,prop) ) 

#define BrowseModeMiscVerbosityScope_get_ModeIndicator(This,prop)	\
    ( (This)->lpVtbl -> get_ModeIndicator(This,prop) ) 

#define BrowseModeMiscVerbosityScope_put_QuickReview(This,prop)	\
    ( (This)->lpVtbl -> put_QuickReview(This,prop) ) 

#define BrowseModeMiscVerbosityScope_get_QuickReview(This,prop)	\
    ( (This)->lpVtbl -> get_QuickReview(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeMiscVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __BrowseModeNewPageVerbosityScope_INTERFACE_DEFINED__
#define __BrowseModeNewPageVerbosityScope_INTERFACE_DEFINED__

/* interface BrowseModeNewPageVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeNewPageVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("6E71A7D0-6353-411F-9382-A259DAD28256")
    BrowseModeNewPageVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_StatusMessage( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_StatusMessage( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfLines( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfLines( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfLinks( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfLinks( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfFrames( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfFrames( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfTables( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfTables( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfLists( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfLists( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfParagraphs( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfParagraphs( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfHeadings( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfHeadings( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfForms( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfForms( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfAcronymsAbbreviations( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfAcronymsAbbreviations( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfBlockQuotes( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfBlockQuotes( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfLanguages( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfLanguages( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfLongDesc( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfLongDesc( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfAnchors( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfAnchors( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfAccessKeys( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfAccessKeys( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfFlashObjects( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfFlashObjects( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfObjects( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfObjects( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_RefreshRate( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_RefreshRate( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadEntirePage( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadEntirePage( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_NumberOfLandmarks( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NumberOfLandmarks( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeNewPageVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeNewPageVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeNewPageVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_StatusMessage )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_StatusMessage )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfLines )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfLines )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfLinks )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfLinks )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfFrames )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfFrames )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfTables )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfTables )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfLists )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfLists )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfParagraphs )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfParagraphs )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfHeadings )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfHeadings )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfForms )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfForms )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfAcronymsAbbreviations )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfAcronymsAbbreviations )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfBlockQuotes )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfBlockQuotes )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfLanguages )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfLanguages )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfLongDesc )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfLongDesc )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfAnchors )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfAnchors )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfAccessKeys )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfAccessKeys )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfFlashObjects )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfFlashObjects )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfObjects )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfObjects )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_RefreshRate )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_RefreshRate )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadEntirePage )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadEntirePage )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_NumberOfLandmarks )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NumberOfLandmarks )( 
            BrowseModeNewPageVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } BrowseModeNewPageVerbosityScopeVtbl;

    interface BrowseModeNewPageVerbosityScope
    {
        CONST_VTBL struct BrowseModeNewPageVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeNewPageVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeNewPageVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeNewPageVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeNewPageVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeNewPageVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeNewPageVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeNewPageVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeNewPageVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeNewPageVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeNewPageVerbosityScope_put_StatusMessage(This,prop)	\
    ( (This)->lpVtbl -> put_StatusMessage(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_StatusMessage(This,prop)	\
    ( (This)->lpVtbl -> get_StatusMessage(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_NumberOfLines(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfLines(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_NumberOfLines(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfLines(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_NumberOfLinks(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfLinks(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_NumberOfLinks(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfLinks(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_NumberOfFrames(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfFrames(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_NumberOfFrames(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfFrames(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_NumberOfTables(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfTables(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_NumberOfTables(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfTables(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_NumberOfLists(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfLists(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_NumberOfLists(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfLists(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_NumberOfParagraphs(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfParagraphs(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_NumberOfParagraphs(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfParagraphs(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_NumberOfHeadings(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfHeadings(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_NumberOfHeadings(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfHeadings(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_NumberOfForms(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfForms(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_NumberOfForms(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfForms(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_NumberOfAcronymsAbbreviations(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfAcronymsAbbreviations(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_NumberOfAcronymsAbbreviations(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfAcronymsAbbreviations(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_NumberOfBlockQuotes(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfBlockQuotes(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_NumberOfBlockQuotes(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfBlockQuotes(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_NumberOfLanguages(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfLanguages(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_NumberOfLanguages(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfLanguages(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_NumberOfLongDesc(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfLongDesc(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_NumberOfLongDesc(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfLongDesc(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_NumberOfAnchors(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfAnchors(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_NumberOfAnchors(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfAnchors(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_NumberOfAccessKeys(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfAccessKeys(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_NumberOfAccessKeys(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfAccessKeys(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_NumberOfFlashObjects(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfFlashObjects(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_NumberOfFlashObjects(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfFlashObjects(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_NumberOfObjects(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfObjects(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_NumberOfObjects(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfObjects(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_RefreshRate(This,prop)	\
    ( (This)->lpVtbl -> put_RefreshRate(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_RefreshRate(This,prop)	\
    ( (This)->lpVtbl -> get_RefreshRate(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_ReadEntirePage(This,prop)	\
    ( (This)->lpVtbl -> put_ReadEntirePage(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_ReadEntirePage(This,prop)	\
    ( (This)->lpVtbl -> get_ReadEntirePage(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_put_NumberOfLandmarks(This,prop)	\
    ( (This)->lpVtbl -> put_NumberOfLandmarks(This,prop) ) 

#define BrowseModeNewPageVerbosityScope_get_NumberOfLandmarks(This,prop)	\
    ( (This)->lpVtbl -> get_NumberOfLandmarks(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeNewPageVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __BrowseModeObjectVerbosityScope_INTERFACE_DEFINED__
#define __BrowseModeObjectVerbosityScope_INTERFACE_DEFINED__

/* interface BrowseModeObjectVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeObjectVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("B3A95367-3F11-4F5F-8396-DD04F9AF54F6")
    BrowseModeObjectVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeType( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeType( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ExposeAlternateText( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ExposeAlternateText( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeObjectVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeObjectVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeObjectVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeObjectVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeObjectVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeObjectVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeObjectVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeObjectVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeObjectVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeObjectVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeType )( 
            BrowseModeObjectVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeType )( 
            BrowseModeObjectVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ExposeAlternateText )( 
            BrowseModeObjectVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ExposeAlternateText )( 
            BrowseModeObjectVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeObjectVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeObjectVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } BrowseModeObjectVerbosityScopeVtbl;

    interface BrowseModeObjectVerbosityScope
    {
        CONST_VTBL struct BrowseModeObjectVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeObjectVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeObjectVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeObjectVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeObjectVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeObjectVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeObjectVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeObjectVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeObjectVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeObjectVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeObjectVerbosityScope_put_IncludeType(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeType(This,prop) ) 

#define BrowseModeObjectVerbosityScope_get_IncludeType(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeType(This,prop) ) 

#define BrowseModeObjectVerbosityScope_put_ExposeAlternateText(This,prop)	\
    ( (This)->lpVtbl -> put_ExposeAlternateText(This,prop) ) 

#define BrowseModeObjectVerbosityScope_get_ExposeAlternateText(This,prop)	\
    ( (This)->lpVtbl -> get_ExposeAlternateText(This,prop) ) 

#define BrowseModeObjectVerbosityScope_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeObjectVerbosityScope_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeObjectVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __BrowseModeParagraphVerbosityScope_INTERFACE_DEFINED__
#define __BrowseModeParagraphVerbosityScope_INTERFACE_DEFINED__

/* interface BrowseModeParagraphVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeParagraphVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("C9F99E65-5A95-431E-AA11-2FE3140F9FED")
    BrowseModeParagraphVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indicate( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indicate( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SmartParagraphMode( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SmartParagraphMode( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ShowBlankLines( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ShowBlankLines( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeParagraphVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeParagraphVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeParagraphVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeParagraphVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeParagraphVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeParagraphVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeParagraphVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeParagraphVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeParagraphVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeParagraphVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indicate )( 
            BrowseModeParagraphVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indicate )( 
            BrowseModeParagraphVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeParagraphVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeParagraphVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SmartParagraphMode )( 
            BrowseModeParagraphVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SmartParagraphMode )( 
            BrowseModeParagraphVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ShowBlankLines )( 
            BrowseModeParagraphVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ShowBlankLines )( 
            BrowseModeParagraphVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } BrowseModeParagraphVerbosityScopeVtbl;

    interface BrowseModeParagraphVerbosityScope
    {
        CONST_VTBL struct BrowseModeParagraphVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeParagraphVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeParagraphVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeParagraphVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeParagraphVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeParagraphVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeParagraphVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeParagraphVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeParagraphVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeParagraphVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeParagraphVerbosityScope_put_Indicate(This,prop)	\
    ( (This)->lpVtbl -> put_Indicate(This,prop) ) 

#define BrowseModeParagraphVerbosityScope_get_Indicate(This,prop)	\
    ( (This)->lpVtbl -> get_Indicate(This,prop) ) 

#define BrowseModeParagraphVerbosityScope_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeParagraphVerbosityScope_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#define BrowseModeParagraphVerbosityScope_put_SmartParagraphMode(This,prop)	\
    ( (This)->lpVtbl -> put_SmartParagraphMode(This,prop) ) 

#define BrowseModeParagraphVerbosityScope_get_SmartParagraphMode(This,prop)	\
    ( (This)->lpVtbl -> get_SmartParagraphMode(This,prop) ) 

#define BrowseModeParagraphVerbosityScope_put_ShowBlankLines(This,prop)	\
    ( (This)->lpVtbl -> put_ShowBlankLines(This,prop) ) 

#define BrowseModeParagraphVerbosityScope_get_ShowBlankLines(This,prop)	\
    ( (This)->lpVtbl -> get_ShowBlankLines(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeParagraphVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __BrowseModeQuoteVerbosityScope_INTERFACE_DEFINED__
#define __BrowseModeQuoteVerbosityScope_INTERFACE_DEFINED__

/* interface BrowseModeQuoteVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeQuoteVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("FB6EC9D5-60A6-43A4-9E69-FD29BF1DBFCB")
    BrowseModeQuoteVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indicate( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indicate( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeQuoteVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeQuoteVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeQuoteVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeQuoteVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeQuoteVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeQuoteVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeQuoteVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeQuoteVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeQuoteVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeQuoteVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indicate )( 
            BrowseModeQuoteVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indicate )( 
            BrowseModeQuoteVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeQuoteVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeQuoteVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } BrowseModeQuoteVerbosityScopeVtbl;

    interface BrowseModeQuoteVerbosityScope
    {
        CONST_VTBL struct BrowseModeQuoteVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeQuoteVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeQuoteVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeQuoteVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeQuoteVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeQuoteVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeQuoteVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeQuoteVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeQuoteVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeQuoteVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeQuoteVerbosityScope_put_Indicate(This,prop)	\
    ( (This)->lpVtbl -> put_Indicate(This,prop) ) 

#define BrowseModeQuoteVerbosityScope_get_Indicate(This,prop)	\
    ( (This)->lpVtbl -> get_Indicate(This,prop) ) 

#define BrowseModeQuoteVerbosityScope_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeQuoteVerbosityScope_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeQuoteVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __BrowseModeTableVerbosityScope_INTERFACE_DEFINED__
#define __BrowseModeTableVerbosityScope_INTERFACE_DEFINED__

/* interface BrowseModeTableVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeTableVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("2FFC450F-E611-4B78-8113-7D32FE6BF0FC")
    BrowseModeTableVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadToEnd( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadToEnd( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeSummary( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeSummary( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeSize( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeSize( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeRowGrouping( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeRowGrouping( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_IncludeMergedCells( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IncludeMergedCells( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_HeaderAttributeOverride( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HeaderAttributeOverride( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indication( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indication( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_DefaultHeaderTo( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DefaultHeaderTo( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeTableVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeTableVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeTableVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeTableVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeTableVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeTableVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeTableVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeTableVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeTableVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeTableVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadToEnd )( 
            BrowseModeTableVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadToEnd )( 
            BrowseModeTableVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeSummary )( 
            BrowseModeTableVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeSummary )( 
            BrowseModeTableVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeSize )( 
            BrowseModeTableVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeSize )( 
            BrowseModeTableVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeRowGrouping )( 
            BrowseModeTableVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeRowGrouping )( 
            BrowseModeTableVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_IncludeMergedCells )( 
            BrowseModeTableVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IncludeMergedCells )( 
            BrowseModeTableVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_HeaderAttributeOverride )( 
            BrowseModeTableVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HeaderAttributeOverride )( 
            BrowseModeTableVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indication )( 
            BrowseModeTableVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indication )( 
            BrowseModeTableVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_DefaultHeaderTo )( 
            BrowseModeTableVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DefaultHeaderTo )( 
            BrowseModeTableVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } BrowseModeTableVerbosityScopeVtbl;

    interface BrowseModeTableVerbosityScope
    {
        CONST_VTBL struct BrowseModeTableVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeTableVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeTableVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeTableVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeTableVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeTableVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeTableVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeTableVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeTableVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeTableVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeTableVerbosityScope_put_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> put_ReadToEnd(This,prop) ) 

#define BrowseModeTableVerbosityScope_get_ReadToEnd(This,prop)	\
    ( (This)->lpVtbl -> get_ReadToEnd(This,prop) ) 

#define BrowseModeTableVerbosityScope_put_IncludeSummary(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeSummary(This,prop) ) 

#define BrowseModeTableVerbosityScope_get_IncludeSummary(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeSummary(This,prop) ) 

#define BrowseModeTableVerbosityScope_put_IncludeSize(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeSize(This,prop) ) 

#define BrowseModeTableVerbosityScope_get_IncludeSize(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeSize(This,prop) ) 

#define BrowseModeTableVerbosityScope_put_IncludeRowGrouping(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeRowGrouping(This,prop) ) 

#define BrowseModeTableVerbosityScope_get_IncludeRowGrouping(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeRowGrouping(This,prop) ) 

#define BrowseModeTableVerbosityScope_put_IncludeMergedCells(This,prop)	\
    ( (This)->lpVtbl -> put_IncludeMergedCells(This,prop) ) 

#define BrowseModeTableVerbosityScope_get_IncludeMergedCells(This,prop)	\
    ( (This)->lpVtbl -> get_IncludeMergedCells(This,prop) ) 

#define BrowseModeTableVerbosityScope_put_HeaderAttributeOverride(This,prop)	\
    ( (This)->lpVtbl -> put_HeaderAttributeOverride(This,prop) ) 

#define BrowseModeTableVerbosityScope_get_HeaderAttributeOverride(This,prop)	\
    ( (This)->lpVtbl -> get_HeaderAttributeOverride(This,prop) ) 

#define BrowseModeTableVerbosityScope_put_Indication(This,prop)	\
    ( (This)->lpVtbl -> put_Indication(This,prop) ) 

#define BrowseModeTableVerbosityScope_get_Indication(This,prop)	\
    ( (This)->lpVtbl -> get_Indication(This,prop) ) 

#define BrowseModeTableVerbosityScope_put_DefaultHeaderTo(This,prop)	\
    ( (This)->lpVtbl -> put_DefaultHeaderTo(This,prop) ) 

#define BrowseModeTableVerbosityScope_get_DefaultHeaderTo(This,prop)	\
    ( (This)->lpVtbl -> get_DefaultHeaderTo(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeTableVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __MouseVerbosityScope_INTERFACE_DEFINED__
#define __MouseVerbosityScope_INTERFACE_DEFINED__

/* interface MouseVerbosityScope */
/* [object][oleautomation][dual][version][uuid] */ 


EXTERN_C const IID IID_MouseVerbosityScope;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("462EC9EF-4EC6-458C-8797-036428907A27")
    MouseVerbosityScope : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_In( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_In( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Out( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Out( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ClipChange( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ClipChange( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_LineChange( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LineChange( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_WindowChange( 
            /* [in] */ SetFileScopes prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_WindowChange( 
            /* [retval][out] */ SetFileScopes *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct MouseVerbosityScopeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            MouseVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            MouseVerbosityScope * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            MouseVerbosityScope * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            MouseVerbosityScope * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            MouseVerbosityScope * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            MouseVerbosityScope * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            MouseVerbosityScope * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            MouseVerbosityScope * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            MouseVerbosityScope * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_In )( 
            MouseVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_In )( 
            MouseVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Out )( 
            MouseVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Out )( 
            MouseVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ClipChange )( 
            MouseVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ClipChange )( 
            MouseVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_LineChange )( 
            MouseVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LineChange )( 
            MouseVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_WindowChange )( 
            MouseVerbosityScope * This,
            /* [in] */ SetFileScopes prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_WindowChange )( 
            MouseVerbosityScope * This,
            /* [retval][out] */ SetFileScopes *prop);
        
        END_INTERFACE
    } MouseVerbosityScopeVtbl;

    interface MouseVerbosityScope
    {
        CONST_VTBL struct MouseVerbosityScopeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define MouseVerbosityScope_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define MouseVerbosityScope_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define MouseVerbosityScope_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define MouseVerbosityScope_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define MouseVerbosityScope_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define MouseVerbosityScope_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define MouseVerbosityScope_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define MouseVerbosityScope_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define MouseVerbosityScope_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define MouseVerbosityScope_put_In(This,prop)	\
    ( (This)->lpVtbl -> put_In(This,prop) ) 

#define MouseVerbosityScope_get_In(This,prop)	\
    ( (This)->lpVtbl -> get_In(This,prop) ) 

#define MouseVerbosityScope_put_Out(This,prop)	\
    ( (This)->lpVtbl -> put_Out(This,prop) ) 

#define MouseVerbosityScope_get_Out(This,prop)	\
    ( (This)->lpVtbl -> get_Out(This,prop) ) 

#define MouseVerbosityScope_put_ClipChange(This,prop)	\
    ( (This)->lpVtbl -> put_ClipChange(This,prop) ) 

#define MouseVerbosityScope_get_ClipChange(This,prop)	\
    ( (This)->lpVtbl -> get_ClipChange(This,prop) ) 

#define MouseVerbosityScope_put_LineChange(This,prop)	\
    ( (This)->lpVtbl -> put_LineChange(This,prop) ) 

#define MouseVerbosityScope_get_LineChange(This,prop)	\
    ( (This)->lpVtbl -> get_LineChange(This,prop) ) 

#define MouseVerbosityScope_put_WindowChange(This,prop)	\
    ( (This)->lpVtbl -> put_WindowChange(This,prop) ) 

#define MouseVerbosityScope_get_WindowChange(This,prop)	\
    ( (This)->lpVtbl -> get_WindowChange(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __MouseVerbosityScope_INTERFACE_DEFINED__ */


#ifndef __Settings2_INTERFACE_DEFINED__
#define __Settings2_INTERFACE_DEFINED__

/* interface Settings2 */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Settings2;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("589E2BBB-AD90-4DD2-91EA-A80BC31372B1")
    Settings2 : public IDispatch
    {
    public:
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Effective( 
            /* [retval][out] */ SetFile **settings) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Global( 
            /* [retval][out] */ SetFile **settings) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct Settings2Vtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Settings2 * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Settings2 * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Settings2 * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Settings2 * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Settings2 * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Settings2 * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Settings2 * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Settings2 * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Settings2 * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Effective )( 
            Settings2 * This,
            /* [retval][out] */ SetFile **settings);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Global )( 
            Settings2 * This,
            /* [retval][out] */ SetFile **settings);
        
        END_INTERFACE
    } Settings2Vtbl;

    interface Settings2
    {
        CONST_VTBL struct Settings2Vtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Settings2_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Settings2_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Settings2_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Settings2_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Settings2_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Settings2_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Settings2_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Settings2_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Settings2_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Settings2_get_Effective(This,settings)	\
    ( (This)->lpVtbl -> get_Effective(This,settings) ) 

#define Settings2_get_Global(This,settings)	\
    ( (This)->lpVtbl -> get_Global(This,settings) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Settings2_INTERFACE_DEFINED__ */


#ifndef ___Mouse_INTERFACE_DEFINED__
#define ___Mouse_INTERFACE_DEFINED__

/* interface _Mouse */
/* [object][oleautomation][dual][hidden][helpcontext][helpstring][version][uuid] */ 


EXTERN_C const IID IID__Mouse;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("C065C0A0-1E67-49F3-BBDD-3E0A9FDD5EDA")
    _Mouse : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Position( 
            /* [retval][out] */ ScreenPoint **point) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Position( 
            /* [in] */ ScreenPoint *point) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BoundingRectangle( 
            /* [retval][out] */ ScreenRectangle **Boundary) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Pointer( 
            /* [retval][out] */ _MousePointer **Pointer) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Window( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LeftButtonState( 
            /* [retval][out] */ VARIANT_BOOL *state) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MiddleButtonState( 
            /* [retval][out] */ VARIANT_BOOL *state) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_RightButtonState( 
            /* [retval][out] */ VARIANT_BOOL *state) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ButtonDown( 
            /* [in] */ MouseButton button) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ButtonUp( 
            /* [in] */ MouseButton button) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Click( 
            /* [in] */ MouseButton button,
            /* [in] */ short Count) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ToFocus( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE DisplaySearchDialog( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ContinueSearch( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Sync( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ButtonDownEx( 
            /* [in] */ VARIANT button,
            /* [in] */ VARIANT KeyboardModifiers) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ClickEx( 
            /* [in] */ VARIANT button,
            /* [in] */ VARIANT ClickCount,
            /* [in] */ VARIANT KeyboardModifiers) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _MouseVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _Mouse * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _Mouse * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _Mouse * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _Mouse * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _Mouse * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _Mouse * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _Mouse * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _Mouse * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _Mouse * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            _Mouse * This,
            /* [retval][out] */ ScreenPoint **point);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            _Mouse * This,
            /* [in] */ ScreenPoint *point);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BoundingRectangle )( 
            _Mouse * This,
            /* [retval][out] */ ScreenRectangle **Boundary);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Pointer )( 
            _Mouse * This,
            /* [retval][out] */ _MousePointer **Pointer);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            _Mouse * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LeftButtonState )( 
            _Mouse * This,
            /* [retval][out] */ VARIANT_BOOL *state);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MiddleButtonState )( 
            _Mouse * This,
            /* [retval][out] */ VARIANT_BOOL *state);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_RightButtonState )( 
            _Mouse * This,
            /* [retval][out] */ VARIANT_BOOL *state);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ButtonDown )( 
            _Mouse * This,
            /* [in] */ MouseButton button);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ButtonUp )( 
            _Mouse * This,
            /* [in] */ MouseButton button);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Click )( 
            _Mouse * This,
            /* [in] */ MouseButton button,
            /* [in] */ short Count);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ToFocus )( 
            _Mouse * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *DisplaySearchDialog )( 
            _Mouse * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ContinueSearch )( 
            _Mouse * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Sync )( 
            _Mouse * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ButtonDownEx )( 
            _Mouse * This,
            /* [in] */ VARIANT button,
            /* [in] */ VARIANT KeyboardModifiers);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ClickEx )( 
            _Mouse * This,
            /* [in] */ VARIANT button,
            /* [in] */ VARIANT ClickCount,
            /* [in] */ VARIANT KeyboardModifiers);
        
        END_INTERFACE
    } _MouseVtbl;

    interface _Mouse
    {
        CONST_VTBL struct _MouseVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _Mouse_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _Mouse_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _Mouse_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _Mouse_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _Mouse_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _Mouse_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _Mouse_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _Mouse_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _Mouse_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _Mouse_get_Position(This,point)	\
    ( (This)->lpVtbl -> get_Position(This,point) ) 

#define _Mouse_put_Position(This,point)	\
    ( (This)->lpVtbl -> put_Position(This,point) ) 

#define _Mouse_get_BoundingRectangle(This,Boundary)	\
    ( (This)->lpVtbl -> get_BoundingRectangle(This,Boundary) ) 

#define _Mouse_get_Pointer(This,Pointer)	\
    ( (This)->lpVtbl -> get_Pointer(This,Pointer) ) 

#define _Mouse_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define _Mouse_get_LeftButtonState(This,state)	\
    ( (This)->lpVtbl -> get_LeftButtonState(This,state) ) 

#define _Mouse_get_MiddleButtonState(This,state)	\
    ( (This)->lpVtbl -> get_MiddleButtonState(This,state) ) 

#define _Mouse_get_RightButtonState(This,state)	\
    ( (This)->lpVtbl -> get_RightButtonState(This,state) ) 

#define _Mouse_ButtonDown(This,button)	\
    ( (This)->lpVtbl -> ButtonDown(This,button) ) 

#define _Mouse_ButtonUp(This,button)	\
    ( (This)->lpVtbl -> ButtonUp(This,button) ) 

#define _Mouse_Click(This,button,Count)	\
    ( (This)->lpVtbl -> Click(This,button,Count) ) 

#define _Mouse_ToFocus(This)	\
    ( (This)->lpVtbl -> ToFocus(This) ) 

#define _Mouse_DisplaySearchDialog(This)	\
    ( (This)->lpVtbl -> DisplaySearchDialog(This) ) 

#define _Mouse_ContinueSearch(This)	\
    ( (This)->lpVtbl -> ContinueSearch(This) ) 

#define _Mouse_Sync(This)	\
    ( (This)->lpVtbl -> Sync(This) ) 

#define _Mouse_ButtonDownEx(This,button,KeyboardModifiers)	\
    ( (This)->lpVtbl -> ButtonDownEx(This,button,KeyboardModifiers) ) 

#define _Mouse_ClickEx(This,button,ClickCount,KeyboardModifiers)	\
    ( (This)->lpVtbl -> ClickEx(This,button,ClickCount,KeyboardModifiers) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___Mouse_INTERFACE_DEFINED__ */


#ifndef ___MousePointer_INTERFACE_DEFINED__
#define ___MousePointer_INTERFACE_DEFINED__

/* interface _MousePointer */
/* [object][oleautomation][dual][hidden][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__MousePointer;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("B3E482A3-7EE0-4B15-AA04-470D0C4D2D13")
    _MousePointer : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_id( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ModuleName( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ExecutableName( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CursorName( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_PointerSilenced( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UseAsSystemPointer( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE DescribePointer( 
            /* [in] */ BSTR Name,
            /* [in] */ VARIANT_BOOL Silence,
            /* [in] */ VARIANT_BOOL IsSystemPointer) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _MousePointerVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _MousePointer * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _MousePointer * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _MousePointer * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _MousePointer * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _MousePointer * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _MousePointer * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _MousePointer * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _MousePointer * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _MousePointer * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            _MousePointer * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ModuleName )( 
            _MousePointer * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ExecutableName )( 
            _MousePointer * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CursorName )( 
            _MousePointer * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_PointerSilenced )( 
            _MousePointer * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UseAsSystemPointer )( 
            _MousePointer * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *DescribePointer )( 
            _MousePointer * This,
            /* [in] */ BSTR Name,
            /* [in] */ VARIANT_BOOL Silence,
            /* [in] */ VARIANT_BOOL IsSystemPointer);
        
        END_INTERFACE
    } _MousePointerVtbl;

    interface _MousePointer
    {
        CONST_VTBL struct _MousePointerVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _MousePointer_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _MousePointer_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _MousePointer_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _MousePointer_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _MousePointer_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _MousePointer_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _MousePointer_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _MousePointer_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _MousePointer_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _MousePointer_get_id(This,prop)	\
    ( (This)->lpVtbl -> get_id(This,prop) ) 

#define _MousePointer_get_ModuleName(This,prop)	\
    ( (This)->lpVtbl -> get_ModuleName(This,prop) ) 

#define _MousePointer_get_ExecutableName(This,prop)	\
    ( (This)->lpVtbl -> get_ExecutableName(This,prop) ) 

#define _MousePointer_get_CursorName(This,prop)	\
    ( (This)->lpVtbl -> get_CursorName(This,prop) ) 

#define _MousePointer_get_PointerSilenced(This,prop)	\
    ( (This)->lpVtbl -> get_PointerSilenced(This,prop) ) 

#define _MousePointer_get_UseAsSystemPointer(This,prop)	\
    ( (This)->lpVtbl -> get_UseAsSystemPointer(This,prop) ) 

#define _MousePointer_DescribePointer(This,Name,Silence,IsSystemPointer)	\
    ( (This)->lpVtbl -> DescribePointer(This,Name,Silence,IsSystemPointer) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___MousePointer_INTERFACE_DEFINED__ */


#ifndef __Cursor_INTERFACE_DEFINED__
#define __Cursor_INTERFACE_DEFINED__

/* interface Cursor */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Cursor;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("E45E7F76-B7BE-4C73-919D-C3894064D228")
    Cursor : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Position( 
            /* [retval][out] */ WindowPoint **prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Visible( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Window( 
            /* [retval][out] */ _Window **prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct CursorVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Cursor * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Cursor * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Cursor * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Cursor * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Cursor * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Cursor * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Cursor * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Cursor * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Cursor * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            Cursor * This,
            /* [retval][out] */ WindowPoint **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            Cursor * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            Cursor * This,
            /* [retval][out] */ _Window **prop);
        
        END_INTERFACE
    } CursorVtbl;

    interface Cursor
    {
        CONST_VTBL struct CursorVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Cursor_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Cursor_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Cursor_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Cursor_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Cursor_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Cursor_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Cursor_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Cursor_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Cursor_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Cursor_get_Position(This,prop)	\
    ( (This)->lpVtbl -> get_Position(This,prop) ) 

#define Cursor_get_Visible(This,prop)	\
    ( (This)->lpVtbl -> get_Visible(This,prop) ) 

#define Cursor_get_Window(This,prop)	\
    ( (This)->lpVtbl -> get_Window(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Cursor_INTERFACE_DEFINED__ */


#ifndef ___Keyboard_INTERFACE_DEFINED__
#define ___Keyboard_INTERFACE_DEFINED__

/* interface _Keyboard */
/* [object][oleautomation][dual][hidden][helpcontext][helpstring][version][uuid] */ 


EXTERN_C const IID IID__Keyboard;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("993F763F-0743-4D46-BC63-9EF2705091A3")
    _Keyboard : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Keys( 
            /* [retval][out] */ Keys **Keys) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FilterActiveWindow( 
            /* [retval][out] */ _Window **var) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FilterActiveWindow( 
            /* [in] */ _Window *var) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FilterFocusedWindow( 
            /* [retval][out] */ _Window **var) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FilterFocusedWindow( 
            /* [in] */ _Window *var) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FilterProcess( 
            /* [retval][out] */ Process **var) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FilterProcess( 
            /* [in] */ Process *var) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_KeyDescriberActive( 
            /* [retval][out] */ VARIANT_BOOL *var) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Down( 
            /* [in] */ long index,
            /* [retval][out] */ VARIANT_BOOL *Down) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Toggled( 
            /* [in] */ long index,
            /* [retval][out] */ VARIANT_BOOL *Toggled) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE InsertKeyDown( 
            /* [in] */ long index) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE InsertKeyUp( 
            /* [in] */ long index) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE InsertKey( 
            /* [in] */ long index,
            /* [in] */ KeyModifiers mods) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Key( 
            /* [in] */ BSTR keyName,
            /* [retval][out] */ _Key **Key) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE RegisterHotkey( 
            /* [in] */ VARIANT Key,
            /* [in] */ VARIANT function,
            /* [optional][in] */ VARIANT filterBy,
            /* [optional][in] */ VARIANT Parameter,
            /* [optional][in] */ VARIANT FirstPress,
            /* [retval][out] */ RegisteredKey **Handle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE RegisterCursorKey( 
            /* [in] */ VARIANT Key,
            /* [in] */ VARIANT function,
            /* [optional][in] */ VARIANT filterBy,
            /* [optional][in] */ VARIANT Parameter,
            /* [retval][out] */ RegisteredKey **Handle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Capture( 
            /* [retval][out] */ _Key **Key) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE WaitForAllKeysUp( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE InsertString( 
            /* [in] */ BSTR str) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE RegisteredHotkeys( 
            /* [retval][out] */ RegisteredHotkeys **RegisteredHotkeys) = 0;
        
        virtual /* [helpstring][id] */ HRESULT STDMETHODCALLTYPE RegisterHotkeyEx( 
            /* [in] */ VARIANT Key,
            /* [in] */ VARIANT function,
            /* [optional][in] */ VARIANT filterBy,
            /* [optional][in] */ VARIANT Parameter,
            /* [optional][in] */ VARIANT FirstPress,
            /* [optional][in] */ VARIANT Priority,
            /* [retval][out] */ RegisteredKey **Handle) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _KeyboardVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _Keyboard * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _Keyboard * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _Keyboard * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _Keyboard * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _Keyboard * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _Keyboard * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _Keyboard * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _Keyboard * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _Keyboard * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Keys )( 
            _Keyboard * This,
            /* [retval][out] */ Keys **Keys);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FilterActiveWindow )( 
            _Keyboard * This,
            /* [retval][out] */ _Window **var);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FilterActiveWindow )( 
            _Keyboard * This,
            /* [in] */ _Window *var);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FilterFocusedWindow )( 
            _Keyboard * This,
            /* [retval][out] */ _Window **var);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FilterFocusedWindow )( 
            _Keyboard * This,
            /* [in] */ _Window *var);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FilterProcess )( 
            _Keyboard * This,
            /* [retval][out] */ Process **var);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FilterProcess )( 
            _Keyboard * This,
            /* [in] */ Process *var);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_KeyDescriberActive )( 
            _Keyboard * This,
            /* [retval][out] */ VARIANT_BOOL *var);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Down )( 
            _Keyboard * This,
            /* [in] */ long index,
            /* [retval][out] */ VARIANT_BOOL *Down);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Toggled )( 
            _Keyboard * This,
            /* [in] */ long index,
            /* [retval][out] */ VARIANT_BOOL *Toggled);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *InsertKeyDown )( 
            _Keyboard * This,
            /* [in] */ long index);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *InsertKeyUp )( 
            _Keyboard * This,
            /* [in] */ long index);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *InsertKey )( 
            _Keyboard * This,
            /* [in] */ long index,
            /* [in] */ KeyModifiers mods);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Key )( 
            _Keyboard * This,
            /* [in] */ BSTR keyName,
            /* [retval][out] */ _Key **Key);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *RegisterHotkey )( 
            _Keyboard * This,
            /* [in] */ VARIANT Key,
            /* [in] */ VARIANT function,
            /* [optional][in] */ VARIANT filterBy,
            /* [optional][in] */ VARIANT Parameter,
            /* [optional][in] */ VARIANT FirstPress,
            /* [retval][out] */ RegisteredKey **Handle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *RegisterCursorKey )( 
            _Keyboard * This,
            /* [in] */ VARIANT Key,
            /* [in] */ VARIANT function,
            /* [optional][in] */ VARIANT filterBy,
            /* [optional][in] */ VARIANT Parameter,
            /* [retval][out] */ RegisteredKey **Handle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Capture )( 
            _Keyboard * This,
            /* [retval][out] */ _Key **Key);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *WaitForAllKeysUp )( 
            _Keyboard * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *InsertString )( 
            _Keyboard * This,
            /* [in] */ BSTR str);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *RegisteredHotkeys )( 
            _Keyboard * This,
            /* [retval][out] */ RegisteredHotkeys **RegisteredHotkeys);
        
        /* [helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *RegisterHotkeyEx )( 
            _Keyboard * This,
            /* [in] */ VARIANT Key,
            /* [in] */ VARIANT function,
            /* [optional][in] */ VARIANT filterBy,
            /* [optional][in] */ VARIANT Parameter,
            /* [optional][in] */ VARIANT FirstPress,
            /* [optional][in] */ VARIANT Priority,
            /* [retval][out] */ RegisteredKey **Handle);
        
        END_INTERFACE
    } _KeyboardVtbl;

    interface _Keyboard
    {
        CONST_VTBL struct _KeyboardVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _Keyboard_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _Keyboard_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _Keyboard_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _Keyboard_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _Keyboard_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _Keyboard_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _Keyboard_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _Keyboard_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _Keyboard_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _Keyboard_get_Keys(This,Keys)	\
    ( (This)->lpVtbl -> get_Keys(This,Keys) ) 

#define _Keyboard_get_FilterActiveWindow(This,var)	\
    ( (This)->lpVtbl -> get_FilterActiveWindow(This,var) ) 

#define _Keyboard_put_FilterActiveWindow(This,var)	\
    ( (This)->lpVtbl -> put_FilterActiveWindow(This,var) ) 

#define _Keyboard_get_FilterFocusedWindow(This,var)	\
    ( (This)->lpVtbl -> get_FilterFocusedWindow(This,var) ) 

#define _Keyboard_put_FilterFocusedWindow(This,var)	\
    ( (This)->lpVtbl -> put_FilterFocusedWindow(This,var) ) 

#define _Keyboard_get_FilterProcess(This,var)	\
    ( (This)->lpVtbl -> get_FilterProcess(This,var) ) 

#define _Keyboard_put_FilterProcess(This,var)	\
    ( (This)->lpVtbl -> put_FilterProcess(This,var) ) 

#define _Keyboard_get_KeyDescriberActive(This,var)	\
    ( (This)->lpVtbl -> get_KeyDescriberActive(This,var) ) 

#define _Keyboard_Down(This,index,Down)	\
    ( (This)->lpVtbl -> Down(This,index,Down) ) 

#define _Keyboard_Toggled(This,index,Toggled)	\
    ( (This)->lpVtbl -> Toggled(This,index,Toggled) ) 

#define _Keyboard_InsertKeyDown(This,index)	\
    ( (This)->lpVtbl -> InsertKeyDown(This,index) ) 

#define _Keyboard_InsertKeyUp(This,index)	\
    ( (This)->lpVtbl -> InsertKeyUp(This,index) ) 

#define _Keyboard_InsertKey(This,index,mods)	\
    ( (This)->lpVtbl -> InsertKey(This,index,mods) ) 

#define _Keyboard_Key(This,keyName,Key)	\
    ( (This)->lpVtbl -> Key(This,keyName,Key) ) 

#define _Keyboard_RegisterHotkey(This,Key,function,filterBy,Parameter,FirstPress,Handle)	\
    ( (This)->lpVtbl -> RegisterHotkey(This,Key,function,filterBy,Parameter,FirstPress,Handle) ) 

#define _Keyboard_RegisterCursorKey(This,Key,function,filterBy,Parameter,Handle)	\
    ( (This)->lpVtbl -> RegisterCursorKey(This,Key,function,filterBy,Parameter,Handle) ) 

#define _Keyboard_Capture(This,Key)	\
    ( (This)->lpVtbl -> Capture(This,Key) ) 

#define _Keyboard_WaitForAllKeysUp(This)	\
    ( (This)->lpVtbl -> WaitForAllKeysUp(This) ) 

#define _Keyboard_InsertString(This,str)	\
    ( (This)->lpVtbl -> InsertString(This,str) ) 

#define _Keyboard_RegisteredHotkeys(This,RegisteredHotkeys)	\
    ( (This)->lpVtbl -> RegisteredHotkeys(This,RegisteredHotkeys) ) 

#define _Keyboard_RegisterHotkeyEx(This,Key,function,filterBy,Parameter,FirstPress,Priority,Handle)	\
    ( (This)->lpVtbl -> RegisterHotkeyEx(This,Key,function,filterBy,Parameter,FirstPress,Priority,Handle) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___Keyboard_INTERFACE_DEFINED__ */


#ifndef __Keys_INTERFACE_DEFINED__
#define __Keys_INTERFACE_DEFINED__

/* interface Keys */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Keys;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("EF11A9E8-F13D-41DC-AA78-353B4A6C96CC")
    Keys : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ _Key **Key) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct KeysVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Keys * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Keys * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Keys * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Keys * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Keys * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Keys * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Keys * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Keys * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Keys * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            Keys * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            Keys * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            Keys * This,
            /* [in] */ long index,
            /* [retval][out] */ _Key **Key);
        
        END_INTERFACE
    } KeysVtbl;

    interface Keys
    {
        CONST_VTBL struct KeysVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Keys_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Keys_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Keys_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Keys_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Keys_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Keys_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Keys_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Keys_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Keys_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Keys_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define Keys_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define Keys_Item(This,index,Key)	\
    ( (This)->lpVtbl -> Item(This,index,Key) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Keys_INTERFACE_DEFINED__ */


#ifndef ___Key_INTERFACE_DEFINED__
#define ___Key_INTERFACE_DEFINED__

/* interface _Key */
/* [object][oleautomation][dual][hidden][helpcontext][helpstring][version][uuid] */ 


EXTERN_C const IID IID__Key;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("EECEC524-6269-43CA-95C0-2B93024B5854")
    _Key : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Name( 
            /* [retval][out] */ BSTR *Name) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_index( 
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Down( 
            /* [retval][out] */ VARIANT_BOOL *Down) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Toggled( 
            /* [retval][out] */ VARIANT_BOOL *Toggled) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FilterActiveWindow( 
            /* [retval][out] */ _Window **var) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FilterActiveWindow( 
            /* [in] */ _Window *var) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FilterFocusedWindow( 
            /* [retval][out] */ _Window **var) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FilterFocusedWindow( 
            /* [in] */ _Window *var) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FilterProcess( 
            /* [retval][out] */ Process **var) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FilterProcess( 
            /* [in] */ Process *var) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FilterShift( 
            /* [retval][out] */ KeyFilterState *kfs) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FilterShift( 
            /* [in] */ KeyFilterState kfs) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FilterControl( 
            /* [retval][out] */ KeyFilterState *kfs) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FilterControl( 
            /* [in] */ KeyFilterState kfs) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FilterAlt( 
            /* [retval][out] */ KeyFilterState *kfs) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FilterAlt( 
            /* [in] */ KeyFilterState kfs) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FilterWindows( 
            /* [retval][out] */ KeyFilterState *kfs) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FilterWindows( 
            /* [in] */ KeyFilterState kfs) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FilterApplication( 
            /* [retval][out] */ KeyFilterState *kfs) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FilterApplication( 
            /* [in] */ KeyFilterState kfs) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FilterInsert( 
            /* [retval][out] */ KeyFilterState *kfs) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FilterInsert( 
            /* [in] */ KeyFilterState kfs) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FilterNumlock( 
            /* [retval][out] */ KeyFilterState *kfs) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FilterNumlock( 
            /* [in] */ KeyFilterState kfs) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FilterNumpad( 
            /* [retval][out] */ KeyFilterState *kfs) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FilterNumpad( 
            /* [in] */ KeyFilterState kfs) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE RequireModifiers( 
            /* [in] */ KeyModifiers Modifiers) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UntranslatedName( 
            /* [retval][out] */ BSTR *Name) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_RequiredModifiers( 
            /* [retval][out] */ KeyModifiers *Modifiers) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Insert( void) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _KeyVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _Key * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _Key * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _Key * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _Key * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _Key * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _Key * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _Key * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Name )( 
            _Key * This,
            /* [retval][out] */ BSTR *Name);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _Key * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _Key * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_index )( 
            _Key * This,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Down )( 
            _Key * This,
            /* [retval][out] */ VARIANT_BOOL *Down);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Toggled )( 
            _Key * This,
            /* [retval][out] */ VARIANT_BOOL *Toggled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FilterActiveWindow )( 
            _Key * This,
            /* [retval][out] */ _Window **var);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FilterActiveWindow )( 
            _Key * This,
            /* [in] */ _Window *var);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FilterFocusedWindow )( 
            _Key * This,
            /* [retval][out] */ _Window **var);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FilterFocusedWindow )( 
            _Key * This,
            /* [in] */ _Window *var);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FilterProcess )( 
            _Key * This,
            /* [retval][out] */ Process **var);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FilterProcess )( 
            _Key * This,
            /* [in] */ Process *var);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FilterShift )( 
            _Key * This,
            /* [retval][out] */ KeyFilterState *kfs);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FilterShift )( 
            _Key * This,
            /* [in] */ KeyFilterState kfs);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FilterControl )( 
            _Key * This,
            /* [retval][out] */ KeyFilterState *kfs);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FilterControl )( 
            _Key * This,
            /* [in] */ KeyFilterState kfs);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FilterAlt )( 
            _Key * This,
            /* [retval][out] */ KeyFilterState *kfs);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FilterAlt )( 
            _Key * This,
            /* [in] */ KeyFilterState kfs);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FilterWindows )( 
            _Key * This,
            /* [retval][out] */ KeyFilterState *kfs);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FilterWindows )( 
            _Key * This,
            /* [in] */ KeyFilterState kfs);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FilterApplication )( 
            _Key * This,
            /* [retval][out] */ KeyFilterState *kfs);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FilterApplication )( 
            _Key * This,
            /* [in] */ KeyFilterState kfs);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FilterInsert )( 
            _Key * This,
            /* [retval][out] */ KeyFilterState *kfs);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FilterInsert )( 
            _Key * This,
            /* [in] */ KeyFilterState kfs);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FilterNumlock )( 
            _Key * This,
            /* [retval][out] */ KeyFilterState *kfs);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FilterNumlock )( 
            _Key * This,
            /* [in] */ KeyFilterState kfs);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FilterNumpad )( 
            _Key * This,
            /* [retval][out] */ KeyFilterState *kfs);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FilterNumpad )( 
            _Key * This,
            /* [in] */ KeyFilterState kfs);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *RequireModifiers )( 
            _Key * This,
            /* [in] */ KeyModifiers Modifiers);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UntranslatedName )( 
            _Key * This,
            /* [retval][out] */ BSTR *Name);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_RequiredModifiers )( 
            _Key * This,
            /* [retval][out] */ KeyModifiers *Modifiers);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Insert )( 
            _Key * This);
        
        END_INTERFACE
    } _KeyVtbl;

    interface _Key
    {
        CONST_VTBL struct _KeyVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _Key_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _Key_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _Key_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _Key_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _Key_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _Key_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _Key_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _Key_get_Name(This,Name)	\
    ( (This)->lpVtbl -> get_Name(This,Name) ) 

#define _Key_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _Key_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _Key_get_index(This,index)	\
    ( (This)->lpVtbl -> get_index(This,index) ) 

#define _Key_get_Down(This,Down)	\
    ( (This)->lpVtbl -> get_Down(This,Down) ) 

#define _Key_get_Toggled(This,Toggled)	\
    ( (This)->lpVtbl -> get_Toggled(This,Toggled) ) 

#define _Key_get_FilterActiveWindow(This,var)	\
    ( (This)->lpVtbl -> get_FilterActiveWindow(This,var) ) 

#define _Key_put_FilterActiveWindow(This,var)	\
    ( (This)->lpVtbl -> put_FilterActiveWindow(This,var) ) 

#define _Key_get_FilterFocusedWindow(This,var)	\
    ( (This)->lpVtbl -> get_FilterFocusedWindow(This,var) ) 

#define _Key_put_FilterFocusedWindow(This,var)	\
    ( (This)->lpVtbl -> put_FilterFocusedWindow(This,var) ) 

#define _Key_get_FilterProcess(This,var)	\
    ( (This)->lpVtbl -> get_FilterProcess(This,var) ) 

#define _Key_put_FilterProcess(This,var)	\
    ( (This)->lpVtbl -> put_FilterProcess(This,var) ) 

#define _Key_get_FilterShift(This,kfs)	\
    ( (This)->lpVtbl -> get_FilterShift(This,kfs) ) 

#define _Key_put_FilterShift(This,kfs)	\
    ( (This)->lpVtbl -> put_FilterShift(This,kfs) ) 

#define _Key_get_FilterControl(This,kfs)	\
    ( (This)->lpVtbl -> get_FilterControl(This,kfs) ) 

#define _Key_put_FilterControl(This,kfs)	\
    ( (This)->lpVtbl -> put_FilterControl(This,kfs) ) 

#define _Key_get_FilterAlt(This,kfs)	\
    ( (This)->lpVtbl -> get_FilterAlt(This,kfs) ) 

#define _Key_put_FilterAlt(This,kfs)	\
    ( (This)->lpVtbl -> put_FilterAlt(This,kfs) ) 

#define _Key_get_FilterWindows(This,kfs)	\
    ( (This)->lpVtbl -> get_FilterWindows(This,kfs) ) 

#define _Key_put_FilterWindows(This,kfs)	\
    ( (This)->lpVtbl -> put_FilterWindows(This,kfs) ) 

#define _Key_get_FilterApplication(This,kfs)	\
    ( (This)->lpVtbl -> get_FilterApplication(This,kfs) ) 

#define _Key_put_FilterApplication(This,kfs)	\
    ( (This)->lpVtbl -> put_FilterApplication(This,kfs) ) 

#define _Key_get_FilterInsert(This,kfs)	\
    ( (This)->lpVtbl -> get_FilterInsert(This,kfs) ) 

#define _Key_put_FilterInsert(This,kfs)	\
    ( (This)->lpVtbl -> put_FilterInsert(This,kfs) ) 

#define _Key_get_FilterNumlock(This,kfs)	\
    ( (This)->lpVtbl -> get_FilterNumlock(This,kfs) ) 

#define _Key_put_FilterNumlock(This,kfs)	\
    ( (This)->lpVtbl -> put_FilterNumlock(This,kfs) ) 

#define _Key_get_FilterNumpad(This,kfs)	\
    ( (This)->lpVtbl -> get_FilterNumpad(This,kfs) ) 

#define _Key_put_FilterNumpad(This,kfs)	\
    ( (This)->lpVtbl -> put_FilterNumpad(This,kfs) ) 

#define _Key_RequireModifiers(This,Modifiers)	\
    ( (This)->lpVtbl -> RequireModifiers(This,Modifiers) ) 

#define _Key_get_UntranslatedName(This,Name)	\
    ( (This)->lpVtbl -> get_UntranslatedName(This,Name) ) 

#define _Key_get_RequiredModifiers(This,Modifiers)	\
    ( (This)->lpVtbl -> get_RequiredModifiers(This,Modifiers) ) 

#define _Key_Insert(This)	\
    ( (This)->lpVtbl -> Insert(This) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___Key_INTERFACE_DEFINED__ */


#ifndef __RegisteredKey_INTERFACE_DEFINED__
#define __RegisteredKey_INTERFACE_DEFINED__

/* interface RegisteredKey */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_RegisteredKey;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("B43442C2-FB5D-4204-A2E2-1BFA0D10D04A")
    RegisteredKey : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Unregister( void) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct RegisteredKeyVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            RegisteredKey * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            RegisteredKey * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            RegisteredKey * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            RegisteredKey * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            RegisteredKey * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            RegisteredKey * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            RegisteredKey * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            RegisteredKey * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            RegisteredKey * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Unregister )( 
            RegisteredKey * This);
        
        END_INTERFACE
    } RegisteredKeyVtbl;

    interface RegisteredKey
    {
        CONST_VTBL struct RegisteredKeyVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define RegisteredKey_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define RegisteredKey_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define RegisteredKey_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define RegisteredKey_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define RegisteredKey_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define RegisteredKey_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define RegisteredKey_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define RegisteredKey_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define RegisteredKey_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define RegisteredKey_Unregister(This)	\
    ( (This)->lpVtbl -> Unregister(This) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __RegisteredKey_INTERFACE_DEFINED__ */


#ifndef __RegisteredHotkeys_INTERFACE_DEFINED__
#define __RegisteredHotkeys_INTERFACE_DEFINED__

/* interface RegisteredHotkeys */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_RegisteredHotkeys;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("07C2E45E-8C6D-4980-AD88-A1A930C9BF5C")
    RegisteredHotkeys : public IDispatch
    {
    public:
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ RegisteredHotkey **RegisteredHotkey) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct RegisteredHotkeysVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            RegisteredHotkeys * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            RegisteredHotkeys * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            RegisteredHotkeys * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            RegisteredHotkeys * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            RegisteredHotkeys * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            RegisteredHotkeys * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            RegisteredHotkeys * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            RegisteredHotkeys * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            RegisteredHotkeys * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            RegisteredHotkeys * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            RegisteredHotkeys * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            RegisteredHotkeys * This,
            /* [in] */ long index,
            /* [retval][out] */ RegisteredHotkey **RegisteredHotkey);
        
        END_INTERFACE
    } RegisteredHotkeysVtbl;

    interface RegisteredHotkeys
    {
        CONST_VTBL struct RegisteredHotkeysVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define RegisteredHotkeys_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define RegisteredHotkeys_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define RegisteredHotkeys_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define RegisteredHotkeys_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define RegisteredHotkeys_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define RegisteredHotkeys_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define RegisteredHotkeys_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define RegisteredHotkeys_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define RegisteredHotkeys_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define RegisteredHotkeys_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define RegisteredHotkeys_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define RegisteredHotkeys_Item(This,index,RegisteredHotkey)	\
    ( (This)->lpVtbl -> Item(This,index,RegisteredHotkey) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __RegisteredHotkeys_INTERFACE_DEFINED__ */


#ifndef __RegisteredHotkey_INTERFACE_DEFINED__
#define __RegisteredHotkey_INTERFACE_DEFINED__

/* interface RegisteredHotkey */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_RegisteredHotkey;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("23587EDB-E616-486A-8510-D1FD46DA24B2")
    RegisteredHotkey : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Key( 
            /* [retval][out] */ _Key **Key) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_filterBy( 
            /* [retval][out] */ VARIANT *FilterByObject) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Script( 
            /* [retval][out] */ LoadedScript **LoadedScript) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Priority( 
            /* [retval][out] */ HotkeyPriority *Priority) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct RegisteredHotkeyVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            RegisteredHotkey * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            RegisteredHotkey * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            RegisteredHotkey * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            RegisteredHotkey * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            RegisteredHotkey * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            RegisteredHotkey * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            RegisteredHotkey * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            RegisteredHotkey * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            RegisteredHotkey * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Key )( 
            RegisteredHotkey * This,
            /* [retval][out] */ _Key **Key);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_filterBy )( 
            RegisteredHotkey * This,
            /* [retval][out] */ VARIANT *FilterByObject);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Script )( 
            RegisteredHotkey * This,
            /* [retval][out] */ LoadedScript **LoadedScript);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Priority )( 
            RegisteredHotkey * This,
            /* [retval][out] */ HotkeyPriority *Priority);
        
        END_INTERFACE
    } RegisteredHotkeyVtbl;

    interface RegisteredHotkey
    {
        CONST_VTBL struct RegisteredHotkeyVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define RegisteredHotkey_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define RegisteredHotkey_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define RegisteredHotkey_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define RegisteredHotkey_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define RegisteredHotkey_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define RegisteredHotkey_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define RegisteredHotkey_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define RegisteredHotkey_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define RegisteredHotkey_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define RegisteredHotkey_get_Key(This,Key)	\
    ( (This)->lpVtbl -> get_Key(This,Key) ) 

#define RegisteredHotkey_get_filterBy(This,FilterByObject)	\
    ( (This)->lpVtbl -> get_filterBy(This,FilterByObject) ) 

#define RegisteredHotkey_get_Script(This,LoadedScript)	\
    ( (This)->lpVtbl -> get_Script(This,LoadedScript) ) 

#define RegisteredHotkey_get_Priority(This,Priority)	\
    ( (This)->lpVtbl -> get_Priority(This,Priority) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __RegisteredHotkey_INTERFACE_DEFINED__ */


#ifndef __LoadedScript_INTERFACE_DEFINED__
#define __LoadedScript_INTERFACE_DEFINED__

/* interface LoadedScript */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_LoadedScript;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("61358C41-8FC6-41AF-95E9-1627FDCE2779")
    LoadedScript : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Filename( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Version( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Name( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Description( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Status( 
            /* [retval][out] */ LoadedScriptStatus *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Association( 
            /* [retval][out] */ BSTR *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct LoadedScriptVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            LoadedScript * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            LoadedScript * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            LoadedScript * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            LoadedScript * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            LoadedScript * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            LoadedScript * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            LoadedScript * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            LoadedScript * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            LoadedScript * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Filename )( 
            LoadedScript * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Version )( 
            LoadedScript * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Name )( 
            LoadedScript * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Description )( 
            LoadedScript * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Status )( 
            LoadedScript * This,
            /* [retval][out] */ LoadedScriptStatus *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Association )( 
            LoadedScript * This,
            /* [retval][out] */ BSTR *prop);
        
        END_INTERFACE
    } LoadedScriptVtbl;

    interface LoadedScript
    {
        CONST_VTBL struct LoadedScriptVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define LoadedScript_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define LoadedScript_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define LoadedScript_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define LoadedScript_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define LoadedScript_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define LoadedScript_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define LoadedScript_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define LoadedScript_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define LoadedScript_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define LoadedScript_get_Filename(This,prop)	\
    ( (This)->lpVtbl -> get_Filename(This,prop) ) 

#define LoadedScript_get_Version(This,prop)	\
    ( (This)->lpVtbl -> get_Version(This,prop) ) 

#define LoadedScript_get_Name(This,prop)	\
    ( (This)->lpVtbl -> get_Name(This,prop) ) 

#define LoadedScript_get_Description(This,prop)	\
    ( (This)->lpVtbl -> get_Description(This,prop) ) 

#define LoadedScript_get_Status(This,prop)	\
    ( (This)->lpVtbl -> get_Status(This,prop) ) 

#define LoadedScript_get_Association(This,prop)	\
    ( (This)->lpVtbl -> get_Association(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __LoadedScript_INTERFACE_DEFINED__ */


#ifndef __CharacterDictionaries_INTERFACE_DEFINED__
#define __CharacterDictionaries_INTERFACE_DEFINED__

/* interface CharacterDictionaries */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_CharacterDictionaries;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("F571E9C7-EC2D-452F-82F9-3E5FB3CBD855")
    CharacterDictionaries : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ CharacterDictionary **dictionary) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SaveAll( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE LookUp( 
            /* [in] */ VARIANT Character,
            /* [optional][in] */ VARIANT FontName,
            /* [retval][out] */ CharacterEntry **Entry) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct CharacterDictionariesVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            CharacterDictionaries * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            CharacterDictionaries * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            CharacterDictionaries * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            CharacterDictionaries * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            CharacterDictionaries * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            CharacterDictionaries * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            CharacterDictionaries * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            CharacterDictionaries * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            CharacterDictionaries * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            CharacterDictionaries * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            CharacterDictionaries * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            CharacterDictionaries * This,
            /* [in] */ long index,
            /* [retval][out] */ CharacterDictionary **dictionary);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SaveAll )( 
            CharacterDictionaries * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *LookUp )( 
            CharacterDictionaries * This,
            /* [in] */ VARIANT Character,
            /* [optional][in] */ VARIANT FontName,
            /* [retval][out] */ CharacterEntry **Entry);
        
        END_INTERFACE
    } CharacterDictionariesVtbl;

    interface CharacterDictionaries
    {
        CONST_VTBL struct CharacterDictionariesVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define CharacterDictionaries_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define CharacterDictionaries_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define CharacterDictionaries_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define CharacterDictionaries_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define CharacterDictionaries_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define CharacterDictionaries_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define CharacterDictionaries_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define CharacterDictionaries_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define CharacterDictionaries_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define CharacterDictionaries_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define CharacterDictionaries_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define CharacterDictionaries_Item(This,index,dictionary)	\
    ( (This)->lpVtbl -> Item(This,index,dictionary) ) 

#define CharacterDictionaries_SaveAll(This)	\
    ( (This)->lpVtbl -> SaveAll(This) ) 

#define CharacterDictionaries_LookUp(This,Character,FontName,Entry)	\
    ( (This)->lpVtbl -> LookUp(This,Character,FontName,Entry) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __CharacterDictionaries_INTERFACE_DEFINED__ */


#ifndef __ColorDictionaries_INTERFACE_DEFINED__
#define __ColorDictionaries_INTERFACE_DEFINED__

/* interface ColorDictionaries */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ColorDictionaries;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("04819B3B-C4D7-4D62-884F-09DF0C436CED")
    ColorDictionaries : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ ColorDictionary **dictionary) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SaveAll( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE LookUp( 
            /* [in] */ Color *Color,
            /* [optional][in] */ VARIANT ClosestMatch,
            /* [retval][out] */ ColorEntry **Entry) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ColorDictionariesVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ColorDictionaries * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ColorDictionaries * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ColorDictionaries * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ColorDictionaries * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ColorDictionaries * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ColorDictionaries * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ColorDictionaries * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ColorDictionaries * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ColorDictionaries * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            ColorDictionaries * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            ColorDictionaries * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            ColorDictionaries * This,
            /* [in] */ long index,
            /* [retval][out] */ ColorDictionary **dictionary);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SaveAll )( 
            ColorDictionaries * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *LookUp )( 
            ColorDictionaries * This,
            /* [in] */ Color *Color,
            /* [optional][in] */ VARIANT ClosestMatch,
            /* [retval][out] */ ColorEntry **Entry);
        
        END_INTERFACE
    } ColorDictionariesVtbl;

    interface ColorDictionaries
    {
        CONST_VTBL struct ColorDictionariesVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ColorDictionaries_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ColorDictionaries_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ColorDictionaries_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ColorDictionaries_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ColorDictionaries_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ColorDictionaries_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ColorDictionaries_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ColorDictionaries_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ColorDictionaries_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ColorDictionaries_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define ColorDictionaries_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define ColorDictionaries_Item(This,index,dictionary)	\
    ( (This)->lpVtbl -> Item(This,index,dictionary) ) 

#define ColorDictionaries_SaveAll(This)	\
    ( (This)->lpVtbl -> SaveAll(This) ) 

#define ColorDictionaries_LookUp(This,Color,ClosestMatch,Entry)	\
    ( (This)->lpVtbl -> LookUp(This,Color,ClosestMatch,Entry) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ColorDictionaries_INTERFACE_DEFINED__ */


#ifndef __ExceptionDictionaries_INTERFACE_DEFINED__
#define __ExceptionDictionaries_INTERFACE_DEFINED__

/* interface ExceptionDictionaries */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ExceptionDictionaries;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("27721E7C-F61F-425B-B859-E91EA9D47D92")
    ExceptionDictionaries : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ ExceptionDictionary **dictionary) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SaveAll( 
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE LookUp( 
            /* [in] */ BSTR Word,
            /* [retval][out] */ ExceptionEntry **Entry) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ExceptionDictionariesVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ExceptionDictionaries * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ExceptionDictionaries * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ExceptionDictionaries * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ExceptionDictionaries * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ExceptionDictionaries * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ExceptionDictionaries * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ExceptionDictionaries * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ExceptionDictionaries * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ExceptionDictionaries * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            ExceptionDictionaries * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            ExceptionDictionaries * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            ExceptionDictionaries * This,
            /* [in] */ long index,
            /* [retval][out] */ ExceptionDictionary **dictionary);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SaveAll )( 
            ExceptionDictionaries * This,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *LookUp )( 
            ExceptionDictionaries * This,
            /* [in] */ BSTR Word,
            /* [retval][out] */ ExceptionEntry **Entry);
        
        END_INTERFACE
    } ExceptionDictionariesVtbl;

    interface ExceptionDictionaries
    {
        CONST_VTBL struct ExceptionDictionariesVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ExceptionDictionaries_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ExceptionDictionaries_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ExceptionDictionaries_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ExceptionDictionaries_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ExceptionDictionaries_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ExceptionDictionaries_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ExceptionDictionaries_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ExceptionDictionaries_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ExceptionDictionaries_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ExceptionDictionaries_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define ExceptionDictionaries_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define ExceptionDictionaries_Item(This,index,dictionary)	\
    ( (This)->lpVtbl -> Item(This,index,dictionary) ) 

#define ExceptionDictionaries_SaveAll(This,success)	\
    ( (This)->lpVtbl -> SaveAll(This,success) ) 

#define ExceptionDictionaries_LookUp(This,Word,Entry)	\
    ( (This)->lpVtbl -> LookUp(This,Word,Entry) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ExceptionDictionaries_INTERFACE_DEFINED__ */


#ifndef __GraphicDictionaries_INTERFACE_DEFINED__
#define __GraphicDictionaries_INTERFACE_DEFINED__

/* interface GraphicDictionaries */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_GraphicDictionaries;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("AD9AD626-FEF9-4F4D-BA55-9906F8038779")
    GraphicDictionaries : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ GraphicDictionary **dictionary) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SaveAll( 
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE LookUp( 
            /* [in] */ long CRC,
            /* [retval][out] */ GraphicEntry **Entry) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct GraphicDictionariesVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            GraphicDictionaries * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            GraphicDictionaries * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            GraphicDictionaries * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            GraphicDictionaries * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            GraphicDictionaries * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            GraphicDictionaries * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            GraphicDictionaries * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            GraphicDictionaries * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            GraphicDictionaries * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            GraphicDictionaries * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            GraphicDictionaries * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            GraphicDictionaries * This,
            /* [in] */ long index,
            /* [retval][out] */ GraphicDictionary **dictionary);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SaveAll )( 
            GraphicDictionaries * This,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *LookUp )( 
            GraphicDictionaries * This,
            /* [in] */ long CRC,
            /* [retval][out] */ GraphicEntry **Entry);
        
        END_INTERFACE
    } GraphicDictionariesVtbl;

    interface GraphicDictionaries
    {
        CONST_VTBL struct GraphicDictionariesVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define GraphicDictionaries_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define GraphicDictionaries_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define GraphicDictionaries_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define GraphicDictionaries_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define GraphicDictionaries_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define GraphicDictionaries_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define GraphicDictionaries_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define GraphicDictionaries_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define GraphicDictionaries_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define GraphicDictionaries_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define GraphicDictionaries_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define GraphicDictionaries_Item(This,index,dictionary)	\
    ( (This)->lpVtbl -> Item(This,index,dictionary) ) 

#define GraphicDictionaries_SaveAll(This,success)	\
    ( (This)->lpVtbl -> SaveAll(This,success) ) 

#define GraphicDictionaries_LookUp(This,CRC,Entry)	\
    ( (This)->lpVtbl -> LookUp(This,CRC,Entry) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __GraphicDictionaries_INTERFACE_DEFINED__ */


#ifndef __KeyLabelDictionaries_INTERFACE_DEFINED__
#define __KeyLabelDictionaries_INTERFACE_DEFINED__

/* interface KeyLabelDictionaries */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_KeyLabelDictionaries;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("57BB2672-9231-4E08-948B-B450A2A5A6D6")
    KeyLabelDictionaries : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ KeyLabelDictionary **dictionary) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SaveAll( 
            /* [retval][out] */ VARIANT_BOOL *success) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE LookUp( 
            /* [in] */ VirtualKeyCode Key,
            KeyModifiers mods,
            /* [retval][out] */ KeyLabelEntry **Entry) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct KeyLabelDictionariesVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            KeyLabelDictionaries * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            KeyLabelDictionaries * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            KeyLabelDictionaries * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            KeyLabelDictionaries * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            KeyLabelDictionaries * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            KeyLabelDictionaries * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            KeyLabelDictionaries * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            KeyLabelDictionaries * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            KeyLabelDictionaries * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            KeyLabelDictionaries * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            KeyLabelDictionaries * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            KeyLabelDictionaries * This,
            /* [in] */ long index,
            /* [retval][out] */ KeyLabelDictionary **dictionary);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SaveAll )( 
            KeyLabelDictionaries * This,
            /* [retval][out] */ VARIANT_BOOL *success);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *LookUp )( 
            KeyLabelDictionaries * This,
            /* [in] */ VirtualKeyCode Key,
            KeyModifiers mods,
            /* [retval][out] */ KeyLabelEntry **Entry);
        
        END_INTERFACE
    } KeyLabelDictionariesVtbl;

    interface KeyLabelDictionaries
    {
        CONST_VTBL struct KeyLabelDictionariesVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define KeyLabelDictionaries_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define KeyLabelDictionaries_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define KeyLabelDictionaries_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define KeyLabelDictionaries_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define KeyLabelDictionaries_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define KeyLabelDictionaries_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define KeyLabelDictionaries_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define KeyLabelDictionaries_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define KeyLabelDictionaries_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define KeyLabelDictionaries_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define KeyLabelDictionaries_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define KeyLabelDictionaries_Item(This,index,dictionary)	\
    ( (This)->lpVtbl -> Item(This,index,dictionary) ) 

#define KeyLabelDictionaries_SaveAll(This,success)	\
    ( (This)->lpVtbl -> SaveAll(This,success) ) 

#define KeyLabelDictionaries_LookUp(This,Key,mods,Entry)	\
    ( (This)->lpVtbl -> LookUp(This,Key,mods,Entry) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __KeyLabelDictionaries_INTERFACE_DEFINED__ */


#ifndef ___BrailleDisplays_INTERFACE_DEFINED__
#define ___BrailleDisplays_INTERFACE_DEFINED__

/* interface _BrailleDisplays */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__BrailleDisplays;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("BDFE3BAB-A96D-4994-BE3D-237966DA8F4A")
    _BrailleDisplays : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Active( 
            /* [retval][out] */ _BrailleDisplay **Active) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Initialized( 
            /* [retval][out] */ VARIANT_BOOL *initalized) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ _BrailleDisplay **brailleDisplay) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _BrailleDisplaysVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _BrailleDisplays * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _BrailleDisplays * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _BrailleDisplays * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _BrailleDisplays * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _BrailleDisplays * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _BrailleDisplays * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _BrailleDisplays * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _BrailleDisplays * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _BrailleDisplays * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            _BrailleDisplays * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Active )( 
            _BrailleDisplays * This,
            /* [retval][out] */ _BrailleDisplay **Active);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Initialized )( 
            _BrailleDisplays * This,
            /* [retval][out] */ VARIANT_BOOL *initalized);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            _BrailleDisplays * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            _BrailleDisplays * This,
            /* [in] */ long index,
            /* [retval][out] */ _BrailleDisplay **brailleDisplay);
        
        END_INTERFACE
    } _BrailleDisplaysVtbl;

    interface _BrailleDisplays
    {
        CONST_VTBL struct _BrailleDisplaysVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _BrailleDisplays_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _BrailleDisplays_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _BrailleDisplays_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _BrailleDisplays_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _BrailleDisplays_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _BrailleDisplays_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _BrailleDisplays_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _BrailleDisplays_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _BrailleDisplays_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _BrailleDisplays_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define _BrailleDisplays_get_Active(This,Active)	\
    ( (This)->lpVtbl -> get_Active(This,Active) ) 

#define _BrailleDisplays_get_Initialized(This,initalized)	\
    ( (This)->lpVtbl -> get_Initialized(This,initalized) ) 

#define _BrailleDisplays_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define _BrailleDisplays_Item(This,index,brailleDisplay)	\
    ( (This)->lpVtbl -> Item(This,index,brailleDisplay) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___BrailleDisplays_INTERFACE_DEFINED__ */


#ifndef ___BrailleDisplay_INTERFACE_DEFINED__
#define ___BrailleDisplay_INTERFACE_DEFINED__

/* interface _BrailleDisplay */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__BrailleDisplay;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("7FD24DC7-78CC-4A03-A295-C449BBC521B8")
    _BrailleDisplay : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Description( 
            /* [retval][out] */ BSTR *Description) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_index( 
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Port( 
            /* [retval][out] */ BSTR *Port) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Port( 
            /* [in] */ BSTR Port) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CommandLine( 
            /* [retval][out] */ BSTR *CommandLine) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_type( 
            /* [retval][out] */ BrailleDisplayType *type) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ApplyChanges( 
            /* [optional][in] */ VARIANT Port) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE InsertKey( 
            /* [in] */ BSTR id) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE CaptureKey( 
            /* [retval][out] */ BrailleHotKey **Hotkey) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Keys( 
            /* [optional][in] */ VARIANT Window,
            /* [retval][out] */ BrailleHotKeys **Keys) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _BrailleDisplayVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _BrailleDisplay * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _BrailleDisplay * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _BrailleDisplay * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _BrailleDisplay * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _BrailleDisplay * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _BrailleDisplay * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _BrailleDisplay * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Description )( 
            _BrailleDisplay * This,
            /* [retval][out] */ BSTR *Description);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _BrailleDisplay * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _BrailleDisplay * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_index )( 
            _BrailleDisplay * This,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Port )( 
            _BrailleDisplay * This,
            /* [retval][out] */ BSTR *Port);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Port )( 
            _BrailleDisplay * This,
            /* [in] */ BSTR Port);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CommandLine )( 
            _BrailleDisplay * This,
            /* [retval][out] */ BSTR *CommandLine);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_type )( 
            _BrailleDisplay * This,
            /* [retval][out] */ BrailleDisplayType *type);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ApplyChanges )( 
            _BrailleDisplay * This,
            /* [optional][in] */ VARIANT Port);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *InsertKey )( 
            _BrailleDisplay * This,
            /* [in] */ BSTR id);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *CaptureKey )( 
            _BrailleDisplay * This,
            /* [retval][out] */ BrailleHotKey **Hotkey);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Keys )( 
            _BrailleDisplay * This,
            /* [optional][in] */ VARIANT Window,
            /* [retval][out] */ BrailleHotKeys **Keys);
        
        END_INTERFACE
    } _BrailleDisplayVtbl;

    interface _BrailleDisplay
    {
        CONST_VTBL struct _BrailleDisplayVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _BrailleDisplay_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _BrailleDisplay_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _BrailleDisplay_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _BrailleDisplay_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _BrailleDisplay_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _BrailleDisplay_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _BrailleDisplay_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _BrailleDisplay_get_Description(This,Description)	\
    ( (This)->lpVtbl -> get_Description(This,Description) ) 

#define _BrailleDisplay_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _BrailleDisplay_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _BrailleDisplay_get_index(This,index)	\
    ( (This)->lpVtbl -> get_index(This,index) ) 

#define _BrailleDisplay_get_Port(This,Port)	\
    ( (This)->lpVtbl -> get_Port(This,Port) ) 

#define _BrailleDisplay_put_Port(This,Port)	\
    ( (This)->lpVtbl -> put_Port(This,Port) ) 

#define _BrailleDisplay_get_CommandLine(This,CommandLine)	\
    ( (This)->lpVtbl -> get_CommandLine(This,CommandLine) ) 

#define _BrailleDisplay_get_type(This,type)	\
    ( (This)->lpVtbl -> get_type(This,type) ) 

#define _BrailleDisplay_ApplyChanges(This,Port)	\
    ( (This)->lpVtbl -> ApplyChanges(This,Port) ) 

#define _BrailleDisplay_InsertKey(This,id)	\
    ( (This)->lpVtbl -> InsertKey(This,id) ) 

#define _BrailleDisplay_CaptureKey(This,Hotkey)	\
    ( (This)->lpVtbl -> CaptureKey(This,Hotkey) ) 

#define _BrailleDisplay_Keys(This,Window,Keys)	\
    ( (This)->lpVtbl -> Keys(This,Window,Keys) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___BrailleDisplay_INTERFACE_DEFINED__ */


#ifndef __BrailleHotKey_INTERFACE_DEFINED__
#define __BrailleHotKey_INTERFACE_DEFINED__

/* interface BrailleHotKey */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleHotKey;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("0B85C700-4EB2-4212-972E-4B2F46CAB832")
    BrailleHotKey : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Description( 
            /* [retval][out] */ BSTR *Description) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_id( 
            /* [retval][out] */ BSTR *id) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Touch( 
            /* [retval][out] */ long *Touch) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Actions( 
            /* [retval][out] */ BrailleActions **Actions) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_StatusTouch( 
            /* [retval][out] */ VARIANT_BOOL *Status) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DoubleTouch( 
            /* [retval][out] */ VARIANT_BOOL *dtouch) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleHotKeyVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleHotKey * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleHotKey * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleHotKey * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleHotKey * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleHotKey * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleHotKey * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleHotKey * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Description )( 
            BrailleHotKey * This,
            /* [retval][out] */ BSTR *Description);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleHotKey * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleHotKey * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            BrailleHotKey * This,
            /* [retval][out] */ BSTR *id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Touch )( 
            BrailleHotKey * This,
            /* [retval][out] */ long *Touch);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Actions )( 
            BrailleHotKey * This,
            /* [retval][out] */ BrailleActions **Actions);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_StatusTouch )( 
            BrailleHotKey * This,
            /* [retval][out] */ VARIANT_BOOL *Status);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DoubleTouch )( 
            BrailleHotKey * This,
            /* [retval][out] */ VARIANT_BOOL *dtouch);
        
        END_INTERFACE
    } BrailleHotKeyVtbl;

    interface BrailleHotKey
    {
        CONST_VTBL struct BrailleHotKeyVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleHotKey_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleHotKey_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleHotKey_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleHotKey_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleHotKey_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleHotKey_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleHotKey_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleHotKey_get_Description(This,Description)	\
    ( (This)->lpVtbl -> get_Description(This,Description) ) 

#define BrailleHotKey_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleHotKey_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleHotKey_get_id(This,id)	\
    ( (This)->lpVtbl -> get_id(This,id) ) 

#define BrailleHotKey_get_Touch(This,Touch)	\
    ( (This)->lpVtbl -> get_Touch(This,Touch) ) 

#define BrailleHotKey_get_Actions(This,Actions)	\
    ( (This)->lpVtbl -> get_Actions(This,Actions) ) 

#define BrailleHotKey_get_StatusTouch(This,Status)	\
    ( (This)->lpVtbl -> get_StatusTouch(This,Status) ) 

#define BrailleHotKey_get_DoubleTouch(This,dtouch)	\
    ( (This)->lpVtbl -> get_DoubleTouch(This,dtouch) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleHotKey_INTERFACE_DEFINED__ */


#ifndef __BrailleActions_INTERFACE_DEFINED__
#define __BrailleActions_INTERFACE_DEFINED__

/* interface BrailleActions */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleActions;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("A217A21C-6183-420B-8E07-87EE9C4DF61E")
    BrailleActions : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ BrailleAction **Action) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleActionsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleActions * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleActions * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleActions * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleActions * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleActions * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleActions * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleActions * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleActions * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleActions * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            BrailleActions * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            BrailleActions * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            BrailleActions * This,
            /* [in] */ long index,
            /* [retval][out] */ BrailleAction **Action);
        
        END_INTERFACE
    } BrailleActionsVtbl;

    interface BrailleActions
    {
        CONST_VTBL struct BrailleActionsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleActions_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleActions_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleActions_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleActions_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleActions_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleActions_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleActions_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleActions_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleActions_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleActions_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define BrailleActions_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define BrailleActions_Item(This,index,Action)	\
    ( (This)->lpVtbl -> Item(This,index,Action) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleActions_INTERFACE_DEFINED__ */


#ifndef __BrailleAction_INTERFACE_DEFINED__
#define __BrailleAction_INTERFACE_DEFINED__

/* interface BrailleAction */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleAction;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("82B55C40-AFF9-4DA3-9BD3-439227DD0289")
    BrailleAction : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Description( 
            /* [retval][out] */ BSTR *Description) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_type( 
            /* [retval][out] */ BrailleActionType *type) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Hotkey( 
            /* [retval][out] */ HotkeyID *HotkeyID) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BrailleHotKey( 
            /* [retval][out] */ BrailleHotkeyID *HotkeyID) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Key( 
            /* [retval][out] */ _Key **Key) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Character( 
            /* [retval][out] */ BSTR *Character) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleActionVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleAction * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleAction * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleAction * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleAction * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleAction * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleAction * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleAction * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Description )( 
            BrailleAction * This,
            /* [retval][out] */ BSTR *Description);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleAction * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleAction * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_type )( 
            BrailleAction * This,
            /* [retval][out] */ BrailleActionType *type);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Hotkey )( 
            BrailleAction * This,
            /* [retval][out] */ HotkeyID *HotkeyID);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BrailleHotKey )( 
            BrailleAction * This,
            /* [retval][out] */ BrailleHotkeyID *HotkeyID);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Key )( 
            BrailleAction * This,
            /* [retval][out] */ _Key **Key);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Character )( 
            BrailleAction * This,
            /* [retval][out] */ BSTR *Character);
        
        END_INTERFACE
    } BrailleActionVtbl;

    interface BrailleAction
    {
        CONST_VTBL struct BrailleActionVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleAction_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleAction_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleAction_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleAction_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleAction_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleAction_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleAction_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleAction_get_Description(This,Description)	\
    ( (This)->lpVtbl -> get_Description(This,Description) ) 

#define BrailleAction_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleAction_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleAction_get_type(This,type)	\
    ( (This)->lpVtbl -> get_type(This,type) ) 

#define BrailleAction_get_Hotkey(This,HotkeyID)	\
    ( (This)->lpVtbl -> get_Hotkey(This,HotkeyID) ) 

#define BrailleAction_get_BrailleHotKey(This,HotkeyID)	\
    ( (This)->lpVtbl -> get_BrailleHotKey(This,HotkeyID) ) 

#define BrailleAction_get_Key(This,Key)	\
    ( (This)->lpVtbl -> get_Key(This,Key) ) 

#define BrailleAction_get_Character(This,Character)	\
    ( (This)->lpVtbl -> get_Character(This,Character) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleAction_INTERFACE_DEFINED__ */


#ifndef __BrailleHotKeys_INTERFACE_DEFINED__
#define __BrailleHotKeys_INTERFACE_DEFINED__

/* interface BrailleHotKeys */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleHotKeys;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("F7A4D96B-0044-4665-91A9-B54041A3ACAF")
    BrailleHotKeys : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ BrailleHotKey **Hotkey) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleHotKeysVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleHotKeys * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleHotKeys * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleHotKeys * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleHotKeys * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleHotKeys * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleHotKeys * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleHotKeys * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleHotKeys * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleHotKeys * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            BrailleHotKeys * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            BrailleHotKeys * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            BrailleHotKeys * This,
            /* [in] */ long index,
            /* [retval][out] */ BrailleHotKey **Hotkey);
        
        END_INTERFACE
    } BrailleHotKeysVtbl;

    interface BrailleHotKeys
    {
        CONST_VTBL struct BrailleHotKeysVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleHotKeys_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleHotKeys_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleHotKeys_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleHotKeys_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleHotKeys_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleHotKeys_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleHotKeys_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleHotKeys_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleHotKeys_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleHotKeys_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define BrailleHotKeys_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define BrailleHotKeys_Item(This,index,Hotkey)	\
    ( (This)->lpVtbl -> Item(This,index,Hotkey) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleHotKeys_INTERFACE_DEFINED__ */


#ifndef __BrailleDisplaysEvents_DISPINTERFACE_DEFINED__
#define __BrailleDisplaysEvents_DISPINTERFACE_DEFINED__

/* dispinterface BrailleDisplaysEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID DIID_BrailleDisplaysEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)

    MIDL_INTERFACE("9A96CCBD-4E68-4D73-B5D7-EBF3D45CC930")
    BrailleDisplaysEvents : public IDispatch
    {
    };
    
#else 	/* C style interface */

    typedef struct BrailleDisplaysEventsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleDisplaysEvents * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleDisplaysEvents * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleDisplaysEvents * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleDisplaysEvents * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleDisplaysEvents * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleDisplaysEvents * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleDisplaysEvents * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        END_INTERFACE
    } BrailleDisplaysEventsVtbl;

    interface BrailleDisplaysEvents
    {
        CONST_VTBL struct BrailleDisplaysEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleDisplaysEvents_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleDisplaysEvents_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleDisplaysEvents_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleDisplaysEvents_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleDisplaysEvents_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleDisplaysEvents_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleDisplaysEvents_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */


#endif 	/* __BrailleDisplaysEvents_DISPINTERFACE_DEFINED__ */


#ifndef ___BrailleDisplaysEvents_INTERFACE_DEFINED__
#define ___BrailleDisplaysEvents_INTERFACE_DEFINED__

/* interface _BrailleDisplaysEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID IID__BrailleDisplaysEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("FA8E9BAF-CA70-48AA-A0F7-8FF19B05320B")
    _BrailleDisplaysEvents
    {
    public:
        BEGIN_INTERFACE
        END_INTERFACE
    };
    
    
#else 	/* C style interface */

    typedef struct _BrailleDisplaysEventsVtbl
    {
        BEGIN_INTERFACE
        
        END_INTERFACE
    } _BrailleDisplaysEventsVtbl;

    interface _BrailleDisplaysEvents
    {
        CONST_VTBL struct _BrailleDisplaysEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___BrailleDisplaysEvents_INTERFACE_DEFINED__ */


#ifndef ___SharedObjects_INTERFACE_DEFINED__
#define ___SharedObjects_INTERFACE_DEFINED__

/* interface _SharedObjects */
/* [object][oleautomation][dual][hidden][helpcontext][helpstring][version][uuid] */ 


EXTERN_C const IID IID__SharedObjects;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("8B8674B2-9D28-4968-8666-234AE4229220")
    _SharedObjects : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Get( 
            /* [in] */ BSTR Name,
            /* [optional][in] */ VARIANT Timeout,
            /* [retval][out] */ IDispatch **Object) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Register( 
            /* [in] */ BSTR Name,
            /* [in] */ IDispatch *Object) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Revoke( 
            /* [in] */ BSTR Name) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **Application) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _SharedObjectsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _SharedObjects * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _SharedObjects * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _SharedObjects * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _SharedObjects * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _SharedObjects * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _SharedObjects * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _SharedObjects * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Get )( 
            _SharedObjects * This,
            /* [in] */ BSTR Name,
            /* [optional][in] */ VARIANT Timeout,
            /* [retval][out] */ IDispatch **Object);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Register )( 
            _SharedObjects * This,
            /* [in] */ BSTR Name,
            /* [in] */ IDispatch *Object);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Revoke )( 
            _SharedObjects * This,
            /* [in] */ BSTR Name);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _SharedObjects * This,
            /* [retval][out] */ _Application **Application);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _SharedObjects * This,
            /* [retval][out] */ IDispatch **Parent);
        
        END_INTERFACE
    } _SharedObjectsVtbl;

    interface _SharedObjects
    {
        CONST_VTBL struct _SharedObjectsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _SharedObjects_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _SharedObjects_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _SharedObjects_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _SharedObjects_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _SharedObjects_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _SharedObjects_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _SharedObjects_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _SharedObjects_Get(This,Name,Timeout,Object)	\
    ( (This)->lpVtbl -> Get(This,Name,Timeout,Object) ) 

#define _SharedObjects_Register(This,Name,Object)	\
    ( (This)->lpVtbl -> Register(This,Name,Object) ) 

#define _SharedObjects_Revoke(This,Name)	\
    ( (This)->lpVtbl -> Revoke(This,Name) ) 

#define _SharedObjects_get_Application(This,Application)	\
    ( (This)->lpVtbl -> get_Application(This,Application) ) 

#define _SharedObjects_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___SharedObjects_INTERFACE_DEFINED__ */


#ifndef ___ClientInformation_INTERFACE_DEFINED__
#define ___ClientInformation_INTERFACE_DEFINED__

/* interface _ClientInformation */
/* [object][oleautomation][dual][hidden][helpcontext][helpstring][version][uuid] */ 


EXTERN_C const IID IID__ClientInformation;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("A5082F96-BC80-4AE8-82CC-51101ED91A11")
    _ClientInformation : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **Application) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ScriptFileName( 
            /* [retval][out] */ BSTR *Name) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ScriptProcess( 
            /* [retval][out] */ Process **clientProcess) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ScriptGlobal( 
            /* [retval][out] */ VARIANT_BOOL *Global) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Overlap( 
            /* [retval][out] */ _Window **Overlap) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ScriptLaunchTime( 
            /* [retval][out] */ DATE *launchTime) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ScriptPath( 
            /* [retval][out] */ BSTR *path) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ApplicationProcess( 
            /* [retval][out] */ Process **Process) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ScriptEmbedded( 
            /* [retval][out] */ VARIANT_BOOL *embedded) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UserProfilePath( 
            /* [retval][out] */ BSTR *path) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ScriptDescription( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ScriptDescription( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ScriptHelp( 
            /* [in] */ VARIANT var) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ScriptHelp( 
            /* [retval][out] */ VARIANT *var) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ScriptVersion( 
            /* [in] */ BSTR var) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ScriptVersion( 
            /* [retval][out] */ BSTR *var) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ScriptName( 
            /* [in] */ BSTR var) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ScriptName( 
            /* [retval][out] */ BSTR *var) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UILanguage( 
            /* [retval][out] */ VARIANT *uiLang) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_UILanguage( 
            /* [in] */ VARIANT uiLang) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Overlap( 
            /* [in] */ _Window *Overlap) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Signature( 
            /* [retval][out] */ Signature **Signature) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _ClientInformationVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _ClientInformation * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _ClientInformation * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _ClientInformation * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _ClientInformation * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _ClientInformation * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _ClientInformation * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _ClientInformation * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _ClientInformation * This,
            /* [retval][out] */ _Application **Application);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _ClientInformation * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ScriptFileName )( 
            _ClientInformation * This,
            /* [retval][out] */ BSTR *Name);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ScriptProcess )( 
            _ClientInformation * This,
            /* [retval][out] */ Process **clientProcess);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ScriptGlobal )( 
            _ClientInformation * This,
            /* [retval][out] */ VARIANT_BOOL *Global);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Overlap )( 
            _ClientInformation * This,
            /* [retval][out] */ _Window **Overlap);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ScriptLaunchTime )( 
            _ClientInformation * This,
            /* [retval][out] */ DATE *launchTime);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ScriptPath )( 
            _ClientInformation * This,
            /* [retval][out] */ BSTR *path);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ApplicationProcess )( 
            _ClientInformation * This,
            /* [retval][out] */ Process **Process);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ScriptEmbedded )( 
            _ClientInformation * This,
            /* [retval][out] */ VARIANT_BOOL *embedded);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UserProfilePath )( 
            _ClientInformation * This,
            /* [retval][out] */ BSTR *path);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ScriptDescription )( 
            _ClientInformation * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ScriptDescription )( 
            _ClientInformation * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ScriptHelp )( 
            _ClientInformation * This,
            /* [in] */ VARIANT var);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ScriptHelp )( 
            _ClientInformation * This,
            /* [retval][out] */ VARIANT *var);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ScriptVersion )( 
            _ClientInformation * This,
            /* [in] */ BSTR var);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ScriptVersion )( 
            _ClientInformation * This,
            /* [retval][out] */ BSTR *var);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ScriptName )( 
            _ClientInformation * This,
            /* [in] */ BSTR var);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ScriptName )( 
            _ClientInformation * This,
            /* [retval][out] */ BSTR *var);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UILanguage )( 
            _ClientInformation * This,
            /* [retval][out] */ VARIANT *uiLang);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_UILanguage )( 
            _ClientInformation * This,
            /* [in] */ VARIANT uiLang);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Overlap )( 
            _ClientInformation * This,
            /* [in] */ _Window *Overlap);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Signature )( 
            _ClientInformation * This,
            /* [retval][out] */ Signature **Signature);
        
        END_INTERFACE
    } _ClientInformationVtbl;

    interface _ClientInformation
    {
        CONST_VTBL struct _ClientInformationVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _ClientInformation_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _ClientInformation_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _ClientInformation_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _ClientInformation_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _ClientInformation_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _ClientInformation_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _ClientInformation_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _ClientInformation_get_Application(This,Application)	\
    ( (This)->lpVtbl -> get_Application(This,Application) ) 

#define _ClientInformation_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _ClientInformation_get_ScriptFileName(This,Name)	\
    ( (This)->lpVtbl -> get_ScriptFileName(This,Name) ) 

#define _ClientInformation_get_ScriptProcess(This,clientProcess)	\
    ( (This)->lpVtbl -> get_ScriptProcess(This,clientProcess) ) 

#define _ClientInformation_get_ScriptGlobal(This,Global)	\
    ( (This)->lpVtbl -> get_ScriptGlobal(This,Global) ) 

#define _ClientInformation_get_Overlap(This,Overlap)	\
    ( (This)->lpVtbl -> get_Overlap(This,Overlap) ) 

#define _ClientInformation_get_ScriptLaunchTime(This,launchTime)	\
    ( (This)->lpVtbl -> get_ScriptLaunchTime(This,launchTime) ) 

#define _ClientInformation_get_ScriptPath(This,path)	\
    ( (This)->lpVtbl -> get_ScriptPath(This,path) ) 

#define _ClientInformation_get_ApplicationProcess(This,Process)	\
    ( (This)->lpVtbl -> get_ApplicationProcess(This,Process) ) 

#define _ClientInformation_get_ScriptEmbedded(This,embedded)	\
    ( (This)->lpVtbl -> get_ScriptEmbedded(This,embedded) ) 

#define _ClientInformation_get_UserProfilePath(This,path)	\
    ( (This)->lpVtbl -> get_UserProfilePath(This,path) ) 

#define _ClientInformation_put_ScriptDescription(This,prop)	\
    ( (This)->lpVtbl -> put_ScriptDescription(This,prop) ) 

#define _ClientInformation_get_ScriptDescription(This,prop)	\
    ( (This)->lpVtbl -> get_ScriptDescription(This,prop) ) 

#define _ClientInformation_put_ScriptHelp(This,var)	\
    ( (This)->lpVtbl -> put_ScriptHelp(This,var) ) 

#define _ClientInformation_get_ScriptHelp(This,var)	\
    ( (This)->lpVtbl -> get_ScriptHelp(This,var) ) 

#define _ClientInformation_put_ScriptVersion(This,var)	\
    ( (This)->lpVtbl -> put_ScriptVersion(This,var) ) 

#define _ClientInformation_get_ScriptVersion(This,var)	\
    ( (This)->lpVtbl -> get_ScriptVersion(This,var) ) 

#define _ClientInformation_put_ScriptName(This,var)	\
    ( (This)->lpVtbl -> put_ScriptName(This,var) ) 

#define _ClientInformation_get_ScriptName(This,var)	\
    ( (This)->lpVtbl -> get_ScriptName(This,var) ) 

#define _ClientInformation_get_UILanguage(This,uiLang)	\
    ( (This)->lpVtbl -> get_UILanguage(This,uiLang) ) 

#define _ClientInformation_put_UILanguage(This,uiLang)	\
    ( (This)->lpVtbl -> put_UILanguage(This,uiLang) ) 

#define _ClientInformation_put_Overlap(This,Overlap)	\
    ( (This)->lpVtbl -> put_Overlap(This,Overlap) ) 

#define _ClientInformation_get_Signature(This,Signature)	\
    ( (This)->lpVtbl -> get_Signature(This,Signature) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___ClientInformation_INTERFACE_DEFINED__ */


#ifndef __Signature_INTERFACE_DEFINED__
#define __Signature_INTERFACE_DEFINED__

/* interface Signature */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Signature;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("AA5276FD-6C33-4209-8710-F0E2A617AA42")
    Signature : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Publisher( 
            /* [retval][out] */ BSTR *Publisher) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ThumbPrint( 
            /* [retval][out] */ BSTR *ThumbPrint) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsValid( 
            /* [retval][out] */ VARIANT_BOOL *IsValid) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct SignatureVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Signature * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Signature * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Signature * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Signature * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Signature * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Signature * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Signature * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Signature * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Signature * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Publisher )( 
            Signature * This,
            /* [retval][out] */ BSTR *Publisher);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ThumbPrint )( 
            Signature * This,
            /* [retval][out] */ BSTR *ThumbPrint);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsValid )( 
            Signature * This,
            /* [retval][out] */ VARIANT_BOOL *IsValid);
        
        END_INTERFACE
    } SignatureVtbl;

    interface Signature
    {
        CONST_VTBL struct SignatureVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Signature_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Signature_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Signature_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Signature_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Signature_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Signature_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Signature_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Signature_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Signature_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Signature_get_Publisher(This,Publisher)	\
    ( (This)->lpVtbl -> get_Publisher(This,Publisher) ) 

#define Signature_get_ThumbPrint(This,ThumbPrint)	\
    ( (This)->lpVtbl -> get_ThumbPrint(This,ThumbPrint) ) 

#define Signature_get_IsValid(This,IsValid)	\
    ( (This)->lpVtbl -> get_IsValid(This,IsValid) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Signature_INTERFACE_DEFINED__ */


#ifndef __WECursor_INTERFACE_DEFINED__
#define __WECursor_INTERFACE_DEFINED__

/* interface WECursor */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WECursor;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("A3DCEA0A-BEFA-4B2E-9C01-6681304BA595")
    WECursor : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Position( 
            /* [retval][out] */ ScreenPoint **point) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Position( 
            /* [in] */ ScreenPoint *point) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Window( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ToFocus( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Sync( void) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WECursorVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WECursor * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WECursor * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WECursor * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WECursor * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WECursor * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WECursor * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WECursor * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WECursor * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WECursor * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            WECursor * This,
            /* [retval][out] */ ScreenPoint **point);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            WECursor * This,
            /* [in] */ ScreenPoint *point);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            WECursor * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ToFocus )( 
            WECursor * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Sync )( 
            WECursor * This);
        
        END_INTERFACE
    } WECursorVtbl;

    interface WECursor
    {
        CONST_VTBL struct WECursorVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WECursor_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WECursor_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WECursor_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WECursor_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WECursor_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WECursor_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WECursor_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WECursor_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WECursor_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WECursor_get_Position(This,point)	\
    ( (This)->lpVtbl -> get_Position(This,point) ) 

#define WECursor_put_Position(This,point)	\
    ( (This)->lpVtbl -> put_Position(This,point) ) 

#define WECursor_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define WECursor_ToFocus(This)	\
    ( (This)->lpVtbl -> ToFocus(This) ) 

#define WECursor_Sync(This)	\
    ( (This)->lpVtbl -> Sync(This) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WECursor_INTERFACE_DEFINED__ */


#ifndef __Text_INTERFACE_DEFINED__
#define __Text_INTERFACE_DEFINED__

/* interface Text */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Text;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("A3E20300-8EF2-4E7F-A485-6FBA94DCDCDA")
    Text : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EnclosingRectangle( 
            /* [retval][out] */ ScreenRectangle **ScreenRectangle) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EnclosingRectangle( 
            /* [in] */ ScreenRectangle *ScreenRectangle) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_point( 
            /* [retval][out] */ ScreenPoint **point) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_point( 
            /* [in] */ ScreenPoint *point) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Clip( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Character( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Word( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Sentence( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Paragraph( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Line( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE PriorClip( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE NextClip( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE PriorCharacter( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE NextCharacter( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE PriorWord( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE NextWord( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE PriorSentence( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE NextSentence( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE PriorParagraph( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE NextParagraph( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE PriorLine( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE NextLine( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SentenceStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ParagraphStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE LineStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE PriorSentenceStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE NextSentenceStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE PriorParagraphStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE NextParagraphStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE PriorLineStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE NextLineStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ClipStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE PriorClipStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE NextClipStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE CharacterStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE PriorCharacterStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE NextCharacterStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE WordStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE PriorWordStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE NextWordStatic( 
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct TextVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Text * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Text * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Text * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Text * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Text * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Text * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Text * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Text * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Text * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EnclosingRectangle )( 
            Text * This,
            /* [retval][out] */ ScreenRectangle **ScreenRectangle);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EnclosingRectangle )( 
            Text * This,
            /* [in] */ ScreenRectangle *ScreenRectangle);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_point )( 
            Text * This,
            /* [retval][out] */ ScreenPoint **point);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_point )( 
            Text * This,
            /* [in] */ ScreenPoint *point);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Clip )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Character )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Word )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Sentence )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Paragraph )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Line )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *PriorClip )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *NextClip )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *PriorCharacter )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *NextCharacter )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *PriorWord )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *NextWord )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *PriorSentence )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *NextSentence )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *PriorParagraph )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *NextParagraph )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *PriorLine )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *NextLine )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SentenceStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ParagraphStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *LineStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *PriorSentenceStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *NextSentenceStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *PriorParagraphStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *NextParagraphStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *PriorLineStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *NextLineStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ Clips **Clips);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ClipStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *PriorClipStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *NextClipStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *CharacterStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *PriorCharacterStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *NextCharacterStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *WordStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *PriorWordStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *NextWordStatic )( 
            Text * This,
            /* [optional][in] */ VARIANT ScreenPoint,
            /* [retval][out] */ TextClip **Clip);
        
        END_INTERFACE
    } TextVtbl;

    interface Text
    {
        CONST_VTBL struct TextVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Text_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Text_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Text_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Text_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Text_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Text_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Text_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Text_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Text_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Text_get_EnclosingRectangle(This,ScreenRectangle)	\
    ( (This)->lpVtbl -> get_EnclosingRectangle(This,ScreenRectangle) ) 

#define Text_put_EnclosingRectangle(This,ScreenRectangle)	\
    ( (This)->lpVtbl -> put_EnclosingRectangle(This,ScreenRectangle) ) 

#define Text_get_point(This,point)	\
    ( (This)->lpVtbl -> get_point(This,point) ) 

#define Text_put_point(This,point)	\
    ( (This)->lpVtbl -> put_point(This,point) ) 

#define Text_Clip(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> Clip(This,ScreenPoint,Clip) ) 

#define Text_Character(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> Character(This,ScreenPoint,Clip) ) 

#define Text_Word(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> Word(This,ScreenPoint,Clip) ) 

#define Text_Sentence(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> Sentence(This,ScreenPoint,Clips) ) 

#define Text_Paragraph(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> Paragraph(This,ScreenPoint,Clips) ) 

#define Text_Line(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> Line(This,ScreenPoint,Clips) ) 

#define Text_PriorClip(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> PriorClip(This,ScreenPoint,Clip) ) 

#define Text_NextClip(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> NextClip(This,ScreenPoint,Clip) ) 

#define Text_PriorCharacter(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> PriorCharacter(This,ScreenPoint,Clip) ) 

#define Text_NextCharacter(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> NextCharacter(This,ScreenPoint,Clip) ) 

#define Text_PriorWord(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> PriorWord(This,ScreenPoint,Clip) ) 

#define Text_NextWord(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> NextWord(This,ScreenPoint,Clip) ) 

#define Text_PriorSentence(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> PriorSentence(This,ScreenPoint,Clips) ) 

#define Text_NextSentence(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> NextSentence(This,ScreenPoint,Clips) ) 

#define Text_PriorParagraph(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> PriorParagraph(This,ScreenPoint,Clips) ) 

#define Text_NextParagraph(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> NextParagraph(This,ScreenPoint,Clips) ) 

#define Text_PriorLine(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> PriorLine(This,ScreenPoint,Clips) ) 

#define Text_NextLine(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> NextLine(This,ScreenPoint,Clips) ) 

#define Text_SentenceStatic(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> SentenceStatic(This,ScreenPoint,Clips) ) 

#define Text_ParagraphStatic(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> ParagraphStatic(This,ScreenPoint,Clips) ) 

#define Text_LineStatic(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> LineStatic(This,ScreenPoint,Clips) ) 

#define Text_PriorSentenceStatic(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> PriorSentenceStatic(This,ScreenPoint,Clips) ) 

#define Text_NextSentenceStatic(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> NextSentenceStatic(This,ScreenPoint,Clips) ) 

#define Text_PriorParagraphStatic(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> PriorParagraphStatic(This,ScreenPoint,Clips) ) 

#define Text_NextParagraphStatic(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> NextParagraphStatic(This,ScreenPoint,Clips) ) 

#define Text_PriorLineStatic(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> PriorLineStatic(This,ScreenPoint,Clips) ) 

#define Text_NextLineStatic(This,ScreenPoint,Clips)	\
    ( (This)->lpVtbl -> NextLineStatic(This,ScreenPoint,Clips) ) 

#define Text_ClipStatic(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> ClipStatic(This,ScreenPoint,Clip) ) 

#define Text_PriorClipStatic(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> PriorClipStatic(This,ScreenPoint,Clip) ) 

#define Text_NextClipStatic(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> NextClipStatic(This,ScreenPoint,Clip) ) 

#define Text_CharacterStatic(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> CharacterStatic(This,ScreenPoint,Clip) ) 

#define Text_PriorCharacterStatic(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> PriorCharacterStatic(This,ScreenPoint,Clip) ) 

#define Text_NextCharacterStatic(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> NextCharacterStatic(This,ScreenPoint,Clip) ) 

#define Text_WordStatic(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> WordStatic(This,ScreenPoint,Clip) ) 

#define Text_PriorWordStatic(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> PriorWordStatic(This,ScreenPoint,Clip) ) 

#define Text_NextWordStatic(This,ScreenPoint,Clip)	\
    ( (This)->lpVtbl -> NextWordStatic(This,ScreenPoint,Clip) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Text_INTERFACE_DEFINED__ */


#ifndef __Screen_INTERFACE_DEFINED__
#define __Screen_INTERFACE_DEFINED__

/* interface Screen */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Screen;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("232A9364-2638-4DA3-833F-6524C6220D6F")
    Screen : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Width( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Height( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ColorDepth( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Outline( 
            /* [retval][out] */ ScreenRectangle **rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Outline( 
            /* [in] */ ScreenRectangle *rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FullScreenRectangle( 
            /* [retval][out] */ ScreenRectangle **obj) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE point( 
            /* [in] */ int X,
            /* [in] */ int Y,
            /* [retval][out] */ ScreenPoint **ScreenPoint) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE rectangle( 
            /* [in] */ int Left,
            /* [in] */ int Top,
            /* [in] */ int Right,
            /* [in] */ int Bottom,
            /* [retval][out] */ ScreenRectangle **ScreenRectangle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE PixelColor( 
            /* [in] */ ScreenPoint *point,
            /* [retval][out] */ Color **Color) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ScreenVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Screen * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Screen * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Screen * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Screen * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Screen * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Screen * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Screen * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Screen * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Screen * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Width )( 
            Screen * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Height )( 
            Screen * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ColorDepth )( 
            Screen * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Outline )( 
            Screen * This,
            /* [retval][out] */ ScreenRectangle **rectangle);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Outline )( 
            Screen * This,
            /* [in] */ ScreenRectangle *rectangle);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FullScreenRectangle )( 
            Screen * This,
            /* [retval][out] */ ScreenRectangle **obj);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *point )( 
            Screen * This,
            /* [in] */ int X,
            /* [in] */ int Y,
            /* [retval][out] */ ScreenPoint **ScreenPoint);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *rectangle )( 
            Screen * This,
            /* [in] */ int Left,
            /* [in] */ int Top,
            /* [in] */ int Right,
            /* [in] */ int Bottom,
            /* [retval][out] */ ScreenRectangle **ScreenRectangle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *PixelColor )( 
            Screen * This,
            /* [in] */ ScreenPoint *point,
            /* [retval][out] */ Color **Color);
        
        END_INTERFACE
    } ScreenVtbl;

    interface Screen
    {
        CONST_VTBL struct ScreenVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Screen_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Screen_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Screen_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Screen_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Screen_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Screen_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Screen_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Screen_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Screen_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Screen_get_Width(This,prop)	\
    ( (This)->lpVtbl -> get_Width(This,prop) ) 

#define Screen_get_Height(This,prop)	\
    ( (This)->lpVtbl -> get_Height(This,prop) ) 

#define Screen_get_ColorDepth(This,prop)	\
    ( (This)->lpVtbl -> get_ColorDepth(This,prop) ) 

#define Screen_get_Outline(This,rectangle)	\
    ( (This)->lpVtbl -> get_Outline(This,rectangle) ) 

#define Screen_put_Outline(This,rectangle)	\
    ( (This)->lpVtbl -> put_Outline(This,rectangle) ) 

#define Screen_get_FullScreenRectangle(This,obj)	\
    ( (This)->lpVtbl -> get_FullScreenRectangle(This,obj) ) 

#define Screen_point(This,X,Y,ScreenPoint)	\
    ( (This)->lpVtbl -> point(This,X,Y,ScreenPoint) ) 

#define Screen_rectangle(This,Left,Top,Right,Bottom,ScreenRectangle)	\
    ( (This)->lpVtbl -> rectangle(This,Left,Top,Right,Bottom,ScreenRectangle) ) 

#define Screen_PixelColor(This,point,Color)	\
    ( (This)->lpVtbl -> PixelColor(This,point,Color) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Screen_INTERFACE_DEFINED__ */


#ifndef ___Synthesizers_INTERFACE_DEFINED__
#define ___Synthesizers_INTERFACE_DEFINED__

/* interface _Synthesizers */
/* [object][oleautomation][dual][hidden][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__Synthesizers;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("34DA2E1F-7A7C-47A9-894D-F3E8673B00D0")
    _Synthesizers : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Active( 
            /* [retval][out] */ _Synthesizer **synthsizer) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ _Synthesizer **synthsizer) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _SynthesizersVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _Synthesizers * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _Synthesizers * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _Synthesizers * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _Synthesizers * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _Synthesizers * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _Synthesizers * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _Synthesizers * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _Synthesizers * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _Synthesizers * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            _Synthesizers * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            _Synthesizers * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Active )( 
            _Synthesizers * This,
            /* [retval][out] */ _Synthesizer **synthsizer);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            _Synthesizers * This,
            /* [in] */ long index,
            /* [retval][out] */ _Synthesizer **synthsizer);
        
        END_INTERFACE
    } _SynthesizersVtbl;

    interface _Synthesizers
    {
        CONST_VTBL struct _SynthesizersVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _Synthesizers_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _Synthesizers_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _Synthesizers_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _Synthesizers_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _Synthesizers_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _Synthesizers_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _Synthesizers_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _Synthesizers_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _Synthesizers_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _Synthesizers_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define _Synthesizers_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define _Synthesizers_get_Active(This,synthsizer)	\
    ( (This)->lpVtbl -> get_Active(This,synthsizer) ) 

#define _Synthesizers_Item(This,index,synthsizer)	\
    ( (This)->lpVtbl -> Item(This,index,synthsizer) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___Synthesizers_INTERFACE_DEFINED__ */


#ifndef ___Synthesizer_INTERFACE_DEFINED__
#define ___Synthesizer_INTERFACE_DEFINED__

/* interface _Synthesizer */
/* [object][oleautomation][dual][hidden][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__Synthesizer;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("BC8EE32D-00B9-4EF1-A58D-F85703B9CD4C")
    _Synthesizer : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_PortValue( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BaudRate( 
            /* [retval][out] */ BaudRate *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DataBits( 
            /* [retval][out] */ DataBits *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_StopBits( 
            /* [retval][out] */ StopBits *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parity( 
            /* [retval][out] */ Parity *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FlowControl( 
            /* [retval][out] */ FlowControl *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SoundCard( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Location( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ActiveEngine( 
            /* [retval][out] */ Engine **Engine) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UseSpecifiedSpeechParameters( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UseEngineSpecifiedSpeechParameters( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UseAsBackupSynthesizer( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CommandLine( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_InstalledEngines( 
            /* [retval][out] */ Engines **Engines) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_PortType( 
            /* [retval][out] */ PortType *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_PortValue( 
            /* [in] */ long prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_BaudRate( 
            /* [in] */ BaudRate prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_DataBits( 
            /* [in] */ DataBits prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_StopBits( 
            /* [in] */ StopBits prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Parity( 
            /* [in] */ Parity prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FlowControl( 
            /* [in] */ FlowControl prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SoundCard( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Location( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_UseSpecifiedSpeechParameters( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_UseEngineSpecifiedSpeechParameters( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_UseAsBackupSynthesizer( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_PortType( 
            /* [in] */ PortType prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Description( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MaxRate( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SuspendSpeech( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SuspendSpeech( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SetSpeechParametersToCurrent( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SaveState( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE RestoreState( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE GetNumberOfSoundCards( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE GetSoundCardDescription( 
            /* [in] */ long index,
            /* [retval][out] */ BSTR *Description) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ApplyChanges( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SwitchEngine( 
            Engine *Engine) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Speak( 
            BSTR Text) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _SynthesizerVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _Synthesizer * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _Synthesizer * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _Synthesizer * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _Synthesizer * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _Synthesizer * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _Synthesizer * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _Synthesizer * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _Synthesizer * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _Synthesizer * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_PortValue )( 
            _Synthesizer * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BaudRate )( 
            _Synthesizer * This,
            /* [retval][out] */ BaudRate *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DataBits )( 
            _Synthesizer * This,
            /* [retval][out] */ DataBits *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_StopBits )( 
            _Synthesizer * This,
            /* [retval][out] */ StopBits *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parity )( 
            _Synthesizer * This,
            /* [retval][out] */ Parity *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FlowControl )( 
            _Synthesizer * This,
            /* [retval][out] */ FlowControl *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SoundCard )( 
            _Synthesizer * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Location )( 
            _Synthesizer * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ActiveEngine )( 
            _Synthesizer * This,
            /* [retval][out] */ Engine **Engine);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UseSpecifiedSpeechParameters )( 
            _Synthesizer * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UseEngineSpecifiedSpeechParameters )( 
            _Synthesizer * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UseAsBackupSynthesizer )( 
            _Synthesizer * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CommandLine )( 
            _Synthesizer * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_InstalledEngines )( 
            _Synthesizer * This,
            /* [retval][out] */ Engines **Engines);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_PortType )( 
            _Synthesizer * This,
            /* [retval][out] */ PortType *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_PortValue )( 
            _Synthesizer * This,
            /* [in] */ long prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_BaudRate )( 
            _Synthesizer * This,
            /* [in] */ BaudRate prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_DataBits )( 
            _Synthesizer * This,
            /* [in] */ DataBits prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_StopBits )( 
            _Synthesizer * This,
            /* [in] */ StopBits prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Parity )( 
            _Synthesizer * This,
            /* [in] */ Parity prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FlowControl )( 
            _Synthesizer * This,
            /* [in] */ FlowControl prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SoundCard )( 
            _Synthesizer * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Location )( 
            _Synthesizer * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_UseSpecifiedSpeechParameters )( 
            _Synthesizer * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_UseEngineSpecifiedSpeechParameters )( 
            _Synthesizer * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_UseAsBackupSynthesizer )( 
            _Synthesizer * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_PortType )( 
            _Synthesizer * This,
            /* [in] */ PortType prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Description )( 
            _Synthesizer * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MaxRate )( 
            _Synthesizer * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SuspendSpeech )( 
            _Synthesizer * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SuspendSpeech )( 
            _Synthesizer * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SetSpeechParametersToCurrent )( 
            _Synthesizer * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SaveState )( 
            _Synthesizer * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *RestoreState )( 
            _Synthesizer * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *GetNumberOfSoundCards )( 
            _Synthesizer * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *GetSoundCardDescription )( 
            _Synthesizer * This,
            /* [in] */ long index,
            /* [retval][out] */ BSTR *Description);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ApplyChanges )( 
            _Synthesizer * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SwitchEngine )( 
            _Synthesizer * This,
            Engine *Engine);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Speak )( 
            _Synthesizer * This,
            BSTR Text);
        
        END_INTERFACE
    } _SynthesizerVtbl;

    interface _Synthesizer
    {
        CONST_VTBL struct _SynthesizerVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _Synthesizer_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _Synthesizer_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _Synthesizer_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _Synthesizer_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _Synthesizer_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _Synthesizer_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _Synthesizer_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _Synthesizer_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _Synthesizer_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _Synthesizer_get_PortValue(This,prop)	\
    ( (This)->lpVtbl -> get_PortValue(This,prop) ) 

#define _Synthesizer_get_BaudRate(This,prop)	\
    ( (This)->lpVtbl -> get_BaudRate(This,prop) ) 

#define _Synthesizer_get_DataBits(This,prop)	\
    ( (This)->lpVtbl -> get_DataBits(This,prop) ) 

#define _Synthesizer_get_StopBits(This,prop)	\
    ( (This)->lpVtbl -> get_StopBits(This,prop) ) 

#define _Synthesizer_get_Parity(This,prop)	\
    ( (This)->lpVtbl -> get_Parity(This,prop) ) 

#define _Synthesizer_get_FlowControl(This,prop)	\
    ( (This)->lpVtbl -> get_FlowControl(This,prop) ) 

#define _Synthesizer_get_SoundCard(This,prop)	\
    ( (This)->lpVtbl -> get_SoundCard(This,prop) ) 

#define _Synthesizer_get_Location(This,prop)	\
    ( (This)->lpVtbl -> get_Location(This,prop) ) 

#define _Synthesizer_get_ActiveEngine(This,Engine)	\
    ( (This)->lpVtbl -> get_ActiveEngine(This,Engine) ) 

#define _Synthesizer_get_UseSpecifiedSpeechParameters(This,prop)	\
    ( (This)->lpVtbl -> get_UseSpecifiedSpeechParameters(This,prop) ) 

#define _Synthesizer_get_UseEngineSpecifiedSpeechParameters(This,prop)	\
    ( (This)->lpVtbl -> get_UseEngineSpecifiedSpeechParameters(This,prop) ) 

#define _Synthesizer_get_UseAsBackupSynthesizer(This,prop)	\
    ( (This)->lpVtbl -> get_UseAsBackupSynthesizer(This,prop) ) 

#define _Synthesizer_get_CommandLine(This,prop)	\
    ( (This)->lpVtbl -> get_CommandLine(This,prop) ) 

#define _Synthesizer_get_InstalledEngines(This,Engines)	\
    ( (This)->lpVtbl -> get_InstalledEngines(This,Engines) ) 

#define _Synthesizer_get_PortType(This,prop)	\
    ( (This)->lpVtbl -> get_PortType(This,prop) ) 

#define _Synthesizer_put_PortValue(This,prop)	\
    ( (This)->lpVtbl -> put_PortValue(This,prop) ) 

#define _Synthesizer_put_BaudRate(This,prop)	\
    ( (This)->lpVtbl -> put_BaudRate(This,prop) ) 

#define _Synthesizer_put_DataBits(This,prop)	\
    ( (This)->lpVtbl -> put_DataBits(This,prop) ) 

#define _Synthesizer_put_StopBits(This,prop)	\
    ( (This)->lpVtbl -> put_StopBits(This,prop) ) 

#define _Synthesizer_put_Parity(This,prop)	\
    ( (This)->lpVtbl -> put_Parity(This,prop) ) 

#define _Synthesizer_put_FlowControl(This,prop)	\
    ( (This)->lpVtbl -> put_FlowControl(This,prop) ) 

#define _Synthesizer_put_SoundCard(This,prop)	\
    ( (This)->lpVtbl -> put_SoundCard(This,prop) ) 

#define _Synthesizer_put_Location(This,prop)	\
    ( (This)->lpVtbl -> put_Location(This,prop) ) 

#define _Synthesizer_put_UseSpecifiedSpeechParameters(This,prop)	\
    ( (This)->lpVtbl -> put_UseSpecifiedSpeechParameters(This,prop) ) 

#define _Synthesizer_put_UseEngineSpecifiedSpeechParameters(This,prop)	\
    ( (This)->lpVtbl -> put_UseEngineSpecifiedSpeechParameters(This,prop) ) 

#define _Synthesizer_put_UseAsBackupSynthesizer(This,prop)	\
    ( (This)->lpVtbl -> put_UseAsBackupSynthesizer(This,prop) ) 

#define _Synthesizer_put_PortType(This,prop)	\
    ( (This)->lpVtbl -> put_PortType(This,prop) ) 

#define _Synthesizer_get_Description(This,prop)	\
    ( (This)->lpVtbl -> get_Description(This,prop) ) 

#define _Synthesizer_get_MaxRate(This,prop)	\
    ( (This)->lpVtbl -> get_MaxRate(This,prop) ) 

#define _Synthesizer_get_SuspendSpeech(This,prop)	\
    ( (This)->lpVtbl -> get_SuspendSpeech(This,prop) ) 

#define _Synthesizer_put_SuspendSpeech(This,prop)	\
    ( (This)->lpVtbl -> put_SuspendSpeech(This,prop) ) 

#define _Synthesizer_SetSpeechParametersToCurrent(This)	\
    ( (This)->lpVtbl -> SetSpeechParametersToCurrent(This) ) 

#define _Synthesizer_SaveState(This)	\
    ( (This)->lpVtbl -> SaveState(This) ) 

#define _Synthesizer_RestoreState(This)	\
    ( (This)->lpVtbl -> RestoreState(This) ) 

#define _Synthesizer_GetNumberOfSoundCards(This,Count)	\
    ( (This)->lpVtbl -> GetNumberOfSoundCards(This,Count) ) 

#define _Synthesizer_GetSoundCardDescription(This,index,Description)	\
    ( (This)->lpVtbl -> GetSoundCardDescription(This,index,Description) ) 

#define _Synthesizer_ApplyChanges(This)	\
    ( (This)->lpVtbl -> ApplyChanges(This) ) 

#define _Synthesizer_SwitchEngine(This,Engine)	\
    ( (This)->lpVtbl -> SwitchEngine(This,Engine) ) 

#define _Synthesizer_Speak(This,Text)	\
    ( (This)->lpVtbl -> Speak(This,Text) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___Synthesizer_INTERFACE_DEFINED__ */


#ifndef __Engine_INTERFACE_DEFINED__
#define __Engine_INTERFACE_DEFINED__

/* interface Engine */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Engine;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("060585CE-18DF-4131-9138-4CEC7A664B43")
    Engine : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Description( 
            /* [retval][out] */ BSTR *param) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Manufacturer( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Product( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Speaker( 
            /* [retval][out] */ BSTR *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct EngineVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Engine * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Engine * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Engine * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Engine * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Engine * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Engine * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Engine * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Engine * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Engine * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Description )( 
            Engine * This,
            /* [retval][out] */ BSTR *param);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Manufacturer )( 
            Engine * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Product )( 
            Engine * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Speaker )( 
            Engine * This,
            /* [retval][out] */ BSTR *prop);
        
        END_INTERFACE
    } EngineVtbl;

    interface Engine
    {
        CONST_VTBL struct EngineVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Engine_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Engine_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Engine_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Engine_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Engine_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Engine_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Engine_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Engine_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Engine_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Engine_get_Description(This,param)	\
    ( (This)->lpVtbl -> get_Description(This,param) ) 

#define Engine_get_Manufacturer(This,prop)	\
    ( (This)->lpVtbl -> get_Manufacturer(This,prop) ) 

#define Engine_get_Product(This,prop)	\
    ( (This)->lpVtbl -> get_Product(This,prop) ) 

#define Engine_get_Speaker(This,prop)	\
    ( (This)->lpVtbl -> get_Speaker(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Engine_INTERFACE_DEFINED__ */


#ifndef __Engines_INTERFACE_DEFINED__
#define __Engines_INTERFACE_DEFINED__

/* interface Engines */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Engines;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("A153F258-36B0-4D4A-ABC8-70CACD323F65")
    Engines : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ Engine **Engine) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct EnginesVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Engines * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Engines * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Engines * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Engines * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Engines * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Engines * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Engines * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Engines * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Engines * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            Engines * This,
            /* [retval][out] */ long *Count);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            Engines * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            Engines * This,
            /* [in] */ long index,
            /* [retval][out] */ Engine **Engine);
        
        END_INTERFACE
    } EnginesVtbl;

    interface Engines
    {
        CONST_VTBL struct EnginesVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Engines_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Engines_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Engines_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Engines_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Engines_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Engines_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Engines_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Engines_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Engines_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Engines_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define Engines_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define Engines_Item(This,index,Engine)	\
    ( (This)->lpVtbl -> Item(This,index,Engine) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Engines_INTERFACE_DEFINED__ */


#ifndef __IniFile_INTERFACE_DEFINED__
#define __IniFile_INTERFACE_DEFINED__

/* interface IniFile */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_IniFile;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("3706B606-932A-4C9D-ADBD-9B93C7BE4E8D")
    IniFile : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Text( 
            /* [in] */ BSTR section,
            /* [in] */ BSTR Key,
            /* [optional][in] */ VARIANT defVal,
            /* [retval][out] */ BSTR *Value) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Text( 
            /* [in] */ BSTR section,
            /* [in] */ BSTR Key,
            /* [optional][in] */ VARIANT defVal,
            /* [in] */ BSTR Value) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Number( 
            /* [in] */ BSTR section,
            /* [in] */ BSTR Key,
            /* [optional][in] */ VARIANT defVal,
            /* [retval][out] */ long *Value) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Number( 
            /* [in] */ BSTR section,
            /* [in] */ BSTR Key,
            /* [optional][in] */ VARIANT defVal,
            /* [in] */ long Value) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE DeleteSection( 
            /* [in] */ BSTR section) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE DeleteEntry( 
            /* [in] */ BSTR section,
            /* [in] */ BSTR Key) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE GetSectionKeys( 
            /* [in] */ BSTR section,
            /* [retval][out] */ BSTR *Sections) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE GetSectionNames( 
            /* [retval][out] */ BSTR *names) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct IniFileVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            IniFile * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            IniFile * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            IniFile * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            IniFile * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            IniFile * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            IniFile * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            IniFile * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            IniFile * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            IniFile * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            IniFile * This,
            /* [in] */ BSTR section,
            /* [in] */ BSTR Key,
            /* [optional][in] */ VARIANT defVal,
            /* [retval][out] */ BSTR *Value);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            IniFile * This,
            /* [in] */ BSTR section,
            /* [in] */ BSTR Key,
            /* [optional][in] */ VARIANT defVal,
            /* [in] */ BSTR Value);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Number )( 
            IniFile * This,
            /* [in] */ BSTR section,
            /* [in] */ BSTR Key,
            /* [optional][in] */ VARIANT defVal,
            /* [retval][out] */ long *Value);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Number )( 
            IniFile * This,
            /* [in] */ BSTR section,
            /* [in] */ BSTR Key,
            /* [optional][in] */ VARIANT defVal,
            /* [in] */ long Value);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *DeleteSection )( 
            IniFile * This,
            /* [in] */ BSTR section);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *DeleteEntry )( 
            IniFile * This,
            /* [in] */ BSTR section,
            /* [in] */ BSTR Key);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *GetSectionKeys )( 
            IniFile * This,
            /* [in] */ BSTR section,
            /* [retval][out] */ BSTR *Sections);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *GetSectionNames )( 
            IniFile * This,
            /* [retval][out] */ BSTR *names);
        
        END_INTERFACE
    } IniFileVtbl;

    interface IniFile
    {
        CONST_VTBL struct IniFileVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define IniFile_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define IniFile_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define IniFile_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define IniFile_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define IniFile_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define IniFile_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define IniFile_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define IniFile_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define IniFile_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define IniFile_get_Text(This,section,Key,defVal,Value)	\
    ( (This)->lpVtbl -> get_Text(This,section,Key,defVal,Value) ) 

#define IniFile_put_Text(This,section,Key,defVal,Value)	\
    ( (This)->lpVtbl -> put_Text(This,section,Key,defVal,Value) ) 

#define IniFile_get_Number(This,section,Key,defVal,Value)	\
    ( (This)->lpVtbl -> get_Number(This,section,Key,defVal,Value) ) 

#define IniFile_put_Number(This,section,Key,defVal,Value)	\
    ( (This)->lpVtbl -> put_Number(This,section,Key,defVal,Value) ) 

#define IniFile_DeleteSection(This,section)	\
    ( (This)->lpVtbl -> DeleteSection(This,section) ) 

#define IniFile_DeleteEntry(This,section,Key)	\
    ( (This)->lpVtbl -> DeleteEntry(This,section,Key) ) 

#define IniFile_GetSectionKeys(This,section,Sections)	\
    ( (This)->lpVtbl -> GetSectionKeys(This,section,Sections) ) 

#define IniFile_GetSectionNames(This,names)	\
    ( (This)->lpVtbl -> GetSectionNames(This,names) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __IniFile_INTERFACE_DEFINED__ */


#ifndef ___Speech_INTERFACE_DEFINED__
#define ___Speech_INTERFACE_DEFINED__

/* interface _Speech */
/* [object][oleautomation][dual][hidden][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__Speech;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("F6344F01-A798-46FE-8736-F321A69F71E6")
    _Speech : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ApplicationSpeakAll( 
            /* [in] */ _Window *Overlap,
            /* [retval][out] */ VARIANT_BOOL *speakall) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ApplicationSpeakAll( 
            /* [in] */ _Window *Overlap,
            /* [in] */ VARIANT_BOOL speakall) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ApplicationSpeech( 
            /* [in] */ _Window *Overlap,
            /* [retval][out] */ VARIANT_BOOL *Speak) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ApplicationSpeech( 
            /* [in] */ _Window *Overlap,
            /* [in] */ VARIANT_BOOL Speak) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_GlobalSpeech( 
            /* [retval][out] */ VARIANT_BOOL *Speak) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_GlobalSpeech( 
            /* [in] */ VARIANT_BOOL Speak) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Silence( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Speak( 
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT Voice) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE CharacterName( 
            /* [in] */ VARIANT vCharacters,
            /* [retval][out] */ BSTR *names) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Phonetic( 
            /* [in] */ VARIANT vCharacters,
            /* [retval][out] */ BSTR *phonetics) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Spell( 
            /* [in] */ VARIANT vCharacters,
            /* [optional][in] */ VARIANT Voice,
            /* [optional][in] */ VARIANT Phonetic) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _SpeechVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _Speech * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _Speech * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _Speech * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _Speech * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _Speech * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _Speech * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _Speech * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _Speech * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _Speech * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ApplicationSpeakAll )( 
            _Speech * This,
            /* [in] */ _Window *Overlap,
            /* [retval][out] */ VARIANT_BOOL *speakall);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ApplicationSpeakAll )( 
            _Speech * This,
            /* [in] */ _Window *Overlap,
            /* [in] */ VARIANT_BOOL speakall);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ApplicationSpeech )( 
            _Speech * This,
            /* [in] */ _Window *Overlap,
            /* [retval][out] */ VARIANT_BOOL *Speak);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ApplicationSpeech )( 
            _Speech * This,
            /* [in] */ _Window *Overlap,
            /* [in] */ VARIANT_BOOL Speak);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_GlobalSpeech )( 
            _Speech * This,
            /* [retval][out] */ VARIANT_BOOL *Speak);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_GlobalSpeech )( 
            _Speech * This,
            /* [in] */ VARIANT_BOOL Speak);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Silence )( 
            _Speech * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Speak )( 
            _Speech * This,
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT Voice);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *CharacterName )( 
            _Speech * This,
            /* [in] */ VARIANT vCharacters,
            /* [retval][out] */ BSTR *names);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Phonetic )( 
            _Speech * This,
            /* [in] */ VARIANT vCharacters,
            /* [retval][out] */ BSTR *phonetics);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Spell )( 
            _Speech * This,
            /* [in] */ VARIANT vCharacters,
            /* [optional][in] */ VARIANT Voice,
            /* [optional][in] */ VARIANT Phonetic);
        
        END_INTERFACE
    } _SpeechVtbl;

    interface _Speech
    {
        CONST_VTBL struct _SpeechVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _Speech_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _Speech_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _Speech_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _Speech_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _Speech_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _Speech_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _Speech_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _Speech_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _Speech_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _Speech_get_ApplicationSpeakAll(This,Overlap,speakall)	\
    ( (This)->lpVtbl -> get_ApplicationSpeakAll(This,Overlap,speakall) ) 

#define _Speech_put_ApplicationSpeakAll(This,Overlap,speakall)	\
    ( (This)->lpVtbl -> put_ApplicationSpeakAll(This,Overlap,speakall) ) 

#define _Speech_get_ApplicationSpeech(This,Overlap,Speak)	\
    ( (This)->lpVtbl -> get_ApplicationSpeech(This,Overlap,Speak) ) 

#define _Speech_put_ApplicationSpeech(This,Overlap,Speak)	\
    ( (This)->lpVtbl -> put_ApplicationSpeech(This,Overlap,Speak) ) 

#define _Speech_get_GlobalSpeech(This,Speak)	\
    ( (This)->lpVtbl -> get_GlobalSpeech(This,Speak) ) 

#define _Speech_put_GlobalSpeech(This,Speak)	\
    ( (This)->lpVtbl -> put_GlobalSpeech(This,Speak) ) 

#define _Speech_Silence(This)	\
    ( (This)->lpVtbl -> Silence(This) ) 

#define _Speech_Speak(This,Text,Voice)	\
    ( (This)->lpVtbl -> Speak(This,Text,Voice) ) 

#define _Speech_CharacterName(This,vCharacters,names)	\
    ( (This)->lpVtbl -> CharacterName(This,vCharacters,names) ) 

#define _Speech_Phonetic(This,vCharacters,phonetics)	\
    ( (This)->lpVtbl -> Phonetic(This,vCharacters,phonetics) ) 

#define _Speech_Spell(This,vCharacters,Voice,Phonetic)	\
    ( (This)->lpVtbl -> Spell(This,vCharacters,Voice,Phonetic) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___Speech_INTERFACE_DEFINED__ */


#ifndef ___MSAAEvents_INTERFACE_DEFINED__
#define ___MSAAEvents_INTERFACE_DEFINED__

/* interface _MSAAEvents */
/* [object][oleautomation][dual][hidden][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__MSAAEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("27EE6A61-246B-45CB-87C0-8A64E577445A")
    _MSAAEvents : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Process( 
            /* [retval][out] */ Process **Process) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Process( 
            /* [in] */ Process *Process) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Window( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Window( 
            /* [in] */ _Window *Window) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Scope( 
            /* [retval][out] */ WindowFilterScope *Scope) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Scope( 
            /* [in] */ WindowFilterScope Scope) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_role( 
            /* [retval][out] */ AccessibleRoleEnum *role) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_role( 
            /* [in] */ AccessibleRoleEnum role) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE WatchEvent( 
            /* [in] */ MSAAEventID id) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE BlockEvent( 
            /* [in] */ MSAAEventID id,
            /* [optional][in] */ VARIANT Process,
            /* [retval][out] */ MSAAEventBlock **block) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _MSAAEventsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _MSAAEvents * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _MSAAEvents * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _MSAAEvents * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _MSAAEvents * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _MSAAEvents * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _MSAAEvents * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _MSAAEvents * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _MSAAEvents * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _MSAAEvents * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Process )( 
            _MSAAEvents * This,
            /* [retval][out] */ Process **Process);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Process )( 
            _MSAAEvents * This,
            /* [in] */ Process *Process);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            _MSAAEvents * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Window )( 
            _MSAAEvents * This,
            /* [in] */ _Window *Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Scope )( 
            _MSAAEvents * This,
            /* [retval][out] */ WindowFilterScope *Scope);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Scope )( 
            _MSAAEvents * This,
            /* [in] */ WindowFilterScope Scope);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_role )( 
            _MSAAEvents * This,
            /* [retval][out] */ AccessibleRoleEnum *role);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_role )( 
            _MSAAEvents * This,
            /* [in] */ AccessibleRoleEnum role);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *WatchEvent )( 
            _MSAAEvents * This,
            /* [in] */ MSAAEventID id);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *BlockEvent )( 
            _MSAAEvents * This,
            /* [in] */ MSAAEventID id,
            /* [optional][in] */ VARIANT Process,
            /* [retval][out] */ MSAAEventBlock **block);
        
        END_INTERFACE
    } _MSAAEventsVtbl;

    interface _MSAAEvents
    {
        CONST_VTBL struct _MSAAEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _MSAAEvents_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _MSAAEvents_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _MSAAEvents_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _MSAAEvents_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _MSAAEvents_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _MSAAEvents_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _MSAAEvents_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _MSAAEvents_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _MSAAEvents_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _MSAAEvents_get_Process(This,Process)	\
    ( (This)->lpVtbl -> get_Process(This,Process) ) 

#define _MSAAEvents_put_Process(This,Process)	\
    ( (This)->lpVtbl -> put_Process(This,Process) ) 

#define _MSAAEvents_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define _MSAAEvents_put_Window(This,Window)	\
    ( (This)->lpVtbl -> put_Window(This,Window) ) 

#define _MSAAEvents_get_Scope(This,Scope)	\
    ( (This)->lpVtbl -> get_Scope(This,Scope) ) 

#define _MSAAEvents_put_Scope(This,Scope)	\
    ( (This)->lpVtbl -> put_Scope(This,Scope) ) 

#define _MSAAEvents_get_role(This,role)	\
    ( (This)->lpVtbl -> get_role(This,role) ) 

#define _MSAAEvents_put_role(This,role)	\
    ( (This)->lpVtbl -> put_role(This,role) ) 

#define _MSAAEvents_WatchEvent(This,id)	\
    ( (This)->lpVtbl -> WatchEvent(This,id) ) 

#define _MSAAEvents_BlockEvent(This,id,Process,block)	\
    ( (This)->lpVtbl -> BlockEvent(This,id,Process,block) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___MSAAEvents_INTERFACE_DEFINED__ */


#ifndef __MSAAEventBlock_INTERFACE_DEFINED__
#define __MSAAEventBlock_INTERFACE_DEFINED__

/* interface MSAAEventBlock */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_MSAAEventBlock;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("11CA4E4B-148D-4311-9B1A-6BCC2C466647")
    MSAAEventBlock : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Unblock( void) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct MSAAEventBlockVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            MSAAEventBlock * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            MSAAEventBlock * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            MSAAEventBlock * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            MSAAEventBlock * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            MSAAEventBlock * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            MSAAEventBlock * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            MSAAEventBlock * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Unblock )( 
            MSAAEventBlock * This);
        
        END_INTERFACE
    } MSAAEventBlockVtbl;

    interface MSAAEventBlock
    {
        CONST_VTBL struct MSAAEventBlockVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define MSAAEventBlock_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define MSAAEventBlock_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define MSAAEventBlock_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define MSAAEventBlock_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define MSAAEventBlock_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define MSAAEventBlock_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define MSAAEventBlock_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define MSAAEventBlock_Unblock(This)	\
    ( (This)->lpVtbl -> Unblock(This) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __MSAAEventBlock_INTERFACE_DEFINED__ */


#ifndef __Clipboard_INTERFACE_DEFINED__
#define __Clipboard_INTERFACE_DEFINED__

/* interface Clipboard */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Clipboard;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("72A9B17D-98B3-44F7-9BE1-A5E892170262")
    Clipboard : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Text( 
            /* [retval][out] */ BSTR *Text) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Text( 
            /* [in] */ BSTR Text) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsText( 
            /* [retval][out] */ VARIANT_BOOL *result) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsBitmap( 
            /* [retval][out] */ VARIANT_BOOL *result) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE AppendText( 
            /* [in] */ BSTR Text) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ClipboardVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Clipboard * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Clipboard * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Clipboard * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Clipboard * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Clipboard * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Clipboard * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Clipboard * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Clipboard * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Clipboard * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            Clipboard * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            Clipboard * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsText )( 
            Clipboard * This,
            /* [retval][out] */ VARIANT_BOOL *result);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsBitmap )( 
            Clipboard * This,
            /* [retval][out] */ VARIANT_BOOL *result);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *AppendText )( 
            Clipboard * This,
            /* [in] */ BSTR Text);
        
        END_INTERFACE
    } ClipboardVtbl;

    interface Clipboard
    {
        CONST_VTBL struct ClipboardVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Clipboard_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Clipboard_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Clipboard_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Clipboard_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Clipboard_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Clipboard_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Clipboard_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Clipboard_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Clipboard_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Clipboard_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define Clipboard_put_Text(This,Text)	\
    ( (This)->lpVtbl -> put_Text(This,Text) ) 

#define Clipboard_get_IsText(This,result)	\
    ( (This)->lpVtbl -> get_IsText(This,result) ) 

#define Clipboard_get_IsBitmap(This,result)	\
    ( (This)->lpVtbl -> get_IsBitmap(This,result) ) 

#define Clipboard_AppendText(This,Text)	\
    ( (This)->lpVtbl -> AppendText(This,Text) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Clipboard_INTERFACE_DEFINED__ */


#ifndef ___BrowseMode_INTERFACE_DEFINED__
#define ___BrowseMode_INTERFACE_DEFINED__

/* interface _BrowseMode */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__BrowseMode;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("7298ACC8-3023-469A-BFC7-7135F9F8561F")
    _BrowseMode : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Lines( 
            /* [retval][out] */ BrowseModeLines **Lines) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SelectionBeginLine( 
            /* [retval][out] */ long *Line) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SelectionBeginColumn( 
            /* [retval][out] */ long *column) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SelectionEndLine( 
            /* [retval][out] */ long *Line) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SelectionEndColumn( 
            /* [retval][out] */ long *column) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_state( 
            /* [retval][out] */ BrowseModeState *state) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CursorLine( 
            /* [retval][out] */ long *Line) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CursorLine( 
            /* [in] */ long Line) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CursorColumn( 
            /* [retval][out] */ long *column) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_CursorColumn( 
            /* [in] */ long column) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Copy( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Reload( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Select( 
            /* [in] */ long startLine,
            /* [in] */ long startColumn,
            /* [in] */ long endLine,
            /* [in] */ long endColumn) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _BrowseModeVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _BrowseMode * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _BrowseMode * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _BrowseMode * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _BrowseMode * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _BrowseMode * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _BrowseMode * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _BrowseMode * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Lines )( 
            _BrowseMode * This,
            /* [retval][out] */ BrowseModeLines **Lines);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _BrowseMode * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _BrowseMode * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SelectionBeginLine )( 
            _BrowseMode * This,
            /* [retval][out] */ long *Line);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SelectionBeginColumn )( 
            _BrowseMode * This,
            /* [retval][out] */ long *column);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SelectionEndLine )( 
            _BrowseMode * This,
            /* [retval][out] */ long *Line);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SelectionEndColumn )( 
            _BrowseMode * This,
            /* [retval][out] */ long *column);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_state )( 
            _BrowseMode * This,
            /* [retval][out] */ BrowseModeState *state);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CursorLine )( 
            _BrowseMode * This,
            /* [retval][out] */ long *Line);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CursorLine )( 
            _BrowseMode * This,
            /* [in] */ long Line);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CursorColumn )( 
            _BrowseMode * This,
            /* [retval][out] */ long *column);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_CursorColumn )( 
            _BrowseMode * This,
            /* [in] */ long column);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Copy )( 
            _BrowseMode * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Reload )( 
            _BrowseMode * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Select )( 
            _BrowseMode * This,
            /* [in] */ long startLine,
            /* [in] */ long startColumn,
            /* [in] */ long endLine,
            /* [in] */ long endColumn);
        
        END_INTERFACE
    } _BrowseModeVtbl;

    interface _BrowseMode
    {
        CONST_VTBL struct _BrowseModeVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _BrowseMode_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _BrowseMode_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _BrowseMode_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _BrowseMode_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _BrowseMode_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _BrowseMode_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _BrowseMode_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _BrowseMode_get_Lines(This,Lines)	\
    ( (This)->lpVtbl -> get_Lines(This,Lines) ) 

#define _BrowseMode_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _BrowseMode_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _BrowseMode_get_SelectionBeginLine(This,Line)	\
    ( (This)->lpVtbl -> get_SelectionBeginLine(This,Line) ) 

#define _BrowseMode_get_SelectionBeginColumn(This,column)	\
    ( (This)->lpVtbl -> get_SelectionBeginColumn(This,column) ) 

#define _BrowseMode_get_SelectionEndLine(This,Line)	\
    ( (This)->lpVtbl -> get_SelectionEndLine(This,Line) ) 

#define _BrowseMode_get_SelectionEndColumn(This,column)	\
    ( (This)->lpVtbl -> get_SelectionEndColumn(This,column) ) 

#define _BrowseMode_get_state(This,state)	\
    ( (This)->lpVtbl -> get_state(This,state) ) 

#define _BrowseMode_get_CursorLine(This,Line)	\
    ( (This)->lpVtbl -> get_CursorLine(This,Line) ) 

#define _BrowseMode_put_CursorLine(This,Line)	\
    ( (This)->lpVtbl -> put_CursorLine(This,Line) ) 

#define _BrowseMode_get_CursorColumn(This,column)	\
    ( (This)->lpVtbl -> get_CursorColumn(This,column) ) 

#define _BrowseMode_put_CursorColumn(This,column)	\
    ( (This)->lpVtbl -> put_CursorColumn(This,column) ) 

#define _BrowseMode_Copy(This)	\
    ( (This)->lpVtbl -> Copy(This) ) 

#define _BrowseMode_Reload(This)	\
    ( (This)->lpVtbl -> Reload(This) ) 

#define _BrowseMode_Select(This,startLine,startColumn,endLine,endColumn)	\
    ( (This)->lpVtbl -> Select(This,startLine,startColumn,endLine,endColumn) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___BrowseMode_INTERFACE_DEFINED__ */


#ifndef __BrowseModeLines_INTERFACE_DEFINED__
#define __BrowseModeLines_INTERFACE_DEFINED__

/* interface BrowseModeLines */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeLines;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("E5EFF012-6802-4DC8-990F-571D6856DCDD")
    BrowseModeLines : public IDispatch
    {
    public:
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ BrowseModeLine **Line) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeLinesVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeLines * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeLines * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeLines * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeLines * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeLines * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeLines * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeLines * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            BrowseModeLines * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeLines * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeLines * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            BrowseModeLines * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            BrowseModeLines * This,
            /* [in] */ long index,
            /* [retval][out] */ BrowseModeLine **Line);
        
        END_INTERFACE
    } BrowseModeLinesVtbl;

    interface BrowseModeLines
    {
        CONST_VTBL struct BrowseModeLinesVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeLines_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeLines_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeLines_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeLines_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeLines_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeLines_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeLines_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeLines_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define BrowseModeLines_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeLines_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeLines_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define BrowseModeLines_Item(This,index,Line)	\
    ( (This)->lpVtbl -> Item(This,index,Line) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeLines_INTERFACE_DEFINED__ */


#ifndef __BrowseModeLine_INTERFACE_DEFINED__
#define __BrowseModeLine_INTERFACE_DEFINED__

/* interface BrowseModeLine */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrowseModeLine;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("08C8FAFF-DFBD-48EF-85A3-6E6886AB8457")
    BrowseModeLine : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Text( 
            /* [retval][out] */ BSTR *Text) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Text( 
            /* [in] */ BSTR Text) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DOMNode( 
            /* [retval][out] */ IDispatch **DOMNode) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_index( 
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsLink( 
            /* [retval][out] */ VARIANT_BOOL *IsLink) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsAbbreviation( 
            /* [retval][out] */ VARIANT_BOOL *IsAbbreviation) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HasAccessKey( 
            /* [retval][out] */ VARIANT_BOOL *HasAccessKey) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsAcronym( 
            /* [retval][out] */ VARIANT_BOOL *IsAcronym) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsBlank( 
            /* [retval][out] */ VARIANT_BOOL *IsBlank) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsGraphic( 
            /* [retval][out] */ VARIANT_BOOL *IsGraphic) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsHeading( 
            /* [retval][out] */ VARIANT_BOOL *IsHeading) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_HasOnClick( 
            /* [retval][out] */ VARIANT_BOOL *HasOnClick) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsQuote( 
            /* [retval][out] */ VARIANT_BOOL *IsQuote) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsSecure( 
            /* [retval][out] */ VARIANT_BOOL *IsSecure) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_IsSilent( 
            /* [retval][out] */ VARIANT_BOOL *IsSilent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Accessible( 
            /* [retval][out] */ Accessible **Accessible) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrowseModeLineVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeLine * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeLine * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeLine * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeLine * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeLine * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeLine * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeLine * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            BrowseModeLine * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            BrowseModeLine * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrowseModeLine * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrowseModeLine * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DOMNode )( 
            BrowseModeLine * This,
            /* [retval][out] */ IDispatch **DOMNode);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_index )( 
            BrowseModeLine * This,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsLink )( 
            BrowseModeLine * This,
            /* [retval][out] */ VARIANT_BOOL *IsLink);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsAbbreviation )( 
            BrowseModeLine * This,
            /* [retval][out] */ VARIANT_BOOL *IsAbbreviation);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HasAccessKey )( 
            BrowseModeLine * This,
            /* [retval][out] */ VARIANT_BOOL *HasAccessKey);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsAcronym )( 
            BrowseModeLine * This,
            /* [retval][out] */ VARIANT_BOOL *IsAcronym);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsBlank )( 
            BrowseModeLine * This,
            /* [retval][out] */ VARIANT_BOOL *IsBlank);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsGraphic )( 
            BrowseModeLine * This,
            /* [retval][out] */ VARIANT_BOOL *IsGraphic);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsHeading )( 
            BrowseModeLine * This,
            /* [retval][out] */ VARIANT_BOOL *IsHeading);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_HasOnClick )( 
            BrowseModeLine * This,
            /* [retval][out] */ VARIANT_BOOL *HasOnClick);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsQuote )( 
            BrowseModeLine * This,
            /* [retval][out] */ VARIANT_BOOL *IsQuote);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsSecure )( 
            BrowseModeLine * This,
            /* [retval][out] */ VARIANT_BOOL *IsSecure);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_IsSilent )( 
            BrowseModeLine * This,
            /* [retval][out] */ VARIANT_BOOL *IsSilent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Accessible )( 
            BrowseModeLine * This,
            /* [retval][out] */ Accessible **Accessible);
        
        END_INTERFACE
    } BrowseModeLineVtbl;

    interface BrowseModeLine
    {
        CONST_VTBL struct BrowseModeLineVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeLine_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeLine_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeLine_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeLine_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeLine_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeLine_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeLine_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrowseModeLine_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define BrowseModeLine_put_Text(This,Text)	\
    ( (This)->lpVtbl -> put_Text(This,Text) ) 

#define BrowseModeLine_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrowseModeLine_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrowseModeLine_get_DOMNode(This,DOMNode)	\
    ( (This)->lpVtbl -> get_DOMNode(This,DOMNode) ) 

#define BrowseModeLine_get_index(This,index)	\
    ( (This)->lpVtbl -> get_index(This,index) ) 

#define BrowseModeLine_get_IsLink(This,IsLink)	\
    ( (This)->lpVtbl -> get_IsLink(This,IsLink) ) 

#define BrowseModeLine_get_IsAbbreviation(This,IsAbbreviation)	\
    ( (This)->lpVtbl -> get_IsAbbreviation(This,IsAbbreviation) ) 

#define BrowseModeLine_get_HasAccessKey(This,HasAccessKey)	\
    ( (This)->lpVtbl -> get_HasAccessKey(This,HasAccessKey) ) 

#define BrowseModeLine_get_IsAcronym(This,IsAcronym)	\
    ( (This)->lpVtbl -> get_IsAcronym(This,IsAcronym) ) 

#define BrowseModeLine_get_IsBlank(This,IsBlank)	\
    ( (This)->lpVtbl -> get_IsBlank(This,IsBlank) ) 

#define BrowseModeLine_get_IsGraphic(This,IsGraphic)	\
    ( (This)->lpVtbl -> get_IsGraphic(This,IsGraphic) ) 

#define BrowseModeLine_get_IsHeading(This,IsHeading)	\
    ( (This)->lpVtbl -> get_IsHeading(This,IsHeading) ) 

#define BrowseModeLine_get_HasOnClick(This,HasOnClick)	\
    ( (This)->lpVtbl -> get_HasOnClick(This,HasOnClick) ) 

#define BrowseModeLine_get_IsQuote(This,IsQuote)	\
    ( (This)->lpVtbl -> get_IsQuote(This,IsQuote) ) 

#define BrowseModeLine_get_IsSecure(This,IsSecure)	\
    ( (This)->lpVtbl -> get_IsSecure(This,IsSecure) ) 

#define BrowseModeLine_get_IsSilent(This,IsSilent)	\
    ( (This)->lpVtbl -> get_IsSilent(This,IsSilent) ) 

#define BrowseModeLine_get_Accessible(This,Accessible)	\
    ( (This)->lpVtbl -> get_Accessible(This,Accessible) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrowseModeLine_INTERFACE_DEFINED__ */


#ifndef __WEDialogs_INTERFACE_DEFINED__
#define __WEDialogs_INTERFACE_DEFINED__

/* interface WEDialogs */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WEDialogs;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("A84F8D09-34E2-4408-82CC-3E8E78BF264C")
    WEDialogs : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE AddRemove( 
            /* [optional][in] */ VARIANT PackagePathAndName,
            /* [optional][in] */ VARIANT DeletePackageAfterInstall,
            /* [optional][in] */ VARIANT WaitForCompletion,
            /* [optional][in] */ VARIANT ParentWindow) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE AppUpdate( 
            /* [optional][in] */ VARIANT ParentWindow) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WEDialogsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WEDialogs * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WEDialogs * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WEDialogs * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WEDialogs * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WEDialogs * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WEDialogs * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WEDialogs * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WEDialogs * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WEDialogs * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *AddRemove )( 
            WEDialogs * This,
            /* [optional][in] */ VARIANT PackagePathAndName,
            /* [optional][in] */ VARIANT DeletePackageAfterInstall,
            /* [optional][in] */ VARIANT WaitForCompletion,
            /* [optional][in] */ VARIANT ParentWindow);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *AppUpdate )( 
            WEDialogs * This,
            /* [optional][in] */ VARIANT ParentWindow);
        
        END_INTERFACE
    } WEDialogsVtbl;

    interface WEDialogs
    {
        CONST_VTBL struct WEDialogsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WEDialogs_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WEDialogs_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WEDialogs_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WEDialogs_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WEDialogs_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WEDialogs_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WEDialogs_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WEDialogs_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WEDialogs_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WEDialogs_AddRemove(This,PackagePathAndName,DeletePackageAfterInstall,WaitForCompletion,ParentWindow)	\
    ( (This)->lpVtbl -> AddRemove(This,PackagePathAndName,DeletePackageAfterInstall,WaitForCompletion,ParentWindow) ) 

#define WEDialogs_AppUpdate(This,ParentWindow)	\
    ( (This)->lpVtbl -> AppUpdate(This,ParentWindow) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WEDialogs_INTERFACE_DEFINED__ */


#ifndef ___Braille_INTERFACE_DEFINED__
#define ___Braille_INTERFACE_DEFINED__

/* interface _Braille */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__Braille;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("6B0AFC3B-4428-483B-8801-89D035C6BCC0")
    _Braille : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_PhysicalCellCount( 
            /* [retval][out] */ long *cellCount) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ActiveCellCount( 
            /* [retval][out] */ long *cellCount) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FirstActiveCell( 
            /* [retval][out] */ long *firstActive) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Character( 
            /* [in] */ long index,
            /* [retval][out] */ BSTR *Character) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DotPattern( 
            /* [in] */ long index,
            /* [retval][out] */ BSTR *DotPattern) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Attributes( 
            /* [in] */ long index,
            /* [retval][out] */ BrailleAttributes **Attributes) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Translated( 
            /* [in] */ long index,
            /* [retval][out] */ VARIANT_BOOL *Translated) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Position( 
            /* [in] */ long index,
            /* [retval][out] */ ScreenPoint **Position) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MousePosition( 
            /* [retval][out] */ long *MousePosition) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CursorPosition( 
            /* [retval][out] */ long *CursorPosition) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_QuickMessageDisplayed( 
            /* [retval][out] */ VARIANT_BOOL *quickMessage) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Line( 
            /* [retval][out] */ BSTR *Line) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EntireDisplay( 
            /* [retval][out] */ _Braille **EntireDisplay) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Word( 
            /* [in] */ long index,
            /* [retval][out] */ BSTR *Word) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE WordStart( 
            /* [in] */ long index,
            /* [retval][out] */ long *startIndex) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE WordEnd( 
            /* [in] */ long index,
            /* [retval][out] */ long *endIndex) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE DotPatternToUnicode( 
            /* [in] */ BSTR DotPattern,
            /* [retval][out] */ BSTR *unicode) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Display( 
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT index,
            /* [optional][in] */ VARIANT Location) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE DisplayQuickMessage( 
            /* [in] */ BSTR Text) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Translate( 
            /* [in] */ BSTR Text,
            /* [retval][out] */ BSTR *translatedText) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Freeze( 
            /* [in] */ BrailleFreezeLocation Location,
            /* [in] */ long Count,
            /* [retval][out] */ _Braille **frozen) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Update( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE UnicodeToDotPattern( 
            /* [in] */ BSTR unicodeCharacter,
            /* [retval][out] */ BSTR *DotPattern) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE DotPatternToUnicodeBraille( 
            /* [in] */ BSTR Character,
            /* [retval][out] */ BSTR *unicode) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE StringToUnicodeBraille( 
            /* [in] */ BSTR Text,
            /* [retval][out] */ BSTR *unicode) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_DisplayMode( 
            /* [in] */ BrailleDisplayMode mode) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DisplayMode( 
            /* [retval][out] */ BrailleDisplayMode *mode) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TranslationTables( 
            /* [retval][out] */ BrailleTables **Table) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BrailleWindow( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_BrailleWindow( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SeparatorDotPattern( 
            /* [retval][out] */ BSTR *pattern) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SeparatorDotPattern( 
            /* [in] */ BSTR pattern) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ControlTypes( 
            /* [retval][out] */ BrailleControlTypes **Types) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SpeechBoxModeCharacterLimit( 
            /* [retval][out] */ long *limit) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SpeechBoxModeCharacterLimit( 
            /* [in] */ long limit) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_LineModeIncludeDataIndicators( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LineModeIncludeDataIndicators( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_LineModeLeftIndicatorDotPattern( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LineModeLeftIndicatorDotPattern( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_LineModeRightIndicatorDotPattern( 
            /* [in] */ BSTR prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LineModeRightIndicatorDotPattern( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UseGraphicSeparator( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_UseGraphicSeparator( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DotPatterns( 
            /* [retval][out] */ BrailleDotPatterns **patterns) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Tracking( 
            /* [in] */ BrailleTracking prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Tracking( 
            /* [retval][out] */ BrailleTracking *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _BrailleVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _Braille * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _Braille * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _Braille * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _Braille * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _Braille * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _Braille * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _Braille * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _Braille * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _Braille * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_PhysicalCellCount )( 
            _Braille * This,
            /* [retval][out] */ long *cellCount);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ActiveCellCount )( 
            _Braille * This,
            /* [retval][out] */ long *cellCount);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FirstActiveCell )( 
            _Braille * This,
            /* [retval][out] */ long *firstActive);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Character )( 
            _Braille * This,
            /* [in] */ long index,
            /* [retval][out] */ BSTR *Character);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DotPattern )( 
            _Braille * This,
            /* [in] */ long index,
            /* [retval][out] */ BSTR *DotPattern);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Attributes )( 
            _Braille * This,
            /* [in] */ long index,
            /* [retval][out] */ BrailleAttributes **Attributes);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Translated )( 
            _Braille * This,
            /* [in] */ long index,
            /* [retval][out] */ VARIANT_BOOL *Translated);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            _Braille * This,
            /* [in] */ long index,
            /* [retval][out] */ ScreenPoint **Position);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MousePosition )( 
            _Braille * This,
            /* [retval][out] */ long *MousePosition);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CursorPosition )( 
            _Braille * This,
            /* [retval][out] */ long *CursorPosition);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_QuickMessageDisplayed )( 
            _Braille * This,
            /* [retval][out] */ VARIANT_BOOL *quickMessage);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Line )( 
            _Braille * This,
            /* [retval][out] */ BSTR *Line);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EntireDisplay )( 
            _Braille * This,
            /* [retval][out] */ _Braille **EntireDisplay);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Word )( 
            _Braille * This,
            /* [in] */ long index,
            /* [retval][out] */ BSTR *Word);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *WordStart )( 
            _Braille * This,
            /* [in] */ long index,
            /* [retval][out] */ long *startIndex);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *WordEnd )( 
            _Braille * This,
            /* [in] */ long index,
            /* [retval][out] */ long *endIndex);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *DotPatternToUnicode )( 
            _Braille * This,
            /* [in] */ BSTR DotPattern,
            /* [retval][out] */ BSTR *unicode);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Display )( 
            _Braille * This,
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT index,
            /* [optional][in] */ VARIANT Location);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *DisplayQuickMessage )( 
            _Braille * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Translate )( 
            _Braille * This,
            /* [in] */ BSTR Text,
            /* [retval][out] */ BSTR *translatedText);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Freeze )( 
            _Braille * This,
            /* [in] */ BrailleFreezeLocation Location,
            /* [in] */ long Count,
            /* [retval][out] */ _Braille **frozen);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Update )( 
            _Braille * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *UnicodeToDotPattern )( 
            _Braille * This,
            /* [in] */ BSTR unicodeCharacter,
            /* [retval][out] */ BSTR *DotPattern);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *DotPatternToUnicodeBraille )( 
            _Braille * This,
            /* [in] */ BSTR Character,
            /* [retval][out] */ BSTR *unicode);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *StringToUnicodeBraille )( 
            _Braille * This,
            /* [in] */ BSTR Text,
            /* [retval][out] */ BSTR *unicode);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_DisplayMode )( 
            _Braille * This,
            /* [in] */ BrailleDisplayMode mode);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DisplayMode )( 
            _Braille * This,
            /* [retval][out] */ BrailleDisplayMode *mode);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TranslationTables )( 
            _Braille * This,
            /* [retval][out] */ BrailleTables **Table);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BrailleWindow )( 
            _Braille * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_BrailleWindow )( 
            _Braille * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SeparatorDotPattern )( 
            _Braille * This,
            /* [retval][out] */ BSTR *pattern);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SeparatorDotPattern )( 
            _Braille * This,
            /* [in] */ BSTR pattern);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ControlTypes )( 
            _Braille * This,
            /* [retval][out] */ BrailleControlTypes **Types);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SpeechBoxModeCharacterLimit )( 
            _Braille * This,
            /* [retval][out] */ long *limit);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SpeechBoxModeCharacterLimit )( 
            _Braille * This,
            /* [in] */ long limit);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_LineModeIncludeDataIndicators )( 
            _Braille * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LineModeIncludeDataIndicators )( 
            _Braille * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_LineModeLeftIndicatorDotPattern )( 
            _Braille * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LineModeLeftIndicatorDotPattern )( 
            _Braille * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_LineModeRightIndicatorDotPattern )( 
            _Braille * This,
            /* [in] */ BSTR prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LineModeRightIndicatorDotPattern )( 
            _Braille * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UseGraphicSeparator )( 
            _Braille * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_UseGraphicSeparator )( 
            _Braille * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DotPatterns )( 
            _Braille * This,
            /* [retval][out] */ BrailleDotPatterns **patterns);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Tracking )( 
            _Braille * This,
            /* [in] */ BrailleTracking prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Tracking )( 
            _Braille * This,
            /* [retval][out] */ BrailleTracking *prop);
        
        END_INTERFACE
    } _BrailleVtbl;

    interface _Braille
    {
        CONST_VTBL struct _BrailleVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _Braille_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _Braille_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _Braille_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _Braille_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _Braille_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _Braille_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _Braille_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _Braille_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _Braille_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _Braille_get_PhysicalCellCount(This,cellCount)	\
    ( (This)->lpVtbl -> get_PhysicalCellCount(This,cellCount) ) 

#define _Braille_get_ActiveCellCount(This,cellCount)	\
    ( (This)->lpVtbl -> get_ActiveCellCount(This,cellCount) ) 

#define _Braille_get_FirstActiveCell(This,firstActive)	\
    ( (This)->lpVtbl -> get_FirstActiveCell(This,firstActive) ) 

#define _Braille_get_Character(This,index,Character)	\
    ( (This)->lpVtbl -> get_Character(This,index,Character) ) 

#define _Braille_get_DotPattern(This,index,DotPattern)	\
    ( (This)->lpVtbl -> get_DotPattern(This,index,DotPattern) ) 

#define _Braille_get_Attributes(This,index,Attributes)	\
    ( (This)->lpVtbl -> get_Attributes(This,index,Attributes) ) 

#define _Braille_get_Translated(This,index,Translated)	\
    ( (This)->lpVtbl -> get_Translated(This,index,Translated) ) 

#define _Braille_get_Position(This,index,Position)	\
    ( (This)->lpVtbl -> get_Position(This,index,Position) ) 

#define _Braille_get_MousePosition(This,MousePosition)	\
    ( (This)->lpVtbl -> get_MousePosition(This,MousePosition) ) 

#define _Braille_get_CursorPosition(This,CursorPosition)	\
    ( (This)->lpVtbl -> get_CursorPosition(This,CursorPosition) ) 

#define _Braille_get_QuickMessageDisplayed(This,quickMessage)	\
    ( (This)->lpVtbl -> get_QuickMessageDisplayed(This,quickMessage) ) 

#define _Braille_get_Line(This,Line)	\
    ( (This)->lpVtbl -> get_Line(This,Line) ) 

#define _Braille_get_EntireDisplay(This,EntireDisplay)	\
    ( (This)->lpVtbl -> get_EntireDisplay(This,EntireDisplay) ) 

#define _Braille_Word(This,index,Word)	\
    ( (This)->lpVtbl -> Word(This,index,Word) ) 

#define _Braille_WordStart(This,index,startIndex)	\
    ( (This)->lpVtbl -> WordStart(This,index,startIndex) ) 

#define _Braille_WordEnd(This,index,endIndex)	\
    ( (This)->lpVtbl -> WordEnd(This,index,endIndex) ) 

#define _Braille_DotPatternToUnicode(This,DotPattern,unicode)	\
    ( (This)->lpVtbl -> DotPatternToUnicode(This,DotPattern,unicode) ) 

#define _Braille_Display(This,Text,index,Location)	\
    ( (This)->lpVtbl -> Display(This,Text,index,Location) ) 

#define _Braille_DisplayQuickMessage(This,Text)	\
    ( (This)->lpVtbl -> DisplayQuickMessage(This,Text) ) 

#define _Braille_Translate(This,Text,translatedText)	\
    ( (This)->lpVtbl -> Translate(This,Text,translatedText) ) 

#define _Braille_Freeze(This,Location,Count,frozen)	\
    ( (This)->lpVtbl -> Freeze(This,Location,Count,frozen) ) 

#define _Braille_Update(This)	\
    ( (This)->lpVtbl -> Update(This) ) 

#define _Braille_UnicodeToDotPattern(This,unicodeCharacter,DotPattern)	\
    ( (This)->lpVtbl -> UnicodeToDotPattern(This,unicodeCharacter,DotPattern) ) 

#define _Braille_DotPatternToUnicodeBraille(This,Character,unicode)	\
    ( (This)->lpVtbl -> DotPatternToUnicodeBraille(This,Character,unicode) ) 

#define _Braille_StringToUnicodeBraille(This,Text,unicode)	\
    ( (This)->lpVtbl -> StringToUnicodeBraille(This,Text,unicode) ) 

#define _Braille_put_DisplayMode(This,mode)	\
    ( (This)->lpVtbl -> put_DisplayMode(This,mode) ) 

#define _Braille_get_DisplayMode(This,mode)	\
    ( (This)->lpVtbl -> get_DisplayMode(This,mode) ) 

#define _Braille_get_TranslationTables(This,Table)	\
    ( (This)->lpVtbl -> get_TranslationTables(This,Table) ) 

#define _Braille_get_BrailleWindow(This,prop)	\
    ( (This)->lpVtbl -> get_BrailleWindow(This,prop) ) 

#define _Braille_put_BrailleWindow(This,prop)	\
    ( (This)->lpVtbl -> put_BrailleWindow(This,prop) ) 

#define _Braille_get_SeparatorDotPattern(This,pattern)	\
    ( (This)->lpVtbl -> get_SeparatorDotPattern(This,pattern) ) 

#define _Braille_put_SeparatorDotPattern(This,pattern)	\
    ( (This)->lpVtbl -> put_SeparatorDotPattern(This,pattern) ) 

#define _Braille_get_ControlTypes(This,Types)	\
    ( (This)->lpVtbl -> get_ControlTypes(This,Types) ) 

#define _Braille_get_SpeechBoxModeCharacterLimit(This,limit)	\
    ( (This)->lpVtbl -> get_SpeechBoxModeCharacterLimit(This,limit) ) 

#define _Braille_put_SpeechBoxModeCharacterLimit(This,limit)	\
    ( (This)->lpVtbl -> put_SpeechBoxModeCharacterLimit(This,limit) ) 

#define _Braille_put_LineModeIncludeDataIndicators(This,prop)	\
    ( (This)->lpVtbl -> put_LineModeIncludeDataIndicators(This,prop) ) 

#define _Braille_get_LineModeIncludeDataIndicators(This,prop)	\
    ( (This)->lpVtbl -> get_LineModeIncludeDataIndicators(This,prop) ) 

#define _Braille_put_LineModeLeftIndicatorDotPattern(This,prop)	\
    ( (This)->lpVtbl -> put_LineModeLeftIndicatorDotPattern(This,prop) ) 

#define _Braille_get_LineModeLeftIndicatorDotPattern(This,prop)	\
    ( (This)->lpVtbl -> get_LineModeLeftIndicatorDotPattern(This,prop) ) 

#define _Braille_put_LineModeRightIndicatorDotPattern(This,prop)	\
    ( (This)->lpVtbl -> put_LineModeRightIndicatorDotPattern(This,prop) ) 

#define _Braille_get_LineModeRightIndicatorDotPattern(This,prop)	\
    ( (This)->lpVtbl -> get_LineModeRightIndicatorDotPattern(This,prop) ) 

#define _Braille_get_UseGraphicSeparator(This,prop)	\
    ( (This)->lpVtbl -> get_UseGraphicSeparator(This,prop) ) 

#define _Braille_put_UseGraphicSeparator(This,prop)	\
    ( (This)->lpVtbl -> put_UseGraphicSeparator(This,prop) ) 

#define _Braille_get_DotPatterns(This,patterns)	\
    ( (This)->lpVtbl -> get_DotPatterns(This,patterns) ) 

#define _Braille_put_Tracking(This,prop)	\
    ( (This)->lpVtbl -> put_Tracking(This,prop) ) 

#define _Braille_get_Tracking(This,prop)	\
    ( (This)->lpVtbl -> get_Tracking(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___Braille_INTERFACE_DEFINED__ */


#ifndef __BrailleAttributes_INTERFACE_DEFINED__
#define __BrailleAttributes_INTERFACE_DEFINED__

/* interface BrailleAttributes */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_BrailleAttributes;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("27A1F779-31DB-466E-8A62-57DBB89D3504")
    BrailleAttributes : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ControlType( 
            /* [retval][out] */ BrailleControlTypeEnum *type) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Bold( 
            /* [retval][out] */ VARIANT_BOOL *Bold) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Highlight( 
            /* [retval][out] */ VARIANT_BOOL *Highlight) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Italic( 
            /* [retval][out] */ VARIANT_BOOL *Italic) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Strikeout( 
            /* [retval][out] */ VARIANT_BOOL *Strikeout) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Underline( 
            /* [retval][out] */ VARIANT_BOOL *Underline) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct BrailleAttributesVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleAttributes * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleAttributes * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleAttributes * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleAttributes * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleAttributes * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleAttributes * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleAttributes * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            BrailleAttributes * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            BrailleAttributes * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ControlType )( 
            BrailleAttributes * This,
            /* [retval][out] */ BrailleControlTypeEnum *type);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Bold )( 
            BrailleAttributes * This,
            /* [retval][out] */ VARIANT_BOOL *Bold);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Highlight )( 
            BrailleAttributes * This,
            /* [retval][out] */ VARIANT_BOOL *Highlight);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Italic )( 
            BrailleAttributes * This,
            /* [retval][out] */ VARIANT_BOOL *Italic);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Strikeout )( 
            BrailleAttributes * This,
            /* [retval][out] */ VARIANT_BOOL *Strikeout);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Underline )( 
            BrailleAttributes * This,
            /* [retval][out] */ VARIANT_BOOL *Underline);
        
        END_INTERFACE
    } BrailleAttributesVtbl;

    interface BrailleAttributes
    {
        CONST_VTBL struct BrailleAttributesVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleAttributes_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleAttributes_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleAttributes_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleAttributes_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleAttributes_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleAttributes_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleAttributes_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define BrailleAttributes_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define BrailleAttributes_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define BrailleAttributes_get_ControlType(This,type)	\
    ( (This)->lpVtbl -> get_ControlType(This,type) ) 

#define BrailleAttributes_get_Bold(This,Bold)	\
    ( (This)->lpVtbl -> get_Bold(This,Bold) ) 

#define BrailleAttributes_get_Highlight(This,Highlight)	\
    ( (This)->lpVtbl -> get_Highlight(This,Highlight) ) 

#define BrailleAttributes_get_Italic(This,Italic)	\
    ( (This)->lpVtbl -> get_Italic(This,Italic) ) 

#define BrailleAttributes_get_Strikeout(This,Strikeout)	\
    ( (This)->lpVtbl -> get_Strikeout(This,Strikeout) ) 

#define BrailleAttributes_get_Underline(This,Underline)	\
    ( (This)->lpVtbl -> get_Underline(This,Underline) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __BrailleAttributes_INTERFACE_DEFINED__ */


#ifndef __OSVersion_INTERFACE_DEFINED__
#define __OSVersion_INTERFACE_DEFINED__

/* interface OSVersion */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_OSVersion;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("9A7FD9FF-D729-4B3E-8D60-DE0B034B07FC")
    OSVersion : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MajorVersion( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_MinorVersion( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_BuildNumber( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ServicePackMajor( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ServicePackMinor( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ProductSuite( 
            /* [retval][out] */ OSProductSuiteMask *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ProductType( 
            /* [retval][out] */ OSProductType *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UserLanguage( 
            /* [retval][out] */ long *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SystemLanguage( 
            /* [retval][out] */ long *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct OSVersionVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            OSVersion * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            OSVersion * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            OSVersion * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            OSVersion * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            OSVersion * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            OSVersion * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            OSVersion * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            OSVersion * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            OSVersion * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MajorVersion )( 
            OSVersion * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_MinorVersion )( 
            OSVersion * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_BuildNumber )( 
            OSVersion * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ServicePackMajor )( 
            OSVersion * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ServicePackMinor )( 
            OSVersion * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ProductSuite )( 
            OSVersion * This,
            /* [retval][out] */ OSProductSuiteMask *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ProductType )( 
            OSVersion * This,
            /* [retval][out] */ OSProductType *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UserLanguage )( 
            OSVersion * This,
            /* [retval][out] */ long *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SystemLanguage )( 
            OSVersion * This,
            /* [retval][out] */ long *prop);
        
        END_INTERFACE
    } OSVersionVtbl;

    interface OSVersion
    {
        CONST_VTBL struct OSVersionVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define OSVersion_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define OSVersion_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define OSVersion_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define OSVersion_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define OSVersion_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define OSVersion_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define OSVersion_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define OSVersion_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define OSVersion_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define OSVersion_get_MajorVersion(This,prop)	\
    ( (This)->lpVtbl -> get_MajorVersion(This,prop) ) 

#define OSVersion_get_MinorVersion(This,prop)	\
    ( (This)->lpVtbl -> get_MinorVersion(This,prop) ) 

#define OSVersion_get_BuildNumber(This,prop)	\
    ( (This)->lpVtbl -> get_BuildNumber(This,prop) ) 

#define OSVersion_get_ServicePackMajor(This,prop)	\
    ( (This)->lpVtbl -> get_ServicePackMajor(This,prop) ) 

#define OSVersion_get_ServicePackMinor(This,prop)	\
    ( (This)->lpVtbl -> get_ServicePackMinor(This,prop) ) 

#define OSVersion_get_ProductSuite(This,prop)	\
    ( (This)->lpVtbl -> get_ProductSuite(This,prop) ) 

#define OSVersion_get_ProductType(This,prop)	\
    ( (This)->lpVtbl -> get_ProductType(This,prop) ) 

#define OSVersion_get_UserLanguage(This,prop)	\
    ( (This)->lpVtbl -> get_UserLanguage(This,prop) ) 

#define OSVersion_get_SystemLanguage(This,prop)	\
    ( (This)->lpVtbl -> get_SystemLanguage(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __OSVersion_INTERFACE_DEFINED__ */


#ifndef ___Dialog_INTERFACE_DEFINED__
#define ___Dialog_INTERFACE_DEFINED__

/* interface _Dialog */
/* [object][oleautomation][dual][hidden][helpcontext][helpstring][version][uuid] */ 


EXTERN_C const IID IID__Dialog;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("6465FE90-C00C-47E3-B2DC-278E486477FC")
    _Dialog : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **Application) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Window( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Control( 
            /* [optional][in] */ VARIANT id,
            /* [retval][out] */ IDispatch **Control) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EventInfo( 
            /* [retval][out] */ VARIANT *info) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Close( void) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Menu( 
            /* [retval][out] */ Menu **Menu) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _DialogVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _Dialog * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _Dialog * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _Dialog * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _Dialog * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _Dialog * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _Dialog * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _Dialog * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _Dialog * This,
            /* [retval][out] */ _Application **Application);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _Dialog * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            _Dialog * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Control )( 
            _Dialog * This,
            /* [optional][in] */ VARIANT id,
            /* [retval][out] */ IDispatch **Control);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EventInfo )( 
            _Dialog * This,
            /* [retval][out] */ VARIANT *info);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Close )( 
            _Dialog * This);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Menu )( 
            _Dialog * This,
            /* [retval][out] */ Menu **Menu);
        
        END_INTERFACE
    } _DialogVtbl;

    interface _Dialog
    {
        CONST_VTBL struct _DialogVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _Dialog_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _Dialog_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _Dialog_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _Dialog_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _Dialog_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _Dialog_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _Dialog_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _Dialog_get_Application(This,Application)	\
    ( (This)->lpVtbl -> get_Application(This,Application) ) 

#define _Dialog_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _Dialog_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define _Dialog_get_Control(This,id,Control)	\
    ( (This)->lpVtbl -> get_Control(This,id,Control) ) 

#define _Dialog_get_EventInfo(This,info)	\
    ( (This)->lpVtbl -> get_EventInfo(This,info) ) 

#define _Dialog_Close(This)	\
    ( (This)->lpVtbl -> Close(This) ) 

#define _Dialog_get_Menu(This,Menu)	\
    ( (This)->lpVtbl -> get_Menu(This,Menu) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___Dialog_INTERFACE_DEFINED__ */


#ifndef __Menu_INTERFACE_DEFINED__
#define __Menu_INTERFACE_DEFINED__

/* interface Menu */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Menu;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("76E1A66A-8A64-4A49-8327-8C8F6334E44F")
    Menu : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Checked( 
            /* [in] */ BSTR Item,
            /* [retval][out] */ VARIANT_BOOL *Checked) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Checked( 
            /* [in] */ BSTR Item,
            /* [in] */ VARIANT_BOOL Checked) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Enabled( 
            /* [in] */ BSTR Item,
            /* [retval][out] */ VARIANT_BOOL *Enabled) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Enabled( 
            /* [in] */ BSTR Item,
            /* [in] */ VARIANT_BOOL Enabled) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Grayed( 
            /* [in] */ BSTR Item,
            /* [retval][out] */ VARIANT_BOOL *Grayed) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Grayed( 
            /* [in] */ BSTR Item,
            /* [in] */ VARIANT_BOOL Grayed) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Text( 
            /* [in] */ BSTR Item,
            /* [retval][out] */ BSTR *Text) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Text( 
            /* [in] */ BSTR Item,
            /* [in] */ BSTR Text) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE InsertItem( 
            /* [in] */ BSTR ItemText,
            /* [in] */ MenuItemFlags Flags,
            /* [in] */ BSTR ItemID,
            /* [optional][in] */ VARIANT Position,
            /* [optional][in] */ VARIANT Accelerator) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE InsertMenu( 
            /* [in] */ BSTR ItemText,
            /* [in] */ BSTR ItemID,
            /* [optional][in] */ VARIANT Position,
            /* [retval][out] */ Menu **SubMenu) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Delete( 
            /* [in] */ BSTR Item) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE GetSubMenu( 
            /* [in] */ BSTR subMenuID,
            /* [retval][out] */ Menu **SubMenu) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct MenuVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Menu * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Menu * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Menu * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Menu * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Menu * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Menu * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Menu * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Menu * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Menu * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Checked )( 
            Menu * This,
            /* [in] */ BSTR Item,
            /* [retval][out] */ VARIANT_BOOL *Checked);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Checked )( 
            Menu * This,
            /* [in] */ BSTR Item,
            /* [in] */ VARIANT_BOOL Checked);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Enabled )( 
            Menu * This,
            /* [in] */ BSTR Item,
            /* [retval][out] */ VARIANT_BOOL *Enabled);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Enabled )( 
            Menu * This,
            /* [in] */ BSTR Item,
            /* [in] */ VARIANT_BOOL Enabled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Grayed )( 
            Menu * This,
            /* [in] */ BSTR Item,
            /* [retval][out] */ VARIANT_BOOL *Grayed);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Grayed )( 
            Menu * This,
            /* [in] */ BSTR Item,
            /* [in] */ VARIANT_BOOL Grayed);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            Menu * This,
            /* [in] */ BSTR Item,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            Menu * This,
            /* [in] */ BSTR Item,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *InsertItem )( 
            Menu * This,
            /* [in] */ BSTR ItemText,
            /* [in] */ MenuItemFlags Flags,
            /* [in] */ BSTR ItemID,
            /* [optional][in] */ VARIANT Position,
            /* [optional][in] */ VARIANT Accelerator);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *InsertMenu )( 
            Menu * This,
            /* [in] */ BSTR ItemText,
            /* [in] */ BSTR ItemID,
            /* [optional][in] */ VARIANT Position,
            /* [retval][out] */ Menu **SubMenu);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Delete )( 
            Menu * This,
            /* [in] */ BSTR Item);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *GetSubMenu )( 
            Menu * This,
            /* [in] */ BSTR subMenuID,
            /* [retval][out] */ Menu **SubMenu);
        
        END_INTERFACE
    } MenuVtbl;

    interface Menu
    {
        CONST_VTBL struct MenuVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Menu_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Menu_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Menu_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Menu_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Menu_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Menu_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Menu_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Menu_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Menu_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Menu_get_Checked(This,Item,Checked)	\
    ( (This)->lpVtbl -> get_Checked(This,Item,Checked) ) 

#define Menu_put_Checked(This,Item,Checked)	\
    ( (This)->lpVtbl -> put_Checked(This,Item,Checked) ) 

#define Menu_get_Enabled(This,Item,Enabled)	\
    ( (This)->lpVtbl -> get_Enabled(This,Item,Enabled) ) 

#define Menu_put_Enabled(This,Item,Enabled)	\
    ( (This)->lpVtbl -> put_Enabled(This,Item,Enabled) ) 

#define Menu_get_Grayed(This,Item,Grayed)	\
    ( (This)->lpVtbl -> get_Grayed(This,Item,Grayed) ) 

#define Menu_put_Grayed(This,Item,Grayed)	\
    ( (This)->lpVtbl -> put_Grayed(This,Item,Grayed) ) 

#define Menu_get_Text(This,Item,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Item,Text) ) 

#define Menu_put_Text(This,Item,Text)	\
    ( (This)->lpVtbl -> put_Text(This,Item,Text) ) 

#define Menu_InsertItem(This,ItemText,Flags,ItemID,Position,Accelerator)	\
    ( (This)->lpVtbl -> InsertItem(This,ItemText,Flags,ItemID,Position,Accelerator) ) 

#define Menu_InsertMenu(This,ItemText,ItemID,Position,SubMenu)	\
    ( (This)->lpVtbl -> InsertMenu(This,ItemText,ItemID,Position,SubMenu) ) 

#define Menu_Delete(This,Item)	\
    ( (This)->lpVtbl -> Delete(This,Item) ) 

#define Menu_GetSubMenu(This,subMenuID,SubMenu)	\
    ( (This)->lpVtbl -> GetSubMenu(This,subMenuID,SubMenu) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Menu_INTERFACE_DEFINED__ */


#ifndef __StartupOptions_INTERFACE_DEFINED__
#define __StartupOptions_INTERFACE_DEFINED__

/* interface StartupOptions */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_StartupOptions;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("3841E91D-3759-4C7F-BD7D-0AF829F07E54")
    StartupOptions : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_RunAtLoginScreen( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_RunAfterLoginScreen( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_RunAfterLoginCurrentUser( 
            /* [retval][out] */ StartupOptionsCurrentUser *prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UnMute( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_UnMute( 
            /* [in] */ VARIANT_BOOL prop) = 0;
        
        virtual /* [helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_StartupMessage( 
            /* [retval][out] */ StartupMessages *prop) = 0;
        
        virtual /* [helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_StartupMessage( 
            /* [in] */ StartupMessages prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct StartupOptionsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            StartupOptions * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            StartupOptions * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            StartupOptions * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            StartupOptions * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            StartupOptions * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            StartupOptions * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            StartupOptions * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            StartupOptions * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            StartupOptions * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_RunAtLoginScreen )( 
            StartupOptions * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_RunAfterLoginScreen )( 
            StartupOptions * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_RunAfterLoginCurrentUser )( 
            StartupOptions * This,
            /* [retval][out] */ StartupOptionsCurrentUser *prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UnMute )( 
            StartupOptions * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_UnMute )( 
            StartupOptions * This,
            /* [in] */ VARIANT_BOOL prop);
        
        /* [helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_StartupMessage )( 
            StartupOptions * This,
            /* [retval][out] */ StartupMessages *prop);
        
        /* [helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_StartupMessage )( 
            StartupOptions * This,
            /* [in] */ StartupMessages prop);
        
        END_INTERFACE
    } StartupOptionsVtbl;

    interface StartupOptions
    {
        CONST_VTBL struct StartupOptionsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define StartupOptions_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define StartupOptions_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define StartupOptions_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define StartupOptions_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define StartupOptions_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define StartupOptions_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define StartupOptions_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define StartupOptions_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define StartupOptions_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define StartupOptions_get_RunAtLoginScreen(This,prop)	\
    ( (This)->lpVtbl -> get_RunAtLoginScreen(This,prop) ) 

#define StartupOptions_get_RunAfterLoginScreen(This,prop)	\
    ( (This)->lpVtbl -> get_RunAfterLoginScreen(This,prop) ) 

#define StartupOptions_get_RunAfterLoginCurrentUser(This,prop)	\
    ( (This)->lpVtbl -> get_RunAfterLoginCurrentUser(This,prop) ) 

#define StartupOptions_get_UnMute(This,prop)	\
    ( (This)->lpVtbl -> get_UnMute(This,prop) ) 

#define StartupOptions_put_UnMute(This,prop)	\
    ( (This)->lpVtbl -> put_UnMute(This,prop) ) 

#define StartupOptions_get_StartupMessage(This,prop)	\
    ( (This)->lpVtbl -> get_StartupMessage(This,prop) ) 

#define StartupOptions_put_StartupMessage(This,prop)	\
    ( (This)->lpVtbl -> put_StartupMessage(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __StartupOptions_INTERFACE_DEFINED__ */


#ifndef __WEPM_INTERFACE_DEFINED__
#define __WEPM_INTERFACE_DEFINED__

/* interface WEPM */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WEPM;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("22FCD0A5-7CF1-46B1-BE14-01CFE86E55C0")
    WEPM : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Name( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Version( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_StopPrompt( 
            /* [retval][out] */ VARIANT_BOOL *prop) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WEPMVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WEPM * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WEPM * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WEPM * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WEPM * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WEPM * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WEPM * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WEPM * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WEPM * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WEPM * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Name )( 
            WEPM * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Version )( 
            WEPM * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_StopPrompt )( 
            WEPM * This,
            /* [retval][out] */ VARIANT_BOOL *prop);
        
        END_INTERFACE
    } WEPMVtbl;

    interface WEPM
    {
        CONST_VTBL struct WEPMVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WEPM_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WEPM_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WEPM_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WEPM_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WEPM_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WEPM_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WEPM_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WEPM_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WEPM_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WEPM_get_Name(This,prop)	\
    ( (This)->lpVtbl -> get_Name(This,prop) ) 

#define WEPM_get_Version(This,prop)	\
    ( (This)->lpVtbl -> get_Version(This,prop) ) 

#define WEPM_get_StopPrompt(This,prop)	\
    ( (This)->lpVtbl -> get_StopPrompt(This,prop) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WEPM_INTERFACE_DEFINED__ */


#ifndef __Profiles_INTERFACE_DEFINED__
#define __Profiles_INTERFACE_DEFINED__

/* interface Profiles */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Profiles;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("B8488AFB-0E95-409D-AD01-CA1410E5EA50")
    Profiles : public IDispatch
    {
    public:
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Active( 
            /* [retval][out] */ Profile **Profile) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ VARIANT index,
            /* [retval][out] */ Profile **Profile) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ProfilesVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Profiles * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Profiles * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Profiles * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Profiles * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Profiles * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Profiles * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Profiles * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            Profiles * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Profiles * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Profiles * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            Profiles * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Active )( 
            Profiles * This,
            /* [retval][out] */ Profile **Profile);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            Profiles * This,
            /* [in] */ VARIANT index,
            /* [retval][out] */ Profile **Profile);
        
        END_INTERFACE
    } ProfilesVtbl;

    interface Profiles
    {
        CONST_VTBL struct ProfilesVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Profiles_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Profiles_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Profiles_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Profiles_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Profiles_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Profiles_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Profiles_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Profiles_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define Profiles_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Profiles_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Profiles_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define Profiles_get_Active(This,Profile)	\
    ( (This)->lpVtbl -> get_Active(This,Profile) ) 

#define Profiles_Item(This,index,Profile)	\
    ( (This)->lpVtbl -> Item(This,index,Profile) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Profiles_INTERFACE_DEFINED__ */


#ifndef __Profile_INTERFACE_DEFINED__
#define __Profile_INTERFACE_DEFINED__

/* interface Profile */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Profile;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("2F930915-3D94-47A1-B55E-D00C92BCC7E1")
    Profile : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Name( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Folder( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_path( 
            /* [retval][out] */ BSTR *prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_type( 
            /* [retval][out] */ ProfileType *prop) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Activate( void) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ProfileVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Profile * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Profile * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Profile * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Profile * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Profile * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Profile * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Profile * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Profile * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Profile * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Name )( 
            Profile * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Folder )( 
            Profile * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_path )( 
            Profile * This,
            /* [retval][out] */ BSTR *prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_type )( 
            Profile * This,
            /* [retval][out] */ ProfileType *prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Activate )( 
            Profile * This);
        
        END_INTERFACE
    } ProfileVtbl;

    interface Profile
    {
        CONST_VTBL struct ProfileVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Profile_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Profile_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Profile_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Profile_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Profile_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Profile_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Profile_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Profile_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Profile_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Profile_get_Name(This,prop)	\
    ( (This)->lpVtbl -> get_Name(This,prop) ) 

#define Profile_get_Folder(This,prop)	\
    ( (This)->lpVtbl -> get_Folder(This,prop) ) 

#define Profile_get_path(This,prop)	\
    ( (This)->lpVtbl -> get_path(This,prop) ) 

#define Profile_get_type(This,prop)	\
    ( (This)->lpVtbl -> get_type(This,prop) ) 

#define Profile_Activate(This)	\
    ( (This)->lpVtbl -> Activate(This) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Profile_INTERFACE_DEFINED__ */


#ifndef __Utilities_INTERFACE_DEFINED__
#define __Utilities_INTERFACE_DEFINED__

/* interface Utilities */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Utilities;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("9E254D3E-8463-40AD-9F09-FC9571053B66")
    Utilities : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE LocaleInfo( 
            /* [in] */ VARIANT LocaleID,
            /* [optional][in] */ VARIANT LCID_SORT,
            /* [retval][out] */ LocaleInformation **LocaleInfo) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE LoadedScripts( 
            /* [retval][out] */ LoadedScripts **LoadedScripts) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ConvertLanguageIDToString( 
            /* [in] */ long __MIDL__Utilities0000,
            /* [retval][out] */ BSTR *String) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ConvertStringToLanguageID( 
            /* [in] */ BSTR String,
            /* [retval][out] */ long *id) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE typeName( 
            /* [in] */ IDispatch *Object,
            /* [retval][out] */ BSTR *typeName) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct UtilitiesVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Utilities * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Utilities * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Utilities * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Utilities * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Utilities * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Utilities * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Utilities * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Utilities * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Utilities * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *LocaleInfo )( 
            Utilities * This,
            /* [in] */ VARIANT LocaleID,
            /* [optional][in] */ VARIANT LCID_SORT,
            /* [retval][out] */ LocaleInformation **LocaleInfo);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *LoadedScripts )( 
            Utilities * This,
            /* [retval][out] */ LoadedScripts **LoadedScripts);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ConvertLanguageIDToString )( 
            Utilities * This,
            /* [in] */ long __MIDL__Utilities0000,
            /* [retval][out] */ BSTR *String);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ConvertStringToLanguageID )( 
            Utilities * This,
            /* [in] */ BSTR String,
            /* [retval][out] */ long *id);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *typeName )( 
            Utilities * This,
            /* [in] */ IDispatch *Object,
            /* [retval][out] */ BSTR *typeName);
        
        END_INTERFACE
    } UtilitiesVtbl;

    interface Utilities
    {
        CONST_VTBL struct UtilitiesVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Utilities_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Utilities_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Utilities_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Utilities_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Utilities_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Utilities_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Utilities_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Utilities_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Utilities_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Utilities_LocaleInfo(This,LocaleID,LCID_SORT,LocaleInfo)	\
    ( (This)->lpVtbl -> LocaleInfo(This,LocaleID,LCID_SORT,LocaleInfo) ) 

#define Utilities_LoadedScripts(This,LoadedScripts)	\
    ( (This)->lpVtbl -> LoadedScripts(This,LoadedScripts) ) 

#define Utilities_ConvertLanguageIDToString(This,__MIDL__Utilities0000,String)	\
    ( (This)->lpVtbl -> ConvertLanguageIDToString(This,__MIDL__Utilities0000,String) ) 

#define Utilities_ConvertStringToLanguageID(This,String,id)	\
    ( (This)->lpVtbl -> ConvertStringToLanguageID(This,String,id) ) 

#define Utilities_typeName(This,Object,typeName)	\
    ( (This)->lpVtbl -> typeName(This,Object,typeName) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Utilities_INTERFACE_DEFINED__ */


#ifndef __LocaleInformation_INTERFACE_DEFINED__
#define __LocaleInformation_INTERFACE_DEFINED__

/* interface LocaleInformation */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_LocaleInformation;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("D0638E9F-6409-4D32-ADDD-2CA5E0A07F56")
    LocaleInformation : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LanguageName( 
            /* [retval][out] */ BSTR *Name) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LocaleName( 
            /* [retval][out] */ BSTR *Country) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_RegionName( 
            /* [retval][out] */ BSTR *Name) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct LocaleInformationVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            LocaleInformation * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            LocaleInformation * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            LocaleInformation * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            LocaleInformation * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            LocaleInformation * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            LocaleInformation * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            LocaleInformation * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            LocaleInformation * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            LocaleInformation * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LanguageName )( 
            LocaleInformation * This,
            /* [retval][out] */ BSTR *Name);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LocaleName )( 
            LocaleInformation * This,
            /* [retval][out] */ BSTR *Country);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_RegionName )( 
            LocaleInformation * This,
            /* [retval][out] */ BSTR *Name);
        
        END_INTERFACE
    } LocaleInformationVtbl;

    interface LocaleInformation
    {
        CONST_VTBL struct LocaleInformationVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define LocaleInformation_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define LocaleInformation_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define LocaleInformation_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define LocaleInformation_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define LocaleInformation_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define LocaleInformation_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define LocaleInformation_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define LocaleInformation_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define LocaleInformation_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define LocaleInformation_get_LanguageName(This,Name)	\
    ( (This)->lpVtbl -> get_LanguageName(This,Name) ) 

#define LocaleInformation_get_LocaleName(This,Country)	\
    ( (This)->lpVtbl -> get_LocaleName(This,Country) ) 

#define LocaleInformation_get_RegionName(This,Name)	\
    ( (This)->lpVtbl -> get_RegionName(This,Name) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __LocaleInformation_INTERFACE_DEFINED__ */


#ifndef __LoadedScripts_INTERFACE_DEFINED__
#define __LoadedScripts_INTERFACE_DEFINED__

/* interface LoadedScripts */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_LoadedScripts;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("7FB8C7BB-7424-440D-B48F-07EDA20EF2B9")
    LoadedScripts : public IDispatch
    {
    public:
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ LoadedScript **LoadedScript) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE FilterByName( 
            /* [in] */ BSTR ScriptName,
            /* [optional][in] */ VARIANT FilterMatch,
            /* [retval][out] */ LoadedScripts **LoadedScripts) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct LoadedScriptsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            LoadedScripts * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            LoadedScripts * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            LoadedScripts * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            LoadedScripts * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            LoadedScripts * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            LoadedScripts * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            LoadedScripts * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            LoadedScripts * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            LoadedScripts * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            LoadedScripts * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            LoadedScripts * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            LoadedScripts * This,
            /* [in] */ long index,
            /* [retval][out] */ LoadedScript **LoadedScript);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *FilterByName )( 
            LoadedScripts * This,
            /* [in] */ BSTR ScriptName,
            /* [optional][in] */ VARIANT FilterMatch,
            /* [retval][out] */ LoadedScripts **LoadedScripts);
        
        END_INTERFACE
    } LoadedScriptsVtbl;

    interface LoadedScripts
    {
        CONST_VTBL struct LoadedScriptsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define LoadedScripts_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define LoadedScripts_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define LoadedScripts_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define LoadedScripts_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define LoadedScripts_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define LoadedScripts_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define LoadedScripts_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define LoadedScripts_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define LoadedScripts_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define LoadedScripts_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define LoadedScripts_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define LoadedScripts_Item(This,index,LoadedScript)	\
    ( (This)->lpVtbl -> Item(This,index,LoadedScript) ) 

#define LoadedScripts_FilterByName(This,ScriptName,FilterMatch,LoadedScripts)	\
    ( (This)->lpVtbl -> FilterByName(This,ScriptName,FilterMatch,LoadedScripts) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __LoadedScripts_INTERFACE_DEFINED__ */


#ifndef __WindowMessage_INTERFACE_DEFINED__
#define __WindowMessage_INTERFACE_DEFINED__

/* interface WindowMessage */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_WindowMessage;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("C8D9D673-EE50-481D-A83B-2D7B9A39C40C")
    WindowMessage : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Detach( void) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct WindowMessageVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WindowMessage * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WindowMessage * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WindowMessage * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WindowMessage * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WindowMessage * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WindowMessage * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WindowMessage * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            WindowMessage * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            WindowMessage * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Detach )( 
            WindowMessage * This);
        
        END_INTERFACE
    } WindowMessageVtbl;

    interface WindowMessage
    {
        CONST_VTBL struct WindowMessageVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WindowMessage_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WindowMessage_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WindowMessage_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WindowMessage_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WindowMessage_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WindowMessage_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WindowMessage_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define WindowMessage_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define WindowMessage_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define WindowMessage_Detach(This)	\
    ( (This)->lpVtbl -> Detach(This) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __WindowMessage_INTERFACE_DEFINED__ */


#ifndef ___WindowEvents_INTERFACE_DEFINED__
#define ___WindowEvents_INTERFACE_DEFINED__

/* interface _WindowEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID IID__WindowEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("17369702-DCC0-4588-883D-F77D1BDC3868")
    _WindowEvents
    {
    public:
        BEGIN_INTERFACE
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnChildDeactivate( 
            /* [in] */ _Window *ChildWindow) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnFocus( 
            /* [in] */ _Window *Window) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnChildFocus( 
            /* [in] */ _Window *ChildWindow) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnBlur( 
            /* [in] */ _Window *Window) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnChildBlur( 
            /* [in] */ _Window *ChildWindow) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnClose( 
            /* [in] */ long Handle) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnChildClose( 
            /* [in] */ long Handle) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnChildCreate( 
            /* [in] */ _Window *ChildWindow) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnFieldName( 
            /* [in] */ _Window *Window,
            /* [retval][out] */ BSTR *FieldName) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnChildFieldName( 
            /* [in] */ _Window *Window,
            /* [retval][out] */ BSTR *FieldName) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnFieldData( 
            /* [in] */ _Window *Window,
            /* [retval][out] */ BSTR *FieldData) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnChildFieldData( 
            /* [in] */ _Window *Window,
            /* [retval][out] */ BSTR *FieldData) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnSummary( 
            /* [in] */ _Window *Window,
            /* [retval][out] */ BSTR *FieldData) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnChildSummary( 
            /* [in] */ _Window *Window,
            /* [retval][out] */ BSTR *FieldData) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnSelection( 
            /* [in] */ _Window *Window,
            /* [retval][out] */ BSTR *Selection) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnChildSelection( 
            /* [in] */ _Window *Window,
            /* [retval][out] */ BSTR *Selection) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnBeforeSelect( 
            /* [in] */ _Window *Window,
            /* [retval][out] */ VARIANT_BOOL *saved) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnAfterSelect( 
            /* [in] */ _Window *Window,
            /* [out] */ BSTR *Selected,
            /* [out] */ BSTR *deselected,
            /* [retval][out] */ VARIANT_BOOL *valid) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnBeforeChildSelect( 
            /* [in] */ _Window *Window,
            /* [retval][out] */ VARIANT_BOOL *saved) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnAfterChildSelect( 
            /* [in] */ _Window *Window,
            /* [out] */ BSTR *Selected,
            /* [out] */ BSTR *deselected,
            /* [retval][out] */ VARIANT_BOOL *valid) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnClipRendered( 
            /* [in] */ Clip *Clip) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnChildClipRendered( 
            /* [in] */ Clip *Clip) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnMDIActivate( 
            /* [in] */ _Window *Window) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnMDIDeactivate( 
            /* [in] */ _Window *Window) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnMDIChildActivate( 
            /* [in] */ _Window *Window) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnMDIChildDeactivate( 
            /* [in] */ _Window *Window) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnMenuOpened( 
            /* [in] */ VARIANT_BOOL IsContextMenu) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnMenuClosed( void) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnMenuSelection( 
            BSTR Text,
            BSTR Shortcut,
            BSTR Accelerator,
            MenuFlags Flags) = 0;
        
        END_INTERFACE
    };
    
    
#else 	/* C style interface */

    typedef struct _WindowEventsVtbl
    {
        BEGIN_INTERFACE
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnChildDeactivate )( 
            _WindowEvents * This,
            /* [in] */ _Window *ChildWindow);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnFocus )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnChildFocus )( 
            _WindowEvents * This,
            /* [in] */ _Window *ChildWindow);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnBlur )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnChildBlur )( 
            _WindowEvents * This,
            /* [in] */ _Window *ChildWindow);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnClose )( 
            _WindowEvents * This,
            /* [in] */ long Handle);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnChildClose )( 
            _WindowEvents * This,
            /* [in] */ long Handle);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnChildCreate )( 
            _WindowEvents * This,
            /* [in] */ _Window *ChildWindow);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnFieldName )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window,
            /* [retval][out] */ BSTR *FieldName);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnChildFieldName )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window,
            /* [retval][out] */ BSTR *FieldName);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnFieldData )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window,
            /* [retval][out] */ BSTR *FieldData);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnChildFieldData )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window,
            /* [retval][out] */ BSTR *FieldData);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnSummary )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window,
            /* [retval][out] */ BSTR *FieldData);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnChildSummary )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window,
            /* [retval][out] */ BSTR *FieldData);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnSelection )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window,
            /* [retval][out] */ BSTR *Selection);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnChildSelection )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window,
            /* [retval][out] */ BSTR *Selection);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnBeforeSelect )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window,
            /* [retval][out] */ VARIANT_BOOL *saved);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnAfterSelect )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window,
            /* [out] */ BSTR *Selected,
            /* [out] */ BSTR *deselected,
            /* [retval][out] */ VARIANT_BOOL *valid);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnBeforeChildSelect )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window,
            /* [retval][out] */ VARIANT_BOOL *saved);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnAfterChildSelect )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window,
            /* [out] */ BSTR *Selected,
            /* [out] */ BSTR *deselected,
            /* [retval][out] */ VARIANT_BOOL *valid);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnClipRendered )( 
            _WindowEvents * This,
            /* [in] */ Clip *Clip);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnChildClipRendered )( 
            _WindowEvents * This,
            /* [in] */ Clip *Clip);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnMDIActivate )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnMDIDeactivate )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnMDIChildActivate )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnMDIChildDeactivate )( 
            _WindowEvents * This,
            /* [in] */ _Window *Window);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnMenuOpened )( 
            _WindowEvents * This,
            /* [in] */ VARIANT_BOOL IsContextMenu);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnMenuClosed )( 
            _WindowEvents * This);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnMenuSelection )( 
            _WindowEvents * This,
            BSTR Text,
            BSTR Shortcut,
            BSTR Accelerator,
            MenuFlags Flags);
        
        END_INTERFACE
    } _WindowEventsVtbl;

    interface _WindowEvents
    {
        CONST_VTBL struct _WindowEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _WindowEvents_OnChildDeactivate(This,ChildWindow)	\
    ( (This)->lpVtbl -> OnChildDeactivate(This,ChildWindow) ) 

#define _WindowEvents_OnFocus(This,Window)	\
    ( (This)->lpVtbl -> OnFocus(This,Window) ) 

#define _WindowEvents_OnChildFocus(This,ChildWindow)	\
    ( (This)->lpVtbl -> OnChildFocus(This,ChildWindow) ) 

#define _WindowEvents_OnBlur(This,Window)	\
    ( (This)->lpVtbl -> OnBlur(This,Window) ) 

#define _WindowEvents_OnChildBlur(This,ChildWindow)	\
    ( (This)->lpVtbl -> OnChildBlur(This,ChildWindow) ) 

#define _WindowEvents_OnClose(This,Handle)	\
    ( (This)->lpVtbl -> OnClose(This,Handle) ) 

#define _WindowEvents_OnChildClose(This,Handle)	\
    ( (This)->lpVtbl -> OnChildClose(This,Handle) ) 

#define _WindowEvents_OnChildCreate(This,ChildWindow)	\
    ( (This)->lpVtbl -> OnChildCreate(This,ChildWindow) ) 

#define _WindowEvents_OnFieldName(This,Window,FieldName)	\
    ( (This)->lpVtbl -> OnFieldName(This,Window,FieldName) ) 

#define _WindowEvents_OnChildFieldName(This,Window,FieldName)	\
    ( (This)->lpVtbl -> OnChildFieldName(This,Window,FieldName) ) 

#define _WindowEvents_OnFieldData(This,Window,FieldData)	\
    ( (This)->lpVtbl -> OnFieldData(This,Window,FieldData) ) 

#define _WindowEvents_OnChildFieldData(This,Window,FieldData)	\
    ( (This)->lpVtbl -> OnChildFieldData(This,Window,FieldData) ) 

#define _WindowEvents_OnSummary(This,Window,FieldData)	\
    ( (This)->lpVtbl -> OnSummary(This,Window,FieldData) ) 

#define _WindowEvents_OnChildSummary(This,Window,FieldData)	\
    ( (This)->lpVtbl -> OnChildSummary(This,Window,FieldData) ) 

#define _WindowEvents_OnSelection(This,Window,Selection)	\
    ( (This)->lpVtbl -> OnSelection(This,Window,Selection) ) 

#define _WindowEvents_OnChildSelection(This,Window,Selection)	\
    ( (This)->lpVtbl -> OnChildSelection(This,Window,Selection) ) 

#define _WindowEvents_OnBeforeSelect(This,Window,saved)	\
    ( (This)->lpVtbl -> OnBeforeSelect(This,Window,saved) ) 

#define _WindowEvents_OnAfterSelect(This,Window,Selected,deselected,valid)	\
    ( (This)->lpVtbl -> OnAfterSelect(This,Window,Selected,deselected,valid) ) 

#define _WindowEvents_OnBeforeChildSelect(This,Window,saved)	\
    ( (This)->lpVtbl -> OnBeforeChildSelect(This,Window,saved) ) 

#define _WindowEvents_OnAfterChildSelect(This,Window,Selected,deselected,valid)	\
    ( (This)->lpVtbl -> OnAfterChildSelect(This,Window,Selected,deselected,valid) ) 

#define _WindowEvents_OnClipRendered(This,Clip)	\
    ( (This)->lpVtbl -> OnClipRendered(This,Clip) ) 

#define _WindowEvents_OnChildClipRendered(This,Clip)	\
    ( (This)->lpVtbl -> OnChildClipRendered(This,Clip) ) 

#define _WindowEvents_OnMDIActivate(This,Window)	\
    ( (This)->lpVtbl -> OnMDIActivate(This,Window) ) 

#define _WindowEvents_OnMDIDeactivate(This,Window)	\
    ( (This)->lpVtbl -> OnMDIDeactivate(This,Window) ) 

#define _WindowEvents_OnMDIChildActivate(This,Window)	\
    ( (This)->lpVtbl -> OnMDIChildActivate(This,Window) ) 

#define _WindowEvents_OnMDIChildDeactivate(This,Window)	\
    ( (This)->lpVtbl -> OnMDIChildDeactivate(This,Window) ) 

#define _WindowEvents_OnMenuOpened(This,IsContextMenu)	\
    ( (This)->lpVtbl -> OnMenuOpened(This,IsContextMenu) ) 

#define _WindowEvents_OnMenuClosed(This)	\
    ( (This)->lpVtbl -> OnMenuClosed(This) ) 

#define _WindowEvents_OnMenuSelection(This,Text,Shortcut,Accelerator,Flags)	\
    ( (This)->lpVtbl -> OnMenuSelection(This,Text,Shortcut,Accelerator,Flags) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___WindowEvents_INTERFACE_DEFINED__ */


#ifndef ___Control_INTERFACE_DEFINED__
#define ___Control_INTERFACE_DEFINED__

/* interface _Control */
/* [object][oleautomation][dual][hidden][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__Control;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("24D22926-0C9C-4021-8152-4960FDD790B5")
    _Control : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Text( 
            /* [retval][out] */ BSTR *Text) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Text( 
            /* [in] */ BSTR Text) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Position( 
            /* [retval][out] */ ClientPoint **Position) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Position( 
            /* [in] */ ClientPoint *Position) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Enabled( 
            /* [retval][out] */ VARIANT_BOOL *Enabled) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Enabled( 
            /* [in] */ VARIANT_BOOL Enabled) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Visible( 
            /* [retval][out] */ VARIANT_BOOL *Visible) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Visible( 
            /* [in] */ VARIANT_BOOL Visible) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Focused( 
            /* [retval][out] */ VARIANT_BOOL *Focused) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Width( 
            /* [retval][out] */ long *Width) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Width( 
            /* [in] */ long Width) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Height( 
            /* [retval][out] */ long *Height) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Height( 
            /* [in] */ long Height) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TabOrder( 
            /* [retval][out] */ long *TabOrder) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TabOrder( 
            /* [in] */ long TabOrder) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TabStop( 
            /* [retval][out] */ VARIANT_BOOL *TabStop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TabStop( 
            /* [in] */ VARIANT_BOOL TabStop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Group( 
            /* [retval][out] */ VARIANT_BOOL *Group) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Group( 
            /* [in] */ VARIANT_BOOL Group) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Window( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_id( 
            /* [retval][out] */ long *id) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Next( 
            /* [retval][out] */ IDispatch **prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Previous( 
            /* [retval][out] */ IDispatch **prop) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Focus( void) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_id( 
            /* [in] */ long id) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _ControlVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _Control * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _Control * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _Control * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _Control * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _Control * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _Control * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _Control * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            _Control * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            _Control * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _Control * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _Control * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            _Control * This,
            /* [retval][out] */ ClientPoint **Position);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            _Control * This,
            /* [in] */ ClientPoint *Position);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Enabled )( 
            _Control * This,
            /* [retval][out] */ VARIANT_BOOL *Enabled);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Enabled )( 
            _Control * This,
            /* [in] */ VARIANT_BOOL Enabled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            _Control * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Visible )( 
            _Control * This,
            /* [in] */ VARIANT_BOOL Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            _Control * This,
            /* [retval][out] */ VARIANT_BOOL *Focused);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Width )( 
            _Control * This,
            /* [retval][out] */ long *Width);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Width )( 
            _Control * This,
            /* [in] */ long Width);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Height )( 
            _Control * This,
            /* [retval][out] */ long *Height);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Height )( 
            _Control * This,
            /* [in] */ long Height);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabOrder )( 
            _Control * This,
            /* [retval][out] */ long *TabOrder);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabOrder )( 
            _Control * This,
            /* [in] */ long TabOrder);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabStop )( 
            _Control * This,
            /* [retval][out] */ VARIANT_BOOL *TabStop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabStop )( 
            _Control * This,
            /* [in] */ VARIANT_BOOL TabStop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Group )( 
            _Control * This,
            /* [retval][out] */ VARIANT_BOOL *Group);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Group )( 
            _Control * This,
            /* [in] */ VARIANT_BOOL Group);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            _Control * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            _Control * This,
            /* [retval][out] */ long *id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Next )( 
            _Control * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Previous )( 
            _Control * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Focus )( 
            _Control * This);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_id )( 
            _Control * This,
            /* [in] */ long id);
        
        END_INTERFACE
    } _ControlVtbl;

    interface _Control
    {
        CONST_VTBL struct _ControlVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _Control_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _Control_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _Control_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _Control_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _Control_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _Control_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _Control_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _Control_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define _Control_put_Text(This,Text)	\
    ( (This)->lpVtbl -> put_Text(This,Text) ) 

#define _Control_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _Control_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _Control_get_Position(This,Position)	\
    ( (This)->lpVtbl -> get_Position(This,Position) ) 

#define _Control_put_Position(This,Position)	\
    ( (This)->lpVtbl -> put_Position(This,Position) ) 

#define _Control_get_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> get_Enabled(This,Enabled) ) 

#define _Control_put_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> put_Enabled(This,Enabled) ) 

#define _Control_get_Visible(This,Visible)	\
    ( (This)->lpVtbl -> get_Visible(This,Visible) ) 

#define _Control_put_Visible(This,Visible)	\
    ( (This)->lpVtbl -> put_Visible(This,Visible) ) 

#define _Control_get_Focused(This,Focused)	\
    ( (This)->lpVtbl -> get_Focused(This,Focused) ) 

#define _Control_get_Width(This,Width)	\
    ( (This)->lpVtbl -> get_Width(This,Width) ) 

#define _Control_put_Width(This,Width)	\
    ( (This)->lpVtbl -> put_Width(This,Width) ) 

#define _Control_get_Height(This,Height)	\
    ( (This)->lpVtbl -> get_Height(This,Height) ) 

#define _Control_put_Height(This,Height)	\
    ( (This)->lpVtbl -> put_Height(This,Height) ) 

#define _Control_get_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> get_TabOrder(This,TabOrder) ) 

#define _Control_put_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> put_TabOrder(This,TabOrder) ) 

#define _Control_get_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> get_TabStop(This,TabStop) ) 

#define _Control_put_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> put_TabStop(This,TabStop) ) 

#define _Control_get_Group(This,Group)	\
    ( (This)->lpVtbl -> get_Group(This,Group) ) 

#define _Control_put_Group(This,Group)	\
    ( (This)->lpVtbl -> put_Group(This,Group) ) 

#define _Control_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define _Control_get_id(This,id)	\
    ( (This)->lpVtbl -> get_id(This,id) ) 

#define _Control_get_Next(This,prop)	\
    ( (This)->lpVtbl -> get_Next(This,prop) ) 

#define _Control_get_Previous(This,prop)	\
    ( (This)->lpVtbl -> get_Previous(This,prop) ) 

#define _Control_Focus(This)	\
    ( (This)->lpVtbl -> Focus(This) ) 

#define _Control_put_id(This,id)	\
    ( (This)->lpVtbl -> put_id(This,id) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___Control_INTERFACE_DEFINED__ */


#ifndef __TreeView_INTERFACE_DEFINED__
#define __TreeView_INTERFACE_DEFINED__

/* interface TreeView */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_TreeView;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("6EA73B9F-A405-4CC4-A1B2-A5FF7CA0D65B")
    TreeView : public _Control
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Checkboxes( 
            /* [retval][out] */ VARIANT_BOOL *Checkboxes) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indent( 
            /* [retval][out] */ long *Indent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indent( 
            /* [in] */ long Indent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Selected( 
            /* [retval][out] */ TreeViewItem **selectedItem) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Root( 
            /* [retval][out] */ TreeViewItem **rootItem) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TopLevel( 
            /* [retval][out] */ TreeViewItems **TopLevel) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Clear( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE NewItem( 
            /* [retval][out] */ TreeViewItem **TreeViewItem) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct TreeViewVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            TreeView * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            TreeView * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            TreeView * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            TreeView * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            TreeView * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            TreeView * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            TreeView * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            TreeView * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            TreeView * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            TreeView * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            TreeView * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            TreeView * This,
            /* [retval][out] */ ClientPoint **Position);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            TreeView * This,
            /* [in] */ ClientPoint *Position);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Enabled )( 
            TreeView * This,
            /* [retval][out] */ VARIANT_BOOL *Enabled);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Enabled )( 
            TreeView * This,
            /* [in] */ VARIANT_BOOL Enabled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            TreeView * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Visible )( 
            TreeView * This,
            /* [in] */ VARIANT_BOOL Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            TreeView * This,
            /* [retval][out] */ VARIANT_BOOL *Focused);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Width )( 
            TreeView * This,
            /* [retval][out] */ long *Width);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Width )( 
            TreeView * This,
            /* [in] */ long Width);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Height )( 
            TreeView * This,
            /* [retval][out] */ long *Height);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Height )( 
            TreeView * This,
            /* [in] */ long Height);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabOrder )( 
            TreeView * This,
            /* [retval][out] */ long *TabOrder);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabOrder )( 
            TreeView * This,
            /* [in] */ long TabOrder);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabStop )( 
            TreeView * This,
            /* [retval][out] */ VARIANT_BOOL *TabStop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabStop )( 
            TreeView * This,
            /* [in] */ VARIANT_BOOL TabStop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Group )( 
            TreeView * This,
            /* [retval][out] */ VARIANT_BOOL *Group);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Group )( 
            TreeView * This,
            /* [in] */ VARIANT_BOOL Group);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            TreeView * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            TreeView * This,
            /* [retval][out] */ long *id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Next )( 
            TreeView * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Previous )( 
            TreeView * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Focus )( 
            TreeView * This);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_id )( 
            TreeView * This,
            /* [in] */ long id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Checkboxes )( 
            TreeView * This,
            /* [retval][out] */ VARIANT_BOOL *Checkboxes);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indent )( 
            TreeView * This,
            /* [retval][out] */ long *Indent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indent )( 
            TreeView * This,
            /* [in] */ long Indent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Selected )( 
            TreeView * This,
            /* [retval][out] */ TreeViewItem **selectedItem);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Root )( 
            TreeView * This,
            /* [retval][out] */ TreeViewItem **rootItem);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TopLevel )( 
            TreeView * This,
            /* [retval][out] */ TreeViewItems **TopLevel);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Clear )( 
            TreeView * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *NewItem )( 
            TreeView * This,
            /* [retval][out] */ TreeViewItem **TreeViewItem);
        
        END_INTERFACE
    } TreeViewVtbl;

    interface TreeView
    {
        CONST_VTBL struct TreeViewVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define TreeView_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define TreeView_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define TreeView_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define TreeView_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define TreeView_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define TreeView_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define TreeView_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define TreeView_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define TreeView_put_Text(This,Text)	\
    ( (This)->lpVtbl -> put_Text(This,Text) ) 

#define TreeView_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define TreeView_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define TreeView_get_Position(This,Position)	\
    ( (This)->lpVtbl -> get_Position(This,Position) ) 

#define TreeView_put_Position(This,Position)	\
    ( (This)->lpVtbl -> put_Position(This,Position) ) 

#define TreeView_get_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> get_Enabled(This,Enabled) ) 

#define TreeView_put_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> put_Enabled(This,Enabled) ) 

#define TreeView_get_Visible(This,Visible)	\
    ( (This)->lpVtbl -> get_Visible(This,Visible) ) 

#define TreeView_put_Visible(This,Visible)	\
    ( (This)->lpVtbl -> put_Visible(This,Visible) ) 

#define TreeView_get_Focused(This,Focused)	\
    ( (This)->lpVtbl -> get_Focused(This,Focused) ) 

#define TreeView_get_Width(This,Width)	\
    ( (This)->lpVtbl -> get_Width(This,Width) ) 

#define TreeView_put_Width(This,Width)	\
    ( (This)->lpVtbl -> put_Width(This,Width) ) 

#define TreeView_get_Height(This,Height)	\
    ( (This)->lpVtbl -> get_Height(This,Height) ) 

#define TreeView_put_Height(This,Height)	\
    ( (This)->lpVtbl -> put_Height(This,Height) ) 

#define TreeView_get_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> get_TabOrder(This,TabOrder) ) 

#define TreeView_put_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> put_TabOrder(This,TabOrder) ) 

#define TreeView_get_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> get_TabStop(This,TabStop) ) 

#define TreeView_put_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> put_TabStop(This,TabStop) ) 

#define TreeView_get_Group(This,Group)	\
    ( (This)->lpVtbl -> get_Group(This,Group) ) 

#define TreeView_put_Group(This,Group)	\
    ( (This)->lpVtbl -> put_Group(This,Group) ) 

#define TreeView_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define TreeView_get_id(This,id)	\
    ( (This)->lpVtbl -> get_id(This,id) ) 

#define TreeView_get_Next(This,prop)	\
    ( (This)->lpVtbl -> get_Next(This,prop) ) 

#define TreeView_get_Previous(This,prop)	\
    ( (This)->lpVtbl -> get_Previous(This,prop) ) 

#define TreeView_Focus(This)	\
    ( (This)->lpVtbl -> Focus(This) ) 

#define TreeView_put_id(This,id)	\
    ( (This)->lpVtbl -> put_id(This,id) ) 


#define TreeView_get_Checkboxes(This,Checkboxes)	\
    ( (This)->lpVtbl -> get_Checkboxes(This,Checkboxes) ) 

#define TreeView_get_Indent(This,Indent)	\
    ( (This)->lpVtbl -> get_Indent(This,Indent) ) 

#define TreeView_put_Indent(This,Indent)	\
    ( (This)->lpVtbl -> put_Indent(This,Indent) ) 

#define TreeView_get_Selected(This,selectedItem)	\
    ( (This)->lpVtbl -> get_Selected(This,selectedItem) ) 

#define TreeView_get_Root(This,rootItem)	\
    ( (This)->lpVtbl -> get_Root(This,rootItem) ) 

#define TreeView_get_TopLevel(This,TopLevel)	\
    ( (This)->lpVtbl -> get_TopLevel(This,TopLevel) ) 

#define TreeView_Clear(This)	\
    ( (This)->lpVtbl -> Clear(This) ) 

#define TreeView_NewItem(This,TreeViewItem)	\
    ( (This)->lpVtbl -> NewItem(This,TreeViewItem) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __TreeView_INTERFACE_DEFINED__ */


#ifndef __TreeViewItem_INTERFACE_DEFINED__
#define __TreeViewItem_INTERFACE_DEFINED__

/* interface TreeViewItem */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_TreeViewItem;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("37F7BC02-68C6-4085-A728-24C2759CC6B1")
    TreeViewItem : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Text( 
            /* [retval][out] */ BSTR *Text) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Text( 
            /* [in] */ BSTR Text) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Selected( 
            /* [retval][out] */ VARIANT_BOOL *Selected) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Selected( 
            /* [in] */ VARIANT_BOOL Selected) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Expanded( 
            /* [retval][out] */ VARIANT_BOOL *Expanded) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Expanded( 
            /* [in] */ VARIANT_BOOL Expanded) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ParentItem( 
            /* [retval][out] */ TreeViewItem **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Children( 
            /* [retval][out] */ TreeViewItems **Children) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Siblings( 
            /* [retval][out] */ TreeViewItems **Siblings) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_UnknownChildren( 
            /* [retval][out] */ VARIANT_BOOL *UnknownChildren) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Checked( 
            /* [retval][out] */ treeviewCheck *Checked) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Checked( 
            /* [in] */ treeviewCheck Checked) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Image( 
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Image( 
            /* [in] */ long index) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_StateImage( 
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_StateImage( 
            /* [in] */ long index) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_OverlayImage( 
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_OverlayImage( 
            /* [in] */ long index) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Data( 
            /* [retval][out] */ long *Data) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Data( 
            /* [in] */ long Data) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Handle( 
            /* [retval][out] */ long *Handle) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Handle( 
            /* [in] */ long Handle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Delete( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ScrollTo( void) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct TreeViewItemVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            TreeViewItem * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            TreeViewItem * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            TreeViewItem * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            TreeViewItem * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            TreeViewItem * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            TreeViewItem * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            TreeViewItem * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            TreeViewItem * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            TreeViewItem * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            TreeViewItem * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            TreeViewItem * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Selected )( 
            TreeViewItem * This,
            /* [retval][out] */ VARIANT_BOOL *Selected);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Selected )( 
            TreeViewItem * This,
            /* [in] */ VARIANT_BOOL Selected);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Expanded )( 
            TreeViewItem * This,
            /* [retval][out] */ VARIANT_BOOL *Expanded);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Expanded )( 
            TreeViewItem * This,
            /* [in] */ VARIANT_BOOL Expanded);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ParentItem )( 
            TreeViewItem * This,
            /* [retval][out] */ TreeViewItem **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Children )( 
            TreeViewItem * This,
            /* [retval][out] */ TreeViewItems **Children);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Siblings )( 
            TreeViewItem * This,
            /* [retval][out] */ TreeViewItems **Siblings);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_UnknownChildren )( 
            TreeViewItem * This,
            /* [retval][out] */ VARIANT_BOOL *UnknownChildren);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Checked )( 
            TreeViewItem * This,
            /* [retval][out] */ treeviewCheck *Checked);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Checked )( 
            TreeViewItem * This,
            /* [in] */ treeviewCheck Checked);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Image )( 
            TreeViewItem * This,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Image )( 
            TreeViewItem * This,
            /* [in] */ long index);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_StateImage )( 
            TreeViewItem * This,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_StateImage )( 
            TreeViewItem * This,
            /* [in] */ long index);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_OverlayImage )( 
            TreeViewItem * This,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_OverlayImage )( 
            TreeViewItem * This,
            /* [in] */ long index);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Data )( 
            TreeViewItem * This,
            /* [retval][out] */ long *Data);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Data )( 
            TreeViewItem * This,
            /* [in] */ long Data);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Handle )( 
            TreeViewItem * This,
            /* [retval][out] */ long *Handle);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Handle )( 
            TreeViewItem * This,
            /* [in] */ long Handle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Delete )( 
            TreeViewItem * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ScrollTo )( 
            TreeViewItem * This);
        
        END_INTERFACE
    } TreeViewItemVtbl;

    interface TreeViewItem
    {
        CONST_VTBL struct TreeViewItemVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define TreeViewItem_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define TreeViewItem_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define TreeViewItem_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define TreeViewItem_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define TreeViewItem_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define TreeViewItem_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define TreeViewItem_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define TreeViewItem_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define TreeViewItem_put_Text(This,Text)	\
    ( (This)->lpVtbl -> put_Text(This,Text) ) 

#define TreeViewItem_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define TreeViewItem_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define TreeViewItem_get_Selected(This,Selected)	\
    ( (This)->lpVtbl -> get_Selected(This,Selected) ) 

#define TreeViewItem_put_Selected(This,Selected)	\
    ( (This)->lpVtbl -> put_Selected(This,Selected) ) 

#define TreeViewItem_get_Expanded(This,Expanded)	\
    ( (This)->lpVtbl -> get_Expanded(This,Expanded) ) 

#define TreeViewItem_put_Expanded(This,Expanded)	\
    ( (This)->lpVtbl -> put_Expanded(This,Expanded) ) 

#define TreeViewItem_get_ParentItem(This,Parent)	\
    ( (This)->lpVtbl -> get_ParentItem(This,Parent) ) 

#define TreeViewItem_get_Children(This,Children)	\
    ( (This)->lpVtbl -> get_Children(This,Children) ) 

#define TreeViewItem_get_Siblings(This,Siblings)	\
    ( (This)->lpVtbl -> get_Siblings(This,Siblings) ) 

#define TreeViewItem_get_UnknownChildren(This,UnknownChildren)	\
    ( (This)->lpVtbl -> get_UnknownChildren(This,UnknownChildren) ) 

#define TreeViewItem_get_Checked(This,Checked)	\
    ( (This)->lpVtbl -> get_Checked(This,Checked) ) 

#define TreeViewItem_put_Checked(This,Checked)	\
    ( (This)->lpVtbl -> put_Checked(This,Checked) ) 

#define TreeViewItem_get_Image(This,index)	\
    ( (This)->lpVtbl -> get_Image(This,index) ) 

#define TreeViewItem_put_Image(This,index)	\
    ( (This)->lpVtbl -> put_Image(This,index) ) 

#define TreeViewItem_get_StateImage(This,index)	\
    ( (This)->lpVtbl -> get_StateImage(This,index) ) 

#define TreeViewItem_put_StateImage(This,index)	\
    ( (This)->lpVtbl -> put_StateImage(This,index) ) 

#define TreeViewItem_get_OverlayImage(This,index)	\
    ( (This)->lpVtbl -> get_OverlayImage(This,index) ) 

#define TreeViewItem_put_OverlayImage(This,index)	\
    ( (This)->lpVtbl -> put_OverlayImage(This,index) ) 

#define TreeViewItem_get_Data(This,Data)	\
    ( (This)->lpVtbl -> get_Data(This,Data) ) 

#define TreeViewItem_put_Data(This,Data)	\
    ( (This)->lpVtbl -> put_Data(This,Data) ) 

#define TreeViewItem_get_Handle(This,Handle)	\
    ( (This)->lpVtbl -> get_Handle(This,Handle) ) 

#define TreeViewItem_put_Handle(This,Handle)	\
    ( (This)->lpVtbl -> put_Handle(This,Handle) ) 

#define TreeViewItem_Delete(This)	\
    ( (This)->lpVtbl -> Delete(This) ) 

#define TreeViewItem_ScrollTo(This)	\
    ( (This)->lpVtbl -> ScrollTo(This) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __TreeViewItem_INTERFACE_DEFINED__ */


#ifndef __TreeViewItems_INTERFACE_DEFINED__
#define __TreeViewItems_INTERFACE_DEFINED__

/* interface TreeViewItems */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_TreeViewItems;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("0F706D22-6B84-4701-99B7-2234A864E317")
    TreeViewItems : public IDispatch
    {
    public:
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ TreeViewItem **Item) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Insert( 
            /* [in] */ TreeViewItem *Item,
            /* [in] */ VARIANT InsertAfter) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct TreeViewItemsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            TreeViewItems * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            TreeViewItems * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            TreeViewItems * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            TreeViewItems * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            TreeViewItems * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            TreeViewItems * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            TreeViewItems * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            TreeViewItems * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            TreeViewItems * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            TreeViewItems * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            TreeViewItems * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            TreeViewItems * This,
            /* [in] */ long index,
            /* [retval][out] */ TreeViewItem **Item);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Insert )( 
            TreeViewItems * This,
            /* [in] */ TreeViewItem *Item,
            /* [in] */ VARIANT InsertAfter);
        
        END_INTERFACE
    } TreeViewItemsVtbl;

    interface TreeViewItems
    {
        CONST_VTBL struct TreeViewItemsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define TreeViewItems_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define TreeViewItems_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define TreeViewItems_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define TreeViewItems_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define TreeViewItems_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define TreeViewItems_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define TreeViewItems_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define TreeViewItems_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define TreeViewItems_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define TreeViewItems_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define TreeViewItems_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define TreeViewItems_Item(This,index,Item)	\
    ( (This)->lpVtbl -> Item(This,index,Item) ) 

#define TreeViewItems_Insert(This,Item,InsertAfter)	\
    ( (This)->lpVtbl -> Insert(This,Item,InsertAfter) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __TreeViewItems_INTERFACE_DEFINED__ */


#ifndef ___ApplicationEvents_INTERFACE_DEFINED__
#define ___ApplicationEvents_INTERFACE_DEFINED__

/* interface _ApplicationEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID IID__ApplicationEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("D3AE5D19-1833-4CC1-8647-194C201270AB")
    _ApplicationEvents
    {
    public:
        BEGIN_INTERFACE
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnCursorKey( 
            /* [in] */ CursorKey *CursorKey,
            /* [in] */ VARIANT_BOOL IsBeforeApplication,
            /* [in] */ VARIANT_BOOL DefaultActionAborted,
            /* [retval][out] */ VARIANT_BOOL *AbortDefaultAction) = 0;
        
        END_INTERFACE
    };
    
    
#else 	/* C style interface */

    typedef struct _ApplicationEventsVtbl
    {
        BEGIN_INTERFACE
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnCursorKey )( 
            _ApplicationEvents * This,
            /* [in] */ CursorKey *CursorKey,
            /* [in] */ VARIANT_BOOL IsBeforeApplication,
            /* [in] */ VARIANT_BOOL DefaultActionAborted,
            /* [retval][out] */ VARIANT_BOOL *AbortDefaultAction);
        
        END_INTERFACE
    } _ApplicationEventsVtbl;

    interface _ApplicationEvents
    {
        CONST_VTBL struct _ApplicationEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _ApplicationEvents_OnCursorKey(This,CursorKey,IsBeforeApplication,DefaultActionAborted,AbortDefaultAction)	\
    ( (This)->lpVtbl -> OnCursorKey(This,CursorKey,IsBeforeApplication,DefaultActionAborted,AbortDefaultAction) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___ApplicationEvents_INTERFACE_DEFINED__ */


#ifndef __ApplicationEvents_DISPINTERFACE_DEFINED__
#define __ApplicationEvents_DISPINTERFACE_DEFINED__

/* dispinterface ApplicationEvents */
/* [helpcontext][uuid] */ 


EXTERN_C const IID DIID_ApplicationEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)

    MIDL_INTERFACE("DCFE6A7D-1541-4EAD-BC3A-BB9108C36C7F")
    ApplicationEvents : public IDispatch
    {
    };
    
#else 	/* C style interface */

    typedef struct ApplicationEventsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ApplicationEvents * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ApplicationEvents * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ApplicationEvents * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ApplicationEvents * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ApplicationEvents * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ApplicationEvents * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ApplicationEvents * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        END_INTERFACE
    } ApplicationEventsVtbl;

    interface ApplicationEvents
    {
        CONST_VTBL struct ApplicationEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ApplicationEvents_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ApplicationEvents_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ApplicationEvents_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ApplicationEvents_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ApplicationEvents_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ApplicationEvents_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ApplicationEvents_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */


#endif 	/* __ApplicationEvents_DISPINTERFACE_DEFINED__ */


#ifndef ___SharedObjectEvents_INTERFACE_DEFINED__
#define ___SharedObjectEvents_INTERFACE_DEFINED__

/* interface _SharedObjectEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID IID__SharedObjectEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("410CB436-7831-4707-A01F-64E425FF1F0E")
    _SharedObjectEvents
    {
    public:
        BEGIN_INTERFACE
        END_INTERFACE
    };
    
    
#else 	/* C style interface */

    typedef struct _SharedObjectEventsVtbl
    {
        BEGIN_INTERFACE
        
        END_INTERFACE
    } _SharedObjectEventsVtbl;

    interface _SharedObjectEvents
    {
        CONST_VTBL struct _SharedObjectEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___SharedObjectEvents_INTERFACE_DEFINED__ */


#ifndef __SharedObjectEvents_DISPINTERFACE_DEFINED__
#define __SharedObjectEvents_DISPINTERFACE_DEFINED__

/* dispinterface SharedObjectEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID DIID_SharedObjectEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)

    MIDL_INTERFACE("88A69BF3-4F8B-406B-9720-29C25542BD9D")
    SharedObjectEvents : public IDispatch
    {
    };
    
#else 	/* C style interface */

    typedef struct SharedObjectEventsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            SharedObjectEvents * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            SharedObjectEvents * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            SharedObjectEvents * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            SharedObjectEvents * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            SharedObjectEvents * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            SharedObjectEvents * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            SharedObjectEvents * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        END_INTERFACE
    } SharedObjectEventsVtbl;

    interface SharedObjectEvents
    {
        CONST_VTBL struct SharedObjectEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define SharedObjectEvents_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define SharedObjectEvents_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define SharedObjectEvents_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define SharedObjectEvents_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define SharedObjectEvents_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define SharedObjectEvents_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define SharedObjectEvents_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */


#endif 	/* __SharedObjectEvents_DISPINTERFACE_DEFINED__ */


#ifndef ___ClientInformationEvents_INTERFACE_DEFINED__
#define ___ClientInformationEvents_INTERFACE_DEFINED__

/* interface _ClientInformationEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID IID__ClientInformationEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("A09E1E00-2B3C-4E53-9005-40D400CE039A")
    _ClientInformationEvents
    {
    public:
        BEGIN_INTERFACE
        END_INTERFACE
    };
    
    
#else 	/* C style interface */

    typedef struct _ClientInformationEventsVtbl
    {
        BEGIN_INTERFACE
        
        END_INTERFACE
    } _ClientInformationEventsVtbl;

    interface _ClientInformationEvents
    {
        CONST_VTBL struct _ClientInformationEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___ClientInformationEvents_INTERFACE_DEFINED__ */


#ifndef __ClientInformationEvents_DISPINTERFACE_DEFINED__
#define __ClientInformationEvents_DISPINTERFACE_DEFINED__

/* dispinterface ClientInformationEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID DIID_ClientInformationEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)

    MIDL_INTERFACE("B571F50F-AB45-487D-B320-66989EDB305A")
    ClientInformationEvents : public IDispatch
    {
    };
    
#else 	/* C style interface */

    typedef struct ClientInformationEventsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ClientInformationEvents * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ClientInformationEvents * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ClientInformationEvents * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ClientInformationEvents * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ClientInformationEvents * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ClientInformationEvents * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ClientInformationEvents * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        END_INTERFACE
    } ClientInformationEventsVtbl;

    interface ClientInformationEvents
    {
        CONST_VTBL struct ClientInformationEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ClientInformationEvents_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ClientInformationEvents_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ClientInformationEvents_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ClientInformationEvents_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ClientInformationEvents_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ClientInformationEvents_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ClientInformationEvents_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */


#endif 	/* __ClientInformationEvents_DISPINTERFACE_DEFINED__ */


#ifndef ___Script_INTERFACE_DEFINED__
#define ___Script_INTERFACE_DEFINED__

/* interface _Script */
/* [object][oleautomation][dual][hidden][helpcontext][helpstring][version][uuid] */ 


EXTERN_C const IID IID__Script;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("9426C016-0F4F-4E32-BEE8-BFD466C8CFCF")
    _Script : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **Application) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE CreateObject( 
            /* [in] */ BSTR progid,
            /* [retval][out] */ IDispatch **pDisp) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ConnectEvent( 
            /* [in] */ VARIANT sourceObject,
            /* [in] */ VARIANT event,
            /* [in] */ VARIANT function,
            /* [retval][out] */ long *connectionHandle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Disconnect( 
            /* [in] */ long connectionHandle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Sleep( 
            /* [in] */ long Milliseconds) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE StartTimer( 
            /* [in] */ long Milliseconds,
            /* [in] */ VARIANT function,
            /* [optional][in] */ VARIANT Parameter,
            /* [retval][out] */ long *TimerID) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE StopTimer( 
            /* [in] */ long TimerID) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ConnectObject( 
            /* [in] */ VARIANT sourceObject,
            /* [in] */ BSTR prefix,
            /* [retval][out] */ long *connectionHandle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE OutputDebugString( 
            /* [in] */ BSTR str) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE CommonFileDialog( 
            /* [in] */ VARIANT_BOOL UseForOpen,
            /* [optional][in] */ VARIANT DialogTitle,
            /* [optional][in] */ VARIANT InitialFileName,
            /* [optional][in] */ VARIANT InitialDirectory,
            /* [optional][in] */ VARIANT Filters,
            /* [optional][in] */ VARIANT FilterIndex,
            /* [optional][in] */ VARIANT DefaultExtension,
            /* [optional][in] */ VARIANT Flags,
            /* [optional][in] */ VARIANT ParentWindow,
            /* [retval][out] */ BSTR *Filename) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE LoadClassInformation( 
            IDispatch *Object) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE StopScript( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE DoEvents( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Queue( 
            /* [in] */ VARIANT function,
            /* [optional][in] */ VARIANT Parameter) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ConnectEventWithParameter( 
            /* [in] */ VARIANT sourceObject,
            /* [in] */ VARIANT event,
            /* [in] */ VARIANT function,
            /* [in] */ VARIANT Parameter,
            /* [retval][out] */ long *connectionHandle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ConnectObjectWithParameter( 
            /* [in] */ VARIANT sourceObject,
            /* [in] */ BSTR prefix,
            /* [in] */ VARIANT Parameter,
            /* [retval][out] */ long *connectionHandle) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _ScriptVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _Script * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _Script * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _Script * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _Script * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _Script * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _Script * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _Script * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _Script * This,
            /* [retval][out] */ _Application **Application);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _Script * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *CreateObject )( 
            _Script * This,
            /* [in] */ BSTR progid,
            /* [retval][out] */ IDispatch **pDisp);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ConnectEvent )( 
            _Script * This,
            /* [in] */ VARIANT sourceObject,
            /* [in] */ VARIANT event,
            /* [in] */ VARIANT function,
            /* [retval][out] */ long *connectionHandle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Disconnect )( 
            _Script * This,
            /* [in] */ long connectionHandle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Sleep )( 
            _Script * This,
            /* [in] */ long Milliseconds);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *StartTimer )( 
            _Script * This,
            /* [in] */ long Milliseconds,
            /* [in] */ VARIANT function,
            /* [optional][in] */ VARIANT Parameter,
            /* [retval][out] */ long *TimerID);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *StopTimer )( 
            _Script * This,
            /* [in] */ long TimerID);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ConnectObject )( 
            _Script * This,
            /* [in] */ VARIANT sourceObject,
            /* [in] */ BSTR prefix,
            /* [retval][out] */ long *connectionHandle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *OutputDebugString )( 
            _Script * This,
            /* [in] */ BSTR str);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *CommonFileDialog )( 
            _Script * This,
            /* [in] */ VARIANT_BOOL UseForOpen,
            /* [optional][in] */ VARIANT DialogTitle,
            /* [optional][in] */ VARIANT InitialFileName,
            /* [optional][in] */ VARIANT InitialDirectory,
            /* [optional][in] */ VARIANT Filters,
            /* [optional][in] */ VARIANT FilterIndex,
            /* [optional][in] */ VARIANT DefaultExtension,
            /* [optional][in] */ VARIANT Flags,
            /* [optional][in] */ VARIANT ParentWindow,
            /* [retval][out] */ BSTR *Filename);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *LoadClassInformation )( 
            _Script * This,
            IDispatch *Object);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *StopScript )( 
            _Script * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *DoEvents )( 
            _Script * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Queue )( 
            _Script * This,
            /* [in] */ VARIANT function,
            /* [optional][in] */ VARIANT Parameter);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ConnectEventWithParameter )( 
            _Script * This,
            /* [in] */ VARIANT sourceObject,
            /* [in] */ VARIANT event,
            /* [in] */ VARIANT function,
            /* [in] */ VARIANT Parameter,
            /* [retval][out] */ long *connectionHandle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ConnectObjectWithParameter )( 
            _Script * This,
            /* [in] */ VARIANT sourceObject,
            /* [in] */ BSTR prefix,
            /* [in] */ VARIANT Parameter,
            /* [retval][out] */ long *connectionHandle);
        
        END_INTERFACE
    } _ScriptVtbl;

    interface _Script
    {
        CONST_VTBL struct _ScriptVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _Script_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _Script_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _Script_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _Script_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _Script_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _Script_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _Script_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _Script_get_Application(This,Application)	\
    ( (This)->lpVtbl -> get_Application(This,Application) ) 

#define _Script_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _Script_CreateObject(This,progid,pDisp)	\
    ( (This)->lpVtbl -> CreateObject(This,progid,pDisp) ) 

#define _Script_ConnectEvent(This,sourceObject,event,function,connectionHandle)	\
    ( (This)->lpVtbl -> ConnectEvent(This,sourceObject,event,function,connectionHandle) ) 

#define _Script_Disconnect(This,connectionHandle)	\
    ( (This)->lpVtbl -> Disconnect(This,connectionHandle) ) 

#define _Script_Sleep(This,Milliseconds)	\
    ( (This)->lpVtbl -> Sleep(This,Milliseconds) ) 

#define _Script_StartTimer(This,Milliseconds,function,Parameter,TimerID)	\
    ( (This)->lpVtbl -> StartTimer(This,Milliseconds,function,Parameter,TimerID) ) 

#define _Script_StopTimer(This,TimerID)	\
    ( (This)->lpVtbl -> StopTimer(This,TimerID) ) 

#define _Script_ConnectObject(This,sourceObject,prefix,connectionHandle)	\
    ( (This)->lpVtbl -> ConnectObject(This,sourceObject,prefix,connectionHandle) ) 

#define _Script_OutputDebugString(This,str)	\
    ( (This)->lpVtbl -> OutputDebugString(This,str) ) 

#define _Script_CommonFileDialog(This,UseForOpen,DialogTitle,InitialFileName,InitialDirectory,Filters,FilterIndex,DefaultExtension,Flags,ParentWindow,Filename)	\
    ( (This)->lpVtbl -> CommonFileDialog(This,UseForOpen,DialogTitle,InitialFileName,InitialDirectory,Filters,FilterIndex,DefaultExtension,Flags,ParentWindow,Filename) ) 

#define _Script_LoadClassInformation(This,Object)	\
    ( (This)->lpVtbl -> LoadClassInformation(This,Object) ) 

#define _Script_StopScript(This)	\
    ( (This)->lpVtbl -> StopScript(This) ) 

#define _Script_DoEvents(This)	\
    ( (This)->lpVtbl -> DoEvents(This) ) 

#define _Script_Queue(This,function,Parameter)	\
    ( (This)->lpVtbl -> Queue(This,function,Parameter) ) 

#define _Script_ConnectEventWithParameter(This,sourceObject,event,function,Parameter,connectionHandle)	\
    ( (This)->lpVtbl -> ConnectEventWithParameter(This,sourceObject,event,function,Parameter,connectionHandle) ) 

#define _Script_ConnectObjectWithParameter(This,sourceObject,prefix,Parameter,connectionHandle)	\
    ( (This)->lpVtbl -> ConnectObjectWithParameter(This,sourceObject,prefix,Parameter,connectionHandle) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___Script_INTERFACE_DEFINED__ */


#ifndef ___ScriptEvents_INTERFACE_DEFINED__
#define ___ScriptEvents_INTERFACE_DEFINED__

/* interface _ScriptEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID IID__ScriptEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("54EA3DCB-7CED-4EE6-A8EB-60197B901657")
    _ScriptEvents
    {
    public:
        BEGIN_INTERFACE
        END_INTERFACE
    };
    
    
#else 	/* C style interface */

    typedef struct _ScriptEventsVtbl
    {
        BEGIN_INTERFACE
        
        END_INTERFACE
    } _ScriptEventsVtbl;

    interface _ScriptEvents
    {
        CONST_VTBL struct _ScriptEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___ScriptEvents_INTERFACE_DEFINED__ */


#ifndef __ScriptEvents_DISPINTERFACE_DEFINED__
#define __ScriptEvents_DISPINTERFACE_DEFINED__

/* dispinterface ScriptEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID DIID_ScriptEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)

    MIDL_INTERFACE("AFAD796F-288A-4947-B442-19541CBF9F9D")
    ScriptEvents : public IDispatch
    {
    };
    
#else 	/* C style interface */

    typedef struct ScriptEventsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ScriptEvents * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ScriptEvents * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ScriptEvents * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ScriptEvents * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ScriptEvents * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ScriptEvents * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ScriptEvents * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        END_INTERFACE
    } ScriptEventsVtbl;

    interface ScriptEvents
    {
        CONST_VTBL struct ScriptEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ScriptEvents_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ScriptEvents_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ScriptEvents_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ScriptEvents_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ScriptEvents_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ScriptEvents_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ScriptEvents_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */


#endif 	/* __ScriptEvents_DISPINTERFACE_DEFINED__ */


#ifndef ___MouseEvents_INTERFACE_DEFINED__
#define ___MouseEvents_INTERFACE_DEFINED__

/* interface _MouseEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID IID__MouseEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("A7F54D0C-4DBD-4B50-A081-E8039352D118")
    _MouseEvents
    {
    public:
        BEGIN_INTERFACE
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnMoveEx( 
            /* [in] */ ScreenPoint *ScreenPoint,
            /* [in] */ _Window *Window,
            /* [in] */ MouseButton button,
            /* [in] */ KeyModifiers KeyModifiers) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnButtonUpEx( 
            /* [in] */ ScreenPoint *ScreenPoint,
            /* [in] */ _Window *Window,
            /* [in] */ MouseButton button,
            /* [in] */ KeyModifiers KeyModifiers) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnButtonDownEx( 
            /* [in] */ ScreenPoint *ScreenPoint,
            /* [in] */ _Window *Window,
            /* [in] */ MouseButton button,
            /* [in] */ KeyModifiers KeyModifiers) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnWheel( 
            /* [in] */ ScreenPoint *ScreenPoint,
            /* [in] */ _Window *Window,
            /* [in] */ MouseButton button,
            /* [in] */ KeyModifiers KeyModifiers,
            /* [in] */ long Delta,
            /* [in] */ MouseWheelDirection direction) = 0;
        
        END_INTERFACE
    };
    
    
#else 	/* C style interface */

    typedef struct _MouseEventsVtbl
    {
        BEGIN_INTERFACE
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnMoveEx )( 
            _MouseEvents * This,
            /* [in] */ ScreenPoint *ScreenPoint,
            /* [in] */ _Window *Window,
            /* [in] */ MouseButton button,
            /* [in] */ KeyModifiers KeyModifiers);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnButtonUpEx )( 
            _MouseEvents * This,
            /* [in] */ ScreenPoint *ScreenPoint,
            /* [in] */ _Window *Window,
            /* [in] */ MouseButton button,
            /* [in] */ KeyModifiers KeyModifiers);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnButtonDownEx )( 
            _MouseEvents * This,
            /* [in] */ ScreenPoint *ScreenPoint,
            /* [in] */ _Window *Window,
            /* [in] */ MouseButton button,
            /* [in] */ KeyModifiers KeyModifiers);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnWheel )( 
            _MouseEvents * This,
            /* [in] */ ScreenPoint *ScreenPoint,
            /* [in] */ _Window *Window,
            /* [in] */ MouseButton button,
            /* [in] */ KeyModifiers KeyModifiers,
            /* [in] */ long Delta,
            /* [in] */ MouseWheelDirection direction);
        
        END_INTERFACE
    } _MouseEventsVtbl;

    interface _MouseEvents
    {
        CONST_VTBL struct _MouseEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _MouseEvents_OnMoveEx(This,ScreenPoint,Window,button,KeyModifiers)	\
    ( (This)->lpVtbl -> OnMoveEx(This,ScreenPoint,Window,button,KeyModifiers) ) 

#define _MouseEvents_OnButtonUpEx(This,ScreenPoint,Window,button,KeyModifiers)	\
    ( (This)->lpVtbl -> OnButtonUpEx(This,ScreenPoint,Window,button,KeyModifiers) ) 

#define _MouseEvents_OnButtonDownEx(This,ScreenPoint,Window,button,KeyModifiers)	\
    ( (This)->lpVtbl -> OnButtonDownEx(This,ScreenPoint,Window,button,KeyModifiers) ) 

#define _MouseEvents_OnWheel(This,ScreenPoint,Window,button,KeyModifiers,Delta,direction)	\
    ( (This)->lpVtbl -> OnWheel(This,ScreenPoint,Window,button,KeyModifiers,Delta,direction) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___MouseEvents_INTERFACE_DEFINED__ */


#ifndef __MouseEvents_DISPINTERFACE_DEFINED__
#define __MouseEvents_DISPINTERFACE_DEFINED__

/* dispinterface MouseEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID DIID_MouseEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)

    MIDL_INTERFACE("C54C5FB5-92D8-4F03-8354-646EB129A62F")
    MouseEvents : public IDispatch
    {
    };
    
#else 	/* C style interface */

    typedef struct MouseEventsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            MouseEvents * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            MouseEvents * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            MouseEvents * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            MouseEvents * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            MouseEvents * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            MouseEvents * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            MouseEvents * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        END_INTERFACE
    } MouseEventsVtbl;

    interface MouseEvents
    {
        CONST_VTBL struct MouseEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define MouseEvents_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define MouseEvents_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define MouseEvents_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define MouseEvents_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define MouseEvents_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define MouseEvents_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define MouseEvents_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */


#endif 	/* __MouseEvents_DISPINTERFACE_DEFINED__ */


#ifndef __IWEConnectionPointSink_INTERFACE_DEFINED__
#define __IWEConnectionPointSink_INTERFACE_DEFINED__

/* interface IWEConnectionPointSink */
/* [object][uuid] */ 


EXTERN_C const IID IID_IWEConnectionPointSink;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("F46B74A4-5BD8-45F5-8DCD-6B9C014AAEFD")
    IWEConnectionPointSink : public IUnknown
    {
    public:
        virtual HRESULT __stdcall WantsEvent( 
            long event) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct IWEConnectionPointSinkVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            IWEConnectionPointSink * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            IWEConnectionPointSink * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            IWEConnectionPointSink * This);
        
        HRESULT ( __stdcall *WantsEvent )( 
            IWEConnectionPointSink * This,
            long event);
        
        END_INTERFACE
    } IWEConnectionPointSinkVtbl;

    interface IWEConnectionPointSink
    {
        CONST_VTBL struct IWEConnectionPointSinkVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define IWEConnectionPointSink_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define IWEConnectionPointSink_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define IWEConnectionPointSink_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define IWEConnectionPointSink_WantsEvent(This,event)	\
    ( (This)->lpVtbl -> WantsEvent(This,event) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __IWEConnectionPointSink_INTERFACE_DEFINED__ */


#ifndef __WindowEvents_DISPINTERFACE_DEFINED__
#define __WindowEvents_DISPINTERFACE_DEFINED__

/* dispinterface WindowEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID DIID_WindowEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)

    MIDL_INTERFACE("609D3D91-10E3-474A-8922-7251AF6157A5")
    WindowEvents : public IDispatch
    {
    };
    
#else 	/* C style interface */

    typedef struct WindowEventsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            WindowEvents * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            WindowEvents * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            WindowEvents * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            WindowEvents * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            WindowEvents * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            WindowEvents * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            WindowEvents * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        END_INTERFACE
    } WindowEventsVtbl;

    interface WindowEvents
    {
        CONST_VTBL struct WindowEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define WindowEvents_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define WindowEvents_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define WindowEvents_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define WindowEvents_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define WindowEvents_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define WindowEvents_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define WindowEvents_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */


#endif 	/* __WindowEvents_DISPINTERFACE_DEFINED__ */


#ifndef ___KeyEvents_INTERFACE_DEFINED__
#define ___KeyEvents_INTERFACE_DEFINED__

/* interface _KeyEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID IID__KeyEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("DAA21096-0EF0-45B5-B242-315B379EF2D8")
    _KeyEvents
    {
    public:
        BEGIN_INTERFACE
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnKeyProcessedUp( 
            /* [in] */ long Key,
            /* [in] */ KeyModifiers Modifers) = 0;
        
        END_INTERFACE
    };
    
    
#else 	/* C style interface */

    typedef struct _KeyEventsVtbl
    {
        BEGIN_INTERFACE
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnKeyProcessedUp )( 
            _KeyEvents * This,
            /* [in] */ long Key,
            /* [in] */ KeyModifiers Modifers);
        
        END_INTERFACE
    } _KeyEventsVtbl;

    interface _KeyEvents
    {
        CONST_VTBL struct _KeyEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _KeyEvents_OnKeyProcessedUp(This,Key,Modifers)	\
    ( (This)->lpVtbl -> OnKeyProcessedUp(This,Key,Modifers) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___KeyEvents_INTERFACE_DEFINED__ */


#ifndef __KeyEvents_DISPINTERFACE_DEFINED__
#define __KeyEvents_DISPINTERFACE_DEFINED__

/* dispinterface KeyEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID DIID_KeyEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)

    MIDL_INTERFACE("67AAE81E-D763-4BB2-B1C0-CD8421375688")
    KeyEvents : public IDispatch
    {
    };
    
#else 	/* C style interface */

    typedef struct KeyEventsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            KeyEvents * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            KeyEvents * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            KeyEvents * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            KeyEvents * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            KeyEvents * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            KeyEvents * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            KeyEvents * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        END_INTERFACE
    } KeyEventsVtbl;

    interface KeyEvents
    {
        CONST_VTBL struct KeyEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define KeyEvents_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define KeyEvents_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define KeyEvents_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define KeyEvents_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define KeyEvents_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define KeyEvents_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define KeyEvents_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */


#endif 	/* __KeyEvents_DISPINTERFACE_DEFINED__ */


#ifndef ___BrailleEvents_INTERFACE_DEFINED__
#define ___BrailleEvents_INTERFACE_DEFINED__

/* interface _BrailleEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID IID__BrailleEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("E864F32C-5E06-49A5-97FA-752EC7AFEF28")
    _BrailleEvents
    {
    public:
        BEGIN_INTERFACE
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnTableActivate( 
            /* [in] */ BrailleTable *Table) = 0;
        
        END_INTERFACE
    };
    
    
#else 	/* C style interface */

    typedef struct _BrailleEventsVtbl
    {
        BEGIN_INTERFACE
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnTableActivate )( 
            _BrailleEvents * This,
            /* [in] */ BrailleTable *Table);
        
        END_INTERFACE
    } _BrailleEventsVtbl;

    interface _BrailleEvents
    {
        CONST_VTBL struct _BrailleEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _BrailleEvents_OnTableActivate(This,Table)	\
    ( (This)->lpVtbl -> OnTableActivate(This,Table) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___BrailleEvents_INTERFACE_DEFINED__ */


#ifndef __BrailleEvents_DISPINTERFACE_DEFINED__
#define __BrailleEvents_DISPINTERFACE_DEFINED__

/* dispinterface BrailleEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID DIID_BrailleEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)

    MIDL_INTERFACE("1FD1FE04-9D70-43D9-ACE3-62DC859BB93D")
    BrailleEvents : public IDispatch
    {
    };
    
#else 	/* C style interface */

    typedef struct BrailleEventsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleEvents * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleEvents * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleEvents * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleEvents * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleEvents * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleEvents * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleEvents * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        END_INTERFACE
    } BrailleEventsVtbl;

    interface BrailleEvents
    {
        CONST_VTBL struct BrailleEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleEvents_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleEvents_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleEvents_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleEvents_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleEvents_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleEvents_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleEvents_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */


#endif 	/* __BrailleEvents_DISPINTERFACE_DEFINED__ */


#ifndef ___BrailleDisplayEvents_INTERFACE_DEFINED__
#define ___BrailleDisplayEvents_INTERFACE_DEFINED__

/* interface _BrailleDisplayEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID IID__BrailleDisplayEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("B8ED3390-BCD5-46EB-BB41-E76E0A187509")
    _BrailleDisplayEvents
    {
    public:
        BEGIN_INTERFACE
        END_INTERFACE
    };
    
    
#else 	/* C style interface */

    typedef struct _BrailleDisplayEventsVtbl
    {
        BEGIN_INTERFACE
        
        END_INTERFACE
    } _BrailleDisplayEventsVtbl;

    interface _BrailleDisplayEvents
    {
        CONST_VTBL struct _BrailleDisplayEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___BrailleDisplayEvents_INTERFACE_DEFINED__ */


#ifndef __BrailleDisplayEvents_DISPINTERFACE_DEFINED__
#define __BrailleDisplayEvents_DISPINTERFACE_DEFINED__

/* dispinterface BrailleDisplayEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID DIID_BrailleDisplayEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)

    MIDL_INTERFACE("7E518B60-494B-4153-BDC0-0501D2322FE5")
    BrailleDisplayEvents : public IDispatch
    {
    };
    
#else 	/* C style interface */

    typedef struct BrailleDisplayEventsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrailleDisplayEvents * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrailleDisplayEvents * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrailleDisplayEvents * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrailleDisplayEvents * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrailleDisplayEvents * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrailleDisplayEvents * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrailleDisplayEvents * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        END_INTERFACE
    } BrailleDisplayEventsVtbl;

    interface BrailleDisplayEvents
    {
        CONST_VTBL struct BrailleDisplayEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrailleDisplayEvents_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrailleDisplayEvents_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrailleDisplayEvents_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrailleDisplayEvents_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrailleDisplayEvents_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrailleDisplayEvents_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrailleDisplayEvents_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */


#endif 	/* __BrailleDisplayEvents_DISPINTERFACE_DEFINED__ */


#ifndef ___SpeechEvents_INTERFACE_DEFINED__
#define ___SpeechEvents_INTERFACE_DEFINED__

/* interface _SpeechEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID IID__SpeechEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("05911DA0-F272-4FDB-8CDB-24B916957F72")
    _SpeechEvents
    {
    public:
        BEGIN_INTERFACE
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnProcessRepeatFilter( 
            /* [in] */ BSTR String,
            /* [out] */ BSTR *NewString,
            /* [retval][out] */ VARIANT_BOOL *Process) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnProcessExceptionDictionaries( 
            /* [in] */ BSTR String,
            /* [out] */ BSTR *NewString,
            /* [retval][out] */ VARIANT_BOOL *Process) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnProcessNumbers( 
            /* [in] */ BSTR String,
            /* [out] */ BSTR *NewString,
            /* [retval][out] */ VARIANT_BOOL *Process) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnProcessSplitAlphaNumeric( 
            /* [in] */ BSTR String,
            /* [out] */ BSTR *NewString,
            /* [retval][out] */ VARIANT_BOOL *Process) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnProcessSplitMixedCaseWords( 
            /* [in] */ BSTR String,
            /* [out] */ BSTR *NewString,
            /* [retval][out] */ VARIANT_BOOL *Process) = 0;
        
        virtual /* [helpcontext][helpstring] */ HRESULT __stdcall OnLanguageSwitch( 
            /* [in] */ BSTR Language) = 0;
        
        END_INTERFACE
    };
    
    
#else 	/* C style interface */

    typedef struct _SpeechEventsVtbl
    {
        BEGIN_INTERFACE
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnProcessRepeatFilter )( 
            _SpeechEvents * This,
            /* [in] */ BSTR String,
            /* [out] */ BSTR *NewString,
            /* [retval][out] */ VARIANT_BOOL *Process);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnProcessExceptionDictionaries )( 
            _SpeechEvents * This,
            /* [in] */ BSTR String,
            /* [out] */ BSTR *NewString,
            /* [retval][out] */ VARIANT_BOOL *Process);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnProcessNumbers )( 
            _SpeechEvents * This,
            /* [in] */ BSTR String,
            /* [out] */ BSTR *NewString,
            /* [retval][out] */ VARIANT_BOOL *Process);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnProcessSplitAlphaNumeric )( 
            _SpeechEvents * This,
            /* [in] */ BSTR String,
            /* [out] */ BSTR *NewString,
            /* [retval][out] */ VARIANT_BOOL *Process);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnProcessSplitMixedCaseWords )( 
            _SpeechEvents * This,
            /* [in] */ BSTR String,
            /* [out] */ BSTR *NewString,
            /* [retval][out] */ VARIANT_BOOL *Process);
        
        /* [helpcontext][helpstring] */ HRESULT ( __stdcall *OnLanguageSwitch )( 
            _SpeechEvents * This,
            /* [in] */ BSTR Language);
        
        END_INTERFACE
    } _SpeechEventsVtbl;

    interface _SpeechEvents
    {
        CONST_VTBL struct _SpeechEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _SpeechEvents_OnProcessRepeatFilter(This,String,NewString,Process)	\
    ( (This)->lpVtbl -> OnProcessRepeatFilter(This,String,NewString,Process) ) 

#define _SpeechEvents_OnProcessExceptionDictionaries(This,String,NewString,Process)	\
    ( (This)->lpVtbl -> OnProcessExceptionDictionaries(This,String,NewString,Process) ) 

#define _SpeechEvents_OnProcessNumbers(This,String,NewString,Process)	\
    ( (This)->lpVtbl -> OnProcessNumbers(This,String,NewString,Process) ) 

#define _SpeechEvents_OnProcessSplitAlphaNumeric(This,String,NewString,Process)	\
    ( (This)->lpVtbl -> OnProcessSplitAlphaNumeric(This,String,NewString,Process) ) 

#define _SpeechEvents_OnProcessSplitMixedCaseWords(This,String,NewString,Process)	\
    ( (This)->lpVtbl -> OnProcessSplitMixedCaseWords(This,String,NewString,Process) ) 

#define _SpeechEvents_OnLanguageSwitch(This,Language)	\
    ( (This)->lpVtbl -> OnLanguageSwitch(This,Language) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___SpeechEvents_INTERFACE_DEFINED__ */


#ifndef __SpeechEvents_DISPINTERFACE_DEFINED__
#define __SpeechEvents_DISPINTERFACE_DEFINED__

/* dispinterface SpeechEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID DIID_SpeechEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)

    MIDL_INTERFACE("A2F3EB81-CFCC-4D1E-8033-21822A9A6F68")
    SpeechEvents : public IDispatch
    {
    };
    
#else 	/* C style interface */

    typedef struct SpeechEventsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            SpeechEvents * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            SpeechEvents * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            SpeechEvents * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            SpeechEvents * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            SpeechEvents * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            SpeechEvents * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            SpeechEvents * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        END_INTERFACE
    } SpeechEventsVtbl;

    interface SpeechEvents
    {
        CONST_VTBL struct SpeechEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define SpeechEvents_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define SpeechEvents_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define SpeechEvents_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define SpeechEvents_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define SpeechEvents_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define SpeechEvents_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define SpeechEvents_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */


#endif 	/* __SpeechEvents_DISPINTERFACE_DEFINED__ */


#ifndef ___MouseCursorEvents_INTERFACE_DEFINED__
#define ___MouseCursorEvents_INTERFACE_DEFINED__

/* interface _MouseCursorEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID IID__MouseCursorEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("737AF3CD-F235-4D32-8263-5D918BA44E3F")
    _MouseCursorEvents
    {
    public:
        BEGIN_INTERFACE
        END_INTERFACE
    };
    
    
#else 	/* C style interface */

    typedef struct _MouseCursorEventsVtbl
    {
        BEGIN_INTERFACE
        
        END_INTERFACE
    } _MouseCursorEventsVtbl;

    interface _MouseCursorEvents
    {
        CONST_VTBL struct _MouseCursorEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___MouseCursorEvents_INTERFACE_DEFINED__ */


#ifndef __MouseCursorEvents_DISPINTERFACE_DEFINED__
#define __MouseCursorEvents_DISPINTERFACE_DEFINED__

/* dispinterface MouseCursorEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID DIID_MouseCursorEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)

    MIDL_INTERFACE("3DAC6012-CCAE-4BC3-923B-1D57DA9BA5E9")
    MouseCursorEvents : public IDispatch
    {
    };
    
#else 	/* C style interface */

    typedef struct MouseCursorEventsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            MouseCursorEvents * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            MouseCursorEvents * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            MouseCursorEvents * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            MouseCursorEvents * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            MouseCursorEvents * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            MouseCursorEvents * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            MouseCursorEvents * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        END_INTERFACE
    } MouseCursorEventsVtbl;

    interface MouseCursorEvents
    {
        CONST_VTBL struct MouseCursorEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define MouseCursorEvents_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define MouseCursorEvents_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define MouseCursorEvents_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define MouseCursorEvents_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define MouseCursorEvents_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define MouseCursorEvents_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define MouseCursorEvents_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */


#endif 	/* __MouseCursorEvents_DISPINTERFACE_DEFINED__ */


#ifndef ___EditBox_INTERFACE_DEFINED__
#define ___EditBox_INTERFACE_DEFINED__

/* interface _EditBox */
/* [object][oleautomation][dual][hidden][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__EditBox;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("F38F6DFC-06E8-42A0-8C64-9138FC36637D")
    _EditBox : public _Control
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FirstVisibleLine( 
            /* [retval][out] */ long *Line) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TextLimit( 
            /* [retval][out] */ long *limit) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TextLimit( 
            /* [in] */ long limit) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LineCount( 
            /* [retval][out] */ long *Lines) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SelectionStart( 
            /* [retval][out] */ long *selStart) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SelectionStart( 
            /* [in] */ long selStart) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SelectionEnd( 
            /* [retval][out] */ long *selEnd) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SelectionEnd( 
            /* [in] */ long selEnd) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ReadOnly( 
            /* [retval][out] */ VARIANT_BOOL *ReadOnly) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ReadOnly( 
            /* [in] */ VARIANT_BOOL ReadOnly) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TextAlignment( 
            /* [retval][out] */ TextAlignment *alignment) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TextAlignment( 
            /* [in] */ TextAlignment alignment) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Multiline( 
            /* [retval][out] */ VARIANT_BOOL *Multiline) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Password( 
            /* [retval][out] */ VARIANT_BOOL *Password) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Cursor( 
            /* [retval][out] */ long *Cursor) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Cursor( 
            /* [in] */ long Cursor) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Selection( 
            /* [retval][out] */ BSTR *Selection) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE CharAtPoint( 
            /* [in] */ ClientPoint *point,
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Line( 
            /* [in] */ long lineNumber,
            /* [retval][out] */ BSTR *Text) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE LineFromChar( 
            /* [in] */ long index,
            /* [retval][out] */ long *lineNumber) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE CharFromLine( 
            /* [in] */ long lineNumber,
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE LineLength( 
            /* [in] */ long lineNumber,
            /* [retval][out] */ long *Length) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ScrollBy( 
            /* [in] */ long X,
            /* [in] */ long Y) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE CharacterPosition( 
            /* [in] */ long index,
            /* [retval][out] */ ClientPoint **point) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Replace( 
            /* [in] */ BSTR Text) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ScrollToCursor( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE SelectAll( void) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _EditBoxVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _EditBox * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _EditBox * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _EditBox * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _EditBox * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _EditBox * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _EditBox * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _EditBox * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            _EditBox * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            _EditBox * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _EditBox * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _EditBox * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            _EditBox * This,
            /* [retval][out] */ ClientPoint **Position);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            _EditBox * This,
            /* [in] */ ClientPoint *Position);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Enabled )( 
            _EditBox * This,
            /* [retval][out] */ VARIANT_BOOL *Enabled);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Enabled )( 
            _EditBox * This,
            /* [in] */ VARIANT_BOOL Enabled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            _EditBox * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Visible )( 
            _EditBox * This,
            /* [in] */ VARIANT_BOOL Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            _EditBox * This,
            /* [retval][out] */ VARIANT_BOOL *Focused);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Width )( 
            _EditBox * This,
            /* [retval][out] */ long *Width);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Width )( 
            _EditBox * This,
            /* [in] */ long Width);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Height )( 
            _EditBox * This,
            /* [retval][out] */ long *Height);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Height )( 
            _EditBox * This,
            /* [in] */ long Height);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabOrder )( 
            _EditBox * This,
            /* [retval][out] */ long *TabOrder);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabOrder )( 
            _EditBox * This,
            /* [in] */ long TabOrder);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabStop )( 
            _EditBox * This,
            /* [retval][out] */ VARIANT_BOOL *TabStop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabStop )( 
            _EditBox * This,
            /* [in] */ VARIANT_BOOL TabStop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Group )( 
            _EditBox * This,
            /* [retval][out] */ VARIANT_BOOL *Group);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Group )( 
            _EditBox * This,
            /* [in] */ VARIANT_BOOL Group);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            _EditBox * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            _EditBox * This,
            /* [retval][out] */ long *id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Next )( 
            _EditBox * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Previous )( 
            _EditBox * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Focus )( 
            _EditBox * This);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_id )( 
            _EditBox * This,
            /* [in] */ long id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FirstVisibleLine )( 
            _EditBox * This,
            /* [retval][out] */ long *Line);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TextLimit )( 
            _EditBox * This,
            /* [retval][out] */ long *limit);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TextLimit )( 
            _EditBox * This,
            /* [in] */ long limit);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LineCount )( 
            _EditBox * This,
            /* [retval][out] */ long *Lines);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SelectionStart )( 
            _EditBox * This,
            /* [retval][out] */ long *selStart);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SelectionStart )( 
            _EditBox * This,
            /* [in] */ long selStart);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SelectionEnd )( 
            _EditBox * This,
            /* [retval][out] */ long *selEnd);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SelectionEnd )( 
            _EditBox * This,
            /* [in] */ long selEnd);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ReadOnly )( 
            _EditBox * This,
            /* [retval][out] */ VARIANT_BOOL *ReadOnly);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ReadOnly )( 
            _EditBox * This,
            /* [in] */ VARIANT_BOOL ReadOnly);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TextAlignment )( 
            _EditBox * This,
            /* [retval][out] */ TextAlignment *alignment);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TextAlignment )( 
            _EditBox * This,
            /* [in] */ TextAlignment alignment);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Multiline )( 
            _EditBox * This,
            /* [retval][out] */ VARIANT_BOOL *Multiline);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Password )( 
            _EditBox * This,
            /* [retval][out] */ VARIANT_BOOL *Password);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Cursor )( 
            _EditBox * This,
            /* [retval][out] */ long *Cursor);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Cursor )( 
            _EditBox * This,
            /* [in] */ long Cursor);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Selection )( 
            _EditBox * This,
            /* [retval][out] */ BSTR *Selection);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *CharAtPoint )( 
            _EditBox * This,
            /* [in] */ ClientPoint *point,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Line )( 
            _EditBox * This,
            /* [in] */ long lineNumber,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *LineFromChar )( 
            _EditBox * This,
            /* [in] */ long index,
            /* [retval][out] */ long *lineNumber);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *CharFromLine )( 
            _EditBox * This,
            /* [in] */ long lineNumber,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *LineLength )( 
            _EditBox * This,
            /* [in] */ long lineNumber,
            /* [retval][out] */ long *Length);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ScrollBy )( 
            _EditBox * This,
            /* [in] */ long X,
            /* [in] */ long Y);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *CharacterPosition )( 
            _EditBox * This,
            /* [in] */ long index,
            /* [retval][out] */ ClientPoint **point);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Replace )( 
            _EditBox * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ScrollToCursor )( 
            _EditBox * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *SelectAll )( 
            _EditBox * This);
        
        END_INTERFACE
    } _EditBoxVtbl;

    interface _EditBox
    {
        CONST_VTBL struct _EditBoxVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _EditBox_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _EditBox_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _EditBox_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _EditBox_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _EditBox_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _EditBox_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _EditBox_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _EditBox_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define _EditBox_put_Text(This,Text)	\
    ( (This)->lpVtbl -> put_Text(This,Text) ) 

#define _EditBox_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _EditBox_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _EditBox_get_Position(This,Position)	\
    ( (This)->lpVtbl -> get_Position(This,Position) ) 

#define _EditBox_put_Position(This,Position)	\
    ( (This)->lpVtbl -> put_Position(This,Position) ) 

#define _EditBox_get_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> get_Enabled(This,Enabled) ) 

#define _EditBox_put_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> put_Enabled(This,Enabled) ) 

#define _EditBox_get_Visible(This,Visible)	\
    ( (This)->lpVtbl -> get_Visible(This,Visible) ) 

#define _EditBox_put_Visible(This,Visible)	\
    ( (This)->lpVtbl -> put_Visible(This,Visible) ) 

#define _EditBox_get_Focused(This,Focused)	\
    ( (This)->lpVtbl -> get_Focused(This,Focused) ) 

#define _EditBox_get_Width(This,Width)	\
    ( (This)->lpVtbl -> get_Width(This,Width) ) 

#define _EditBox_put_Width(This,Width)	\
    ( (This)->lpVtbl -> put_Width(This,Width) ) 

#define _EditBox_get_Height(This,Height)	\
    ( (This)->lpVtbl -> get_Height(This,Height) ) 

#define _EditBox_put_Height(This,Height)	\
    ( (This)->lpVtbl -> put_Height(This,Height) ) 

#define _EditBox_get_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> get_TabOrder(This,TabOrder) ) 

#define _EditBox_put_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> put_TabOrder(This,TabOrder) ) 

#define _EditBox_get_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> get_TabStop(This,TabStop) ) 

#define _EditBox_put_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> put_TabStop(This,TabStop) ) 

#define _EditBox_get_Group(This,Group)	\
    ( (This)->lpVtbl -> get_Group(This,Group) ) 

#define _EditBox_put_Group(This,Group)	\
    ( (This)->lpVtbl -> put_Group(This,Group) ) 

#define _EditBox_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define _EditBox_get_id(This,id)	\
    ( (This)->lpVtbl -> get_id(This,id) ) 

#define _EditBox_get_Next(This,prop)	\
    ( (This)->lpVtbl -> get_Next(This,prop) ) 

#define _EditBox_get_Previous(This,prop)	\
    ( (This)->lpVtbl -> get_Previous(This,prop) ) 

#define _EditBox_Focus(This)	\
    ( (This)->lpVtbl -> Focus(This) ) 

#define _EditBox_put_id(This,id)	\
    ( (This)->lpVtbl -> put_id(This,id) ) 


#define _EditBox_get_FirstVisibleLine(This,Line)	\
    ( (This)->lpVtbl -> get_FirstVisibleLine(This,Line) ) 

#define _EditBox_get_TextLimit(This,limit)	\
    ( (This)->lpVtbl -> get_TextLimit(This,limit) ) 

#define _EditBox_put_TextLimit(This,limit)	\
    ( (This)->lpVtbl -> put_TextLimit(This,limit) ) 

#define _EditBox_get_LineCount(This,Lines)	\
    ( (This)->lpVtbl -> get_LineCount(This,Lines) ) 

#define _EditBox_get_SelectionStart(This,selStart)	\
    ( (This)->lpVtbl -> get_SelectionStart(This,selStart) ) 

#define _EditBox_put_SelectionStart(This,selStart)	\
    ( (This)->lpVtbl -> put_SelectionStart(This,selStart) ) 

#define _EditBox_get_SelectionEnd(This,selEnd)	\
    ( (This)->lpVtbl -> get_SelectionEnd(This,selEnd) ) 

#define _EditBox_put_SelectionEnd(This,selEnd)	\
    ( (This)->lpVtbl -> put_SelectionEnd(This,selEnd) ) 

#define _EditBox_get_ReadOnly(This,ReadOnly)	\
    ( (This)->lpVtbl -> get_ReadOnly(This,ReadOnly) ) 

#define _EditBox_put_ReadOnly(This,ReadOnly)	\
    ( (This)->lpVtbl -> put_ReadOnly(This,ReadOnly) ) 

#define _EditBox_get_TextAlignment(This,alignment)	\
    ( (This)->lpVtbl -> get_TextAlignment(This,alignment) ) 

#define _EditBox_put_TextAlignment(This,alignment)	\
    ( (This)->lpVtbl -> put_TextAlignment(This,alignment) ) 

#define _EditBox_get_Multiline(This,Multiline)	\
    ( (This)->lpVtbl -> get_Multiline(This,Multiline) ) 

#define _EditBox_get_Password(This,Password)	\
    ( (This)->lpVtbl -> get_Password(This,Password) ) 

#define _EditBox_get_Cursor(This,Cursor)	\
    ( (This)->lpVtbl -> get_Cursor(This,Cursor) ) 

#define _EditBox_put_Cursor(This,Cursor)	\
    ( (This)->lpVtbl -> put_Cursor(This,Cursor) ) 

#define _EditBox_get_Selection(This,Selection)	\
    ( (This)->lpVtbl -> get_Selection(This,Selection) ) 

#define _EditBox_CharAtPoint(This,point,index)	\
    ( (This)->lpVtbl -> CharAtPoint(This,point,index) ) 

#define _EditBox_Line(This,lineNumber,Text)	\
    ( (This)->lpVtbl -> Line(This,lineNumber,Text) ) 

#define _EditBox_LineFromChar(This,index,lineNumber)	\
    ( (This)->lpVtbl -> LineFromChar(This,index,lineNumber) ) 

#define _EditBox_CharFromLine(This,lineNumber,index)	\
    ( (This)->lpVtbl -> CharFromLine(This,lineNumber,index) ) 

#define _EditBox_LineLength(This,lineNumber,Length)	\
    ( (This)->lpVtbl -> LineLength(This,lineNumber,Length) ) 

#define _EditBox_ScrollBy(This,X,Y)	\
    ( (This)->lpVtbl -> ScrollBy(This,X,Y) ) 

#define _EditBox_CharacterPosition(This,index,point)	\
    ( (This)->lpVtbl -> CharacterPosition(This,index,point) ) 

#define _EditBox_Replace(This,Text)	\
    ( (This)->lpVtbl -> Replace(This,Text) ) 

#define _EditBox_ScrollToCursor(This)	\
    ( (This)->lpVtbl -> ScrollToCursor(This) ) 

#define _EditBox_SelectAll(This)	\
    ( (This)->lpVtbl -> SelectAll(This) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___EditBox_INTERFACE_DEFINED__ */


#ifndef ___Pushbutton_INTERFACE_DEFINED__
#define ___Pushbutton_INTERFACE_DEFINED__

/* interface _Pushbutton */
/* [object][oleautomation][dual][hidden][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__Pushbutton;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("F7125E61-8E5D-44AA-86DC-E2BC92D40C73")
    _Pushbutton : public _Control
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TextAlignment( 
            /* [retval][out] */ TextAlignment *align) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TextAlignment( 
            /* [in] */ TextAlignment align) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_VerticalAlignment( 
            /* [retval][out] */ VerticalAlignment *align) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_VerticalAlignment( 
            /* [in] */ VerticalAlignment align) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Default( 
            /* [retval][out] */ VARIANT_BOOL *def) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Default( 
            /* [in] */ VARIANT_BOOL def) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Click( void) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _PushbuttonVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _Pushbutton * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _Pushbutton * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _Pushbutton * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _Pushbutton * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _Pushbutton * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _Pushbutton * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _Pushbutton * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            _Pushbutton * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            _Pushbutton * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _Pushbutton * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _Pushbutton * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            _Pushbutton * This,
            /* [retval][out] */ ClientPoint **Position);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            _Pushbutton * This,
            /* [in] */ ClientPoint *Position);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Enabled )( 
            _Pushbutton * This,
            /* [retval][out] */ VARIANT_BOOL *Enabled);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Enabled )( 
            _Pushbutton * This,
            /* [in] */ VARIANT_BOOL Enabled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            _Pushbutton * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Visible )( 
            _Pushbutton * This,
            /* [in] */ VARIANT_BOOL Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            _Pushbutton * This,
            /* [retval][out] */ VARIANT_BOOL *Focused);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Width )( 
            _Pushbutton * This,
            /* [retval][out] */ long *Width);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Width )( 
            _Pushbutton * This,
            /* [in] */ long Width);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Height )( 
            _Pushbutton * This,
            /* [retval][out] */ long *Height);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Height )( 
            _Pushbutton * This,
            /* [in] */ long Height);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabOrder )( 
            _Pushbutton * This,
            /* [retval][out] */ long *TabOrder);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabOrder )( 
            _Pushbutton * This,
            /* [in] */ long TabOrder);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabStop )( 
            _Pushbutton * This,
            /* [retval][out] */ VARIANT_BOOL *TabStop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabStop )( 
            _Pushbutton * This,
            /* [in] */ VARIANT_BOOL TabStop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Group )( 
            _Pushbutton * This,
            /* [retval][out] */ VARIANT_BOOL *Group);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Group )( 
            _Pushbutton * This,
            /* [in] */ VARIANT_BOOL Group);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            _Pushbutton * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            _Pushbutton * This,
            /* [retval][out] */ long *id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Next )( 
            _Pushbutton * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Previous )( 
            _Pushbutton * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Focus )( 
            _Pushbutton * This);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_id )( 
            _Pushbutton * This,
            /* [in] */ long id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TextAlignment )( 
            _Pushbutton * This,
            /* [retval][out] */ TextAlignment *align);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TextAlignment )( 
            _Pushbutton * This,
            /* [in] */ TextAlignment align);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_VerticalAlignment )( 
            _Pushbutton * This,
            /* [retval][out] */ VerticalAlignment *align);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_VerticalAlignment )( 
            _Pushbutton * This,
            /* [in] */ VerticalAlignment align);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Default )( 
            _Pushbutton * This,
            /* [retval][out] */ VARIANT_BOOL *def);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Default )( 
            _Pushbutton * This,
            /* [in] */ VARIANT_BOOL def);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Click )( 
            _Pushbutton * This);
        
        END_INTERFACE
    } _PushbuttonVtbl;

    interface _Pushbutton
    {
        CONST_VTBL struct _PushbuttonVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _Pushbutton_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _Pushbutton_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _Pushbutton_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _Pushbutton_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _Pushbutton_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _Pushbutton_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _Pushbutton_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _Pushbutton_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define _Pushbutton_put_Text(This,Text)	\
    ( (This)->lpVtbl -> put_Text(This,Text) ) 

#define _Pushbutton_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _Pushbutton_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _Pushbutton_get_Position(This,Position)	\
    ( (This)->lpVtbl -> get_Position(This,Position) ) 

#define _Pushbutton_put_Position(This,Position)	\
    ( (This)->lpVtbl -> put_Position(This,Position) ) 

#define _Pushbutton_get_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> get_Enabled(This,Enabled) ) 

#define _Pushbutton_put_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> put_Enabled(This,Enabled) ) 

#define _Pushbutton_get_Visible(This,Visible)	\
    ( (This)->lpVtbl -> get_Visible(This,Visible) ) 

#define _Pushbutton_put_Visible(This,Visible)	\
    ( (This)->lpVtbl -> put_Visible(This,Visible) ) 

#define _Pushbutton_get_Focused(This,Focused)	\
    ( (This)->lpVtbl -> get_Focused(This,Focused) ) 

#define _Pushbutton_get_Width(This,Width)	\
    ( (This)->lpVtbl -> get_Width(This,Width) ) 

#define _Pushbutton_put_Width(This,Width)	\
    ( (This)->lpVtbl -> put_Width(This,Width) ) 

#define _Pushbutton_get_Height(This,Height)	\
    ( (This)->lpVtbl -> get_Height(This,Height) ) 

#define _Pushbutton_put_Height(This,Height)	\
    ( (This)->lpVtbl -> put_Height(This,Height) ) 

#define _Pushbutton_get_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> get_TabOrder(This,TabOrder) ) 

#define _Pushbutton_put_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> put_TabOrder(This,TabOrder) ) 

#define _Pushbutton_get_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> get_TabStop(This,TabStop) ) 

#define _Pushbutton_put_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> put_TabStop(This,TabStop) ) 

#define _Pushbutton_get_Group(This,Group)	\
    ( (This)->lpVtbl -> get_Group(This,Group) ) 

#define _Pushbutton_put_Group(This,Group)	\
    ( (This)->lpVtbl -> put_Group(This,Group) ) 

#define _Pushbutton_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define _Pushbutton_get_id(This,id)	\
    ( (This)->lpVtbl -> get_id(This,id) ) 

#define _Pushbutton_get_Next(This,prop)	\
    ( (This)->lpVtbl -> get_Next(This,prop) ) 

#define _Pushbutton_get_Previous(This,prop)	\
    ( (This)->lpVtbl -> get_Previous(This,prop) ) 

#define _Pushbutton_Focus(This)	\
    ( (This)->lpVtbl -> Focus(This) ) 

#define _Pushbutton_put_id(This,id)	\
    ( (This)->lpVtbl -> put_id(This,id) ) 


#define _Pushbutton_get_TextAlignment(This,align)	\
    ( (This)->lpVtbl -> get_TextAlignment(This,align) ) 

#define _Pushbutton_put_TextAlignment(This,align)	\
    ( (This)->lpVtbl -> put_TextAlignment(This,align) ) 

#define _Pushbutton_get_VerticalAlignment(This,align)	\
    ( (This)->lpVtbl -> get_VerticalAlignment(This,align) ) 

#define _Pushbutton_put_VerticalAlignment(This,align)	\
    ( (This)->lpVtbl -> put_VerticalAlignment(This,align) ) 

#define _Pushbutton_get_Default(This,def)	\
    ( (This)->lpVtbl -> get_Default(This,def) ) 

#define _Pushbutton_put_Default(This,def)	\
    ( (This)->lpVtbl -> put_Default(This,def) ) 

#define _Pushbutton_Click(This)	\
    ( (This)->lpVtbl -> Click(This) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___Pushbutton_INTERFACE_DEFINED__ */


#ifndef ___CheckBox_INTERFACE_DEFINED__
#define ___CheckBox_INTERFACE_DEFINED__

/* interface _CheckBox */
/* [object][oleautomation][dual][hidden][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__CheckBox;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("FDF989F9-5D53-46B1-A31B-4E1EC662D316")
    _CheckBox : public _Control
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TextAlignment( 
            /* [retval][out] */ TextAlignment *align) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TextAlignment( 
            /* [in] */ TextAlignment align) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ThreeState( 
            /* [retval][out] */ VARIANT_BOOL *ThreeState) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ThreeState( 
            /* [in] */ VARIANT_BOOL ThreeState) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LeftText( 
            /* [retval][out] */ VARIANT_BOOL *LeftText) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_LeftText( 
            /* [in] */ VARIANT_BOOL LeftText) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Checked( 
            /* [retval][out] */ CheckState *Checked) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Checked( 
            /* [in] */ CheckState Checked) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Click( void) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _CheckBoxVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _CheckBox * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _CheckBox * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _CheckBox * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _CheckBox * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _CheckBox * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _CheckBox * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _CheckBox * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            _CheckBox * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            _CheckBox * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _CheckBox * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _CheckBox * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            _CheckBox * This,
            /* [retval][out] */ ClientPoint **Position);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            _CheckBox * This,
            /* [in] */ ClientPoint *Position);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Enabled )( 
            _CheckBox * This,
            /* [retval][out] */ VARIANT_BOOL *Enabled);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Enabled )( 
            _CheckBox * This,
            /* [in] */ VARIANT_BOOL Enabled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            _CheckBox * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Visible )( 
            _CheckBox * This,
            /* [in] */ VARIANT_BOOL Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            _CheckBox * This,
            /* [retval][out] */ VARIANT_BOOL *Focused);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Width )( 
            _CheckBox * This,
            /* [retval][out] */ long *Width);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Width )( 
            _CheckBox * This,
            /* [in] */ long Width);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Height )( 
            _CheckBox * This,
            /* [retval][out] */ long *Height);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Height )( 
            _CheckBox * This,
            /* [in] */ long Height);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabOrder )( 
            _CheckBox * This,
            /* [retval][out] */ long *TabOrder);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabOrder )( 
            _CheckBox * This,
            /* [in] */ long TabOrder);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabStop )( 
            _CheckBox * This,
            /* [retval][out] */ VARIANT_BOOL *TabStop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabStop )( 
            _CheckBox * This,
            /* [in] */ VARIANT_BOOL TabStop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Group )( 
            _CheckBox * This,
            /* [retval][out] */ VARIANT_BOOL *Group);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Group )( 
            _CheckBox * This,
            /* [in] */ VARIANT_BOOL Group);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            _CheckBox * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            _CheckBox * This,
            /* [retval][out] */ long *id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Next )( 
            _CheckBox * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Previous )( 
            _CheckBox * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Focus )( 
            _CheckBox * This);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_id )( 
            _CheckBox * This,
            /* [in] */ long id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TextAlignment )( 
            _CheckBox * This,
            /* [retval][out] */ TextAlignment *align);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TextAlignment )( 
            _CheckBox * This,
            /* [in] */ TextAlignment align);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ThreeState )( 
            _CheckBox * This,
            /* [retval][out] */ VARIANT_BOOL *ThreeState);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ThreeState )( 
            _CheckBox * This,
            /* [in] */ VARIANT_BOOL ThreeState);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LeftText )( 
            _CheckBox * This,
            /* [retval][out] */ VARIANT_BOOL *LeftText);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_LeftText )( 
            _CheckBox * This,
            /* [in] */ VARIANT_BOOL LeftText);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Checked )( 
            _CheckBox * This,
            /* [retval][out] */ CheckState *Checked);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Checked )( 
            _CheckBox * This,
            /* [in] */ CheckState Checked);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Click )( 
            _CheckBox * This);
        
        END_INTERFACE
    } _CheckBoxVtbl;

    interface _CheckBox
    {
        CONST_VTBL struct _CheckBoxVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _CheckBox_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _CheckBox_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _CheckBox_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _CheckBox_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _CheckBox_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _CheckBox_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _CheckBox_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _CheckBox_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define _CheckBox_put_Text(This,Text)	\
    ( (This)->lpVtbl -> put_Text(This,Text) ) 

#define _CheckBox_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _CheckBox_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _CheckBox_get_Position(This,Position)	\
    ( (This)->lpVtbl -> get_Position(This,Position) ) 

#define _CheckBox_put_Position(This,Position)	\
    ( (This)->lpVtbl -> put_Position(This,Position) ) 

#define _CheckBox_get_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> get_Enabled(This,Enabled) ) 

#define _CheckBox_put_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> put_Enabled(This,Enabled) ) 

#define _CheckBox_get_Visible(This,Visible)	\
    ( (This)->lpVtbl -> get_Visible(This,Visible) ) 

#define _CheckBox_put_Visible(This,Visible)	\
    ( (This)->lpVtbl -> put_Visible(This,Visible) ) 

#define _CheckBox_get_Focused(This,Focused)	\
    ( (This)->lpVtbl -> get_Focused(This,Focused) ) 

#define _CheckBox_get_Width(This,Width)	\
    ( (This)->lpVtbl -> get_Width(This,Width) ) 

#define _CheckBox_put_Width(This,Width)	\
    ( (This)->lpVtbl -> put_Width(This,Width) ) 

#define _CheckBox_get_Height(This,Height)	\
    ( (This)->lpVtbl -> get_Height(This,Height) ) 

#define _CheckBox_put_Height(This,Height)	\
    ( (This)->lpVtbl -> put_Height(This,Height) ) 

#define _CheckBox_get_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> get_TabOrder(This,TabOrder) ) 

#define _CheckBox_put_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> put_TabOrder(This,TabOrder) ) 

#define _CheckBox_get_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> get_TabStop(This,TabStop) ) 

#define _CheckBox_put_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> put_TabStop(This,TabStop) ) 

#define _CheckBox_get_Group(This,Group)	\
    ( (This)->lpVtbl -> get_Group(This,Group) ) 

#define _CheckBox_put_Group(This,Group)	\
    ( (This)->lpVtbl -> put_Group(This,Group) ) 

#define _CheckBox_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define _CheckBox_get_id(This,id)	\
    ( (This)->lpVtbl -> get_id(This,id) ) 

#define _CheckBox_get_Next(This,prop)	\
    ( (This)->lpVtbl -> get_Next(This,prop) ) 

#define _CheckBox_get_Previous(This,prop)	\
    ( (This)->lpVtbl -> get_Previous(This,prop) ) 

#define _CheckBox_Focus(This)	\
    ( (This)->lpVtbl -> Focus(This) ) 

#define _CheckBox_put_id(This,id)	\
    ( (This)->lpVtbl -> put_id(This,id) ) 


#define _CheckBox_get_TextAlignment(This,align)	\
    ( (This)->lpVtbl -> get_TextAlignment(This,align) ) 

#define _CheckBox_put_TextAlignment(This,align)	\
    ( (This)->lpVtbl -> put_TextAlignment(This,align) ) 

#define _CheckBox_get_ThreeState(This,ThreeState)	\
    ( (This)->lpVtbl -> get_ThreeState(This,ThreeState) ) 

#define _CheckBox_put_ThreeState(This,ThreeState)	\
    ( (This)->lpVtbl -> put_ThreeState(This,ThreeState) ) 

#define _CheckBox_get_LeftText(This,LeftText)	\
    ( (This)->lpVtbl -> get_LeftText(This,LeftText) ) 

#define _CheckBox_put_LeftText(This,LeftText)	\
    ( (This)->lpVtbl -> put_LeftText(This,LeftText) ) 

#define _CheckBox_get_Checked(This,Checked)	\
    ( (This)->lpVtbl -> get_Checked(This,Checked) ) 

#define _CheckBox_put_Checked(This,Checked)	\
    ( (This)->lpVtbl -> put_Checked(This,Checked) ) 

#define _CheckBox_Click(This)	\
    ( (This)->lpVtbl -> Click(This) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___CheckBox_INTERFACE_DEFINED__ */


#ifndef ___RadioButton_INTERFACE_DEFINED__
#define ___RadioButton_INTERFACE_DEFINED__

/* interface _RadioButton */
/* [object][oleautomation][dual][hidden][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__RadioButton;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("C8508D98-ADC2-47E3-A51E-CEB0436592DF")
    _RadioButton : public _Control
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TextAlignment( 
            /* [retval][out] */ TextAlignment *align) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TextAlignment( 
            /* [in] */ TextAlignment align) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_LeftText( 
            /* [retval][out] */ VARIANT_BOOL *LeftText) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_LeftText( 
            /* [in] */ VARIANT_BOOL LeftText) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Checked( 
            /* [retval][out] */ CheckState *Checked) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Checked( 
            /* [in] */ CheckState Checked) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Click( void) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _RadioButtonVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _RadioButton * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _RadioButton * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _RadioButton * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _RadioButton * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _RadioButton * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _RadioButton * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _RadioButton * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            _RadioButton * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            _RadioButton * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _RadioButton * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _RadioButton * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            _RadioButton * This,
            /* [retval][out] */ ClientPoint **Position);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            _RadioButton * This,
            /* [in] */ ClientPoint *Position);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Enabled )( 
            _RadioButton * This,
            /* [retval][out] */ VARIANT_BOOL *Enabled);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Enabled )( 
            _RadioButton * This,
            /* [in] */ VARIANT_BOOL Enabled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            _RadioButton * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Visible )( 
            _RadioButton * This,
            /* [in] */ VARIANT_BOOL Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            _RadioButton * This,
            /* [retval][out] */ VARIANT_BOOL *Focused);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Width )( 
            _RadioButton * This,
            /* [retval][out] */ long *Width);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Width )( 
            _RadioButton * This,
            /* [in] */ long Width);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Height )( 
            _RadioButton * This,
            /* [retval][out] */ long *Height);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Height )( 
            _RadioButton * This,
            /* [in] */ long Height);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabOrder )( 
            _RadioButton * This,
            /* [retval][out] */ long *TabOrder);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabOrder )( 
            _RadioButton * This,
            /* [in] */ long TabOrder);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabStop )( 
            _RadioButton * This,
            /* [retval][out] */ VARIANT_BOOL *TabStop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabStop )( 
            _RadioButton * This,
            /* [in] */ VARIANT_BOOL TabStop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Group )( 
            _RadioButton * This,
            /* [retval][out] */ VARIANT_BOOL *Group);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Group )( 
            _RadioButton * This,
            /* [in] */ VARIANT_BOOL Group);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            _RadioButton * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            _RadioButton * This,
            /* [retval][out] */ long *id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Next )( 
            _RadioButton * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Previous )( 
            _RadioButton * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Focus )( 
            _RadioButton * This);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_id )( 
            _RadioButton * This,
            /* [in] */ long id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TextAlignment )( 
            _RadioButton * This,
            /* [retval][out] */ TextAlignment *align);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TextAlignment )( 
            _RadioButton * This,
            /* [in] */ TextAlignment align);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_LeftText )( 
            _RadioButton * This,
            /* [retval][out] */ VARIANT_BOOL *LeftText);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_LeftText )( 
            _RadioButton * This,
            /* [in] */ VARIANT_BOOL LeftText);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Checked )( 
            _RadioButton * This,
            /* [retval][out] */ CheckState *Checked);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Checked )( 
            _RadioButton * This,
            /* [in] */ CheckState Checked);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Click )( 
            _RadioButton * This);
        
        END_INTERFACE
    } _RadioButtonVtbl;

    interface _RadioButton
    {
        CONST_VTBL struct _RadioButtonVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _RadioButton_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _RadioButton_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _RadioButton_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _RadioButton_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _RadioButton_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _RadioButton_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _RadioButton_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _RadioButton_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define _RadioButton_put_Text(This,Text)	\
    ( (This)->lpVtbl -> put_Text(This,Text) ) 

#define _RadioButton_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _RadioButton_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _RadioButton_get_Position(This,Position)	\
    ( (This)->lpVtbl -> get_Position(This,Position) ) 

#define _RadioButton_put_Position(This,Position)	\
    ( (This)->lpVtbl -> put_Position(This,Position) ) 

#define _RadioButton_get_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> get_Enabled(This,Enabled) ) 

#define _RadioButton_put_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> put_Enabled(This,Enabled) ) 

#define _RadioButton_get_Visible(This,Visible)	\
    ( (This)->lpVtbl -> get_Visible(This,Visible) ) 

#define _RadioButton_put_Visible(This,Visible)	\
    ( (This)->lpVtbl -> put_Visible(This,Visible) ) 

#define _RadioButton_get_Focused(This,Focused)	\
    ( (This)->lpVtbl -> get_Focused(This,Focused) ) 

#define _RadioButton_get_Width(This,Width)	\
    ( (This)->lpVtbl -> get_Width(This,Width) ) 

#define _RadioButton_put_Width(This,Width)	\
    ( (This)->lpVtbl -> put_Width(This,Width) ) 

#define _RadioButton_get_Height(This,Height)	\
    ( (This)->lpVtbl -> get_Height(This,Height) ) 

#define _RadioButton_put_Height(This,Height)	\
    ( (This)->lpVtbl -> put_Height(This,Height) ) 

#define _RadioButton_get_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> get_TabOrder(This,TabOrder) ) 

#define _RadioButton_put_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> put_TabOrder(This,TabOrder) ) 

#define _RadioButton_get_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> get_TabStop(This,TabStop) ) 

#define _RadioButton_put_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> put_TabStop(This,TabStop) ) 

#define _RadioButton_get_Group(This,Group)	\
    ( (This)->lpVtbl -> get_Group(This,Group) ) 

#define _RadioButton_put_Group(This,Group)	\
    ( (This)->lpVtbl -> put_Group(This,Group) ) 

#define _RadioButton_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define _RadioButton_get_id(This,id)	\
    ( (This)->lpVtbl -> get_id(This,id) ) 

#define _RadioButton_get_Next(This,prop)	\
    ( (This)->lpVtbl -> get_Next(This,prop) ) 

#define _RadioButton_get_Previous(This,prop)	\
    ( (This)->lpVtbl -> get_Previous(This,prop) ) 

#define _RadioButton_Focus(This)	\
    ( (This)->lpVtbl -> Focus(This) ) 

#define _RadioButton_put_id(This,id)	\
    ( (This)->lpVtbl -> put_id(This,id) ) 


#define _RadioButton_get_TextAlignment(This,align)	\
    ( (This)->lpVtbl -> get_TextAlignment(This,align) ) 

#define _RadioButton_put_TextAlignment(This,align)	\
    ( (This)->lpVtbl -> put_TextAlignment(This,align) ) 

#define _RadioButton_get_LeftText(This,LeftText)	\
    ( (This)->lpVtbl -> get_LeftText(This,LeftText) ) 

#define _RadioButton_put_LeftText(This,LeftText)	\
    ( (This)->lpVtbl -> put_LeftText(This,LeftText) ) 

#define _RadioButton_get_Checked(This,Checked)	\
    ( (This)->lpVtbl -> get_Checked(This,Checked) ) 

#define _RadioButton_put_Checked(This,Checked)	\
    ( (This)->lpVtbl -> put_Checked(This,Checked) ) 

#define _RadioButton_Click(This)	\
    ( (This)->lpVtbl -> Click(This) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___RadioButton_INTERFACE_DEFINED__ */


#ifndef __GroupBox_INTERFACE_DEFINED__
#define __GroupBox_INTERFACE_DEFINED__

/* interface GroupBox */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_GroupBox;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("6182416D-F3D5-4602-8B5A-47561BEE7ECB")
    GroupBox : public _Control
    {
    public:
    };
    
    
#else 	/* C style interface */

    typedef struct GroupBoxVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            GroupBox * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            GroupBox * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            GroupBox * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            GroupBox * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            GroupBox * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            GroupBox * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            GroupBox * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            GroupBox * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            GroupBox * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            GroupBox * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            GroupBox * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            GroupBox * This,
            /* [retval][out] */ ClientPoint **Position);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            GroupBox * This,
            /* [in] */ ClientPoint *Position);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Enabled )( 
            GroupBox * This,
            /* [retval][out] */ VARIANT_BOOL *Enabled);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Enabled )( 
            GroupBox * This,
            /* [in] */ VARIANT_BOOL Enabled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            GroupBox * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Visible )( 
            GroupBox * This,
            /* [in] */ VARIANT_BOOL Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            GroupBox * This,
            /* [retval][out] */ VARIANT_BOOL *Focused);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Width )( 
            GroupBox * This,
            /* [retval][out] */ long *Width);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Width )( 
            GroupBox * This,
            /* [in] */ long Width);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Height )( 
            GroupBox * This,
            /* [retval][out] */ long *Height);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Height )( 
            GroupBox * This,
            /* [in] */ long Height);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabOrder )( 
            GroupBox * This,
            /* [retval][out] */ long *TabOrder);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabOrder )( 
            GroupBox * This,
            /* [in] */ long TabOrder);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabStop )( 
            GroupBox * This,
            /* [retval][out] */ VARIANT_BOOL *TabStop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabStop )( 
            GroupBox * This,
            /* [in] */ VARIANT_BOOL TabStop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Group )( 
            GroupBox * This,
            /* [retval][out] */ VARIANT_BOOL *Group);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Group )( 
            GroupBox * This,
            /* [in] */ VARIANT_BOOL Group);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            GroupBox * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            GroupBox * This,
            /* [retval][out] */ long *id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Next )( 
            GroupBox * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Previous )( 
            GroupBox * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Focus )( 
            GroupBox * This);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_id )( 
            GroupBox * This,
            /* [in] */ long id);
        
        END_INTERFACE
    } GroupBoxVtbl;

    interface GroupBox
    {
        CONST_VTBL struct GroupBoxVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define GroupBox_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define GroupBox_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define GroupBox_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define GroupBox_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define GroupBox_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define GroupBox_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define GroupBox_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define GroupBox_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define GroupBox_put_Text(This,Text)	\
    ( (This)->lpVtbl -> put_Text(This,Text) ) 

#define GroupBox_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define GroupBox_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define GroupBox_get_Position(This,Position)	\
    ( (This)->lpVtbl -> get_Position(This,Position) ) 

#define GroupBox_put_Position(This,Position)	\
    ( (This)->lpVtbl -> put_Position(This,Position) ) 

#define GroupBox_get_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> get_Enabled(This,Enabled) ) 

#define GroupBox_put_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> put_Enabled(This,Enabled) ) 

#define GroupBox_get_Visible(This,Visible)	\
    ( (This)->lpVtbl -> get_Visible(This,Visible) ) 

#define GroupBox_put_Visible(This,Visible)	\
    ( (This)->lpVtbl -> put_Visible(This,Visible) ) 

#define GroupBox_get_Focused(This,Focused)	\
    ( (This)->lpVtbl -> get_Focused(This,Focused) ) 

#define GroupBox_get_Width(This,Width)	\
    ( (This)->lpVtbl -> get_Width(This,Width) ) 

#define GroupBox_put_Width(This,Width)	\
    ( (This)->lpVtbl -> put_Width(This,Width) ) 

#define GroupBox_get_Height(This,Height)	\
    ( (This)->lpVtbl -> get_Height(This,Height) ) 

#define GroupBox_put_Height(This,Height)	\
    ( (This)->lpVtbl -> put_Height(This,Height) ) 

#define GroupBox_get_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> get_TabOrder(This,TabOrder) ) 

#define GroupBox_put_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> put_TabOrder(This,TabOrder) ) 

#define GroupBox_get_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> get_TabStop(This,TabStop) ) 

#define GroupBox_put_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> put_TabStop(This,TabStop) ) 

#define GroupBox_get_Group(This,Group)	\
    ( (This)->lpVtbl -> get_Group(This,Group) ) 

#define GroupBox_put_Group(This,Group)	\
    ( (This)->lpVtbl -> put_Group(This,Group) ) 

#define GroupBox_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define GroupBox_get_id(This,id)	\
    ( (This)->lpVtbl -> get_id(This,id) ) 

#define GroupBox_get_Next(This,prop)	\
    ( (This)->lpVtbl -> get_Next(This,prop) ) 

#define GroupBox_get_Previous(This,prop)	\
    ( (This)->lpVtbl -> get_Previous(This,prop) ) 

#define GroupBox_Focus(This)	\
    ( (This)->lpVtbl -> Focus(This) ) 

#define GroupBox_put_id(This,id)	\
    ( (This)->lpVtbl -> put_id(This,id) ) 


#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __GroupBox_INTERFACE_DEFINED__ */


#ifndef __Static_INTERFACE_DEFINED__
#define __Static_INTERFACE_DEFINED__

/* interface Static */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_Static;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("CF8CAE4C-15C6-4226-8984-1C66352171DE")
    Static : public _Control
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TextAlignment( 
            /* [retval][out] */ TextAlignment *align) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TextAlignment( 
            /* [in] */ TextAlignment align) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_prefix( 
            /* [retval][out] */ VARIANT_BOOL *prefix) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_prefix( 
            /* [in] */ VARIANT_BOOL prefix) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct StaticVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            Static * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            Static * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            Static * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            Static * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            Static * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            Static * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            Static * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            Static * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            Static * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            Static * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            Static * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            Static * This,
            /* [retval][out] */ ClientPoint **Position);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            Static * This,
            /* [in] */ ClientPoint *Position);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Enabled )( 
            Static * This,
            /* [retval][out] */ VARIANT_BOOL *Enabled);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Enabled )( 
            Static * This,
            /* [in] */ VARIANT_BOOL Enabled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            Static * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Visible )( 
            Static * This,
            /* [in] */ VARIANT_BOOL Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            Static * This,
            /* [retval][out] */ VARIANT_BOOL *Focused);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Width )( 
            Static * This,
            /* [retval][out] */ long *Width);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Width )( 
            Static * This,
            /* [in] */ long Width);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Height )( 
            Static * This,
            /* [retval][out] */ long *Height);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Height )( 
            Static * This,
            /* [in] */ long Height);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabOrder )( 
            Static * This,
            /* [retval][out] */ long *TabOrder);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabOrder )( 
            Static * This,
            /* [in] */ long TabOrder);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabStop )( 
            Static * This,
            /* [retval][out] */ VARIANT_BOOL *TabStop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabStop )( 
            Static * This,
            /* [in] */ VARIANT_BOOL TabStop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Group )( 
            Static * This,
            /* [retval][out] */ VARIANT_BOOL *Group);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Group )( 
            Static * This,
            /* [in] */ VARIANT_BOOL Group);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            Static * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            Static * This,
            /* [retval][out] */ long *id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Next )( 
            Static * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Previous )( 
            Static * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Focus )( 
            Static * This);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_id )( 
            Static * This,
            /* [in] */ long id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TextAlignment )( 
            Static * This,
            /* [retval][out] */ TextAlignment *align);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TextAlignment )( 
            Static * This,
            /* [in] */ TextAlignment align);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_prefix )( 
            Static * This,
            /* [retval][out] */ VARIANT_BOOL *prefix);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_prefix )( 
            Static * This,
            /* [in] */ VARIANT_BOOL prefix);
        
        END_INTERFACE
    } StaticVtbl;

    interface Static
    {
        CONST_VTBL struct StaticVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define Static_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define Static_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define Static_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define Static_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define Static_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define Static_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define Static_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define Static_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define Static_put_Text(This,Text)	\
    ( (This)->lpVtbl -> put_Text(This,Text) ) 

#define Static_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define Static_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define Static_get_Position(This,Position)	\
    ( (This)->lpVtbl -> get_Position(This,Position) ) 

#define Static_put_Position(This,Position)	\
    ( (This)->lpVtbl -> put_Position(This,Position) ) 

#define Static_get_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> get_Enabled(This,Enabled) ) 

#define Static_put_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> put_Enabled(This,Enabled) ) 

#define Static_get_Visible(This,Visible)	\
    ( (This)->lpVtbl -> get_Visible(This,Visible) ) 

#define Static_put_Visible(This,Visible)	\
    ( (This)->lpVtbl -> put_Visible(This,Visible) ) 

#define Static_get_Focused(This,Focused)	\
    ( (This)->lpVtbl -> get_Focused(This,Focused) ) 

#define Static_get_Width(This,Width)	\
    ( (This)->lpVtbl -> get_Width(This,Width) ) 

#define Static_put_Width(This,Width)	\
    ( (This)->lpVtbl -> put_Width(This,Width) ) 

#define Static_get_Height(This,Height)	\
    ( (This)->lpVtbl -> get_Height(This,Height) ) 

#define Static_put_Height(This,Height)	\
    ( (This)->lpVtbl -> put_Height(This,Height) ) 

#define Static_get_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> get_TabOrder(This,TabOrder) ) 

#define Static_put_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> put_TabOrder(This,TabOrder) ) 

#define Static_get_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> get_TabStop(This,TabStop) ) 

#define Static_put_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> put_TabStop(This,TabStop) ) 

#define Static_get_Group(This,Group)	\
    ( (This)->lpVtbl -> get_Group(This,Group) ) 

#define Static_put_Group(This,Group)	\
    ( (This)->lpVtbl -> put_Group(This,Group) ) 

#define Static_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define Static_get_id(This,id)	\
    ( (This)->lpVtbl -> get_id(This,id) ) 

#define Static_get_Next(This,prop)	\
    ( (This)->lpVtbl -> get_Next(This,prop) ) 

#define Static_get_Previous(This,prop)	\
    ( (This)->lpVtbl -> get_Previous(This,prop) ) 

#define Static_Focus(This)	\
    ( (This)->lpVtbl -> Focus(This) ) 

#define Static_put_id(This,id)	\
    ( (This)->lpVtbl -> put_id(This,id) ) 


#define Static_get_TextAlignment(This,align)	\
    ( (This)->lpVtbl -> get_TextAlignment(This,align) ) 

#define Static_put_TextAlignment(This,align)	\
    ( (This)->lpVtbl -> put_TextAlignment(This,align) ) 

#define Static_get_prefix(This,prefix)	\
    ( (This)->lpVtbl -> get_prefix(This,prefix) ) 

#define Static_put_prefix(This,prefix)	\
    ( (This)->lpVtbl -> put_prefix(This,prefix) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __Static_INTERFACE_DEFINED__ */


#ifndef ___ComboBox_INTERFACE_DEFINED__
#define ___ComboBox_INTERFACE_DEFINED__

/* interface _ComboBox */
/* [object][oleautomation][dual][hidden][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__ComboBox;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("86E0723D-A77F-44D8-AE7B-427D4AEB9FA5")
    _ComboBox : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Text( 
            /* [optional][in] */ VARIANT index,
            /* [retval][out] */ BSTR *Text) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Text( 
            /* [optional][in] */ VARIANT index,
            /* [in] */ BSTR Text) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Position( 
            /* [retval][out] */ ClientPoint **Position) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Position( 
            /* [in] */ ClientPoint *Position) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Enabled( 
            /* [retval][out] */ VARIANT_BOOL *Enabled) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Enabled( 
            /* [in] */ VARIANT_BOOL Enabled) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Visible( 
            /* [retval][out] */ VARIANT_BOOL *Visible) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Visible( 
            /* [in] */ VARIANT_BOOL Visible) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Focused( 
            /* [retval][out] */ VARIANT_BOOL *Focused) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Width( 
            /* [retval][out] */ long *Width) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Width( 
            /* [in] */ long Width) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Height( 
            /* [retval][out] */ long *Height) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Height( 
            /* [in] */ long Height) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TabOrder( 
            /* [retval][out] */ long *TabOrder) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TabOrder( 
            /* [in] */ long TabOrder) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TabStop( 
            /* [retval][out] */ VARIANT_BOOL *TabStop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TabStop( 
            /* [in] */ VARIANT_BOOL TabStop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Group( 
            /* [retval][out] */ VARIANT_BOOL *Group) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Group( 
            /* [in] */ VARIANT_BOOL Group) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Window( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_id( 
            /* [retval][out] */ long *id) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Next( 
            /* [retval][out] */ IDispatch **prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Previous( 
            /* [retval][out] */ IDispatch **prop) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Focus( void) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_type( 
            /* [retval][out] */ ComboBoxType *type) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Sort( 
            /* [retval][out] */ VARIANT_BOOL *Sort) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DropdownVisible( 
            /* [retval][out] */ VARIANT_BOOL *Visible) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_DropdownVisible( 
            /* [in] */ VARIANT_BOOL Visible) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_selectedItem( 
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_selectedItem( 
            /* [in] */ long index) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EditSelectionStart( 
            /* [retval][out] */ long *start) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EditSelectionStart( 
            /* [in] */ long start) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_EditSelectionEnd( 
            /* [retval][out] */ long *end) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_EditSelectionEnd( 
            /* [in] */ long end) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Data( 
            /* [optional][in] */ VARIANT index,
            /* [retval][out] */ long *Data) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Data( 
            /* [optional][in] */ VARIANT index,
            /* [in] */ long Data) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Add( 
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT Data,
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Delete( 
            /* [in] */ long index) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE FindStart( 
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT startIndex,
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Find( 
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT startIndex,
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE rectangle( 
            /* [optional][in] */ VARIANT index,
            /* [retval][out] */ ClientRectangle **rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Insert( 
            /* [in] */ long index,
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT Data,
            /* [retval][out] */ long *newIndex) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Clear( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE FindData( 
            /* [in] */ long Data,
            /* [optional][in] */ VARIANT startIndex,
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Replace( 
            /* [in] */ BSTR Text) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_id( 
            /* [in] */ long id) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _ComboBoxVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _ComboBox * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _ComboBox * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _ComboBox * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _ComboBox * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _ComboBox * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _ComboBox * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _ComboBox * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            _ComboBox * This,
            /* [optional][in] */ VARIANT index,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            _ComboBox * This,
            /* [optional][in] */ VARIANT index,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _ComboBox * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _ComboBox * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            _ComboBox * This,
            /* [retval][out] */ ClientPoint **Position);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            _ComboBox * This,
            /* [in] */ ClientPoint *Position);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Enabled )( 
            _ComboBox * This,
            /* [retval][out] */ VARIANT_BOOL *Enabled);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Enabled )( 
            _ComboBox * This,
            /* [in] */ VARIANT_BOOL Enabled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            _ComboBox * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Visible )( 
            _ComboBox * This,
            /* [in] */ VARIANT_BOOL Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            _ComboBox * This,
            /* [retval][out] */ VARIANT_BOOL *Focused);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Width )( 
            _ComboBox * This,
            /* [retval][out] */ long *Width);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Width )( 
            _ComboBox * This,
            /* [in] */ long Width);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Height )( 
            _ComboBox * This,
            /* [retval][out] */ long *Height);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Height )( 
            _ComboBox * This,
            /* [in] */ long Height);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabOrder )( 
            _ComboBox * This,
            /* [retval][out] */ long *TabOrder);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabOrder )( 
            _ComboBox * This,
            /* [in] */ long TabOrder);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabStop )( 
            _ComboBox * This,
            /* [retval][out] */ VARIANT_BOOL *TabStop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabStop )( 
            _ComboBox * This,
            /* [in] */ VARIANT_BOOL TabStop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Group )( 
            _ComboBox * This,
            /* [retval][out] */ VARIANT_BOOL *Group);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Group )( 
            _ComboBox * This,
            /* [in] */ VARIANT_BOOL Group);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            _ComboBox * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            _ComboBox * This,
            /* [retval][out] */ long *id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Next )( 
            _ComboBox * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Previous )( 
            _ComboBox * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Focus )( 
            _ComboBox * This);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_type )( 
            _ComboBox * This,
            /* [retval][out] */ ComboBoxType *type);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Sort )( 
            _ComboBox * This,
            /* [retval][out] */ VARIANT_BOOL *Sort);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DropdownVisible )( 
            _ComboBox * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_DropdownVisible )( 
            _ComboBox * This,
            /* [in] */ VARIANT_BOOL Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            _ComboBox * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_selectedItem )( 
            _ComboBox * This,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_selectedItem )( 
            _ComboBox * This,
            /* [in] */ long index);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EditSelectionStart )( 
            _ComboBox * This,
            /* [retval][out] */ long *start);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EditSelectionStart )( 
            _ComboBox * This,
            /* [in] */ long start);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_EditSelectionEnd )( 
            _ComboBox * This,
            /* [retval][out] */ long *end);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_EditSelectionEnd )( 
            _ComboBox * This,
            /* [in] */ long end);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Data )( 
            _ComboBox * This,
            /* [optional][in] */ VARIANT index,
            /* [retval][out] */ long *Data);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Data )( 
            _ComboBox * This,
            /* [optional][in] */ VARIANT index,
            /* [in] */ long Data);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Add )( 
            _ComboBox * This,
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT Data,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Delete )( 
            _ComboBox * This,
            /* [in] */ long index);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *FindStart )( 
            _ComboBox * This,
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT startIndex,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Find )( 
            _ComboBox * This,
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT startIndex,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *rectangle )( 
            _ComboBox * This,
            /* [optional][in] */ VARIANT index,
            /* [retval][out] */ ClientRectangle **rectangle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Insert )( 
            _ComboBox * This,
            /* [in] */ long index,
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT Data,
            /* [retval][out] */ long *newIndex);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Clear )( 
            _ComboBox * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *FindData )( 
            _ComboBox * This,
            /* [in] */ long Data,
            /* [optional][in] */ VARIANT startIndex,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Replace )( 
            _ComboBox * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_id )( 
            _ComboBox * This,
            /* [in] */ long id);
        
        END_INTERFACE
    } _ComboBoxVtbl;

    interface _ComboBox
    {
        CONST_VTBL struct _ComboBoxVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _ComboBox_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _ComboBox_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _ComboBox_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _ComboBox_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _ComboBox_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _ComboBox_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _ComboBox_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _ComboBox_get_Text(This,index,Text)	\
    ( (This)->lpVtbl -> get_Text(This,index,Text) ) 

#define _ComboBox_put_Text(This,index,Text)	\
    ( (This)->lpVtbl -> put_Text(This,index,Text) ) 

#define _ComboBox_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _ComboBox_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _ComboBox_get_Position(This,Position)	\
    ( (This)->lpVtbl -> get_Position(This,Position) ) 

#define _ComboBox_put_Position(This,Position)	\
    ( (This)->lpVtbl -> put_Position(This,Position) ) 

#define _ComboBox_get_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> get_Enabled(This,Enabled) ) 

#define _ComboBox_put_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> put_Enabled(This,Enabled) ) 

#define _ComboBox_get_Visible(This,Visible)	\
    ( (This)->lpVtbl -> get_Visible(This,Visible) ) 

#define _ComboBox_put_Visible(This,Visible)	\
    ( (This)->lpVtbl -> put_Visible(This,Visible) ) 

#define _ComboBox_get_Focused(This,Focused)	\
    ( (This)->lpVtbl -> get_Focused(This,Focused) ) 

#define _ComboBox_get_Width(This,Width)	\
    ( (This)->lpVtbl -> get_Width(This,Width) ) 

#define _ComboBox_put_Width(This,Width)	\
    ( (This)->lpVtbl -> put_Width(This,Width) ) 

#define _ComboBox_get_Height(This,Height)	\
    ( (This)->lpVtbl -> get_Height(This,Height) ) 

#define _ComboBox_put_Height(This,Height)	\
    ( (This)->lpVtbl -> put_Height(This,Height) ) 

#define _ComboBox_get_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> get_TabOrder(This,TabOrder) ) 

#define _ComboBox_put_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> put_TabOrder(This,TabOrder) ) 

#define _ComboBox_get_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> get_TabStop(This,TabStop) ) 

#define _ComboBox_put_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> put_TabStop(This,TabStop) ) 

#define _ComboBox_get_Group(This,Group)	\
    ( (This)->lpVtbl -> get_Group(This,Group) ) 

#define _ComboBox_put_Group(This,Group)	\
    ( (This)->lpVtbl -> put_Group(This,Group) ) 

#define _ComboBox_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define _ComboBox_get_id(This,id)	\
    ( (This)->lpVtbl -> get_id(This,id) ) 

#define _ComboBox_get_Next(This,prop)	\
    ( (This)->lpVtbl -> get_Next(This,prop) ) 

#define _ComboBox_get_Previous(This,prop)	\
    ( (This)->lpVtbl -> get_Previous(This,prop) ) 

#define _ComboBox_Focus(This)	\
    ( (This)->lpVtbl -> Focus(This) ) 

#define _ComboBox_get_type(This,type)	\
    ( (This)->lpVtbl -> get_type(This,type) ) 

#define _ComboBox_get_Sort(This,Sort)	\
    ( (This)->lpVtbl -> get_Sort(This,Sort) ) 

#define _ComboBox_get_DropdownVisible(This,Visible)	\
    ( (This)->lpVtbl -> get_DropdownVisible(This,Visible) ) 

#define _ComboBox_put_DropdownVisible(This,Visible)	\
    ( (This)->lpVtbl -> put_DropdownVisible(This,Visible) ) 

#define _ComboBox_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define _ComboBox_get_selectedItem(This,index)	\
    ( (This)->lpVtbl -> get_selectedItem(This,index) ) 

#define _ComboBox_put_selectedItem(This,index)	\
    ( (This)->lpVtbl -> put_selectedItem(This,index) ) 

#define _ComboBox_get_EditSelectionStart(This,start)	\
    ( (This)->lpVtbl -> get_EditSelectionStart(This,start) ) 

#define _ComboBox_put_EditSelectionStart(This,start)	\
    ( (This)->lpVtbl -> put_EditSelectionStart(This,start) ) 

#define _ComboBox_get_EditSelectionEnd(This,end)	\
    ( (This)->lpVtbl -> get_EditSelectionEnd(This,end) ) 

#define _ComboBox_put_EditSelectionEnd(This,end)	\
    ( (This)->lpVtbl -> put_EditSelectionEnd(This,end) ) 

#define _ComboBox_get_Data(This,index,Data)	\
    ( (This)->lpVtbl -> get_Data(This,index,Data) ) 

#define _ComboBox_put_Data(This,index,Data)	\
    ( (This)->lpVtbl -> put_Data(This,index,Data) ) 

#define _ComboBox_Add(This,Text,Data,index)	\
    ( (This)->lpVtbl -> Add(This,Text,Data,index) ) 

#define _ComboBox_Delete(This,index)	\
    ( (This)->lpVtbl -> Delete(This,index) ) 

#define _ComboBox_FindStart(This,Text,startIndex,index)	\
    ( (This)->lpVtbl -> FindStart(This,Text,startIndex,index) ) 

#define _ComboBox_Find(This,Text,startIndex,index)	\
    ( (This)->lpVtbl -> Find(This,Text,startIndex,index) ) 

#define _ComboBox_rectangle(This,index,rectangle)	\
    ( (This)->lpVtbl -> rectangle(This,index,rectangle) ) 

#define _ComboBox_Insert(This,index,Text,Data,newIndex)	\
    ( (This)->lpVtbl -> Insert(This,index,Text,Data,newIndex) ) 

#define _ComboBox_Clear(This)	\
    ( (This)->lpVtbl -> Clear(This) ) 

#define _ComboBox_FindData(This,Data,startIndex,index)	\
    ( (This)->lpVtbl -> FindData(This,Data,startIndex,index) ) 

#define _ComboBox_Replace(This,Text)	\
    ( (This)->lpVtbl -> Replace(This,Text) ) 

#define _ComboBox_put_id(This,id)	\
    ( (This)->lpVtbl -> put_id(This,id) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___ComboBox_INTERFACE_DEFINED__ */


#ifndef ___ListBox_INTERFACE_DEFINED__
#define ___ListBox_INTERFACE_DEFINED__

/* interface _ListBox */
/* [object][oleautomation][dual][hidden][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__ListBox;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("F8209E01-C266-4FC1-96F7-61EA7C30BF4A")
    _ListBox : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Text( 
            /* [optional][in] */ VARIANT index,
            /* [retval][out] */ BSTR *Text) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Text( 
            /* [optional][in] */ VARIANT index,
            /* [in] */ BSTR Text) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Position( 
            /* [retval][out] */ ClientPoint **Position) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Position( 
            /* [in] */ ClientPoint *Position) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Enabled( 
            /* [retval][out] */ VARIANT_BOOL *Enabled) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Enabled( 
            /* [in] */ VARIANT_BOOL Enabled) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Visible( 
            /* [retval][out] */ VARIANT_BOOL *Visible) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Visible( 
            /* [in] */ VARIANT_BOOL Visible) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Focused( 
            /* [retval][out] */ VARIANT_BOOL *Focused) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Width( 
            /* [retval][out] */ long *Width) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Width( 
            /* [in] */ long Width) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Height( 
            /* [retval][out] */ long *Height) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Height( 
            /* [in] */ long Height) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TabOrder( 
            /* [retval][out] */ long *TabOrder) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TabOrder( 
            /* [in] */ long TabOrder) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TabStop( 
            /* [retval][out] */ VARIANT_BOOL *TabStop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TabStop( 
            /* [in] */ VARIANT_BOOL TabStop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Group( 
            /* [retval][out] */ VARIANT_BOOL *Group) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Group( 
            /* [in] */ VARIANT_BOOL Group) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Window( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_id( 
            /* [retval][out] */ long *id) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Next( 
            /* [retval][out] */ IDispatch **prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Previous( 
            /* [retval][out] */ IDispatch **prop) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Focus( void) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_type( 
            /* [retval][out] */ ListBoxType *type) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Sort( 
            /* [retval][out] */ VARIANT_BOOL *Sort) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SelectedCount( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Data( 
            /* [optional][in] */ VARIANT index,
            /* [retval][out] */ long *Data) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Data( 
            /* [optional][in] */ VARIANT index,
            /* [in] */ long Data) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Selected( 
            /* [in] */ long index,
            /* [retval][out] */ VARIANT_BOOL *Selected) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Selected( 
            /* [in] */ long index,
            /* [in] */ VARIANT_BOOL Selected) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FocusedIndex( 
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FocusedIndex( 
            /* [in] */ long index) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Add( 
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT Data,
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Delete( 
            /* [in] */ long index) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE FindStart( 
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT startIndex,
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Find( 
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT startIndex,
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE rectangle( 
            /* [in] */ long index,
            /* [retval][out] */ ClientRectangle **rectangle) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Insert( 
            /* [in] */ long index,
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT Data,
            /* [retval][out] */ long *newIndex) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Clear( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE FindData( 
            /* [in] */ long Data,
            /* [optional][in] */ VARIANT startIndex,
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE selectedItem( 
            /* [optional][in] */ VARIANT itemNumber,
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_id( 
            /* [in] */ long id) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _ListBoxVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _ListBox * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _ListBox * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _ListBox * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _ListBox * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _ListBox * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _ListBox * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _ListBox * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            _ListBox * This,
            /* [optional][in] */ VARIANT index,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            _ListBox * This,
            /* [optional][in] */ VARIANT index,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _ListBox * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _ListBox * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            _ListBox * This,
            /* [retval][out] */ ClientPoint **Position);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            _ListBox * This,
            /* [in] */ ClientPoint *Position);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Enabled )( 
            _ListBox * This,
            /* [retval][out] */ VARIANT_BOOL *Enabled);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Enabled )( 
            _ListBox * This,
            /* [in] */ VARIANT_BOOL Enabled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            _ListBox * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Visible )( 
            _ListBox * This,
            /* [in] */ VARIANT_BOOL Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            _ListBox * This,
            /* [retval][out] */ VARIANT_BOOL *Focused);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Width )( 
            _ListBox * This,
            /* [retval][out] */ long *Width);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Width )( 
            _ListBox * This,
            /* [in] */ long Width);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Height )( 
            _ListBox * This,
            /* [retval][out] */ long *Height);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Height )( 
            _ListBox * This,
            /* [in] */ long Height);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabOrder )( 
            _ListBox * This,
            /* [retval][out] */ long *TabOrder);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabOrder )( 
            _ListBox * This,
            /* [in] */ long TabOrder);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabStop )( 
            _ListBox * This,
            /* [retval][out] */ VARIANT_BOOL *TabStop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabStop )( 
            _ListBox * This,
            /* [in] */ VARIANT_BOOL TabStop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Group )( 
            _ListBox * This,
            /* [retval][out] */ VARIANT_BOOL *Group);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Group )( 
            _ListBox * This,
            /* [in] */ VARIANT_BOOL Group);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            _ListBox * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            _ListBox * This,
            /* [retval][out] */ long *id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Next )( 
            _ListBox * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Previous )( 
            _ListBox * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Focus )( 
            _ListBox * This);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_type )( 
            _ListBox * This,
            /* [retval][out] */ ListBoxType *type);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Sort )( 
            _ListBox * This,
            /* [retval][out] */ VARIANT_BOOL *Sort);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            _ListBox * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SelectedCount )( 
            _ListBox * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Data )( 
            _ListBox * This,
            /* [optional][in] */ VARIANT index,
            /* [retval][out] */ long *Data);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Data )( 
            _ListBox * This,
            /* [optional][in] */ VARIANT index,
            /* [in] */ long Data);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Selected )( 
            _ListBox * This,
            /* [in] */ long index,
            /* [retval][out] */ VARIANT_BOOL *Selected);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Selected )( 
            _ListBox * This,
            /* [in] */ long index,
            /* [in] */ VARIANT_BOOL Selected);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FocusedIndex )( 
            _ListBox * This,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FocusedIndex )( 
            _ListBox * This,
            /* [in] */ long index);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Add )( 
            _ListBox * This,
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT Data,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Delete )( 
            _ListBox * This,
            /* [in] */ long index);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *FindStart )( 
            _ListBox * This,
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT startIndex,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Find )( 
            _ListBox * This,
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT startIndex,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *rectangle )( 
            _ListBox * This,
            /* [in] */ long index,
            /* [retval][out] */ ClientRectangle **rectangle);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Insert )( 
            _ListBox * This,
            /* [in] */ long index,
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT Data,
            /* [retval][out] */ long *newIndex);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Clear )( 
            _ListBox * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *FindData )( 
            _ListBox * This,
            /* [in] */ long Data,
            /* [optional][in] */ VARIANT startIndex,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *selectedItem )( 
            _ListBox * This,
            /* [optional][in] */ VARIANT itemNumber,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_id )( 
            _ListBox * This,
            /* [in] */ long id);
        
        END_INTERFACE
    } _ListBoxVtbl;

    interface _ListBox
    {
        CONST_VTBL struct _ListBoxVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _ListBox_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _ListBox_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _ListBox_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _ListBox_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _ListBox_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _ListBox_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _ListBox_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _ListBox_get_Text(This,index,Text)	\
    ( (This)->lpVtbl -> get_Text(This,index,Text) ) 

#define _ListBox_put_Text(This,index,Text)	\
    ( (This)->lpVtbl -> put_Text(This,index,Text) ) 

#define _ListBox_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _ListBox_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _ListBox_get_Position(This,Position)	\
    ( (This)->lpVtbl -> get_Position(This,Position) ) 

#define _ListBox_put_Position(This,Position)	\
    ( (This)->lpVtbl -> put_Position(This,Position) ) 

#define _ListBox_get_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> get_Enabled(This,Enabled) ) 

#define _ListBox_put_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> put_Enabled(This,Enabled) ) 

#define _ListBox_get_Visible(This,Visible)	\
    ( (This)->lpVtbl -> get_Visible(This,Visible) ) 

#define _ListBox_put_Visible(This,Visible)	\
    ( (This)->lpVtbl -> put_Visible(This,Visible) ) 

#define _ListBox_get_Focused(This,Focused)	\
    ( (This)->lpVtbl -> get_Focused(This,Focused) ) 

#define _ListBox_get_Width(This,Width)	\
    ( (This)->lpVtbl -> get_Width(This,Width) ) 

#define _ListBox_put_Width(This,Width)	\
    ( (This)->lpVtbl -> put_Width(This,Width) ) 

#define _ListBox_get_Height(This,Height)	\
    ( (This)->lpVtbl -> get_Height(This,Height) ) 

#define _ListBox_put_Height(This,Height)	\
    ( (This)->lpVtbl -> put_Height(This,Height) ) 

#define _ListBox_get_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> get_TabOrder(This,TabOrder) ) 

#define _ListBox_put_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> put_TabOrder(This,TabOrder) ) 

#define _ListBox_get_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> get_TabStop(This,TabStop) ) 

#define _ListBox_put_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> put_TabStop(This,TabStop) ) 

#define _ListBox_get_Group(This,Group)	\
    ( (This)->lpVtbl -> get_Group(This,Group) ) 

#define _ListBox_put_Group(This,Group)	\
    ( (This)->lpVtbl -> put_Group(This,Group) ) 

#define _ListBox_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define _ListBox_get_id(This,id)	\
    ( (This)->lpVtbl -> get_id(This,id) ) 

#define _ListBox_get_Next(This,prop)	\
    ( (This)->lpVtbl -> get_Next(This,prop) ) 

#define _ListBox_get_Previous(This,prop)	\
    ( (This)->lpVtbl -> get_Previous(This,prop) ) 

#define _ListBox_Focus(This)	\
    ( (This)->lpVtbl -> Focus(This) ) 

#define _ListBox_get_type(This,type)	\
    ( (This)->lpVtbl -> get_type(This,type) ) 

#define _ListBox_get_Sort(This,Sort)	\
    ( (This)->lpVtbl -> get_Sort(This,Sort) ) 

#define _ListBox_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define _ListBox_get_SelectedCount(This,Count)	\
    ( (This)->lpVtbl -> get_SelectedCount(This,Count) ) 

#define _ListBox_get_Data(This,index,Data)	\
    ( (This)->lpVtbl -> get_Data(This,index,Data) ) 

#define _ListBox_put_Data(This,index,Data)	\
    ( (This)->lpVtbl -> put_Data(This,index,Data) ) 

#define _ListBox_get_Selected(This,index,Selected)	\
    ( (This)->lpVtbl -> get_Selected(This,index,Selected) ) 

#define _ListBox_put_Selected(This,index,Selected)	\
    ( (This)->lpVtbl -> put_Selected(This,index,Selected) ) 

#define _ListBox_get_FocusedIndex(This,index)	\
    ( (This)->lpVtbl -> get_FocusedIndex(This,index) ) 

#define _ListBox_put_FocusedIndex(This,index)	\
    ( (This)->lpVtbl -> put_FocusedIndex(This,index) ) 

#define _ListBox_Add(This,Text,Data,index)	\
    ( (This)->lpVtbl -> Add(This,Text,Data,index) ) 

#define _ListBox_Delete(This,index)	\
    ( (This)->lpVtbl -> Delete(This,index) ) 

#define _ListBox_FindStart(This,Text,startIndex,index)	\
    ( (This)->lpVtbl -> FindStart(This,Text,startIndex,index) ) 

#define _ListBox_Find(This,Text,startIndex,index)	\
    ( (This)->lpVtbl -> Find(This,Text,startIndex,index) ) 

#define _ListBox_rectangle(This,index,rectangle)	\
    ( (This)->lpVtbl -> rectangle(This,index,rectangle) ) 

#define _ListBox_Insert(This,index,Text,Data,newIndex)	\
    ( (This)->lpVtbl -> Insert(This,index,Text,Data,newIndex) ) 

#define _ListBox_Clear(This)	\
    ( (This)->lpVtbl -> Clear(This) ) 

#define _ListBox_FindData(This,Data,startIndex,index)	\
    ( (This)->lpVtbl -> FindData(This,Data,startIndex,index) ) 

#define _ListBox_selectedItem(This,itemNumber,index)	\
    ( (This)->lpVtbl -> selectedItem(This,itemNumber,index) ) 

#define _ListBox_put_id(This,id)	\
    ( (This)->lpVtbl -> put_id(This,id) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___ListBox_INTERFACE_DEFINED__ */


#ifndef ___UpDown_INTERFACE_DEFINED__
#define ___UpDown_INTERFACE_DEFINED__

/* interface _UpDown */
/* [object][oleautomation][dual][hidden][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__UpDown;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("B34DF9B0-5781-4CB5-BF97-CE7004AE7FA5")
    _UpDown : public _Control
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_AlignLeft( 
            /* [retval][out] */ VARIANT_BOOL *Left) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_AlignLeft( 
            /* [in] */ VARIANT_BOOL Left) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Horizontal( 
            /* [retval][out] */ VARIANT_BOOL *Horizontal) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Horizontal( 
            /* [in] */ VARIANT_BOOL Horizontal) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ArrowKeys( 
            /* [retval][out] */ VARIANT_BOOL *ArrowKeys) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_ArrowKeys( 
            /* [in] */ VARIANT_BOOL ArrowKeys) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Wrap( 
            /* [retval][out] */ VARIANT_BOOL *Wrap) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Wrap( 
            /* [in] */ VARIANT_BOOL Wrap) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Auto( 
            /* [retval][out] */ VARIANT_BOOL *isAuto) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Auto( 
            /* [in] */ VARIANT_BOOL isAuto) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Commas( 
            /* [retval][out] */ VARIANT_BOOL *Commas) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Commas( 
            /* [in] */ VARIANT_BOOL Commas) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Base( 
            /* [retval][out] */ long *Base) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Base( 
            /* [in] */ long Base) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Buddy( 
            /* [retval][out] */ _Control **Control) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Buddy( 
            /* [in] */ IDispatch *Control) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Value( 
            /* [retval][out] */ long *Value) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Value( 
            /* [in] */ long Value) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Minimum( 
            /* [retval][out] */ long *Minimum) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Minimum( 
            /* [in] */ long Minimum) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Maximum( 
            /* [retval][out] */ long *Maximum) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Maximum( 
            /* [in] */ long Maximum) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _UpDownVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _UpDown * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _UpDown * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _UpDown * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _UpDown * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _UpDown * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _UpDown * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _UpDown * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            _UpDown * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            _UpDown * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _UpDown * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _UpDown * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            _UpDown * This,
            /* [retval][out] */ ClientPoint **Position);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            _UpDown * This,
            /* [in] */ ClientPoint *Position);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Enabled )( 
            _UpDown * This,
            /* [retval][out] */ VARIANT_BOOL *Enabled);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Enabled )( 
            _UpDown * This,
            /* [in] */ VARIANT_BOOL Enabled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            _UpDown * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Visible )( 
            _UpDown * This,
            /* [in] */ VARIANT_BOOL Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            _UpDown * This,
            /* [retval][out] */ VARIANT_BOOL *Focused);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Width )( 
            _UpDown * This,
            /* [retval][out] */ long *Width);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Width )( 
            _UpDown * This,
            /* [in] */ long Width);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Height )( 
            _UpDown * This,
            /* [retval][out] */ long *Height);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Height )( 
            _UpDown * This,
            /* [in] */ long Height);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabOrder )( 
            _UpDown * This,
            /* [retval][out] */ long *TabOrder);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabOrder )( 
            _UpDown * This,
            /* [in] */ long TabOrder);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabStop )( 
            _UpDown * This,
            /* [retval][out] */ VARIANT_BOOL *TabStop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabStop )( 
            _UpDown * This,
            /* [in] */ VARIANT_BOOL TabStop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Group )( 
            _UpDown * This,
            /* [retval][out] */ VARIANT_BOOL *Group);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Group )( 
            _UpDown * This,
            /* [in] */ VARIANT_BOOL Group);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            _UpDown * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            _UpDown * This,
            /* [retval][out] */ long *id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Next )( 
            _UpDown * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Previous )( 
            _UpDown * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Focus )( 
            _UpDown * This);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_id )( 
            _UpDown * This,
            /* [in] */ long id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_AlignLeft )( 
            _UpDown * This,
            /* [retval][out] */ VARIANT_BOOL *Left);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_AlignLeft )( 
            _UpDown * This,
            /* [in] */ VARIANT_BOOL Left);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Horizontal )( 
            _UpDown * This,
            /* [retval][out] */ VARIANT_BOOL *Horizontal);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Horizontal )( 
            _UpDown * This,
            /* [in] */ VARIANT_BOOL Horizontal);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ArrowKeys )( 
            _UpDown * This,
            /* [retval][out] */ VARIANT_BOOL *ArrowKeys);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_ArrowKeys )( 
            _UpDown * This,
            /* [in] */ VARIANT_BOOL ArrowKeys);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Wrap )( 
            _UpDown * This,
            /* [retval][out] */ VARIANT_BOOL *Wrap);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Wrap )( 
            _UpDown * This,
            /* [in] */ VARIANT_BOOL Wrap);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Auto )( 
            _UpDown * This,
            /* [retval][out] */ VARIANT_BOOL *isAuto);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Auto )( 
            _UpDown * This,
            /* [in] */ VARIANT_BOOL isAuto);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Commas )( 
            _UpDown * This,
            /* [retval][out] */ VARIANT_BOOL *Commas);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Commas )( 
            _UpDown * This,
            /* [in] */ VARIANT_BOOL Commas);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Base )( 
            _UpDown * This,
            /* [retval][out] */ long *Base);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Base )( 
            _UpDown * This,
            /* [in] */ long Base);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Buddy )( 
            _UpDown * This,
            /* [retval][out] */ _Control **Control);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Buddy )( 
            _UpDown * This,
            /* [in] */ IDispatch *Control);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Value )( 
            _UpDown * This,
            /* [retval][out] */ long *Value);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Value )( 
            _UpDown * This,
            /* [in] */ long Value);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Minimum )( 
            _UpDown * This,
            /* [retval][out] */ long *Minimum);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Minimum )( 
            _UpDown * This,
            /* [in] */ long Minimum);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Maximum )( 
            _UpDown * This,
            /* [retval][out] */ long *Maximum);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Maximum )( 
            _UpDown * This,
            /* [in] */ long Maximum);
        
        END_INTERFACE
    } _UpDownVtbl;

    interface _UpDown
    {
        CONST_VTBL struct _UpDownVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _UpDown_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _UpDown_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _UpDown_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _UpDown_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _UpDown_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _UpDown_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _UpDown_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _UpDown_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define _UpDown_put_Text(This,Text)	\
    ( (This)->lpVtbl -> put_Text(This,Text) ) 

#define _UpDown_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _UpDown_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _UpDown_get_Position(This,Position)	\
    ( (This)->lpVtbl -> get_Position(This,Position) ) 

#define _UpDown_put_Position(This,Position)	\
    ( (This)->lpVtbl -> put_Position(This,Position) ) 

#define _UpDown_get_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> get_Enabled(This,Enabled) ) 

#define _UpDown_put_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> put_Enabled(This,Enabled) ) 

#define _UpDown_get_Visible(This,Visible)	\
    ( (This)->lpVtbl -> get_Visible(This,Visible) ) 

#define _UpDown_put_Visible(This,Visible)	\
    ( (This)->lpVtbl -> put_Visible(This,Visible) ) 

#define _UpDown_get_Focused(This,Focused)	\
    ( (This)->lpVtbl -> get_Focused(This,Focused) ) 

#define _UpDown_get_Width(This,Width)	\
    ( (This)->lpVtbl -> get_Width(This,Width) ) 

#define _UpDown_put_Width(This,Width)	\
    ( (This)->lpVtbl -> put_Width(This,Width) ) 

#define _UpDown_get_Height(This,Height)	\
    ( (This)->lpVtbl -> get_Height(This,Height) ) 

#define _UpDown_put_Height(This,Height)	\
    ( (This)->lpVtbl -> put_Height(This,Height) ) 

#define _UpDown_get_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> get_TabOrder(This,TabOrder) ) 

#define _UpDown_put_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> put_TabOrder(This,TabOrder) ) 

#define _UpDown_get_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> get_TabStop(This,TabStop) ) 

#define _UpDown_put_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> put_TabStop(This,TabStop) ) 

#define _UpDown_get_Group(This,Group)	\
    ( (This)->lpVtbl -> get_Group(This,Group) ) 

#define _UpDown_put_Group(This,Group)	\
    ( (This)->lpVtbl -> put_Group(This,Group) ) 

#define _UpDown_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define _UpDown_get_id(This,id)	\
    ( (This)->lpVtbl -> get_id(This,id) ) 

#define _UpDown_get_Next(This,prop)	\
    ( (This)->lpVtbl -> get_Next(This,prop) ) 

#define _UpDown_get_Previous(This,prop)	\
    ( (This)->lpVtbl -> get_Previous(This,prop) ) 

#define _UpDown_Focus(This)	\
    ( (This)->lpVtbl -> Focus(This) ) 

#define _UpDown_put_id(This,id)	\
    ( (This)->lpVtbl -> put_id(This,id) ) 


#define _UpDown_get_AlignLeft(This,Left)	\
    ( (This)->lpVtbl -> get_AlignLeft(This,Left) ) 

#define _UpDown_put_AlignLeft(This,Left)	\
    ( (This)->lpVtbl -> put_AlignLeft(This,Left) ) 

#define _UpDown_get_Horizontal(This,Horizontal)	\
    ( (This)->lpVtbl -> get_Horizontal(This,Horizontal) ) 

#define _UpDown_put_Horizontal(This,Horizontal)	\
    ( (This)->lpVtbl -> put_Horizontal(This,Horizontal) ) 

#define _UpDown_get_ArrowKeys(This,ArrowKeys)	\
    ( (This)->lpVtbl -> get_ArrowKeys(This,ArrowKeys) ) 

#define _UpDown_put_ArrowKeys(This,ArrowKeys)	\
    ( (This)->lpVtbl -> put_ArrowKeys(This,ArrowKeys) ) 

#define _UpDown_get_Wrap(This,Wrap)	\
    ( (This)->lpVtbl -> get_Wrap(This,Wrap) ) 

#define _UpDown_put_Wrap(This,Wrap)	\
    ( (This)->lpVtbl -> put_Wrap(This,Wrap) ) 

#define _UpDown_get_Auto(This,isAuto)	\
    ( (This)->lpVtbl -> get_Auto(This,isAuto) ) 

#define _UpDown_put_Auto(This,isAuto)	\
    ( (This)->lpVtbl -> put_Auto(This,isAuto) ) 

#define _UpDown_get_Commas(This,Commas)	\
    ( (This)->lpVtbl -> get_Commas(This,Commas) ) 

#define _UpDown_put_Commas(This,Commas)	\
    ( (This)->lpVtbl -> put_Commas(This,Commas) ) 

#define _UpDown_get_Base(This,Base)	\
    ( (This)->lpVtbl -> get_Base(This,Base) ) 

#define _UpDown_put_Base(This,Base)	\
    ( (This)->lpVtbl -> put_Base(This,Base) ) 

#define _UpDown_get_Buddy(This,Control)	\
    ( (This)->lpVtbl -> get_Buddy(This,Control) ) 

#define _UpDown_put_Buddy(This,Control)	\
    ( (This)->lpVtbl -> put_Buddy(This,Control) ) 

#define _UpDown_get_Value(This,Value)	\
    ( (This)->lpVtbl -> get_Value(This,Value) ) 

#define _UpDown_put_Value(This,Value)	\
    ( (This)->lpVtbl -> put_Value(This,Value) ) 

#define _UpDown_get_Minimum(This,Minimum)	\
    ( (This)->lpVtbl -> get_Minimum(This,Minimum) ) 

#define _UpDown_put_Minimum(This,Minimum)	\
    ( (This)->lpVtbl -> put_Minimum(This,Minimum) ) 

#define _UpDown_get_Maximum(This,Maximum)	\
    ( (This)->lpVtbl -> get_Maximum(This,Maximum) ) 

#define _UpDown_put_Maximum(This,Maximum)	\
    ( (This)->lpVtbl -> put_Maximum(This,Maximum) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___UpDown_INTERFACE_DEFINED__ */


#ifndef __ProgressBar_INTERFACE_DEFINED__
#define __ProgressBar_INTERFACE_DEFINED__

/* interface ProgressBar */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ProgressBar;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("20134F69-7A84-4198-9560-DA2CA87016CA")
    ProgressBar : public _Control
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Vertical( 
            /* [retval][out] */ VARIANT_BOOL *Vertical) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Vertical( 
            /* [in] */ VARIANT_BOOL Vertical) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Value( 
            /* [retval][out] */ long *Value) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Value( 
            /* [in] */ long Value) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Minimum( 
            /* [retval][out] */ long *Minimum) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Minimum( 
            /* [in] */ long Minimum) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Maximum( 
            /* [retval][out] */ long *Maximum) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Maximum( 
            /* [in] */ long Maximum) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Step( 
            /* [retval][out] */ long *Step) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Step( 
            /* [in] */ long Step) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Marquee( 
            /* [in] */ long rhs) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Increase( 
            /* [in] */ long amount) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE IncreaseStep( void) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ProgressBarVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ProgressBar * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ProgressBar * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ProgressBar * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ProgressBar * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ProgressBar * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ProgressBar * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ProgressBar * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            ProgressBar * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            ProgressBar * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ProgressBar * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ProgressBar * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            ProgressBar * This,
            /* [retval][out] */ ClientPoint **Position);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            ProgressBar * This,
            /* [in] */ ClientPoint *Position);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Enabled )( 
            ProgressBar * This,
            /* [retval][out] */ VARIANT_BOOL *Enabled);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Enabled )( 
            ProgressBar * This,
            /* [in] */ VARIANT_BOOL Enabled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            ProgressBar * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Visible )( 
            ProgressBar * This,
            /* [in] */ VARIANT_BOOL Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            ProgressBar * This,
            /* [retval][out] */ VARIANT_BOOL *Focused);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Width )( 
            ProgressBar * This,
            /* [retval][out] */ long *Width);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Width )( 
            ProgressBar * This,
            /* [in] */ long Width);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Height )( 
            ProgressBar * This,
            /* [retval][out] */ long *Height);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Height )( 
            ProgressBar * This,
            /* [in] */ long Height);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabOrder )( 
            ProgressBar * This,
            /* [retval][out] */ long *TabOrder);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabOrder )( 
            ProgressBar * This,
            /* [in] */ long TabOrder);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabStop )( 
            ProgressBar * This,
            /* [retval][out] */ VARIANT_BOOL *TabStop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabStop )( 
            ProgressBar * This,
            /* [in] */ VARIANT_BOOL TabStop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Group )( 
            ProgressBar * This,
            /* [retval][out] */ VARIANT_BOOL *Group);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Group )( 
            ProgressBar * This,
            /* [in] */ VARIANT_BOOL Group);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            ProgressBar * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            ProgressBar * This,
            /* [retval][out] */ long *id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Next )( 
            ProgressBar * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Previous )( 
            ProgressBar * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Focus )( 
            ProgressBar * This);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_id )( 
            ProgressBar * This,
            /* [in] */ long id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Vertical )( 
            ProgressBar * This,
            /* [retval][out] */ VARIANT_BOOL *Vertical);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Vertical )( 
            ProgressBar * This,
            /* [in] */ VARIANT_BOOL Vertical);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Value )( 
            ProgressBar * This,
            /* [retval][out] */ long *Value);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Value )( 
            ProgressBar * This,
            /* [in] */ long Value);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Minimum )( 
            ProgressBar * This,
            /* [retval][out] */ long *Minimum);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Minimum )( 
            ProgressBar * This,
            /* [in] */ long Minimum);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Maximum )( 
            ProgressBar * This,
            /* [retval][out] */ long *Maximum);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Maximum )( 
            ProgressBar * This,
            /* [in] */ long Maximum);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Step )( 
            ProgressBar * This,
            /* [retval][out] */ long *Step);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Step )( 
            ProgressBar * This,
            /* [in] */ long Step);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Marquee )( 
            ProgressBar * This,
            /* [in] */ long rhs);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Increase )( 
            ProgressBar * This,
            /* [in] */ long amount);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *IncreaseStep )( 
            ProgressBar * This);
        
        END_INTERFACE
    } ProgressBarVtbl;

    interface ProgressBar
    {
        CONST_VTBL struct ProgressBarVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ProgressBar_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ProgressBar_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ProgressBar_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ProgressBar_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ProgressBar_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ProgressBar_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ProgressBar_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ProgressBar_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define ProgressBar_put_Text(This,Text)	\
    ( (This)->lpVtbl -> put_Text(This,Text) ) 

#define ProgressBar_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ProgressBar_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ProgressBar_get_Position(This,Position)	\
    ( (This)->lpVtbl -> get_Position(This,Position) ) 

#define ProgressBar_put_Position(This,Position)	\
    ( (This)->lpVtbl -> put_Position(This,Position) ) 

#define ProgressBar_get_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> get_Enabled(This,Enabled) ) 

#define ProgressBar_put_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> put_Enabled(This,Enabled) ) 

#define ProgressBar_get_Visible(This,Visible)	\
    ( (This)->lpVtbl -> get_Visible(This,Visible) ) 

#define ProgressBar_put_Visible(This,Visible)	\
    ( (This)->lpVtbl -> put_Visible(This,Visible) ) 

#define ProgressBar_get_Focused(This,Focused)	\
    ( (This)->lpVtbl -> get_Focused(This,Focused) ) 

#define ProgressBar_get_Width(This,Width)	\
    ( (This)->lpVtbl -> get_Width(This,Width) ) 

#define ProgressBar_put_Width(This,Width)	\
    ( (This)->lpVtbl -> put_Width(This,Width) ) 

#define ProgressBar_get_Height(This,Height)	\
    ( (This)->lpVtbl -> get_Height(This,Height) ) 

#define ProgressBar_put_Height(This,Height)	\
    ( (This)->lpVtbl -> put_Height(This,Height) ) 

#define ProgressBar_get_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> get_TabOrder(This,TabOrder) ) 

#define ProgressBar_put_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> put_TabOrder(This,TabOrder) ) 

#define ProgressBar_get_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> get_TabStop(This,TabStop) ) 

#define ProgressBar_put_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> put_TabStop(This,TabStop) ) 

#define ProgressBar_get_Group(This,Group)	\
    ( (This)->lpVtbl -> get_Group(This,Group) ) 

#define ProgressBar_put_Group(This,Group)	\
    ( (This)->lpVtbl -> put_Group(This,Group) ) 

#define ProgressBar_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define ProgressBar_get_id(This,id)	\
    ( (This)->lpVtbl -> get_id(This,id) ) 

#define ProgressBar_get_Next(This,prop)	\
    ( (This)->lpVtbl -> get_Next(This,prop) ) 

#define ProgressBar_get_Previous(This,prop)	\
    ( (This)->lpVtbl -> get_Previous(This,prop) ) 

#define ProgressBar_Focus(This)	\
    ( (This)->lpVtbl -> Focus(This) ) 

#define ProgressBar_put_id(This,id)	\
    ( (This)->lpVtbl -> put_id(This,id) ) 


#define ProgressBar_get_Vertical(This,Vertical)	\
    ( (This)->lpVtbl -> get_Vertical(This,Vertical) ) 

#define ProgressBar_put_Vertical(This,Vertical)	\
    ( (This)->lpVtbl -> put_Vertical(This,Vertical) ) 

#define ProgressBar_get_Value(This,Value)	\
    ( (This)->lpVtbl -> get_Value(This,Value) ) 

#define ProgressBar_put_Value(This,Value)	\
    ( (This)->lpVtbl -> put_Value(This,Value) ) 

#define ProgressBar_get_Minimum(This,Minimum)	\
    ( (This)->lpVtbl -> get_Minimum(This,Minimum) ) 

#define ProgressBar_put_Minimum(This,Minimum)	\
    ( (This)->lpVtbl -> put_Minimum(This,Minimum) ) 

#define ProgressBar_get_Maximum(This,Maximum)	\
    ( (This)->lpVtbl -> get_Maximum(This,Maximum) ) 

#define ProgressBar_put_Maximum(This,Maximum)	\
    ( (This)->lpVtbl -> put_Maximum(This,Maximum) ) 

#define ProgressBar_get_Step(This,Step)	\
    ( (This)->lpVtbl -> get_Step(This,Step) ) 

#define ProgressBar_put_Step(This,Step)	\
    ( (This)->lpVtbl -> put_Step(This,Step) ) 

#define ProgressBar_put_Marquee(This,rhs)	\
    ( (This)->lpVtbl -> put_Marquee(This,rhs) ) 

#define ProgressBar_Increase(This,amount)	\
    ( (This)->lpVtbl -> Increase(This,amount) ) 

#define ProgressBar_IncreaseStep(This)	\
    ( (This)->lpVtbl -> IncreaseStep(This) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ProgressBar_INTERFACE_DEFINED__ */


#ifndef ___TrackBar_INTERFACE_DEFINED__
#define ___TrackBar_INTERFACE_DEFINED__

/* interface _TrackBar */
/* [object][oleautomation][dual][hidden][helpcontext][version][uuid] */ 


EXTERN_C const IID IID__TrackBar;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("FEA9E0B5-7AEB-4F65-9DF7-3696876076AD")
    _TrackBar : public _Control
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Ticks( 
            /* [retval][out] */ TickPlacement *TickPlacement) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Ticks( 
            /* [in] */ TickPlacement TickPlacement) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TickLocation( 
            /* [retval][out] */ TickLocation *TickLocation) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TickLocation( 
            /* [in] */ TickLocation TickLocation) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_DownIsLeft( 
            /* [retval][out] */ VARIANT_BOOL *DownIsLeft) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_DownIsLeft( 
            /* [in] */ VARIANT_BOOL DownIsLeft) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Vertical( 
            /* [retval][out] */ VARIANT_BOOL *Vertical) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Vertical( 
            /* [in] */ VARIANT_BOOL Vertical) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Reversed( 
            /* [retval][out] */ VARIANT_BOOL *Reversed) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Reversed( 
            /* [in] */ VARIANT_BOOL Reversed) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Value( 
            /* [retval][out] */ long *Value) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Value( 
            /* [in] */ long Value) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Minimum( 
            /* [retval][out] */ long *Minimum) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Minimum( 
            /* [in] */ long Minimum) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Maximum( 
            /* [retval][out] */ long *Maximum) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Maximum( 
            /* [in] */ long Maximum) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Line( 
            /* [retval][out] */ long *Value) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Line( 
            /* [in] */ long Value) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Page( 
            /* [retval][out] */ long *Value) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Page( 
            /* [in] */ long Value) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TickInterval( 
            /* [in] */ long rhs) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TickCount( 
            /* [retval][out] */ long *Value) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ClearTicks( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE AddTick( 
            /* [in] */ long Position) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Tick( 
            /* [in] */ long index,
            /* [retval][out] */ long *Position) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct _TrackBarVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            _TrackBar * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            _TrackBar * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            _TrackBar * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            _TrackBar * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            _TrackBar * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            _TrackBar * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            _TrackBar * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            _TrackBar * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            _TrackBar * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            _TrackBar * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            _TrackBar * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            _TrackBar * This,
            /* [retval][out] */ ClientPoint **Position);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            _TrackBar * This,
            /* [in] */ ClientPoint *Position);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Enabled )( 
            _TrackBar * This,
            /* [retval][out] */ VARIANT_BOOL *Enabled);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Enabled )( 
            _TrackBar * This,
            /* [in] */ VARIANT_BOOL Enabled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            _TrackBar * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Visible )( 
            _TrackBar * This,
            /* [in] */ VARIANT_BOOL Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            _TrackBar * This,
            /* [retval][out] */ VARIANT_BOOL *Focused);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Width )( 
            _TrackBar * This,
            /* [retval][out] */ long *Width);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Width )( 
            _TrackBar * This,
            /* [in] */ long Width);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Height )( 
            _TrackBar * This,
            /* [retval][out] */ long *Height);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Height )( 
            _TrackBar * This,
            /* [in] */ long Height);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabOrder )( 
            _TrackBar * This,
            /* [retval][out] */ long *TabOrder);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabOrder )( 
            _TrackBar * This,
            /* [in] */ long TabOrder);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabStop )( 
            _TrackBar * This,
            /* [retval][out] */ VARIANT_BOOL *TabStop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabStop )( 
            _TrackBar * This,
            /* [in] */ VARIANT_BOOL TabStop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Group )( 
            _TrackBar * This,
            /* [retval][out] */ VARIANT_BOOL *Group);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Group )( 
            _TrackBar * This,
            /* [in] */ VARIANT_BOOL Group);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            _TrackBar * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            _TrackBar * This,
            /* [retval][out] */ long *id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Next )( 
            _TrackBar * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Previous )( 
            _TrackBar * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Focus )( 
            _TrackBar * This);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_id )( 
            _TrackBar * This,
            /* [in] */ long id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Ticks )( 
            _TrackBar * This,
            /* [retval][out] */ TickPlacement *TickPlacement);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Ticks )( 
            _TrackBar * This,
            /* [in] */ TickPlacement TickPlacement);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TickLocation )( 
            _TrackBar * This,
            /* [retval][out] */ TickLocation *TickLocation);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TickLocation )( 
            _TrackBar * This,
            /* [in] */ TickLocation TickLocation);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_DownIsLeft )( 
            _TrackBar * This,
            /* [retval][out] */ VARIANT_BOOL *DownIsLeft);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_DownIsLeft )( 
            _TrackBar * This,
            /* [in] */ VARIANT_BOOL DownIsLeft);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Vertical )( 
            _TrackBar * This,
            /* [retval][out] */ VARIANT_BOOL *Vertical);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Vertical )( 
            _TrackBar * This,
            /* [in] */ VARIANT_BOOL Vertical);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Reversed )( 
            _TrackBar * This,
            /* [retval][out] */ VARIANT_BOOL *Reversed);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Reversed )( 
            _TrackBar * This,
            /* [in] */ VARIANT_BOOL Reversed);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Value )( 
            _TrackBar * This,
            /* [retval][out] */ long *Value);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Value )( 
            _TrackBar * This,
            /* [in] */ long Value);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Minimum )( 
            _TrackBar * This,
            /* [retval][out] */ long *Minimum);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Minimum )( 
            _TrackBar * This,
            /* [in] */ long Minimum);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Maximum )( 
            _TrackBar * This,
            /* [retval][out] */ long *Maximum);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Maximum )( 
            _TrackBar * This,
            /* [in] */ long Maximum);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Line )( 
            _TrackBar * This,
            /* [retval][out] */ long *Value);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Line )( 
            _TrackBar * This,
            /* [in] */ long Value);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Page )( 
            _TrackBar * This,
            /* [retval][out] */ long *Value);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Page )( 
            _TrackBar * This,
            /* [in] */ long Value);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TickInterval )( 
            _TrackBar * This,
            /* [in] */ long rhs);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TickCount )( 
            _TrackBar * This,
            /* [retval][out] */ long *Value);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ClearTicks )( 
            _TrackBar * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *AddTick )( 
            _TrackBar * This,
            /* [in] */ long Position);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Tick )( 
            _TrackBar * This,
            /* [in] */ long index,
            /* [retval][out] */ long *Position);
        
        END_INTERFACE
    } _TrackBarVtbl;

    interface _TrackBar
    {
        CONST_VTBL struct _TrackBarVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define _TrackBar_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define _TrackBar_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define _TrackBar_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define _TrackBar_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define _TrackBar_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define _TrackBar_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define _TrackBar_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define _TrackBar_get_Text(This,Text)	\
    ( (This)->lpVtbl -> get_Text(This,Text) ) 

#define _TrackBar_put_Text(This,Text)	\
    ( (This)->lpVtbl -> put_Text(This,Text) ) 

#define _TrackBar_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define _TrackBar_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define _TrackBar_get_Position(This,Position)	\
    ( (This)->lpVtbl -> get_Position(This,Position) ) 

#define _TrackBar_put_Position(This,Position)	\
    ( (This)->lpVtbl -> put_Position(This,Position) ) 

#define _TrackBar_get_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> get_Enabled(This,Enabled) ) 

#define _TrackBar_put_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> put_Enabled(This,Enabled) ) 

#define _TrackBar_get_Visible(This,Visible)	\
    ( (This)->lpVtbl -> get_Visible(This,Visible) ) 

#define _TrackBar_put_Visible(This,Visible)	\
    ( (This)->lpVtbl -> put_Visible(This,Visible) ) 

#define _TrackBar_get_Focused(This,Focused)	\
    ( (This)->lpVtbl -> get_Focused(This,Focused) ) 

#define _TrackBar_get_Width(This,Width)	\
    ( (This)->lpVtbl -> get_Width(This,Width) ) 

#define _TrackBar_put_Width(This,Width)	\
    ( (This)->lpVtbl -> put_Width(This,Width) ) 

#define _TrackBar_get_Height(This,Height)	\
    ( (This)->lpVtbl -> get_Height(This,Height) ) 

#define _TrackBar_put_Height(This,Height)	\
    ( (This)->lpVtbl -> put_Height(This,Height) ) 

#define _TrackBar_get_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> get_TabOrder(This,TabOrder) ) 

#define _TrackBar_put_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> put_TabOrder(This,TabOrder) ) 

#define _TrackBar_get_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> get_TabStop(This,TabStop) ) 

#define _TrackBar_put_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> put_TabStop(This,TabStop) ) 

#define _TrackBar_get_Group(This,Group)	\
    ( (This)->lpVtbl -> get_Group(This,Group) ) 

#define _TrackBar_put_Group(This,Group)	\
    ( (This)->lpVtbl -> put_Group(This,Group) ) 

#define _TrackBar_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define _TrackBar_get_id(This,id)	\
    ( (This)->lpVtbl -> get_id(This,id) ) 

#define _TrackBar_get_Next(This,prop)	\
    ( (This)->lpVtbl -> get_Next(This,prop) ) 

#define _TrackBar_get_Previous(This,prop)	\
    ( (This)->lpVtbl -> get_Previous(This,prop) ) 

#define _TrackBar_Focus(This)	\
    ( (This)->lpVtbl -> Focus(This) ) 

#define _TrackBar_put_id(This,id)	\
    ( (This)->lpVtbl -> put_id(This,id) ) 


#define _TrackBar_get_Ticks(This,TickPlacement)	\
    ( (This)->lpVtbl -> get_Ticks(This,TickPlacement) ) 

#define _TrackBar_put_Ticks(This,TickPlacement)	\
    ( (This)->lpVtbl -> put_Ticks(This,TickPlacement) ) 

#define _TrackBar_get_TickLocation(This,TickLocation)	\
    ( (This)->lpVtbl -> get_TickLocation(This,TickLocation) ) 

#define _TrackBar_put_TickLocation(This,TickLocation)	\
    ( (This)->lpVtbl -> put_TickLocation(This,TickLocation) ) 

#define _TrackBar_get_DownIsLeft(This,DownIsLeft)	\
    ( (This)->lpVtbl -> get_DownIsLeft(This,DownIsLeft) ) 

#define _TrackBar_put_DownIsLeft(This,DownIsLeft)	\
    ( (This)->lpVtbl -> put_DownIsLeft(This,DownIsLeft) ) 

#define _TrackBar_get_Vertical(This,Vertical)	\
    ( (This)->lpVtbl -> get_Vertical(This,Vertical) ) 

#define _TrackBar_put_Vertical(This,Vertical)	\
    ( (This)->lpVtbl -> put_Vertical(This,Vertical) ) 

#define _TrackBar_get_Reversed(This,Reversed)	\
    ( (This)->lpVtbl -> get_Reversed(This,Reversed) ) 

#define _TrackBar_put_Reversed(This,Reversed)	\
    ( (This)->lpVtbl -> put_Reversed(This,Reversed) ) 

#define _TrackBar_get_Value(This,Value)	\
    ( (This)->lpVtbl -> get_Value(This,Value) ) 

#define _TrackBar_put_Value(This,Value)	\
    ( (This)->lpVtbl -> put_Value(This,Value) ) 

#define _TrackBar_get_Minimum(This,Minimum)	\
    ( (This)->lpVtbl -> get_Minimum(This,Minimum) ) 

#define _TrackBar_put_Minimum(This,Minimum)	\
    ( (This)->lpVtbl -> put_Minimum(This,Minimum) ) 

#define _TrackBar_get_Maximum(This,Maximum)	\
    ( (This)->lpVtbl -> get_Maximum(This,Maximum) ) 

#define _TrackBar_put_Maximum(This,Maximum)	\
    ( (This)->lpVtbl -> put_Maximum(This,Maximum) ) 

#define _TrackBar_get_Line(This,Value)	\
    ( (This)->lpVtbl -> get_Line(This,Value) ) 

#define _TrackBar_put_Line(This,Value)	\
    ( (This)->lpVtbl -> put_Line(This,Value) ) 

#define _TrackBar_get_Page(This,Value)	\
    ( (This)->lpVtbl -> get_Page(This,Value) ) 

#define _TrackBar_put_Page(This,Value)	\
    ( (This)->lpVtbl -> put_Page(This,Value) ) 

#define _TrackBar_put_TickInterval(This,rhs)	\
    ( (This)->lpVtbl -> put_TickInterval(This,rhs) ) 

#define _TrackBar_get_TickCount(This,Value)	\
    ( (This)->lpVtbl -> get_TickCount(This,Value) ) 

#define _TrackBar_ClearTicks(This)	\
    ( (This)->lpVtbl -> ClearTicks(This) ) 

#define _TrackBar_AddTick(This,Position)	\
    ( (This)->lpVtbl -> AddTick(This,Position) ) 

#define _TrackBar_Tick(This,index,Position)	\
    ( (This)->lpVtbl -> Tick(This,index,Position) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___TrackBar_INTERFACE_DEFINED__ */


#ifndef __TabControl_INTERFACE_DEFINED__
#define __TabControl_INTERFACE_DEFINED__

/* interface TabControl */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_TabControl;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("13C16134-27DB-4D84-BEB2-14842AEE4B40")
    TabControl : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Text( 
            /* [optional][in] */ VARIANT index,
            /* [retval][out] */ BSTR *Text) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Text( 
            /* [optional][in] */ VARIANT index,
            /* [in] */ BSTR Text) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Position( 
            /* [retval][out] */ ClientPoint **Position) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Position( 
            /* [in] */ ClientPoint *Position) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Enabled( 
            /* [retval][out] */ VARIANT_BOOL *Enabled) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Enabled( 
            /* [in] */ VARIANT_BOOL Enabled) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Visible( 
            /* [retval][out] */ VARIANT_BOOL *Visible) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Visible( 
            /* [in] */ VARIANT_BOOL Visible) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Focused( 
            /* [retval][out] */ VARIANT_BOOL *Focused) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Width( 
            /* [retval][out] */ long *Width) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Width( 
            /* [in] */ long Width) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Height( 
            /* [retval][out] */ long *Height) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Height( 
            /* [in] */ long Height) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TabOrder( 
            /* [retval][out] */ long *TabOrder) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TabOrder( 
            /* [in] */ long TabOrder) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TabStop( 
            /* [retval][out] */ VARIANT_BOOL *TabStop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TabStop( 
            /* [in] */ VARIANT_BOOL TabStop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Group( 
            /* [retval][out] */ VARIANT_BOOL *Group) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Group( 
            /* [in] */ VARIANT_BOOL Group) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Window( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_id( 
            /* [retval][out] */ long *id) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Next( 
            /* [retval][out] */ IDispatch **prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Previous( 
            /* [retval][out] */ IDispatch **prop) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Focus( void) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Multiline( 
            /* [retval][out] */ VARIANT_BOOL *Multiline) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Multiline( 
            /* [in] */ VARIANT_BOOL Multiline) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Justified( 
            /* [retval][out] */ VARIANT_BOOL *Justified) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Justified( 
            /* [in] */ VARIANT_BOOL Justified) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_FocusedIndex( 
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_FocusedIndex( 
            /* [in] */ long index) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_SelectedIndex( 
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_SelectedIndex( 
            /* [in] */ long index) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Rows( 
            /* [retval][out] */ long *Rows) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Clear( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Delete( 
            /* [in] */ long index) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Insert( 
            /* [in] */ long index,
            /* [in] */ BSTR Text) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_id( 
            /* [in] */ long id) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct TabControlVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            TabControl * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            TabControl * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            TabControl * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            TabControl * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            TabControl * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            TabControl * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            TabControl * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            TabControl * This,
            /* [optional][in] */ VARIANT index,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            TabControl * This,
            /* [optional][in] */ VARIANT index,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            TabControl * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            TabControl * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            TabControl * This,
            /* [retval][out] */ ClientPoint **Position);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            TabControl * This,
            /* [in] */ ClientPoint *Position);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Enabled )( 
            TabControl * This,
            /* [retval][out] */ VARIANT_BOOL *Enabled);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Enabled )( 
            TabControl * This,
            /* [in] */ VARIANT_BOOL Enabled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            TabControl * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Visible )( 
            TabControl * This,
            /* [in] */ VARIANT_BOOL Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            TabControl * This,
            /* [retval][out] */ VARIANT_BOOL *Focused);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Width )( 
            TabControl * This,
            /* [retval][out] */ long *Width);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Width )( 
            TabControl * This,
            /* [in] */ long Width);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Height )( 
            TabControl * This,
            /* [retval][out] */ long *Height);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Height )( 
            TabControl * This,
            /* [in] */ long Height);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabOrder )( 
            TabControl * This,
            /* [retval][out] */ long *TabOrder);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabOrder )( 
            TabControl * This,
            /* [in] */ long TabOrder);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabStop )( 
            TabControl * This,
            /* [retval][out] */ VARIANT_BOOL *TabStop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabStop )( 
            TabControl * This,
            /* [in] */ VARIANT_BOOL TabStop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Group )( 
            TabControl * This,
            /* [retval][out] */ VARIANT_BOOL *Group);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Group )( 
            TabControl * This,
            /* [in] */ VARIANT_BOOL Group);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            TabControl * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            TabControl * This,
            /* [retval][out] */ long *id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Next )( 
            TabControl * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Previous )( 
            TabControl * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Focus )( 
            TabControl * This);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Multiline )( 
            TabControl * This,
            /* [retval][out] */ VARIANT_BOOL *Multiline);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Multiline )( 
            TabControl * This,
            /* [in] */ VARIANT_BOOL Multiline);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Justified )( 
            TabControl * This,
            /* [retval][out] */ VARIANT_BOOL *Justified);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Justified )( 
            TabControl * This,
            /* [in] */ VARIANT_BOOL Justified);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            TabControl * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_FocusedIndex )( 
            TabControl * This,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_FocusedIndex )( 
            TabControl * This,
            /* [in] */ long index);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_SelectedIndex )( 
            TabControl * This,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_SelectedIndex )( 
            TabControl * This,
            /* [in] */ long index);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Rows )( 
            TabControl * This,
            /* [retval][out] */ long *Rows);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Clear )( 
            TabControl * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Delete )( 
            TabControl * This,
            /* [in] */ long index);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Insert )( 
            TabControl * This,
            /* [in] */ long index,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_id )( 
            TabControl * This,
            /* [in] */ long id);
        
        END_INTERFACE
    } TabControlVtbl;

    interface TabControl
    {
        CONST_VTBL struct TabControlVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define TabControl_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define TabControl_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define TabControl_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define TabControl_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define TabControl_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define TabControl_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define TabControl_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define TabControl_get_Text(This,index,Text)	\
    ( (This)->lpVtbl -> get_Text(This,index,Text) ) 

#define TabControl_put_Text(This,index,Text)	\
    ( (This)->lpVtbl -> put_Text(This,index,Text) ) 

#define TabControl_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define TabControl_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define TabControl_get_Position(This,Position)	\
    ( (This)->lpVtbl -> get_Position(This,Position) ) 

#define TabControl_put_Position(This,Position)	\
    ( (This)->lpVtbl -> put_Position(This,Position) ) 

#define TabControl_get_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> get_Enabled(This,Enabled) ) 

#define TabControl_put_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> put_Enabled(This,Enabled) ) 

#define TabControl_get_Visible(This,Visible)	\
    ( (This)->lpVtbl -> get_Visible(This,Visible) ) 

#define TabControl_put_Visible(This,Visible)	\
    ( (This)->lpVtbl -> put_Visible(This,Visible) ) 

#define TabControl_get_Focused(This,Focused)	\
    ( (This)->lpVtbl -> get_Focused(This,Focused) ) 

#define TabControl_get_Width(This,Width)	\
    ( (This)->lpVtbl -> get_Width(This,Width) ) 

#define TabControl_put_Width(This,Width)	\
    ( (This)->lpVtbl -> put_Width(This,Width) ) 

#define TabControl_get_Height(This,Height)	\
    ( (This)->lpVtbl -> get_Height(This,Height) ) 

#define TabControl_put_Height(This,Height)	\
    ( (This)->lpVtbl -> put_Height(This,Height) ) 

#define TabControl_get_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> get_TabOrder(This,TabOrder) ) 

#define TabControl_put_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> put_TabOrder(This,TabOrder) ) 

#define TabControl_get_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> get_TabStop(This,TabStop) ) 

#define TabControl_put_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> put_TabStop(This,TabStop) ) 

#define TabControl_get_Group(This,Group)	\
    ( (This)->lpVtbl -> get_Group(This,Group) ) 

#define TabControl_put_Group(This,Group)	\
    ( (This)->lpVtbl -> put_Group(This,Group) ) 

#define TabControl_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define TabControl_get_id(This,id)	\
    ( (This)->lpVtbl -> get_id(This,id) ) 

#define TabControl_get_Next(This,prop)	\
    ( (This)->lpVtbl -> get_Next(This,prop) ) 

#define TabControl_get_Previous(This,prop)	\
    ( (This)->lpVtbl -> get_Previous(This,prop) ) 

#define TabControl_Focus(This)	\
    ( (This)->lpVtbl -> Focus(This) ) 

#define TabControl_get_Multiline(This,Multiline)	\
    ( (This)->lpVtbl -> get_Multiline(This,Multiline) ) 

#define TabControl_put_Multiline(This,Multiline)	\
    ( (This)->lpVtbl -> put_Multiline(This,Multiline) ) 

#define TabControl_get_Justified(This,Justified)	\
    ( (This)->lpVtbl -> get_Justified(This,Justified) ) 

#define TabControl_put_Justified(This,Justified)	\
    ( (This)->lpVtbl -> put_Justified(This,Justified) ) 

#define TabControl_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define TabControl_get_FocusedIndex(This,index)	\
    ( (This)->lpVtbl -> get_FocusedIndex(This,index) ) 

#define TabControl_put_FocusedIndex(This,index)	\
    ( (This)->lpVtbl -> put_FocusedIndex(This,index) ) 

#define TabControl_get_SelectedIndex(This,index)	\
    ( (This)->lpVtbl -> get_SelectedIndex(This,index) ) 

#define TabControl_put_SelectedIndex(This,index)	\
    ( (This)->lpVtbl -> put_SelectedIndex(This,index) ) 

#define TabControl_get_Rows(This,Rows)	\
    ( (This)->lpVtbl -> get_Rows(This,Rows) ) 

#define TabControl_Clear(This)	\
    ( (This)->lpVtbl -> Clear(This) ) 

#define TabControl_Delete(This,index)	\
    ( (This)->lpVtbl -> Delete(This,index) ) 

#define TabControl_Insert(This,index,Text)	\
    ( (This)->lpVtbl -> Insert(This,index,Text) ) 

#define TabControl_put_id(This,id)	\
    ( (This)->lpVtbl -> put_id(This,id) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __TabControl_INTERFACE_DEFINED__ */


#ifndef __TreeViewEventInfo_INTERFACE_DEFINED__
#define __TreeViewEventInfo_INTERFACE_DEFINED__

/* interface TreeViewEventInfo */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_TreeViewEventInfo;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("723D6972-A7FE-46AC-BABF-69F37D28C953")
    TreeViewEventInfo : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_OldItem( 
            /* [retval][out] */ TreeViewItem **OldItem) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_NewItem( 
            /* [retval][out] */ TreeViewItem **NewItem) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Action( 
            /* [retval][out] */ long *Action) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct TreeViewEventInfoVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            TreeViewEventInfo * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            TreeViewEventInfo * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            TreeViewEventInfo * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            TreeViewEventInfo * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            TreeViewEventInfo * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            TreeViewEventInfo * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            TreeViewEventInfo * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            TreeViewEventInfo * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            TreeViewEventInfo * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_OldItem )( 
            TreeViewEventInfo * This,
            /* [retval][out] */ TreeViewItem **OldItem);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_NewItem )( 
            TreeViewEventInfo * This,
            /* [retval][out] */ TreeViewItem **NewItem);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Action )( 
            TreeViewEventInfo * This,
            /* [retval][out] */ long *Action);
        
        END_INTERFACE
    } TreeViewEventInfoVtbl;

    interface TreeViewEventInfo
    {
        CONST_VTBL struct TreeViewEventInfoVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define TreeViewEventInfo_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define TreeViewEventInfo_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define TreeViewEventInfo_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define TreeViewEventInfo_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define TreeViewEventInfo_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define TreeViewEventInfo_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define TreeViewEventInfo_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define TreeViewEventInfo_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define TreeViewEventInfo_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define TreeViewEventInfo_get_OldItem(This,OldItem)	\
    ( (This)->lpVtbl -> get_OldItem(This,OldItem) ) 

#define TreeViewEventInfo_get_NewItem(This,NewItem)	\
    ( (This)->lpVtbl -> get_NewItem(This,NewItem) ) 

#define TreeViewEventInfo_get_Action(This,Action)	\
    ( (This)->lpVtbl -> get_Action(This,Action) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __TreeViewEventInfo_INTERFACE_DEFINED__ */


#ifndef __ListViewItem_INTERFACE_DEFINED__
#define __ListViewItem_INTERFACE_DEFINED__

/* interface ListViewItem */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ListViewItem;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("BD6EA316-F119-4E3B-B6D5-0F89AA4CEB32")
    ListViewItem : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Text( 
            /* [optional][in] */ VARIANT subitem,
            /* [retval][out] */ BSTR *Text) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Text( 
            /* [optional][in] */ VARIANT subitem,
            /* [in] */ BSTR Text) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Indent( 
            /* [retval][out] */ long *Indent) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Indent( 
            /* [in] */ long Indent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Data( 
            /* [retval][out] */ long *Data) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Data( 
            /* [in] */ long Data) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Focused( 
            /* [retval][out] */ VARIANT_BOOL *Focused) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Focused( 
            /* [in] */ VARIANT_BOOL Focused) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Selected( 
            /* [retval][out] */ VARIANT_BOOL *Selected) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Selected( 
            /* [in] */ VARIANT_BOOL Selected) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Image( 
            /* [optional][in] */ VARIANT subitem,
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Image( 
            /* [optional][in] */ VARIANT subitem,
            /* [in] */ long index) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_OverlayImage( 
            /* [optional][in] */ VARIANT subitem,
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_OverlayImage( 
            /* [optional][in] */ VARIANT subitem,
            /* [in] */ long index) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_StateImage( 
            /* [optional][in] */ VARIANT subitem,
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_StateImage( 
            /* [optional][in] */ VARIANT subitem,
            /* [in] */ long index) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Checked( 
            /* [retval][out] */ listviewCheck *Checked) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Checked( 
            /* [in] */ listviewCheck Checked) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_index( 
            /* [retval][out] */ long *index) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Delete( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE ScrollTo( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Insert( 
            /* [in] */ long index) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ListViewItemVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ListViewItem * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ListViewItem * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ListViewItem * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ListViewItem * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ListViewItem * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ListViewItem * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ListViewItem * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            ListViewItem * This,
            /* [optional][in] */ VARIANT subitem,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            ListViewItem * This,
            /* [optional][in] */ VARIANT subitem,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ListViewItem * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ListViewItem * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Indent )( 
            ListViewItem * This,
            /* [retval][out] */ long *Indent);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Indent )( 
            ListViewItem * This,
            /* [in] */ long Indent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Data )( 
            ListViewItem * This,
            /* [retval][out] */ long *Data);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Data )( 
            ListViewItem * This,
            /* [in] */ long Data);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            ListViewItem * This,
            /* [retval][out] */ VARIANT_BOOL *Focused);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Focused )( 
            ListViewItem * This,
            /* [in] */ VARIANT_BOOL Focused);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Selected )( 
            ListViewItem * This,
            /* [retval][out] */ VARIANT_BOOL *Selected);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Selected )( 
            ListViewItem * This,
            /* [in] */ VARIANT_BOOL Selected);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Image )( 
            ListViewItem * This,
            /* [optional][in] */ VARIANT subitem,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Image )( 
            ListViewItem * This,
            /* [optional][in] */ VARIANT subitem,
            /* [in] */ long index);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_OverlayImage )( 
            ListViewItem * This,
            /* [optional][in] */ VARIANT subitem,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_OverlayImage )( 
            ListViewItem * This,
            /* [optional][in] */ VARIANT subitem,
            /* [in] */ long index);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_StateImage )( 
            ListViewItem * This,
            /* [optional][in] */ VARIANT subitem,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_StateImage )( 
            ListViewItem * This,
            /* [optional][in] */ VARIANT subitem,
            /* [in] */ long index);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Checked )( 
            ListViewItem * This,
            /* [retval][out] */ listviewCheck *Checked);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Checked )( 
            ListViewItem * This,
            /* [in] */ listviewCheck Checked);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_index )( 
            ListViewItem * This,
            /* [retval][out] */ long *index);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Delete )( 
            ListViewItem * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *ScrollTo )( 
            ListViewItem * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Insert )( 
            ListViewItem * This,
            /* [in] */ long index);
        
        END_INTERFACE
    } ListViewItemVtbl;

    interface ListViewItem
    {
        CONST_VTBL struct ListViewItemVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ListViewItem_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ListViewItem_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ListViewItem_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ListViewItem_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ListViewItem_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ListViewItem_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ListViewItem_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ListViewItem_get_Text(This,subitem,Text)	\
    ( (This)->lpVtbl -> get_Text(This,subitem,Text) ) 

#define ListViewItem_put_Text(This,subitem,Text)	\
    ( (This)->lpVtbl -> put_Text(This,subitem,Text) ) 

#define ListViewItem_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ListViewItem_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ListViewItem_get_Indent(This,Indent)	\
    ( (This)->lpVtbl -> get_Indent(This,Indent) ) 

#define ListViewItem_put_Indent(This,Indent)	\
    ( (This)->lpVtbl -> put_Indent(This,Indent) ) 

#define ListViewItem_get_Data(This,Data)	\
    ( (This)->lpVtbl -> get_Data(This,Data) ) 

#define ListViewItem_put_Data(This,Data)	\
    ( (This)->lpVtbl -> put_Data(This,Data) ) 

#define ListViewItem_get_Focused(This,Focused)	\
    ( (This)->lpVtbl -> get_Focused(This,Focused) ) 

#define ListViewItem_put_Focused(This,Focused)	\
    ( (This)->lpVtbl -> put_Focused(This,Focused) ) 

#define ListViewItem_get_Selected(This,Selected)	\
    ( (This)->lpVtbl -> get_Selected(This,Selected) ) 

#define ListViewItem_put_Selected(This,Selected)	\
    ( (This)->lpVtbl -> put_Selected(This,Selected) ) 

#define ListViewItem_get_Image(This,subitem,index)	\
    ( (This)->lpVtbl -> get_Image(This,subitem,index) ) 

#define ListViewItem_put_Image(This,subitem,index)	\
    ( (This)->lpVtbl -> put_Image(This,subitem,index) ) 

#define ListViewItem_get_OverlayImage(This,subitem,index)	\
    ( (This)->lpVtbl -> get_OverlayImage(This,subitem,index) ) 

#define ListViewItem_put_OverlayImage(This,subitem,index)	\
    ( (This)->lpVtbl -> put_OverlayImage(This,subitem,index) ) 

#define ListViewItem_get_StateImage(This,subitem,index)	\
    ( (This)->lpVtbl -> get_StateImage(This,subitem,index) ) 

#define ListViewItem_put_StateImage(This,subitem,index)	\
    ( (This)->lpVtbl -> put_StateImage(This,subitem,index) ) 

#define ListViewItem_get_Checked(This,Checked)	\
    ( (This)->lpVtbl -> get_Checked(This,Checked) ) 

#define ListViewItem_put_Checked(This,Checked)	\
    ( (This)->lpVtbl -> put_Checked(This,Checked) ) 

#define ListViewItem_get_index(This,index)	\
    ( (This)->lpVtbl -> get_index(This,index) ) 

#define ListViewItem_Delete(This)	\
    ( (This)->lpVtbl -> Delete(This) ) 

#define ListViewItem_ScrollTo(This)	\
    ( (This)->lpVtbl -> ScrollTo(This) ) 

#define ListViewItem_Insert(This,index)	\
    ( (This)->lpVtbl -> Insert(This,index) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ListViewItem_INTERFACE_DEFINED__ */


#ifndef __ListViewItems_INTERFACE_DEFINED__
#define __ListViewItems_INTERFACE_DEFINED__

/* interface ListViewItems */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ListViewItems;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("1189BF9D-C69F-43CB-AD9A-39B77D2744CD")
    ListViewItems : public IDispatch
    {
    public:
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Selected( 
            /* [retval][out] */ ListViewItems **selectedItems) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Focused( 
            /* [retval][out] */ ListViewItem **focusedItem) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ ListViewItem **Item) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE FindText( 
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT partial,
            /* [retval][out] */ ListViewItems **items) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE FindData( 
            /* [in] */ long Data,
            /* [retval][out] */ ListViewItems **items) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ListViewItemsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ListViewItems * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ListViewItems * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ListViewItems * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ListViewItems * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ListViewItems * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ListViewItems * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ListViewItems * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            ListViewItems * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ListViewItems * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ListViewItems * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            ListViewItems * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Selected )( 
            ListViewItems * This,
            /* [retval][out] */ ListViewItems **selectedItems);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            ListViewItems * This,
            /* [retval][out] */ ListViewItem **focusedItem);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            ListViewItems * This,
            /* [in] */ long index,
            /* [retval][out] */ ListViewItem **Item);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *FindText )( 
            ListViewItems * This,
            /* [in] */ BSTR Text,
            /* [optional][in] */ VARIANT partial,
            /* [retval][out] */ ListViewItems **items);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *FindData )( 
            ListViewItems * This,
            /* [in] */ long Data,
            /* [retval][out] */ ListViewItems **items);
        
        END_INTERFACE
    } ListViewItemsVtbl;

    interface ListViewItems
    {
        CONST_VTBL struct ListViewItemsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ListViewItems_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ListViewItems_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ListViewItems_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ListViewItems_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ListViewItems_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ListViewItems_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ListViewItems_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ListViewItems_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define ListViewItems_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ListViewItems_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ListViewItems_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define ListViewItems_get_Selected(This,selectedItems)	\
    ( (This)->lpVtbl -> get_Selected(This,selectedItems) ) 

#define ListViewItems_get_Focused(This,focusedItem)	\
    ( (This)->lpVtbl -> get_Focused(This,focusedItem) ) 

#define ListViewItems_Item(This,index,Item)	\
    ( (This)->lpVtbl -> Item(This,index,Item) ) 

#define ListViewItems_FindText(This,Text,partial,items)	\
    ( (This)->lpVtbl -> FindText(This,Text,partial,items) ) 

#define ListViewItems_FindData(This,Data,items)	\
    ( (This)->lpVtbl -> FindData(This,Data,items) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ListViewItems_INTERFACE_DEFINED__ */


#ifndef __ListViewColumn_INTERFACE_DEFINED__
#define __ListViewColumn_INTERFACE_DEFINED__

/* interface ListViewColumn */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ListViewColumn;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("202E439E-5524-4CFA-926D-A9BD4A602B05")
    ListViewColumn : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Header( 
            /* [retval][out] */ BSTR *Text) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Header( 
            /* [in] */ BSTR Text) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_alignment( 
            /* [retval][out] */ TextAlignment *align) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_alignment( 
            /* [in] */ TextAlignment align) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Width( 
            /* [retval][out] */ long *Width) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Width( 
            /* [in] */ long Width) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_subitem( 
            /* [retval][out] */ long *subitem) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_subitem( 
            /* [in] */ long subitem) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_order( 
            /* [retval][out] */ long *order) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_order( 
            /* [in] */ long order) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Delete( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Insert( 
            /* [in] */ long index) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ListViewColumnVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ListViewColumn * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ListViewColumn * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ListViewColumn * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ListViewColumn * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ListViewColumn * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ListViewColumn * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ListViewColumn * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Header )( 
            ListViewColumn * This,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Header )( 
            ListViewColumn * This,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ListViewColumn * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ListViewColumn * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_alignment )( 
            ListViewColumn * This,
            /* [retval][out] */ TextAlignment *align);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_alignment )( 
            ListViewColumn * This,
            /* [in] */ TextAlignment align);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Width )( 
            ListViewColumn * This,
            /* [retval][out] */ long *Width);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Width )( 
            ListViewColumn * This,
            /* [in] */ long Width);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_subitem )( 
            ListViewColumn * This,
            /* [retval][out] */ long *subitem);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_subitem )( 
            ListViewColumn * This,
            /* [in] */ long subitem);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_order )( 
            ListViewColumn * This,
            /* [retval][out] */ long *order);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_order )( 
            ListViewColumn * This,
            /* [in] */ long order);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Delete )( 
            ListViewColumn * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Insert )( 
            ListViewColumn * This,
            /* [in] */ long index);
        
        END_INTERFACE
    } ListViewColumnVtbl;

    interface ListViewColumn
    {
        CONST_VTBL struct ListViewColumnVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ListViewColumn_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ListViewColumn_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ListViewColumn_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ListViewColumn_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ListViewColumn_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ListViewColumn_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ListViewColumn_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ListViewColumn_get_Header(This,Text)	\
    ( (This)->lpVtbl -> get_Header(This,Text) ) 

#define ListViewColumn_put_Header(This,Text)	\
    ( (This)->lpVtbl -> put_Header(This,Text) ) 

#define ListViewColumn_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ListViewColumn_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ListViewColumn_get_alignment(This,align)	\
    ( (This)->lpVtbl -> get_alignment(This,align) ) 

#define ListViewColumn_put_alignment(This,align)	\
    ( (This)->lpVtbl -> put_alignment(This,align) ) 

#define ListViewColumn_get_Width(This,Width)	\
    ( (This)->lpVtbl -> get_Width(This,Width) ) 

#define ListViewColumn_put_Width(This,Width)	\
    ( (This)->lpVtbl -> put_Width(This,Width) ) 

#define ListViewColumn_get_subitem(This,subitem)	\
    ( (This)->lpVtbl -> get_subitem(This,subitem) ) 

#define ListViewColumn_put_subitem(This,subitem)	\
    ( (This)->lpVtbl -> put_subitem(This,subitem) ) 

#define ListViewColumn_get_order(This,order)	\
    ( (This)->lpVtbl -> get_order(This,order) ) 

#define ListViewColumn_put_order(This,order)	\
    ( (This)->lpVtbl -> put_order(This,order) ) 

#define ListViewColumn_Delete(This)	\
    ( (This)->lpVtbl -> Delete(This) ) 

#define ListViewColumn_Insert(This,index)	\
    ( (This)->lpVtbl -> Insert(This,index) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ListViewColumn_INTERFACE_DEFINED__ */


#ifndef __ListViewColumns_INTERFACE_DEFINED__
#define __ListViewColumns_INTERFACE_DEFINED__

/* interface ListViewColumns */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ListViewColumns;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("88E0CD20-399A-4D9E-8EE1-616A4049ADA3")
    ListViewColumns : public IDispatch
    {
    public:
        virtual /* [hidden][propget][id] */ HRESULT STDMETHODCALLTYPE get__NewEnum( 
            /* [retval][out] */ IUnknown **pEnum) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Count( 
            /* [retval][out] */ long *Count) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Selected( 
            /* [retval][out] */ ListViewColumn **Selected) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Item( 
            /* [in] */ long index,
            /* [retval][out] */ ListViewColumn **Item) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ListViewColumnsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ListViewColumns * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ListViewColumns * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ListViewColumns * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ListViewColumns * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ListViewColumns * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ListViewColumns * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ListViewColumns * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [hidden][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get__NewEnum )( 
            ListViewColumns * This,
            /* [retval][out] */ IUnknown **pEnum);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ListViewColumns * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ListViewColumns * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Count )( 
            ListViewColumns * This,
            /* [retval][out] */ long *Count);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Selected )( 
            ListViewColumns * This,
            /* [retval][out] */ ListViewColumn **Selected);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Item )( 
            ListViewColumns * This,
            /* [in] */ long index,
            /* [retval][out] */ ListViewColumn **Item);
        
        END_INTERFACE
    } ListViewColumnsVtbl;

    interface ListViewColumns
    {
        CONST_VTBL struct ListViewColumnsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ListViewColumns_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ListViewColumns_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ListViewColumns_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ListViewColumns_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ListViewColumns_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ListViewColumns_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ListViewColumns_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ListViewColumns_get__NewEnum(This,pEnum)	\
    ( (This)->lpVtbl -> get__NewEnum(This,pEnum) ) 

#define ListViewColumns_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ListViewColumns_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ListViewColumns_get_Count(This,Count)	\
    ( (This)->lpVtbl -> get_Count(This,Count) ) 

#define ListViewColumns_get_Selected(This,Selected)	\
    ( (This)->lpVtbl -> get_Selected(This,Selected) ) 

#define ListViewColumns_Item(This,index,Item)	\
    ( (This)->lpVtbl -> Item(This,index,Item) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ListViewColumns_INTERFACE_DEFINED__ */


#ifndef __ListView_INTERFACE_DEFINED__
#define __ListView_INTERFACE_DEFINED__

/* interface ListView */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_ListView;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("4FD84575-2DE1-4823-89FD-CA13BE72ABA5")
    ListView : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Text( 
            /* [optional][in] */ VARIANT index,
            /* [optional][in] */ VARIANT subindex,
            /* [retval][out] */ BSTR *Text) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Text( 
            /* [optional][in] */ VARIANT index,
            /* [optional][in] */ VARIANT subindex,
            /* [in] */ BSTR Text) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Position( 
            /* [retval][out] */ ClientPoint **Position) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Position( 
            /* [in] */ ClientPoint *Position) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Enabled( 
            /* [retval][out] */ VARIANT_BOOL *Enabled) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Enabled( 
            /* [in] */ VARIANT_BOOL Enabled) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Visible( 
            /* [retval][out] */ VARIANT_BOOL *Visible) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Visible( 
            /* [in] */ VARIANT_BOOL Visible) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Focused( 
            /* [retval][out] */ VARIANT_BOOL *Focused) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Width( 
            /* [retval][out] */ long *Width) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Width( 
            /* [in] */ long Width) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Height( 
            /* [retval][out] */ long *Height) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Height( 
            /* [in] */ long Height) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TabOrder( 
            /* [retval][out] */ long *TabOrder) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TabOrder( 
            /* [in] */ long TabOrder) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_TabStop( 
            /* [retval][out] */ VARIANT_BOOL *TabStop) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_TabStop( 
            /* [in] */ VARIANT_BOOL TabStop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Group( 
            /* [retval][out] */ VARIANT_BOOL *Group) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_Group( 
            /* [in] */ VARIANT_BOOL Group) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Window( 
            /* [retval][out] */ _Window **Window) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_id( 
            /* [retval][out] */ long *id) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Next( 
            /* [retval][out] */ IDispatch **prop) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Previous( 
            /* [retval][out] */ IDispatch **prop) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Focus( void) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_View( 
            /* [retval][out] */ listviewView *View) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_View( 
            /* [in] */ listviewView View) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_ColumnHeaders( 
            /* [retval][out] */ VARIANT_BOOL *headers) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Sort( 
            /* [retval][out] */ listviewSort *Sort) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Checkboxes( 
            /* [retval][out] */ VARIANT_BOOL *Checkboxes) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_items( 
            /* [retval][out] */ ListViewItems **items) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Columns( 
            /* [retval][out] */ ListViewColumns **Columns) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE Clear( void) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE NewItem( 
            /* [retval][out] */ ListViewItem **ListViewItem) = 0;
        
        virtual /* [helpcontext][helpstring][id] */ HRESULT STDMETHODCALLTYPE NewColumn( 
            /* [retval][out] */ ListViewColumn **ListViewColumn) = 0;
        
        virtual /* [helpcontext][helpstring][propput][id] */ HRESULT STDMETHODCALLTYPE put_id( 
            /* [in] */ long id) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct ListViewVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            ListView * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            ListView * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            ListView * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            ListView * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            ListView * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            ListView * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            ListView * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Text )( 
            ListView * This,
            /* [optional][in] */ VARIANT index,
            /* [optional][in] */ VARIANT subindex,
            /* [retval][out] */ BSTR *Text);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Text )( 
            ListView * This,
            /* [optional][in] */ VARIANT index,
            /* [optional][in] */ VARIANT subindex,
            /* [in] */ BSTR Text);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            ListView * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            ListView * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Position )( 
            ListView * This,
            /* [retval][out] */ ClientPoint **Position);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Position )( 
            ListView * This,
            /* [in] */ ClientPoint *Position);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Enabled )( 
            ListView * This,
            /* [retval][out] */ VARIANT_BOOL *Enabled);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Enabled )( 
            ListView * This,
            /* [in] */ VARIANT_BOOL Enabled);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Visible )( 
            ListView * This,
            /* [retval][out] */ VARIANT_BOOL *Visible);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Visible )( 
            ListView * This,
            /* [in] */ VARIANT_BOOL Visible);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Focused )( 
            ListView * This,
            /* [retval][out] */ VARIANT_BOOL *Focused);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Width )( 
            ListView * This,
            /* [retval][out] */ long *Width);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Width )( 
            ListView * This,
            /* [in] */ long Width);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Height )( 
            ListView * This,
            /* [retval][out] */ long *Height);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Height )( 
            ListView * This,
            /* [in] */ long Height);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabOrder )( 
            ListView * This,
            /* [retval][out] */ long *TabOrder);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabOrder )( 
            ListView * This,
            /* [in] */ long TabOrder);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_TabStop )( 
            ListView * This,
            /* [retval][out] */ VARIANT_BOOL *TabStop);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_TabStop )( 
            ListView * This,
            /* [in] */ VARIANT_BOOL TabStop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Group )( 
            ListView * This,
            /* [retval][out] */ VARIANT_BOOL *Group);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_Group )( 
            ListView * This,
            /* [in] */ VARIANT_BOOL Group);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Window )( 
            ListView * This,
            /* [retval][out] */ _Window **Window);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_id )( 
            ListView * This,
            /* [retval][out] */ long *id);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Next )( 
            ListView * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Previous )( 
            ListView * This,
            /* [retval][out] */ IDispatch **prop);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Focus )( 
            ListView * This);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_View )( 
            ListView * This,
            /* [retval][out] */ listviewView *View);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_View )( 
            ListView * This,
            /* [in] */ listviewView View);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_ColumnHeaders )( 
            ListView * This,
            /* [retval][out] */ VARIANT_BOOL *headers);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Sort )( 
            ListView * This,
            /* [retval][out] */ listviewSort *Sort);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Checkboxes )( 
            ListView * This,
            /* [retval][out] */ VARIANT_BOOL *Checkboxes);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_items )( 
            ListView * This,
            /* [retval][out] */ ListViewItems **items);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Columns )( 
            ListView * This,
            /* [retval][out] */ ListViewColumns **Columns);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Clear )( 
            ListView * This);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *NewItem )( 
            ListView * This,
            /* [retval][out] */ ListViewItem **ListViewItem);
        
        /* [helpcontext][helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *NewColumn )( 
            ListView * This,
            /* [retval][out] */ ListViewColumn **ListViewColumn);
        
        /* [helpcontext][helpstring][propput][id] */ HRESULT ( STDMETHODCALLTYPE *put_id )( 
            ListView * This,
            /* [in] */ long id);
        
        END_INTERFACE
    } ListViewVtbl;

    interface ListView
    {
        CONST_VTBL struct ListViewVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define ListView_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define ListView_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define ListView_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define ListView_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define ListView_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define ListView_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define ListView_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define ListView_get_Text(This,index,subindex,Text)	\
    ( (This)->lpVtbl -> get_Text(This,index,subindex,Text) ) 

#define ListView_put_Text(This,index,subindex,Text)	\
    ( (This)->lpVtbl -> put_Text(This,index,subindex,Text) ) 

#define ListView_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define ListView_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define ListView_get_Position(This,Position)	\
    ( (This)->lpVtbl -> get_Position(This,Position) ) 

#define ListView_put_Position(This,Position)	\
    ( (This)->lpVtbl -> put_Position(This,Position) ) 

#define ListView_get_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> get_Enabled(This,Enabled) ) 

#define ListView_put_Enabled(This,Enabled)	\
    ( (This)->lpVtbl -> put_Enabled(This,Enabled) ) 

#define ListView_get_Visible(This,Visible)	\
    ( (This)->lpVtbl -> get_Visible(This,Visible) ) 

#define ListView_put_Visible(This,Visible)	\
    ( (This)->lpVtbl -> put_Visible(This,Visible) ) 

#define ListView_get_Focused(This,Focused)	\
    ( (This)->lpVtbl -> get_Focused(This,Focused) ) 

#define ListView_get_Width(This,Width)	\
    ( (This)->lpVtbl -> get_Width(This,Width) ) 

#define ListView_put_Width(This,Width)	\
    ( (This)->lpVtbl -> put_Width(This,Width) ) 

#define ListView_get_Height(This,Height)	\
    ( (This)->lpVtbl -> get_Height(This,Height) ) 

#define ListView_put_Height(This,Height)	\
    ( (This)->lpVtbl -> put_Height(This,Height) ) 

#define ListView_get_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> get_TabOrder(This,TabOrder) ) 

#define ListView_put_TabOrder(This,TabOrder)	\
    ( (This)->lpVtbl -> put_TabOrder(This,TabOrder) ) 

#define ListView_get_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> get_TabStop(This,TabStop) ) 

#define ListView_put_TabStop(This,TabStop)	\
    ( (This)->lpVtbl -> put_TabStop(This,TabStop) ) 

#define ListView_get_Group(This,Group)	\
    ( (This)->lpVtbl -> get_Group(This,Group) ) 

#define ListView_put_Group(This,Group)	\
    ( (This)->lpVtbl -> put_Group(This,Group) ) 

#define ListView_get_Window(This,Window)	\
    ( (This)->lpVtbl -> get_Window(This,Window) ) 

#define ListView_get_id(This,id)	\
    ( (This)->lpVtbl -> get_id(This,id) ) 

#define ListView_get_Next(This,prop)	\
    ( (This)->lpVtbl -> get_Next(This,prop) ) 

#define ListView_get_Previous(This,prop)	\
    ( (This)->lpVtbl -> get_Previous(This,prop) ) 

#define ListView_Focus(This)	\
    ( (This)->lpVtbl -> Focus(This) ) 

#define ListView_get_View(This,View)	\
    ( (This)->lpVtbl -> get_View(This,View) ) 

#define ListView_put_View(This,View)	\
    ( (This)->lpVtbl -> put_View(This,View) ) 

#define ListView_get_ColumnHeaders(This,headers)	\
    ( (This)->lpVtbl -> get_ColumnHeaders(This,headers) ) 

#define ListView_get_Sort(This,Sort)	\
    ( (This)->lpVtbl -> get_Sort(This,Sort) ) 

#define ListView_get_Checkboxes(This,Checkboxes)	\
    ( (This)->lpVtbl -> get_Checkboxes(This,Checkboxes) ) 

#define ListView_get_items(This,items)	\
    ( (This)->lpVtbl -> get_items(This,items) ) 

#define ListView_get_Columns(This,Columns)	\
    ( (This)->lpVtbl -> get_Columns(This,Columns) ) 

#define ListView_Clear(This)	\
    ( (This)->lpVtbl -> Clear(This) ) 

#define ListView_NewItem(This,ListViewItem)	\
    ( (This)->lpVtbl -> NewItem(This,ListViewItem) ) 

#define ListView_NewColumn(This,ListViewColumn)	\
    ( (This)->lpVtbl -> NewColumn(This,ListViewColumn) ) 

#define ListView_put_id(This,id)	\
    ( (This)->lpVtbl -> put_id(This,id) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __ListView_INTERFACE_DEFINED__ */


#ifndef __UpDownEventInfo_INTERFACE_DEFINED__
#define __UpDownEventInfo_INTERFACE_DEFINED__

/* interface UpDownEventInfo */
/* [object][oleautomation][dual][helpcontext][version][uuid] */ 


EXTERN_C const IID IID_UpDownEventInfo;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("1FFA0BD0-F1B0-436D-98EB-1ACB5FFBE1FF")
    UpDownEventInfo : public IDispatch
    {
    public:
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Application( 
            /* [retval][out] */ _Application **app) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Parent( 
            /* [retval][out] */ IDispatch **Parent) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_CurrentPosition( 
            /* [retval][out] */ long *CurrentPosition) = 0;
        
        virtual /* [helpcontext][helpstring][propget][id] */ HRESULT STDMETHODCALLTYPE get_Delta( 
            /* [retval][out] */ long *Delta) = 0;
        
    };
    
    
#else 	/* C style interface */

    typedef struct UpDownEventInfoVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            UpDownEventInfo * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            UpDownEventInfo * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            UpDownEventInfo * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            UpDownEventInfo * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            UpDownEventInfo * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            UpDownEventInfo * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            UpDownEventInfo * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Application )( 
            UpDownEventInfo * This,
            /* [retval][out] */ _Application **app);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Parent )( 
            UpDownEventInfo * This,
            /* [retval][out] */ IDispatch **Parent);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_CurrentPosition )( 
            UpDownEventInfo * This,
            /* [retval][out] */ long *CurrentPosition);
        
        /* [helpcontext][helpstring][propget][id] */ HRESULT ( STDMETHODCALLTYPE *get_Delta )( 
            UpDownEventInfo * This,
            /* [retval][out] */ long *Delta);
        
        END_INTERFACE
    } UpDownEventInfoVtbl;

    interface UpDownEventInfo
    {
        CONST_VTBL struct UpDownEventInfoVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define UpDownEventInfo_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define UpDownEventInfo_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define UpDownEventInfo_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define UpDownEventInfo_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define UpDownEventInfo_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define UpDownEventInfo_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define UpDownEventInfo_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 


#define UpDownEventInfo_get_Application(This,app)	\
    ( (This)->lpVtbl -> get_Application(This,app) ) 

#define UpDownEventInfo_get_Parent(This,Parent)	\
    ( (This)->lpVtbl -> get_Parent(This,Parent) ) 

#define UpDownEventInfo_get_CurrentPosition(This,CurrentPosition)	\
    ( (This)->lpVtbl -> get_CurrentPosition(This,CurrentPosition) ) 

#define UpDownEventInfo_get_Delta(This,Delta)	\
    ( (This)->lpVtbl -> get_Delta(This,Delta) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* __UpDownEventInfo_INTERFACE_DEFINED__ */


#ifndef ___SynthesizersEvents_INTERFACE_DEFINED__
#define ___SynthesizersEvents_INTERFACE_DEFINED__

/* interface _SynthesizersEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID IID__SynthesizersEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("529A02D5-8E8C-4C07-8DD2-B7A916390AA5")
    _SynthesizersEvents
    {
    public:
        BEGIN_INTERFACE
        END_INTERFACE
    };
    
    
#else 	/* C style interface */

    typedef struct _SynthesizersEventsVtbl
    {
        BEGIN_INTERFACE
        
        END_INTERFACE
    } _SynthesizersEventsVtbl;

    interface _SynthesizersEvents
    {
        CONST_VTBL struct _SynthesizersEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___SynthesizersEvents_INTERFACE_DEFINED__ */


#ifndef __SynthesizersEvents_DISPINTERFACE_DEFINED__
#define __SynthesizersEvents_DISPINTERFACE_DEFINED__

/* dispinterface SynthesizersEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID DIID_SynthesizersEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)

    MIDL_INTERFACE("93280A3E-5C14-4B12-8B79-75CB454000DD")
    SynthesizersEvents : public IDispatch
    {
    };
    
#else 	/* C style interface */

    typedef struct SynthesizersEventsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            SynthesizersEvents * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            SynthesizersEvents * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            SynthesizersEvents * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            SynthesizersEvents * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            SynthesizersEvents * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            SynthesizersEvents * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            SynthesizersEvents * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        END_INTERFACE
    } SynthesizersEventsVtbl;

    interface SynthesizersEvents
    {
        CONST_VTBL struct SynthesizersEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define SynthesizersEvents_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define SynthesizersEvents_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define SynthesizersEvents_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define SynthesizersEvents_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define SynthesizersEvents_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define SynthesizersEvents_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define SynthesizersEvents_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */


#endif 	/* __SynthesizersEvents_DISPINTERFACE_DEFINED__ */


#ifndef ____MSAAEvents_INTERFACE_DEFINED__
#define ____MSAAEvents_INTERFACE_DEFINED__

/* interface __MSAAEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID IID___MSAAEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("8D8B648D-6C23-48A9-A544-6BCD010FA13D")
    __MSAAEvents
    {
    public:
        BEGIN_INTERFACE
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemMenuStart( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemMenuEnd( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemMenuPopupStart( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemMenuPopupEnd( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemCaptureStart( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemCaptureEnd( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemMoveSizeStart( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemMoveSizeEnd( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemContextHelpStart( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemContextHelpEnd( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemDragDropStart( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemDragDropEnd( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemDialogStart( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemDialogEnd( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemScrollingStart( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemScrollingEnd( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemSwitchStart( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemSwitchEnd( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemMinimizeStart( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnSystemMinimizeEnd( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnConsoleCaret( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnConsoleUpdateRegion( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnConsoleUpdateSimple( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnConsoleUpdateScroll( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnConsoleLayout( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnConsoleStartApplication( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnConsoleEndApplication( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectShow( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectHide( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectReorder( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectFocus( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectSelection( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectSelectionAdd( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectSelectionRemove( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectSelectionWithin( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectStateChange( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectLocationChange( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectNameChange( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectDescriptionChange( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectValueChange( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectParentChange( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectHelpChange( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectDefActionChange( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectAcceleratorChange( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectTextSelectionChanged( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectContentScrolled( 
            /* [in] */ Accessible *Accessible) = 0;
        
        virtual /* [helpcontext] */ HRESULT __stdcall OnObjectInvoked( 
            /* [in] */ Accessible *Accessible) = 0;
        
        END_INTERFACE
    };
    
    
#else 	/* C style interface */

    typedef struct __MSAAEventsVtbl
    {
        BEGIN_INTERFACE
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemMenuStart )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemMenuEnd )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemMenuPopupStart )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemMenuPopupEnd )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemCaptureStart )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemCaptureEnd )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemMoveSizeStart )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemMoveSizeEnd )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemContextHelpStart )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemContextHelpEnd )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemDragDropStart )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemDragDropEnd )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemDialogStart )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemDialogEnd )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemScrollingStart )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemScrollingEnd )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemSwitchStart )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemSwitchEnd )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemMinimizeStart )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnSystemMinimizeEnd )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnConsoleCaret )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnConsoleUpdateRegion )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnConsoleUpdateSimple )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnConsoleUpdateScroll )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnConsoleLayout )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnConsoleStartApplication )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnConsoleEndApplication )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectShow )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectHide )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectReorder )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectFocus )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectSelection )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectSelectionAdd )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectSelectionRemove )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectSelectionWithin )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectStateChange )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectLocationChange )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectNameChange )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectDescriptionChange )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectValueChange )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectParentChange )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectHelpChange )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectDefActionChange )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectAcceleratorChange )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectTextSelectionChanged )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectContentScrolled )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        /* [helpcontext] */ HRESULT ( __stdcall *OnObjectInvoked )( 
            __MSAAEvents * This,
            /* [in] */ Accessible *Accessible);
        
        END_INTERFACE
    } __MSAAEventsVtbl;

    interface __MSAAEvents
    {
        CONST_VTBL struct __MSAAEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define __MSAAEvents_OnSystemMenuStart(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemMenuStart(This,Accessible) ) 

#define __MSAAEvents_OnSystemMenuEnd(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemMenuEnd(This,Accessible) ) 

#define __MSAAEvents_OnSystemMenuPopupStart(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemMenuPopupStart(This,Accessible) ) 

#define __MSAAEvents_OnSystemMenuPopupEnd(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemMenuPopupEnd(This,Accessible) ) 

#define __MSAAEvents_OnSystemCaptureStart(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemCaptureStart(This,Accessible) ) 

#define __MSAAEvents_OnSystemCaptureEnd(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemCaptureEnd(This,Accessible) ) 

#define __MSAAEvents_OnSystemMoveSizeStart(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemMoveSizeStart(This,Accessible) ) 

#define __MSAAEvents_OnSystemMoveSizeEnd(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemMoveSizeEnd(This,Accessible) ) 

#define __MSAAEvents_OnSystemContextHelpStart(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemContextHelpStart(This,Accessible) ) 

#define __MSAAEvents_OnSystemContextHelpEnd(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemContextHelpEnd(This,Accessible) ) 

#define __MSAAEvents_OnSystemDragDropStart(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemDragDropStart(This,Accessible) ) 

#define __MSAAEvents_OnSystemDragDropEnd(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemDragDropEnd(This,Accessible) ) 

#define __MSAAEvents_OnSystemDialogStart(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemDialogStart(This,Accessible) ) 

#define __MSAAEvents_OnSystemDialogEnd(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemDialogEnd(This,Accessible) ) 

#define __MSAAEvents_OnSystemScrollingStart(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemScrollingStart(This,Accessible) ) 

#define __MSAAEvents_OnSystemScrollingEnd(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemScrollingEnd(This,Accessible) ) 

#define __MSAAEvents_OnSystemSwitchStart(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemSwitchStart(This,Accessible) ) 

#define __MSAAEvents_OnSystemSwitchEnd(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemSwitchEnd(This,Accessible) ) 

#define __MSAAEvents_OnSystemMinimizeStart(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemMinimizeStart(This,Accessible) ) 

#define __MSAAEvents_OnSystemMinimizeEnd(This,Accessible)	\
    ( (This)->lpVtbl -> OnSystemMinimizeEnd(This,Accessible) ) 

#define __MSAAEvents_OnConsoleCaret(This,Accessible)	\
    ( (This)->lpVtbl -> OnConsoleCaret(This,Accessible) ) 

#define __MSAAEvents_OnConsoleUpdateRegion(This,Accessible)	\
    ( (This)->lpVtbl -> OnConsoleUpdateRegion(This,Accessible) ) 

#define __MSAAEvents_OnConsoleUpdateSimple(This,Accessible)	\
    ( (This)->lpVtbl -> OnConsoleUpdateSimple(This,Accessible) ) 

#define __MSAAEvents_OnConsoleUpdateScroll(This,Accessible)	\
    ( (This)->lpVtbl -> OnConsoleUpdateScroll(This,Accessible) ) 

#define __MSAAEvents_OnConsoleLayout(This,Accessible)	\
    ( (This)->lpVtbl -> OnConsoleLayout(This,Accessible) ) 

#define __MSAAEvents_OnConsoleStartApplication(This,Accessible)	\
    ( (This)->lpVtbl -> OnConsoleStartApplication(This,Accessible) ) 

#define __MSAAEvents_OnConsoleEndApplication(This,Accessible)	\
    ( (This)->lpVtbl -> OnConsoleEndApplication(This,Accessible) ) 

#define __MSAAEvents_OnObjectShow(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectShow(This,Accessible) ) 

#define __MSAAEvents_OnObjectHide(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectHide(This,Accessible) ) 

#define __MSAAEvents_OnObjectReorder(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectReorder(This,Accessible) ) 

#define __MSAAEvents_OnObjectFocus(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectFocus(This,Accessible) ) 

#define __MSAAEvents_OnObjectSelection(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectSelection(This,Accessible) ) 

#define __MSAAEvents_OnObjectSelectionAdd(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectSelectionAdd(This,Accessible) ) 

#define __MSAAEvents_OnObjectSelectionRemove(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectSelectionRemove(This,Accessible) ) 

#define __MSAAEvents_OnObjectSelectionWithin(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectSelectionWithin(This,Accessible) ) 

#define __MSAAEvents_OnObjectStateChange(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectStateChange(This,Accessible) ) 

#define __MSAAEvents_OnObjectLocationChange(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectLocationChange(This,Accessible) ) 

#define __MSAAEvents_OnObjectNameChange(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectNameChange(This,Accessible) ) 

#define __MSAAEvents_OnObjectDescriptionChange(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectDescriptionChange(This,Accessible) ) 

#define __MSAAEvents_OnObjectValueChange(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectValueChange(This,Accessible) ) 

#define __MSAAEvents_OnObjectParentChange(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectParentChange(This,Accessible) ) 

#define __MSAAEvents_OnObjectHelpChange(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectHelpChange(This,Accessible) ) 

#define __MSAAEvents_OnObjectDefActionChange(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectDefActionChange(This,Accessible) ) 

#define __MSAAEvents_OnObjectAcceleratorChange(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectAcceleratorChange(This,Accessible) ) 

#define __MSAAEvents_OnObjectTextSelectionChanged(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectTextSelectionChanged(This,Accessible) ) 

#define __MSAAEvents_OnObjectContentScrolled(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectContentScrolled(This,Accessible) ) 

#define __MSAAEvents_OnObjectInvoked(This,Accessible)	\
    ( (This)->lpVtbl -> OnObjectInvoked(This,Accessible) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ____MSAAEvents_INTERFACE_DEFINED__ */


#ifndef __MSAAEvents_DISPINTERFACE_DEFINED__
#define __MSAAEvents_DISPINTERFACE_DEFINED__

/* dispinterface MSAAEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID DIID_MSAAEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)

    MIDL_INTERFACE("324CC008-E06C-477B-96B7-071C50C6FECE")
    MSAAEvents : public IDispatch
    {
    };
    
#else 	/* C style interface */

    typedef struct MSAAEventsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            MSAAEvents * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            MSAAEvents * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            MSAAEvents * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            MSAAEvents * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            MSAAEvents * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            MSAAEvents * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            MSAAEvents * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        END_INTERFACE
    } MSAAEventsVtbl;

    interface MSAAEvents
    {
        CONST_VTBL struct MSAAEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define MSAAEvents_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define MSAAEvents_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define MSAAEvents_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define MSAAEvents_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define MSAAEvents_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define MSAAEvents_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define MSAAEvents_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */


#endif 	/* __MSAAEvents_DISPINTERFACE_DEFINED__ */


#ifndef ___BrowseModeEvents_INTERFACE_DEFINED__
#define ___BrowseModeEvents_INTERFACE_DEFINED__

/* interface _BrowseModeEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID IID__BrowseModeEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)
    
    MIDL_INTERFACE("14397875-7E87-42A6-B338-0D053FC08CA4")
    _BrowseModeEvents
    {
    public:
        BEGIN_INTERFACE
        END_INTERFACE
    };
    
    
#else 	/* C style interface */

    typedef struct _BrowseModeEventsVtbl
    {
        BEGIN_INTERFACE
        
        END_INTERFACE
    } _BrowseModeEventsVtbl;

    interface _BrowseModeEvents
    {
        CONST_VTBL struct _BrowseModeEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#endif /* COBJMACROS */


#endif 	/* C style interface */




#endif 	/* ___BrowseModeEvents_INTERFACE_DEFINED__ */


#ifndef __BrowseModeEvents_DISPINTERFACE_DEFINED__
#define __BrowseModeEvents_DISPINTERFACE_DEFINED__

/* dispinterface BrowseModeEvents */
/* [helpcontext][version][uuid] */ 


EXTERN_C const IID DIID_BrowseModeEvents;

#if defined(__cplusplus) && !defined(CINTERFACE)

    MIDL_INTERFACE("D12B2FFD-52B7-43D9-A18D-A55745CCCB67")
    BrowseModeEvents : public IDispatch
    {
    };
    
#else 	/* C style interface */

    typedef struct BrowseModeEventsVtbl
    {
        BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )( 
            BrowseModeEvents * This,
            /* [in] */ REFIID riid,
            /* [annotation][iid_is][out] */ 
            _COM_Outptr_  void **ppvObject);
        
        ULONG ( STDMETHODCALLTYPE *AddRef )( 
            BrowseModeEvents * This);
        
        ULONG ( STDMETHODCALLTYPE *Release )( 
            BrowseModeEvents * This);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfoCount )( 
            BrowseModeEvents * This,
            /* [out] */ UINT *pctinfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetTypeInfo )( 
            BrowseModeEvents * This,
            /* [in] */ UINT iTInfo,
            /* [in] */ LCID lcid,
            /* [out] */ ITypeInfo **ppTInfo);
        
        HRESULT ( STDMETHODCALLTYPE *GetIDsOfNames )( 
            BrowseModeEvents * This,
            /* [in] */ REFIID riid,
            /* [size_is][in] */ LPOLESTR *rgszNames,
            /* [range][in] */ UINT cNames,
            /* [in] */ LCID lcid,
            /* [size_is][out] */ DISPID *rgDispId);
        
        /* [local] */ HRESULT ( STDMETHODCALLTYPE *Invoke )( 
            BrowseModeEvents * This,
            /* [annotation][in] */ 
            _In_  DISPID dispIdMember,
            /* [annotation][in] */ 
            _In_  REFIID riid,
            /* [annotation][in] */ 
            _In_  LCID lcid,
            /* [annotation][in] */ 
            _In_  WORD wFlags,
            /* [annotation][out][in] */ 
            _In_  DISPPARAMS *pDispParams,
            /* [annotation][out] */ 
            _Out_opt_  VARIANT *pVarResult,
            /* [annotation][out] */ 
            _Out_opt_  EXCEPINFO *pExcepInfo,
            /* [annotation][out] */ 
            _Out_opt_  UINT *puArgErr);
        
        END_INTERFACE
    } BrowseModeEventsVtbl;

    interface BrowseModeEvents
    {
        CONST_VTBL struct BrowseModeEventsVtbl *lpVtbl;
    };

    

#ifdef COBJMACROS


#define BrowseModeEvents_QueryInterface(This,riid,ppvObject)	\
    ( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) ) 

#define BrowseModeEvents_AddRef(This)	\
    ( (This)->lpVtbl -> AddRef(This) ) 

#define BrowseModeEvents_Release(This)	\
    ( (This)->lpVtbl -> Release(This) ) 


#define BrowseModeEvents_GetTypeInfoCount(This,pctinfo)	\
    ( (This)->lpVtbl -> GetTypeInfoCount(This,pctinfo) ) 

#define BrowseModeEvents_GetTypeInfo(This,iTInfo,lcid,ppTInfo)	\
    ( (This)->lpVtbl -> GetTypeInfo(This,iTInfo,lcid,ppTInfo) ) 

#define BrowseModeEvents_GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId)	\
    ( (This)->lpVtbl -> GetIDsOfNames(This,riid,rgszNames,cNames,lcid,rgDispId) ) 

#define BrowseModeEvents_Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr)	\
    ( (This)->lpVtbl -> Invoke(This,dispIdMember,riid,lcid,wFlags,pDispParams,pVarResult,pExcepInfo,puArgErr) ) 

#endif /* COBJMACROS */


#endif 	/* C style interface */


#endif 	/* __BrowseModeEvents_DISPINTERFACE_DEFINED__ */


EXTERN_C const CLSID CLSID_BrailleDisplays;

#ifdef __cplusplus

class DECLSPEC_UUID("589C86A7-883E-43B0-960E-D55EE8E3DA84")
BrailleDisplays;
#endif

EXTERN_C const CLSID CLSID_Application;

#ifdef __cplusplus

class DECLSPEC_UUID("76D4F4E6-3E5D-43E8-B9D9-FC58684A6B8C")
Application;
#endif

EXTERN_C const CLSID CLSID_Keyboard;

#ifdef __cplusplus

class DECLSPEC_UUID("43C8DC45-4F69-4898-8D9D-783DE615AFD7")
Keyboard;
#endif

EXTERN_C const CLSID CLSID_Key;

#ifdef __cplusplus

class DECLSPEC_UUID("B7D5D28F-A998-4EFB-B2E9-8BDBA3086F36")
Key;
#endif

EXTERN_C const CLSID CLSID_Script;

#ifdef __cplusplus

class DECLSPEC_UUID("86352B61-B01A-4A1B-BB40-DC92121F6781")
Script;
#endif

EXTERN_C const CLSID CLSID_Window;

#ifdef __cplusplus

class DECLSPEC_UUID("37018039-132B-43EF-B079-B6A13D2C02D1")
Window;
#endif

EXTERN_C const CLSID CLSID_ClientInformation;

#ifdef __cplusplus

class DECLSPEC_UUID("BDA6DF44-CFD0-4DEC-947F-364D83C4B079")
ClientInformation;
#endif

EXTERN_C const CLSID CLSID_Mouse;

#ifdef __cplusplus

class DECLSPEC_UUID("E9C3515D-58AB-4B80-9D6B-A47E2578EED7")
Mouse;
#endif

EXTERN_C const CLSID CLSID_Dialog;

#ifdef __cplusplus

class DECLSPEC_UUID("B2ADCAFA-C744-43CF-8591-4AAB3E6FEDED")
Dialog;
#endif

EXTERN_C const CLSID CLSID_MousePointer;

#ifdef __cplusplus

class DECLSPEC_UUID("45B8228B-2DCD-4F35-9AE8-13F6272941E4")
MousePointer;
#endif

EXTERN_C const CLSID CLSID_Control;

#ifdef __cplusplus

class DECLSPEC_UUID("5F118E43-C234-4879-A655-2A8F0041B463")
Control;
#endif

EXTERN_C const CLSID CLSID_EditBox;

#ifdef __cplusplus

class DECLSPEC_UUID("9ABF4CC0-A154-4248-BD84-B53D2AAD51E7")
EditBox;
#endif

EXTERN_C const CLSID CLSID_Pushbutton;

#ifdef __cplusplus

class DECLSPEC_UUID("E9B4A950-BDAD-443A-BA63-637075835D76")
Pushbutton;
#endif

EXTERN_C const CLSID CLSID_CheckBox;

#ifdef __cplusplus

class DECLSPEC_UUID("61C99FFB-1A53-4106-BAB4-04D888ACAD1E")
CheckBox;
#endif

EXTERN_C const CLSID CLSID_RadioButton;

#ifdef __cplusplus

class DECLSPEC_UUID("648C7A43-CB4B-47B5-9DBD-C64A3E60B966")
RadioButton;
#endif

EXTERN_C const CLSID CLSID_ComboBox;

#ifdef __cplusplus

class DECLSPEC_UUID("A9CC725C-1A13-4780-BC7E-AEF3898D18CA")
ComboBox;
#endif

EXTERN_C const CLSID CLSID_ListBox;

#ifdef __cplusplus

class DECLSPEC_UUID("ABD450AB-9039-4BA6-8396-E97B2C67B016")
ListBox;
#endif

EXTERN_C const CLSID CLSID_UpDown;

#ifdef __cplusplus

class DECLSPEC_UUID("CADB2DDD-9D9E-4EAE-AC1C-C55E4A000337")
UpDown;
#endif

EXTERN_C const CLSID CLSID_TrackBar;

#ifdef __cplusplus

class DECLSPEC_UUID("876B6B1A-C823-4BBF-8B5B-D57D08800773")
TrackBar;
#endif

EXTERN_C const CLSID CLSID_Synthesizer;

#ifdef __cplusplus

class DECLSPEC_UUID("D353FC81-02D3-4967-B11E-A541A06AD1B5")
Synthesizer;
#endif

EXTERN_C const CLSID CLSID_Speech;

#ifdef __cplusplus

class DECLSPEC_UUID("215AA611-9C86-48B3-89E5-EC1452ADAE8B")
Speech;
#endif

EXTERN_C const CLSID CLSID_MSAAEventSource;

#ifdef __cplusplus

class DECLSPEC_UUID("A330FEFD-F752-4A37-BAE4-EECE1D998EA3")
MSAAEventSource;
#endif

EXTERN_C const CLSID CLSID_BrowseMode;

#ifdef __cplusplus

class DECLSPEC_UUID("EAACAB81-2159-4768-8208-42210CC6D484")
BrowseMode;
#endif

EXTERN_C const CLSID CLSID_SharedObjects;

#ifdef __cplusplus

class DECLSPEC_UUID("CDC696E2-2B07-434C-8F11-87924E56983D")
SharedObjects;
#endif

EXTERN_C const CLSID CLSID_brailleDisplay;

#ifdef __cplusplus

class DECLSPEC_UUID("34FE5E9B-27E3-4CB9-8C5D-5E8DD1FE771D")
brailleDisplay;
#endif

EXTERN_C const CLSID CLSID_Braille;

#ifdef __cplusplus

class DECLSPEC_UUID("4DB82E86-92F3-4CD1-BEDF-17D963D325AB")
Braille;
#endif

EXTERN_C const CLSID CLSID_Synthesizers;

#ifdef __cplusplus

class DECLSPEC_UUID("88EB8543-46B6-4642-B017-63C7DAAC13D0")
Synthesizers;
#endif
#endif /* __WindowEyes_LIBRARY_DEFINED__ */

/* Additional Prototypes for ALL interfaces */

/* end of Additional Prototypes */

#ifdef __cplusplus
}
#endif

#endif


